//*** uart.c ***/
// Aus www.mikrocontroller.net

#include <avr/io.h>
#include "usart.h"

void usart_init(void)
{
  UBRR0H = (unsigned char)(UBRR_VAL>>8);
  UBRR0L = (unsigned char)(UBRR_VAL);
  UCSR0B = (1<<TXEN0) | (1<<RXEN0);      // tx & rx enable
  UCSR0C = (1<<UCSZ01) | (1<<UCSZ00);    // Asynchron 8N1
}

void usart_write_char(unsigned char c)
{
  while (!(UCSR0A & (1<<UDRE0)))
  {
    /* warten bis Senden moeglich */
  }
  UDR0 = c;
}


void usart_write_str(char *str)
{
	while (*str)
	{
		usart_write_char(*str++);
	}
}

void usart_write_long(char *str, long n)
{
	usart_write_str(str);
	if(n<0)
		{ usart_write_char('-'); usart_write_ulong(0, -n); }
	else
		usart_write_ulong(0, n);
}

void usart_write_ulong(char *str, unsigned long n)
{
	if(0!=str)
		usart_write_str(str);
	char buffer[15];
	ultoa(n, buffer, 10);
	usart_write_str(buffer);
}

void usart_write_hex(char *str, unsigned long n)
{
	if(0!=str)
		usart_write_str(str);
	char buffer[9];
	for(uint8_t i=0; i<8; i++)
	{
		if((buffer[i]=(n>>(28-4*i))&15)<10)
			buffer[i]+='0';
		else
			buffer[i]+='a'-10;
	}
	buffer[8]=0;
	usart_write_str(buffer);
}

void usart_write_int(char *str, int n)
{
	usart_write_long(str, (long)n);
}

void usart_write_uint(char *str, unsigned int n)
{
	usart_write_ulong(str, (unsigned long)n);
}
