#include "BytecodeTargetedCompiler.h"

using BytecodeInterpreter::Value;
using BytecodeInterpreter::uint8;
using BytecodeInterpreter::uint16;
using BytecodeInterpreter::uint32;
using BytecodeInterpreter::uint64;
using BytecodeInterpreter::int8;
using BytecodeInterpreter::int16;
using BytecodeInterpreter::int32;
using BytecodeInterpreter::int64;


uint32 nibble_codesize=0;
uint8 *CodeOfProgram=0;

uint8 get_code_nibble(PROGRAM_COUNTER_TYPE pc)
{
#ifdef USE_ASSERT
	if(pc>=nibble_codesize)
		throw BytecodeInterpreter::RuntimeError("get_code_nibble Error", pc);
#endif
	return (pc&1) ? ((uint8)(CodeOfProgram[pc>>1]&15)) : ((uint8)((CodeOfProgram[pc>>1]>>4)&15));
}
uint8 get_code_byte(PROGRAM_COUNTER_TYPE pc)
{
#ifdef USE_ASSERT
	if(pc>=nibble_codesize)
		throw BytecodeInterpreter::RuntimeError("get_code_nibble Error", pc);
#endif
	if(0==(pc&1))
		return CodeOfProgram[pc>>1];
	else
		return (((uint8)(CodeOfProgram[pc>>1]&15))<<4) + ((uint8)((CodeOfProgram[(pc+1)>>1]>>4)&15));
}

namespace extern_function_names { enum
{	// C/C++ - Funktionen, die vom Programm, welches auf dem Bytecode-Interpreter luft, aufgerufen werden knnen
	// ("externe Funktionen").
	prints
}; };

void extern_function_invoker(VARIABLE_INDEX_TYPE function_number, BytecodeInterpreter::BytecodeInterpreter& ip)
{	// Diese Funktion wird vom Bytecode-Interpreter aufgerufen, wenn das Programm eine externe Funktion aufruft.
	namespace nn=extern_function_names;
	switch(function_number)
	{
		case nn::prints:
		{
			Value v;
			ip.pop_stack(&v);
			char *str=(char*)ip.var_addr_to_mem_addr(v.u16);
			printf(str);
			break;
		}
	}
}

#include "language_defines.h"

BEGIN_GLOBAL_SCOPE(TestFloat64Bibliothek)
// ********************* HIER BEGINNT DAS EIGENTLICHE PROGRAMM *************************
	declare_void_function(Main, params());
	declare_extern_void_function(prints, params(pointer(uint8)));
	declare_function(pointer(uint8), copystr, params(pointer(uint8)));

	void_function(Main)
		while(1)
			prints(copystr("Hallo, Welt !\r\n"));
				// Weil der String "Hallo, Welt !\n" als Konstante im Bytecode abgelegt wird, muss er erst in
				// ein globales Array kopiert werden (copystr()), damit die externe Funktion prints() darauf
				// zugreifen kann. Das globale Array ist in dem Speicher lokalisiert, dessen Startadresse dem Konstruktor
				// BytecodeInterpreter::BytecodeInterpreter bergeben wurde (Parameter __memory).
		end
	end

	function(copystr)
		param s(pointer(uint8)); // Deklaration des Funktionsarguments
		var mem(array(uint8, 128)); // !!! Speicher im SRAM (entspricht einem statischen Array in C/C++)
		local p(pointer(uint8), &mem[0]);
		while(0!=(*(p++)=*(s++))) end
		return(&mem[0]); // Im Gegensatz zu ANSI C darf man hier nicht return(mem) schreiben.
	end
// ********************* HIER ENDET DAS EIGENTLICHE PROGRAMM ***************************
END_GLOBAL_SCOPE

#include "language_undefines.h"

int main()
{
	printf("Starte Kompilierung ...\n");
	BytecodeTargetedCompiler::Compiler compiler(TestFloat64Bibliothek);
	nibble_codesize=compiler.nibble_codesize;
	CodeOfProgram=compiler.bytecode;
	printf("\nKompilierung fertig.\n\n");

	printf("code size = %u bytes = 0x%x bytes = %u nibbles = 0x%x nibbles\r\n\r\n", (compiler.nibble_codesize+1)/2,
			(compiler.nibble_codesize+1)/2, compiler.nibble_codesize, compiler.nibble_codesize);
	printf("code_begin=%u   =0x%x\r\n\r\n", compiler.function_begin_addresses["Main"]);

	FILE *file=fopen("d:\\program.txt", "wb");

	if(file)
	{
		fprintf(file, "code size = %u bytes = %x bytes = %u nibbles = %x nibbles\r\n\r\n", (compiler.nibble_codesize+1)/2,
			(compiler.nibble_codesize+1)/2, compiler.nibble_codesize, compiler.nibble_codesize);
		fprintf(file, "code_begin=%u   =0x%x\r\n\r\n", compiler.function_begin_addresses["Main"]);
		for(uint32 i=0; i<(compiler.nibble_codesize+1)/2; i++)
		{
			fprintf(file, "%3u, ", CodeOfProgram[i]);
			if(19==(i%20)) fprintf(file, "\r\n");
		}
		fflush(file);
	}

	const uint16 stack_size=65535, memory_size=65535;
	static uint8 stack[stack_size], memory[memory_size];

	BytecodeInterpreter::BytecodeInterpreter ip(stack, stack_size, memory, memory_size, get_code_nibble, get_code_byte, &extern_function_invoker);
	PROGRAM_COUNTER_TYPE code_begin=compiler.function_begin_addresses["Main"];

	try
	{
		ip.execute_bytecode(code_begin);
	}
	catch(BytecodeInterpreter::RuntimeError err)
	{
		err;
		printf("\n\n***** Laufzeit FEHLER *****\n\n");
		printf("===> %s\n", err.message);
		printf("===> PC = 0x%x\n\n", err.pc);
		while(_getch()==0);
		return 0;
	}
#ifdef CALCULATE_MAXIMUM_STACK_USAGE
	printf("Maximaler benutzter Stackspeicher: %u Bytes\n", ip.get_maximum_stack_usage());
	fprintf(file, "\r\nMaximaler benutzter Stackspeicher: %u Bytes\r\n", ip.get_maximum_stack_usage());
#endif
	if(file) fclose(file);
	while(_getch()==0);
	return 0;
}
