// utils.h

//***********************************************************************************************************//
//***********************************************************************************************************//
//                                                                                                           //
// BytecodeInterpreter.h, BytecodeTargetedCompiler.h, language_defines.h, language_undefines.h, utils.h:     //
// These files contain a bytecode interpreter and a compiler converting a C-like program into bytecode.      //
//                                                                                                           //
// Copyright 2009 by Florian Knigstein, mail@virgusta.eu                                                    //
//                                                                                                           //
//      Commercial use of this code without my permission is prohibited.                                     //
//      You may use this program free of charge for non-commercial purposes provided you agree to the        //
//      following License:                                                                                   //
//                                                                                                           //
//      The Original Work is provided under this License on an "AS IS" BASIS and WITHOUT WARRANTY,           //
//      either express or implied, including, without limitation, the warranties of                          //
//      non-infringement, merchantability or fitness for a particular purpose. THE ENTIRE RISK AS TO THE     //
//      QUALITY OF THE ORIGINAL WORK IS WITH YOU. This DISCLAIMER OF WARRANTY constitutes an essential part  //
//      of this License. No license to the Original Work is granted by this License except under this        //
//      disclaimer.                                                                                          //
//                                                                                                           //
//      Limitation of Liability. Under no circumstances and under no legal theory, whether in tort           //
//      (including negligence), contract, or otherwise, shall the Licensor be liable to anyone for any       //
//      indirect, special, incidental, or consequential damages of any character arising as a result of      //
//      this License or the use of the Original Work including, without limitation, damages for loss of      //
//      goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages    //
//      or losses. This limitation of liability shall not apply to the extent applicable law prohibits such  //
//      limitation.                                                                                          //
//                                                                                                           //
//***********************************************************************************************************//
//***********************************************************************************************************//

// ***** ATTENTION *****
// This code is not well-tested. There may be several bugs.


// Wenn nlist.h in mehr als einer Quelldatei eingebunden wird, muss in allen auer in einer der Quelldateien vor der
// #include "nlist.h"-Direktive das #define
//    #define NLIST_H_DONT_DEFINE_CODE_WHICH_MUST_NOT_BE_DEFINED_IN_MORE_THAN_ONE_SOURCE_FILE
// erfolgen.


#ifndef NLIST_INCLUDED
#define NLIST_INCLUDED

#include <assert.h>
#include <list>
//#include "xstringnounion.h"

#if defined(__GNUC__)
#include <c++\3.4.2\bits\stl_function.h>
#include <c++\3.4.2\string>
#endif

template<typename _Tp>
class mem_allocator
{
public:
	typedef size_t     size_type;
	typedef ptrdiff_t  difference_type;
	typedef _Tp*       pointer;
	typedef const _Tp* const_pointer;
	typedef _Tp&       reference;
	typedef const _Tp& const_reference;
	typedef _Tp        value_type;

	template<typename _Tp1>
	struct rebind
	{
		typedef mem_allocator<_Tp1> other;
	};

	mem_allocator() throw() { }

	mem_allocator(const mem_allocator&) throw() { }

	template<typename _Tp1>
	mem_allocator(const mem_allocator<_Tp1>&) throw() { }

	~mem_allocator() throw() { }

	pointer address(reference __x) const { return &__x; }

	const_pointer address(const_reference __x) const { return &__x; }

	pointer allocate(size_type _N, const void *hint=0)
	{
		pointer mem = static_cast<_Tp*>(malloc(sizeof(_Tp)*_N));
		if(!mem) throw std::bad_alloc();
		return mem;
	}

	void deallocate(pointer _P, size_type)
		{ free(static_cast<void*>(_P)); }

	size_type max_size() const throw() 
		{ return size_t(-1) / sizeof(_Tp); }

	void  construct(pointer __p, const _Tp& __val) 
		{ ::new(__p) value_type(__val); }

	void destroy(pointer __p) { __p->~_Tp(); }
};

template<typename _Tp>
inline bool
operator==(const mem_allocator<_Tp>&, const mem_allocator<_Tp>&)
{ return true; }

template<typename _Tp>
inline bool
operator!=(const mem_allocator<_Tp>&, const mem_allocator<_Tp>&)
{ return false; }

typedef std::basic_string<char, std::char_traits<char>, mem_allocator<char> > nstring;

template<class T> struct BaumKeineAttrBerechnen;

template<class T> struct RotSchwarzBaum_Node;

template<class T>
struct RotSchwarzBaum_Node_Base
{
	RotSchwarzBaum_Node<T> *p, *links, *rechts;  // p: Eltern-Knoten
	bool rot;
};

template<class T>
struct RotSchwarzBaum_Node : public RotSchwarzBaum_Node_Base<T>
{
public:
	T t;
	RotSchwarzBaum_Node(const T& _t) : t(_t) { }
	RotSchwarzBaum_Node() { }
};

template<class T>
struct BaumKeineAttrBerechnen
{
	enum { soll_attribut_berechnen=0 };
	inline void calculate_attribute(T& akt) { }
	inline void calculate_attribute(const T& links, const T& rechts, T& akt) { }
	inline void calculate_attribute1(const T& links, T& akt) { }
	inline void calculate_attribute2(const T& rechts, T& akt) { }
};

template<class T>
struct TLessEqual : public BaumKeineAttrBerechnen<T>
{
	inline bool less(const T& x, const T& y) const { return x.operator<(y); }
	inline bool equal(const T& x, const T& y) const { return x.operator==(y); }
};

template<class T>
struct LessEqual : public BaumKeineAttrBerechnen<T>
{
	//inline bool less(const T& x, const T& y) const { return ::operator<(x,y); }
	//inline bool equal(const T& x, const T& y) const { return ::operator==(x,y); }
	inline bool less(const T& x, const T& y) const { return x<y; }
	inline bool equal(const T& x, const T& y) const { return x<y; }
};

template<>
struct LessEqual<unsigned long> : public BaumKeineAttrBerechnen<unsigned long>
{
	bool less(const unsigned long& x, const unsigned long& y) const { return x<y; }
	bool equal(const unsigned long& x, const unsigned long& y) const { return x==y; }
};

template<>
struct LessEqual<long> : public BaumKeineAttrBerechnen<long>
{
	bool less(const long& x, const long& y) const { return x<y; }
	bool equal(const long& x, const long& y) const { return x==y; }
};

template<>
struct LessEqual<unsigned int> : public BaumKeineAttrBerechnen<unsigned int>
{
	bool less(const unsigned int& x, const unsigned int& y) const { return x<y; }
	bool equal(const unsigned int& x, const unsigned int& y) const { return x==y; }
};

template<>
struct LessEqual<int> : public BaumKeineAttrBerechnen<int>
{
	bool less(const int& x, const int& y) const { return x<y; }
	bool equal(const int& x, const int& y) const { return x==y; }
};

template<>
struct LessEqual<unsigned short> : public BaumKeineAttrBerechnen<unsigned short>
{
	bool less(const unsigned short& x, const unsigned short& y) const { return x<y; }
	bool equal(const unsigned short& x, const unsigned short& y) const { return x==y; }
};

template<>
struct LessEqual<short> : public BaumKeineAttrBerechnen<short>
{
	bool less(const short& x, const short& y) const { return x<y; }
	bool equal(const short& x, const short& y) const { return x==y; }
};

template<>
struct LessEqual<unsigned char> : public BaumKeineAttrBerechnen<unsigned char>
{
	bool less(const unsigned char& x, const unsigned char& y) const { return x<y; }
	bool equal(const unsigned char& x, const unsigned char& y) const { return x==y; }
};

template<>
struct LessEqual<char> : public BaumKeineAttrBerechnen<char>
{
	bool less(const char& x, const char& y) const { return x<y; }
	bool equal(const char& x, const char& y) const { return x==y; }
};

template<>
struct LessEqual<nstring> : public BaumKeineAttrBerechnen<nstring>
{
	bool less(const nstring& x, const nstring& y) const { return 0<x.compare(y); }
	bool equal(const nstring& x, const nstring& y) const { return 0==x.compare(y); }
};

template<class T, class __LessEqualOPs=LessEqual<T>, class Allocator=mem_allocator<RotSchwarzBaum_Node<T> > >
class RotSchwarzBaum
{
	__LessEqualOPs LessEqualOPs;
	Allocator allct;

	inline bool LessOperator(const T& t1, const T& t2) const
		{ return LessEqualOPs.less(t1,t2); }
	inline bool EqualOperator(const T& t1, const T& t2) const
		{ return LessEqualOPs.equal(t1,t2); }

public:
	class iterator;
	class const_iterator;
	friend class iterator;
	friend class const_iterator;

	typedef RotSchwarzBaum_Node<T> Node;

private:
	Node *wurzel;
	unsigned long anz_elemente;

	static Node *nil()	// ACHTUNG: ES DARF KEIN ZUGRIFF AUF nil()->t ERFOLGEN (==> ABSTURZ) !!!
			{
				static RotSchwarzBaum_Node_Base<T> n;
				return static_cast<Node *>(&n);
			}
			// Es darf kein static nil-Objekt in RotSchwarzBaum angelegt
			// werden, weil dessen rechtzeitige Initialisierung nicht sichergestellt ist, falls andere Klassen
			// statische Elemente haben, die einen RotSchwarzBaum enthalten. Wenn static Node dagegen innerhalb einer
			// Funktion (hier nil()) definiert wird, wird es genau beim ersten Aufruf dieser Funktion initialisiert.
	static bool& copying_t_allowed() { static bool al=false; return al; }
				// Standardmig wird diese Variable auf false gesetzt.
				// In diesem Fall wird die Funktion erase() den Aufruf des Copy-Konstruktors oder
				// des Zuweisungsoperators von class T vermeiden, was jedoch zustzliche Zeiger-
				// zuweisungen erfordert. Diese Einstellung ist dann sinnvoll, wenn der
				// Copy-Konstruktor oder der Zuweisungsoperastor von class T zeitaufwendig arbeiten.

	inline Node *new_node(const Node *node)
			{
				return new Node(node);
				/*Node *nd=allct.allocate(1, 0);
				if(0!=nd) allct.construct(nd, node);
				return nd;*/
			}
	inline Node *new_node(const T &t)
			{
				return new Node(t);
				/*Node *nd=allct.allocate(1, 0);
				if(0!=nd) allct.construct(nd, Node(t));
				return nd;*/
			}
	inline void delete_node(Node *node)
			{
				delete node;
				//allct.destroy(node);
				//allct.deallocate(node,1);
			}

	Node *copy(const Node *n)
			{
				if(n==0 || n==nil()) return nil();
				Node *m=new_node(n->t);
				m->t=n->t; m->p=nil(); m->rot=n->rot;
				m->links=copy(n->links, m); m->rechts=copy(n->rechts, m);
				return m;
			}
	Node *copy(const Node *n, Node *p)
			{
				if(n==0 || n==nil()) return nil();
				Node *m=new_node(n->t);
				m->t=n->t; m->p=p; m->rot=n->rot;
				m->links=copy(n->links, m); m->rechts=copy(n->rechts, m);
				return m;
			}
	void destruct(Node *n)
			{		// n!=nil() vorausgesetzt !
				if(n->links!=nil()) destruct(n->links);
				if(n->rechts!=nil()) destruct(n->rechts);
				delete_node(n);
			}
	inline static Node *tree_minimizer(Node *n)
			{
				if(nil()==n) return nil();
				while(n->links!=nil()) n=n->links;
				return n;
			}
	inline static Node *tree_maximizer(Node *n)
			{
				if(nil()==n) return nil();
				while(n->rechts!=nil()) n=n->rechts;
				return n;
			}
	inline static Node *successor(Node *n)
			{
				if(nil()==n) return nil();
				if(nil()!=n->rechts)
				{
					// return tree_minimizer(n->rechts);
					n=n->rechts;
					while(n->links!=nil()) n=n->links;
					return n;
				}
				Node *y=n->p;
				while(nil()!=y && n==y->rechts)
					{ n=y; y=y->p; }
				return y;
			}
	inline static Node *predecessor(Node *n)
			{
				if(nil()==n) return nil();
				if(nil()!=n->links)
				{
					// return tree_maximizer(n->links);
					n=n->links;
					while(n->rechts!=nil()) n=n->rechts;
					return n;
				}
				Node *y=n->p;
				while(nil()!=y && n==y->links)
					{ n=y; y=y->p; }
				return y;
			}
	void left_rotate(Node *x)
			{
				if(nil()==x->rechts) return;
				Node *y=x->rechts;
				x->rechts=y->links;
				if(nil()!=y->links) y->links->p=x;
				y->p=x->p;
				if(nil()==x->p)
					wurzel=y;
				else if(x==x->p->links)
					x->p->links=y;
				else
					x->p->rechts=y;
				y->links=x;
				x->p=y;
				if(LessEqualOPs.soll_attribut_berechnen)
				{
					if(nil()==x->links)
					{
						if(nil()==x->rechts)
							LessEqualOPs.calculate_attribute(x->t);
						else
							LessEqualOPs.calculate_attribute2(x->rechts->t, x->t);
					}
					else if(nil()==x->rechts)
						LessEqualOPs.calculate_attribute1(x->links->t, x->t);
					else
						LessEqualOPs.calculate_attribute(x->links->t, x->rechts->t, x->t);

					// y->links!=nil() (y->links=x)
					if(nil()==y->rechts)
						LessEqualOPs.calculate_attribute1(y->links->t, y->t);
					else
						LessEqualOPs.calculate_attribute(y->links->t, y->rechts->t, y->t);

					while(nil() != (y = (y->p)))
					{
						if(nil()==y->links) // dann muss y->rechts!=nil() sein
							LessEqualOPs.calculate_attribute2(y->rechts->t, y->t);
						else if(nil()==y->rechts)
							LessEqualOPs.calculate_attribute1(y->links->t, y->t);
						else
							LessEqualOPs.calculate_attribute(y->links->t, y->rechts->t, y->t);
					}
				}
			}
	void right_rotate(Node *x)
			{
				if(nil()==x->links) return;
				Node *y=x->links;
				x->links=y->rechts;
				if(nil()!=y->rechts) y->rechts->p=x;
				y->p=x->p;
				if(nil()==x->p)
					wurzel=y;
				else if(x==x->p->rechts)
					x->p->rechts=y;
				else
					x->p->links=y;
				y->rechts=x;
				x->p=y;
				if(LessEqualOPs.soll_attribut_berechnen)
				{
					if(nil()==x->links)
					{
						if(nil()==x->rechts)
							LessEqualOPs.calculate_attribute(x->t);
						else
							LessEqualOPs.calculate_attribute2(x->rechts->t, x->t);
					}
					else if(nil()==x->rechts)
						LessEqualOPs.calculate_attribute1(x->links->t, x->t);
					else
						LessEqualOPs.calculate_attribute(x->links->t, x->rechts->t, x->t);

					// y->rechts!=nil() (y->rechts=x)
					if(nil()==y->rechts)
						LessEqualOPs.calculate_attribute1(y->links->t, y->t);
					else
						LessEqualOPs.calculate_attribute(y->links->t, y->rechts->t, y->t);

					while(nil() != (y = (y->p)))
					{
						if(nil()==y->links) // dann muss y->rechts!=nil() sein
							LessEqualOPs.calculate_attribute2(y->rechts->t, y->t);
						else if(nil()==y->rechts)
							LessEqualOPs.calculate_attribute1(y->links->t, y->t);
						else
							LessEqualOPs.calculate_attribute(y->links->t, y->rechts->t, y->t);
					}
				}
			}
	void insert_fixup(Node *z)
			{
				while(nil()!=z->p && z->p->rot)
				{
					Node *y;
					if(z->p==z->p->p->links)
					{
						y=z->p->p->rechts;
						if(y->rot)
						{
							z->p->rot=false;
							y->rot=false;
							(z = z->p->p)->rot=true;
						}
						else
						{
							if(z==z->p->rechts)
								left_rotate( z = z->p );
							z->p->rot=false;
							z->p->p->rot=true;
							right_rotate(z->p->p);
						}
					}
					else
					{
						y=z->p->p->links;
						if(y->rot)
						{
							z->p->rot=false;
							y->rot=false;
							(z = z->p->p)->rot=true;
						}
						else
						{
							if(z==z->p->links)
								right_rotate( z = z->p );
							z->p->rot=false;
							z->p->p->rot=true;
							left_rotate(z->p->p);
						}
					}
				}
				wurzel->rot=false;
			}
	void erase_fixup(Node *x)
			{
				while(x!=wurzel && false==x->rot)
				{
					Node *w;
					if(x==x->p->links)
					{
						w=x->p->rechts;
						if(w->rot)
						{
							w->rot=false;
							x->p->rot=true;
							left_rotate(x->p);
							w=x->p->rechts;
						}
						if(false==w->links->rot && false==w->rechts->rot)
							{ w->rot=true; x=x->p; }
						else
						{
							if(false==w->rechts->rot)
							{
								w->links->rot=false;
								w->rot=true;
								right_rotate(w);
								w=x->p->rechts;
							}
							w->rot=x->p->rot;
							x->p->rot=false;
							w->rechts->rot=false;
							left_rotate(x->p);
							x=wurzel;
						}
					}
					else
					{
						w=x->p->links;
						if(w->rot)
						{
							w->rot=false;
							x->p->rot=true;
							right_rotate(x->p);
							w=x->p->links;
						}
						if(false==w->rechts->rot && false==w->links->rot)
							{ w->rot=true; x=x->p; }
						else
						{
							if(false==w->links->rot)
							{
								w->rechts->rot=false;
								w->rot=true;
								left_rotate(w);
								w=x->p->links;
							}
							w->rot=x->p->rot;
							x->p->rot=false;
							w->links->rot=false;
							right_rotate(x->p);
							x=wurzel;
						}
					}
				}
				x->rot=false;
			}

public:
	RotSchwarzBaum(__LessEqualOPs _less_equal=__LessEqualOPs(), Allocator _allct=Allocator())
			: LessEqualOPs(_less_equal), allct(_allct)
		{ nil()->rot=false; nil()->p=nil()->links=nil()->rechts=nil();
			wurzel=nil(); anz_elemente=0; }
	RotSchwarzBaum(const RotSchwarzBaum &b) : LessEqualOPs(b.LessEqualOPs), allct(b.allct)
		{ if(b.wurzel!=nil()) wurzel=copy(b.wurzel); else wurzel=nil();
			nil()->rot=false; anz_elemente=b.anz_elemente; }
	RotSchwarzBaum& operator=(const RotSchwarzBaum &b)
		{ if(this!=&b) { if(wurzel!=nil()) destruct(wurzel);
				if(b.wurzel!=nil()) wurzel=copy(b.wurzel); else wurzel=nil();
				LessEqualOPs=b.LessEqualOPs; allct=b.allct; nil()->rot=false;
				anz_elemente=b.anz_elemente; }
			return *this; }
	~RotSchwarzBaum() { if(wurzel!=nil()) destruct(wurzel); }

	void clear() { if(wurzel!=nil()) destruct(wurzel); wurzel=nil(); anz_elemente=0; }
		// ACHTUNG: Im Gegensatz zu den Iteratoren fr Listen wird ein iterator
		// von RotSchwarzBaum ungltig, wenn ein beliebiges Element aus dem Baum mit
		// erase_copy_t entfernt wird. Wird dagegen ein anderes Element als das
		// durch den Iterator bezeichnete mit erase_dont_copy_t entfernt, oder wird
		// ein beliebiges Element eingefgt, so bleibt der Iterator gltig.

	class iterator
	{		// Achtung: Nach auen hin verwendet iterator fr den Null-Iterator den Wert 0;
			// tatschlich jedoch wird in node dann der Zeiger nil() abgespeichert.
		Node *node;
		friend class RotSchwarzBaum<T, __LessEqualOPs, Allocator>;
		friend class const_iterator;
		iterator(Node *n) : node(n) { }

	public:
		static iterator from_node(Node *n) { return iterator(n); }
		iterator(const int dummy=0) { node=nil(); } // NICHT explicit verwenden, weil sonst
			// bei void f(iterator) { }  void g() { f(0); } der Konstruktur iterator(Node *n) mit n=0 aufgerufen wird.
		iterator(const iterator &i) : node(i.node) { }
		iterator& operator=(const iterator &i) { if(this!=&i) node=i.node; return *this; }
		iterator& operator=(const int dummy) { node=nil(); return *this; }

		T& operator*() const { assert(nil()!=node); return node->t; }
			// ACHTUNG: Nicht T operator*() definieren (kein Objekt, sondern Referenz zurckgeben),
			// denn sonst erhlt man mit dem Ausdruck &(*iter) einen Zeiger auf ein temporres Objekt,
			// welches in der folgenden Anweisung nicht mehr existiert.
		T* operator->() const { assert(nil()!=node); return &(node->t); }
		iterator operator++() // prefix ++
			{ if(node!=nil()) node=successor(node); return *this; } // ++maximizer() ergibt end() (node==0)
		iterator operator++(int) // postfix ++  ;  maximizer()-- ergibt end() (node==0)
			{ iterator save=*this; if(node!=nil()) node=successor(node); return save; }
		iterator operator--() // ACHTUNG:: --end() ergibt NICHT maximizer(), sondern wieder end() !
			{ if(node!=nil()) node=predecessor(node); return *this; }
		iterator operator--(int) // ACHTUNG:: end()-- ergibt NICHT maximizer(), sondern wieder end() !
			{ iterator save=*this; if(node!=nil()) node=predecessor(node); return save; }
		bool operator==(const iterator &i) const { return node==i.node; }
		bool operator==(const int x) const { return x==0 && node==nil(); }
		bool operator!=(const iterator &i) const { return node!=i.node; }
		bool operator!=(const int x) const { return x!=0 || node!=nil(); }
		bool operator<(const iterator &i) const  // ACHTUNG: Vergleich der Speicheradressen von node !
				{ if(node==nil()) return i.node!=nil();
					else return i.node!=nil() && ((unsigned long)node) < ((unsigned long)i.node); }
		bool operator<=(const iterator &i) const  // ACHTUNG: Vergleich der Speicheradressen von node !
				{ return node==nil() || (i.node!=nil() && ((unsigned long)node) <= ((unsigned long)i.node)); }
		bool operator>(const iterator &i) const  // ACHTUNG: Vergleich der Speicheradressen von node !
				{ if(i.node==nil()) return node!=nil();
					else return node!=nil() && ((unsigned long)i.node) < ((unsigned long)node); }
		bool operator>=(const iterator &i) const  // ACHTUNG: Vergleich der Speicheradressen von node !
				{ return i.node==nil() || (node!=nil() && ((unsigned long)i.node) <= ((unsigned long)node)); }
		bool is_uninitialized() const { return node==nil(); }
		bool is_initialized() const { return node!=nil(); }
		iterator parent() const { if(nil()==node || nil()==node->p) return iterator(); else return iterator(node->p); }
		iterator left() const { if(nil()==node || nil()==node->links) return iterator(); else return iterator(node->links); }
		iterator right() const  { if(nil()==node || nil()==node->rechts) return iterator(); else return iterator(node->rechts); }
		Node *get_node() const { return node; }
	};

	class const_iterator
	{		// Achtung: Nach auen hin verwendet iterator fr den Null-Iterator den Wert 0;
			// tatschlich jedoch wird in node dann der Zeiger nil() abgespeichert.
		Node *node;
		friend class RotSchwarzBaum<T, __LessEqualOPs, Allocator>;
		friend class iterator;
		const_iterator(Node *n) : node(n) { }

	public:
		static const_iterator from_node(Node *n) { return const_iterator(n); }
		const_iterator(const int dummy=0) { node=nil(); } // NICHT explicit verwenden, weil sonst
			// bei void f(iterator) { }  void g() { f(0); } der Konstruktur iterator(Node *n) mit n=0 aufgerufen wird.
		const_iterator(const const_iterator &i) : node(i.node) { }
		const_iterator& operator=(const const_iterator &i) { if(this!=&i) node=i.node; return *this; }
		const_iterator& operator=(const int dummy) { node=nil(); return *this; }
		const_iterator(const iterator &i) : node(i.node) { }

		const T& operator*() const { assert(nil()!=node); return node->t; }
			// ACHTUNG: Nicht T operator*() definieren (kein Objekt, sondern Referenz zurckgeben),
			// denn sonst erhlt man mit dem Ausdruck &(*iter) einen Zeiger auf ein temporres Objekt,
			// welches in der folgenden Anweisung nicht mehr existiert.
		const T* operator->() const { assert(nil()!=node); return &(node->t); }
		const_iterator operator++() // prefix ++
			{ if(node!=nil()) node=successor(node); return *this; } // ++maximizer() ergibt end() (node==0)
		const_iterator operator++(int) // postfix ++  ;  ++maximizer() ergibt end() (node==0)
			{ const_iterator save=*this; if(node!=nil()) node=successor(node); return save; }
		const_iterator operator--() // ACHTUNG:: --end() ergibt NICHT maximizer(), sondern wieder end() !
			{ if(node!=nil()) node=predecessor(node); return *this; }
		const_iterator operator--(int) // ACHTUNG:: end()-- ergibt NICHT maximizer(), sondern wieder end() !
			{ const_iterator save=*this; if(node!=nil()) node=predecessor(node); return save; }
		bool operator==(const const_iterator &i) const { return node==i.node; }
		bool operator==(const iterator &i) const { return node==i.node; }
		bool operator==(const int x) const { return x==0 && node==nil(); }
		bool operator!=(const const_iterator &i) const { return node!=i.node; }
		bool operator!=(const iterator &i) const { return node!=i.node; }
		bool operator!=(const int x) const { return x!=0 || node!=nil(); }
		bool operator<(const const_iterator &i) const  // ACHTUNG: Vergleich der Speicheradressen von node !
				{ if(node==nil()) return i.node!=nil();
					else return i.node!=nil() && ((unsigned long)node) < ((unsigned long)i.node); }
		bool operator<=(const const_iterator &i) const  // ACHTUNG: Vergleich der Speicheradressen von node !
				{ return node==nil() || (i.node!=nil() && ((unsigned long)node) <= ((unsigned long)i.node)); }
		bool operator>(const const_iterator &i) const  // ACHTUNG: Vergleich der Speicheradressen von node !
				{ if(i.node==nil()) return node!=nil();
					else return node!=nil() && ((unsigned long)i.node) < ((unsigned long)node); }
		bool operator>=(const const_iterator &i) const  // ACHTUNG: Vergleich der Speicheradressen von node !
				{ return i.node==nil() || (node!=nil() && ((unsigned long)i.node) <= ((unsigned long)node)); }
		bool is_uninitialized() const { return node==nil(); }
		bool is_initialized() const { return node!=nil(); }
		const_iterator parent() const { if(nil()==node || nil()==node->p) return const_iterator(); else return const_iterator(node->p); }
		const_iterator left() const { if(nil()==node || nil()==node->links) return const_iterator(); else return const_iterator(node->links); }
		const_iterator right() const { if(nil()==node || nil()==node->rechts) return const_iterator(); else return const_iterator(node->rechts); }
		iterator ConstCast() { return iterator(node); }
		Node *get_node() const { return node; }
	};

	inline iterator root() { return iterator(wurzel); }
	inline iterator root() const { return const_iterator(wurzel); }
	inline iterator maximizer() const { return iterator(tree_maximizer(wurzel)); }
	inline iterator minimizer() const { return iterator(tree_minimizer(wurzel)); }
	inline iterator end() { return iterator(0); }
		// ACHTUNG: --(end()) ergibt wieder end() und nicht einen
		// Iterator auf das letzte (grte) Element !
	inline const_iterator end() const { return iterator(0); }
		// ACHTUNG: --(end()) ergibt wieder end() und nicht einen
		// Iterator auf das letzte (grte) Element !
	inline iterator begin() { return iterator(tree_minimizer(wurzel)); }
	inline const_iterator begin() const { return const_iterator(tree_minimizer(wurzel)); }

	void attr_changed_recalculate(const_iterator _z)
			{
				if(LessEqualOPs.soll_attribut_berechnen)
					for(Node *z=_z.node ; nil()!=z ; z=z->p)
					{
						if(nil()==z->links)
						{
							if(nil()==z->rechts)
								LessEqualOPs.calculate_attribute(z->t);
							else
								LessEqualOPs.calculate_attribute2(z->rechts->t, z->t);
						}
						else if(nil()==z->rechts)
							LessEqualOPs.calculate_attribute1(z->links->t, z->t);
						else
							LessEqualOPs.calculate_attribute(z->links->t, z->rechts->t, z->t);
					}
			}
	iterator suche(const T &t) const
			{
				Node *n=wurzel;
				while(n!=nil() && !EqualOperator(n->t, t))
					if(LessOperator(t, n->t)) n=n->links; else n=n->rechts;
				return iterator(n);
			}
	iterator suche_minus_iter_bei_gleichen(const T &t) const
			{		// DIE LAUFZEIT DIESER FUNKTION IST PROPORTIONAL ZUM LOGARITHMUS AUS DER ANZAHL DER ELEMENTE DES BAUMS.
					// Es wird ein Iterator gesucht, welcher auf ein Baumelement mit dem Wert t
					// zeigt. Falls keiner existiert, wird der Nulliterator zurckgegeben.
					// Ansonsten wird ein Iterator zurckgegeben, welcher auf t zeigt. Falls
					// es mehrere Baumelemente mit dem Wert t gibt, wird derjenige Iterator auf
					// eines dieser Elemente zurckgegeben, welcher bei einer folgenden Inorder-
					// Transversierung des Baums gefunden wrde:
					//    iterator it=begin();
					//    while(it!=end() && !EqualOperator(it.node->t, t))
					//       ++it;
					//    return it;
				Node *n=wurzel, *n2;
				while(n!=nil())
				{
					if(EqualOperator(n->t, t))
						while(1)
						{
							while(n->links!=nil() && EqualOperator(n->links->t, t))
								n=n->links;
							if(nil()==n->links)
								return iterator(n);
							n2=n;
							n=n->links;
							while(n->rechts!=nil() && !EqualOperator( (n=n->rechts)->t , t ))
								;
							if(!EqualOperator( n->t, t ))
								return iterator(n2);
						}
					if(LessOperator(t, n->t))
						n=n->links;
					else
						n=n->rechts;
				}
				return iterator(0);
				// Der vorstehende Code ist quivalent zu folgendem (weniger zeiteffizientem) Code
				// Node *n=wurzel;
				// while(n!=nil() && !EqualOperator(n->t, t))
				//    if(LessOperator(t, n->t)) n=n->links; else n=n->rechts;
				// iterator it(n), it2;
				// if(it!=0)
				//    for(it2=it ; (--it2).node!=nil() && EqualOperator(it2.node->t, t) ; it=it2)
				//       ;
				// return it;
			}
	long anzahl_spez_elem_in_baum(const T &t) const
			{
				Node *n=wurzel;
				while(n!=nil() && !EqualOperator(n->t, t))
					if(LessOperator(t, n->t)) n=n->links; else n=n->rechts;
				long anz=0;
				if(n!=nil())
				{
					//check_rot_schwarz();
					iterator it(n), it2;
					for(it2=it ; (--it2).node!=nil() && EqualOperator(it2.node->t, t) ; it=it2)
						;
					while(it.node!=nil() && EqualOperator(it.node->t, t))
						{ ++anz; ++it; }
				}
				return anz;
			}
	iterator suche_naechstes_nichtgroesseres(const T &t) const
			{		// Falls t kleiner als alle Elemente des Baums ist, wird nil() zurckgegeben.
				Node *n;
				if((n=wurzel)!=nil())
					while(!EqualOperator(n->t, t))
					{
						if(LessOperator(t, n->t))
						{
							if(nil()==n->links) return iterator(predecessor(n));
							n=n->links;
						}
						else
						{
							if(nil()==n->rechts) return iterator(n);
							n=n->rechts;
						}
					}
				return iterator(n);
			}
	iterator suche_naechstes_nichtkleineres(const T &t) const
			{		// Falls t grer als alle Elemente des Baums ist, wird nil() zurckgegeben.
				Node *n;
				if((n=wurzel)!=nil())
					while(!EqualOperator(n->t, t))
					{
						if(LessOperator(n->t, t))
						{
							if(nil()==n->rechts) return iterator(successor(n));
							n=n->rechts;
						}
						else
						{
							if(nil()==n->links) return iterator(n);
							n=n->links;
						}
					}
				return iterator(n);
			}
	iterator suche_naechstes_echt_groesseres(const T &t) const
			{		// Falls t grer als oder gleich wie alle Elemente des Baums ist, wird nil() zurckgegeben.
				Node *n;
				if((n=wurzel)!=nil())
					while(1)
					{
						if(LessOperator(t, n->t))
						{
							if(nil()==n->links) return iterator(n);
							n=n->links;
						}
						else
						{
							if(nil()==n->rechts) return iterator(successor(n));
							n=n->rechts;
						}
					}
				return iterator(n);
			}
	iterator suche_naechstes_echt_kleineres(const T &t) const
			{		// Falls t kleiner als oder gleich wie alle Elemente des Baums ist, wird nil() zurckgegeben.
				Node *n;
				if((n=wurzel)!=nil())
					while(1)
					{
						if(LessOperator(n->t, t))
						{
							if(nil()==n->rechts) return iterator(n);
							n=n->rechts;
						}
						else
						{
							if(nil()==n->links) return iterator(predecessor(n));
							n=n->links;
						}
					}
				return iterator(n);
			}

	iterator insert(const T &t)
			{		// Fgt ein neues Element mit dem Wert t in den Baum ein.
					// Falls bereits ein Element t vorhanden ist, wird es trotzdem eingefgt, und zwar
					// an das "Ende" der vorhandenen t-Elemente, d.h. mit dem von iter=insert(t) gelieferten
					// Iterator iter gilt *iter==t, und falls im Baum noch grere Elemente als t vorhanden
					// sind, gilt *(++iter)>t .
				Node *z=new_node(t);
				return insert_inplace(const_iterator(z), t);
			}
	iterator insert_inplace(const_iterator _z, const T &t)
			{
				if(nil()==_z.node) return nil();
				_z.node->t=t;
				return insert_inplace(_z);
			}
	iterator insert_inplace(const_iterator _z)
			{
				Node *z=_z.node;
				if(nil()==z) return nil();
				z->links=z->rechts=nil(); z->rot=true;
				if(LessEqualOPs.soll_attribut_berechnen)
					LessEqualOPs.calculate_attribute(z->t);
				Node *y=nil(), *x=wurzel;
				while(nil()!=x)
				{
					y=x;
					if(LessOperator(z->t, x->t)) x=x->links;
						else x=x->rechts;
				}
				if(nil()==(z->p=y))
					wurzel=z;
				else
				{
					if(LessOperator(z->t, y->t))
						y->links=z;
					else
						y->rechts=z;
					if(LessEqualOPs.soll_attribut_berechnen)
						do
						{
							if(nil()==y->links) // dann muss y->rechts!=nil() sein
								LessEqualOPs.calculate_attribute2(y->rechts->t, y->t);
							else if(nil()==y->rechts)
								LessEqualOPs.calculate_attribute1(y->links->t, y->t);
							else
								LessEqualOPs.calculate_attribute(y->links->t, y->rechts->t, y->t);
						} while(nil() != (y = (y->p)));
				}
				insert_fixup(z);
				++anz_elemente;
				return iterator(z);
			}

	iterator insert_first_if_equal(const T &t)
			{		// Fgt ein neues Element mit dem Wert t in den Baum ein.
					// Falls bereits ein Element t vorhanden ist, wird es trotzdem eingefgt, und zwar
					// an den "Anfang" der vorhandenen t-Elemente, d.h. mit dem von iter=insert(t) gelieferten
					// Iterator iter gilt *iter==t, und falls im Baum noch kleinere Elemente als t vorhanden
					// sind, gilt *(--iter)<t .
				Node *z=new_node(t);
				return insert_inplace_first_if_equal(const_iterator(z), t);
			}
	iterator insert_inplace_first_if_equal(const_iterator _z, const T &t)
			{
				Node *z=_z.node;
				if(nil()==z) return nil();
				z->t=t;
				z->links=z->rechts=nil(); z->rot=true;
				if(LessEqualOPs.soll_attribut_berechnen)
					LessEqualOPs.calculate_attribute(z->t);
				Node *y=nil(), *x=wurzel;
				while(nil()!=x)
				{
					y=x;
					if(LessOperator(x->t, t)) x=x->rechts;
						else x=x->links;
				}
				if(nil()==(z->p=y))
					wurzel=z;
				else
				{
					if(LessOperator(t, y->t))
						y->links=z;
					else
						y->rechts=z;
					if(LessEqualOPs.soll_attribut_berechnen)
						do
						{
							if(nil()==y->links) // dann muss y->rechts!=nil() sein
								LessEqualOPs.calculate_attribute2(y->rechts->t, y->t);
							else if(nil()==y->rechts)
								LessEqualOPs.calculate_attribute1(y->links->t, y->t);
							else
								LessEqualOPs.calculate_attribute(y->links->t, y->rechts->t, y->t);
						} while(nil() != (y = (y->p)));
				}
				insert_fixup(z);
				++anz_elemente;
				return iterator(z);
			}

	iterator insert_before(const_iterator pos, const T &t)
			{		// Fgt ein neues Element mit dem Wert t in den Baum vor das Element mit dem Iterator pos ein,
					// d.h. nach dem Einfgen x=insert_before(pos, t) gilt: ++x==pos.
					// ACHTUNG: Diese Funktion darf anstatt von insert() nur dann verwendet werden, wenn eine der
					// folgenden Bedingungen erfllt ist:
					//      1. Es ist sichergestellt, dass der Schlssel von t tatschlich zwischen dem Schlssel
					//         von pos und dem Schlssel von --pos liegt.
					//      2. Es werden prinzipiell keine Schlssel-Vergleiche durchgefhrt, so dass der Baum
					//         als Liste verwendet wird.
					// ********** Wenn pos der Nulliterator ist (pos==end()), dann wird t nach den
					// ********** maximizer() eingefgt.
				Node *z=new_node(t);
				return insert_inplace_before(pos, const_iterator(z), t);
			}

	iterator push_back(const T &t)
			{		// Siehe die WARNUNGEN bei insert_before() !
				Node *z=new_node(t);
				return insert_inplace_before(nil(), z, t);
			}

	iterator insert_inplace_before(const_iterator pos, const_iterator _z, const T &t)
			{		// Wenn pos==0 ist, wird t NACH dem 'maximizer' eingefgt.
				if(0==pos.node) return insert_inplace_after(tree_maximizer(wurzel), _z, t);
				Node *z=_z.node;
				if(nil()==z) return nil();
				z->t=t;
				z->links=z->rechts=nil(); z->rot=true;
				if(LessEqualOPs.soll_attribut_berechnen)
					LessEqualOPs.calculate_attribute(z->t);
				if(nil()==wurzel)
					{ wurzel=z; z->p=nil(); }
				else
				{
					Node *y=pos.node;
					if(y==nil()) // bedeutet (nur HIER): nach dem 'maximizer'
					{
						y=wurzel;
						while(nil()!=y->rechts)
							y=y->rechts;					
					}
					else if(nil()!=y->links)
					{
						y=y->links;
						while(nil()!=y->rechts)
							y=y->rechts;
					}
					z->p=y;

					if(y==pos.node)
						y->links=z;
					else
						y->rechts=z;
					if(LessEqualOPs.soll_attribut_berechnen)
						do
						{
							if(nil()==y->links) // dann muss y->rechts!=nil() sein
								LessEqualOPs.calculate_attribute2(y->rechts->t, y->t);
							else if(nil()==y->rechts)
								LessEqualOPs.calculate_attribute1(y->links->t, y->t);
							else
								LessEqualOPs.calculate_attribute(y->links->t, y->rechts->t, y->t);
						} while(nil() != (y = (y->p)));
				}
				insert_fixup(z);
				++anz_elemente;
				return iterator(z);
			}

	iterator insert_after(const_iterator pos, const T &t)
			{		// Fgt ein neues Element mit dem Wert t in den Baum vor das Element mit dem Iterator pos ein,
					// d.h. nach dem Einfgen x=insert_before(pos, t) gilt: ++x==pos.
					// ACHTUNG: Diese Funktion darf anstatt von insert() nur dann verwendet werden, wenn eine der
					// folgenden Bedingungen erfllt ist:
					//      1. Es ist sichergestellt, dass der Schlssel von t tatschlich zwischen dem Schlssel
					//         von pos und dem Schlssel von --pos liegt.
					//      2. Es werden prinzipiell keine Schlssel-Vergleiche durchgefhrt, so dass der Baum
					//         als Liste verwendet wird.
					// ********** Wenn pos der Nulliterator ist (pos==end()), dann wird t nach den
					// ********** maximizer() eingefgt.
				Node *z=new_node(t);
				return insert_inplace_after(pos, const_iterator(z), t);
			}

	iterator insert_inplace_after(const_iterator pos, const_iterator _z, const T &t)
			{		// Wenn pos==0 ist, wird t nach dem 'maximizer' eingefgt.
				if(0==pos.node) pos.node=tree_maximizer(wurzel);
				Node *z=_z.node;
				if(nil()==z) return nil();
				z->t=t;
				z->links=z->rechts=nil(); z->rot=true;
				if(LessEqualOPs.soll_attribut_berechnen)
					LessEqualOPs.calculate_attribute(z->t);
				Node *y=pos.node;
				if(nil()!=y->rechts)
				{
					y=y->rechts;
					while(nil()!=y->links)
						y=y->links;
				}
				z->p=y;
				if(nil()==wurzel)
					{ wurzel=z; z->p=nil(); }
				else
				{
					if(y==pos.node)
						y->rechts=z;
					else
						y->links=z;
					if(LessEqualOPs.soll_attribut_berechnen)
						do
						{
							if(nil()==y->links) // dann muss y->rechts!=nil() sein
								LessEqualOPs.calculate_attribute2(y->rechts->t, y->t);
							else if(nil()==y->rechts)
								LessEqualOPs.calculate_attribute1(y->links->t, y->t);
							else
								LessEqualOPs.calculate_attribute(y->links->t, y->rechts->t, y->t);
						} while(nil() != (y = (y->p)));
				}
				insert_fixup(z);
				++anz_elemente;
				return iterator(z);
			}

	void erase_copy_t(iterator _z)
		// ACHTUNG: Im Gegensatz zu den Iteratoren fr Listen wird ein iterator
		// von RotSchwarzBaum ungltig, wenn ein beliebiges Element aus dem Baum mit
		// erase_copy_t entfernt wird. Wird dagegen ein anderes Element als das
		// durch den Iterator bezeichnete mit erase_dont_copy_t entfernt, oder wird
		// ein beliebiges Element eingefgt, so bleibt der Iterator gltig.
			{
				//if(nil()!=_z.node) TRACE("erase %d", _z.node->t);
				if(nil()==_z.node) return;
				Node *z=_z.node, *x, *y;
				if(nil()==z->links || nil()==z->rechts) y=z;
					else y=successor(z);
				if(nil()!=y->links) x=y->links;
					else x=y->rechts;
				x->p=y->p;
				if(nil()==y->p)
					wurzel=x;
				else if(y==y->p->links)
					y->p->links=x;
				else
					y->p->rechts=x;
				if(y!=z)
					z->t=y->t;
				if(LessEqualOPs.soll_attribut_berechnen)
				{
					for(z=y ; nil()!=(z = (z->p)) ; )
					{		// vor erase_fixup() !
						if(nil()==z->links)
						{
							if(nil()==z->rechts)
								LessEqualOPs.calculate_attribute(z->t);
							else
								LessEqualOPs.calculate_attribute2(z->rechts->t, z->t);
						}
						else if(nil()==z->rechts)
							LessEqualOPs.calculate_attribute1(z->links->t, z->t);
						else
							LessEqualOPs.calculate_attribute(z->links->t, z->rechts->t, z->t);
					}
				}
				if(false==y->rot)
					erase_fixup(x);
				if(y!=nil()) { delete_node(y); --anz_elemente; }
				//check_rot_schwarz_silent(); // ====> NUR ZUM DEBUGGEN <====
			}
	const_iterator change_and_reorder(const_iterator _z, const T& t)
	{		// Das Element, auf welches _z zeigt, erhlt den Wert t und wird ggf. neu in den
			// Baum einsortiert. Im Gegensatz zur Befehlsfolge erase() und insert() bleiben
			// bei dieser Funktion change_and_reorder alle Iteratoren, insbesondere _z, gltig.
			// ACHTUNG: OBWOHL _z ein const_iterator ist, wird das Element verndert.
		return insert_inplace(erase_dont_delete(iterator(_z.node)), t);
	}
	const_iterator changed_do_reorder(const_iterator _z)
	{		// Das Element, auf welches _z zeigt, wird nach einer ZUVOR erfolgten nderung ggf. neu in den
			// Baum einsortiert. Im Gegensatz zur Befehlsfolge erase() und insert() bleiben
			// bei dieser Funktion change_and_reorder alle Iteratoren, insbesondere _z, gltig.
			// ACHTUNG: OBWOHL _z ein const_iterator ist, wird das Element verndert.
		return insert_inplace(erase_dont_delete(iterator(_z.node)));
	}
	void erase(iterator _z)
		// ACHTUNG: Im Gegensatz zu den Iteratoren fr Listen wird ein iterator
		// von RotSchwarzBaum ungltig, wenn ein beliebiges Element aus dem Baum mit
		// erase_copy_t entfernt wird. Wird dagegen ein anderes Element als das
		// durch den Iterator bezeichnete mit erase_dont_copy_t entfernt, oder wird
		// ein beliebiges Element eingefgt, so bleibt der Iterator gltig.
			{
				//if(nil()!=_z.node) TRACE("erase %d", _z.node->t);
				if(nil()==_z.node) return;
				Node *z=_z.node, *x, *y;
				if(nil()==z->links || nil()==z->rechts) y=z;
					else y=successor(z);
				if(nil()!=y->links) x=y->links;
					else x=y->rechts;
				x->p=y->p;
				if(nil()==y->p)
					wurzel=x;
				else if(y==y->p->links)
					y->p->links=x;
				else
					y->p->rechts=x;
				if(copying_t_allowed())
				{
					if(y!=z)
						z->t=y->t;
					if(LessEqualOPs.soll_attribut_berechnen)
					{
						for(z=y ; nil()!=(z = (z->p)) ; )
						{		// vor erase_fixup() !
							if(nil()==z->links)
							{
								if(nil()==z->rechts)
									LessEqualOPs.calculate_attribute(z->t);
								else
									LessEqualOPs.calculate_attribute2(z->rechts->t, z->t);
							}
							else if(nil()==z->rechts)
								LessEqualOPs.calculate_attribute1(z->links->t, z->t);
							else
								LessEqualOPs.calculate_attribute(z->links->t, z->rechts->t, z->t);
						}
					}
					if(false==y->rot)
						erase_fixup(x);
					if(y!=nil()) { delete_node(y); --anz_elemente; }					
				}
				else
				{
					if(y!=z && y!=nil())
					{
						if(nil()==z->p)
							wurzel=y;
						else if(z==z->p->links)
							z->p->links=y;
						else
							z->p->rechts=y;
						if(nil() != (y->links=z->links)) y->links->p=y;
						if(nil() != (y->rechts=z->rechts)) y->rechts->p=y;
						y->p=z->p;
						if(x->p==z) x->p=y;  // Wenn x==nil() ist, wird der in nil()->p gespeicherte
											 // Wert von erase_fixup() benutzt.
					}
					if(y->rot)
						{ y->rot=z->rot; y=x->p; x=0; } // x=0: kein erase_fixup(); ACHTUNG: KEIN x=nil() GEWHLT !!!
					else
						{ y->rot=z->rot; y=x->p; }
					if(LessEqualOPs.soll_attribut_berechnen)
					{
						for( ; nil()!=y ; y=y->p)
						{		// vor erase_fixup() !
							if(nil()==y->links)
							{
								if(nil()==y->rechts)
									LessEqualOPs.calculate_attribute(y->t);
								else
									LessEqualOPs.calculate_attribute2(y->rechts->t, y->t);
							}
							else if(nil()==y->rechts)
								LessEqualOPs.calculate_attribute1(y->links->t, y->t);
							else
								LessEqualOPs.calculate_attribute(y->links->t, y->rechts->t, y->t);
						}
					}
					if(x) // ACHTUNG: kein x==nil() (siehe oben: x=0)
						erase_fixup(x);
					delete_node(z); --anz_elemente;					
				}
				//check_rot_schwarz(); // ====> NUR ZUM DEBUGGEN <====
			}
	void erase_dont_copy_t(iterator _z)
		{
			if(_z.node!=nil())
			{
				erase_dont_delete(_z);
				delete_node(_z);
			}
		}
	iterator erase_dont_delete(iterator _z) // und dont_copy_t
			{
				//if(nil()!=_z.node) TRACE("erase %d", _z.node->t);
				if(nil()==_z.node) return _z;
				Node *z=_z.node, *x, *y;
				if(nil()==z->links || nil()==z->rechts) y=z;
					else y=successor(z);
				if(nil()!=y->links) x=y->links;
					else x=y->rechts;
				x->p=y->p;
				if(nil()==y->p)
					wurzel=x;
				else if(y==y->p->links)
					y->p->links=x;
				else
					y->p->rechts=x;
				if(y!=z && y!=nil())
				{
					if(nil()==z->p)
						wurzel=y;
					else if(z==z->p->links)
						z->p->links=y;
					else
						z->p->rechts=y;
					if(nil() != (y->links=z->links)) y->links->p=y;
					if(nil() != (y->rechts=z->rechts)) y->rechts->p=y;
					y->p=z->p;
					if(x->p==z) x->p=y;  // Wenn x==nil() ist, wird der in nil()->p gespeicherte
										 // Wert von erase_fixup() benutzt.
				}
				if(y->rot)
					{ y->rot=z->rot; y=x->p; x=0; } // x=0: kein erase_fixup(); ACHTUNG: KEIN x=nil() GEWHLT !!!
				else
					{ y->rot=z->rot; y=x->p; }
				if(LessEqualOPs.soll_attribut_berechnen)
				{
					for( ; nil()!=y ; y=y->p)
					{		// vor erase_fixup() !
						if(nil()==y->links)
						{
							if(nil()==y->rechts)
								LessEqualOPs.calculate_attribute(y->t);
							else
								LessEqualOPs.calculate_attribute2(y->rechts->t, y->t);
						}
						else if(nil()==y->rechts)
							LessEqualOPs.calculate_attribute1(y->links->t, y->t);
						else
							LessEqualOPs.calculate_attribute(y->links->t, y->rechts->t, y->t);
					}
				}
				if(x) // ACHTUNG: kein x==nil() (siehe oben: x=0)
					erase_fixup(x);
				--anz_elemente;
				//check_rot_schwarz_silent(); // ====> NUR ZUM DEBUGGEN <====
				return _z;
			}
	void erase(const T &t) { erase(suche(t)); }
	void erase_all_equals(const T &t) { iterator it; while((it=suche(t))!=0) erase(it); }
			// erase_all entfernt alle Elemente, die gleich t sind.

	unsigned long size() const { return anz_elemente; }
	bool empty() const { return 0==anz_elemente; }

	const __LessEqualOPs& get_LessEqualOPs() const { return LessEqualOPs; }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////

template<class T> struct nlist_Node;

template<class T>
struct nlist_Node_Base
{
	nlist_Node<T> *prev, *next;
	nlist_Node_Base() : next(0), prev(0) { }
	nlist_Node_Base(const nlist_Node_Base &l) : prev(l.prev), next(l.next) { }
	nlist_Node_Base& operator=(const nlist_Node_Base &l) { prev=l.prev; next=l.next; return *this; }
};

template<class T>
struct nlist_Node : public nlist_Node_Base<T>
{
	T t;
	nlist_Node(const T &_t=T()) : t(_t) { }
	nlist_Node(const nlist_Node &l) : nlist_Node_Base<T>(l), t(l.t) { }
	nlist_Node& operator=(const nlist_Node &l) { nlist_Node_Base<T>::operator=(l); t=l.t; return *this; }
	static nlist_Node<T> *node_address_from_t_address(T *t)
	{	// Diese Funktion ist abhngig von der Compiler-Implementierung und sollte mit Vorsicht benutzt werden.
		assert(t == &((nlist_Node<T> *)(((char*)t)-offsetof(nlist_Node<T>, t)))->t);
		return (nlist_Node<T> *)(((char*)t)-offsetof(nlist_Node<T>, t));
	}
};

template<class T, class nlist_Node_Allocator = mem_allocator<nlist_Node<T> > >
class nlist
{
	typedef nlist_Node<T> Node;
	nlist_Node_Allocator allct;

	size_t anz_elem;
	nlist_Node_Base<T> base_endobj;
	Node *endobj() { return static_cast<Node*>(&base_endobj); }
		// ACHTUNG: Auf endobj()->t darf nicht zugegriffen werden !
	const Node *endobj() const { return static_cast<const Node*>(&base_endobj); }
		// ACHTUNG: Auf endobj()->t darf nicht zugegriffen werden !

public:
	class iterator;
	class const_iterator;
	class Iterator;
	class Const_Iterator;
	friend class iterator;
	friend class const_iterator;
	friend class Iterator;
	friend class Const_Iterator;
private:

	/*inline Node *new_node(const Node *node)
			{
				return new Node(node);
			}
	inline Node *new_node(const T &t)
			{
				return new Node(t);
			}
	inline void delete_node(Node *node)
			{
				delete node;
			}*/
	inline Node *new_node(const Node *node)
			{
				Node *nd=allct.allocate(1, 0);
				if(0!=nd) allct.construct(nd, node);
				return nd;
			}
	inline Node *new_node(const T &t)
			{
				Node *nd=allct.allocate(1, 0);
				if(0!=nd) allct.construct(nd, Node(t));
				return nd;
			}
	inline void delete_node(Node *node)
			{
				allct.destroy(node);
				allct.deallocate(node,1);
			}

public:

	class iterator
	{		// iterator==0 bei lst=0
		nlist_Node<T> *n;
#ifdef _DEBUG
		nlist_Node<T> *EndObj;
#endif
		friend class nlist<T, nlist_Node_Allocator>;
		friend class const_iterator;
		friend class Iterator;

#ifdef _DEBUG
		iterator(nlist_Node<T> *_n, nlist_Node<T> *_EndObj) : n(_n), EndObj(_EndObj) { }
#else
		iterator(nlist_Node<T> *_n) : n(_n) { }
#endif

	public:
#ifdef _DEBUG
		iterator(long dummy) { n=0; EndObj=0; }
		iterator(int dummy=0) { n=0; EndObj=0; }
			// damit der Wert 0 explizit bergeben werden kann.
		iterator(const iterator &i) : n(i.n), EndObj(i.EndObj) { }
		iterator& operator=(const iterator &i) { n=i.n; EndObj=i.EndObj; return *this; }
		iterator& operator=(long dummy) { n=0; EndObj=0; return *this; }

		T& operator*() const { assert(0!=n && EndObj!=n); return n->t; }
			// ACHTUNG: Der verbotene Zugriff auf *end() wird mit assert() hier nicht erkannt !
			// const ist zulssig, weil der Operator selbst nicht verndert wird
		T* operator->() const { assert(0!=n && EndObj!=n); return &(n->t); }
			// ACHTUNG: Der verbotene Zugriff auf *end() wird mit assert() hier nicht erkannt !
			// const ist zulssig, weil der Operator selbst nicht verndert wird

		iterator operator++() // prefix ++
			{ assert(0!=n && EndObj!=n); n=n->next; return *this; }
		iterator operator++(int) // postfix ++
			{ iterator save=*this; assert(0!=n && EndObj!=n); n=n->next; return save; }
		iterator operator--() { assert(0!=n && EndObj!=n->prev); n=n->prev; return *this; }
		iterator operator--(int) { iterator save=*this; assert(0!=n && EndObj!=n->prev); n=n->prev; return save; }
#else
		iterator(long dummy) { n=0; }
		iterator(int dummy=0) { n=0; }
			// damit der Wert 0 explizit bergeben werden kann.
		iterator(const iterator &i) : n(i.n) { }
		iterator& operator=(const iterator &i) { n=i.n; return *this; }
		iterator& operator=(long dummy) { n=0; return *this; }

		T& operator*() const { assert(0!=n); return n->t; }
			// ACHTUNG: Der verbotene Zugriff auf *end() wird mit assert() hier nicht erkannt !
			// const ist zulssig, weil der Operator selbst nicht verndert wird
		T* operator->() const { assert(0!=n); return &(n->t); }
			// ACHTUNG: Der verbotene Zugriff auf *end() wird mit assert() hier nicht erkannt !
			// const ist zulssig, weil der Operator selbst nicht verndert wird

		iterator operator++() // prefix ++
			{ n=n->next; return *this; }
		iterator operator++(int) // postfix ++
			{ iterator save=*this; n=n->next; return save; }
		iterator operator--() { n=n->prev; return *this; }
		iterator operator--(int) { iterator save=*this; n=n->prev; return save; }
#endif

		bool operator==(const const_iterator &i) const { return i.n==n; }
		bool operator==(const iterator &i) const { return i.n==n; }
		bool operator==(const long x) const { return x==0 && n==0; }
		bool operator!=(const const_iterator &i) const { return n!=i.n; }
		bool operator!=(const iterator &i) const { return n!=i.n; }
		bool operator!=(const long x) const { return x!=0 || n!=0; }

		Node *get_node() const { return n; }
		bool is_valid_and_not_end(const nlist<T> &lst) const { return 0!=n && lst.endobj()!=n; }
	};

	class const_iterator
	{		// const_iterator==0 bei lst=0, const_iterator==end() bei lst!=0, it=0
		nlist_Node<T> *n;
#ifdef _DEBUG
		nlist_Node<T> *EndObj;
#endif
		friend class nlist<T, nlist_Node_Allocator>;
		friend class iterator;
		friend class Const_Iterator;

#ifdef _DEBUG
		const_iterator(const nlist_Node<T> *_n, const nlist_Node<T> *_EndObj)
			: n(const_cast<nlist_Node<T>*>(_n)), EndObj(const_cast<nlist_Node<T>*>(_EndObj)) { }
#else
		const_iterator(const nlist_Node<T> *_n) : n(const_cast<nlist_Node<T>*>(_n)) { }
#endif

	public:
#ifdef _DEBUG
		const_iterator(long dummy) { n=0; EndObj=0; }
		const_iterator(int dummy=0) { n=0; EndObj=0; }
			// damit der Wert 0 explizit bergeben werden kann.
		const_iterator(const const_iterator &i) : n(i.n), EndObj(i.EndObj) { }
		const_iterator(const iterator &i) : n(i.n), EndObj(i.EndObj) { }
		const_iterator& operator=(const const_iterator &i) { n=i.n; EndObj=i.EndObj; return *this; }
		const_iterator& operator=(const iterator &i) { n=i.n; EndObj=i.EndObj; return *this; }
		const_iterator& operator=(long dummy) { n=0; EndObj=0; return *this; }

		const T& operator*() const { assert(0!=n && EndObj!=n); return n->t; }
			// ACHTUNG: Der verbotene Zugriff auf *end() wird mit assert() hier nicht erkannt !
		const T* operator->() const { assert(0!=n && EndObj!=n); return &(n->t); }
			// ACHTUNG: Der verbotene Zugriff auf *end() wird mit assert() hier nicht erkannt !
		const_iterator operator++() // prefix ++
			{ assert(0!=n && EndObj!=n); n=n->next; return *this; }
		const_iterator operator++(int) // postfix ++
			{ const_iterator save=*this; assert(0!=n && EndObj!=n); n=n->next; return save; }
		const_iterator operator--() { assert(0!=n && EndObj!=n->prev); n=n->prev; return *this; }
		const_iterator operator--(int) { const_iterator save=*this;
				assert(0!=n && EndObj!=n->prev); n=n->prev; return save; }
#else
		const_iterator(long dummy) { n=0; }
		const_iterator(int dummy=0) { n=0; }
			// damit der Wert 0 explizit bergeben werden kann.
		const_iterator(const const_iterator &i) : n(i.n) { }
		const_iterator(const iterator &i) : n(i.n) { }
		const_iterator& operator=(const const_iterator &i) { n=i.n; return *this; }
		const_iterator& operator=(const iterator &i) { n=i.n; return *this; }
		const_iterator& operator=(long dummy) { n=0; return *this; }

		const T& operator*() const { assert(0!=n); return n->t; }
			// ACHTUNG: Der verbotene Zugriff auf *end() wird mit assert() hier nicht erkannt !
		const T* operator->() const { assert(0!=n); return &(n->t); }
			// ACHTUNG: Der verbotene Zugriff auf *end() wird mit assert() hier nicht erkannt !
		const_iterator operator++() // prefix ++
			{ n=n->next; return *this; }
		const_iterator operator++(int) // postfix ++
			{ const_iterator save=*this; n=n->next; return save; }
		const_iterator operator--() { n=n->prev; return *this; }
		const_iterator operator--(int) { const_iterator save=*this; n=n->prev; return save; }
#endif

		bool operator==(const const_iterator &i) const { return i.n==n; }
		bool operator==(const iterator &i) const { return i.n==n; }
		bool operator==(const long x) const { return x==0 && n==0; }
		bool operator!=(const const_iterator &i) const { return n!=i.n; }
		bool operator!=(const iterator &i) const { return n!=i.n; }
		bool operator!=(const long x) const { return x!=0 || n!=0; }
		iterator ConstCast() { return iterator(n); }

		const Node *get_node() const { return n; }
		bool is_valid_and_not_end(const nlist<T> &lst) const { return 0!=n && lst.endobj()!=n; }
	};

	class Iterator : public iterator
	{
		nlist<T> *lst;
		Iterator(nlist_Node<T> *_n, nlist<T> *_lst) : iterator(_n), lst(_lst) { }
		friend class Const_Iterator;
	public:
		Iterator(long dummy) : iterator(dummy), lst(0) { }
		Iterator(int dummy=0) : iterator(dummy), lst(0) { } // damit der Wert 0 explizit bergeben werden kann.
		Iterator(const Iterator &i) : iterator(i), lst(i.lst) { }
		Iterator(const iterator &i, nlist<T> &_lst) : iterator(i), lst(&_lst) { }
		Iterator& operator=(const Iterator &i) { iterator::operator=(i); lst=i.lst; return *this; }
		Iterator& operator=(long dummy) { iterator::operator=(dummy); lst=0; return *this; }
		bool is_end() const { return 0!=lst && iterator::n==lst->endobj(); }
	};

	class Const_Iterator : public const_iterator
	{
		const nlist<T> *lst;
	public:
		Const_Iterator(long dummy) : const_iterator(dummy), lst(0) { }
		Const_Iterator(int dummy=0) : const_iterator(dummy), lst(0) { } // damit der Wert 0 explizit bergeben werden kann.
		Const_Iterator(const Const_Iterator &i) : const_iterator(i), lst(i.lst) { }
		Const_Iterator(const Iterator &i) : const_iterator(i), lst(i.lst) { }
		Const_Iterator(const const_iterator &i, const nlist<T> &_lst) : const_iterator(i), lst(&_lst) { }
		Const_Iterator(const iterator &i, const nlist<T> &_lst) : const_iterator(i), lst(&_lst) { }
		Const_Iterator& operator=(const Const_Iterator &i) { const_iterator::operator=(i); lst=i.lst; return *this; }
		Const_Iterator& operator=(const Iterator &i) { const_iterator::operator=(i); lst=i.lst; return *this; }
		Const_Iterator& operator=(long dummy) { const_iterator::operator=(dummy); lst=0; return *this; }
		Iterator ConstCast() { return Iterator(const_iterator::n, lst); }
		bool is_end() const { return 0!=lst && const_iterator::n==lst->endobj(); }
	};

public:

	~nlist() { clear(); }
	nlist(const nlist_Node_Allocator &_a=nlist_Node_Allocator()) : allct(_a)
		{ endobj()->next=endobj()->prev=endobj(); anz_elem=0; }
	nlist(const nlist<T, nlist_Node_Allocator> &l) : allct(l.allct)
			{
				const Node *nd;
				Node *n=endobj();
				for(nd=l.endobj()->next ; nd!=l.endobj() ; nd=nd->next)
				{
					n->next=new_node(nd->t);
					n->next->prev=n;
					n=n->next;
				}
				endobj()->prev=n;
				n->next=endobj();
				anz_elem=l.anz_elem;
			}
	nlist& operator=(const nlist<T, nlist_Node_Allocator>& l)
		{ if(this!=&l)
			{
				allct=l.allct;
				const Node *nd;
				Node *n;
				for(n=endobj()->next ; n!=endobj() ; )
					{ Node *nach=n->next; delete_node(n); n=nach; }
				n=endobj();
				for(nd=l.endobj()->next ; nd!=l.endobj() ; nd=nd->next)
				{
					n->next=new_node(nd->t);
					n->next->prev=n;
					n=n->next;
				}
				endobj()->prev=n;
				n->next=endobj();
				anz_elem=l.anz_elem;
			}
			return *this;
		}

#ifdef _DEBUG
	inline iterator begin()
		{ return iterator(endobj()->next, endobj()); }
	inline const_iterator begin() const
		{ return const_iterator(endobj()->next, endobj()); }
	inline iterator end()
		{ return iterator(endobj(), endobj()); }
	inline const_iterator end() const
		{ return const_iterator(endobj(), endobj()); }
#else
	inline iterator begin()
		{ return iterator(endobj()->next); }
	inline const_iterator begin() const
		{ return const_iterator(endobj()->next); }
	inline iterator end()
		{ return iterator(endobj()); }
	inline const_iterator end() const
		{ return const_iterator(endobj()); }
#endif

	inline size_t size() const { return anz_elem; }
	bool empty() const { return 0==anz_elem; }

	iterator insert(const_iterator it, const T& x=T())
		{
			if(it.n!=0)
			{
				Node *n=new_node(x);
				n->prev=it.n->prev; it.n->prev->next=n;
				it.n->prev=n; n->next=it.n;
				++anz_elem;
#ifdef _DEBUG
				return iterator(n, endobj());
#else
				return iterator(n);
#endif
			}
#ifdef _DEBUG
			return iterator(0,0);
#else
			return iterator(0);
#endif
		}

	iterator insert(const_iterator it, const_iterator list2_first, const_iterator list2_last)
		{
			iterator first;
			if(list2_first!=list2_last)
				first=insert(it, *(list2_first++));
			while(list2_first!=list2_last)
				insert(it, *(list2_first++));
			return first;
		}

	iterator insert(const_iterator it, const nlist<T> &lst)
		{ return insert(it, lst.begin(), lst.end()); }

	iterator insert_inplace(const_iterator it, iterator obj)
		{
			if(obj.n!=0)
			{
				obj.n->prev=it.n->prev; it.n->prev->next=obj.n;
				it.n->prev=obj.n; obj.n->next=it.n;
				++anz_elem;
			}
			return obj;
		}

	iterator erase(iterator it)
		{
			if(it.n!=0 && it.n!=endobj())
			{
				Node *n=it.n->next;
				it.n->prev->next=n;
				n->prev=it.n->prev;
				delete_node(it.n);
				--anz_elem;
#ifdef _DEBUG
				return iterator(n, endobj());
			}
			return iterator(0,0);
#else
				return iterator(n);
			}
			return iterator(0);
#endif
		}

	iterator erase_dont_delete(iterator it)
		{
			if(it.n!=0 && it.n!=endobj())
			{
				Node *n=it.n->next;
				it.n->prev->next=n;
				n->prev=it.n->prev;
				--anz_elem;
#ifdef _DEBUG
				return iterator(n, endobj());
			}
			return iterator(0,0);
#else
				return iterator(n);
			}
			return iterator(0);
#endif
		}

	void clear()
		{
			for(Node *n=endobj()->next; n!=endobj() ; )
				{ Node *nach=n->next; delete_node(n); n=nach; }
			endobj()->next=endobj()->prev=endobj();
			anz_elem=0;
		}
	void push_back(const T& x=T()) { insert(end(),x); }
	void push_front(const T& x=T()) { insert(begin(),x); }

#ifdef _DEBUG
	void pop_back() { assert(0!=anz_elem); erase(iterator(endobj()->prev, endobj())); }
	void pop_front() { assert(0!=anz_elem); erase(iterator(endobj()->next, endobj())); }
#else
	void pop_back() { assert(0!=anz_elem); erase(iterator(endobj()->prev)); }
	void pop_front() { assert(0!=anz_elem); erase(iterator(endobj()->next)); }
#endif

	T& back() { assert(0!=anz_elem); return endobj()->prev->t; }
				// back() darf nicht auf eine leere Liste angewendet werden !
	T& front() { assert(0!=anz_elem); return endobj()->next->t; }
				// front() darf nicht auf eine leere Liste angewendet werden !
	const T& back() const { assert(0!=anz_elem); return endobj()->prev->t; }
				// back() darf nicht auf eine leere Liste angewendet werden !
	const T& front() const { assert(0!=anz_elem); return endobj()->next->t; }
				// front() darf nicht auf eine leere Liste angewendet werden !

	/*struct sort_LessEqual : public BaumKeineAttrBerechnen<iterator>
	{
		std::greater<T> gr;
		bool less(const iterator& t1, const iterator& t2) const { return gr(*t2, *t1); }
		bool equal(const iterator& t1, const iterator& t2) const { return !gr(*t1, *t2) && !gr(*t2, *t1); }
	};

	void sort(const std::greater<T>& gr)
	{
		RotSchwarzBaum<iterator, sort_LessEqual> b;
		iterator it;
		for(it=begin(); it!=end(); ++it)
			b.insert(it);
		it=begin();f
        typename RotSchwarzBaum<iterator, sort_LessEqual>::iterator it2;
		for(it2=b.begin(); it2!=0; ++it2)
		{
			iterator it3=insert(it, **it2);
			erase(*it2);
			it=it3;
			++it;
		}
	}*/

#ifdef _DEBUG
	inline iterator iter_from_node(Node *n) { return iterator(n, endobj()); }
#else
	inline iterator iter_from_node(Node *n) { return iterator(n); }
#endif
};

/////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////


template<class T>
class AutoSizeVector
{
	T* v;
	size_t dim,mem;
	void get_mem(size_t n)
	{
		if(0==v || n+1 > mem)
		{
			if(0==v) dim=mem=0;
			size_t n2=(n+1) + (n+1)/2;
			T* v2=new T[n2];
			size_t i;
			for(i=0; i<dim; i++)
				v2[i]=v[i];
			for( ; i<n2; i++)
				v2[i]=T();
			mem=n2;
			dim=n+1;
			if(v) delete[] v;
			v=v2;
		}
		else if(n+1 > dim)
			dim=n+1;
	}

public:
	~AutoSizeVector() { if(v) delete[] v; }

	AutoSizeVector() : v(0), dim(0), mem(0) { }

	AutoSizeVector(size_t X);
	AutoSizeVector(const T* CArr, size_t Y);
			// ACHTUNG: C_Matrix mu auf Y T-Werte zeigen !
	AutoSizeVector(const T& init_fuer_alle_elemente, size_t n);
	AutoSizeVector(const AutoSizeVector &m);
	AutoSizeVector& operator=(const AutoSizeVector& m);

	AutoSizeVector<T>& copy_block(size_t y, const AutoSizeVector<T>& m);
		// copy_block() ist identisch wie setze_block()
	AutoSizeVector<T>& setze_block(size_t y, const AutoSizeVector<T>& m);
		// setze_block() ist identisch wie copy_block()
	AutoSizeVector<T> read_block(size_t y0, size_t dy) const;
	void push_back(const T &x) { (*this)[dim]=x; }
	T& front() { if(0==dim) get_mem(0); return v[0]; }
	T& back() { if(0==dim) get_mem(0); return v[dim-1]; }

	size_t size() const { return dim; }

	T& operator() (size_t y)
			// Index-Operator im Fortran-Stil
		{
			get_mem(y);
			return v[y];
		}

	T& operator[] (size_t y)
		{
			get_mem(y);
			return v[y];
		}

	const T& operator[](size_t y) const
		{
			get_mem(y);
			return v[y];
		}

	void resize(size_t N);
	void delete_element(size_t n);
	void insert_element_before(size_t n);
	void insert_element_before(size_t n, const T &elem);

	T* array() { return v; }
	const T* array() const { return v; }
};

template<class T>
		AutoSizeVector<T>::AutoSizeVector(size_t N)
{
	dim=mem=N;
	v = new T[N];
	for(size_t i=0; i<dim; i++)
		v[i] = T();
}

template<class T>
		AutoSizeVector<T>::AutoSizeVector(const T& init_fuer_alle_elemente, size_t N)
{
	dim=N; mem=N;
	v = new T[dim];
	for(size_t i=0 ; i<dim ; i++)
		v[i] = init_fuer_alle_elemente;
}

template<class T>
		AutoSizeVector<T>::AutoSizeVector(const T* C_Array, size_t Y)
{
	dim=Y; mem=Y;
	v = new T[dim];
	if(C_Array)
	{
		for(size_t i=0; i<dim; i++)
			v[i]=C_Array[i];
	}
}

template<class T>
		AutoSizeVector<T>::AutoSizeVector(const AutoSizeVector& m)
{
	if(0==m.v)
	{
		v=0;
		dim=mem=0;
	}
	else
	{
		dim=mem=m.dim;
		v = new T[m.dim];
		for(size_t i=0; i<m.dim; i++) v[i]=m[i];
	}
}

template<class T>
		AutoSizeVector<T>& AutoSizeVector<T>::operator=(const AutoSizeVector& m)
{
	size_t i;
	if(this != &m)
	{
		if(0==v)
		{
			if(0!=m.v)
			{
				dim=mem=m.dim;
				v = new T[m.dim];
				for(i=0; i<m.dim; i++) v[i]=m[i];
			}
		}
		else if(0!=m.v)
		{
			delete[] v; v=0; dim=mem=0;
			v = new T[m.dim];
			dim=mem=m.dim;
			for(i=0; i<m.dim; i++) v[i]=m[i];
		}
		else
		{
			delete[] v;
			v=0;
			dim=mem=0;
		}
	}
	return *this;
}

template<class T>
void AutoSizeVector<T>::resize(size_t N)
{
	size_t i;
	if(0==v)
	{
		if(0!=N)
		{
			dim=mem=N;
			v = new T[N];
			for(i=0; i<N; i++)
				v[i]=T();
		}
	}
	else if(0!=N)
	{
		if(N != dim)
		{
			size_t dim_alt=dim;
			T* v_alt = v;
			dim=mem=N;
			v = new T[N];
			for(i=0; i<N && i<dim_alt; i++)
				v[i]=v_alt[i];
			for( ; i<N; i++)
				v[i]=T();
			delete[] v_alt;
		}
	}
	else
	{
		delete[] v;
		v=0;
		dim=mem=0;
	}
}

template<class T>
void AutoSizeVector<T>::delete_element(size_t n)
{
	size_t i;
	if(0!=v && n<dim)
	{
		for(i=n; i+1<dim; i++)
			v[i] = v[i+1];
		--dim;
	}
}

template<class T>
void AutoSizeVector<T>::insert_element_before(size_t n)
{
	size_t i;
	if(n<=dim)
	{
		get_mem(dim); // Hier wird dim inkrementiert.
		for(i=dim; --i>n; )
			v[i] = v[i-1];
		v[n]=T();
	}
}

template<class T>
void AutoSizeVector<T>::insert_element_before(size_t n, const T &elem)
{
	size_t i;
	if(n<=dim)
	{
		get_mem(dim); // Hier wird dim inkrementiert.
		for(i=dim; --i>n; )
			v[i] = v[i-1];
		v[n]=elem;
	}
}

template<class T>
AutoSizeVector<T>& AutoSizeVector<T>::setze_block(size_t y, const AutoSizeVector<T>& m)
	// setze_block() ist identisch wie copy_block()
{
	get_mem(y + m.size());
	for(size_t i=0; i<m.size(); i++)
		v[y+i] = m[i];
	return (*this);
}

template<class T>
AutoSizeVector<T>& AutoSizeVector<T>::copy_block(size_t y, const AutoSizeVector<T>& m)
	// copy_block() ist identisch wie setze_block()
{
	get_mem(y + m.size());
	for(size_t i=0; i<m.size(); i++)
		v[y+i] = m[i];
	return (*this);
}

template<class T>
AutoSizeVector<T> AutoSizeVector<T>::read_block(size_t y0, size_t dy) const
{
	if(0==v)
		return *this;
	long dy2=dy;
	if(y0 >= dim)
		return AutoSizeVector<T>();
	if(y0+dy > dim)
		dy2=dim-y0;

	AutoSizeVector<T> m(dy2);
	for(size_t i=0; i<dy2; i++)
		m(i)=v(y0+i);
	return m;
}

template<class T>
AutoSizeVector<T> duplicate_all(const AutoSizeVector<T>& v)
{		// Der Aufruf von duplicate_all() ist sinnvoll, wenn T ein Zeiger-Typ ist, d.h ein
		// C/C++-Zeiger oder ein count_ptr bzw. ein auto_ptr_list. Die Elemente des Vektors werden
		// ebenfalls mit Hilfe der Funktion duplicate_all() kopiert. Wenn T ein Zeiger-Typ ist,
		// werden die Speicherbereiche, auf die jeder Zeiger im Array zeigt, selbst kopiert, statt
		// nur die Zeiger zu kopieren.
	AutoSizeVector<T> v2(v.size());
	for(long i=0; i<v.size(); i++)
		v2[i]=duplicate_all(v[i]);
	return v2;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////

template<class X, class Y, class __X_LessEqualOPs=LessEqual<X> > class mapping;

template<class X, class Y, class __X_LessEqualOPs>
class mapping_XYComp : public BaumKeineAttrBerechnen<std::pair<X,Y> >
{
	typedef std::pair<X,Y> Pair;
	typedef RotSchwarzBaum<Pair, mapping_XYComp<X,Y,__X_LessEqualOPs> > baum_typ;
	typedef std::pair<X,Y> Pair;
	friend class mapping<X,Y,__X_LessEqualOPs>;
	friend class RotSchwarzBaum<Pair, mapping_XYComp<X,Y,__X_LessEqualOPs> >;
	__X_LessEqualOPs LessEqualOPs;
	mapping_XYComp() : LessEqualOPs(__X_LessEqualOPs()) { }
	mapping_XYComp(const mapping_XYComp<X,Y,__X_LessEqualOPs>& m) : LessEqualOPs(m.LessEqualOPs) { }
	mapping_XYComp<X,Y,__X_LessEqualOPs>& operator=(const mapping_XYComp<X,Y,__X_LessEqualOPs>& m)
			{ LessEqualOPs=m.LessEqualOPs; return *this; }
	bool less(const Pair& p1, const Pair& p2) const
			{ return LessEqualOPs.less(p1.first, p2.first); }
	bool equal(const Pair& p1, const Pair& p2) const
			{ return LessEqualOPs.equal(p1.first, p2.first); }
};

template<class X, class Y, class __X_LessEqualOPs >
class mapping
{
	typedef std::pair<X,Y> Pair;
	mapping_XYComp<X,Y,__X_LessEqualOPs> xycomp;
	typedef RotSchwarzBaum<Pair, mapping_XYComp<X,Y,__X_LessEqualOPs> > baum_typ;
	baum_typ baum;
	static Y dummy_y;

public:
	typedef typename baum_typ::iterator iterator;
	typedef typename baum_typ::const_iterator const_iterator;

	mapping() { }
	mapping(const mapping& mp) : xycomp(mp.xycomp), baum(mp.baum) { }
	mapping& operator=(const mapping& mp) { xycomp=mp.xycomp; baum=mp.baum; return *this; }
	Y& operator[](const X& x)
		{
			typename baum_typ::iterator it;
			if((it=baum.suche(Pair(x,Y()))) == 0)
				it=baum.insert(Pair(x,Y()));
			return it->second;
		}
	const Y& operator[](const X& x) const
		{
			typename baum_typ::iterator it;
			if(!((it=baum.suche(Pair(x,Y()))) == 0))
				return it->second;
			else
				return dummy_y;
		}
	iterator find(const X& x) { return baum.suche(Pair(x,Y())); }
	const_iterator find(const X& x) const { return baum.suche(Pair(x,Y())); }
	iterator suche_naechstes_nichtgroesseres(const X& x) { return baum.suche_naechstes_nichtgroesseres(Pair(x,Y())); }
	size_t size() const  { return baum.size(); }
	std::pair<iterator,bool> insert(const X& x)
		{
			iterator it=baum.suche(x);
			return it==0 ? (std::make_pair(baum.insert(x)),true) : Pair(it,false);
		}
	void erase(iterator it) { baum.erase(it); }
	size_t erase(const X& x)
		{
			iterator it=baum.suche(x);
			return (it==0) ? 0 : (baum.erase(it), 1);
		}
	void clear() { baum.clear(); }
	iterator begin() { return baum.begin(); }
	iterator end() { return baum.end(); }
	const_iterator begin() const { return baum.begin(); }
	const_iterator end() const { return baum.end(); }
};

template<class X, class Y, class __X_LessEqualOPs> Y mapping<X,Y,__X_LessEqualOPs>::dummy_y;

/////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////

template<class T, class __LessEqualOPs=LessEqual<T> >
class ListAndTree : private nlist<RotSchwarzBaum_Node<T> >
{	// private-Ableitung erforderlich, um zu verhindern, dass der Benutzer von ListAndTree direkt auf
	// nlist<RotSchwarzBaum_Node<T> > zugreift, insbesondere, falls diese Klasse erweitert wird, ohne die entsprechenden
	// Funktionen durch ListAndTree zu berschreiben.
		// ListAndTree verhlt sich wie eine nlist<T>, mit der Erweiterung, dass
		// eine Suchfunktion find() mit einer Zeit zur Verfgung steht, die proportional zum Logarithmus aus
		// der Anzahl der Listenelemente ist.
	typedef RotSchwarzBaum_Node<T> RSNode;
	typedef nlist<RotSchwarzBaum_Node<T> > lbase;
	typedef RotSchwarzBaum<T, __LessEqualOPs> rsbase;
	typedef nlist_Node<RotSchwarzBaum_Node<T> > Node;

	private: void sort(const std::greater<T>& gr) { } // Zugriff verhindern. Statt dessen: parameterloses sort() !

	rsbase baum;

	typedef typename lbase::iterator lbase_iterator; // Visual C++ gewhrt aufgrund der private-Ableitung ListAndTree : private nlist<RotSchwarzBaum_Node<T> >
		// nur Zugriff auf lbase_iterator, nicht auf lbase::iterator
	typedef typename lbase::const_iterator lbase_const_iterator;

	class diterator : public lbase_iterator
		{		// Der abgeleitete iterator muss hier zunchst anders als iterator genannt werden,
				// weil sonst Visual C++ beim Aufruf lbase::iterator::operator=(it); statt den Basis-operator=()
				// den operator=() dieser Klasse diterator aufruft, was zu einem Stackberlauf fhrt. Unten
				// wird diterator per typedef in iterator umgenannt.
			typedef lbase_iterator baseiter;
			friend class ListAndTree;
			inline Node *get_node() const { return baseiter::get_node(); } // stellt sicher,
				// dass lbase::iterator::get_node() verfgbar ist (keine andere berschreibung).
		public:
			diterator(baseiter it) : baseiter(it) { }
			diterator() { }
			diterator(const diterator &it) : baseiter(it) { }
			diterator& operator=(const diterator &it) { baseiter::operator=(it); return *this; }
			T& operator*() const { return get_node()->t.t; }
			T* operator->() const { return &(get_node()->t.t); }
			diterator operator++() { return baseiter::operator++(); }
			diterator operator++(int dummy) { return baseiter::operator++(dummy); }
			diterator operator--() { return baseiter::operator--(); }
			diterator operator--(int dummy) { return baseiter::operator--(dummy); }
			bool is_valid_and_not_end(const ListAndTree& lst) const { return baseiter::is_valid_and_not_end(lst); }
		};

	class dconst_iterator : public lbase_const_iterator
		{
			typedef lbase_const_iterator baseciter;
			friend class ListAndTree;
			inline const Node *get_node() const { return baseciter::get_node(); } // stellt sicher,
				// dass lbase::iterator::get_node() verfgbar ist (keine andere berschreibung).
		public:
			// ACHTUNG: lbase::iterator::get_node() darf nicht berladen werden, weil es unten benutzt wird !
			dconst_iterator(baseciter it) : baseciter(it) { }
			dconst_iterator(lbase_iterator it) : baseciter(it) { }
			dconst_iterator() { }
			dconst_iterator(const dconst_iterator &it) : baseciter(it) { }
			//dconst_iterator(const iterator &it) : baseciter(it) { }
			dconst_iterator& operator=(const dconst_iterator &it) { baseciter::operator=(it); return *this; }
			//dconst_iterator& operator=(const iterator &it) { baseciter::operator=(it); return *this; }
			const T& operator*() const { return get_node()->t.t; }
			const T* operator->() const { return &(get_node()->t.t); }
			dconst_iterator operator++() { return baseciter::operator++(); }
			dconst_iterator operator++(int dummy) { return baseciter::operator++(dummy); }
			dconst_iterator operator--() { return baseciter::operator--(); }
			dconst_iterator operator--(int dummy) { return baseciter::operator--(dummy); }
			bool is_valid_and_not_end(const ListAndTree& lst) const { return baseciter::is_valid_and_not_end(lst); }
		};

public:

	typedef diterator iterator;
	typedef dconst_iterator const_iterator;

	typedef typename RotSchwarzBaum<T, __LessEqualOPs>::iterator tree_iterator;
	typedef typename RotSchwarzBaum<T, __LessEqualOPs>::const_iterator tree_const_iterator;
	tree_iterator tree_begin() { return baum.minimizer(); }
	tree_iterator tree_minimizer() { return baum.minimizer(); }
	tree_iterator tree_maximizer() { return baum.maximizer(); }
	iterator minimizer() { return iter_from_tree_iter(baum.minimizer()); }
	iterator maximizer() { return iter_from_tree_iter(baum.maximizer()); }

	ListAndTree(__LessEqualOPs _less_equal=__LessEqualOPs()) : baum(_less_equal) { }
	ListAndTree(const ListAndTree<T, __LessEqualOPs>& lt) : baum(lt.baum.get_LessEqualOPs())
		{
			for(const_iterator it=lt.begin(); it!=lt.end(); it++)
				insert(lbase::end(), *it);
		}
	ListAndTree& operator=(const ListAndTree<T, __LessEqualOPs>& lt)
		{
			if(this!=&lt)
			{
				clear();
				baum=rsbase(lt.baum.get_LessEqualOPs());
 				for(const_iterator it=lt.begin(); it!=lt.end(); it++)
					insert(lbase::end(), *it);
			}
			return *this;
		}
	~ListAndTree()
		{ clear(); }
	void clear()
		{
			for(iterator it=begin(); it!=end(); )
				{ iterator it2=it; ++it; erase(it2); }
		}

	inline iterator begin()
		{ return lbase::begin(); }
	inline const_iterator begin() const
		{ return lbase::begin(); }
	inline iterator end()
		{ return lbase::end(); }
	inline const_iterator end() const
		{ return lbase::end(); }

	iterator insert(const_iterator it, const T& x=T())
		{
			RSNode rsn; rsn.t=x;
			typename lbase::iterator it2=lbase::insert(it, rsn);
			baum.insert_inplace(rsbase::const_iterator::from_node(&(*it2)));
			return iterator(it2);
		}
	void insert(const_iterator it, const_iterator list2_first, const_iterator list2_last)
		{
			while(list2_first!=list2_last)
				insert(it, *(list2_first++));
		}
	iterator insert_inplace(const_iterator it, iterator obj)
		{
			typename lbase::iterator it2=lbase::insert_inplace(it, obj);
			baum.insert_inplace(rsbase::const_iterator::from_node(&(*it2)));
			return iterator(it2);			
		}

	iterator erase(iterator it)
		{
			baum.erase_dont_delete(tree_iterator::from_node(&(it.get_node()->t)));
			return lbase::erase(it);
		}
	iterator erase_dont_delete(iterator it)
		{
			baum.erase_dont_delete(tree_iterator::from_node(&(it.get_node()->t)));
			return lbase::erase_dont_delete(it);
		}

	void push_front(const T& x=T())
		{ insert(begin(), x); }
	void push_back(const T& x=T())
		{ insert(end(), x); }
	void pop_front()
		{ erase(begin()); }
	void pop_back()
		{ erase(--end()); }
	T& back() { return lbase::back().t; }
	T& front() { return lbase::front().t; }
	const T& back() const { return lbase::back().t; }
	const T& front() const { return lbase::front().t; }
	size_t size() const { return lbase::size(); }

	iterator iter_from_tree_iter(tree_iterator it)
		{	// Wenn it==0 ist, wird 0 und NICHT end() geliefert !
			// Im Folgenden muss it==0 statt it.get_base_node()==0 abgefregt werden, weil im RotSchwarzBaum
			// der Nulliterator durch nil() dargestellt wird.
			return (it==0) ? iterator(0) :
					iterator(lbase::iter_from_node(Node::node_address_from_t_address(it.get_node())));
		}
	iterator find(const T &x) // Liefert iterator(0), wenn nicht gefunden.
		{
			return iter_from_tree_iter(baum.suche_minus_iter_bei_gleichen(x));
		}
	void sort()
		{
			typename lbase::iterator it=lbase::begin();
			for(typename rsbase::iterator it2=baum.begin(); it2!=0; ++it2)
			{
				typename lbase::iterator it3=iter_from_tree_iter(it2);
				if(it3!=it)
				{
					lbase::erase_dont_delete(it3);
					it=lbase::insert_inplace(it, it3);
				}
				++it;
			}
		}
};

///////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
/////  SumList0 : Jedes Listenelement enthlt nur die Listenposition und
/////             zustzlichen KEINEN Summanden
/////  (SumList1, SumList2): ... ein bzw. zwei Summanden
/////  SumList0, SumList1 und SumList2 stellen Iteratoren zur Verfgung,
/////     in denen zustzlich die Listenposition abgespeichert ist.
////////////////////////////////////////////////////////////////////////


template<class T, class Allocator> class SumList0;

template<class T, class Allocator = mem_allocator<nlist_Node<T> > >
class SumList0
{
	typedef nlist<T, Allocator> base;
	typedef SumList0<T, Allocator> clss;

public:
	class iterator;
	class const_iterator;
	friend class iterator;
	friend class const_iterator;

private:
	base *lst;
	bool new_reserved;
	nlist<iterator*> iters;
	nlist<const_iterator*> const_iters;
	long max_anzahl_markers;

public:
	class iterator
	{
		typename nlist<iterator*>::iterator iters_iter;
		clss *lst;
		typename base::iterator it;
		long listpos;
		friend class SumList0<T, Allocator>;
	public:
		iterator(long dummy=0) { it=0; listpos=-1; lst=0; iters_iter=0; }
			// damit der Wert 0 explizit bergeben werden kann.
		iterator(const iterator &i) : it(i.it), listpos(i.listpos), lst(i.lst)
				{ if(lst!=0 && it!=0) { iters_iter=lst->iters.insert(lst->iters.end(), this); } }
		iterator& operator=(const iterator &i)
			{ if(this!=&i) { if(lst!=i.lst) { if(0!=lst && iters_iter!=0) lst->iters.erase(iters_iter);
								if(i.lst!=0) iters_iter=i.lst->iters.insert(i.lst->iters.end(), this);
									else iters_iter=0; }
							 it=i.it; listpos=i.listpos; lst=i.lst; }
				return *this; } // iters_iter darf nicht zugewiesen werden, wenn es sich nicht
								// um einen uninitialisierten Iterator handelt, denn der
								// Speicherort this, welcher in lst->iters abgespeichert wird,
								// bleibt unverndert.
		iterator& operator=(long dummy)
					{ if(0!=lst && iters_iter!=0) lst->iters.erase(iters_iter);
						iters_iter=0; it=0; listpos=-1; lst=0; return *this; }
		~iterator() { if(iters_iter!=0 && lst!=0) lst->iters.erase(iters_iter); }

		const T& operator*() const
			{ assert(it!=0 && lst!=0 && it!=lst->lst->end()); return *it; }
		T& operator*()
			{ assert(it!=0 && lst!=0 && it!=lst->lst->end()); return *it; }
		const T* operator->() const
			{ assert(it!=0 && lst!=0 && it!=lst->lst->end()); return &(*it); }
		T* operator->()
			{ assert(it!=0 && lst!=0 && it!=lst->lst->end()); return &(*it); }
		iterator operator++() // prefix ++
			{ if(it!=0 && lst!=0 && it!=lst->lst->end())
				{ ++it; ++listpos; } return *this; }
		iterator operator++(int) // postfix ++
			{ iterator save=*this;
				if(it!=0 && lst!=0 && it!=lst->lst->end()) { ++it; ++listpos; }
					return save; }
		iterator operator--()
			{ if(it!=0 && lst!=0 && it!=lst->lst->begin())
				{ --it; --listpos; } return *this; }
		iterator operator--(int)
			{ iterator save=*this;
				if(it!=0 && lst!=0 && it!=lst->lst->begin())
					{ --it; --listpos; } return save; }
		bool operator==(const iterator &i) const
			{ return lst==i.lst && it==i.it; }
		bool operator==(const long x) const { return x==0 && it==0; }
		bool operator!=(const iterator &i) const
			{ return lst!=i.lst || it!=i.it; }
		bool operator!=(const long x) const { return x!=0 || it!=0; }
		bool is_valid_and_not_end() const { return it!=0 && lst!=0 && it!=lst->lst->end(); }
		bool is_end() const { return it!=0 && lst!=0 && it==lst->lst->end(); }
		bool is_begin() const { return it!=0 && lst!=0 && it==lst->lst->begin(); }
		bool is_uninitialized() const { return it==0 || lst==0; }
		long get_listpos() const { if(it!=0) return listpos; else return -1; }
		SumList0<T, Allocator> *get_list() { return lst; }
		typename base::iterator get_list_iterator() { return it; }
	};

	class const_iterator
	{
		typename nlist<const_iterator*>::iterator iters_iter;
		clss *lst;
		typename base::const_iterator it;
		long listpos;
		friend class SumList0<T, Allocator>;
		friend class iterator;
	public:
		const_iterator(long dummy=0) { it=0; listpos=-1; lst=0; iters_iter=0; }
		const_iterator(const const_iterator &i) : it(i.it), listpos(i.listpos), lst(i.lst)
				{ if(lst!=0 && it!=0) { iters_iter=lst->const_iters.insert(lst->const_iters.end(), this); } }
		const_iterator(const iterator &i) : it(i.it), listpos(i.listpos), lst(i.lst)
				{ if(lst!=0 && it!=0)
					{ iters_iter=lst->const_iters.insert(lst->const_iters.end(), this); } }
		const_iterator& operator=(const const_iterator &i)
			{ if(this!=&i) { if(lst!=i.lst) { if(0!=lst && iters_iter!=0) lst->const_iters.erase(iters_iter);
								if(i.lst!=0) iters_iter=i.lst->const_iters.insert(i.lst->const_iters.end(), this);
									else iters_iter=0; }
							 it=i.it; listpos=i.listpos; lst=i.lst; }
				return *this; } // iters_iter darf nicht zugewiesen werden, wenn es sich nicht
								// um einen uninitialisierten Iterator handelt, denn der
								// Speicherort this, welcher in lst->const_iters abgespeichert wird,
								// bleibt unverndert.
		const_iterator& operator=(long dummy)
					{ if(0!=lst && iters_iter!=0) lst->iters.erase(iters_iter);
						iters_iter=0; it=0; listpos=-1; lst=0; return *this; }
		~const_iterator() { if(iters_iter!=0 && lst!=0) lst->const_iters.erase(iters_iter); }

		T operator*() const
			{ assert(it!=0 && lst!=0 && it!=lst->lst->end()); return *it; }
		const T* operator->() const
			{ assert(it!=0 && lst!=0 && it!=lst->lst->end()); return &(*it); }
		const_iterator operator++() // prefix ++
			{ if(it!=0 && lst!=0 && it!=lst->lst->end())
				{ ++it; ++listpos; } return *this; }
		const_iterator operator++(int) // postfix ++
			{ const_iterator save=*this;
				if(it!=0 && lst!=0 && it!=lst->lst->end()) { ++it; ++listpos; }
					return save; }
		const_iterator operator--()
			{ if(it!=0 && lst!=0 && it!=lst->lst->begin())
				{ --it; --listpos; } return *this; }
		const_iterator operator--(int)
			{ const_iterator save=*this;
				if(it!=0 && lst!=0 && it!=lst->lst->begin())
					{ --it; --listpos; } return save; }
		bool operator==(const iterator &i) const { return lst==i.lst && it==i.it; }
		bool operator==(const const_iterator &i) const { return lst==i.lst && it==i.it; }
		bool operator==(const long x) const { return x==0 && it==0; }
		bool operator!=(const iterator &i) const { return lst!=i.lst || it!=i.it; }
		bool operator!=(const const_iterator &i) const { return lst!=i.lst || it!=i.it; }
		bool operator!=(const long x) const { return x!=0 || it!=0; }
		bool is_valid_and_not_end() const { return it!=0 && lst!=0 && it!=lst->lst->end(); }
		bool is_end() const { return it!=0 && lst!=0 && it==lst->lst->end(); }
		bool is_begin() const { return it!=0 && lst!=0 && it==lst->lst->begin(); }
		bool is_uninitialized() const { return it==0 || lst==0; }
		long get_listpos() const { if(it!=0) return listpos; else return -1; }
		const SumList0<T, Allocator> *get_list() const { return lst; }
		typename base::iterator get_list_iterator() { return it; }
	};

private:
	nlist<iterator> marker;
			// Es wird vorausgesetzt, dass die Marker-Liste Iteratoren von
			// Listenelementen mit nicht absteigenden Positionsnummern enthlt.
	iterator _end_iter;

public:
	struct MaxAnzahlMarkers { long anz; MaxAnzahlMarkers(long n) : anz(n) { } };
	struct OnlyReferenceDontCopy
			{ base& l; OnlyReferenceDontCopy(nlist<T, Allocator> &_lst) : l(_lst) { } };
	SumList0(MaxAnzahlMarkers max_mark, const Allocator &_A = Allocator())
				: lst(new base(_A)), iters(), const_iters(), marker()
			{ max_anzahl_markers=max_mark.anz; _end_iter=begin(); new_reserved=true; }
	SumList0(Allocator& _A = Allocator())
				: lst(new base(_A)), iters(), const_iters(), marker()
					{ max_anzahl_markers=3; _end_iter=begin(); new_reserved=true; }
	SumList0(const nlist<T, Allocator>& _lst)
				: lst(new base(_lst)), new_reserved(true),
					iters(), const_iters(), marker(), max_anzahl_markers(_lst.max_anzahl_markers)
			{ _end_iter.lst=this; _end_iter.it=lst->end(); _end_iter.listpos=lst->size();
				_end_iter.iters_iter=iters.insert(iters.end(), &_end_iter); }
	SumList0(OnlyReferenceDontCopy _lst)
				: lst(&(_lst.l)), new_reserved(false), iters(), const_iters(),
					marker(), max_anzahl_markers(3)
			{ _end_iter.lst=this; _end_iter.it=lst->end(); _end_iter.listpos=lst->size();
				_end_iter.iters_iter=iters.insert(iters.end(), &_end_iter); }
	SumList0(OnlyReferenceDontCopy _lst, MaxAnzahlMarkers max_mark)
				: lst(_lst.l), new_reserved(false), iters(), const_iters(),
					marker(), max_anzahl_markers(max_mark.anz)
			{ _end_iter.lst=this; _end_iter.it=lst->end(); _end_iter.listpos=lst->size();
				_end_iter.iters_iter=iters.insert(iters.end(), &_end_iter); }
	SumList0(const SumList0 &l)
				: lst(l.new_reserved ? new base(*(l.lst)) : l.lst),
					new_reserved(l.new_reserved),
					iters(), const_iters(), marker(),  // iters und marker nicht kopieren !
					max_anzahl_markers(l.max_anzahl_markers)
			{ _end_iter.lst=this; _end_iter.it=lst->end(); _end_iter.listpos=lst->size();
				_end_iter.iters_iter=iters.insert(iters.end(), &_end_iter); }
	SumList0& operator=(const SumList0& l)
			{ if(this!=&l) {
					for(typename nlist<iterator*>::iterator it=iters.begin(); it!=iters.end(); it++)
						{ (*it)->lst=0; (*it)->it=0; (*it)->iters_iter=0; (*it)->listpos=-1; }
					for(typename nlist<const_iterator*>::iterator it2=const_iters.begin(); it2!=const_iters.end(); it2++)
						{ (*it2)->lst=0; (*it2)->it=0; (*it2)->iters_iter=0; (*it2)->listpos=-1; }
					if(lst && new_reserved) delete lst;
					lst=l.new_reserved ? new base(*(l.lst)) : l.lst;
					new_reserved=l.new_reserved;
					iters.clear(); const_iters.clear();
												// iters nicht von l zuweisen !
					marker.clear(); max_anzahl_markers=l.max_anzahl_markers;
					_end_iter.lst=this; _end_iter.it=l.lst->end(); _end_iter.listpos=l.lst->size();
					_end_iter.iters_iter=iters.insert(iters.end(), &_end_iter);
			} return *this; }
	~SumList0() {
			for(typename nlist<iterator*>::iterator it=iters.begin(); it!=iters.end(); it++)
				{ (*it)->lst=0; (*it)->it=0; (*it)->iters_iter=0; (*it)->listpos=-1; }
			for(typename nlist<const_iterator*>::iterator it2=const_iters.begin(); it2!=const_iters.end(); it2++)
				{ (*it2)->lst=0; (*it2)->it=0; (*it2)->iters_iter=0; (*it2)->listpos=-1; }
			if(lst && new_reserved) delete lst;
		}

	iterator begin()
		{ iterator it; it.it=lst->begin(); it.listpos=0; it.lst=this;
			it.iters_iter=iters.insert(iters.end(), &it); return it; }
	const_iterator begin() const
		{ const_iterator it; it.it=base::begin(); it.listpos=0;
			it.lst=const_cast<SumList0<T> >(this);
			it.iters_iter=(const_cast<SumList0<T>*>(this)->const_iters).insert(
				(const_cast<SumList0<T>*>(this)->const_iters).end(), &it); return it; }
	iterator end() { return _end_iter; }
	const_iterator end() const { return _end_iter; }

	bool check_is_iterator_from_list(const iterator &it) { return it.lst==this; }
	bool check_is_iterator_from_list(const const_iterator &it) { return it.lst==this; }

	bool check_is_itr_from_list_and_not_end(const iterator &it)
			{ return it->lst==this && it->it!=0 && it->it!=base::end(); }
	bool check_is_itr_from_list_and_not_end(const const_iterator &it)
			{ return it->lst==this && it->it!=0 && it->it!=base::end(); }

	long size() const { return lst->size(); }
	iterator insert(iterator it, const T& x=T())
		{ if(it.lst==this && it.it!=0 && it.iters_iter!=0)
		  {
			iterator it2(it);
			it2.it=lst->insert(it.it, x);
			long pos=it2.listpos; // ACHTUNG: it und it2 werden in der folgenden Schleife ebenfalls
								  // verndert, weil sie in der Liste iters stehen !
			for(typename nlist<iterator*>::iterator it3=iters.begin(); it3!=iters.end(); it3++)
				if((*it3)->listpos>=pos) ++(*it3)->listpos;
			for(typename nlist<const_iterator*>::iterator it4=const_iters.begin();
					it4!=const_iters.end(); it4++)
				if((*it4)->listpos>=pos) ++(*it4)->listpos;
			--it2.listpos; // Diese Anweisungen knnen nicht ersetzt werden,
					// indem in der Schleife auf (*it3)->listpos>pos abgefragt wird !
			return it2;
		  }
		  return iterator();
		}
	iterator erase(iterator it)
		{ if(it.lst==this && it.it!=0 && it.iters_iter!=0)
		  {
			if(it.it==lst->end()) return it;
			typename nlist<iterator>::iterator iNach, i2;
			for(typename nlist<iterator>::iterator i=marker.begin(); i!=marker.end(); i=iNach)
				if(iNach=i, ++iNach, it==*i)
				{		// Es wird vorausgesetzt, dass die Marker-Liste Iteratoren von
						// Listenelementen mit nicht absteigenden Positionsnummern enthlt.
					if(iNach==marker.end())
					{
						if(i==marker.begin())
						{
							iterator it2;
							if(it2=*i, ++it2.it!=lst->end()) ++(*i);
							else if(i->it!=lst->begin()) --(*i);
							else marker.erase(i);
						}
						else if(i2=i, *(--i2)==*i)
							marker.erase(i);
						else --(*i);
					}
					else if(*iNach==*i)
						marker.erase(i);
					else ++(*i);
				}
			iterator it2(it);
			++it2.listpos;  // wird in der Schleife unten wieder dekrementiert.
			it2.it=lst->erase(it.it);
			long pos=it.listpos; // ACHTUNG: it und it2 werden in der folgenden Schleife ebenfalls
								  // verndert, weil sie in der Liste iters stehen !
			it.it=0; it.lst=0; // damit der Destruktor von it nicht auf it.it zugreift.
			typename nlist<iterator*>::iterator it3Nach;
			for(typename nlist<iterator*>::iterator it3=iters.begin(); it3!=iters.end(); it3=it3Nach)
			{
				it3Nach=it3; ++it3Nach;
				if((*it3)->listpos==pos) { (*it3)->lst=0; (*it3)->it=0; iters.erase(it3); }
				else if((*it3)->listpos>pos) --(*it3)->listpos;
			}
			typename nlist<const_iterator*>::iterator it4Nach;
			for(typename nlist<const_iterator*>::iterator it4=const_iters.begin(); it4!=const_iters.end(); it4=it4Nach)
			{
				it4Nach=it4; ++it4Nach;
				if((*it4)->listpos==pos) { (*it4)->lst=0; (*it4)->it=0; const_iters.erase(it4); }
				else if((*it4)->listpos>pos) --(*it4)->listpos;
			}
			return it2;
		  }
		  return iterator();
		}
	void clear()
	{
		for(typename nlist<iterator*>::iterator it=iters.begin(); it!=iters.end(); it++)
			(*it)->lst=0;
		for(typename nlist<const_iterator*>::iterator it2=const_iters.begin(); it2!=const_iters.end(); it2++)
			(*it2)->lst=0;
		_end_iter.lst=0;  // Diese lst=0 Anweisungen sind notwendig, weil beim folgenden
						  // iters.clear() die Destruktoren der Iteratoren aufgerufen werden. Durch
						  // lst=0 knnen die Destruktoren keinen Schaden anrichten.
		marker.clear(); iters.clear(); const_iters.clear();
		lst->clear(); _end_iter=begin();
	}
	void push_back(const T& x=T()) { insert(end(),x); }
	void push_front(const T& x=T()) { insert(begin(),x); }
	void pop_back() { if(lst->begin()!=lst->end()) erase(--end()); }
	void pop_front() { if(lst->begin()!=lst->end()) erase(begin()); }
	T& back() { assert(lst->begin()!=lst->end()); return *(--(lst->end())); }
	T& front() { assert(lst->begin()!=lst->end()); return *(lst->begin()); }
	iterator find_iter_from_listpos(long listpos)
		{
			if(listpos<0 || listpos>=lst->size())
				return listpos==lst->size() ? _end_iter : iterator();
			typename nlist<iterator>::iterator it, it2;
			if(marker.size()<2)
			{
				if(marker.size()==0)
				{
					it=marker.insert(marker.end(), _end_iter);
					while(it->listpos > listpos && (*it).it!=lst->begin()) --(*it);
					return *it;
				}
				// Hier marker.size()==1
				if(listpos<(it=marker.begin())->listpos)
				{
					if(marker.size()<max_anzahl_markers)
						it=marker.insert(it, *it);
					while(listpos < it->listpos && (*it).it!=lst->begin()) --(*it);
				}
				else
				{
					if(marker.size()<max_anzahl_markers)
						marker.insert(it, *it);					
					while(listpos > it->listpos && (*it).it!=lst->end()) ++(*it);
				}
				return *it;
			}
			else for(it=marker.begin(); (it2=it, ++it2!=marker.end()) ; it=it2)
					if(*it==*it2) marker.erase(it);

			if(lst->size()==listpos) return _end_iter;
			if(0==listpos) return begin();
			for(it=marker.begin(); it!=marker.end() && it->listpos<listpos; it++) ;
			bool rueck=false;
			if(it==marker.end())
			{
				--it;  // marker.size()==0 schon oben abgefangen
				if(marker.size()<max_anzahl_markers)
					{ iterator i=*(it++); it=marker.insert(it, i); }  // ++it nicht in insert() !!!
			}
			else
			{
				if(it->listpos==listpos) return *it;
				if(it==marker.begin())
				{
					rueck=true;
					if(marker.size()<max_anzahl_markers)
						it=marker.insert(it, *it);
				}
				else if((--it)->listpos==listpos)
					return *it;
				else if(2*listpos > it->listpos + (it2=it, ++it2)->listpos)
				{
					++it; rueck=true;
					if(marker.size()<max_anzahl_markers)
						it=marker.insert(it, *it);
				}
				else if(marker.size()<max_anzahl_markers)
					{ iterator i=*(it++); it=marker.insert(it, i); }  // ++it nicht in insert() !!!
			}

			//print_marker();
			//TRACE("listpos=%d, soll=%d", it->listpos, listpos);
			if(rueck)
			{
				while(it->listpos > listpos && (*it).it!=lst->begin()) --(*it);
				return *it;
			}
			else
			{
				while(it->listpos < listpos && (*it).it!=lst->end()) ++(*it);
				return *it;
			}
		}
	void reserve_n_markers_keep_m_in_intervall(long n, long itv_posstart, long itv_posend, long m=1)
			// itv_posend ist ausschlielich. Diese Funktion lscht ggf. soviele Marker, dass
			// die Summe der im Intervall [itv_posstart, itv_posend[ befindlichen Marker und der
			//		max_anzahl_markers - marker.size()
			// noch freien Marker mindestens gleich n ist (falls n nicht grer als max_anzahl_markers
			// ist). Falls sich bereits m oder mehr Marker im Intervall befinden, werden zunchst
			// Marker aus diesem Intervall gelscht, bis nur noch m Marker in diesem Intervall sind.
			// Danach werden -- falls ntig -- weitere Marker auerhalb des Intervalls gelscht.
			// Maximal m der im Intervall befindlichen Marker zhlen als "reservierte Marker".
			// Wenn also m2<=m Marker im Invervall verbleiben, werden weitere Marker nur so lange
			// gelscht, bis
			//      max_anzahl_markers - marker.size() + m2 >= n
			// ist.
	{
		typename nlist<iterator>::iterator it, it2, itNach;
		long anz=0, res=max_anzahl_markers-marker.size();
		for(it=marker.begin(); it!=marker.end() && it->listpos<itv_posstart; it++) ;
		for(it2=it; it2!=marker.end() && it2->listpos<itv_posend; it2++) ++anz;
		if(anz>=m) n-=m; else n-=anz;
		for(it2=it; it2!=marker.end() && it2->listpos<itv_posend && anz>m && res<n; it2=itNach)
			{ itNach=it2; ++itNach; marker.erase(it2); --anz; ++res; }
		for(it=marker.begin(); it!=marker.end() && res<n && it->listpos<itv_posstart; it=itNach)
			{ itNach=it; ++itNach; marker.erase(it); ++res; }
		if(res<n)
		{
			for( ; it!=marker.end() && it->listpos<itv_posend; it++) ;
			for( ; it!=marker.end() && res<n; it=itNach)
				{ itNach=it; ++itNach; marker.erase(it); ++res; }
		}
	}
	iterator find_iter_from_listpos2(long listpos)
	{		// INEFFIZIENT, nur zur Kontrolle von find_iter_from_listpos()
		long c=listpos;
		if(listpos<0 || listpos>lst->size()) return iterator();
		typename base::iterator it;
		for(it=lst->begin(); it!=lst->end() && --c>=0; it++) ;
		iterator it2;
		it2.it=it; it2.listpos=listpos; it2.lst=this;
		it2.iters_iter=iters.insert(iters.end(), &it2);
		return it2;
	}
};

template<class T> class SumList;

template<class T>
class SumList_TExt
{
	friend class SumList<T>;
	SumList_TExt(const T &_t=T(), long _to_add1=0)
		: t(_t), to_add1(_to_add1), anz_elem(0), psum1(0) { }

public:
	T t;
	long to_add1, psum1, anz_elem;
	SumList_TExt(const SumList_TExt &l) : t(l.t), to_add1(l.to_add1),
		psum1(l.psum1), anz_elem(l.anz_elem) { }
	SumList_TExt& operator=(const SumList_TExt &l)
		{ if(this!=&l) { t=l.t; to_add1=l.to_add1; psum1=l.psum1; anz_elem=l.anz_elem; } return *this; }
};

template<class T>
struct SumList_Compare
{
	typedef SumList_TExt<T> X;
	bool less(const X& x, const X& y) const { return true; } // wird nicht benutzt
	bool equal(const X& x, const X& y) const { return true; } // wird nicht benutzt

	enum { soll_attribut_berechnen=1 };
	void calculate_attribute(X& akt) { akt.psum1=akt.to_add1; akt.anz_elem=1; }
	void calculate_attribute(const X& links, const X& rechts, X& akt)
		{ akt.psum1=akt.to_add1+links.psum1+rechts.psum1; akt.anz_elem=1+links.anz_elem+rechts.anz_elem; }
	void calculate_attribute1(const X& links, X& akt)
		{ akt.psum1=akt.to_add1+links.psum1; akt.anz_elem=1+links.anz_elem; }
	void calculate_attribute2(const X& rechts, X& akt)
		{ akt.psum1=akt.to_add1+rechts.psum1; akt.anz_elem=1+rechts.anz_elem; }
};

template<class T>
class SumList : public RotSchwarzBaum<SumList_TExt<T>, SumList_Compare<T> >
{
	typedef SumList_TExt<T> TExt;
	typedef RotSchwarzBaum<SumList_TExt<T>, SumList_Compare<T> > base;
	typedef SumList<T> clss;

public:
	class iterator;
	class const_iterator;
	friend class iterator;
	friend class const_iterator;

	class iterator
	{		// iterator==0 bei lst=0, iterator==end() bei lst!=0, it=0
		clss *lst;
		typename base::iterator it;
		friend class SumList<T>;
		friend class const_iterator;
		iterator(clss *_lst) : lst(_lst), it(0) { }
		iterator(clss *_lst, typename base::iterator _it) : lst(_lst), it(_it) { }
	public:
		iterator(long dummy) { it=0; lst=0; }
		iterator(int dummy=0) { it=0; lst=0; }
			// damit der Wert 0 explizit bergeben werden kann.
		iterator(const iterator &i) : lst(i.lst), it(i.it) { }
		iterator& operator=(const iterator &i) { lst=i.lst; it=i.it; return *this; }
		iterator& operator=(long dummy) { it=0; lst=0; return *this; }


		const T& operator*() const
			{ assert(it!=0 && lst!=0); return (*it).t; }
		T& operator*()
			{ assert(it!=0 && lst!=0); return (*it).t; }
		T* operator->() const
			{ assert(it!=0 && lst!=0); return &((*it).t); }

		iterator operator++() // prefix ++
			{ if(lst!=0 && it!=0) ++it; return *this; }
		iterator operator++(int) // postfix ++
			{ iterator save=*this; if(lst!=0 && it!=0) ++it; return save; }
		iterator operator--()
			{ if(lst!=0) { if(it==0) it=lst->maximizer(); else if(it!=lst->base::begin()) --it; } return *this; }
		iterator operator--(int)
			{ iterator save=*this; if(lst!=0) { if(it==0) it=lst->maximizer(); else if(it!=lst->base::begin()) --it; }
				return save; }
		bool operator==(const iterator &i) const { if(0==lst) return 0==i.lst; else return lst==i.lst && it==i.it; }
		bool operator==(const const_iterator &i) const { if(0==lst) return 0==i.lst; else return lst==i.lst && it==i.it; }
		bool operator==(const long x) const { return x==0 && lst==0; }
		bool operator!=(const iterator &i) const { if(0==lst) return 0!=i.lst; else return lst!=i.lst || it!=i.it; }
		bool operator!=(const const_iterator &i) const { return lst!=i.lst || it!=i.it; }
		bool operator!=(const long x) const { return x!=0 || lst!=0; }
		bool operator<(const iterator &i) const
			{ return (0==lst) ? (0!=it.lst) : ((0==it.lst) ? true : (it<i.it)); }
		bool operator<(const const_iterator &i) const
			{ return (0==lst) ? (0!=it.lst) : ((0==it.lst) ? true : (it<i.it)); }
		bool is_valid_and_not_end() const { return it!=0 && lst!=0; }
		bool is_end() const { return lst!=0 && it==0; } // siehe obige Definition !
		bool is_begin() const { return lst!=0 && it==lst->base::begin(); }
		bool is_uninitialized() const { return lst==0; }
		long set_summand(long n) const
			{ if(it!=0 && lst!=0) it->to_add1=n; lst->attr_changed_recalculate(it); return n; }
		long get_summand() const
			{ if(it!=0 && lst!=0) return it->to_add1; else return 0; }
		long get_partialsum() const
			{		// Das erste Listenelement hat die Partialsumme 0, das end()-Element hat die Summe der
					// to_add2-Elemente aller Listenelemente.
				if(lst==0) return 0;
				if(it==0) return (lst->root()==0) ? 0 : (lst->root()->psum1);
				long sum=(it.left()!=0) ? (it.left()->psum1) : 0;
				typename base::iterator it2=it;
				while(it2.parent()!=0)
				{
					if(it2.parent().right()==it2)
					{
						sum += (it2=it2.parent())->to_add1;
						if(it2.left()!=0) sum+=it2.left()->psum1;
					}
					else
						it2=it2.parent();
				}
				return sum;
			}
		long get_listpos() const
			{		// Das erste Listenelement hat die Position 0, das end()-Element hat die Position
					// size().
				if(lst==0) return -1;
				if(it==0) return (lst->root()==0) ? 0 : (lst->root()->anz_elem);
				long sum=(it.left()!=0) ? (it.left()->anz_elem) : 0;
				typename base::iterator it2=it;
				while(it2.parent()!=0)
				{
					if(it2.parent().right()==it2)
					{
						++sum;
						if((it2=it2.parent()).left()!=0)
							sum+=it2.left()->anz_elem;
					}
					else
						it2=it2.parent();
				}
				return sum;
			}
		SumList<T> *get_list() { return lst; }
	};

	class const_iterator
	{		// const_iterator==0 bei lst=0, const_iterator==end() bei lst!=0, it=0
		clss *lst;
		typename base::const_iterator it;
		friend class SumList<T>;
		friend class iterator;
		const_iterator(clss *_lst) : lst(_lst), it(0) { }
		const_iterator(clss *_lst, typename base::const_iterator _it) : lst(_lst), it(_it) { }
	public:
		const_iterator(long dummy) { it=0; lst=0; }
		const_iterator(int dummy=0) { it=0; lst=0; }
			// damit der Wert 0 explizit bergeben werden kann.
		const_iterator(const const_iterator &i) : lst(i.lst), it(i.it) { }
		const_iterator(const iterator &i) : lst(i.lst), it(i.it) { }
		const_iterator& operator=(const const_iterator &i) { lst=i.lst; it=i.it; return *this; }
		const_iterator& operator=(const iterator &i) { lst=i.lst; it=i.it; return *this; }
		const_iterator& operator=(long dummy) { it=0; lst=0; return *this; }

		const T& operator*() const
			{ assert(it!=0 && lst!=0); return (*it).t; }
		const T* operator->() const
			{ assert(it!=0 && lst!=0); return &((*it).t); }
		const_iterator operator++() // prefix ++
			{ if(lst!=0 && it!=0) ++it; return *this; }
		const_iterator operator++(int) // postfix ++
			{ const_iterator save=*this; if(lst!=0 && it!=0) ++it; return save; }
		const_iterator operator--()
			{ if(lst!=0) { if(it==0) it=lst->maximizer(); else if(it!=lst->base::begin()) --it; } return *this; }
		const_iterator operator--(int)
			{ const_iterator save=*this; if(lst!=0) { if(it==0) it=lst->maximizer(); else if(it!=lst->base::begin()) --it; }
				return save; }
		bool operator==(const const_iterator &i) const { if(0==lst) return 0==i.lst; else return lst==i.lst && it==i.it; }
		bool operator==(const iterator &i) const { if(0==lst) return 0==i.lst; else return lst==i.lst && it==i.it; }
		bool operator==(const long x) const { return x==0 && lst==0; }
		bool operator!=(const const_iterator &i) const { if(0==lst) return 0!=i.lst; else return lst!=i.lst || it!=i.it; }
		bool operator!=(const iterator &i) const { return lst!=i.lst || it!=i.it; }
		bool operator!=(const long x) const { return x!=0 || lst!=0; }
		bool operator<(const iterator &i) const
			{ return (0==lst) ? (0!=it.lst) : ((0==it.lst) ? true : (it<i.it)); }
		bool operator<(const const_iterator &i) const
			{ return (0==lst) ? (0!=it.lst) : ((0==it.lst) ? true : (it<i.it)); }
		bool is_valid_and_not_end() const { return it!=0 && lst!=0; }
		bool is_end() const { return lst!=0 && it==0; } // siehe obige Definition !
		bool is_begin() const { return lst!=0 && it==lst->base::begin(); }
		bool is_uninitialized() const { return lst==0; }
		long set_summand(long n) const
			{ if(it!=0 && lst!=0) it->to_add1=n; lst->attr_changed_recalculate(it); return n; }
		long get_summand() const
			{ if(it!=0 && lst!=0) return it->to_add1; else return 0; }
		long get_partialsum() const
			{		// Das erste Listenelement hat die Partialsumme 0, das end()-Element hat die Summe der
					// to_add2-Elemente aller Listenelemente.
				if(lst==0) return 0;
				if(it==0) return (lst->root()==0) ? 0 : (lst->root()->psum1);
				long sum=(it.left()!=0) ? (it.left()->psum1) : 0;
				typename base::iterator it2=it;
				while(it2.parent()!=0)
				{
					if(it2.parent().right()==it2)
					{
						sum += (it2=it2.parent())->to_add1;
						if(it2.left()!=0) sum+=it2.left()->psum1;
					}
					else
						it2=it2.parent();
				}
				return sum;
			}
		long get_listpos() const
			{		// Das erste Listenelement hat die Position 0, das end()-Element hat die Position
					// size().
				if(lst==0) return -1;
				if(it==0) return (lst->root()==0) ? 0 : (lst->root()->anz_elem);
				long sum=(it.left()!=0) ? (it.left()->anz_elem) : 0;
				typename base::iterator it2=it;
				while(it2.parent()!=0)
				{
					if(it2.parent().right()==it2)
					{
						++sum;
						if((it2=it2.parent()).left()!=0)
							sum+=it2.left()->anz_elem;
					}
					else
						it2=it2.parent();
				}
				return sum;
			}
		SumList<T> *get_list() { return lst; }
		iterator ConstCast() { return iterator(lst, it); }
	};

public:
	SumList() { }
	SumList(const SumList &l) : base(l) { }
	SumList& operator=(const SumList& l)
		{ if(this!=&l) base::operator=(l); return *this; }

	iterator begin()
		{ iterator it; it.lst=this; it.it=base::begin(); return it; }
	const_iterator begin() const
		{ const_iterator it; it.lst=const_cast<SumList<T>*>(this); it.it=base::begin(); return it; }
	iterator end()
		{ iterator it; it.lst=this; it.it=0; return it; }
	const_iterator end() const
		{ iterator it; it.lst=const_cast<SumList<T>*>(this); it.it=0; return it; }

	bool check_is_iterator_from_list(const iterator &it) const { return it.lst==this; }
	bool check_is_iterator_from_list(const const_iterator &it) const { return it.lst==this; }

	bool check_is_itr_from_list_and_not_end(const iterator &it) const
			{ return it->lst==this && it->it!=0; }
	bool check_is_itr_from_list_and_not_end(const const_iterator &it) const
			{ return it->lst==this && it->it!=0; }

	long size() const { return base::size(); }
	iterator insert(const_iterator it, const T& x=T(), long _to_add=0)
		{ if(it.lst==this)
		  {
			if(it.it==0)
				return iterator(this, base::push_back(TExt(x, _to_add)));
			else
				return iterator(this, base::insert_before(it.it, TExt(x, _to_add)));
		  }
		  return iterator();
		}
	iterator erase(iterator it)
		{ if(it.lst==this && it.it!=0)
		  {
			iterator it2(it);
			++it2.it;
			base::erase(it.it);
			return it2;
		  }
		  return iterator();
		}
	void clear()
		{ base::clear(); }
	void push_back(const T& x=T()) { base::push_back(TExt(x,0)); }
	void push_front(const T& x=T()) { base::insert_before(base::begin(), TExt(x,0)); }
	void pop_back() { assert(base::root()!=0); base::erase(base::maximizer()); } // Achtung: --base::end() ist nicht
			// statt base::maximizer() zulssig, weil end() beim RotSchwarzBaum ein Nulliterator ist und
			// --end()==end ist.
	void pop_front() { assert(base::root()!=0); base::erase(base::begin()); }
	T& back() { assert(base::root()!=0); return (base::maximizer())->t; }
			// Achtung: (--base::end())->t ist nicht statt base::maximizer()->t zulssig, weil end() beim
			// RotSchwarzBaum ein Nulliterator ist und --end()==end ist.
	T& front() { assert(base::root()!=0); return (base::begin())->t; }
	iterator find_iter_from_listpos(long listpos)
		{
			if(base::root()==0)
				return (0==listpos) ? end() : iterator();
			typename base::iterator it=base::root();
			while(it!=0)
			{
				long r=(it.left()!=0) ? (it.left()->anz_elem) : 0;
				if(listpos==r)
					return iterator(this, it);
				else if(listpos<r)
					it=it.left();
				else
				{
					it=it.right();
					listpos -= (r+1);
				}
			}
			return (0==listpos) ? end() : iterator();
		}
	iterator find_iter_from_partialsum(long psum)
		{		// Sucht den Iterator fr das erste Listenelement, so dass die Partialsumme aus
				// den Summanden aller frheren Listenelemente kleiner oder gleich 'psum' ist
				// und die Summe der Summanden aller frheren Listenelemente zuzglich des
				// Summanden des durch den Iterator bezeichneten Listenelements grer oder gleich als
				// 'psum' ist. Wenn partialsum1 kleiner als Null oder grer oder gleich
				// der 'to_add1'-Summe aller Listenelemente ist, wird der Nulliterator geliefert.
				// Wenn psum zwischen einschlielich 0 und ausschlielich der Summe der 'sum1'-Elemente
				// der Listenelemente ist, existiert ein Iterator gem obiger Regel stets, wenn alle
				// 'to_add1'-Elemente nichtnegativ sind, insbesondere auch dann, wenn einige Null sind.
				// ACHTUNG: DIESE FUNKTION SETZT VORAUS, DASS ALLE SUMMANDEN to_add1, DIE MIT
				// iterator::set_summand1() GESETZT WORDEN SIND, NICHTNEGATIV SIND.
			if(base::root()==0)
				return iterator();
			typename base::iterator it=base::root();
			long r=0;
			while(it!=0)
			{
				r=(it.left()!=0) ? (it.left()->psum1) : 0;
				if(psum>=r && psum<r+it->to_add1)
					return iterator(this, it);
				else if(psum<r)
					it=it.left();
				else
				{
					psum -= (r + it->to_add1); // vor it=it.right() !!!
					it=it.right();
				}
			}
			return iterator();
		}
};

template<class T> class SumList2;

template<class T>
class SumList2_TExt
{
	friend class SumList2<T>;
	SumList2_TExt(const T &_t=T(), long _to_add1=0, long _to_add2=0)
		: t(_t), to_add1(_to_add1), to_add2(_to_add2), anz_elem(0), psum1(0), psum2(0) { }

public:
	T t;
	long to_add1, to_add2, psum1, psum2, anz_elem;
	SumList2_TExt(const SumList2_TExt &l) : t(l.t), to_add1(l.to_add1), to_add2(l.to_add2),
		psum1(l.psum1), psum2(l.psum2), anz_elem(l.anz_elem) { }
	SumList2_TExt& operator=(const SumList2_TExt &l)
		{ if(this!=&l) { t=l.t; to_add1=l.to_add1; to_add2=l.to_add2; psum1=l.psum1; psum2=l.psum2;
				anz_elem=l.anz_elem; } return *this; }
};

template<class T>
struct SumList2_Compare
{
	typedef SumList2_TExt<T> X;
	bool less(const X& x, const X& y) const { return true; } // wird nicht benutzt
	bool equal(const X& x, const X& y) const { return true; } // wird nicht benutzt

	enum { soll_attribut_berechnen=1 };
	void calculate_attribute(X& akt) { akt.psum1=akt.to_add1; akt.psum2=akt.to_add2; akt.anz_elem=1; }
	void calculate_attribute(const X& links, const X& rechts, X& akt)
		{ akt.psum1=akt.to_add1+links.psum1+rechts.psum1; akt.psum2=akt.to_add2+links.psum2+rechts.psum2;
			akt.anz_elem=1+links.anz_elem+rechts.anz_elem; }
	void calculate_attribute1(const X& links, X& akt)
		{ akt.psum1=akt.to_add1+links.psum1; akt.psum2=akt.to_add2+links.psum2; akt.anz_elem=1+links.anz_elem; }
	void calculate_attribute2(const X& rechts, X& akt)
		{ akt.psum1=akt.to_add1+rechts.psum1; akt.psum2=akt.to_add2+rechts.psum2; akt.anz_elem=1+rechts.anz_elem; }
};

template<class T>
class SumList2 : public RotSchwarzBaum<SumList2_TExt<T>, SumList2_Compare<T> >
{
	typedef SumList2_TExt<T> TExt;
	typedef RotSchwarzBaum<SumList2_TExt<T>, SumList2_Compare<T> > base;
	typedef SumList2<T> clss;

public:
	class iterator;
	class const_iterator;
	friend class iterator;
	friend class const_iterator;

	class iterator
	{		// iterator==0 bei lst=0, iterator==end() bei lst!=0, it=0
		clss *lst;
		typename base::iterator it;
		friend class SumList2<T>;
		friend class const_iterator;
		iterator(clss *_lst) : lst(_lst), it(0) { }
		iterator(clss *_lst, typename base::iterator _it) : lst(_lst), it(_it) { }
	public:
		iterator(long dummy) { it=0; lst=0; }
		iterator(int dummy=0) { it=0; lst=0; }
			// damit der Wert 0 explizit bergeben werden kann.
		iterator(const iterator &i) : lst(i.lst), it(i.it) { }
		iterator& operator=(const iterator &i) { lst=i.lst; it=i.it; return *this; }
		iterator& operator=(long dummy) { it=0; lst=0; return *this; }
		bool operator<(const iterator &i) const
			{ return (0==lst) ? (0!=i.lst) : ((0==i.lst) ? true : (it<i.it)); }
		bool operator<(const const_iterator &i) const
			{ return (0==lst) ? (0!=i.lst) : ((0==i.lst) ? true : (it<i.it)); }

		const T& operator*() const
			{ assert(it!=0 && lst!=0); return (*it).t; }
		T& operator*()
			{ assert(it!=0 && lst!=0); return (*it).t; }
		T* operator->() const
			{ assert(it!=0 && lst!=0); return &((*it).t); }
		iterator operator++() // prefix ++
			{ if(lst!=0 && it!=0) ++it; return *this; }
		iterator operator++(int) // postfix ++
			{ iterator save=*this; if(lst!=0 && it!=0) ++it; return save; }
		iterator operator--()
			{ if(lst!=0) { if(it==0) it=lst->maximizer(); else if(it!=lst->base::begin()) --it; } return *this; }
		iterator operator--(int)
			{ iterator save=*this; if(lst!=0) { if(it==0) it=lst->maximizer(); else if(it!=lst->base::begin()) --it; }
				return save; }
		bool operator==(const iterator &i) const { if(0==lst) return 0==i.lst; else return lst==i.lst && it==i.it; }
		bool operator==(const const_iterator &i) const { if(0==lst) return 0==i.lst; else return lst==i.lst && it==i.it; }
		bool operator==(const long x) const { return x==0 && lst==0; }
		bool operator!=(const iterator &i) const { if(0==lst) return 0!=i.lst; else return lst!=i.lst || it!=i.it; }
		bool operator!=(const const_iterator &i) const { return lst!=i.lst || it!=i.it; }
		bool operator!=(const long x) const { return x!=0 || lst!=0; }
		bool is_valid_and_not_end() const { return it!=0 && lst!=0; }
		bool is_end() const { return lst!=0 && it==0; } // siehe obige Definition !
		bool is_begin() const { return lst!=0 && it==lst->base::begin(); }
		bool is_uninitialized() const { return lst==0; }
		long set_summand1(long n1) const
			{ if(it!=0 && lst!=0) it->to_add1=n1; lst->attr_changed_recalculate(it); return n1; }
		long set_summand2(long n2) const
			{ if(it!=0 && lst!=0) it->to_add2=n2; lst->attr_changed_recalculate(it); return n2; }
		long get_summand1() const
			{ if(it!=0 && lst!=0) return it->to_add1; else return 0; }
		long get_summand2() const
			{ if(it!=0 && lst!=0) return it->to_add2; else return 0; }
		long get_partialsum1() const
			{		// Das erste Listenelement hat die Partialsumme 0, das end()-Element hat die Summe der
					// to_add1-Elemente aller Listenelemente.
				if(lst==0) return 0;
				if(it==0) return (lst->root()==0) ? 0 : (lst->root()->psum1);
				long sum=(it.left()!=0) ? (it.left()->psum1) : 0;
				typename base::iterator it2=it;
				while(it2.parent()!=0)
				{
					if(it2.parent().right()==it2)
					{
						sum += (it2=it2.parent())->to_add1;
						if(it2.left()!=0) sum+=it2.left()->psum1;
					}
					else
						it2=it2.parent();
				}
				return sum;
			}
		long get_partialsum2() const
			{		// Das erste Listenelement hat die Partialsumme 0, das end()-Element hat die Summe der
					// to_add2-Elemente aller Listenelemente.
				if(lst==0) return 0;
				if(it==0) return (lst->root()==0) ? 0 : (lst->root()->psum2);
				long sum=(it.left()!=0) ? (it.left()->psum2) : 0;
				typename base::iterator it2=it;
				while(it2.parent()!=0)
				{
					if(it2.parent().right()==it2)
					{
						sum += (it2=it2.parent())->to_add2;
						if(it2.left()!=0) sum+=it2.left()->psum2;
					}
					else
						it2=it2.parent();
				}
				return sum;
			}
		long get_listpos() const
			{		// Das erste Listenelement hat die Position 0, das end()-Element hat die Position
					// size().
				if(lst==0) return -1;
				if(it==0) return (lst->root()==0) ? 0 : (lst->root()->anz_elem);
				long sum=(it.left()!=0) ? (it.left()->anz_elem) : 0;
				typename base::iterator it2=it;
				while(it2.parent()!=0)
				{
					if(it2.parent().right()==it2)
					{
						++sum;
						if((it2=it2.parent()).left()!=0)
							sum+=it2.left()->anz_elem;
					}
					else
						it2=it2.parent();
				}
				return sum;
			}
		SumList2<T> *get_list() { return lst; }
	};

	class const_iterator
	{		// const_iterator==0 bei lst=0, const_iterator==end() bei lst!=0, it=0
		clss *lst;
		typename base::const_iterator it;
		friend class SumList2<T>;
		friend class iterator;
		const_iterator(clss *_lst) : lst(_lst), it(0) { }
		const_iterator(clss *_lst, typename base::const_iterator _it) : lst(_lst), it(_it) { }
	public:
		const_iterator(long dummy) { it=0; lst=0; }
		const_iterator(int dummy=0) { it=0; lst=0; }
			// damit der Wert 0 explizit bergeben werden kann.
		const_iterator(const const_iterator &i) : lst(i.lst), it(i.it) { }
		const_iterator(const iterator &i) : lst(i.lst), it(i.it) { }
		const_iterator& operator=(const const_iterator &i) { lst=i.lst; it=i.it; return *this; }
		const_iterator& operator=(const iterator &i) { lst=i.lst; it=i.it; return *this; }
		const_iterator& operator=(long dummy) { it=0; lst=0; return *this; }
		bool operator<(const iterator &i) const
			{ return (0==lst) ? (0!=it.lst) : ((0==it.lst) ? true : (it<i.it)); }
		bool operator<(const const_iterator &i) const
			{ return (0==lst) ? (0!=it.lst) : ((0==it.lst) ? true : (it<i.it)); }

		const T& operator*() const
			{ assert(it!=0 && lst!=0); return (*it).t; }
		const T* operator->() const
			{ assert(it!=0 && lst!=0); return &((*it).t); }
		const_iterator operator++() // prefix ++
			{ if(lst!=0 && it!=0) ++it; return *this; }
		const_iterator operator++(int) // postfix ++
			{ const_iterator save=*this; if(lst!=0 && it!=0) ++it; return save; }
		const_iterator operator--()
			{ if(lst!=0) { if(it==0) it=lst->maximizer(); else if(it!=lst->base::begin()) --it; } return *this; }
		const_iterator operator--(int)
			{ const_iterator save=*this; if(lst!=0) { if(it==0) it=lst->maximizer(); else if(it!=lst->base::begin()) --it; }
				return save; }
		bool operator==(const const_iterator &i) const { if(0==lst) return 0==i.lst; else return lst==i.lst && it==i.it; }
		bool operator==(const iterator &i) const { if(0==lst) return 0==i.lst; else return lst==i.lst && it==i.it; }
		bool operator==(const long x) const { return x==0 && lst==0; }
		bool operator!=(const const_iterator &i) const { if(0==lst) return 0!=i.lst; else return lst!=i.lst || it!=i.it; }
		bool operator!=(const iterator &i) const { return lst!=i.lst || it!=i.it; }
		bool operator!=(const long x) const { return x!=0 || lst!=0; }
		bool is_valid_and_not_end() const { return it!=0 && lst!=0; }
		bool is_end() const { return lst!=0 && it==0; } // siehe obige Definition !
		bool is_begin() const { return lst!=0 && it==lst->base::begin(); }
		bool is_uninitialized() const { return lst==0; }
		long set_summand1(long n1) const
			{ if(it!=0 && lst!=0) it->to_add1=n1; lst->attr_changed_recalculate(it); return n1; }
		long set_summand2(long n2) const
			{ if(it!=0 && lst!=0) it->to_add2=n2; lst->attr_changed_recalculate(it); return n2; }
		long get_summand1() const
			{ if(it!=0 && lst!=0) return it->to_add1; else return 0; }
		long get_summand2() const
			{ if(it!=0 && lst!=0) return it->to_add2; else return 0; }
		long get_partialsum1() const
			{		// Das erste Listenelement hat die Partialsumme 0, das end()-Element hat die Summe der
					// to_add1-Elemente aller Listenelemente.
				if(lst==0) return 0;
				if(it==0) return (lst->root()==0) ? 0 : (lst->root()->psum1);
				long sum=(it.left()!=0) ? (it.left()->psum1) : 0;
				typename base::iterator it2=it;
				while(it2.parent()!=0)
				{
					if(it2.parent().right()==it2)
					{
						sum += (it2=it2.parent())->to_add1;
						if(it2.left()!=0) sum+=it2.left()->psum1;
					}
					else
						it2=it2.parent();
				}
				return sum;
			}
		long get_partialsum2() const
			{		// Das erste Listenelement hat die Partialsumme 0, das end()-Element hat die Summe der
					// to_add2-Elemente aller Listenelemente.
				if(lst==0) return 0;
				if(it==0) return (lst->root()==0) ? 0 : (lst->root()->psum2);
				long sum=(it.left()!=0) ? (it.left()->psum2) : 0;
				typename base::iterator it2=it;
				while(it2.parent()!=0)
				{
					if(it2.parent().right()==it2)
					{
						sum += (it2=it2.parent())->to_add2;
						if(it2.left()!=0) sum+=it2.left()->psum2;
					}
					else
						it2=it2.parent();
				}
				return sum;
			}
		long get_listpos() const
			{		// Das erste Listenelement hat die Position 0, das end()-Element hat die Position
					// size().
				if(lst==0) return -1;
				if(it==0) return (lst->root()==0) ? 0 : (lst->root()->anz_elem);
				long sum=(it.left()!=0) ? (it.left()->anz_elem) : 0;
				typename base::iterator it2=it;
				while(it2.parent()!=0)
				{
					if(it2.parent().right()==it2)
					{
						++sum;
						if((it2=it2.parent()).left()!=0)
							sum+=it2.left()->anz_elem;
					}
					else
						it2=it2.parent();
				}
				return sum;
			}
		SumList2<T> *get_list() { return lst; }
		iterator ConstCast() { return iterator(lst, it); }
	};

public:
	SumList2() { }
	SumList2(const SumList2 &l) : base(l) { }
	SumList2& operator=(const SumList2& l)
		{ if(this!=&l) base::operator=(l); return *this; }

	iterator begin()
		{ iterator it; it.lst=this; it.it=base::begin(); return it; }
	const_iterator begin() const
		{ const_iterator it; it.lst=const_cast<SumList2<T>*>(this); it.it=base::begin(); return it; }
	iterator end()
		{ iterator it; it.lst=this; it.it=0; return it; }
	const_iterator end() const
		{ iterator it; it.lst=const_cast<SumList2<T>*>(this); it.it=0; return it; }

	bool check_is_iterator_from_list(const iterator &it) const { return it.lst==this; }
	bool check_is_iterator_from_list(const const_iterator &it) const { return it.lst==this; }

	bool check_is_itr_from_list_and_not_end(const iterator &it) const
			{ return it->lst==this && it->it!=0; }
	bool check_is_itr_from_list_and_not_end(const const_iterator &it) const
			{ return it->lst==this && it->it!=0; }

	long size() const { return base::size(); }
	iterator insert(const_iterator it, const T& x=T(), long _to_add1=0, long _to_add2=0)
		{ if(it.lst==this)
		  {
			if(it.it==0)
				return iterator(this, base::push_back(TExt(x, _to_add1, _to_add2)));
			else
				return iterator(this, base::insert_before(it.it, TExt(x, _to_add1, _to_add2)));
		  }
		  return iterator();
		}
	iterator erase(iterator it)
		{ if(it.lst==this && it.it!=0)
		  {
			iterator it2(it);
			++it2.it;
			base::erase(it.it);
			return it2;
		  }
		  return iterator();
		}
	void clear()
		{ base::clear(); }
	void push_back(const T& x=T()) { base::push_back(TExt(x,0)); }
	void push_front(const T& x=T()) { base::insert_before(base::begin(), TExt(x,0)); }
	void pop_back() { assert(base::root()!=0); base::erase(base::maximizer()); } // Achtung: --base::end() ist nicht
			// statt base::maximizer() zulssig, weil end() beim RotSchwarzBaum ein Nulliterator ist und
			// --end()==end ist.
	void pop_front() { assert(base::root()!=0); base::erase(base::begin()); }
	T& back() { assert(base::root()!=0); return (base::maximizer())->t; }
			// Achtung: (--base::end())->t ist nicht statt base::maximizer()->t zulssig, weil end() beim
			// RotSchwarzBaum ein Nulliterator ist und --end()==end ist.
	T& front() { assert(base::root()!=0); return (base::begin())->t; }
	iterator find_iter_from_listpos(long listpos)
		{
			if(base::root()==0)
				return (0==listpos) ? end() : iterator();
			typename base::iterator it=base::root();
			while(it!=0)
			{
				long r=(it.left()!=0) ? (it.left()->anz_elem) : 0;
				if(listpos==r)
					return iterator(this, it);
				else if(listpos<r)
					it=it.left();
				else
				{
					it=it.right();
					listpos -= (r+1);
				}
			}
			return (0==listpos) ? end() : iterator();
		}
	iterator find_iter_from_partialsum1(long psum)
		{		// Sucht den Iterator fr das erste Listenelement, so dass die Partialsumme aus
				// den Summanden aller frheren Listenelemente kleiner oder gleich 'psum' ist
				// und die Summe der Summanden aller frheren Listenelemente zuzglich des
				// Summanden des durch den Iterator bezeichneten Listenelements grer oder gleich als
				// 'psum' ist. Wenn partialsum1 kleiner als Null oder grer oder gleich
				// der 'to_add1'-Summe aller Listenelemente ist, wird der Nulliterator geliefert.
				// Wenn psum zwischen einschlielich 0 und ausschlielich der Summe der 'sum1'-Elemente
				// der Listenelemente ist, existiert ein Iterator gem obiger Regel stets, wenn alle
				// 'to_add1'-Elemente nichtnegativ sind, insbesondere auch dann, wenn einige Null sind.
				// ACHTUNG: DIESE FUNKTION SETZT VORAUS, DASS ALLE SUMMANDEN to_add1, DIE MIT
				// iterator::set_summand1() GESETZT WORDEN SIND, NICHTNEGATIV SIND.
			if(base::root()==0)
				return iterator();
			typename base::iterator it=base::root();
			long r=0;
			while(it!=0)
			{
				r=(it.left()!=0) ? (it.left()->psum1) : 0;
				if(psum>=r && psum<r+it->to_add1)
					return iterator(this, it);
				else if(psum<r)
					it=it.left();
				else
				{
					psum -= (r + it->to_add1); // vor it=it.right() !!!
					it=it.right();
				}
			}
			return iterator();
		}
	iterator find_iter_from_partialsum2(long psum)
		{		// siehe auch find_iter_from_partialsum1()
			if(base::root()==0)
				return iterator();
			typename base::iterator it=base::root();
			long r=0;
			while(it!=0)
			{
				r=(it.left()!=0) ? (it.left()->psum2) : 0;
				if(psum>=r && psum<r+it->to_add2)
					return iterator(this, it);
				else if(psum<r)
					it=it.left();
				else
				{
					psum -= (r + it->to_add2); // vor it=it.right() !!!
					it=it.right();
				}
			}
			return iterator();
		}
};

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

template<class T> class checkptr;
template<class T> class checkptr_c;
template<class T, class S> bool operator==(checkptr<T> p1, checkptr<S> p2);
template<class T, class S> bool operator==(checkptr<T> p1, const S *p2);
template<class T, class S> bool operator==(const T *p1, checkptr<S> p2);
template<class T> bool operator==(checkptr<T> p, int x);
template<class T> bool operator==(int x, checkptr<T> p);
template<class T, class S> bool operator!=(checkptr<T> p1, checkptr<S> p2);
template<class T, class S> bool operator!=(checkptr<T> p1, const S *p2);
template<class T, class S> bool operator!=(const T *p1, checkptr<S> p2);
template<class T> bool operator!=(checkptr<T> p, int x);
template<class T> bool operator!=(int x, checkptr<T> p);
template<typename Y, typename X> checkptr<Y> StaticCast(const checkptr<X> &ptr);
template<typename Y, typename X> checkptr<Y> DynamicCast(const checkptr<X> &ptr);

class checkptr_UID;
template<class T> class checkptr;

struct checkptr_UID_LessEqual : public BaumKeineAttrBerechnen<checkptr_UID>
{
	bool less(const checkptr_UID& t1, const checkptr_UID& t2) const;
	bool equal(const checkptr_UID& t1, const checkptr_UID& t2) const;
};

class checkptr_UID
{	// Das Objekt, welches berwacht werden soll, muss checkptr_UID als Member haben.
	static RotSchwarzBaum<checkptr_UID, checkptr_UID_LessEqual > alive_object_uids;
	unsigned long long uid; // Wenn Bit 0 gesetzt ist, handelt es sich um eine Kopie
	static unsigned long long next_unused_uid;
	friend class RotSchwarzBaum<checkptr_UID, checkptr_UID_LessEqual>;
	friend struct RotSchwarzBaum_Node<checkptr_UID >;
	friend struct checkptr_UID_LessEqual;
	template<class T> friend class checkptr;
	template<class T> friend class checkptr_c;
	private: checkptr_UID(const checkptr_UID& u) : uid(u.uid)
			 { uid |= 1; } // Als Kopie markieren, Zugriff auf den Copy-Konstruktor von auen verhindern.
	private: checkptr_UID& operator=(const checkptr_UID& u)
			 { uid=(u.uid | 1); return *this; } // Zugriff verhindern
	checkptr_UID(unsigned long long u) { uid=u; if(alive_object_uids.suche(*this)==0) alive_object_uids.insert(*this); }
public:
	checkptr_UID() { uid=next_unused_uid; next_unused_uid+=2; alive_object_uids.insert(*this); }
	~checkptr_UID()
	{
		if(0==(uid & 1))
			alive_object_uids.erase(*this);
	}
};

#ifndef NLIST_H_DONT_DEFINE_CODE_WHICH_MUST_NOT_BE_DEFINED_IN_MORE_THAN_ONE_SOURCE_FILE
bool checkptr_UID_LessEqual::less(const checkptr_UID& t1, const checkptr_UID& t2) const
{
	return (t1.uid & (~((unsigned long long)1))) < (t2.uid & (~((unsigned long long)1)));
}
bool checkptr_UID_LessEqual::equal(const checkptr_UID& t1, const checkptr_UID& t2) const
{
	return (t1.uid & (~((unsigned long long)1))) == (t2.uid & (~((unsigned long long)1)));
}

unsigned long long checkptr_UID::next_unused_uid=2; // uid==0 und uid==1 sind reserviert fr "uid undefiniert"
RotSchwarzBaum<checkptr_UID, checkptr_UID_LessEqual > checkptr_UID::alive_object_uids;
#endif

template<class T>
class checkptr
{
private:
	T *ptr;
	checkptr_UID uid;
	template<class U> friend class checkptr_c;
public:
	struct AccessToNonExistingObject { };
	checkptr(T *pointer_to_object) : ptr(pointer_to_object), uid(checkptr_UID(0)) { }
	checkptr(T *pointer_to_object, const checkptr_UID& uid_of_object=0) : ptr(pointer_to_object), uid(uid_of_object) { }
	checkptr() : ptr(0), uid(0) { }
	checkptr(const checkptr& p) : ptr(p.ptr), uid(p.uid) { }
	checkptr& operator=(const checkptr& p) { ptr=p.ptr; uid=p.uid; return *this; }
	bool is_alive() const // Wenn checkptr aus einem Zeiger ohne zugehrende UID generiert wurde, kann nicht festgestellt werden,
			// ob das Objekt, aus das der Zeiger zeigt, noch existiert.
			// ******* In diesem Fall liefern SOWOHL is_alive() als auch is_dead() den Wert false. *******
		{ return 0!=ptr && (0!=(uid.uid & (~((unsigned long long)1))) && checkptr_UID::alive_object_uids.suche(uid)!=0); }
	bool is_dead() const
		{ return 0==ptr || (0!=(uid.uid & (~((unsigned long long)1))) && checkptr_UID::alive_object_uids.suche(uid)==0); }
	T& operator*() const
	{
		if(is_dead())
			throw AccessToNonExistingObject();
		return (*ptr);
	}
	T *operator->() const
	{
		if(is_dead())
			throw AccessToNonExistingObject();
		return ptr;
	}
	operator const T*() const
	{
		if(is_dead())
			throw AccessToNonExistingObject();
		return ptr;
	}
	T* get() const
	{
		if(is_dead())
			throw AccessToNonExistingObject();
		return ptr;
	}

	template<class U, class V> friend bool operator==(checkptr<U> p1, checkptr<V> p2);
	template<class U, class V> friend bool operator==(checkptr<U> p1, const V *p2);
	template<class U, class V> friend bool operator==(const U *p1, checkptr<V> p2);
	template<class U> friend bool operator==(checkptr<U> p, int x);
	template<class U> friend bool operator==(int x, checkptr<U> p);
	template<class U, class V> friend bool operator!=(checkptr<U> p1, checkptr<V> p2);
	template<class U, class V> friend bool operator!=(checkptr<U> p1, const V *p2);
	template<class U, class V> friend bool operator!=(const U *p1, checkptr<V> p2);
	template<class U> friend bool operator!=(checkptr<U> p, int x);
	template<class U> friend bool operator!=(int x, checkptr<U> p);
	template<typename Y, typename X> friend checkptr<Y> StaticCast(const checkptr<X> &ptr);
	template<typename Y, typename X> friend checkptr<Y> DynamicCast(const checkptr<X> &ptr);

	template<class U> operator checkptr<U>() const
		{ return checkptr<U>(ptr, uid); } // Diese return-Anweisung lsst sich nur compilieren, wenn es eine
		// EINGEBAUTE von const T* nach const U* gibt, d.h. wenn U eine Basisklasse von T ist. Falls jedoch U eine
		// abgeleitete Klasse von T ist, kann man ber operator checkptr_c<U>() nicht konvertieren. Falls der Zeiger
		// auf T jedoch tatschlich (auch) auf ein Objekt der Klasse U zeigt, kann man dies mit DynamicCast<U> feststellen
		// und die Konvertierung nach checkptr_c<U> mit StaticCast<U> durchfhren (siehe unten).
};

template<class T, class S> bool operator==(checkptr<T> p1, checkptr<S> p2) { return p1.ptr==p2.ptr; }
template<class T, class S> bool operator==(checkptr<T> p1, const S *p2) { return p1.ptr==p2; }
template<class T, class S> bool operator==(const T *p1, checkptr<S> p2) { return p1==p2.ptr; }
template<class T> bool operator==(checkptr<T> p, int x) { return (0==x) ? p.is_dead() : false; }
template<class T> bool operator==(int x, checkptr<T> p) { return (0==x) ? p.is_dead() : false; }
template<class T, class S> bool operator!=(checkptr<T> p1, checkptr<S> p2) { return p1.ptr!=p2.ptr; }
template<class T, class S> bool operator!=(checkptr<T> p1, const S *p2) { return p1.ptr!=p2; }
template<class T, class S> bool operator!=(const T *p1, checkptr<S> p2) { return p1!=p2.ptr; }
template<class T> bool operator!=(checkptr<T> p, int x) { return (0==x) ? !p.is_dead() : true; }
template<class T> bool operator!=(int x, checkptr<T> p) { return (0==x) ? !p.is_dead() : true; }

template<typename Y, typename X>
checkptr<Y> StaticCast(const checkptr<X> &ptr)
{
	return checkptr<Y>(static_cast<Y*>(ptr.ptr), ptr.uid);
}

template<typename Y, typename X>
checkptr<Y> DynamicCast(const checkptr<X> &ptr)
{
	return checkptr<Y>(dynamic_cast<Y*>(ptr.ptr), ptr.uid);
}

template<typename T>
struct AllowUseOf_checkptr
{	// AllowUseOf_checkptr<T> MUSS immer als Basisklasse einer anderen Klasse T sein,
	// z.B. struct ExampleClass : public AllowUseOf_checkptr<ExampleClass> { };
	private: typename checkptr<T>::UID uid;
	public:
	checkptr<T> operator&()
	{
		// Falls AllowUseOf_checkptr<T> irrtmlicherweise keine Basisklasse der Klasse T ist, meldet der Compiler beim
		// folgenden static_cast seine Fehlermeldung. Daher braucht die Zulssigkeit des static_cast nicht vorher mit dynamic_cast
		// getestet zu werden, und AllowUseOf_checkptr sowie T mssen nicht unbedingt virtuell sein.
		return checkptr<T>(static_cast<T*>(this), &uid);
	}
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////

template<class T> class checkptr_c;
template<class T, class S> bool operator==(checkptr_c<T> p1, checkptr_c<S> p2);
template<class T, class S> bool operator==(checkptr_c<T> p1, checkptr<S> p2);
template<class T, class S> bool operator==(checkptr<T> p1, checkptr_c<S> p2);
template<class T, class S> bool operator==(checkptr_c<T> p1, const S *p2);
template<class T, class S> bool operator==(const T *p1, checkptr_c<S> p2);
template<class T> bool operator==(checkptr_c<T> p, int x);
template<class T> bool operator==(int x, checkptr_c<T> p);
template<class T, class S> bool operator!=(checkptr_c<T> p1, checkptr_c<S> p2);
template<class T, class S> bool operator!=(checkptr_c<T> p1, checkptr<S> p2);
template<class T, class S> bool operator!=(checkptr<T> p1, checkptr_c<S> p2);
template<class T, class S> bool operator!=(checkptr_c<T> p1, const S *p2);
template<class T, class S> bool operator!=(const T *p1, checkptr_c<S> p2);
template<class T> bool operator!=(checkptr_c<T> p, int x);
template<class T> bool operator!=(int x, checkptr_c<T> p);
template<typename Y, typename X> checkptr_c<Y> StaticCast(const checkptr_c<X> &ptr);
template<typename Y, typename X> checkptr_c<Y> DynamicCast(const checkptr_c<X> &ptr);

template<class T>
class checkptr_c
{
private:
	const T *ptr;
	checkptr_UID uid;
public:
	struct AccessToNonExistingObject { };
	checkptr_c(const T *pointer_to_object) : ptr(pointer_to_object), uid(checkptr_UID(0)) { }
	checkptr_c(const T *pointer_to_object, const checkptr_UID& uid_of_object) : ptr(pointer_to_object), uid(uid_of_object) { }
	checkptr_c() : ptr(0), uid(0) { }
	checkptr_c(const checkptr_c& p) : ptr(p.ptr), uid(p.uid) { }
	checkptr_c(const checkptr<T>& p) : ptr(p.ptr), uid(p.uid) { }
	checkptr_c& operator=(const checkptr_c& p) { ptr=p.ptr; uid=p.uid; return *this; }
	bool is_alive() const // Wenn checkptr aus einem Zeiger ohne zugehrende UID generiert wurde, kann nicht festgestellt werden,
			// ob das Objekt, aus das der Zeiger zeigt, noch existiert.
			// ******* In diesem Fall liefern SOWOHL is_alive() als auch is_dead() den Wert false. *******
		{ return 0!=ptr && (0!=(uid.uid & (~((unsigned long long)1))) && checkptr_UID::alive_object_uids.suche(uid)!=0); }
	bool is_dead() const
		{ return 0==ptr || (0!=(uid.uid & (~((unsigned long long)1))) && checkptr_UID::alive_object_uids.suche(uid)==0); }
	const T& operator*() const
	{
		if(is_dead())
			throw AccessToNonExistingObject();
		return (*ptr);
	}
	const T *operator->() const
	{
		if(is_dead())
			throw AccessToNonExistingObject();
		return ptr;
	}
	operator const T*() const
	{
		if(is_dead())
			throw AccessToNonExistingObject();
		return ptr;
	}
	const T* get() const
	{
		if(is_dead())
			throw AccessToNonExistingObject();
		return ptr;
	}

	template<class U, class V> friend bool operator==(checkptr_c<U> p1, checkptr_c<V> p2);
	template<class U, class V> friend bool operator==(checkptr_c<U> p1, checkptr<V> p2);
	template<class U, class V> friend bool operator==(checkptr<U> p1, checkptr_c<V> p2);
	template<class U, class V> friend bool operator==(checkptr_c<U> p1, const V *p2);
	template<class U, class V> friend bool operator==(const U *p1, checkptr_c<V> p2);
	template<class U> friend bool operator==(checkptr_c<U> p, int x);
	template<class U> friend bool operator==(int x, checkptr_c<U> p);
	template<class U, class V> friend bool operator!=(checkptr_c<U> p1, checkptr_c<V> p2);
	template<class U, class V> friend bool operator!=(checkptr_c<U> p1, checkptr<V> p2);
	template<class U, class V> friend bool operator!=(checkptr<U> p1, checkptr_c<V> p2);
	template<class U, class V> friend bool operator!=(checkptr_c<U> p1, const V *p2);
	template<class U, class V> friend bool operator!=(const U *p1, checkptr_c<V> p2);
	template<class U> friend bool operator!=(checkptr_c<U> p, int x);
	template<class U> friend bool operator!=(int x, checkptr_c<U> p);
	template<typename Y, typename X> friend checkptr_c<Y> StaticCast(const checkptr_c<X> &ptr);
	template<typename Y, typename X> friend checkptr_c<Y> DynamicCast(const checkptr_c<X> &ptr);

	template<class U> operator checkptr_c<U>() const
		{ return checkptr_c<U>(ptr, uid); } // Diese return-Anweisung lsst sich nur compilieren, wenn es eine
		// EINGEBAUTE von const T* nach const U* gibt, d.h. wenn U eine Basisklasse von T ist. Falls jedoch U eine
		// abgeleitete Klasse von T ist, kann man ber operator checkptr_c<U>() nicht konvertieren. Falls der Zeiger
		// auf T jedoch tatschlich (auch) auf ein Objekt der Klasse U zeigt, kann man dies mit DynamicCast<U> feststellen
		// und die Konvertierung nach checkptr_c<U> mit StaticCast<U> durchfhren (siehe unten).
};

template<class T, class S> bool operator==(checkptr_c<T> p1, checkptr_c<S> p2) { return p1.ptr==p2.ptr; }
template<class T, class S> bool operator==(checkptr_c<T> p1, checkptr<S> p2) { return p1.ptr==p2.ptr; }
template<class T, class S> bool operator==(checkptr<T> p1, checkptr_c<S> p2) { return p1.ptr==p2.ptr; }
template<class T, class S> bool operator==(checkptr_c<T> p1, const S *p2) { return p1.ptr==p2; }
template<class T, class S> bool operator==(const T *p1, checkptr_c<S> p2) { return p1==p2.ptr; }
template<class T> bool operator==(checkptr_c<T> p, int x) { return (0==x) ? p.is_dead() : false; }
template<class T> bool operator==(int x, checkptr_c<T> p) { return (0==x) ? p.is_dead() : false; }
template<class T, class S> bool operator!=(checkptr_c<T> p1, checkptr_c<S> p2) { return p1.ptr!=p2.ptr; }
template<class T, class S> bool operator!=(checkptr_c<T> p1, checkptr<S> p2) { return p1.ptr!=p2.ptr; }
template<class T, class S> bool operator!=(checkptr<T> p1, checkptr_c<S> p2) { return p1.ptr!=p2.ptr; }
template<class T, class S> bool operator!=(checkptr_c<T> p1, const S *p2) { return p1.ptr!=p2; }
template<class T, class S> bool operator!=(const T *p1, checkptr_c<S> p2) { return p1!=p2.ptr; }
template<class T> bool operator!=(checkptr_c<T> p, int x) { return (0==x) ? !p.is_dead() : true; }
template<class T> bool operator!=(int x, checkptr_c<T> p) { return (0==x) ? !p.is_dead() : true; }

template<typename Y, typename X>
checkptr_c<Y> StaticCast(const checkptr_c<X> &ptr)
{
	return checkptr_c<Y>(static_cast<const Y*>(ptr.ptr), ptr.uid);
}

template<typename Y, typename X>
checkptr_c<Y> DynamicCast(const checkptr_c<X> &ptr)
{
	return checkptr_c<Y>(dynamic_cast<const Y*>(ptr.ptr), ptr.uid);
}

template<typename Y, typename X>
checkptr<Y> StaticConstCast(const checkptr_c<X> &ptr)
{
	return checkptr_c<Y>(const_cast<Y*>(static_cast<const Y*>(ptr.ptr)), ptr.uid);
}

template<typename Y, typename X>
checkptr<Y> DynamicConstCast(const checkptr_c<X> &ptr)
{
	return checkptr_c<Y>(const_cast<Y*>(dynamic_cast<const Y*>(ptr.ptr)), ptr.uid);
}

template<typename T>
struct AllowUseOf_checkptr_c
{	// AllowUseOf_checkptr<T> MUSS immer als Basisklasse einer anderen Klasse T sein,
	// z.B. struct ExampleClass : public AllowUseOf_checkptr<ExampleClass> { };
	private: typename checkptr_c<T>::UID uid;
	public:
	checkptr_c<T> operator&()
	{
		// Falls AllowUseOf_checkptr<T> irrtmlicherweise keine Basisklasse der Klasse T ist, meldet der Compiler beim
		// folgenden static_cast seine Fehlermeldung. Daher braucht die Zulssigkeit des static_cast nicht vorher mit dynamic_cast
		// getestet zu werden, und AllowUseOf_checkptr sowie T mssen nicht unbedingt virtuell sein.
		return checkptr_c<T>(static_cast<T*>(this), &uid);
	}
};

////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////

/*template<class _Ty> class count_ptr_ext;

template<class _Ty>
class count_ptr_TExt
{
	friend class count_ptr_ext<_Ty>;
	long count;
public:
	_Ty t;
	count_ptr_TExt(const _Ty &_t=_Ty()) : t(_t) { } // Keine Initialisierung von count !
	count_ptr_TExt(const count_ptr_TExt &te) : t(te.t), count(te.count) { }
	count_ptr_TExt& operator=(const count_ptr_TExt& te) { t=te.t; count=te.count; return *this; }
};

template<class _Ty>
class count_ptr_ext
{
public:
	typedef count_ptr_TExt<_Ty> TExt;
protected:
	explicit count_ptr_ext(TExt *_P, long count_init=1)
		: _Ptr(_P) { if(0!=_P) _P->count=count_init; }
public:
	count_ptr_ext() { _Ptr=0; }
	count_ptr_ext(const count_ptr_ext<_Ty>& _Y)
		{ if(0!=(_Ptr=_Y._Ptr)) ++_Ptr->count; }
	count_ptr_ext<_Ty>& operator=(const count_ptr_ext<_Ty>& _Y)
		{if (this != &_Y)
		  {	if(_Ptr != _Y._Ptr)
				{if(0!=_Ptr && --_Ptr->count==0)
					delete _Ptr;
				 if((_Ptr=_Y._Ptr)!=0) ++_Ptr->count; }
		  }
		 return (*this);
		}
	~count_ptr_ext()
		{ if(0!=_Ptr) { if(0==--_Ptr->count) delete _Ptr; } }
			// Nicht auf _Ptr->count<=0 abfragen, weil count_ptr_ext_static die Variable count auf
			// Null initialisiert !
	_Ty& operator*() const
		{return (*get()); }
	_Ty *operator->() const
		{return (get()); }
	_Ty *get() const
		{return &(_Ptr->t); }
	_Ty& get_ref() const
		{return (_Ptr->t); }

	bool operator==(const count_ptr_ext<_Ty> &p) const { return _Ptr==p._Ptr; }
	bool operator!=(const count_ptr_ext<_Ty> &p) const { return _Ptr!=p._Ptr; }
	bool operator<(const count_ptr_ext<_Ty> &p) const { return ((unsigned long)_Ptr)<((unsigned long)p._Ptr); }
		// operator< zum Einordnen von Zeigern in eine geordnete Liste oder einen RotSchwarzBaum
	bool operator==(unsigned long null) const { return _Ptr==0 && null==0; }
	bool operator!=(unsigned long null) const { return _Ptr!=0 || null!=0; }
	bool is_nullptr() const { return _Ptr==0; }
	bool is_valid() const { return _Ptr!=0; }
	bool is_last_reference() { return _Ptr!=0 && _Ptr->count==1; }

protected:
	TExt *_Ptr;
};

template<class _Ty>
class count_ptr_ext_simple_new : public count_ptr_ext<_Ty>
{		// Nur verwenden, wenn der Konstruktor count_ptr_ext_simple_new(const _Ty &t)
		// NICHT mit einem Objekt t einer Klasse verwendet wird, welche von _Ty
		// abgeleitet ist. Ansonsten muss count_ptr_ext_new verwendet werden.
public:
	typedef count_ptr_TExt<_Ty> TExt;
	count_ptr_ext_simple_new() : count_ptr_ext<_Ty>(new TExt) { }
	count_ptr_ext_simple_new(const TExt &text) : count_ptr_ext<_Ty>(new TExt(text)) { }
	count_ptr_ext_simple_new(const _Ty &t) : count_ptr_ext<_Ty>(new TExt(t)) { }
	count_ptr_ext_simple_new(const _Ty *t) : count_ptr_ext<_Ty>(new TExt(*t)) { }
	count_ptr_ext_simple_new(const count_ptr_ext<_Ty> &p) : count_ptr_ext<_Ty>(new TExt(p.get_ref())) { }
};

template<class _Ty>
class count_ptr_ext_static : public count_ptr_ext<_Ty>
{
public:
	typedef count_ptr_TExt<_Ty> TExt;
	count_ptr_ext_static(TExt *_P = 0) : count_ptr_ext<_Ty>(_P) { }
	count_ptr_ext_static(const TExt& _P) : count_ptr_ext<_Ty>(&_P) { }
};

template<class _Ty>
class count_ptr_ext_ref : public count_ptr_ext<_Ty>
{
public:
	typedef count_ptr_TExt<_Ty> TExt;
	count_ptr_ext_ref(TExt *_P = 0) : count_ptr_ext<_Ty>(_P) { }
};

////////////////////////////////////////////////////////////////////////////

template<class T> T* vcopy(const T *t) { return 0!=t ? (new T(*t)) : 0; }

template<class _Ty>
class count_ptr
{
protected:
	explicit count_ptr(_Ty *_P)
		: _Ptr(_P) { if(0!=_P) { ref_count=new unsigned long; *ref_count=1; } else ref_count=0; }
public:
	count_ptr() { _Ptr=0; ref_count=0; }
	count_ptr(const count_ptr<_Ty>& _Y)
		{ _Ptr=_Y._Ptr; ref_count=_Y.ref_count; if(0!=_Ptr && 0!=ref_count) ++(*ref_count); }
	count_ptr<_Ty>& operator=(const count_ptr<_Ty>& _Y)
		{if (this != &_Y)
		  {	if(_Ptr != _Y._Ptr)
				{if(0!=ref_count && (*ref_count)>0 && --(*(ref_count))==0)
					{ delete (_Ty*)_Ptr; delete ref_count; }
				 _Ptr=_Y._Ptr; ref_count=_Y.ref_count; if(0!=ref_count && 0!=_Ptr) ++(*ref_count); }
			else if(0==ref_count && 0!=_Y.ref_count && (*_Y.ref_count)>0)
				{ ref_count=new unsigned long; ++(*ref_count); }
			else if(0!=ref_count && 0==_Y.ref_count)
			{
				if((--(*ref_count))==0) delete ref_count;
				ref_count=0;
			}
		  }
		 return (*this);
		}
	~count_ptr() { if(ref_count) { if((*ref_count)>0 && 0==(--(*ref_count))) { delete (_Ty*)_Ptr; delete ref_count; } } }
	_Ty& operator*() const
		{return (*get()); }
	_Ty *operator->() const
		{return (get()); }
	_Ty *get() const
		{return (_Ptr); }
	_Ty& get_ref() const
		{return (*_Ptr); }

	void free_reference()
	{		// this->erase() entspricht dem Aufruf this=count_ptr<_Ty>()
		if(ref_count)
		{
			if((*ref_count)>0 && 0==(--(*ref_count)))
				{ if(_Ptr) delete (_Ty*)_Ptr; delete ref_count; _Ptr=0; }
			ref_count=0;
		}
		_Ptr=0;
	}

	_Ty *release_dont_delete_obj()
			// Nach der Ausfhrung von release() "hlt" das count_ptr-Objekt
			// den Zeiger nicht mehr, d.h. beim Aufruf des Destruktors des count_ptr-Objekts
			// wird nicht mehr berprft, ob es noch Referenzen auf den Zeiger gibt, und der
			// Speicher, auf den der Zeiger zeigt, wird nicht freigegeben. Das count_ptr-Objekt
			// hat keine Mglichkeit festzustellen, ob der Speicher, auf den der Zeiger zeigt,
			// von einem anderen count_ptr-Objekt, welches den gleichen Zeiger benutzt,
			// freigegeben wurde.
			// ACHTUNG: release_dont_delete_obj() ruft NICHT delete _Ptr auf, wenn das aktuelle count_ptr-Objekt
			// die letzte Referenz darstellt.
	{
		if(ref_count)
		{
			if((*ref_count)>0) --(*ref_count);
			ref_count=0;
		}
		return (_Ptr);
	}

	_Ty *release_dont_delete_obj_delete_refcount_if_last()
			// Nach der Ausfhrung von release() "hlt" das count_ptr-Objekt
			// den Zeiger nicht mehr, d.h. beim Aufruf des Destruktors des count_ptr-Objekts
			// wird nicht mehr berprft, ob es noch Referenzen auf den Zeiger gibt, und der
			// Speicher, auf den der Zeiger zeigt, wird nicht freigegeben. Das count_ptr-Objekt
			// hat keine Mglichkeit festzustellen, ob der Speicher, auf den der Zeiger zeigt,
			// von einem anderen count_ptr-Objekt, welches den gleichen Zeiger benutzt,
			// freigegeben wurde.
	{
		if(ref_count)
		{
			if((*ref_count)>0)
				if(0==--(*ref_count))
					delete ref_count;
			ref_count=0; // Auch auf Null setzen, wenn KEIN delete ref_count erfolgt ist.
		}
		return (_Ptr);
	}

	bool operator==(const count_ptr<_Ty> &p) const { return _Ptr==p._Ptr; }
	bool operator!=(const count_ptr<_Ty> &p) const { return _Ptr!=p._Ptr; }
	bool operator==(unsigned long null) const { return _Ptr==0 && null==0; }
	bool operator!=(unsigned long null) const { return _Ptr!=0 || null!=0; }
	bool is_nullptr() const { return _Ptr==0; }
	bool is_valid() const { return _Ptr!=0; }
	bool is_last_reference() { return ref_count!=0 && 1==*ref_count; }

private:
		// ===========================================================
		// ====> ACHTUNG: free_memory() mit groer Vorsicht anwenden !
		// ===========================================================
		// Bislang gibt es keine sinnvolle und ungefhrliche Anwendung von free_memory(), daher private.
	void free_memory()	// Macht ALLE count_ptr-Objekte, die auf das Objekt *_Ptr zeigen,
						// ungltig. ACHTUNG: Diese Ungltigkeit kann nur im aktuellen count_ptr-
						// Objekt *this durch _Ptr=0 vermerkt werden !
	{
		if(_Ptr) { delete (_Ty*)_Ptr; _Ptr=0; }
		if(ref_count) delete ref_count;
	}

protected:
	unsigned long *ref_count;
	_Ty *_Ptr;
};

template<class _Ty>
class count_ptr_new : public count_ptr<_Ty>
{
public:
	count_ptr_new() : count_ptr<_Ty>(new _Ty) { }
	count_ptr_new(const _Ty &t) : count_ptr<_Ty>(vcopy<_Ty>(&t)) { }
	count_ptr_new(const _Ty *t) : count_ptr<_Ty>(0==t ? 0 : vcopy<_Ty>(t)) { }
	count_ptr_new(const count_ptr<_Ty> &p) : count_ptr<_Ty>(0==p.get() ? 0 : vcopy<_Ty>(p.get())) { }
};

template<class _Ty>
class count_ptr_simple_new : public count_ptr<_Ty>
{		// Nur verwenden, wenn der Konstruktor count_ptr_simple_new(const _Ty &t)
		// NICHT mit einem Objekt t einer Klasse verwendet wird, welche von _Ty
		// abgeleitet ist. Ansonsten muss count_ptr_new verwendet werden.
public:
	count_ptr_simple_new() : count_ptr<_Ty>(new _Ty) { }
	count_ptr_simple_new(const _Ty &t) : count_ptr<_Ty>(new _Ty(t)) { }
	count_ptr_simple_new(const _Ty *t) : count_ptr<_Ty>(new _Ty(*t)) { }
	count_ptr_simple_new(const count_ptr<_Ty> &p) : count_ptr<_Ty>(new _Ty(*(p.get()))) { }
};

template<class _Ty>
class count_ptr_static : public count_ptr<_Ty>
{
public:
	count_ptr_static(_Ty *_P = 0) : count_ptr<_Ty>()
		{ count_ptr<_Ty>::_Ptr=_P; if(count_ptr<_Ty>::ref_count) { delete count_ptr<_Ty>::ref_count; count_ptr<_Ty>::ref_count=0; } }
};

template<class _Ty>
class count_ptr_ref : public count_ptr<_Ty>
{
public:
	count_ptr_ref(_Ty *_P = 0) : count_ptr<_Ty>(_P) { }
};




template<>
class count_ptr<void>
{
protected:
	explicit count_ptr(void *_P)
		: _Ptr(_P) { if(0!=_P) { ref_count=new unsigned long; *ref_count=1; } else ref_count=0; }
public:
	count_ptr() { _Ptr=0; ref_count=0; }
	count_ptr(const count_ptr<void>& _Y)
		{ _Ptr=_Y._Ptr; ref_count=_Y.ref_count; if(0!=_Ptr && 0!=ref_count) ++(*ref_count); }
	count_ptr<void>& operator=(const count_ptr<void>& _Y)
		{if (this != &_Y)
		  {	if(_Ptr != _Y._Ptr)
				{if(0!=ref_count && (*ref_count)>0 && --(*(ref_count))==0)
					{ delete _Ptr; delete ref_count; }
				 _Ptr=_Y._Ptr; ref_count=_Y.ref_count; if(0!=ref_count && 0!=_Ptr) ++(*ref_count); }
			else if(0==ref_count && 0!=_Y.ref_count && (*_Y.ref_count)>0)
				{ ref_count=new unsigned long; ++(*ref_count); }
			else if(0!=ref_count && 0==_Y.ref_count)
			{
				if((--(*ref_count))==0) delete ref_count;
				ref_count=0;
			}
		  }
		 return (*this);
		}
	~count_ptr() { if(ref_count) { if((*ref_count)>0 && 0==(--(*ref_count))) { delete _Ptr; delete ref_count; } } }
	void *operator->() const
		{return (get()); }
	void *get() const
		{return (_Ptr); }

	void free_reference()
	{		// this->erase() entspricht dem Aufruf this=count_ptr<void>()
		if(ref_count)
		{
			if((*ref_count)>0 && 0==(--(*ref_count)))
				{ if(_Ptr) delete _Ptr; delete ref_count; _Ptr=0; }
			ref_count=0;
		}
		_Ptr=0;
	}

	void *release_dont_delete_obj()
			// Nach der Ausfhrung von release() "hlt" das count_ptr-Objekt
			// den Zeiger nicht mehr, d.h. beim Aufruf des Destruktors des count_ptr-Objekts
			// wird nicht mehr berprft, ob es noch Referenzen auf den Zeiger gibt, und der
			// Speicher, auf den der Zeiger zeigt, wird nicht freigegeben. Das count_ptr-Objekt
			// hat keine Mglichkeit festzustellen, ob der Speicher, auf den der Zeiger zeigt,
			// von einem anderen count_ptr-Objekt, welches den gleichen Zeiger benutzt,
			// freigegeben wurde.
			// ACHTUNG: release_dont_delete_obj() ruft NICHT delete _Ptr auf, wenn das aktuelle count_ptr-Objekt
			// die letzte Referenz darstellt.
	{
		if(ref_count)
		{
			if((*ref_count)>0) --(*ref_count);
			ref_count=0;
		}
		return (_Ptr);
	}

	void *release_dont_delete_obj_delete_refcount_if_last()
			// Nach der Ausfhrung von release() "hlt" das count_ptr-Objekt
			// den Zeiger nicht mehr, d.h. beim Aufruf des Destruktors des count_ptr-Objekts
			// wird nicht mehr berprft, ob es noch Referenzen auf den Zeiger gibt, und der
			// Speicher, auf den der Zeiger zeigt, wird nicht freigegeben. Das count_ptr-Objekt
			// hat keine Mglichkeit festzustellen, ob der Speicher, auf den der Zeiger zeigt,
			// von einem anderen count_ptr-Objekt, welches den gleichen Zeiger benutzt,
			// freigegeben wurde.
	{
		if(ref_count)
		{
			if((*ref_count)>0)
				if(0==--(*ref_count))
					delete ref_count;
			ref_count=0; // Auch auf Null setzen, wenn KEIN delete ref_count erfolgt ist.
		}
		return (_Ptr);
	}

	bool operator==(const count_ptr<void> &p) const { return _Ptr==p._Ptr; }
	bool operator!=(const count_ptr<void> &p) const { return _Ptr!=p._Ptr; }
	bool operator==(unsigned long null) const { return _Ptr==0 && null==0; }
	bool operator!=(unsigned long null) const { return _Ptr!=0 || null!=0; }
	bool is_nullptr() const { return _Ptr==0; }
	bool is_valid() const { return _Ptr!=0; }
	bool is_last_reference() { return ref_count!=0 && 1==*ref_count; }

private:
		// ===========================================================
		// ====> ACHTUNG: free_memory() mit groer Vorsicht anwenden !
		// ===========================================================
		// Bislang gibt es keine sinnvolle und ungefhrliche Anwendung von free_memory(), daher private.
	void free_memory()	// Macht ALLE count_ptr-Objekte, die auf das Objekt *_Ptr zeigen,
						// ungltig. ACHTUNG: Diese Ungltigkeit kann nur im aktuellen count_ptr-
						// Objekt *this durch _Ptr=0 vermerkt werden !
	{
		if(_Ptr) { delete _Ptr; _Ptr=0; }
		if(ref_count) delete ref_count;
	}

protected:
	unsigned long *ref_count;
	void *_Ptr;
};

template<class _Ty>
	count_ptr<void> count_ptr_void_new()
{
	return count_ptr<void>(new _Ty);
}

template<class _Ty>
	count_ptr<void> count_ptr_void_new(const _Ty& t)
{
	return count_ptr<void>(new _Ty(t));
}


template<>
class count_ptr_ref<void> : public count_ptr<void>
{
public:
	count_ptr_ref(void *_P = 0) : count_ptr<void>(_P) { }
};



template<class T>
count_ptr<T> duplicate_all(const count_ptr<T>& p)
{
	return 0==p.get() ? count_ptr_new<T>() : count_ptr_new<T>(*p);
}*/


#endif  // #ifndef NLIST_INCLUDED
