//************************************************************************************************************
// main.cpp : Testet den Bytecode-Interpreter mit Hilfe der nach Bytecode kompilierten float64-Bibliothek.
// Wenn nderungen am Programm, welches der Bytecode-Interpreter ausfhren soll, vorgenommen werden sollen,
// muss unten zum einen das Array mem[] mit dem neuen Bytecode ersetzt werden, und zum anderen 
// muss die neue Startadresse der auszufhrenden Funktion bei der Anweisung ip.execute_bytecode( ... );
// bergeben werden (diese Anweisung steht unten in main2()).
//************************************************************************************************************

#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <avr/pgmspace.h>

extern "C"
{
#include "usart.h"
#include "avr_f64.h"
}

#define cast_real_to_largest_uint(ValueX) ((ValueX).u64=f_float64_to_long((ValueX).u64))
#define cast_real_to_largest_sint(ValueX) ((ValueX).i64=f_float64_to_long((ValueX).u64))
#define cast_largest_sint_to_real(ValueX) ((ValueX).u64=f_long_to_float64((ValueX).i64))
#define cast_largest_uint_to_real(ValueX) ((ValueX).u64=f_long_to_float64((ValueX).u64))
#define real_add(x, y)              (x.u64=f_add((x.u64),(y.u64)))
#define real_sub(x, y)              (x.u64=f_sub((x.u64),(y.u64)))
#define real_mult(x, y)             (x.u64=f_mult((x.u64),(y.u64)))
#define real_div(x, y)              (x.u64=f_div((x.u64),(y.u64)))
#define real_modulo(x, y)           (x.u64=f_mod((x.u64),(y.u64), 0))
#define real_equal(x, y)            (f_compare((x.u64),(y.u64))==0)
#define real_not_equal(x, y)        (f_compare((x.u64),(y.u64))!=0)
#define real_less(x, y)             (f_compare((x.u64),(y.u64))<0)
#define real_greater(x, y)          (f_compare((x.u64),(y.u64))>0)
#define real_less_or_equal(x, y)    (f_compare((x.u64),(y.u64))<=0)
#define real_greater_or_equal       (f_compare((x.u64),(y.u64))>=0)
#define real_greater_or_equal(x, y) (f_compare((x.u64),(y.u64))>=0)
#define real_negate(x)              ((x).u64 ^= 0x8000000000000000ULL)

#include "BytecodeInterpreter.h"

using BytecodeInterpreter::Value;
using BytecodeInterpreter::uint8;
using BytecodeInterpreter::uint16;
using BytecodeInterpreter::uint32;
using BytecodeInterpreter::uint64;
using BytecodeInterpreter::int8;
using BytecodeInterpreter::int16;
using BytecodeInterpreter::int32;
using BytecodeInterpreter::int64;


float64 _f_sin(float64 x) { return f_sin(x); }
float64 _f_cos(float64 x) { return f_cos(x); }
float64 _f_tan(float64 x) { return f_tan(x); }
float64 _f_arcsin(float64 x) { return f_arcsin(x); }
float64 _f_arccos(float64 x) { return f_arccos(x); }
float64 _f_arctan(float64 x) { return f_arctan(x); }

namespace extern_function_names { enum
{	// C/C++ - Funktionen, die vom Programm, welches auf dem Bytecode-Interpreter luft, aufgerufen werden knnen
	// ("externe Funktionen").
	printsu, printu, printsr, prints, waitforkey,

	sqrt, exp, log, sin, cos, tan, arcsin, arccos, arctan,
	abs, cut_noninteger_fraction,

	float64_to_long, long_to_float64,
	compare, fmod
}; };

void extern_function_invoker(VARIABLE_INDEX_TYPE function_number, BytecodeInterpreter::BytecodeInterpreter& ip)
{
	namespace nn=extern_function_names;
	switch(function_number)
	{
		case nn::waitforkey:
		{	// eigentlich msste hier eine Ausgabe erfolgen. Ist hier aber nicht implementiert.
			Value v;
			ip.pop_stack(&v);
			break;
		}
		case nn::printsu:
		{
			Value v;

			ip.pop_stack(&v);
			char *str=(char*)ip.var_addr_to_mem_addr(v.u16);
			usart_write_str(str);
			ip.pop_stack(&v);
			ip.cast(&v, Value::t_u64);
			usart_write_hex("  0x", (uint32_t)(v.u64>>32));
			usart_write_hex("  ", (uint32_t)(v.u64&0xffffffff));
			usart_write_str("\r\n");
			break;
		}
		case nn::printsr:
		{	// eigentlich msste hier eine Ausgabe erfolgen. Ist hier aber nicht implementiert.
			Value v;
			ip.pop_stack(&v);
			ip.pop_stack(&v);
			break;
		}
		case nn::prints:
		{	// String auf die serielle Schnittstelle ausgeben
			Value v;
			ip.pop_stack(&v);
			char *str=(char*)ip.var_addr_to_mem_addr(v.u16);
			usart_write_str("\r\n");
			usart_write_str(str);
			usart_write_str("\r\n");
			break;
		}
		case nn::sqrt:
		case nn::exp:
		case nn::log:
		case nn::sin:
		case nn::cos:
		case nn::tan:
		case nn::arcsin:
		case nn::arccos:
		case nn::arctan:
		case nn::abs:
		case nn::cut_noninteger_fraction:
		{
			typedef uint64 (*function_address_type)(uint64 float64_param);
			function_address_type funcs[]=
			{
				f_sqrt, f_exp, f_log, _f_sin, _f_cos, _f_tan, _f_arcsin, _f_arccos, _f_arctan,
				f_abs, f_cut_noninteger_fraction
				// Im enum ExternFunctionIDs mssen diese Funktionen genau in dieser Reihenfolge definiert sein, weil
				// die dem enum zugeordnete Zahl unten als Index fr dieses Array funcs[] verwendet wird.
			};
			Value v;
			ip.pop_stack(&v);
			ip.cast(&v, Value::t_u64);
			v.u64=funcs[function_number-nn::sqrt](v.u64); // Funktion aus der ANSI-C float64-Bibliothek aufrufen
			v.type=Value::t_u64;
			ip.push_stack(&v);
			break;
		}
		case nn::float64_to_long:
		{
			Value v;
			ip.pop_stack(&v);
			ip.cast(&v, Value::t_u64);
			v.i32=f_float64_to_long(v.u64);
			v.type=Value::t_i32;
			ip.push_stack(&v);
			break;
		}
		case nn::long_to_float64:
		{
			Value v;
			ip.pop_stack(&v);
			ip.cast(&v, Value::t_i32);
			v.u64=f_long_to_float64(v.i32);
			v.type=Value::t_u64;
			ip.push_stack(&v);
			break;
		}
		case nn::compare:
		{
			Value v1, v2;
			ip.pop_stack(&v1);
			ip.cast(&v1, Value::t_u64);
			ip.pop_stack(&v2);
			ip.cast(&v2, Value::t_u64);
			v1.i8=f_compare(v1.u64, v2.u64);
			v1.type=Value::t_i8;
			ip.push_stack(&v1);
			break;
		}
		case nn::fmod:
		{
			Value v1, v2, v3;
			ip.pop_stack(&v1);
			ip.cast(&v1, Value::t_u64);
			ip.pop_stack(&v2);
			ip.cast(&v1, Value::t_u64);
			ip.pop_stack(&v3);
			ip.cast(&v1, Value::VarIndexTypeNr);
			v1.u64=f_mod(v1.u64, v2.u64, (uint64*)ip.var_addr_to_mem_addr(v3.u16));
			v1.type=Value::t_u64;
			ip.push_stack(&v1);
			break;
		}
	}
}

int main2();

int main()
{
	
	for(long i=0; i<1000000; i++)
		;
	usart_init();
	for(long i=0; i<1000000; i++)
		;
	usart_write_str("START\n");
	while(1)
	{
		main2();
		usart_write_str("\r\n---------------\r\n");
	}
}

const uint8 mem[] PROGMEM =
{	// Der Bytecode: Implementiert und testet die 64-Bit-Gleitkommabibliothek
	118,   0,  17,   0, 209,   3, 152,   7, 254,   0,  20,   1,  42,   1, 160,   0, 177,   0,  12,  61, 
	 46,   0,   0,   0,   0,   0,   0,   0,   0,   0, 240,   8,  32,   9,  20,  10, 204,  10, 204,  10, 
	204,  10, 204,  10, 204,  10, 204,  10,  21,  39,  43,  56,  11,  14,  11, 114,  10,  26,   9,   0, 
	  0,   1,   0,   1,   0,   1,   0,   1,   0,   1,   0,   1,   0,   1,   0,  17,   0,  17,   0,  17, 
	  0,  17,   0,  17,   0,  17,   0,  17,   0, 177,   0, 177,   0, 177,   0, 177,   0, 177,   0, 177, 
	  0, 209,   3, 209,   3, 209,   3, 209,   3, 209,   3,  24,   6,  24,   6,  24,   6,  24,   6,  24, 
	  6,  24,   6, 152,   7, 152,   7,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48, 
	 48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48,  48, 
	  2,   1,   0,   0,   0,   2,   0,   0,   0,   4,   0,   0,   0,   8,   0,   0,   0,  49,  50,  51, 
	 46,  57,  56,  55,  32,  43,  32,  55,  51,  49,  50,  46,  53,  52,  51,  32,  61,   0,  49,  50, 
	 51,  46,  55,  55,  55,  32,  47,  32,  52,  53,  54,  46,  49,  49,  49,  32,  61,   0, 115, 105, 
	110,  40,  49,  50,  51,  46,  52,  52,  52,  41,  61,   0,  13,  10,  83, 116,  97, 114, 116, 101, 
	 32,  49,  48,  48,  32,  65, 100, 100, 105, 116, 105, 111, 110, 101, 110,  32, 109, 105, 116,  32, 
	 90, 117, 102,  97, 108, 108, 115, 122,  97, 104, 108, 101, 110,  32,  46,  46,  46,  13,  10,   0, 
	 49,  48,  48,  32,  65, 100, 100, 105, 116, 105, 111, 110, 101, 110,  32, 102, 101, 114, 116, 105, 
	103,  46,  13,  10,  13,  10,   0,  83, 116,  97, 114, 116, 101,  32,  49,  48,  48,  32,  68, 105, 
	118, 105, 115, 105, 111, 110, 101, 110,  32, 109, 105, 116,  32,  90, 117, 102,  97, 108, 108, 115, 
	122,  97, 104, 108, 101, 110,  32,  46,  46,  46,  13,  10,   0,  49,  48,  48,  32,  68, 105, 118, 
	105, 115, 105, 111, 110, 101, 110,  32, 102, 101, 114, 116, 105, 103,  46,  13,  10,  13,  10,   0, 
	 83, 116,  97, 114, 116, 101,  32,  53,  48,  32,  83, 105, 110, 117, 115,  45,  87, 101, 114, 116, 
	101,  32, 118, 111, 110,  32,  90, 117, 102,  97, 108, 108, 115, 122,  97, 104, 108, 101, 110,  32, 
	 46,  46,  46,  13,  10,   0,  53,  48,  32,  83, 105, 110, 117, 115,  45,  87, 101, 114, 116, 101, 
	 32, 102, 101, 114, 116, 105, 103,  46,  13,  10,  13,  10,   0,  84, 101, 115, 116, 101,  32,  54, 
	 52,  32,  66, 105, 116,  32,  71, 108, 101, 105, 116, 107, 111, 109, 109,  97,  98, 105,  98, 108, 
	105, 111, 116, 104, 101, 107,  32, 109, 105, 116,  32,  90, 117, 102,  97, 108, 108, 115, 122,  97, 
	104, 108, 101, 110,  46,  13,  10,  13,  10,   0, 105,  32,  61,  32,   0,  35,  35,  32,  49,   0, 
	 35,  35,  32,  50,   0,  35,  35,  32,  51,   0,  35,  35,  32,  52,   0,  35,  35,  32,  53,   0, 
	 35,  35,  32,  54,   0,  35,  35,  32,  55,   0,  35,  35,  32,  56,   0,  35,  35,  32,  57,   0, 
	 35,  35,  32,  49,  48,   0,  35,  35,  32,  49,  49,   0,  35,  35,  32,  49,  50,   0,  35,  35, 
	 32,  49,  51,   0,  35,  35,  32,  49,  52,   0,  35,  35,  32,  49,  53,   0,  35,  35,  32,  49, 
	 54,   0,  35,  35,  32,  49,  55,   0,  35,  35,  32,  49,  56,   0,  35,  35,  32,  49,  57,   0, 
	 45,  45,  45,  45,  45,  45,  45,  45,  45,  45,  45,  45,  70,  69,  82,  84,  73,  71,  45,  45, 
	 45,  45,  45,  45,  45,  45,  45,  45,  45,  45,  45,  45,  45,  45,  45,  45,  45,   0,  42,  42, 
	 42,  42,  42,  42,  42,  42,  42,  42,  42,  32,  70,  69,  72,  76,  69,  82,  32,  42,  42,  42, 
	 42,  42,  42,  42,  42,  42,  42,  42,  42,  42,   0, 101, 114, 114, 110, 114,  32,  61,  32,   0, 
	114, 110, 100, 115, 101, 101, 100,  95, 120,  32,  61,  32,   0, 114, 110, 100, 115, 101, 101, 100, 
	 95, 121,  32,  61,  32,   0, 114, 110, 100, 115, 101, 101, 100,  95, 122,  32,  61,  32,   0, 114, 
	110, 100, 115, 101, 101, 100,  95, 119,  32,  61,  32,   0,  78,  97,  78,   0,  73,  78,  70,   0, 
	 65, 242,  47, 244,  37,   0,   0,   0, 241, 135, 147, 231,  18,   1,   0,   0, 243,  15, 167,  37, 
	165,  11,   0,   0, 203, 145,  13, 206, 247, 103,   0,   0, 145,  88,  92,  99,  58,  64,   3,   0, 
	238, 213,  40,  65, 106, 193,  22,   0, 237, 134,   8, 225, 136, 136, 136,   0,  62,  29,  37, 160, 
	170, 170, 170,   2, 192, 161, 102, 171, 170, 170, 170,  10,  77,  67, 249, 255, 255, 255, 255,  31, 
	 11,  24,   0,   0,   0,   0,   0,  64, 242, 255, 255, 255, 255, 255, 255,  63, 192,  89, 148, 178, 
	  6,   3,   2,   0, 106, 109,  18, 116, 169, 200,  89,   0, 228, 184,  45,  12,  38,  40, 246,   3, 
	165, 132, 149, 216, 210, 165, 223,  17, 204,  45, 218, 250, 244, 241,  92,  37, 223, 107,  74, 125, 
	146,  72,  51,  36, 229, 145,  14,  33, 113, 233,  34,  13, 128,  24, 156, 173,  48,  94,   0,   0, 
	168, 133, 140, 208, 242, 104,  26,   0, 126, 254,  71, 231,  83, 153, 176,   1, 238, 104,   0,  73, 
	149,  75, 242,  10, 219,  39,  71, 130, 208, 247,  21,  34, 142,  39, 190, 118,  31,  88,  94,  54, 
	 60, 181, 209,  13,  75, 189, 196,  42, 230, 145,  14,  33, 113, 233,  34,  13,  21, 129, 176,  43, 
	  0,   0,   0,   0,  57,  74,  29,  26, 229, 255, 255, 255, 114, 249, 168, 241, 142,  11,   0,   0, 
	 74, 218, 200, 252, 203, 191, 252, 255,  83, 152, 135, 136, 136, 136, 136,   0, 177,  94,  85,  85, 
	 85,  85,  85, 245, 240, 255, 255, 255, 255, 255, 255,  63, 231, 113, 222, 144, 245,  81,   0,   0, 
	193, 205, 208, 201,  32,  33,  37,   0,  80, 194, 255, 129, 121, 137, 119,   2,  16, 226, 191, 179, 
	 91, 162,  69,  14, 173,   8, 118, 146,  30, 174,  96,  35, 136,  25, 208, 194,  94, 252, 211,  38, 
	200, 219,  34, 200, 130,  49, 127,  15, 217, 218,  75, 155, 131,  71,   3,   0,  77, 107, 142, 224, 
	181, 231, 130,   0, 157, 180,  46,   6, 210, 103, 107,   5, 229,  25, 170, 100, 110, 176,  85,  23, 
	152,  91, 108, 142, 141,  94, 241,  46, 174,  86, 134,   5,  53,  98, 254,  43, 206, 219,  34, 200, 
	130,  49, 127,  15,  14,   0,   4,   0,   8,   0,  17,   0,   0,   0,  35,   0,   3,   0,  12,   0, 
	  5,   0,  50,   0,  49,   0,  54,   0,  41,   0,   1,   0,  48,   0,  42,   0,  27,   0,  55,   0, 
	 56,   0,   2,   0,  39,   0,  28,   0,   6,   0,  26,   0,  31,   0,  63,   0,  53,   0, 233,  55, 
	 43,   0,  23,   0,  59,   0,  15,   0,  37,   0,  60,   0,  45,   0,  44,   0,  34,   0,  13,   0, 
	 16,   0,  38,   0,  43,  56,  22,   0,  66,   0,  24,   0,   7,   0,  21,   0, 184,  56,  67,   0, 
	130,  57, 203,  56,  71,   0,  47,   0,  11,   0, 108,  57,  51,  56,  75,   0,  70,   0,  46,   0, 
	 20,   0, 209,  56,  83,  57,  71,  56,  57,   0,  27,  57,  19,  56,  67,  57,  20,  57,  65,  56, 
	 35,  56, 122,  57,  58,  57, 237,  56, 229,  56, 222,  56, 114,  57,   3,  57, 245,  56, 215,  56, 
	177,  56,  58,  56,  69,   0,   9,   0,  89,  57,  76,  57,  58,   0,  97,  57,  33,  57, 191,  56, 
	 25,  56, 239,  55,  51,   0,  40,   0,  33,   0, 193,  57, 170,  57, 158,  57, 135,  57,  44,  57, 
	  9,  57,  65,   0,  61,   0,  18,   0, 185,  57, 176,  57, 164,  57, 151,  57, 145,  57, 253,  56, 
	160,  56,  76,  56, 250,  55,  68,   0,  64,   0,  62,   0,  29,   0,  19,   0,  73,   0,  36,   0, 
	 30,   0, 208,  14,  90,  23, 234,  22, 222,  20, 230,  23, 198,  21, 144,  19, 250,  23, 126,  27, 
	144,  32, 192,  14,  50,  50, 152,  27,  92,  45,  64,  42,  74,  21, 176,  43, 186,  52, 244,  25, 
	222,  19, 250,  28, 228,  31,  88,  27,   8,  26, 238,  29, 228,  18, 192,  19,  96,  21,  17,  51, 
	  3,  32,  49,   6, 255,  51,  69, 177,  49,  18,  64, 144,  22,   2,  50, 227,  45,  49,  70,  10, 
	 64, 134,  64,  99, 243,  63,  47, 245, 255,   7,  79, 240,  53, 255,   3, 111,  80,  53, 254,   3, 
	 99,  48,  70, 253,   3,  95, 255, 245, 253,   3, 111, 240,  54, 244,   3,  91, 119, 133,  80,  70, 
	 81, 224,  84, 255,   7,  79, 224,  52, 221,   3,  75, 208,  55, 255, 255, 255, 255, 111, 255, 240, 
	  0,   6, 221, 188,  27,  79, 108, 208, 208,  16,   9, 255, 255, 255, 255, 255, 255, 255, 127, 144, 
	  0,   0,   0,   0,   0,   0,   8,   9,   0,   0,   0,   0,   0,   0, 240, 127, 159, 255, 255, 255, 
	255, 255, 255, 255, 249,   0,   0,   0,   0,   0,   0,   0, 160, 159, 255, 255, 255, 255, 255, 240, 
	240,   9,   0,   0,   0,   0,   0,   0, 240, 255, 144,   0,   0,   0,   0,   0,   0,  15, 249,  26, 
	 97, 180,  16,  81, 237, 135, 100, 144,   0,   0,   0,   0,   0,  14,  15, 249,  53, 194, 104,  33, 
	162, 218,  15, 201, 154, 199, 156, 253,  31, 113, 119,  43,  25,  35, 219, 249, 126, 106, 220,  94, 
	 64, 145, 130, 212,  69,  79, 178,  31, 147, 249,   0,   0,   0,   0,   0,   0, 240,  63, 152, 211, 
	  5, 168, 138, 143, 100,  51,  32,   0,   8, 101,   3, 209,  64,  12, 248, 129,  14, 105,  64, 188, 
	144, 139,   2,  12,  73, 186, 105,  64,  94, 255,  43,   2,  12,  73, 186,  64, 196,   9, 207, 101, 
	  3, 230,  64,  12, 248, 129,  14, 105,  64, 124, 129, 198, 167, 239, 157, 178, 105,  64,  94, 241, 
	186,  94,  53,  63, 125,  65,  68,   9, 207, 101,   3, 250,  64,  12, 248, 129,  14, 129,  55,  64, 
	180,   9, 207, 136,  16, 225,  55,  40, 200,  64, 156, 254, 128,   1, 144, 135,   6,  38,  74, 194, 
	 63,  13,  64,  97,  48, 198,  64,  97,  39, 198, 203, 114, 175, 248,  17, 123, 203, 145,   6, 210, 
	242,  64,  97,  44, 198, 242, 254,  12, 237, 208,  44,  70,  80,  64, 132,   0, 207, 135,   6,  38, 
	 74, 194,  19,  13, 255,  84,  12,  80, 253, 208,  44,  70,  80,  67, 132,   0, 207, 101,   4,  83, 
	 64,  12, 248, 112,  98, 100, 172,  33,  48, 223, 245,  65,  69,  15, 221,   2, 196, 101,   4, 129, 
	 64,  12, 246,  80,  73, 196,   0, 207, 135,   6,  35,  42, 194,  22,  13, 143, 224, 201,  64, 181, 
	 15, 221,   2, 196, 101,   4, 202,  64,  12, 246,  80,  78,  84,   0, 207,  72, 119, 172,  26,   3, 
	  0,   3,   0, 255, 153, 193,   9,   3,   0,   3,   0, 255, 154, 193,   5,   3,   0,   6, 179,  30, 
	 16,   0, 201, 255, 160,  31, 225, 194,   6,  29, 107,  49, 193,   0,  58, 170, 169,   2, 136, 232, 
	 56,  23, 169,   8, 216, 119, 196,   7, 113,  39, 172,  47, 240,  53,   0, 208, 119,  18,  36,   0, 
	140, 130,  28, 149, 135, 244,  89, 114, 156, 149, 167,  45, 201,  91, 116,  25, 216,  63,  64, 165, 
	 15,  65, 157, 132, 244,  10,  81, 246,  80,  82,  52,   0, 253, 251, 248,  64, 196,  10, 156,   2, 
	  2,  29, 204, 127, 240,  95,  38,  80,  82, 132,   0, 253, 252,  95, 132, 142,   0,  64, 169, 192, 
	 32,  34, 220, 199, 255,   5, 160, 101,   5,  45,  64,  15, 223, 213, 248,  72, 225, 228,  10, 156, 
	  2,   2,  61, 204, 127, 240,  84, 230,  80,  83,  36,   0, 253, 255, 235, 248,  65,  68,  10, 156, 
	  2,   2,  77, 204, 127, 240,  79, 230,  80,  83, 116,   0, 234,  37, 253, 100, 151,  70, 156,   2, 
	  2,  93, 204, 127, 240,  75,  38,  80,  83, 196,   0, 234,  38, 253, 100, 151, 152, 156,   2,   2, 
	 93, 204, 127, 240,  70, 102,  80,  84,  20,   0, 234,  39, 253, 100, 152,  58, 156,   2,   2,  93, 
	204, 127, 240,  65, 166,  80,  84, 100,   0, 234,  40, 200, 128,  63,  64, 185, 192,  32,  37, 220, 
	199, 255,   3, 204, 101,   5,  75,  64,  14, 162, 156, 130,  16,  63,  64, 185, 192,  32,  38, 220, 
	199, 255,   3, 124, 101,   5,  80,  64,  14, 162, 172, 130,  32,  63,  64, 185, 192,  32,  39, 220, 
	199, 255,   3,  44, 101,   5,  86,  64,  14, 162, 188, 136,   3, 244,  17, 156,   2,   2, 141, 204, 
	127, 240,  45, 230,  80,  85, 196,   0, 234,  44, 200,  33,   3, 244,  17, 156,   2,   2, 157, 204, 
	127, 240,  40, 230,  80,  86,  36,   0, 234,  45, 200,  34,   3, 244,  17, 156,   2,   2, 173, 204, 
	127, 240,  35, 230,  80,  86, 132,   0, 234,  46, 253, 100,  18, 156,   2,   2, 173, 204, 127, 240, 
	 31, 102,  80,  86, 228,   0, 234,  47, 253, 100, 142, 210, 156,   2,   2, 173, 204, 127, 240,  26, 
	166,  80,  87,  68,   0, 234,  16,  79, 214,  65, 137, 192,  32,  43, 220, 199, 255,   1,  96, 101, 
	  5, 122,  64,  12, 255, 134,  49,  44, 143, 132, 143,  56, 156,   2,   2, 205, 204, 127, 240,  16, 
	166,  80,  88,   4,   0, 207,  53, 255, 134,  49,  60, 131, 111, 248,  73, 105, 137, 215,  11,   5, 
	240, 111, 156,   1, 130, 221, 204, 121,  46, 133, 193,  96, 101,   5, 134,  64,  12, 240, 191,  16, 
	140, 128, 191,  72, 205,  73, 192,  24,  46, 220, 199,  52, 209,   7, 118,  80,  88, 196,   0, 207, 
	 36, 200, 193, 101,   5, 178,  64,  12, 240, 119,  18,  36,   0, 140, 135,   6,  80,  93,  52,   0, 
	140, 128, 135, 101,   5, 220,  64,   8, 200,   9, 118,  80,  94, 148,   0, 140, 128, 167, 101,   5, 
	246,  64,   8, 200,  11, 118,  80,  96,  52,   0, 140, 129,   0, 194,   6,   5, 193,   1,  20,   0, 
	  4,   6,  64, 102,  55, 252, 108,  49,  44, 169, 193,  64,  98, 252, 109, 128,  54,  32, 234, 231, 
	  6,  31,  64, 111, 124,  33, 146, 161,  23,  64,  97,  21, 198,   0,  63,  21,   3,  17, 116,   6, 
	 17,  92,  96,   3, 205, 197,  80,  52,   6, 247,  17,  61,  32,   3, 203, 145,  21, 198,  16, 109, 
	 47,  39,  17,  48, 198, 242, 193,   0,   0,   2,  18,  28, 226,  34,  92, 226,  66, 156, 226, 130, 
	220, 236,  16,   1,  25,   0,  15, 244,  33, 255,  60, 185,  16,  61,  37,  15, 255,  67,  15,  33, 
	255,  63, 213,   0, 252,  16,  85,   4,   0,  14,  16,   4, 201, 225,   0,  15, 225, 194,   6,  29, 
	193,   5,   3,   0,  12, 165, 225,   0,  12, 156,  32, 160, 221, 208,  39,  12,  21,   3,  22,   0, 
	 12, 161,  16, 221,  48,   9, 195,  28,  33, 204,  13,  16,   0, 211,   0, 194,  30,   7, 225,   0, 
	 12, 150, 195,  11,   0, 113,  15, 213, 181,   7, 112, 192,  24,   0, 125, 245,   7,   0, 124,  16, 
	 86,   4,   0,  16,   0, 226, 192,  44,  16, 222,  16,   8, 206,   0,  13, 245,   0,   0, 142, 128, 
	  1,  48,  16,  32,   0,  42, 255, 139, 255, 160,  28, 227,   0,  42, 255, 188,  64,   0, 195,  63, 
	143, 227,  15, 236, 128, 201,  80,  29,  48, 143, 227,   0,   1, 212,   8, 112,   0, 137, 208,  12, 
	112, 238,   0, 137, 195,  85, 193,   3,  85,  85,  10,   0,  13,  48,   1,  48, 198, 212,  12, 211, 
	  0, 253, 239, 252,   8, 194,  42,  10,  17,  93,  48, 137, 192,  56,   0, 198, 144,   1,   0,   0, 
	  0,   0,   0,   0,  15, 213, 141,  64, 192,  12,   1,  14,  93,  48,   1,  19, 205, 247, 212,   4, 
	193,   4,  85,  85,  10,   0,   8,   0, 192,   8,   0,  64,   0,  64,  61,  48, 141,   0, 193,  21, 
	211,   8, 156,  10,   1,  26, 211,   8, 197,   1,  12,  85,  16,  98, 192,   1,  10, 208,  11,   1, 
	  1,   9, 172,  34,  48, 160,  16, 226, 194,  16,  16,   0, 192,  16, 223, 229,   0, 192,  16, 239, 
	141,  64, 204,  16,  84,  53,   8,   0,   0,  10, 217,  15, 227,   8, 217,  23,  12,  44,  49, 231, 
	 38, 159, 255, 254,   0,   0,   0,   0,   0,  12, 110, 113,  34, 189, 164,  40, 229, 209, 240, 103, 
	 33,  49, 198, 231,  18,  13, 218,  66,  30,  93,  24,   6, 114,  19,  44, 108,  33, 224, 215,  66, 
	140, 108, 177, 204,  93, 164,  36, 239, 209,  48, 103,  33,  52, 198, 194,  28,  13, 116, 247, 203, 
	 28, 197, 218,  66,  30, 253,  16,   6, 115, 192,  62, 221,   2, 114,  19,  76, 108,   2,  13, 162, 
	 33, 239, 209,   0, 108, 162,   0, 106, 195,  24, 133,   6, 140, 194,   0,  97,  35, 172,   1,   8, 
	 80,  96,   6,  18,  58, 195,  28,  17,  85,   6, 140, 192,   0,  76, 185,  17,  61,  32,   6, 203, 
	145,   6, 210, 242, 113,  19,  12, 111,  45,  64,  12,  21,   3,   0, 140, 162, 232,   0,  25,   1, 
	 50, 216,  15, 196,   0, 253,  48,  12, 109,  96, 167,   1,   9, 172,  33,  18,  80,   0,  40, 229, 
	221, 114,  18, 205, 128, 252,  64,  15, 211,   0, 198, 214,  10, 112,  16, 154, 194,  17,  37,   0, 
	  2,  30,  93, 208,  39,  12,  19,  85,   7,   0,  13,  48,  15, 212, 211,   4, 253,  77,  93,  48, 
	 15, 212, 211,   4,  18, 124, 109,  81,   3, 241, 211,   4, 253,  77,  48,   1,  39, 198, 213,  16, 
	 63,  16,   8, 253,  13,   0, 221,  48,  12, 184, 226, 192,  26, 218,  13,  48,  78,  48,   8,  16, 
	188, 109,   0, 221,  48,  76, 184, 226, 192,  26, 218,  13,  48,  14,  55,  12,  19, 149,   7,   0, 
	  0,  20,  48,  64,   0,  64,  28,  21,   3,   1, 189,  48,  14, 128,   3, 153,  80, 112, 253,  76, 
	183, 232,  18,  55, 115,   7,  28, 183, 103, 128,   0,   0,   0, 198, 237,   9, 254, 192,  12,  42, 
	176, 135,  33,   4, 205,  82,  48,  35, 224, 183, 255, 252, 108,  38, 212,  65,  28,   2,  57, 194, 
	 89,  14,  33,  18, 112,  35,  33, 205, 197,   2,  63, 235, 188, 181,  82,  48,  35, 203, 113,   8, 
	210, 119,   2,  48,  35, 203, 118, 112,   0,  32,   0,  11, 122, 203, 157,  93,  81,   3, 241, 197, 
	 33, 197,  82, 183,  77, 245,  43, 114,  17, 204, 108,  49, 104, 116, 197, 193, 114,  17, 204,  98, 
	 45,  45, 245,  43, 133,  43, 105,   0,   0,   0,   1,   0,   0,   0,   0,   2, 188, 185, 242, 113, 
	213, 113,  33, 241, 216,  31, 216,  51, 196,   1, 255, 212, 216,  53, 196,   3,  92,  34,  80,  80, 
	 51, 115, 176,  51, 172,   1,  45,  16,  53,  51,  55,  61, 237,  16,  43,  18, 192,  43, 203, 150, 
	 49, 253,  47,  37,  15, 128,  15, 197, 135,  50,  29,  35,  15,  64,  76,  85,  31, 131,  31,  48, 
	244,   1,   1, 241,  44, 198, 195,  22,   0, 252, 112, 236, 169, 181,  15,   0, 242,  27,  33, 205, 
	193,   0,   6,   3,  25,   0,   0,   0, 226, 192,  26,   0,  13, 252, 112, 224,   0, 203, 145,  14, 
	210, 218,   6,  80,  65, 240,   0, 226, 192,  22,  16,  12, 112, 236, 166,  48, 244,   5,   0, 252, 
	 25,   3,   0,   0,   0,  18, 188, 108,  21,  57, 149,  10,  83,  68, 153,   0,   2, 163,  25, 252, 
	 48,  68,   3,  42,  50, 147,   9,  48,  19,  20,  64,  56, 212,  38,  17,  80,   5, 156,  33, 166, 
	 65,  21,   0, 153, 195,  26,   1,  77,  64,  12,  16,  25, 214,  25,  17,  80,   9, 157, 112, 224, 
	 41, 214,   7,   2,  76,  32, 208, 224,   4, 212,   0,  18, 189,  64,  12,  16,   5, 208,  11, 251, 
	 17,  27, 172,  50,  96,  20, 212,   0,  33, 212,  38, 193,   0, 157,   0, 207, 211,  17,  27, 172, 
	 49, 160,   4, 212,   0, 193,   2,  76,  35,  56, 207, 211, 172,  33, 177, 192,  41,   1, 159, 188, 
	 95,  21,  25,   0, 149,   5,  49, 144,  41, 253,  60,  95, 242, 251, 173,  97,  63, 211, 157, 112, 
	176,  25,   2, 154, 194,  34,  28,   2, 144,  25, 251, 253,  44,  85,  25,  33, 212,  38,   0, 149, 
	  5,  49, 144,  41, 253,  63, 210, 227,  49, 144,   5,  16, 252,  88,   0,   4,   5, 193,  37,   4, 
	  0,   0,   4, 194,  15,  22,   0,   1,  44, 253,  80,   0,  18, 191, 246, 196, 193,   9,   3,   0, 
	  3,   0, 217,   1,  19, 205, 247, 203,  60,  16, 153,   5, 163,  25,   0,   3,  23,   0,   4,   8, 
	 80,  16,  16,  64, 133,   3, 254, 156,  34,  66,  99,   7, 128,  16,   0,   7, 100,  23,   0, 112, 
	  1, 253, 191, 208,  80,  83,   7, 246,   0,   7,  68,  23,   0,  21,   5,   0,  87,  20,  22,   1, 
	124,  16, 153,   5,  19,   0, 128,  16, 232,   1,  25,   7,   4,   8,  80,  19,   1,  33, 243,   0, 
	 29, 161,  65, 103,   0,   0,  64, 197,  17,   1,  28,  16, 153,   5,  99,  52,  73, 144,   0,  43, 
	253, 212,   3,  43, 254, 132,   3, 250, 194,  98,  64, 245, 194,  59,  62,   0, 125,   0, 112,  11, 
	194,  38,  10,   0, 112,  11, 190,  11,   4,   9, 197,  80, 120,  50, 179,  27,  64,  21,  27, 202, 
	156,  16,  43, 214,   7,   0, 124,  49,  33,  43, 193,   0, 181,   7,   2, 181,  27,   1, 189, 112, 
	223,  93,  96, 112,  43, 214,   7,   0, 188,  49,  33,  43, 193,  48,  16,   3, 243,  49, 176,   7, 
	  0,  19,   0,  64,  80,   0, 193, 153,   5, 115,  52,  73, 153,   0,   2, 191, 221,  64,  50, 191, 
	232,  64,  48,  11, 194, 105,  10, 250, 194,  60,  74, 245, 194,  35,  34, 131,  43,  64, 115,  27, 
	 64,  69,  59,  48, 113,  30,   0, 188,  93, 224,  43, 195,  18, 202, 156,  17,  43, 193,   1, 189, 
	 96, 111,  92,  49,  33,  43, 193, 254, 156,   1,  97,  49, 199,  14,  18, 220,  17,  43, 193,  51, 
	176,   7,  16, 252,  95, 233,  48,   4,   5,   0,  12,  16,   9,   3,   0,   0,   0, 253, 236, 180, 
	112, 208,  13, 112, 100,   4,  51, 172,  49,  32,   0, 193, 112,  17, 106, 192,  18, 202, 156,  16, 
	  0,  18, 225,  26, 115, 197, 210, 198, 193,   9,   3,  51,  73,   0,   8,  49,  95, 218,  64, 240, 
	  5, 100,   4,  31, 172,  49, 108, 166, 199,  58,   0,  28,   2,   0,  21, 203, 109, 252, 113,  64, 
	 21, 203, 108,  25,   3,   0,   1,  45,   0,   1,  45, 198, 157,  96, 192,   0,  19,  12, 108, 177, 
	203,  60,  25,   3,   0,   1,  45,   0,   1,  45, 198, 156, 179, 193, 153,   5,  99,  52,  73, 144, 
	  0, 143, 221,  64,  56, 254, 132,   3, 250, 214,   7,   1, 189, 113,  15,  93,  96, 112,  43, 203, 
	 28,   1,  44, 162, 193, 250, 195, 148, 245, 194,  28,  80,   0,  28,   1,  97,   5, 199,  14,  16, 
	 12, 178, 193,   0,  16,   3, 156,  49, 108, 166, 199,  42,   0,  28,   1,  97,   5, 199,  14,  16, 
	 12, 178, 193, 245, 195,  52,   0,  60,   1,  97,   0, 199,  14,  16,  92, 178, 193,   0, 124,  53, 
	224,  11, 215,   7,   2, 188,  49, 108, 166, 199,  42,   0,  60,   1,  97,   0, 199,  14,  16,  92, 
	178, 193,   0, 190, 208, 176,   1,   0,  57, 215,  11,   0, 176,   7, 172,   3,  64,   1, 192,  22, 
	 16,  92, 112, 225,   0, 203,  44,  16,   7,   0, 186, 192,  52,   0,  28,   1,  97,   0, 199,  14, 
	 16,  92, 178, 193,   1, 176,  43, 156,  49, 108, 166, 199, 106,   2, 176,  27, 174, 208, 112,   1, 
	237,  21,   1, 176,  43, 173,   0, 112,   1, 203,  28,   1,  97,   0, 199,  14,  16,  92, 178, 193, 
	 88,   5,   0,   1,  47, 202,  78,   4,  56,  18, 205,  65,   0,   0, 226, 192,  34,  33, 204,   0, 
	  0, 223,  80,   0,   0, 194,  85,   5,   0,  15, 124,   4, 104, 218,  96,  16,  64,  29,  65,  13, 
	164, 116, 240,   1,   4,   2, 197, 222, 141, 166, 218, 116,   1, 204,  87,  50,  29,  47,  13, 166, 
	 64,  44,  92, 195,  48,   2,  30, 247,  12,   3, 160,  16,  64, 125,  65,   7,  45, 240,  16,  64, 
	 44,  92, 194, 114, 193,  83,  57,   9, 163,  68,  73, 144,   0,  98,  17,   1,  10, 192,  20,  16, 
	133,  16,   1,   6,  32,  26, 192,  18,  33,  81,   2, 179,  31,  48, 191, 217,  64,  48,  11, 195, 
	 48,  16,  33,   1, 206, 134, 177,  44, 225,   1, 193, 245, 195, 140,   1, 252,   2, 166,  80,  97, 
	  1,   8,  65, 161,   1, 193,   0,  28,   1, 134, 194, 220, 113,   6, 194, 177,   1, 206, 101,   6, 
	 20,  99,  18,  65, 161,   1, 193,  48, 177,  23, 227, 133,   3,   0,  28,   1,  65,  13, 253, 192, 
	 11, 226, 194,  29,  52,   0, 189, 252, 181, 224, 105, 213,  37, 205, 223,  80, 240,  11, 203,  81, 
	 15, 213, 108,  31, 178,  92, 220, 181,  80, 243,  11,  50, 240,  15, 223,  65,  93, 232,  50, 243, 
	 31,  64,  21,  31,  48, 177,   0,  49, 244,   2, 197, 222,   0, 190,  44,  34, 144, 152,  18, 243, 
	 31,  64,  69,  31,  48, 190,   6,  67,  31,  64,  44,  93, 237,  24,  15,   0, 182,  32,  74, 208, 
	 41,   0, 178,  57, 215,  31,   1, 246, 159,   0,   0,   0,   0,   0,   0,   0,  12,  97,  47, 174, 
	115, 118,  24, 105, 204, 204, 204, 204, 204, 204, 204, 204,  49, 244,   4,  81, 243,  11,  49, 244, 
	  2, 224,  99, 190,  61,  16,  15,   0,  53,   5, 209,   0,  51,  16,  33, 254, 162,  27, 194,  97, 
	 21,  16,  32,   3, 233, 206,   0, 190,  44,  53,  32,   3, 233,   1, 241,  18,   0, 188,  95, 242, 
	 49, 241,   0,   0, 187, 229,  17,  69,  11, 209,   0,  56,  18, 243,  31,  64,  69,  31,  48, 190, 
	  6,  67,  31,  64,  44,  93, 224,  11, 226, 215,  21,   1, 241,  18,   0, 191, 210,  98,   5, 172, 
	 63,  96,   3, 216,   7, 196,   0,  94, 232,   7, 172,  34, 241,  16,   7, 233, 201,  99,  57, 156, 
	 33,  49,  64,   7, 233,  33, 222, 253, 129,  16,  32,   7, 233, 206,   0, 112,   5, 156,  53, 237, 
	 16,  15,  99,  49, 238,   0, 126, 156, 238, 224,   7,   0,  58, 194,  16,  17,  16,  32,   7, 233, 
	206,   0,  95,  12, 181, 233, 201,   0,  94, 156, 225,  12,   0,  95,  12, 181, 233, 206,   1,  44, 
	  1,   8, 253, 198,  52,  95, 220, 202,  32,  15, 172,   1, 102,  50, 191, 220,   0,  62, 144,  15, 
	 72, 162, 160,  20, 192,  22,   0, 253,  65,  65,   1, 193,  51, 144, 113,  64,   2,   0,   1,  44, 
	198, 192,  32,   1,   1,   0, 197, 199,  20,   1,  12, 182, 203,  55,  15,  12, 179, 202,  94, 128, 
	 34,  51,   8, 253,  16,   5,  98,   3, 173,   0, 151,  39,  41, 194, 184,  33, 115, 117, 172,   3, 
	  3,   3,  33, 117,   0,   4,   9, 204,  47, 216,   3,   3,  33, 118,   0,   4,   9, 204,  32,   0, 
	253, 225,  21, 157,   0, 192,   0,  18, 220, 108,  49,   7,  44,  16,   3,   1,  45, 241,   0, 197, 
	172,   1,  15, 216,   7,  76, 195,   0,   4,   8, 192,  32,   1,   1,   0, 197, 199,  20,   1,  12, 
	182, 204,  64,   3, 226, 194,  19,  32, 218,  96,   3,  16,  12,  93, 224,   3, 240, 117, 172,   1, 
	 13, 216,  77,  16,   5,   0,   1,  45, 198, 237,  13,   0,  48,  18, 223, 173,  96, 160,   3, 115, 
	172,  47, 240,  24,  79, 240,  34,  64,   0,  64, 140,   1,  13, 208,  32,   3, 226, 194, 158, 255, 
	  1, 110, 116,  65,  63, 209, 202,  46,  24,   5, 174, 208, 193,   2, 249, 201, 158, 112, 242, 152, 
	  0,  87, 139, 206, 217,  71,  82,  27, 206, 238,   0,  92,  76, 162,   0,  90, 194,  32,  13,   0, 
	 87, 123, 201,   0,  80,   3, 197, 120, 188, 237,  24,   5,  16,  47, 227,  65,  12, 117,  33, 188, 
	226,  47, 209, 238,   0,  62,  44,  33,  48, 209,   2, 225,   0,  87, 139, 206,  33, 253,  16,   3, 
	  0,  90, 231,  68,  13,   0,  95,  12, 181, 117, 188, 159, 159, 225, 195,  78,   0,  48,   5, 174, 
	113,  48, 209,   2, 225,   0,  87, 139, 206, 143, 156, 239, 216,  13,  16,   5,   0,  95,   0,  16, 
	173,   0, 176,   5, 115, 228, 194,  74,  18,  16, 207, 156, 231,  68,  19, 253,  28, 162, 225, 128, 
	 90, 237,  12,  16,  47, 156, 153, 231,  15,  41, 128,   5, 120, 188, 225,  12, 249, 201, 156,  49, 
	 40, 249, 206, 128,   5, 116, 188, 224,   0,  64, 140,   1,  13, 216,  32,  16, 224, 100, 197,  80, 
	 80,   0,  64, 140,   1,  45,  24,   5,   0,  15, 222,  80,  48,   3,  17, 106, 192,  18, 209, 128, 
	 86,  64,  66,  16,   3, 173, 112, 176,   3,  18,  90, 192,  18, 209, 128,  80,   5,  98,   1, 172, 
	  1,  34,  21,   5, 136,   0,  80,   0,  64, 159, 253,  12,  64,  16, 113,  65,  58, 194,  24,  51, 
	238, 128,  86,  32,  26, 192,  16, 253, 128,  34, 253,  16,   5, 113, 228, 208,  11,  99,  69, 254, 
	125,   0, 182,  54,  95, 231, 194,  11,  57, 209,   0,  80,   5,  80, 126, 232,   5,  98,   4, 173, 
	  1,  96,   0,  64, 141, 112, 192,   5,  98,   3, 173, 112, 161,   2, 254, 126, 112, 101,  62,  16, 
	  7, 113, 228, 238,   0,  87,  43, 254,  28,  32,  98, 119,  12,  21,  80,  80,  28,   0,   8, 142, 
	128,   5,  73, 148, 144, 217,  46, 209,  93, 146,  16,  57, 208,  10, 217,  34, 153, 231,  10,  47, 
	221,   4, 217,  38,  54, 233, 208,  12, 217,  38,  52, 233, 214,  33, 114,  33, 228,  99,  97, 157, 
	  0, 247,  34,  30,  70,  52,  25, 254,  45,   2,  23,  34,  46,  70,  54, 233, 208,  15, 114,  34, 
	228,  99,  78, 159, 226, 192,  32, 221,  34, 199, 255,   4,  40, 217,  33,  13, 156,  33,  97,  45, 
	148,  99,  43, 156,  49,  13, 208,  66,  28, 193, 221,   2, 217,  38,  54, 153, 208,  12, 217,  38, 
	 52, 153, 214,  33, 114,  33, 228,  99, 110, 157,   0, 247,  34,  30,  70,  52, 233, 254,  45,   2, 
	 23,  34,  46,  70,  54, 105, 208,  15, 114,  34, 228,  99,  70, 159, 226, 192,  86,   0,  76,   2, 
	 78,   6,  60, 181, 115, 189,  64,  71,  28,   1,  97,  49, 199,  14,  18, 220,  24,  82, 120,  81, 
	 50,  53,   3,  16,  12,  43,  64,  93, 148,  16, 201, 195,  60, 221,   4, 114, 192,  30, 255,  28, 
	127, 240,  38, 226,  28, 194, 217,  70,  35,  10, 215,  12,  98,  57, 217,  90, 192,  16, 253, 128, 
	 98,  59,   0,  58, 194,  17, 160, 116, 195,  18, 209,   2, 119,  44,   1,  45,  24,  39, 129,  47, 
	 49,  52,   4,  81,  48,  19, 192,  18, 209,  48,  51,  19, 217,  92, 185,  16,  44,  81,  18,   0, 
	 60,  93,  45, 224,  19, 192,  30,  48,  51,  19,  64,  46,  61, 208,  77, 146,  99,  69, 157,   0, 
	205, 146,  99, 101, 156,  57, 131,  39, 254, 192,  52, 137, 254, 222, 217,  33,  13, 157,   0, 205, 
	146,  99,  43, 156,  49,  13, 208,  45, 146,  98,  48, 173, 112, 198,  35, 157, 147, 174, 112, 162, 
	125, 208,  67,   3,  50,  48,  39,  65,  95,  13, 232,  50,  51,  19,  64,  21,  19,  49,  49,  31, 
	  0,  59, 115,  51, 116,   5,   0,  76,   1,  71,  45,  64,  64,  55, 193,   0,  76,   1,  71,  45, 
	 64,  65,  43, 193,   3,  83,  57,  11,  57, 147,   0,   0,  16, 194, 100,  42, 141, 128, 252,  64, 
	 24,  16, 188, 110, 208, 160,  16, 253,  12,   2,  67,  15, 211,  20, 255, 124, 113, 195,  15, 211, 
	 20, 254, 172,  78,  23,  20, 196, 238, 129,  12, 177, 194,  24, 125,   1, 143, 208,  48,   3,  15, 
	 64,  21,  15, 105,   8,   0,   0,   0,   0,   0,   0,   0,  80, 240,  18, 195,  18,   0, 252,  24, 
	216,  31, 196,   1, 129,  11, 198, 237,  10,   1,  47, 208, 192,  36,  49, 253,  49,  79, 247, 199, 
	 28,  49, 253,  49,  79, 234, 196, 225, 113,  76,  78, 232,  18, 203,  28,  33, 135, 208,  24, 253, 
	  3,   0,  49, 244,   1,  81, 248,  82,  17,  50,   1, 252, 110, 114, 112, 243,  31,  40, 229,  19, 
	 32,  15, 198, 194,  11,  14, 209, 114,  19,  15,  40, 229,  18, 192,  31, 198, 231,  39,  15,  49, 
	242,  30,  81,  44,   0, 252, 108,  32, 176, 237,  16,  33,  48, 242,  30,  83,  31,  64, 117,  31, 
	131,  31,  48, 244,   1,   2,  29,  44,  21,  84,  57,  11, 115,  52,  73, 153,   0,   2, 191, 221, 
	 64,  48,  26, 192,  18, 141,  65, 160,   7, 195,  18, 202, 156,  31, 173,   0, 177,  21,   1,  41, 
	208,  13,   1,  45,  49, 111,  44,  50, 224,  26, 192,  22,  18, 189,  65, 161,  43, 193,  42, 216, 
	  3, 196,   0, 112,  27, 242, 237,  34,   1,  32,   7, 173,  97, 112,   7,   1,  41, 215,  12,   1, 
	189,  49, 106, 214,   9, 238, 128,  60,  43, 118,  88,   1, 100,   7,  37, 197,  33, 210,  49, 180, 
	  4,  37, 197,  83, 176,   7,   1,  41, 215,  12,  18, 192,  59, 198, 195,  20,  18, 197,  59,   0, 
	112,  18, 197,  80, 182,  35, 240,  11, 172,  51,  35,  59,  18, 225,  18,   0, 188,  93,  47, 246, 
	196,  49, 184,   1,  99,  59,  64,  18,  29,  46,  48,  26, 192,  86,  51, 177,  31,   0, 187,   0, 
	 16,  16, 253, 179,  43,  64,  80,  43, 211,  26,  64, 205,  65, 163,   7,  49, 180,   2, 227,  49, 
	176,   7, 224, 107, 197,   0,  19,   0,  64,  80,  27, 212,  22,   0,  60,  49,  44, 169, 193,   0, 
	 12,  21, 153,   7,  51,  73,   0,   2, 179,  21, 252,  49,   4,   3,   2,   3,  21,   0,  80,   1, 
	  0,   4,  16, 193, 153,   5,   1,  28, 169, 202, 158, 128,   3, 153,  48,  99, 128, 216,  33, 196, 
	  2,  28,  46,  96,  93,  25,  33, 114,  80, 247,  53,  31,   2,  17,   9, 172,  34, 162,  64,  31, 
	 16, 160,  33, 210, 223, 172,  49,  45,  16,  15,  49, 241,  10,   2,  29,  45, 227,  15,  16, 160, 
	 33,  17,  28,  93,  45, 237, 163,  33, 239, 114, 247, 192,  26, 218,  49,  44, 253,  93, 162,  33, 
	239,   0, 240,   0, 173,  97,  96,  15,   0,   9, 215,  11,   1,   0,  31, 172,  58,  67,   0,   0, 
	254,  48,  16,   1, 250, 192,  18, 209, 128,   3,  16,   1, 254,  48,  33,  16, 154, 194,  20,  38, 
	218,  81,  10,   2,  29,  47, 213, 218,  33,  10,   2,  17,  17, 197, 210, 253,  80,   0, 214,  10, 
	  1,   7,  42, 214,  10,  18, 231,  41, 192,  26, 113,  33, 188, 112, 199,  28,  25,   3,  67,  73, 
	144,   0,  43,  49,  95, 218,  64,  48,   5, 195,  18, 202, 156,  17,  21,   0,  89, 195,  56,   1, 
	 93, 112, 112,   1, 195,  18,   0,  12,  17,  43, 193,   0,  28,   1,  33,  43, 193, 133,  37,   0, 
	 95, 208, 192,  52,   1,  95, 124,   1,  65,  44,  82,  83,  21,  33, 239,   2,  80,  21,  64, 213, 
	 21,  49,  80,   5, 240,  16, 191, 235, 224, 176,  31,  75, 131,   0,  64,  80,   0, 193,   9,   3, 
	131,  52, 153,  73, 144,   0, 131,  59,  48, 127, 217,  64,  48,   7, 195,  18,  19, 156,  31, 172, 
	 52,  96,  59, 214,  14,   3, 189, 112, 112,   1, 195,  18,   0,  12,  28, 169, 193,   0, 118,  64, 
	 64, 154, 195,  46,   0,  28,   1, 108, 169, 199,  14,  18, 220,  17,  54,  83, 179,  23,  51, 177, 
	 25, 128,   0,  65,  36,  16,  82, 118,  52,   3,  75,  48, 115,   3,  50, 116,  15, 100,   3, 190, 
	  0, 122, 194,  23,   8,  51, 190,  11,   3, 254,   0, 124,  94, 248,  83, 176,  23,  65, 133,  43, 
	  0,  60,  32, 162,  40,  83, 176,   7,  17, 106, 195,  20,  18, 229,  59, 134,  80,  97, 136,  44, 
	  3, 180,  14,  83, 176,   1, 192,  48,  50, 179,  59,  64,  46,  51,  59,  64, 117,  59,  51, 176, 
	  1, 192,  40, 224, 176,  63,  48,  43, 197, 199,  22,  17, 128,  43, 184,  49, 116,   5,   1, 124, 
	 25,   3,  19,   1,  76, 162, 232,   1,  68,  73, 144,  43,  49, 159, 218,  64,  48,   5, 195,  18, 
	 19,  28,  17,  21,   0,  89, 237,   7,   0,  28,   3, 128,  25, 215,   7,   0,  28,  49,  32,   0, 
	193,  18, 188,  17,  25,   0,  89, 215,  26,   1, 150, 158,   0,   0,   0,   0,   0,   0,   0,  10, 
	194,  68,  30,   0,  81,  30, 197,  82, 144,   5,  17, 106, 194,  21,  32,   2, 157, 245,  41,  98, 
	254, 204,  65,  22,   0,  90, 192,  18,  33, 204,  35,  25,  33, 229, 128,  25, 197,  81, 145,  20, 
	204,   1,  24,  49, 148,   2, 197,  80,  80,  25, 192, 120, 129,  20, 114, 156,  49, 142,   6,  60, 
	112, 236, 166, 101,   6, 120,  40,  39,   1, 145,  24,   0,  95, 210,  64, 227,  25,  64,  69,  25, 
	112, 208,  10,  17,  71,  25, 194, 113,  12,  17, 163,  41,  64,  44,  85,   9, 129,  54,  50, 148, 
	  4,  82, 159, 186, 194,  16,  64,  49, 159, 211, 197, 229, 251, 197,  16, 154, 195,  26, 141, 129, 
	156, 113, 163,  25, 251, 197, 235, 196, 114, 226, 194,  17,  16,   2, 144,  25, 197,  81, 147,  25, 
	  2, 157, 224,   5,  80, 147,  25,   0, 151,  46,  44,   1,  97,   0, 199,  14, 202,  99,   0,  64, 
	 80,   0, 193,  57,   5,  83,  73, 147,   0,  72, 246, 156,  49,  97,   0, 199,  14, 202, 108, 179, 
	131,  21, 253, 164,   3,   0,  92,  50, 239, 105, 195,  22,  19, 156, 112, 236, 169, 193,  17,  80, 
	  5, 156,  49,  33,  43, 193,   0,  29,   0, 127, 105, 192,  22,  16,  12, 112, 236, 166, 216,  39, 
	204,  17,  53,  82,  88,  50,  86,  80,  64,  24,   0,   4,  18,  65,   5,   0,  99,  61,  49,  95, 
	218,  64, 240,  37,  81,  86,  64,  60,  16,   5, 172,  33, 112, 131,  37, 224, 176,  64,  16,   5, 
	197, 239, 133,  37,   2,  81,  51, 172,  50,  99,  39,  33, 243,  50,  81,  51, 227,   2,  81,  58, 
	172,  52, 141, 161,  33, 243, 246, 156,  49,  99,  39,  33, 243,  19,  48,  37, 197,  82,  80,  37, 
	105,  25,  33, 251,  84,  68,  45,  24,  70, 172,  50, 205, 160,  33, 243,  19, 160,  37, 197,  82, 
	 83,  37,  35, 229,   0,  81,  36, 172,  32, 225,  67,   5,  17, 158,  48,  37,  81,  88,  80,  83, 
	  5,  49,  84,   2, 227, 130,  22,  80, 111,   8,  39, 131,  37,  50,  84,   1,  64, 227,  21,  64, 
	 69,  21,  18, 192,  21, 198, 192,  16, 133,  21, 112, 214,  10,  34,   1,   9, 195, 255,   1, 140, 
	  1,  85,  37,  98, 192,   0,  90, 194,  16,   8,  49,  80,   5, 223, 239, 133,  21,  18, 225,  46, 
	131,  21,  49,  84,   1,  36, 210, 197,  64, 213,  21,  34,   1,   9, 194,  13, 218,  98, 254,  80, 
	 94, 129,  66,  85,  13,  17,   5, 112, 194,  26,  24,  49,  85,  43,  50,  85,  47,   0,  93, 245, 
	  5,  50,  85,  43,  49,  85,  47, 211,  47, 195,  18,  18, 188,  19,   5,   2, 180,   2, 227,  48, 
	 80,  47,  64,  45, 232,   2, 244,   7,   2, 180,   4,  81,  87,  21,  39,  49,  81,  24,   0,  91, 
	  2, 115,   0,  64,  80,   0, 193,  57,   5,  19,   1,  56, 232,   1,  67,  73, 144,  43,  49, 115, 
	  7, 253, 148,   3,   0, 124,  50, 239, 105, 195,  22,  19, 140, 112, 236, 169, 193, 250, 195, 168, 
	  1, 124,   1,  32,   0, 193,   0,  28,  34, 178, 194,  32,  16, 156,  51,  70, 155, 255, 146,  31, 
	181,  68,  66, 209, 140, 112, 225,  43, 193,  34,   1,   9, 195,  22,  19, 140, 112, 225,  43, 193, 
	 34,   1,   9, 192, 255,   2, 204,   0, 113,  22, 173,   2,  80,   7,  17, 153, 215,  26,   1, 118, 
	155,  80,  79,  51,  47, 157, 230,  72,  74, 194, 101, 255,   1, 154,  99,  64,  50, 115,   7, 253, 
	148,  15,  18, 232,  50, 115,  39,  64,  28,  85,  39,  50, 116,   7,  82, 120,   2, 116,  13,  82, 
	120,  50, 115,  23,  64,  21,  23, 209,   0, 127, 105, 195,  22,  16,  12, 112, 236, 166, 204,  33, 
	 22,   0, 122, 215,  22,   0, 113,  23, 157,  96, 177,  44,   1, 122, 192,  18,  18, 188,  16,   7, 
	 17, 121, 194, 130,  20, 128,  23, 197,  82, 120, 216,   3, 196, 105,  96,   0,   0,   0,   0,   0, 
	  0,   0,   2, 124, 110, 113,  34, 211,  39,  34, 229, 209,   0,  51,  23,  64, 117,  23, 136,   2, 
	114,  29,  40,  50, 115,  39,  64,  17,  11,   0,  61,  92, 220,  84,  13,  49, 116,   4,  81, 126, 
	 11,   4,   0,   0,  60,  85,   7, 133,  23,  17, 213,   7,   1,  12,  49,  97,   0, 199,  14, 202, 
	108, 192,   1, 124,   1, 227,   7,  49, 116,   2, 227, 209, 128, 112,   7,  18,  74, 195,  82,  49, 
	116,   7,  81, 126,  11,   7, 252,   0, 124,  83,  23,  64,  44,  85,   7,  33, 204,   1,  25,   0, 
	121, 194, 184,  42, 136, 101,   7,  40,  39,  39,  18,  80,   7, 172,  49, 108, 169, 199,  56, 131, 
	 23,  49, 116,   1,  34,  17, 176,   7, 197, 213, 205,  64, 227,  23,  64,  69,  23, 114, 192,  96, 
	 19,  49,  38,   0, 122, 195,  22, 202, 156, 114,   0,  23,  17, 176,   7, 253,  44,  85,  23,  17, 
	213,   7, 246, 157,  96, 112,   1, 192, 104,  19,  81,  38,   0, 122, 195,  22, 202, 156, 114,   0, 
	 23,  17, 176,   7, 253,  44,  85,  23,  17, 213,   7, 133,   1, 105,  25,  33, 251,  84,  68,  45, 
	 24,  71,  81, 115,  23,   0, 126,   6, 140,  80,   1,  48,   4,   5,   0,  12,  18,  60, 138, 151, 
	 49,  16, 255,  11, 169, 108,  99, 115, 197, 203,  82, 143, 240, 228, 254,  12, 154, 146,  92, 154, 
	156, 240,  15,   1, 250, 144,  79,   3, 250, 156, 194, 255,  26, 156, 240,  63, 169, 200,   3, 250, 
	146,  28, 106, 156, 219, 169,  18, 162,  28, 157,  81,  41, 178,  28, 239, 230,  45, 254, 255, 230, 
	 16, 140, 223,  63, 230,  37, 254, 241,  33, 254,  95, 238, 254, 223,  18, 156, 225,  32, 254,  79, 
	238, 254,  65,   4, 241,  45, 206, 169, 201, 244, 191, 237, 210, 191, 228, 186, 159, 215, 196, 169, 
	202,  42, 170, 145,   6, 205,  17,  92, 106, 158, 108,  74, 158, 188,  74, 145,   3, 205, 169,  16, 
	 12, 106, 147,   5,  48,  26, 144,   9,   0,  90, 144,   5,   0, 154, 156, 221, 234, 156,  92, 218, 
	144,   1,   0,  63, 219, 169,  33,   1,  10, 158, 204,  74, 147,  27,  48, 127, 217, 169,  50, 179, 
	 11,  48,  51,  16, 169,  17,  80,   7, 154, 145,  21,   0, 185, 169, 252,  48,  10, 145,   0, 186, 
	147,   1,  48,  10, 158,  16,   3, 233, 206, 169,  16,  27, 169, 216,   5, 196, 169,   0,  87, 107, 
	169, 188, 154, 144,   5, 114, 228, 154, 146,  45, 186, 157,  81,   7, 169,  37, 244, 169, 210, 243, 
	169,  41, 201, 169,  18, 140, 109,  90, 159, 229,  16,  74, 146, 220, 154, 144
};

uint8 get_code_nibble(PROGRAM_COUNTER_TYPE pc)
{
	return (pc&1) ? ((uint8)(pgm_read_byte(&mem[pc>>1])&15)) : ((uint8)((pgm_read_byte(&mem[pc>>1])>>4)&15));
}

uint8 get_code_byte(PROGRAM_COUNTER_TYPE pc)
{
	//gleichwertig wre: return (get_code_nibble(pc)<<4)+get_code_nibble(pc+1);
	if(0==(pc&1))
		return pgm_read_byte(&mem[pc>>1]);
	else
		return (((uint8)(pgm_read_byte(&mem[pc>>1])&15))<<4) + ((uint8)((pgm_read_byte(&mem[(pc+1)>>1])>>4)&15));
}


int main2()
{
	// Die Gre von Stack (stack_size) und Speicher fr globale Variablen (memory_size) wird
	// im Folgenden festgelegt. Die bentigte Speichergre fr globale Variablen wird vom Compiler ermittelt
	// (Compiler::required_memory_size nach der Compilierung), und die bentigte Stackgre wird von Bytecode-Interpreter
	// nach der Ausfhrung des Programms bereitgestellt (BytecodeInterpreter::get_maximum_stack_usage(), sofern
	// folgendes #define in BytecodeInterpreter.h ausgefhrt wurde: #define CALCULATE_MAXIMUM_STACK_USAGE .
	const uint16 stack_size=512, memory_size=1024;
	static uint8 stack[stack_size], memory[memory_size];

	BytecodeInterpreter::BytecodeInterpreter ip(stack, stack_size, memory, memory_size, get_code_nibble,
		get_code_byte, &extern_function_invoker);
	ip.execute_bytecode(3335); // Funktion Main
	return 0;
}
