// BytecodeInterpreter.h

//***********************************************************************************************************//
//***********************************************************************************************************//
//                                                                                                           //
// BytecodeInterpreter.h, BytecodeTargetedCompiler.h, language_defines.h, language_undefines.h, utils.h:     //
// These files contain a bytecode interpreter and a compiler converting a C-like program into bytecode.      //
//                                                                                                           //
// Copyright 2009 by Florian Knigstein, mail@virgusta.eu                                                    //
//                                                                                                           //
//      Commercial use of this code without my permission is prohibited.                                     //
//      You may use this program free of charge for non-commercial purposes provided you agree to the        //
//      following License:                                                                                   //
//                                                                                                           //
//      The Original Work is provided under this License on an "AS IS" BASIS and WITHOUT WARRANTY,           //
//      either express or implied, including, without limitation, the warranties of                          //
//      non-infringement, merchantability or fitness for a particular purpose. THE ENTIRE RISK AS TO THE     //
//      QUALITY OF THE ORIGINAL WORK IS WITH YOU. This DISCLAIMER OF WARRANTY constitutes an essential part  //
//      of this License. No license to the Original Work is granted by this License except under this        //
//      disclaimer.                                                                                          //
//                                                                                                           //
//      Limitation of Liability. Under no circumstances and under no legal theory, whether in tort           //
//      (including negligence), contract, or otherwise, shall the Licensor be liable to anyone for any       //
//      indirect, special, incidental, or consequential damages of any character arising as a result of      //
//      this License or the use of the Original Work including, without limitation, damages for loss of      //
//      goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages    //
//      or losses. This limitation of liability shall not apply to the extent applicable law prohibits such  //
//      limitation.                                                                                          //
//                                                                                                           //
//***********************************************************************************************************//
//***********************************************************************************************************//

// ***** ATTENTION *****
// This code is not well-tested. There may be several bugs.

#ifndef BytecodeInterpreter_h_included
#define BytecodeInterpreter_h_included


#define MSCPP // Nur definieren, wenn unter Microsoft Visual Studio compiliert wird

#ifdef MSCPP
#pragma warning( push )
#pragma warning( disable : 4244 ) // Warning: conversion from ... to ... ==> possible loss of data
#endif

namespace BytecodeInterpreter
{

#define CALCULATE_MAXIMUM_STACK_USAGE

#define USE_CODE_MAKRO_COMPRESSION
//#define ALLOW_BRANCHES_INSIDE_MAKROS

#define DEBUG_MODE

#define ASSUME_VALUE_UNION_EQUAL_ADDRESSES

#define FUNCTION_ADDRESS_ALIGNMENT_NR_OF_RIGHT_ZERO_BITS 1

#define USE_TABLE_FOR_CONSTANTS

//#define USE_STACK_TYPE_INFO_VARIANT

//#define WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION

#define USE_ASSERT
#ifdef USE_ASSERT
	#define ASSERT(condition, message, pc) { if(0==(condition)) { throw RuntimeError(message, pc); } }
#else
	#define ASSERT(condition, message, pc) ;
#endif

// Wenn PRINT_COMMANDS_AT_RUNTIME definiert ist, mssen unten das Makro PRINT_COMMAND_IF_ENABLED sowie die
// Funktion bi_print_value() ggf. angepasst werden.
//#define PRINT_COMMANDS_AT_RUNTIME

struct RuntimeError
{
	const char *message;
	unsigned long pc;
	RuntimeError(const char *_message="", unsigned long _pc=(unsigned long)(-1)) : message(_message), pc(_pc) { }
};

typedef unsigned char uint8;
typedef unsigned short uint16;
typedef unsigned long uint32;
typedef unsigned long long uint64;
typedef signed char int8;
typedef short int16;
typedef long int32;
typedef long long int64;

#define BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE   // Auf Little Endian Maschinen wird kompakterer Code generiert (sofern dann auch 
	// BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE definiert ist). WARNUNG: Der Konstruktor der Klasse BytecodeInterpreter fhrt im Falle
	// eines definierten BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE eine Laufzeitberprufung durch, ob es sich tatschlich um eine Little
	// Endian Maschine handelt. Falls dies nicht zutrifft, wird absichtlich eine Endlosschleife ausgefhrt
	// (als eine Art "Fehlermeldung").

#define MAX_NR_BYTES_FOR_VARIABLE_INDEX 2  // Darf nur 1, 2 oder 4 sein
#define MAX_NR_BYTES_FOR_PROGRAM_COUNTER 2  // Darf nur 1, 2 oder 4 sein

#define WITH_INTEGER8
#define WITH_INTEGER16
#define WITH_INTEGER32
#define WITH_INTEGER64

#define WITH_REAL
#define WITH_REAL_ADDITION
#define WITH_REAL_SUBTRACTION
#define WITH_REAL_MULTIPLICATION
#define WITH_REAL_DIVISION

#define WITH_INTEGER32_MULTIPLICATION
#define WITH_INTEGER32_DIVISION
#define WITH_INTEGER32_MODULO
#define WITH_INTEGER64_MULTIPLICATION
//#define WITH_INTEGER64_DIVISION
//#define WITH_INTEGER64_MODULO


typedef uint64 real_type;
#define real_type_is_integer_type

//ifndef cast_real_to_largest_uint
//	#define cast_real_to_largest_uint(ValueX) ((ValueX).u64=(uint64)(ValueX).r)
//#endif
//ifndef cast_real_to_largest_sint
//	#define cast_real_to_largest_sint(ValueX) ((ValueX).i64=(int64)(ValueX).r)
//#endif
//ifndef cast_largest_sint_to_real
//	#define cast_largest_sint_to_real(ValueX) ((ValueX).r=(real_type)(ValueX).i64)
//#endif
//ifndef cast_largest_uint_to_real
//	#define cast_largest_uint_to_real(ValueX) ((ValueX).r=(real_type)(ValueX).u64)
//#endif

#ifndef real_add
	#define real_add(ValueX, ValueY) ((ValueX.r)+(ValueY.r))
#endif
#ifndef real_sub
	#define real_sub(ValueX, ValueY) ((ValueX.r)-(ValueY.r))
#endif
#ifndef real_mult
	#define real_mult(ValueX, ValueY) ((ValueX.r)*(ValueY.r))
#endif
#ifndef real_div
	#define real_div(ValueX, ValueY) ((ValueX.r)/(ValueY.r))
#endif
#ifndef real_negate
#define real_negate(ValueX) ((ValueX).r=-(ValueX).r)
#endif
//#ifndef real_modulo
	//#define real_modulo(ValueX, ValueY) (fmod((ValueX.r),(ValueY.r)))
//#endif

// Es knnen ggf. real_and(ValueX, ValueY), real_or(ValueX, ValueY), real_xor(ValueX, ValueY), real_shift_left(ValueX, ValueY), real_shift_right(ValueX, ValueY)
// definiert werden (obwohl dies in der Regel nicht sinnvoll ist). 

#ifndef real_equal
#define real_equal(ValueX, ValueY) ((ValueX.r)==(ValueY.r))
#endif
#ifndef real_not_equal
#define real_not_equal(ValueX, ValueY) ((ValueX.r)!=(ValueY.r))
#endif
//#ifndef real_less
	//#define real_less(ValueX, ValueY) ((ValueX.r)<(ValueY.r))
//#endif
//#ifndef real_greater
	//#define real_greater(ValueX, ValueY) ((ValueX.r)>(ValueY.r))
//#endif
//#ifndef real_less_or_equal
	//#define real_less_or_equal(ValueX, ValueY) ((ValueX.r)<=(ValueY.r))
//#endif
//#ifndef real_greater_or_equal
	//#define real_greater_or_equal(ValueX, ValueY) ((ValueX.r)>=(ValueY.r))
//#endif

#ifndef uint64_mult
#define uint64_mult(x, y) ((x)*(y))
#endif
#ifndef uint64_div
#define uint64_div(x, y) ((x)/(y))
#endif
#ifndef uint64_modulo
#define uint64_modulo(x, y) ((x)%(y))
#endif
#ifndef int64_mult
#define int64_mult(x, y) ((x)*(y))
#endif
#ifndef int64_div
#define int64_div(x, y) ((x)/(y))
#endif
#ifndef int64_modulo
#define int64_modulo(x, y) ((x)%(y))
#endif

#define WITH_OPCODE_LT
//#define WITH_OPCODE_LE
//#define WITH_OPCODE_GT
//#define WITH_OPCODE_GE
//#define WITH_OPCODE_EQ
#define WITH_OPCODE_NE

//#define WITH_OPCODE_SETPREVC_AND_SETLOCALC

#define WITH_OPCODE_COM
#define WITH_OPCODE_NEG
#define WITH_OPCODE_NOT

//#define WITH_OPCODE_JMPI
//#define WITH_OPCODE_CALLI

#define WITH_OPCODE_GETPREV
#define WITH_OPCODE_SETPREV
#define WITH_OPCODE_GETLOCAL
//#define WITH_OPCODE_GETLOCALI_AND_LOCALADDRI
#define WITH_OPCODE_SETLOCAL
//#define WITH_OPCODE_SETLOCALI

//#ifdef WITH_OPCODE_SETPREVT
//#ifdef WITH_OPCODE_SETLOCALT


//#define WITH_OPCODE_STORET

#define WITH_OPCODE_POPN
//#define  WITH_OPCODE_PUSHN

//#define WITH_OPCODE_GETTYPE

#define AUTO_TYPEINFO_FOR_UNINITIALIZED_LOCALS

#if MAX_NR_BYTES_FOR_VARIABLE_INDEX==1
	#define VARIABLE_INDEX_TYPE uint8
	#define SIGNED_VARIABLE_INDEX_TYPE int8
	#define CHOOSE_VI_TYPE(v) ((v).u8)
	#define CHOOSE_SIGNED_VI_TYPE(v) ((v).i8)
#elif MAX_NR_BYTES_FOR_VARIABLE_INDEX==2
	#define VARIABLE_INDEX_TYPE uint16
	#define SIGNED_VARIABLE_INDEX_TYPE int16
	#define CHOOSE_VI_TYPE(v) ((v).u16)
	#define CHOOSE_SIGNED_VI_TYPE(v) ((v).i16)
#else
	#define VARIABLE_INDEX_TYPE uint32
	#define SIGNED_VARIABLE_INDEX_TYPE int32
	#define CHOOSE_VI_TYPE(v) ((v).u32)
	#define CHOOSE_SIGNED_VI_TYPE(v) ((v).i32)
#endif

#if MAX_NR_BYTES_FOR_PROGRAM_COUNTER==1
	#define PROGRAM_COUNTER_TYPE uint8
	#define SIGNED_PROGRAM_COUNTER_TYPE int8
	#define CHOOSE_PC_TYPE(v) ((v).u8)
#elif MAX_NR_BYTES_FOR_PROGRAM_COUNTER==2
	#define PROGRAM_COUNTER_TYPE uint16
	#define SIGNED_PROGRAM_COUNTER_TYPE int16
	#define CHOOSE_PC_TYPE(v) ((v).u16)
#else
	#define PROGRAM_COUNTER_TYPE uint32
	#define SIGNED_PROGRAM_COUNTER_TYPE int32
	#define CHOOSE_PC_TYPE(v) ((v).u32)
#endif


#if defined(WITH_INTEGER64)
	#define LARGEST_UINTEGER_TYPE uint64
	#define LARGEST_SINTEGER_TYPE int64
	#define NR_BYTES_FOR_LARGEST_AVAILABLE_INTEGER 8
	#define CHOOSE_LARGEST_SINTEGER(v) ((v).i64)
	#define CHOOSE_LARGEST_UINTEGER(v) ((v).u64)
#elif defined(WITH_INTEGER32)
	#define LARGEST_UINTEGER_TYPE uint32
	#define LARGEST_SINTEGER_TYPE int32
	#define NR_BYTES_FOR_LARGEST_AVAILABLE_INTEGER 4
	#define CHOOSE_LARGEST_SINTEGER(v) ((v).i32)
	#define CHOOSE_LARGEST_UINTEGER(v) ((v).u32)
#elif defined(WITH_INTEGER16)
	#define LARGEST_UINTEGER_TYPE uint16
	#define LARGEST_SINTEGER_TYPE int16
	#define NR_BYTES_FOR_LARGEST_AVAILABLE_INTEGER 2
	#define CHOOSE_LARGEST_SINTEGER(v) ((v).i16)
	#define CHOOSE_LARGEST_UINTEGER(v) ((v).u16)
#elif defined(WITH_INTEGER8)
	#define LARGEST_UINTEGER_TYPE uint8
	#define LARGEST_SINTEGER_TYPE int8
	#define NR_BYTES_FOR_LARGEST_AVAILABLE_INTEGER 1
	#define CHOOSE_LARGEST_SINTEGER(v) ((v).i8)
	#define CHOOSE_LARGEST_UINTEGER(v) ((v).u8)
#endif

struct Value
{
	enum
	{	// Die Reihenfolge und die Werte drfen nicht verndert werden !
		t_void=0, t_bool=1, t_i8=2, t_u8=3, t_i16=4, t_u16=5, t_i32=6, t_u32=7, t_i64=8, t_u64=9, t_real=10,
		anz_user_types=11, // Die Anzahl der untersttzten elementaren Typen

		t_dummytype=anz_user_types-1,

		t_pc_marker_address,
		anz_type_indices,

		t_unknown_type=anz_type_indices-1,
		t_nil,
		t_byte_block_length, // Dies ist nur eine Markierung (ein Byte lang), in der NICHT die Lnge abgespeichert wird !
		t_unused_byte,

		t_i16_but8used=0, t_u16_but8used=11, t_i32_but8used=12, t_u32_but8used=13, t_i64_but8used=14, t_u64_but8used=15,
		t_i32_but16used=11, t_u32_but16used=12, t_i64_but32used=13, t_u64_but32used=14
	};
	enum
	{
		t_uchar=(sizeof(unsigned char)==1 ? t_u8 : (sizeof(unsigned char)==2 ? t_u16 : (sizeof(unsigned char)==4 ? t_u32 : t_u64))),
		t_char=(sizeof(signed char)==1 ? t_i8 : (sizeof(signed char)==2 ? t_i16 : (sizeof(signed char)==4 ? t_i32 : t_i64))),
		t_ushort=(sizeof(unsigned short)==1 ? t_u8 : (sizeof(unsigned short)==2 ? t_u16 : (sizeof(unsigned short)==4 ? t_u32 : t_u64))),
		t_short=(sizeof(short)==1 ? t_i8 : (sizeof(short)==2 ? t_i16 : (sizeof(short)==4 ? t_i32 : t_i64))),
		t_uint=(sizeof(unsigned int)==1 ? t_u8 : (sizeof(unsigned int)==2 ? t_u16 : (sizeof(unsigned int)==4 ? t_u32 : t_u64))),
		t_int=(sizeof(int)==1 ? t_i8 : (sizeof(int)==2 ? t_i16 : (sizeof(int)==4 ? t_i32 : t_i64))),
		t_ulong=(sizeof(unsigned long)==1 ? t_u8 : (sizeof(unsigned long)==2 ? t_u16 : (sizeof(unsigned long)==4 ? t_u32 : t_u64))),
		t_long=(sizeof(long)==1 ? t_i8 : (sizeof(long)==2 ? t_i16 : (sizeof(long)==4 ? t_i32 : t_i64))),
		t_ulonglong=(sizeof(unsigned long long)==1 ? t_u8 : (sizeof(unsigned long long)==2 ? t_u16 : (sizeof(unsigned long long)==4 ? t_u32 : t_u64))),
		t_longlong=(sizeof(long long)==1 ? t_i8 : (sizeof(long long)==2 ? t_i16 : (sizeof(long long)==4 ? t_i32 : t_i64))),

		VarIndexTypeNr=(1==MAX_NR_BYTES_FOR_VARIABLE_INDEX) ? t_u8 :
						((2==MAX_NR_BYTES_FOR_VARIABLE_INDEX) ? t_u16 : t_u32),
		SignedVarIndexTypeNr=(1==MAX_NR_BYTES_FOR_VARIABLE_INDEX) ? t_i8 :
						((2==MAX_NR_BYTES_FOR_VARIABLE_INDEX) ? t_i16 : t_i32),
		ProgramCounterTypeNr=(1==MAX_NR_BYTES_FOR_PROGRAM_COUNTER) ? t_u8 :
						((2==MAX_NR_BYTES_FOR_PROGRAM_COUNTER) ? t_u16 : t_u32),
		LargestAvailableUIntegerTypeNr=(1==NR_BYTES_FOR_LARGEST_AVAILABLE_INTEGER) ? t_u8 :
						((2==NR_BYTES_FOR_LARGEST_AVAILABLE_INTEGER) ? t_u16 :
						((4==NR_BYTES_FOR_LARGEST_AVAILABLE_INTEGER) ? t_u32 : t_u64)),
		LargestAvailableSIntegerTypeNr=(1==NR_BYTES_FOR_LARGEST_AVAILABLE_INTEGER) ? t_i8 :
						((2==NR_BYTES_FOR_LARGEST_AVAILABLE_INTEGER) ? t_i16 :
						((4==NR_BYTES_FOR_LARGEST_AVAILABLE_INTEGER) ? t_i32 : t_i64)),
	};
	uint8 type; // Muss einen Wert aus dem enum haben (t_void, t_bool, ... , t_real, t_pc_marker_address).
	union
	{	// Wenn type==t_void ist, sind die Elemente dieser Union bedeutungslos.
		// Wenn type==t_bool ist, ist nur das Element u8 von Bedeutung, und es darf ausschlielich die Werte 0 oder 1 annehmen.
		uint8 u8; int8 i8;
		uint16 u16; int16 i16;
		uint32 u32; int32 i32;
		uint64 u64; int64 i64;
		real_type r;
	};
	static uint8 sbytesize(uint8 typ)
	{	// t_void bentigt ein Byte !
		static uint8 sz[anz_type_indices]={ 1, 1, 1, 1, 2, 2, 4, 4, 8, 8, sizeof(real_type), MAX_NR_BYTES_FOR_PROGRAM_COUNTER };
		return typ>=anz_type_indices ? 1 : sz[typ];
	}
	uint8 bytesize() { return sbytesize(type); }
	static bool s_is_integer_type(uint8 typ) { return typ>=t_i8 && typ<t_real; }
	bool is_signed_integer() { return type>=t_i8 && type<t_real && 0==(type&1); }
	bool is_unsigned_integer() { return type>=t_i8 && type<t_real && 0!=(type&1); }
	static bool s_is_signed_integer(uint8 typ) { return typ>=t_i8 && typ<t_real && 0==(typ&1); }
	static bool s_is_unsigned_integer(uint8 typ) { return typ>=t_i8 && typ<t_real && 0!=(typ&1); }
};

#ifdef PRINT_COMMANDS_AT_RUNTIME
	#define PRINT_COMMAND_IF_ENABLED(command_name, pc, sp) \
		printf("PC=%.4x  SP=%.4d  Command: %-10s", pc, sp, command_name);
	#define PRINT_NEWLINE_IF_PRINT_COMMANDS_AT_RUNTIME printf("\r\n");
	void bi_print_value(const Value& v)
	{
		switch(v.type)
		{
			case Value::t_void:
				printf("void");				
				break;
			case Value::t_bool:
				printf("bool: %s", v.u8 ? "true" : "false");
				break;
			case Value::t_u8:
				printf("u8: %u  0x%x", v.u8, v.u8);
				break;
			case Value::t_i8:
				printf("i8: %d  0x%x", v.i8, v.i8);
				break;
			case Value::t_u16:
				printf("u16: %u  0x%x", v.u16, v.u16);
				break;
			case Value::t_i16:
				printf("i16: %d  0x%x", v.i16, v.i16);
				break;
			case Value::t_u32:
				printf("u32: %u  0x%x", v.u32, v.u32);
				break;
			case Value::t_i32:
				printf("i32: %d  0x%x", v.i32, v.i32);
				break;
			case Value::t_u64:
				printf("u64: 0x%x %x", (uint32)(v.u64>>32), (uint32)(v.u64&0xffffffff));
				break;
			case Value::t_i64:
				printf("i64: 0x%x %x", (uint32)(v.u64>>32), (uint32)(v.u64&0xffffffff));
				break;
			case Value::t_real:
				printf("r  : %.16g", v.r);
				break;
		}
		printf("      ");
	}
#endif

#define DENSE_PACKED_STRUCT_INIT enum { sum0=0 };
#define DENSE_PACKED_STRUCT_NEW_ELEMENT(nr_minus_1, nr, type, name) ;enum { size##nr=sizeof(type), sum##nr=sum##nr_minus_1+size##nr };\
			;inline type& name() { return *(type*)(((uint8*)this)+sum##nr_minus_1); };
#define DENSE_PACKED_STRUCT_NEW_ARRAY(nr_minus_1, nr, type, name, arrsize) ;enum { size##nr=sizeof(type)*(arrsize), sum##nr=sum##nr_minus_1+size##nr };\
			;inline type& name(uint32 index) { return ((type*)(((uint8*)this)+sum##nr_minus_1))[index]; };
#define DENSE_PACKED_STRUCT_END(maximal_occured_nr) ;uint8 internal_memory[sum##maximal_occured_nr];

class BytecodeInterpreter
{
	private: BytecodeInterpreter(const BytecodeInterpreter& bi) { }
			 // Copy-Konstruktor nicht vogesehen, Zugriff verhindern
	private: BytecodeInterpreter& operator=(const BytecodeInterpreter& bi) { return *this; }
			 // Zuweisungsoperator nicht vogesehen, Zugriff verhindern

#ifdef CALCULATE_MAXIMUM_STACK_USAGE
	VARIABLE_INDEX_TYPE minimum_stack_pointer;
#endif

public:

	void read_byte_block(PROGRAM_COUNTER_TYPE codepos, uint8 bytesize, uint8 *mem)
	{	// Es muss size>=1 sein !. Achtung: Es knnen maximal 255 Bytes gelesen werden (bytesize ist uint8).
		do {
			*(mem++)=get_code_byte(codepos);
			codepos+=2;
		} while(--bytesize);
	}
#ifdef CALCULATE_MAXIMUM_STACK_USAGE
	uint32 get_maximum_stack_usage()
	{
		return ((VARIABLE_INDEX_TYPE)(-minimum_stack_pointer)) + ((3+(VARIABLE_INDEX_TYPE)(-minimum_stack_pointer))>>2) + 2;
	}
#endif

	struct BytecodeProgramHeader
	{
		DENSE_PACKED_STRUCT_INIT;
		DENSE_PACKED_STRUCT_NEW_ELEMENT(0, 1, PROGRAM_COUNTER_TYPE, global_begin_addrs_of_elementary_types_start);
		DENSE_PACKED_STRUCT_NEW_ARRAY(1, 2, VARIABLE_INDEX_TYPE, addr_of_first_global_non_elementary_variable, 3);
		DENSE_PACKED_STRUCT_NEW_ARRAY(2, 3, PROGRAM_COUNTER_TYPE, global_var_type_info_start, 3);
		DENSE_PACKED_STRUCT_NEW_ARRAY(3, 4, PROGRAM_COUNTER_TYPE, global_var_initvalues_start, 2);
		DENSE_PACKED_STRUCT_NEW_ELEMENT(4, 5, uint8, limit_for_4bit_opcodes);
		DENSE_PACKED_STRUCT_NEW_ELEMENT(5, 6, uint8, limit_for_8bit_opcodes);
		DENSE_PACKED_STRUCT_NEW_ARRAY(6, 7, uint16, limit_for_nbit_opcodes, 5);
		DENSE_PACKED_STRUCT_NEW_ARRAY(7, 8, PROGRAM_COUNTER_TYPE, opcode_transformation_tables, 8);
		DENSE_PACKED_STRUCT_NEW_ELEMENT(8, 9, PROGRAM_COUNTER_TYPE, function_address_table);
#ifdef USE_TABLE_FOR_CONSTANTS
		DENSE_PACKED_STRUCT_NEW_ELEMENT(9, 10, uint8, limit_for_1bytes_constants);
		DENSE_PACKED_STRUCT_NEW_ELEMENT(10, 11, uint8, limit_for_2bytes_constants);
		DENSE_PACKED_STRUCT_NEW_ELEMENT(11, 12, uint8, limit_for_4bytes_constants);
		DENSE_PACKED_STRUCT_NEW_ELEMENT(12, 13, PROGRAM_COUNTER_TYPE, constants1bytes_table);
		DENSE_PACKED_STRUCT_NEW_ELEMENT(13, 14, PROGRAM_COUNTER_TYPE, constants2bytes_table);
		DENSE_PACKED_STRUCT_NEW_ELEMENT(14, 15, PROGRAM_COUNTER_TYPE, constants4bytes_table);
		DENSE_PACKED_STRUCT_NEW_ELEMENT(15, 16, PROGRAM_COUNTER_TYPE, constants8bytes_table);
		DENSE_PACKED_STRUCT_END(16);
#else
		DENSE_PACKED_STRUCT_END(9);
#endif
	} ProgramHeader;

	VARIABLE_INDEX_TYPE global_begin_addrs_of_elementary_types[3*Value::anz_user_types+1];
		// global_begin_addrs_of_elementary_types[3*Value::anz_user_types] enthlt die erste unbenutzte Variablen-Adresse.

	uint8 global_elementary_var_addr_to_type_nr(VARIABLE_INDEX_TYPE var_addr)
	{
		VARIABLE_INDEX_TYPE t=(3*Value::anz_user_types) >> 1;
		VARIABLE_INDEX_TYPE dt=((t==1) ? 1 : (t>>1));
		ASSERT(0!=dt, "var_addr_to_type_nr Error", pc);
		while(1)
		{
			if(var_addr>=global_begin_addrs_of_elementary_types[t])
			{
				if(t==3*Value::anz_user_types-1 || var_addr<global_begin_addrs_of_elementary_types[t+1])
					break;
				if(t+dt<3*Value::anz_user_types)
					t+=dt;
			}
			else if(t>=dt)
				t-=dt;
			dt=(dt+1)>>1;
		}
		return t % Value::anz_user_types;
	}

	uint8 get_stack_byte_associated_type_info(VARIABLE_INDEX_TYPE byte_nr)
	{
		byte_nr=-byte_nr;
		return (stack[byte_nr>>2]>>((byte_nr&3)<<1))&3;
	}
#if defined(USE_STACK_TYPE_INFO_VARIANT) && defined(BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE)
	uint16 get_stack_word_associated_type_info(VARIABLE_INDEX_TYPE byte_nr)
	{
		byte_nr=-byte_nr;
		return (*(uint16*)(&stack[(byte_nr>>2)-1]))<<((3-(byte_nr&3))<<1);
	}
#endif
	void set_stack_byte_associated_type_info(VARIABLE_INDEX_TYPE byte_nr, uint8 type_info)
	{
		byte_nr=-byte_nr;
		stack[byte_nr>>2] = (stack[byte_nr>>2] & (~(3<<((byte_nr&3)<<1)))) | (type_info<<((byte_nr&3)<<1));
	}
#if defined(USE_STACK_TYPE_INFO_VARIANT) && defined(BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE)
	void set_stack_word_associated_type_info(VARIABLE_INDEX_TYPE byte_nr, uint8 type_info, uint8 anz_bytes)
	{
		byte_nr=-byte_nr;
		uint8 nr1=((byte_nr&3)<<1)+26;
		uint32 mask=~((32==nr1 ? 0 : (((uint32)1LU)<<nr1)) - (((uint32)1LU)<<(nr1-(anz_bytes<<1))));
		uint32 *mem=(uint32*)(&stack[(byte_nr>>2)-3]);
		(*mem) &= mask;
		(*mem) |= ((uint32)type_info)<<(nr1-2);
	}
#endif

	uint8 *var_addr_to_mem_addr(VARIABLE_INDEX_TYPE var_addr)
	{
		return (0!=sp && var_addr>=sp) ? ((uint8*)(stack_end-(VARIABLE_INDEX_TYPE)(-var_addr))) : (((uint8*)memory)+var_addr);
	}

	int8 get_var(VARIABLE_INDEX_TYPE var_addr, Value *value, bool read_type_given_in_value=false)
	{
		Value tempval;
		uint8 i=0, type=0; // type: Vorinitialisierung fr den Fall read_type_given_in_value==true
		if(read_type_given_in_value)
			i=value->bytesize(); // value->bytesize() darf nur im Fall read_type_given_in_value==true aufgerufen werden, weil value
				// sonst nicht initialisiert ist.
		uint8 type2=255, *mem=var_addr_to_mem_addr(var_addr);
		if(0!=sp && var_addr>=sp)
		{
			if(!read_type_given_in_value) // // Wenn read_type_given_in_value gesetzt ist, drfen i und type nicht verndert werden,
			{	// und es darf es auch keine Fehlermeldung geben, wenn get_stack_byte_associated_type_info() falsche Typeninformationen liefert.
				ASSERT(0!=get_stack_byte_associated_type_info(var_addr), "get_var: invalid var_addr", pc);
#if defined(USE_STACK_TYPE_INFO_VARIANT) && defined(BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE)
				uint16 b;
				type=((b=get_stack_word_associated_type_info(var_addr))>>14)&3;
				if(0!=(b&((3<<12)|(3<<10))))
					i=(0!=(b&(3<<12))) ? 1 : 2;
				else
					i=(0!=(b&(3<<6))) ? 4 : 8;
#else
				uint8 b;
				for(i=0; 0==(b=get_stack_byte_associated_type_info(var_addr+i)) || (0==i && (type=b, true)); i++) // Die Abfrage 0==i muss nach der
					{ ASSERT(i<=8, "get_var: invalid type information in the stack", pc); }
#endif
			}
		}		
		else
		{
			uint8 section=(var_addr<global_begin_addrs_of_elementary_types[Value::anz_user_types]) ? 0 :
							((var_addr>=global_begin_addrs_of_elementary_types[2*Value::anz_user_types]) ? 2 : 1);
			if(var_addr<ProgramHeader.addr_of_first_global_non_elementary_variable(section))
			{
				if(!read_type_given_in_value) // // Wenn read_type_given_in_value gesetzt ist, drfen i und type nicht verndert werden.
				{
					type2=global_elementary_var_addr_to_type_nr(var_addr);
					i=Value::sbytesize(type2);
					type=0;
				}
			}
			else
			{
				if(!read_type_given_in_value) // // Wenn read_type_given_in_value gesetzt ist, drfen i und type nicht verndert werden,
				{	// und es darf es auch keine Fehlermeldung geben, wenn get_stack_byte_associated_type_info() falsche Typeninformationen liefert.
					var_addr-=ProgramHeader.addr_of_first_global_non_elementary_variable(section);
					ASSERT(0!=((get_code_nibble(ProgramHeader.global_var_type_info_start(section)+(var_addr>>1))>>((var_addr&1)<<1))&3),
							"get_var: invalid var_addr", pc);
					uint8 b;
					for(i=0; 0==(b=(get_code_nibble(ProgramHeader.global_var_type_info_start(section)+((var_addr+i)>>1))>>(((var_addr+i)&1)<<1))&3)
							|| (0==i && (type=b, true)); i++) // Die Abfrage 0==i muss nach der Zuweisung an b erfolgen !
						;
					ASSERT(i<=8, "get_var: invalid type information in the stack", pc);
					var_addr+=ProgramHeader.addr_of_first_global_non_elementary_variable(section);
				}
			}

			if(2==section)
			{	// Im folgenden ASSERT wird (zu Recht) davon ausgegangen, dass nicht der Typ Value::t_pc_marker_address vorliegt.
				ASSERT(var_addr+i<= global_begin_addrs_of_elementary_types[3*Value::anz_user_types], "Invalid variable address", pc);
				read_byte_block(2*(ProgramHeader.global_var_initvalues_start(1)+
							(var_addr-global_begin_addrs_of_elementary_types[2*Value::anz_user_types])), i, (uint8*)(&tempval.u64));
				mem=(uint8*)(&tempval.u64);
			}
		}
		if(0==value)
			value=&tempval;
		switch(i)
		{
			case 1:
				value->u8 = *((uint8*)mem);
				if(1==t_bool && 2==t_i8 && 3==t_u8)
				{
					if(1==type)
					{
						if(2==value->u8) type=t_void;
						if(3==value->u8) type=t_byte_block_length;
						if(4==value->u8) type=t_unused_byte;
					}
				}
				else if(2==type)
					type=t_i8;
				else if(3==type)
					type=t_u8;
				else // if(1==type)
				{
					if(2==value->u8) type=t_void;
					if(3==value->u8) type=t_byte_block_length;
				}
				break;
#ifdef WITH_INTEGER16
			case 2:
				if(1==type)
				{
					type=t_pc_marker_address;
#ifdef BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE
					CHOOSE_PC_TYPE(*value)=*((PROGRAM_COUNTER_TYPE*)mem);
#elif MAX_NR_BYTES_FOR_PROGRAM_COUNTER==1
					value->u8 =*mem;
#elif MAX_NR_BYTES_FOR_PROGRAM_COUNTER==2
					value->u16 =*(mem++);
					value->u16 |= (*(mem++))<<8;
#elif MAX_NR_BYTES_FOR_PROGRAM_COUNTER==4
					for(i=0, value->u32=0; i<4; i++)
						value->u32 |= (*(mem++))<<(8*i);
#elif MAX_NR_BYTES_FOR_PROGRAM_COUNTER==8
					for(i=0, value->u64=0; i<8; i++)
						value->u64 |= (*(mem++))<<(8*i);
#endif
				}
				else
				{
					if(t_i16==4 && t_u16==5)
						type+=2;
					else if(2==type)
						type=t_i16;
					else
						type=t_u16;
#ifdef BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE
					value->u16=*((uint16*)mem);
#else
					value->u16 =*(mem++);
					value->u16 |= (*(mem++))<<8;
#endif
				}
				break;
#endif
#ifdef WITH_INTEGER32
			case 4:
				if(t_i32==6 && t_u32==7)
					type+=4;
				else if(2==type)
					type=t_i32;
				else
					type=t_u32;
#ifdef BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE
				value->u32=*((uint32*)mem);
#else
				for(i=0, value->u32=0; i<4; i++)
					value->u32 |= (*(mem++))<<(8*i);
#endif
				break;
#endif
			case 8:
				if(1==type)
				{
#ifdef WITH_REAL
					type=t_real;
#ifdef BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE
					value->r = *((real_type*)mem);
#else
					if(sizeof(real_type)==4)
						for(i=0, *((uint32*)&value->r)=0; i<4; i++)
							*((uint32*)&value->r) |= (*(mem++))<<(8*i);
					else // hier muss sizeof(real_type)==8 sein !
						for(i=0, *((uint64*)&value->r)=0; i<8; i++)
							*((uint64*)&value->r) |= (*(mem++))<<(8*i);
#endif
#endif
				}
#ifdef WITH_INTEGER64
				else
				{
					if(t_i64==8 && t_u64==9)
						type+=6;
					else if(2==type)
						type=t_i64;
					else
						type=t_u64;
#ifdef BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE
					value->u64=*((uint64*)mem);
#else
					for(i=0, value->u64=0; i<8; i++)
						value->u64 |= (*(mem++))<<(8*i);
#endif
				}
				break;
#endif
		}
		if(255!=type2)
			type=type2;
		if(!read_type_given_in_value)
			value->type=type;
		return value->type;
	}
	void set_var(VARIABLE_INDEX_TYPE var_addr, Value *value, bool convert_value_to_existing_type=false, bool dont_change_typeinfo=false)
	{	// ACHTUNG: Wenn convert_value_to_existing_type==true bergeben wird, wird value auf den Typ gecastet, der
		// an der Speicheradresse var_addr momentan gespeichert ist.
		if(convert_value_to_existing_type)
			cast(value, get_var(var_addr, 0));
		if(0==var_addr)
			return; // Zuweisung an die Null-Adresse ignorieren (der Nullzeiger ist reserviert).
		uint8 t=value->type, j=Value::sbytesize(t);
		uint8 *mem=var_addr_to_mem_addr(var_addr);
		ASSERT((0!=sp && var_addr>=sp) || var_addr+j <= global_begin_addrs_of_elementary_types[2*Value::anz_user_types], "Invalid variable address", pc);
		if(t_pc_marker_address==t)
			{ t=ProgramCounterTypeNr; j=MAX_NR_BYTES_FOR_PROGRAM_COUNTER; }
		switch(t)
		{
			case t_void:
				*((uint8*)mem)=2;
				goto set_1_byte_typeinfo;
			case t_byte_block_length:
				*((uint8*)mem)=3;
				goto set_1_byte_typeinfo;
			case t_unused_byte:
				*((uint8*)mem)=4;
				goto set_1_byte_typeinfo;
			case t_bool: // Wenn t==t_bool ist, dann darf value->u8 nur 0 oder 1 sein !
			case t_u8:
			case t_i8:
				*((uint8*)mem)=value->u8;
set_1_byte_typeinfo:
				if(t_i8==2 && t_u8==3)
				{
					if(t_u8!=t && t_i8!=t)
						t=1; // t==t_bool oder t==t_void oder t==t_byte_block_length
				}
				else if(t_i8==t)
					t=2;
				else if(t_u8==t)
					t=3;
				else
					t=1; // t==t_bool oder t==t_void oder t==t_byte_block_length
				break;
#ifdef WITH_INTEGER16
			case t_u16:
			case t_i16:
#ifdef BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE
				*((uint16*)mem)=value->u16;
#else
				*(mem++)=value->u16&255; value->u16>>=8;
				*mem=value->u16;
#endif
				if(t_i16==4 && t_u16==5)
					t-=2;
				else if(t_i16==t)
					t=2;
				else
					t=3;
				break;
#endif
#ifdef WITH_INTEGER32
			case t_u32:
			case t_i32:
#ifdef BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE
				*((uint32*)mem)=value->u32;
#else
				for(uint8 i=4+1; --i; )
					{ *(mem++)=value->u32&255; value->u32>>=8; }
#endif
				if(t_i32==6 && t_u32==7)
					t-=4;
				else if(t_i32==t)
					t=2;
				else
					t=3;
				break;
#endif
#ifdef WITH_INTEGER64
			case t_u64:
			case t_i64:
#ifdef BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE
				*((uint64*)mem)=value->u64;
#else
				for(uint8 i=8+1; --i; )
					{ *(mem++)=value->u64&255; value->u64>>=8; }
#endif
				if(t_i64==8 && t_u64==9)
					t-=6;
				else if(t_i64==t)
					t=2;
				else
					t=3;
				break;
#endif
#ifdef WITH_REAL
			case t_real:
#ifdef BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE
				*((real_type*)mem)=value->r;
#else
				if(sizeof(real_type)==4)
					for(uint8 i=sizeof(real_type)+1; --i; )
						{ *(mem++)=((uint32)value->r)&255; *((uint32*)&value->r)>>=8; }
				else // es muss sizeof(real_type)==8 sein !
					for(uint8 i=sizeof(real_type)+1; --i; )
						{ *(mem++)=((uint64)value->r)&255; *((uint64*)&value->r)>>=8; }
#endif
				t=1;
				break;
#endif
		}
		if(t_pc_marker_address==value->type)
			t=1;
#if defined(USE_STACK_TYPE_INFO_VARIANT) && defined(BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE)
		if(0!=sp && var_addr>=sp && !dont_change_typeinfo)
		{
			if(((PROGRAM_COUNTER_TYPE)(var_addr+16))>=var_addr)
				set_stack_word_associated_type_info(var_addr, t, j);
			else
				for(uint8 k=0; k<j; t=0, k++)
					set_stack_byte_associated_type_info(var_addr+k, t);
		}
#else
		if(0!=sp && var_addr>=sp && !dont_change_typeinfo)
			for(uint8 k=0; k<j; t=0, k++)
				set_stack_byte_associated_type_info(var_addr+k, t);
#endif
	}

#ifdef DEBUG_MODE
	VARIABLE_INDEX_TYPE find_previous_stack_element_verify(VARIABLE_INDEX_TYPE diff)
	{
		VARIABLE_INDEX_TYPE start_addr=sp;
		Value v;
		while(diff)
		{
			if(1==get_stack_byte_associated_type_info(start_addr) && (get_var(start_addr, &v), v.type==t_byte_block_length))
			{
				get_var(++start_addr, &v);
				start_addr+=MAX_NR_BYTES_FOR_VARIABLE_INDEX;
				start_addr+=CHOOSE_VI_TYPE(v);
				--diff;
				continue;
			}
			while(!get_stack_byte_associated_type_info(++start_addr))
				;
			--diff;
		}
		return start_addr;
	}
#endif

	VARIABLE_INDEX_TYPE find_previous_stack_element(VARIABLE_INDEX_TYPE diff)
	{
#ifdef DEBUG_MODE
		VARIABLE_INDEX_TYPE orig_diff=diff;
#endif
		uint8 d=((last_stack_addrs_akt-last_stack_addrs_begin)&31);
		if(diff<d)
		{
#ifdef DEBUG_MODE
			if(last_stack_addrs[(last_stack_addrs_akt-1-diff)&31] != find_previous_stack_element_verify(diff))
			{
				ASSERT(last_stack_addrs[(last_stack_addrs_akt-1-diff)&31] == find_previous_stack_element_verify(diff),
					"Internal Bytecode Interpreter Error: find_previous_stack_element() returns invalid address.", pc);
			}
#endif
			return last_stack_addrs[(last_stack_addrs_akt-1-diff)&31];
		}
		// ddprintf("##### Suche: sp=%u   d=%u    akt=%d    begin=%d\r\n", sp, d, last_stack_addrs_akt, last_stack_addrs_begin);
#ifdef USE_ASSERT
		Value v;
#endif
		VARIABLE_INDEX_TYPE start_addr;
		if(0==d)
		{
			last_stack_addrs_akt=last_stack_addrs_begin+1;
			start_addr=sp;
			++d;
			// Gleichwertig wre    last_stack_addrs[(last_stack_addrs_akt++)&31]=sp;   ++d;
			// und danach die Ausfhrung Codes im else-Teil (Fall 0!=d).
		}
		else
		{
			start_addr=last_stack_addrs[last_stack_addrs_begin&31];
			diff-=d-1;
		}
		for( ; ((d<31 ? (++d, last_stack_addrs[last_stack_addrs_begin&31]=start_addr) : 0), true) && 0!=diff ; --last_stack_addrs_begin)
		{
#ifdef USE_ASSERT
			ASSERT(!(1==get_stack_byte_associated_type_info(start_addr) && (get_var(start_addr, &v), v.type==t_byte_block_length)),
				"Internal Bytecode Interpreter Error: find_previous_stack_element() traverses the section in the stack used for uninitialized variables.", pc);
			while(0==get_stack_byte_associated_type_info(++start_addr))
				;
#else
			//get_stack_byte_associated_type_info(start_addr);
			//while(0==get_stack_byte_associated_type_info(++start_addr))
			//	;
#if defined(USE_STACK_TYPE_INFO_VARIANT) && defined(BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE)
			uint16 b=get_stack_word_associated_type_info(++start_addr);
			if(0==(b&0xfffc))
				start_addr+=7;
			else
				while(0==(b&(3<<14)))
					{ b<<=2; ++start_addr; }
#else
			get_stack_byte_associated_type_info(start_addr);
			while(0==get_stack_byte_associated_type_info(++start_addr))
				;
#endif
#endif
			--diff;
			/*ddprintf("Suche: last_stack_addrs[%u]=%u,  last_stack_addrs_begin(neu)=%u\r\n",
				(last_stack_addrs_begin)&31, last_stack_addrs[last_stack_addrs_begin&31], last_stack_addrs_begin-1);*/
		}
#ifdef DEBUG_MODE
			ASSERT(start_addr == find_previous_stack_element_verify(orig_diff),
				"Internal Bytecode Interpreter Error: find_previous_stack_element() returns invalid address.", pc);
#endif
		return start_addr;
	}

	enum
	{	// Die Reihenfolge und die Werte drfen nicht verndert werden !
		t_void=Value::t_void, t_bool=Value::t_bool, t_i8=Value::t_i8, t_u8=Value::t_u8, t_i16=Value::t_i16, t_u16=Value::t_u16,
		t_i32=Value::t_i32, t_u32=Value::t_u32, t_i64=Value::t_i64, t_u64=Value::t_u64, t_real=Value::t_real,
		anz_user_types=Value::anz_user_types,

		t_pc_marker_address=Value::t_pc_marker_address,

		t_byte_block_length=Value::t_byte_block_length, // Dies ist nur eine Markierung (ein Byte lang), in der NICHT die Lnge abgespeichert wird !
		t_unknown_type=Value::t_unknown_type,
		t_unused_byte=Value::t_unused_byte,
		
		anz_type_indices=Value::anz_type_indices,

		VarIndexTypeNr=Value::VarIndexTypeNr,
		ProgramCounterTypeNr=Value::ProgramCounterTypeNr,
		LargestAvailableUIntegerTypeNr=Value::LargestAvailableUIntegerTypeNr,
		LargestAvailableSIntegerTypeNr=Value::LargestAvailableSIntegerTypeNr,

		t_ushort=Value::t_ushort,
		t_short=Value::t_short,
		t_uint=Value::t_uint,
		t_int=Value::t_int,
		t_ulong=Value::t_ulong,
		t_long=Value::t_long,
		t_ulonglong=Value::t_ulonglong,
		t_longlong=Value::t_longlong,

		t_i16_but8used=Value::t_i16_but8used, t_u16_but8used=Value::t_u16_but8used, t_i32_but8used=Value::t_i32_but8used,
		t_u32_but8used=Value::t_u32_but8used, t_i64_but8used=Value::t_i64_but8used, t_u64_but8used=Value::t_u64_but8used,
		t_i32_but16used=Value::t_i32_but16used, t_u32_but16used=Value::t_u32_but16used,
		t_i64_but32used=Value::t_i64_but32used, t_u64_but32used=Value::t_u64_but32used

	};

	enum Opcode
	{
		CALL=0,         // call function at pc address following the CALL opcode
#ifdef WITH_OPCODE_CALLI
		CALLI,        // call function at pc address stored on the top of the stack. Eigentlich ist CALLI ein "unary operand"-Befehl.
					  // Um ihn jedoch zusammen mit CALL mit gemeinsamen Code abarbeiten zu knnen, wird er nicht nach
					  // BEGIN_UNARY_OPERATORS aufgefhrt.
#endif

		RET,

		BR,

		PUSHC,        // Push constant onto stack
#ifdef USE_TABLE_FOR_CONSTANTS
		PUSHCT,
#endif
		PUSHC0U8,
		PUSHC0,
		PUSHCS,
		PUSHCSU8,

		BRK,		  // BREAK. For use after executing IFELSE command ONLY IF THE IFELSE-CONDITION (OPERAND OF IFELSE COMMAND)
					  // HAS BEEN TRUE, I.E. IF THE IF-PART OF THE IFELSE-COMMAND IS EXECUTED.
					  // When a BRK opcode occures (within the IF-Part), the execution is continued at the
					  // beginning of the ELSE-Part of the IFELSE Command, i.d. at the position PC0+d1 (see description
					  // of the IFELSE Opcode).
		CONT,		  // CONTINUE. For use after executing IFELSE command ONLY IF THE IFELSE-CONDITION (OPERAND OF IFELSE COMMAND)
					  // HAS BEEN TRUE, I.E. IF THE IF-PART OF THE IFELSE-COMMAND IS EXECUTED.
					  // When a CONT opcode occures (within the IF-Part), the execution is continued after the
					  // end of the ELSE-Part of the IFELSE Command, i.d. at the position PC0+d1+d2 (see description
					  // of the IFELSE Opcode).
		SETTYPES,

#ifdef WITH_OPCODE_GETPREV
		GETPREV,
		GETPREVLOAD,
#endif
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
		SETPREVC,
		SETPREVCPOP,
#endif
#ifdef WITH_OPCODE_GETLOCAL
		GETLOCAL,
		GETLOCALLOAD,
#endif
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
		SETLOCALC,
		SETLOCALCPOP,
#endif
		STACKADDR,
		LOCALADDR,

		ADDCPREV,
		PREVADDC,
		ADDCLOCAL,
		LOCALADDC,
		ADDCPREVPOP,
		ADDCLOCALPOP,

		// Unary operators: pop value from stack, compute, push result
		BEGIN_UNARY_OPERATORS,
		DUMMY_OPCODE1=BEGIN_UNARY_OPERATORS-1,

#ifdef WITH_OPCODE_NEG
		NEG,          // Negate top element on stack (arithmetically)
#endif
#ifdef WITH_OPCODE_COM
		COM,          // bitwise complement the top element on stack
#endif
		CAST,		 // Pop the top stack element, cast it to the type given by the nibble following the CAST opcode
					 // and push the casted value back onto the stack.
#ifdef WITH_OPCODE_GETTYPE
		GETTYPE,	 // Pop the top stack element and push its type integer onto the stack (see the enum in struct Value).
#endif
		POP,          // discard top element of stack

		LOAD,
		LOADT,
#ifdef WITH_OPCODE_SETPREVT
	    SETPREVT,
#endif
#ifdef WITH_OPCODE_SETPREV
	    SETPREV,
	    SETPREVPOP,
		GETPREVSTORE,
		GETPREVSTOREPOP,
#endif
#ifdef WITH_OPCODE_SETLOCALT
	    SETLOCALT,
#endif
#ifdef WITH_OPCODE_SETLOCAL
		SETLOCAL,
		SETLOCALPOP,
		GETLOCALSTORE,
		GETLOCALSTOREPOP,
#endif

#ifdef WITH_OPCODE_GETLOCALI_AND_LOCALADDRI
		GETLOCALI,
		LOCALADDRI,
		STACKADDRI,
#endif

#ifdef WITH_OPCODE_PUSHN
		PUSHN,
#endif
#ifdef WITH_OPCODE_POPN
		POPN,
#endif
#ifdef WITH_OPCODE_JMPI
		JMPI,		 // Jump to the address stored on the top of the stack
#endif
		CALLEXT,

		BEGIN_UNARY_OPERATORS_WITH_BOOLEAN_OPERAND,
		DUMMY_OPCODE2=BEGIN_UNARY_OPERATORS_WITH_BOOLEAN_OPERAND-1,

#ifdef WITH_OPCODE_NOT
		NOT,          // Logical NOT of top element on stack
#endif

		// ***** Nur wenn WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION definiert ist, gilt *****:
		// Bei den folgenden Kommandos BRF und BRT wird der zu testende Operand immer vom stack geholt (POP), wenn der
		// Sprung nicht ausgefhrt wird. Wenn der Sprung ausgefhrt wird und wenn das Kommando, welches direkt am Sprungziel
		// steht (ohne Bercksichtigung einer PC-Marke, die ggf. durch IFELSE oder IFNOTELSE gesetzt wurde), einen
		// Opcode >=BEGIN_UNARY_OPERATORS hat, wird der zu testende Operand nicht vom stack geholt (kein POP), aber er wird 
		// nach bool konvertiert und das oberste Stackelement mit dem nach bool konertierten Wert berschrieben.
		// In den meisten Fllen ist es die gewnschte Eigenschaft, dass der Operand von BRF und BRT bei einem ausgefhrten
		// Sprung und einem am Sprungziel befindlichen Opcode >=BEGIN_UNARY_OPERATORS nach Konversion nach bool auf dem stack verbleibt.
		// Wenn dieses Verhalten in Ausnahmefllen nicht gewnscht ist, d.h. wenn man mchte, dass BRF und BRT ihren
		// Operanden in jedem Fall vom Stack POPen, oder wenn man mchte, dass der Operand im Fall eines ausgefhrten
		// Sprungs nicht vom Stack geholt wird (POP), obwohl der am Sprungziel stehende Opcode <BEGIN_UNARY_OPERATORS ist,
		// kann man am Sprungziel vor die gewnschten Anweisungen noch einen Befehl einfgen, welcher fr sich alleine
		// wirkungslos ist, aber einen Opcode <BEGIN_UNARY_OPERATORS hat (==> BRF und BRT holen den Operand immer vom Stack (POP))
		// oder einen Opcode >=BEGIN_UNARY_OPERATORS hat (==> BRF und BRT holen den Operand genau dann vom Stack (POP), wenn
		// der Sprung ausgefhrt wurde. Eine fr sich alleine wirkungslose Anweisung mit einem Opcode <BEGIN_UNARY_OPERATORS
		// ist der unbedingte relative Sprung (Opcode BR) um die Distanz Null, so dass der Sprung zum folgenden Befehl erfolgt.
		// Eine fr sich alleine wirkungslose Anweisung mit einem Opcode >=BEGIN_UNARY_OPERATORS
		// ist der CAST des obersten Stackelements auf den Typ, den der Operand zuvor ohnehin schon hatte. Weil
		// BRF und BRT nach bool casten, muss es sich hier ebenfalls um einen CAST nach bool handeln.

		BRF,          // Pop top stack element, convert it to bool and branch if false.
					  // ***** Nur wenn WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION definiert ist, gilt *****:
					  // If the branch is taken and the opcode found at the new pc position is at least BEGIN_UNARY_OPERATORS,
					  // i.e. if die corresponding command will POP at least one operand from the stack,
					  // re-push the boolean element POPed by the BRF command before executing the next command.
					  // The bytecode interpreter always examines the new command addressed relatively to the actual PC by the
					  // branch distance, without condidering a PC marker which could be set by a preceding
					  // IFELSE or IFNOTELSE command. If the new PC addesss (for the case the BRF branch is taken) is
					  // equal to the PC marker set by a preceding IFELSE or IFNOTELSE command, the next executed command
					  // in not the command at this PC address. But the opcode beeing testes for >=BEGIN_UNARY_OPERATORS
					  // is nevertheless the opcode found at the new PC address.

		BRT,          // Pop top stack element, convert it to bool and branch if true.
					  // ***** Nur wenn WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION definiert ist, gilt *****:
					  // If the branch is taken and the opcode found at the new pc position is at least BEGIN_UNARY_OPERATORS,
					  // i.e. if die corresponding command will POP at least one operand from the stack,
					  // re-push the boolean element POPed by the BRT command before executing the next command.
					  // The bytecode interpreter always examines the new command addressed relatively to the actual PC by the
					  // branch distance, without condidering a PC marker which could be set by a preceding
					  // IFELSE or IFNOTELSE command. If the new PC addesss (for the case the BRT branch is taken) is
					  // equal to the PC marker set by a preceding IFELSE or IFNOTELSE command, the next executed command
					  // in not the command at this PC address. But the opcode beeing testes for >=BEGIN_UNARY_OPERATORS
					  // is nevertheless the opcode found at the new PC address.
#ifndef WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION
		BRAND,		  // branch on false; pop top stack element only if branch is not taken.
					  // If branch is taken, convert the top stack element to bool
					  // and leave it on the stack
		BROR,		  // branch on true; pop top stack element only if branch is not taken.
					  // If branch is taken, convert the top stack element to bool
					  // and leave it on the stack
	    BRANDNOT,	  // Gleich wie BRAND mit der Ausnahme, dass der Operand vor der Verwendung logisch negiert wird.
					  // Wenn der Sprung ausgefhrt wird, wird also auch der Operand auf dem Stack durch seine logische
					  // Negierung ersetzt.
	    BRORNOT,		  // Gleich wie BROR mit der Ausnahme, dass der Operand vor der Verwendung logisch negiert wird.
					  // Wenn der Sprung ausgefhrt wird, wird also auch der Operand auf dem Stack durch seine logische
					  // Negierung ersetzt.
#endif

	    IFNOTELSE,	  // The same as IFELSE exept that the Operand POPed from the top of the stack is logically negated
					  // before used.
		IFELSE,		  // Read bytes after the IFELSE Opcode: Read d1 = first branch distance after IFELSE Opcode,
					  // read d2 = second branch distance after IFELSE Opcode.
					  // POP top stack element; if false, branch relatively by distance d1;
					  // If true, don't branch but set pc_marker = PC + d1 (PC is the PC of the IFELSE command
					  // If PC will reach pc_marker after
					  // executing following commands, automatically branch to PC0 + d1 + d2 .
					  // Note that IFELSE pushes two elements onto stack (after POPing the condition).
					  // The IFELSE command can also be used to space-efficiently implement while loops:
					  // To implement the loop while(cond) { commands; } produce these Opcodes:
					  //   ((1))  Evaluate cond ==> boolean is on top of stack
					  //   ((2))  IFELSE d1, d2     where d1=(positive) PC-difference ((4))-((3)) ,
					  //                                  d2=(negative) PC-difference ((1))-((4))
					  //   ((3))  commands
					  //   ((4))  ... (commands following the loop) ...
					  // Use the Opcodes BREAK or CONTINUE to break or continue the while loop.

		// Binary operators: pop two values off stack, compute, push result
		BEGIN_BINARY_OPERATORS,
		DUMMY_OPCODE3=BEGIN_BINARY_OPERATORS-1,
		// Comparison operators: pop two values off stack, push true or false
#ifdef WITH_OPCODE_LT
		LT,           // <
#endif
#ifdef WITH_OPCODE_LE
		LE,           // <=
#endif
#ifdef WITH_OPCODE_GT
		GT,           // >
#endif
#ifdef WITH_OPCODE_GE
		GE,           // >=
#endif
#ifdef WITH_OPCODE_EQ
		EQ,           // ==
#endif
#ifdef WITH_OPCODE_NE
		NE,           // !=
#endif

#ifdef WITH_OPCODE_STORET
		STORET,
#endif
	    STORE,
		STORED,
	    STOREPOP,
#ifdef WITH_OPCODE_SETLOCALI
		SETLOCALI,
		SETLOCALIT,
#endif

		BEGIN_ARITHMETIC_OPERATORS,
		DUMMY_OPCODE4=BEGIN_ARITHMETIC_OPERATORS-1,
		ADD,          // Add
		SUB,          // Subtract
		AND,          // And
		OR,          // Inclusive or
		XOR,          // Exclusive or
		SL,           // Left shift

		BEGIN_ARITHMETIC_OPERATORS_SIGNEDNESS_SENSITIVE,
		DUMMY_OPCODE5=BEGIN_ARITHMETIC_OPERATORS_SIGNEDNESS_SENSITIVE-1,
		MUL,          // Multiply
		DIV,          // Divide
		MOD,          // Remainder
		SR,           // Right shift


		/*
		* Binary operators on memory:
		*   pop top stack element,
		*   pop pointer from stack
		*   fetch value through pointer
		*   compute,
		*   then push result and also store it through pointer
		*/
		BEGIN_BINARY_OPERATORS_ON_MEMORY,
		DUMMY_OPCODE6=BEGIN_BINARY_OPERATORS_ON_MEMORY-1,
		ADDM,         // Add to memory (variable)
		SUBM,         // Subtract from memory (variable)
		ADDMPOP,         // Add to memory (variable)
		SUBMPOP,         // Subtract from memory (variable)
		ANDM,         // And to memory (variable)
		ORM,          // Inclusive or to memory (variable)
		XORM,         // Exclusive or to memory (variable)
		SLM,          // Left shift memory (variable)

		BEGIN_BIN_OP_ON_MEM_SIGNEDNESS_SENSITIVE,
		DUMMY_OPCODE7=BEGIN_BIN_OP_ON_MEM_SIGNEDNESS_SENSITIVE-1,
		MULM,         // Multiply with memory (variable)
		DIVM,         // Divide by memory (variable)
		MODM,         // Remainder to memory
		SRM,          // Right shift memory (variable)


		MAX_OPCODE_NUMBER,
		BEGINN_CONTROL_OPCODES_ONLY_USED_BY_COMPILER=MAX_OPCODE_NUMBER,
		DUMMY_OPCODE8=BEGINN_CONTROL_OPCODES_ONLY_USED_BY_COMPILER-1

#ifdef WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION
		,BRAND,		  // branch on false; pop top stack element only if branch is not taken.
					  // If branch is taken, convert the top stack element to bool ( uint8(0) or uint8(1) )
					  // and leave it on the stack
		BROR,		  // branch on true; pop top stack element only if branch is not taken.
					  // If branch is taken, convert the top stack element to bool ( uint8(0) or uint8(1) )
					  // and leave it on the stack
	    BRANDNOT,	  // Gleich wie BRAND mit der Ausnahme, dass der Operand vor der Verwendung logisch negiert wird.
					  // Wenn der Sprung ausgefhrt wird, wird also auch der Operand auf dem Stack durch seine logische
					  // Negierung ersetzt.
	    BRORNOT		  // Gleich wie BROR mit der Ausnahme, dass der Operand vor der Verwendung logisch negiert wird.
					  // Wenn der Sprung ausgefhrt wird, wird also auch der Operand auf dem Stack durch seine logische
					  // Negierung ersetzt.
#endif
	};

#ifdef PRINT_COMMANDS_AT_RUNTIME
	static void print_command(int cmd, PROGRAM_COUNTER_TYPE pc_of_command, VARIABLE_INDEX_TYPE sp)
	{
#define MAKE_COMMAND_AND_STRING(cmd) { cmd, #cmd }
		static struct CmdName
		{
			int cmd;
			const char *name;
		} CommandNames[]=
		{
			MAKE_COMMAND_AND_STRING(CALL),
#ifdef WITH_OPCODE_CALLI
			MAKE_COMMAND_AND_STRING(CALLI),
#endif

			MAKE_COMMAND_AND_STRING(RET),

			MAKE_COMMAND_AND_STRING(BR),

			MAKE_COMMAND_AND_STRING(PUSHC),
#ifdef USE_TABLE_FOR_CONSTANTS
			MAKE_COMMAND_AND_STRING(PUSHCT),
#endif
			MAKE_COMMAND_AND_STRING(PUSHC0U8),
			MAKE_COMMAND_AND_STRING(PUSHC0),
			MAKE_COMMAND_AND_STRING(PUSHCS),
			MAKE_COMMAND_AND_STRING(PUSHCSU8),

			MAKE_COMMAND_AND_STRING(BRK),
			MAKE_COMMAND_AND_STRING(CONT),
			MAKE_COMMAND_AND_STRING(SETTYPES),

#ifdef WITH_OPCODE_GETPREV
			MAKE_COMMAND_AND_STRING(GETPREV),
			MAKE_COMMAND_AND_STRING(GETPREVLOAD),
#endif
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
			MAKE_COMMAND_AND_STRING(SETPREVC),
			MAKE_COMMAND_AND_STRING(SETPREVCPOP),
#endif
#ifdef WITH_OPCODE_GETLOCAL
			MAKE_COMMAND_AND_STRING(GETLOCAL),
			MAKE_COMMAND_AND_STRING(GETLOCALLOAD),
#endif
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
			MAKE_COMMAND_AND_STRING(SETLOCALC),
			MAKE_COMMAND_AND_STRING(SETLOCALCPOP),
#endif
			MAKE_COMMAND_AND_STRING(STACKADDR),
			MAKE_COMMAND_AND_STRING(LOCALADDR),

			MAKE_COMMAND_AND_STRING(ADDCPREV),
			MAKE_COMMAND_AND_STRING(PREVADDC),
			MAKE_COMMAND_AND_STRING(ADDCLOCAL),
			MAKE_COMMAND_AND_STRING(LOCALADDC),
			MAKE_COMMAND_AND_STRING(ADDCPREVPOP),
			MAKE_COMMAND_AND_STRING(ADDCLOCALPOP),

#ifdef WITH_OPCODE_NEG
			MAKE_COMMAND_AND_STRING(NEG),
#endif
#ifdef WITH_OPCODE_COM
			MAKE_COMMAND_AND_STRING(COM),
#endif
			MAKE_COMMAND_AND_STRING(CAST),
#ifdef WITH_OPCODE_GETTYPE
			MAKE_COMMAND_AND_STRING(GETTYPE),
#endif
			MAKE_COMMAND_AND_STRING(POP),

			MAKE_COMMAND_AND_STRING(LOAD),
			MAKE_COMMAND_AND_STRING(LOADT),
#ifdef WITH_OPCODE_SETPREVT
			MAKE_COMMAND_AND_STRING(SETPREVT),
#endif
#ifdef WITH_OPCODE_SETPREV
			MAKE_COMMAND_AND_STRING(SETPREV),
			MAKE_COMMAND_AND_STRING(SETPREVPOP),
			MAKE_COMMAND_AND_STRING(GETPREVSTORE),
			MAKE_COMMAND_AND_STRING(GETPREVSTOREPOP),
#endif
#ifdef WITH_OPCODE_SETLOCALT
			MAKE_COMMAND_AND_STRING(SETLOCALT),
#endif
#ifdef WITH_OPCODE_SETLOCAL
			MAKE_COMMAND_AND_STRING(SETLOCAL),
			MAKE_COMMAND_AND_STRING(SETLOCALPOP),
			MAKE_COMMAND_AND_STRING(GETLOCALSTORE),
			MAKE_COMMAND_AND_STRING(GETLOCALSTOREPOP),
#endif

#ifdef WITH_OPCODE_GETLOCALI_AND_LOCALADDRI
			MAKE_COMMAND_AND_STRING(GETLOCALI),
			MAKE_COMMAND_AND_STRING(LOCALADDRI),
			MAKE_COMMAND_AND_STRING(STACKADDRI),
#endif

#ifdef WITH_OPCODE_PUSHN
			MAKE_COMMAND_AND_STRING(PUSHN),
#endif
#ifdef WITH_OPCODE_POPN
			MAKE_COMMAND_AND_STRING(POPN),
#endif
#ifdef WITH_OPCODE_JMPI
			MAKE_COMMAND_AND_STRING(JMPI),
#endif
			MAKE_COMMAND_AND_STRING(CALLEXT),

#ifdef WITH_OPCODE_NOT
			MAKE_COMMAND_AND_STRING(NOT),
#endif

			MAKE_COMMAND_AND_STRING(BRF),
			MAKE_COMMAND_AND_STRING(BRT),
#ifndef WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION
			MAKE_COMMAND_AND_STRING(BRAND),
			MAKE_COMMAND_AND_STRING(BROR),
			MAKE_COMMAND_AND_STRING(BRANDNOT),
			MAKE_COMMAND_AND_STRING(BRORNOT),
#endif

			MAKE_COMMAND_AND_STRING(IFNOTELSE),
			MAKE_COMMAND_AND_STRING(IFELSE),
#ifdef WITH_OPCODE_LT
			MAKE_COMMAND_AND_STRING(LT),
#endif
#ifdef WITH_OPCODE_LE
			MAKE_COMMAND_AND_STRING(LE),
#endif
#ifdef WITH_OPCODE_GT
			MAKE_COMMAND_AND_STRING(GT),
#endif
#ifdef WITH_OPCODE_GE
			MAKE_COMMAND_AND_STRING(GE),
#endif
#ifdef WITH_OPCODE_EQ
			MAKE_COMMAND_AND_STRING(EQ),
#endif
#ifdef WITH_OPCODE_NE
			MAKE_COMMAND_AND_STRING(NE),
#endif

#ifdef WITH_OPCODE_STORET
			MAKE_COMMAND_AND_STRING(STORET),
#endif
			MAKE_COMMAND_AND_STRING(STORE),
			MAKE_COMMAND_AND_STRING(STORED),
			MAKE_COMMAND_AND_STRING(STOREPOP),
#ifdef WITH_OPCODE_SETLOCALI
			MAKE_COMMAND_AND_STRING(SETLOCALI),
			MAKE_COMMAND_AND_STRING(SETLOCALIT),
#endif

			MAKE_COMMAND_AND_STRING(ADD),
			MAKE_COMMAND_AND_STRING(SUB),
			MAKE_COMMAND_AND_STRING(AND),
			MAKE_COMMAND_AND_STRING(OR),
			MAKE_COMMAND_AND_STRING(XOR),
			MAKE_COMMAND_AND_STRING(SL),

			MAKE_COMMAND_AND_STRING(MUL),
			MAKE_COMMAND_AND_STRING(DIV),
			MAKE_COMMAND_AND_STRING(MOD),
			MAKE_COMMAND_AND_STRING(SR),

			MAKE_COMMAND_AND_STRING(ADDM),
			MAKE_COMMAND_AND_STRING(SUBM),
			MAKE_COMMAND_AND_STRING(ADDMPOP),
			MAKE_COMMAND_AND_STRING(SUBMPOP),
			MAKE_COMMAND_AND_STRING(ANDM),
			MAKE_COMMAND_AND_STRING(ORM),
			MAKE_COMMAND_AND_STRING(XORM),
			MAKE_COMMAND_AND_STRING(SLM),

			MAKE_COMMAND_AND_STRING(MULM),
			MAKE_COMMAND_AND_STRING(DIVM),
			MAKE_COMMAND_AND_STRING(MODM),
			MAKE_COMMAND_AND_STRING(SRM),

			-1, 0
		};
#undef MAKE_COMMAND_AND_STRING
		for(int n=0 ; -1!=CommandNames[n].cmd && 0!=CommandNames[n].name; n++)
			if(n!=CommandNames[n].cmd)
			{
				PRINT_COMMAND_IF_ENABLED("****** opcodes in print_command() differ from opcodes in enum Opcodes.", 0, 0);
				return;
			}
		for(int n=0 ; -1!=CommandNames[n].cmd && 0!=CommandNames[n].name; n++)
			if(cmd==CommandNames[n].cmd)
			{
				PRINT_COMMAND_IF_ENABLED(CommandNames[n].name, pc_of_command, -(SIGNED_VARIABLE_INDEX_TYPE)sp);
				return;
			}
		PRINT_COMMAND_IF_ENABLED("(Unknown opcode)", pc_of_command, -(SIGNED_VARIABLE_INDEX_TYPE)sp);
	}
#endif

	typedef uint8(*get_code_nibble_proc)(PROGRAM_COUNTER_TYPE pc);
	typedef uint8(*get_code_byte_proc)(PROGRAM_COUNTER_TYPE pc);
	get_code_nibble_proc get_code_nibble;
	get_code_byte_proc get_code_byte;
	uint8 *stack, *stack_end, *memory;
	VARIABLE_INDEX_TYPE stack_size, sp, memory_size;
#if defined(WITH_OPCODE_GETLOCAL) || defined(WITH_OPCODE_SETLOCAL)
	VARIABLE_INDEX_TYPE act_function_stackbase;
	PROGRAM_COUNTER_TYPE act_function_codebase;
#endif
	PROGRAM_COUNTER_TYPE pc;
	PROGRAM_COUNTER_TYPE pc_marker;
	PROGRAM_COUNTER_TYPE pc_vor_opcode;

	enum { max_makro_recursion_depth=8 };
	enum { makro_escape_nibble=10, makro_escape_escape=makro_escape_nibble,
			makro_escape_ret=9, makro_escape_varsize_param=8, makro_escape_fixedsize_param=7 };
			// Es muss makro_escape_escape==makro_escape_nibble sein, und die anderen makro_escape_...-enums mssen verschieden
			// von makro_escape_nibble sein.
	PROGRAM_COUNTER_TYPE makro_pc_stack[max_makro_recursion_depth];
	uint8 makro_sp;
	PROGRAM_COUNTER_TYPE akt_makro_size;
	PROGRAM_COUNTER_TYPE table4bit_opcodes[16];
	bool makro_escape_sequences_temporarily_disabled, reading_makro_parameter;
	PROGRAM_COUNTER_TYPE pc_vor_code_nibble;

	uint8 get_next_code_nibble()
	{
		if(0!=akt_makro_size)
		{
			test_makro_size:
			if(0==(--akt_makro_size))
			{
				if(!makro_escape_sequences_temporarily_disabled)
					goto makro_ret;
				makro_escape_sequences_temporarily_disabled=false;
			}
			// Hier muss nicht auf pc==pc_marker getestet werden, denn es ist nicht erlaubt, dass eine durch eine IFELSE-Anweisung
			// definierte Marke (= die PC-Position des ersten Befehls nach dem Ende des if-Teils) in einen nicht durch Escape-Sequenzen
			// codierten Bereich zeigt.
			pc_vor_code_nibble=pc;
			return get_code_nibble(pc++);
		}
		else while(1)
		{
			if(pc==pc_marker)
				return 255;
			pc_vor_code_nibble=pc;
			test_escape:
			uint8 i; i=get_code_nibble(pc++);
			if(0==makro_sp || makro_escape_nibble!=i || makro_escape_escape==(i=get_code_nibble(pc++)))
				return i;
			akt_makro_size=0;
			if(makro_escape_ret==i)
			{
				makro_ret:
				//dprintf("%d ==> %d ######## Ret  makro %x  ==>   %x\r\n", makro_sp, makro_sp-1, pc, makro_pc_stack[makro_sp-1]);
				if(reading_makro_parameter)
				{
					reading_makro_parameter=false;
					makro_pc_stack[makro_sp-2]=pc;
				}
				pc=makro_pc_stack[--makro_sp];
				continue;
			}
			if(makro_escape_fixedsize_param==i)
			{
				if(15==(akt_makro_size=get_code_nibble(pc++)))
				{
					makro_escape_sequences_temporarily_disabled=true; // Escape-Sequenzen deaktivieren
					akt_makro_size=get_code_byte(pc);
					pc+=2;
						// ... und zwar fr die nun ausgelesene Zahl von Nibbles.
					goto test_makro_size;
				}
				else if(0==akt_makro_size)
				{
					read_byte_block(pc, MAX_NR_BYTES_FOR_PROGRAM_COUNTER, (uint8*)(&akt_makro_size));
					pc+=2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER;
					// Wenn nun fr akt_makro_size der Wert 0 gelesen wurde, handelt es sich um einen Fehler.
					ASSERT(0!=akt_makro_size, "Nonzero makro parameter size expected.", pc);
				}
			}
			else if(makro_escape_varsize_param!=i)
			{	// Es liegt keine Escape-Sequenz vor. Geliefert wird genau das, was durch von get_code_nibble() gelesen wurde.
				--pc; // beim nchsten get_nibble()-Aufruf wird i zurckgeliefert.
				// pc_vor_code_nibble hat schon den korrekten Wert
				return makro_escape_nibble;
			}
			makro_pc_stack[makro_sp++]=pc;
			pc_vor_code_nibble=pc=makro_pc_stack[makro_sp-2];
			reading_makro_parameter=true;
			if(makro_escape_varsize_param==i)
				goto test_escape; // pc_vor_code_nibble muss vor diesem goto zugewiesen werden
			return get_code_nibble(pc++); // Im Fall makro_escape_fixedsize_param ist der Parameter immer mindestens 1
				// Nibble lang.
		}
	}
	uint8 get_next_opcode()
	{
		uint8 i;
		PROGRAM_COUNTER_TYPE b;
		while(1)
		{
			if(255==(i=get_next_code_nibble()))
				return 255;
			pc_vor_opcode=pc_vor_code_nibble; // // pc_vor_opcode darf nicht vor get_next_code_nibble() zugewiesen werden, weil ein
				// Makro-Rcksprung durch get_next_code_nibble() erfolgt sein knnte.
			if(i<ProgramHeader.limit_for_4bit_opcodes())
				b=table4bit_opcodes[i];
			else
			{
				i=((i-ProgramHeader.limit_for_4bit_opcodes())<<4) + get_next_code_nibble();
				if(i<ProgramHeader.limit_for_8bit_opcodes())
					read_byte_block(ProgramHeader.opcode_transformation_tables(1)+2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER*i,
						MAX_NR_BYTES_FOR_PROGRAM_COUNTER, (uint8*)(&b));
				else
				{
					b=(((PROGRAM_COUNTER_TYPE)(i-ProgramHeader.limit_for_8bit_opcodes()))<<4) + get_next_code_nibble();
					if(b<ProgramHeader.limit_for_nbit_opcodes(0))
						read_byte_block(ProgramHeader.opcode_transformation_tables(2)+2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER*b,
							MAX_NR_BYTES_FOR_PROGRAM_COUNTER, (uint8*)(&b));
					else
					{
						for(i=1; i<=5; i++)
						{
							b=((b-ProgramHeader.limit_for_nbit_opcodes(i-1))<<4) + get_next_code_nibble();
							if(5==i || b<ProgramHeader.limit_for_nbit_opcodes(i))
							{
								read_byte_block(ProgramHeader.opcode_transformation_tables(2+i)+2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER*b,
												MAX_NR_BYTES_FOR_PROGRAM_COUNTER, (uint8*)(&b));
								return (uint8)b;
							}
						}
					}
				}
			}
			if(b>=MAX_OPCODE_NUMBER) // Es wird davon ausgegangen, dass ein Opcode (also ein Wert im enum Opcode) keinen Wert von MAX_OPCODE_NUMBER oder
			{		// grer haben kann und dass die PC-Adresse, an der Makros stehen, mindestens den Wert MAX_OPCODE_NUMBER hat.
				//dprintf("%d ==> %d ######## Call makro %x  ==>   %x\r\n", makro_sp, makro_sp+1, pc, (uint32)b);
				makro_pc_stack[makro_sp++]=pc;
				pc=b;
				akt_makro_size=0; // Makro-Ende wird durch die Nibblefolge makro_escape_nibble , makro_escape_ret angezeigt.
				continue;
			}
			return (uint8)b;
		}
	}

	VARIABLE_INDEX_TYPE last_stack_addrs[32];
	uint8 last_stack_addrs_begin, last_stack_addrs_akt;

	bool pop_stack(Value *val, bool set_pc_if_marker=false, bool without_delete_last_stack_addr=false)
	{
		static bool last_ret=false;
		ASSERT(0!=sp, "Unexpected POP", pc);
		uint8 t=get_var(sp, val);
		// ddprintf("\r\nPop  stack type %u:   sp_vor=%u    sp_nach=%u\r\n", t, sp, sp+Value::sbytesize(t));
		sp+=Value::sbytesize(t);
		if(t_byte_block_length==val->type)
		{
			get_var(sp, val);
			ASSERT(VarIndexTypeNr==val->type, "Expected type VarIndexTypeNr on top of stack", pc);
			sp+=MAX_NR_BYTES_FOR_VARIABLE_INDEX + CHOOSE_VI_TYPE(*val);
		}
		else
		{
			if(!without_delete_last_stack_addr)
				if(0!=((last_stack_addrs_akt-last_stack_addrs_begin)&31))
				{
					// ddprintf("--------- Delete  last_stack_addr. ------------ sp_vor=%u   akt_index=%u   begin_index=%u\r\n", sp-Value::sbytesize(t), last_stack_addrs_akt-1, last_stack_addrs_begin);
					--last_stack_addrs_akt;
				}
			if(t_pc_marker_address==val->type)
			{
				pc_marker=CHOOSE_PC_TYPE(*val);
				ASSERT(false==last_ret, "Expected type 'PROGRAM_COUNTER_TYPE' on top of stack", pc);
				//check_last_stack_addrs();
				return (last_ret=true);
			}
			if(last_ret)
			{
				ASSERT(val->type==ProgramCounterTypeNr, "Expected type 'PROGRAM_COUNTER_TYPE' on top of stack", pc);
				if(set_pc_if_marker)
					pc=CHOOSE_PC_TYPE(*val);
			}
			//check_last_stack_addrs();
		}
		return (last_ret=false);
	}

	void push_stack_without_save_last_stack_addr(Value *val)
	{
		// ddprintf("\r\nPush stack type %u:   sp_vor=%u    sp_nach=%u\r\n", val->type, sp, sp-Value::sbytesize(val->type));
		sp-=Value::sbytesize(val->type);
#ifdef CALCULATE_MAXIMUM_STACK_USAGE
		if(sp<minimum_stack_pointer)
			minimum_stack_pointer=sp;
#endif
		ASSERT(sp>=VARIABLE_INDEX_TYPE(-stack_size), "Out of stack memory", pc);
		set_var(sp, val); // Im Falle von OUT_OF_STACK_MEMORY muss set_var dafr sorgen, dass nicht
			// in einen unerlaubten Speicherbereich geschrieben wird.
	}

	void push_stack(Value *val)
	{
		push_stack_without_save_last_stack_addr(val);
		// ddprintf("--------- Save to last_stack_addr. ------------ sp_nach=%u   akt_index=%u    begin_index=%u\r\n", sp, last_stack_addrs_akt, last_stack_addrs_begin);
		last_stack_addrs[(last_stack_addrs_akt++)&31]=sp;
		if(0==((last_stack_addrs_akt-last_stack_addrs_begin)&31))
			++last_stack_addrs_begin;
		//check_last_stack_addrs();
	}

	void cast(Value *val, uint8 typ)
	{
#ifdef BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE
		if(((signed char*)(&val->u8))-((signed char*)(&val->u64))==0 && ((signed char*)(&val->u16))-((signed char*)(&val->u64))==0 &&
				((signed char*)(&val->u32))-((signed char*)(&val->u64))==0)
		{
			if(typ<t_real)
			{
				if(t_bool==typ)
					cast_to_bool(val);
				else if(val->type<t_real)
				{
					uint8 i=(Value::t_void==val->type) ? 0 : val->bytesize();
					if(Value::sbytesize(typ) > i)
					{
						uint8 fillbytes = (val->is_signed_integer() && 0!=(((uint8*)&val->u64)[i-1]&128)) ? 255 : 0;
						while(i<8) // arbeitet auch korrekt, wenn val vom Typ t_void ist.
							((uint8*)&val->u64)[i++]=fillbytes;
					}
				}
				else if(Value::s_is_signed_integer(typ))
				{
#ifdef cast_real_to_largest_sint
					cast_real_to_largest_sint(*val);
#endif
				}
				else
				{
#ifdef cast_real_to_largest_uint
					cast_real_to_largest_uint(*val);
#endif
				}
			}
			else
			{
				if(val->type<t_real)
				{
					uint8 i=(Value::t_void==val->type) ? 0 : val->bytesize();
					uint64 x=(val->is_signed_integer() && 0!=(((uint8*)&val->u64)[i-1]&128)) ? ((uint64)(-1)) : 0;
					while(0!=(i--))
						x=(x<<8) | (((uint8*)&val->u64)[i]);
					if(Value::s_is_signed_integer(val->type))
					{
#ifdef cast_largest_sint_to_real
						cast_largest_sint_to_real(*val);
#endif
					}
					else
					{
#ifdef cast_largest_uint_to_real
						cast_largest_uint_to_real(*val);
#endif
					}
				}
			}
		}
		else
#endif
		{ // Der folgende uerst speicherintensive Code wird dann NICHT ausgefhrt und compiliert,
			// wenn es sich um eine Little Endian Maschine handelt (BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE definiert) und die
			// Startadressen, an denen die verschiedenen Typen der union gespeichert werden, bereinstimmen.
			const unsigned int t_anz=anz_user_types;
			switch(val->type*t_anz + typ)
			{
				case t_u8 *t_anz + t_bool:								val->u8=(0!=val->t_u8) ? 1 : 0; break;
				case t_u8 *t_anz + t_u16: case t_u8 *t_anz + t_i16:		val->u16=(uint16)val->u8; break;
				case t_u8 *t_anz + t_u32: case t_u8 *t_anz + t_i32:		val->u32=(uint32)val->u8; break;
				case t_u8 *t_anz + t_u64: case t_u8 *t_anz + t_i64:		val->u64=(uint64)val->u8; break;
				case t_u8 *t_anz + t_real:								val->r=(real_type)val->u8; break;

				case t_i8 *t_anz + t_bool:								val->u8=(0!=val->t_i8) ? 1 : 0; break;
				case t_i8 *t_anz + t_u16: case t_i8 *t_anz + t_i16:		val->u16=(int16)val->i8; break;
				case t_i8 *t_anz + t_u32: case t_i8 *t_anz + t_i32:		val->u32=(int32)val->i8; break;
				case t_i8 *t_anz + t_u64: case t_i8 *t_anz + t_i64:		val->u64=(int64)val->i8; break;
				case t_i8 *t_anz + t_real:								val->r=(real_type)val->i8; break;

				case t_u16*t_anz + t_bool:								val->u8=(0!=val->t_u16) ? 1 : 0; break;
				case t_u16*t_anz + t_u8 : case t_u16*t_anz + t_i8 :		val->u8=(uint8)val->u16; break;
				case t_u16*t_anz + t_u32: case t_u16*t_anz + t_i32:		val->u32=(uint32)val->u16; break;
				case t_u16*t_anz + t_u64: case t_u16*t_anz + t_i64:		val->u64=(uint64)val->u16; break;
				case t_u16*t_anz + t_real:								val->r=(real_type)val->u16; break;

				case t_i16*t_anz + t_bool:								val->u8=(0!=val->t_i16) ? 1 : 0; break;
				case t_i16*t_anz + t_u8 : case t_i16*t_anz + t_i8 :		val->u8=(int8)val->i16; break;
				case t_i16*t_anz + t_u32: case t_i16*t_anz + t_i32:		val->u32=(int32)val->i16; break;
				case t_i16*t_anz + t_u64: case t_i16*t_anz + t_i64:		val->u64=(int64)val->i16; break;
				case t_i16*t_anz + t_real:								val->r=(real_type)val->i16; break;

				case t_u32*t_anz + t_bool:								val->u8=(0!=val->t_u32) ? 1 : 0; break;
				case t_u32*t_anz + t_u8 : case t_u32*t_anz + t_i8 :		val->u8=(uint8)val->u32; break;
				case t_u32*t_anz + t_u16: case t_u32*t_anz + t_i16:		val->u16=(uint16)val->u32; break;
				case t_u32*t_anz + t_u64: case t_u32*t_anz + t_i64:		val->u64=(uint64)val->u32; break;
				case t_u32*t_anz + t_real:								val->r=(real_type)val->u32; break;

				case t_i32*t_anz + t_bool:								val->u8=(0!=val->t_i32) ? 1 : 0; break;
				case t_i32*t_anz + t_u8 : case t_i32*t_anz + t_i8 :		val->u8=(int8)val->i32; break;
				case t_i32*t_anz + t_u16: case t_i32*t_anz + t_i16:		val->u16=(int16)val->i32; break;
				case t_i32*t_anz + t_u64: case t_i32*t_anz + t_i64:		val->u64=(int64)val->i32; break;
				case t_i32*t_anz + t_real:								val->r=(real_type)val->i32; break;

				case t_u64*t_anz + t_bool:								val->u8=(0!=val->t_u64) ? 1 : 0; break;
				case t_u64*t_anz + t_u8 : case t_u64*t_anz + t_i8 :		val->u8=(uint8)val->u64; break;
				case t_u64*t_anz + t_u16: case t_u64*t_anz + t_i16:		val->u16=(uint16)val->u64; break;
				case t_u64*t_anz + t_u32: case t_u64*t_anz + t_i32:		val->u32=(uint32)val->u64; break;
				case t_u64*t_anz + t_real:								val->r=(real_type)val->u64; break;

				case t_i64*t_anz + t_bool:								val->u8=(0!=val->t_i64) ? 1 : 0; break;
				case t_i64*t_anz + t_u8 : case t_i64*t_anz + t_i8 :		val->u8=(int8)val->i64; break;
				case t_i64*t_anz + t_u16: case t_i64*t_anz + t_i16:		val->u16=(int16)val->i64; break;
				case t_i64*t_anz + t_u32: case t_i64*t_anz + t_i32:		val->u32=(int32)val->i64; break;
				case t_i64*t_anz + t_real:								val->r=(real_type)val->i64; break;

				case t_real*t_anz + t_bool:	val->u8=(0!=val->r) ? 1 : 0; break;
				case t_real*t_anz + t_u8 :	val->u8=(uint8)val->r; break;
				case t_real*t_anz + t_i8 :	val->i8=(int8)val->r; break;
				case t_real*t_anz + t_u16:	val->u16=(uint16)val->r; break;
				case t_real*t_anz + t_i16:	val->i16=(int16)val->r; break;
				case t_real*t_anz + t_u32:	val->u32=(uint32)val->r; break;
				case t_real*t_anz + t_i32:	val->i32=(int32)val->r; break;
				case t_real*t_anz + t_u64:	val->u64=(uint64)val->r; break;
				case t_real*t_anz + t_i64:	val->i64=(int64)val->r; break;

				default:
				if(t_void==val->type)
					{ if(t_real==typ) val->r=0; else val->u64=0; }
			}
		}
		val->type=typ;
	}
	void cast_up(Value *val1, Value *val2)
	{	// Genau einer der Werte *val1 oder *val2 wird auf den Typ des anderen Werts gecastet; der andere Wert bleibt immer
		// unverndert (auch in Bezug auf signed oder unsigned). Es wird auf die hhere der beiden Bitzahlen gecastet.
		// Im Zweifelsfall hat unsigned Vorrang vor signed: Falls der Wert aus *val1, *val2, der die grere Bitzahl besitzt,
		// unsigned ist und der Wert mit der kleineren Bitzahl signed ist, wird auf unsigned gecastet (mit einer unter Umstnden
		// ungewollten Konversion eines ursprnglich negativen Werts in einen positiven Wert).
		// Falls sowohl *val1 als auch *val2 die gleiche Bitzahl haben, sich aber in der Eigenschaft 'signed' bzw. 'unsigned'
		// unterscheiden, wird der signed-Typ auf 'unsigned' gecastet.
		if(val1->type > val2->type)
			cast(val2, val1->type);
		else
			cast(val1, val2->type);
	}

	void cast_to_bool(Value *v)
	{
#if defined(ASSUME_VALUE_UNION_EQUAL_ADDRESSES) && defined(BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE)
			for(uint8 i=v->bytesize(); i-- ; )
				if(0!=((uint8*)(&v->u64))[i])
					{ v->u8=1; break; }
#else
		switch(v->type>>1)
		{
			case t_void>>1: // Hier kann v->type==t_void oder t_bool sein, daher muss noch mal abgefragt werden.
				if(t_void==v->type) v->u8=0; break;
	#if defined(WITH_INTEGER8)
			case t_u8>>1: v->u8=uint8(0!=v->u8 ? 1 : 0); break;
	#endif
	#if defined(WITH_INTEGER16)
			case t_u16>>1: v->u8=uint8(0!=v->u16 ? 1 : 0); break;
	#endif
	#if defined(WITH_INTEGER32)
			case t_u32>>1: v->u8=uint8(0!=v->u32 ? 1 : 0); break;
	#endif
	#if defined(WITH_INTEGER64)
			case t_u64>>1: v->u8=uint8(0!=v->u64 ? 1 : 0); break;
	#endif
	#if defined(WITH_REAL)
			case t_real>>1: v->u8=uint8(0!=v->r ? 1 : 0); break;
	#endif
		}
#endif
		v->type=t_bool;
	}

	uint64 read_coded_integer(uint8 max_val_if_not_escape, uint8 nr_nibbles_if_escape, uint8 _signed)
	{	// Falls nr_bytes_if_escape!=0 ist, gilt:
		// Bei der Decodierung wird zunchst ein Byte gelesen, falls max_val_if_not_escape==255 ist, bzw. ein Nibble,
		// falls max_val_if_not_escape==15 ist (andere Werte sind fr max_val_if_not_escape nicht zulssig).
		// Wenn der gelesene Wert nicht gleich max_val_if_not_escape ist, wird er als Integer
		// fr die weitere Verarbeitung verwendet.
		// Wenn er jedoch gleich max_val_if_not_escape ist, wird er verworfen,
		// und es werden nr_bytes_if_escape Bytes gelesen, welche eine Integer Zahl im Big Endian Format darstellen.
		// Wenn signed!=0 bergeben wird, wird der aus dem einem oder den mehreren gelesenen Bytes erzeugte
		// Integer genau dann logisch rechts geshiftet und dann logisch invertiert (~ , fhrt zu einer negativen Zahl),
		// wenn sein niederwertigstes Bit gesetzt ist; andernfalls wird der Integer nur logisch nach rechts geshiftet (fhrt zu einer
		// nicht-negativen Zahl).
		uint64 t=get_next_code_nibble();
		if(255==max_val_if_not_escape)
			t=(t<<4) | get_next_code_nibble();
		if((0==nr_nibbles_if_escape) ? (t>=max_val_if_not_escape-8 &&
				(nr_nibbles_if_escape=((t==max_val_if_not_escape) ? 16 : (t+9-max_val_if_not_escape)), true))
				: (max_val_if_not_escape==t))
		{
			t=0;
			while(nr_nibbles_if_escape--)
				t=(t<<4)+get_next_code_nibble(); // ACHTUNG: Nibbles mssen im Big Endian Format abgelegt sein.
		}
		if(_signed)
		{	// Der Rechtsshift eines vorzeichenbehafteten Integers ist Compiler-abhngig. Daher wird hier
			// zunchst ein Rechtsshift eines vorzeichenlosen Integers ausgefhrt und dann das Vorzeichen angepasst.
			if(0!=(t&1))
				return (uint64)((int64)~((((uint64)t) >> 1)));
			else
				return (uint64)((int64)(((uint64)t) >> 1));
		}
		else
			return t;
	}

#define read_coded_sintegerVI(max_val_if_not_escape) ((SIGNED_VARIABLE_INDEX_TYPE)(read_coded_integer((max_val_if_not_escape), 2*MAX_NR_BYTES_FOR_VARIABLE_INDEX, true)))
#define read_coded_uintegerVI(max_val_if_not_escape) ((VARIABLE_INDEX_TYPE)read_coded_integer((max_val_if_not_escape), 2*MAX_NR_BYTES_FOR_VARIABLE_INDEX, false))

	void functioncall(Value &v1)
	{
		cast(&v1, ProgramCounterTypeNr);
		Value v;
		CHOOSE_PC_TYPE(v)=pc;
		pc=CHOOSE_PC_TYPE(v1);

		// Im Folgenden werden die auf dem Stack bergebenen Funktionsparameter in die Typen konvertiert,
		// die fr die Funktion erwartet werden. Wenn sich dabei die Byte-Gren der Parameter ndern, wird
		// die Stackgre auch entsprechend angepasst.
		union
		{
			struct { uint8 anzparam, i; } pm;
			VARIABLE_INDEX_TYPE n;
		};
		VARIABLE_INDEX_TYPE size_soll=0, sp2=sp;
		pm.anzparam=get_code_nibble(pc-2)<<3;	// Die beiden get_code_nibble()-Aufrufe drfen nicht in einem Ausdruck erscheinen,
		pm.anzparam|=get_code_nibble(pc-1)>>1;	// weil der C++-Standard keine Reihenfolge fr die Auswertung von x | y vorschreibt.
		if(pm.anzparam>=31)
			last_stack_addrs_akt=last_stack_addrs_begin;
		else
		{
			ASSERT(((last_stack_addrs_akt-last_stack_addrs_begin)&31)>=pm.anzparam, "Function parameters expected in array last_stack_addrs[]", pc);
			last_stack_addrs_akt-=pm.anzparam;
		}
	
		uint8 types_differ=false;
		for(pm.i=0; pm.i<pm.anzparam; pm.i++)
		{
			uint8 t=get_code_nibble(pc-3-pm.i);
			if(get_var(sp2, &v1)!=t)
				types_differ=true;
			sp2+=(CHOOSE_VI_TYPE(v1)=Value::sbytesize(v1.type));
			//size_vorh+=CHOOSE_VI_TYPE(v1);
			VARIABLE_INDEX_TYPE v2=Value::sbytesize(t);
			size_soll+=(v2>CHOOSE_VI_TYPE(v1)) ? v2 : CHOOSE_VI_TYPE(v1);
		}
		if(types_differ)
		{
			// size_vorh=sp2-sp
			size_soll=sp2-size_soll;
			//size_vorh=sp;
			sp2=sp;
			sp=size_soll;
			for(pm.i=0; pm.i<pm.anzparam; pm.i++)
			{
				get_var(sp2, &v1);
				sp2+=Value::sbytesize(v1.type);
				cast(&v1, get_code_nibble(pc-3-pm.i));
				set_var(size_soll, &v1, false);
				size_soll+=Value::sbytesize(v1.type);
			}
			// size_vorh ==> sp2
			if(sp2!=size_soll) // ggf. Ausfhrungszeit sparen
			{
				n=size_soll;
				size_soll-=sp;
				sp=sp2;
				/*sp2=size_soll;
				size_soll-=sp;
				sp=size_vorh;*/
				while(size_soll--)
				{
					*((uint8*)(stack_end-(VARIABLE_INDEX_TYPE)(-(--sp))))=*((uint8*)(stack_end-(VARIABLE_INDEX_TYPE)(-(--n))));
					set_stack_byte_associated_type_info(sp, get_stack_byte_associated_type_info(n));
				}
			}
		}
		v.type=ProgramCounterTypeNr;
		push_stack_without_save_last_stack_addr(&v); // Rcksprungadresse
		CHOOSE_VI_TYPE(v)=act_function_codebase;
		push_stack_without_save_last_stack_addr(&v); // sichere altes act_function_codebase
		act_function_codebase=pc;
#if defined(WITH_OPCODE_GETLOCAL) || defined(WITH_OPCODE_SETLOCAL)
		CHOOSE_VI_TYPE(v)=act_function_stackbase;
		v.type=VarIndexTypeNr;
		push_stack_without_save_last_stack_addr(&v); // sichere altes act_function_stackbase
		act_function_stackbase=sp;
#endif	
#ifdef AUTO_TYPEINFO_FOR_UNINITIALIZED_LOCALS
		v.u64=0; // Null-Werte auf Stack PUSHen
		VARIABLE_INDEX_TYPE d;
		while(0!=(d=read_coded_integer(15, 2*MAX_NR_BYTES_FOR_VARIABLE_INDEX, 0))) // d als unsigned lesen
		{
			if(d<10)
			{
				while(d--)
				{
					v.type=get_next_code_nibble();
					push_stack_without_save_last_stack_addr(&v);
				}
			}
			else
			{
				v.type=get_next_code_nibble();
				for(d-=5; --d; ) // ACHTUNG: d=10 erzeugt 4 Variablen vom durch get_next_code_nibble() bestimmten Typ.
					push_stack_without_save_last_stack_addr(&v);
			}
		}
		v1.i32=read_coded_integer(255, 2*MAX_NR_BYTES_FOR_VARIABLE_INDEX, 0);
		v1.type=t_i32;
		cast(&v1, VarIndexTypeNr);
		CHOOSE_VI_TYPE(v)=CHOOSE_VI_TYPE(v1);
		sp-=CHOOSE_VI_TYPE(v);
				// Die folgende Schleife muss dann ausgefhrt werden, wenn Typeninformationen fr nicht-initialisierte
				// lokale Variablen noch nicht gleich am Anfang (d.h. hier in functioncall()) geschrieben werden knnen (siehe Variablen
				// max_byte_addressed_local_variable_count und first_uninit_vars_created_bytesize in ProgramStructure_Function
				// und ProgramStructure_VoidFunction).
		while(CHOOSE_VI_TYPE(v)--)
			set_stack_byte_associated_type_info(sp+CHOOSE_VI_TYPE(v), 2);
		CHOOSE_VI_TYPE(v1)=act_function_stackbase-sp; // Alle uninitialisierten Variablen mssen spter mit einem einzigen POP-Befehl
			// entfernt werden, weil nur so aufgrund des Typs "t_byte_block_length" verhindert werden kann, dass beim Entfernen
			// Vernderungen an last_stack_addrs[] und den Indizes last_stack_addrs_akt und last_stack_addrs_begin vorgenommen werden.
		push_stack_without_save_last_stack_addr(&v1);
		v1.type=t_byte_block_length; // Obwohl Typ t_byte_block_length, wird nur ein Byte geschrieben
		push_stack_without_save_last_stack_addr(&v1);
#endif
#ifdef CALCULATE_MAXIMUM_STACK_USAGE
		if(sp<minimum_stack_pointer)
			minimum_stack_pointer=sp;
#endif
	}

	void (*extern_function_invoker)(VARIABLE_INDEX_TYPE function_number, BytecodeInterpreter& interp);

	BytecodeInterpreter(uint8 *__stack, VARIABLE_INDEX_TYPE __stack_size, uint8 *__memory, VARIABLE_INDEX_TYPE __memory_size,
						get_code_nibble_proc __get_code_nibble, get_code_byte_proc __get_code_byte,
						void (*__extern_function_invoker)(PROGRAM_COUNTER_TYPE function_number, BytecodeInterpreter& interp)=0)
		: stack(__stack), stack_end(__stack+__stack_size), stack_size(__stack_size), memory(__memory), memory_size(__memory_size),
			get_code_nibble(__get_code_nibble), get_code_byte(__get_code_byte), extern_function_invoker(__extern_function_invoker)
	{
		ASSERT(stack_size>0, "Stack size is zero", pc);
#ifdef BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE
		unsigned short test_little_endian=1u;
		if(((unsigned char)1u) != *((unsigned char*)(&test_little_endian)))
		{
			ASSERT(false, "BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE is defined on a big-endian machine.", pc);
			while(1) // ENDLOSSCHLEIFE !!!! Wenn diese Schleife ausgefhrt wird, ist BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE
				;	 // definiert, obwohl es sich nicht um eine Little Endian Maschine handelt.
		}
#endif
	}
	void execute_bytecode(PROGRAM_COUNTER_TYPE __pc);
};

void BytecodeInterpreter::execute_bytecode(PROGRAM_COUNTER_TYPE __pc)
{
	stack[0]=3; // Datentyp Markierung
	last_stack_addrs_akt=last_stack_addrs_begin=0;
	for(VARIABLE_INDEX_TYPE i=0; i<memory_size; i++)
		memory[i]=0;
	read_byte_block(0, sizeof(BytecodeProgramHeader), (uint8*)&ProgramHeader.internal_memory[0]);
	read_byte_block(ProgramHeader.global_begin_addrs_of_elementary_types_start(),
		(3*Value::anz_user_types+1)*MAX_NR_BYTES_FOR_VARIABLE_INDEX, (uint8*)(&global_begin_addrs_of_elementary_types[0]));
	ASSERT(memory_size>=global_begin_addrs_of_elementary_types[2*Value::anz_user_types], "Insufficient memory for global variables", pc);
	read_byte_block(2*ProgramHeader.global_var_initvalues_start(0),
			ProgramHeader.global_var_initvalues_start(1)-ProgramHeader.global_var_initvalues_start(0),
			((uint8*)memory)+global_begin_addrs_of_elementary_types[0]);

	makro_sp=akt_makro_size=0;
	read_byte_block(ProgramHeader.opcode_transformation_tables(0),
			((uint8)(ProgramHeader.opcode_transformation_tables(1)-ProgramHeader.opcode_transformation_tables(0)))>>1,
			(uint8*)(&table4bit_opcodes[0]));

	const unsigned int t_anz=anz_user_types;
	act_function_codebase=pc=0;
	makro_escape_sequences_temporarily_disabled=reading_makro_parameter=false;
	pc_marker=(PROGRAM_COUNTER_TYPE)0;
	sp=0;
#ifdef CALCULATE_MAXIMUM_STACK_USAGE
	minimum_stack_pointer=(VARIABLE_INDEX_TYPE)(0xffffffff);
#endif
#if defined(WITH_OPCODE_GETLOCAL) || defined(WITH_OPCODE_SETLOCAL)
	act_function_stackbase=0;
#endif

	Value v1, v2, v;
	CHOOSE_PC_TYPE(v1)=__pc;
	v1.type=VarIndexTypeNr;
	functioncall(v1);
	bool v1_contains_top_stack_element=false;

	while(1)
	{
		uint8 op;
		if(255==(op=get_next_opcode()))
		{
			if(v1_contains_top_stack_element)
				push_stack(&v1);
			if(false==pop_stack(&v)) // Dieser Fall kann eintreten, wenn vor Erreichen von pc_marker die aktuelle Funktion rekursiv
			{						// aufgerufen wird und dann der Fall pc==pc_marker erreicht wird, ohne dass pc_marker zuvor
				push_stack(&v);		// durch die rekursiv aufgerufene Funktion selbst gesetzt wurde.
				PROGRAM_COUNTER_TYPE save_pc_marker=pc_marker;
				pc_marker=0;
				op=get_next_opcode();	
				pc_marker=save_pc_marker;
			}
			else
			{
				pop_stack(&v, true);
				continue; // Es knnen mehrere Marken unmittelbar nacheinander erreicht werden !
			}
			v1_contains_top_stack_element=false;
		}
#ifdef PRINT_COMMANDS_AT_RUNTIME
		print_command(op, pc_vor_opcode, sp);
#endif

		if(op>=BEGIN_UNARY_OPERATORS)
		{
			if(!v1_contains_top_stack_element)
				pop_stack(&v1);
			v1_contains_top_stack_element=false;
#ifdef PRINT_COMMANDS_AT_RUNTIME
			if(op<BEGIN_BINARY_OPERATORS)
				bi_print_value(v1);
#endif
			if(op>=BEGIN_BINARY_OPERATORS)
			{
				pop_stack(&v2);
#ifdef PRINT_COMMANDS_AT_RUNTIME
				bi_print_value(v2);
				bi_print_value(v1);
#endif

				if(op<BEGIN_ARITHMETIC_OPERATORS)
				{
					switch(op)
					{
#ifdef WITH_OPCODE_STORET
						case STORET:
#endif
						case STORE: case STOREPOP:
						case STORED:
						{
							cast(&v1, VarIndexTypeNr);
							set_var(CHOOSE_VI_TYPE(v1), &v2, STORE==op || STOREPOP==op, STORED==op);
							if(STOREPOP!=op)
								{ v1=v2; v1_contains_top_stack_element=true; }
							break;
						}
#ifdef WITH_OPCODE_SETLOCALI
						case SETLOCALI:
						case SETLOCALIT:
						{
							cast(&v1, VarIndexTypeNr);
							if(CHOOSE_SIGNED_VI_TYPE(v1)>=0)
								CHOOSE_SIGNED_VI_TYPE(v1)+=2*MAX_NR_BYTES_FOR_VARIABLE_INDEX + MAX_NR_BYTES_FOR_PROGRAM_COUNTER;
							set_var(CHOOSE_SIGNED_VI_TYPE(v1)+act_function_stackbase, &v2, SETLOCALI==op);
							v1=v2; v1_contains_top_stack_element=true;
							break;
						}
#endif

					default:
						cast_up(&v1, &v2);
						switch(((uint16)op)*t_anz + v1.type)
						{
#ifdef WITH_OPCODE_LT
							case LT*t_anz+t_void:
#endif
#ifdef WITH_OPCODE_GT
							case GT*t_anz+t_void:
#endif
#ifdef WITH_OPCODE_NE
							case NE*t_anz+t_void:
#endif
#if defined(WITH_OPCODE_LT) || defined(WITH_OPCODE_GT) || defined(WITH_OPCODE_NE)
							v2.u8=false;
							break;
#endif

#ifdef WITH_OPCODE_LE
							case LE*t_anz+t_void:
#endif
#ifdef WITH_OPCODE_GE
							case GE*t_anz+t_void:
#endif
#ifdef WITH_OPCODE_EQ
							case EQ*t_anz+t_void:
#endif
#if defined(WITH_OPCODE_LE) || defined(WITH_OPCODE_GE) || defined(WITH_OPCODE_Eq)
							v2.u8=true;
							break;
#endif

#ifdef WITH_OPCODE_LT
#if defined(WITH_INTEGER8)
							case LT*t_anz+t_u8:		case LT*t_anz+t_bool:	v2.u8=(v2.u8<v1.u8); break;
							case LT*t_anz+t_i8:		v2.u8=(v2.i8<v1.i8); break;
#endif
#if defined(WITH_INTEGER16)
							case LT*t_anz+t_u16:	v2.u8=(v2.u16<v1.u16); break;
							case LT*t_anz+t_i16:	v2.u8=(v2.i16<v1.i16); break;
#endif
#if defined(WITH_INTEGER32)
							case LT*t_anz+t_u32:	v2.u8=(v2.u32<v1.u32); break;
							case LT*t_anz+t_i32:	v2.u8=(v2.i32<v1.i32); break;
#endif
#if defined(WITH_INTEGER64)
							case LT*t_anz+t_u64:	v2.u8=(v2.u64<v1.u64); break;
							case LT*t_anz+t_i64:	v2.u8=(v2.i64<v1.i64); break;
#endif
#if defined(WITH_REAL) && defined(real_less)
							case LT*t_anz+t_real:	v2.u8=real_less(v2, v1); break;
#endif
#endif

#ifdef WITH_OPCODE_LE
#if defined(WITH_INTEGER8)
							case LE*t_anz+t_u8:		case LE*t_anz+t_bool:	v2.u8=(v2.u8<=v1.u8); break;
							case LE*t_anz+t_i8:		v2.u8=(v2.i8<=v1.i8); break;
#endif
#if defined(WITH_INTEGER16)
							case LE*t_anz+t_u16:	v2.u8=(v2.u16<=v1.u16); break;
							case LE*t_anz+t_i16:	v2.u8=(v2.i16<=v1.i16); break;
#endif
#if defined(WITH_INTEGER32)
							case LE*t_anz+t_u32:	v2.u8=(v2.u32<=v1.u32); break;
							case LE*t_anz+t_i32:	v2.u8=(v2.i32<=v1.i32); break;
#endif
#if defined(WITH_INTEGER64)
							case LE*t_anz+t_u64:	v2.u8=(v2.u64<=v1.u64); break;
							case LE*t_anz+t_i64:	v2.u8=(v2.i64<=v1.i64); break;
#endif
#if defined(WITH_REAL) && defined(real_less_or_equal)
							case LE*t_anz+t_real:	v2.u8=real_less_or_equal(v2, v1); break;
#endif
#endif

#ifdef WITH_OPCODE_GT
#if defined(WITH_INTEGER8)
							case GT*t_anz+t_u8:		case GT*t_anz+t_bool:	v2.u8=(v2.u8>v1.u8); break;
							case GT*t_anz+t_i8:		v2.u8=(v2.i8>v1.i8); break;
#endif
#if defined(WITH_INTEGER16)
							case GT*t_anz+t_u16:	v2.u8=(v2.u16>v1.u16); break;
							case GT*t_anz+t_i16:	v2.u8=(v2.i16>v1.i16); break;
#endif
#if defined(WITH_INTEGER32)
							case GT*t_anz+t_u32:	v2.u8=(v2.u32>v1.u32); break;
							case GT*t_anz+t_i32:	v2.u8=(v2.i32>v1.i32); break;
#endif
#if defined(WITH_INTEGER64)
							case GT*t_anz+t_u64:	v2.u8=(v2.u64>v1.u64); break;
							case GT*t_anz+t_i64:	v2.u8=(v2.i64>v1.i64); break;
#endif
#if defined(WITH_REAL) && defined(real_greater)
							case GT*t_anz+t_real:	v2.u8=real_greater(v2, v1); break;
#endif
#endif

#ifdef WITH_OPCODE_GE
#if defined(WITH_INTEGER8)
							case GE*t_anz+t_u8:		case GE*t_anz+t_bool:	v2.u8=(v2.u8>=v1.u8); break;
							case GE*t_anz+t_i8:		v2.u8=(v2.i8>=v1.i8); break;
#endif
#if defined(WITH_INTEGER16)
							case GE*t_anz+t_u16:	v2.u8=(v2.u16>=v1.u16); break;
							case GE*t_anz+t_i16:	v2.u8=(v2.i16>=v1.i16); break;
#endif
#if defined(WITH_INTEGER32)
							case GE*t_anz+t_u32:	v2.u8=(v2.u32>=v1.u32); break;
							case GE*t_anz+t_i32:	v2.u8=(v2.i32>=v1.i32); break;
#endif
#if defined(WITH_INTEGER64)
							case GE*t_anz+t_u64:	v2.u8=(v2.u64>=v1.u64); break;
							case GE*t_anz+t_i64:	v2.u8=(v2.i64>=v1.i64); break;
#endif
#if defined(WITH_REAL) && defined(real_greater_or_equal)
							case GE*t_anz+t_real:	v2.u8=real_greater_or_equal(v2, v1); break;
#endif
#endif

#ifdef WITH_OPCODE_EQ
#if defined(WITH_INTEGER8)
							case EQ*t_anz+t_u8: case EQ*t_anz+t_bool: case EQ*t_anz+t_i8:	v2.u8=(v2.u8==v1.u8); break;
#endif
#if defined(WITH_INTEGER16)
							case EQ*t_anz+t_u16:	case EQ*t_anz+t_i16:	v2.u8=(v2.u16==v1.u16); break;
#endif
#if defined(WITH_INTEGER32)
							case EQ*t_anz+t_u32:	case EQ*t_anz+t_i32:	v2.u8=(v2.u32==v1.u32); break;						
#endif
#if defined(WITH_INTEGER64)
							case EQ*t_anz+t_u64:	case EQ*t_anz+t_i64:	v2.u8=(v2.u64==v1.u64); break;							
#endif
#if defined(WITH_REAL) && defined(real_equal)
							case EQ*t_anz+t_real:	v2.u8=real_equal(v2, v1); break;
#endif
#endif

#ifdef WITH_OPCODE_NE
#if defined(WITH_INTEGER8)
							case NE*t_anz+t_u8: case NE*t_anz+t_bool: case NE*t_anz+t_i8:	v2.u8=(v2.u8!=v1.u8); break;
#endif
#if defined(WITH_INTEGER16)
							case NE*t_anz+t_u16:	case NE*t_anz+t_i16:	v2.u8=(v2.u16!=v1.u16); break;
#endif
#if defined(WITH_INTEGER32)
							case NE*t_anz+t_u32:	case NE*t_anz+t_i32:	v2.u8=(v2.u32!=v1.u32); break;						
#endif
#if defined(WITH_INTEGER64)
							case NE*t_anz+t_u64:	case NE*t_anz+t_i64:	v2.u8=(v2.u64!=v1.u64); break;							
#endif
#if defined(WITH_REAL) && defined(real_not_equal)
							case NE*t_anz+t_real:	v2.u8=real_not_equal(v2, v1); break;
#endif
#endif
						}
						v2.type=t_bool;
						if(v2.u8) v2.u8=1; // Mglicherweise erzeugen ltere C-Compiler fr TRUE eine -1 statt einer 1.
						v1=v2; v1_contains_top_stack_element=true;
						break;
					}
				}
				else // if(op>=BEGIN_ARITHMETIC_OPERATORS)
				{
					if(op>=BEGIN_BINARY_OPERATORS_ON_MEMORY)
					{
						ASSERT(op<BEGINN_CONTROL_OPCODES_ONLY_USED_BY_COMPILER, "Unknown Opcode", pc);
						v=v2;
						cast(&v, VarIndexTypeNr);
						get_var(CHOOSE_VI_TYPE(v), &v2);
						if(SRM!=op && SLM!=op)
							cast(&v1, v2.type);
					}
					else if(SR!=op && SL!=op)
						cast_up(&v1, &v2);
					if((op>=BEGIN_ARITHMETIC_OPERATORS_SIGNEDNESS_SENSITIVE && op<BEGIN_BINARY_OPERATORS_ON_MEMORY)
							|| op>=BEGIN_BIN_OP_ON_MEM_SIGNEDNESS_SENSITIVE)
						switch(((uint16)op)*t_anz + v2.type) // darf nicht ...+v1.type sein !
					{
#if defined(WITH_INTEGER8)
						case MULM*t_anz+t_bool: case MUL*t_anz+t_bool: case MULM*t_anz+t_u8: case MUL*t_anz+t_u8:
																		v2.u8*=v1.u8; break;
						case MULM*t_anz+t_i8:	case MUL*t_anz+t_i8:	v2.i8*=v1.i8; break;
#endif
#if defined(WITH_INTEGER16)
						case MULM*t_anz+t_u16:	case MUL*t_anz+t_u16:	v2.u16*=v1.u16; break;
						case MULM*t_anz+t_i16:	case MUL*t_anz+t_i16:	v2.i16*=v1.i16; break;
#endif
#if defined(WITH_INTEGER32) && defined(WITH_INTEGER32_MULTIPLICATION)
						case MULM*t_anz+t_u32:	case MUL*t_anz+t_u32:	v2.u32*=v1.u32; break;
						case MULM*t_anz+t_i32:	case MUL*t_anz+t_i32:	v2.i32*=v1.i32; break;
#endif
#if defined(WITH_INTEGER64) && defined(WITH_INTEGER64_MULTIPLICATION)
						case MULM*t_anz+t_u64:	case MUL*t_anz+t_u64:	v2.u64=uint64_mult(v2.u64, v1.u64); break;
						case MULM*t_anz+t_i64:	case MUL*t_anz+t_i64:	v2.i64=int64_mult(v2.i64, v1.i64); break;
#endif
#if defined(WITH_REAL) && defined(WITH_REAL_MULTIPLICATION)
						case MULM*t_anz+t_real:	case MUL*t_anz+t_real:	{ real_mult(v2, v1); } break;
#endif

#if defined(WITH_INTEGER8)
						case DIVM*t_anz+t_u8:	case DIV*t_anz+t_u8:
							ASSERT(0!=v1.u8, "Division by zero", pc);
							v2.u8/=v1.u8; break;
						case DIVM*t_anz+t_i8:	case DIV*t_anz+t_i8:
							ASSERT(0!=v1.i8, "Division by zero", pc);
							v2.i8/=v1.i8; break;
#endif
#if defined(WITH_INTEGER16)
						case DIVM*t_anz+t_u16:	case DIV*t_anz+t_u16:
							ASSERT(0!=v1.u16, "Division by zero", pc);
							v2.u16/=v1.u16; break;
						case DIVM*t_anz+t_i16:	case DIV*t_anz+t_i16:
							ASSERT(0!=v1.i16, "Division by zero", pc);
							v2.i16/=v1.i16; break;
#endif
#if defined(WITH_INTEGER32) && defined(WITH_INTEGER32_DIVISION)
						case DIVM*t_anz+t_u32:	case DIV*t_anz+t_u32:
							ASSERT(0!=v1.u32, "Division by zero", pc);
							v2.u32/=v1.u32; break;
						case DIVM*t_anz+t_i32:	case DIV*t_anz+t_i32:
							ASSERT(0!=v1.i32, "Division by zero", pc);
							v2.i32/=v1.i32; break;
#endif
#if defined(WITH_INTEGER64) && defined(WITH_INTEGER64_DIVISION)
						case DIVM*t_anz+t_u64:	case DIV*t_anz+t_u64:
							ASSERT(0!=v1.u64, "Division by zero", pc);
							v2.u64=uint64_div(v2.u64, v1.u64); break;
						case DIVM*t_anz+t_i64:	case DIV*t_anz+t_i64:
							ASSERT(0!=v1.i64, "Division by zero", pc);
							v2.i64=int64_div(v2.i64, v1.i64); break;
#endif
#if defined(WITH_REAL) && defined(WITH_REAL_DIVISION)
						case DIVM*t_anz+t_real:	case DIV*t_anz+t_real:
							// ASSERT(0!=v1.r, "Division by zero", pc);
							{ real_div(v2, v1); } break;
#endif

#if defined(WITH_INTEGER8)
						case MODM*t_anz+t_u8:	case MOD*t_anz+t_u8:
							ASSERT(0!=v1.u8, "Division by zero", pc);
							v2.u8%=v1.u8; break;
						case MODM*t_anz+t_i8:	case MOD*t_anz+t_i8:
							ASSERT(0!=v1.i8, "Division by zero", pc);
							v2.i8%=v1.i8; break;
#endif
#if defined(WITH_INTEGER16)
						case MODM*t_anz+t_u16:	case MOD*t_anz+t_u16:
							ASSERT(0!=v1.u16, "Division by zero", pc);
							v2.u16%=v1.u16; break;
						case MODM*t_anz+t_i16:	case MOD*t_anz+t_i16:
							ASSERT(0!=v1.i16, "Division by zero", pc);
							v2.i16%=v1.i16; break;
#endif
#if defined(WITH_INTEGER32)
						case MODM*t_anz+t_u32:	case MOD*t_anz+t_u32:
							ASSERT(0!=v1.u32, "Division by zero", pc);
							v2.u32%=v1.u32; break;
						case MODM*t_anz+t_i32:	case MOD*t_anz+t_i32:
							ASSERT(0!=v1.i32, "Division by zero", pc);
							v2.i32%=v1.i32; break;
#endif
#if defined(WITH_INTEGER64) && defined(WITH_INTEGER64_MODULO)
						case MODM*t_anz+t_u64:	case MOD*t_anz+t_u64:
							ASSERT(0!=v1.u64, "Division by zero", pc);
							v2.u64=uint64_modulo(v2.u64, v1.u64); break;
						case MODM*t_anz+t_i64:	case MOD*t_anz+t_i64:
							ASSERT(0!=v1.i64, "Division by zero", pc);
							v2.i64=int64_modulo(v2.i64, v1.i64); break;
#endif
#if defined(WITH_REAL) && defined(real_modulo)
						case MODM*t_anz+t_real:	case MOD*t_anz+t_real:
							// ASSERT(0!=v1.r, "Division by zero", pc);
							{ real_modulo(v2, v1); } break;
#endif

#if defined(WITH_INTEGER8)
						case SRM*t_anz+t_u8:	case SR*t_anz+t_u8:		v2.u8>>=(uint8)v1.u8; break;
						case SRM*t_anz+t_i8:	case SR*t_anz+t_i8:		v2.i8>>=(uint8)v1.i8; break;
#endif
#if defined(WITH_INTEGER16)
						case SRM*t_anz+t_u16:	case SR*t_anz+t_u16:	v2.u16>>=(uint8)v1.u8; break;
						case SRM*t_anz+t_i16:	case SR*t_anz+t_i16:	v2.i16>>=(uint8)v1.i8; break;
#endif
#if defined(WITH_INTEGER32)
						case SRM*t_anz+t_u32:	case SR*t_anz+t_u32:	v2.u32>>=(uint8)v1.u8; break;
						case SRM*t_anz+t_i32:	case SR*t_anz+t_i32:	v2.i32>>=(uint8)v1.i8; break;
#endif
#if defined(WITH_INTEGER64)
						case SRM*t_anz+t_u64:	case SR*t_anz+t_u64:	v2.u64>>=(uint8)v1.u8; break;
						case SRM*t_anz+t_i64:	case SR*t_anz+t_i64:	v2.i64>>=(uint8)v1.i8; break;
#endif
#if defined(WITH_REAL) && defined(real_shift_right)
						case SRM*t_anz+t_real:	case SR*t_anz+t_real:	{ real_shift_right(v2, v1); } break;
#endif
					}
#if defined(ASSUME_VALUE_UNION_EQUAL_ADDRESSES) && defined(BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE)
					else
					{
						switch(op)
						{
							case ADDM: case ADDMPOP: case ADD:
#if defined(WITH_REAL) && defined(WITH_REAL_ADDITION)
								if(Value::t_real==v2.type)
									{ real_add(v2, v1); }
								else
#endif
									CHOOSE_LARGEST_UINTEGER(v2)+=CHOOSE_LARGEST_UINTEGER(v1);
								break;
							case SUBM: case SUBMPOP: case SUB:
#if defined(WITH_REAL) && defined(WITH_REAL_SUBTRACTION)
								if(Value::t_real==v2.type)
									{ real_sub(v2, v1); }
								else
#endif
									CHOOSE_LARGEST_UINTEGER(v2)-=CHOOSE_LARGEST_UINTEGER(v1);
								break;
							case ANDM: case AND:
								CHOOSE_LARGEST_UINTEGER(v2)&=CHOOSE_LARGEST_UINTEGER(v1); break;
							case ORM: case OR:
								CHOOSE_LARGEST_UINTEGER(v2)|=CHOOSE_LARGEST_UINTEGER(v1); break;
							case XORM: case XOR:
								CHOOSE_LARGEST_UINTEGER(v2)^=CHOOSE_LARGEST_UINTEGER(v1); break;
							case SLM: case SL:
								CHOOSE_LARGEST_UINTEGER(v2)<<=(uint8)v1.u8; break;
						}
						if(t_bool==v2.type)
							v2.u8=(0!=v2.u8 ? 1 : 0);
					}
#else
					else switch(((uint16)op)*fkt + (v2.type>>1)) // darf nicht ...+v1.type sein !
					{
						// ----- Addition -----
#if defined(WITH_INTEGER8)
					case ADDM*fkt+(t_bool>>1):	case ADDMPOP*fkt+(t_bool>>1):	case ADD*fkt+(t_bool>>1):	v2.u8=v2.u8 || v1.u8; break;
					case ADDM*fkt+(t_u8>>1):	case ADDMPOP*fkt+(t_u8>>1):		case ADD*fkt+(t_u8>>1):		v2.u8+=v1.u8; break;
#endif
#if defined(WITH_INTEGER16)
						case ADDM*fkt+(t_u16>>1):	case ADDMPOP*fkt+(t_u16>>1):	case ADD*fkt+(t_u16>>1):	v2.u16+=v1.u16; break;
#endif
#if defined(WITH_INTEGER32)
						case ADDM*fkt+(t_u32>>1):	case ADDMPOP*fkt+(t_u32>>1):	case ADD*fkt+(t_u32>>1):	v2.u32+=v1.u32; break;
#endif
#if defined(WITH_INTEGER64)
						case ADDM*fkt+(t_u64>>1):	case ADDMPOP*fkt+(t_u64>>1):	case ADD*fkt+(t_u64>>1):	v2.u64+=v1.u64; break;
#endif
#if defined(WITH_REAL)
						case ADDM*fkt+(t_real>>1):	case ADDMPOP*fkt+(t_real>>1):	case ADD*fkt+(t_real>>1):	{ real_add(v2, v1); } break;
#endif

						// ----- Subtraktion -----
#if defined(WITH_INTEGER8)
						case SUBM*fkt+(t_bool>>1):	case SUBMPOP*fkt+(t_bool>>1):	case SUB*fkt+(t_bool>>1):	if(v1.u8) v2.u8=0; break;
						case SUBM*fkt+(t_u8>>1):	case SUBMPOP*fkt+(t_u8>>1):		case SUB*fkt+(t_u8>>1):		v2.u8-=v1.u8; break;
#endif
#if defined(WITH_INTEGER16)
						case SUBM*fkt+(t_u16>>1):	case SUBMPOP*fkt+(t_u16>>1):	case SUB*fkt+(t_u16>>1):	v2.u16-=v1.u16; break;
#endif
#if defined(WITH_INTEGER32)
						case SUBM*fkt+(t_u32>>1):	case SUBMPOP*fkt+(t_u32>>1):	case SUB*fkt+(t_u32>>1):	v2.u32-=v1.u32; break;
#endif
#if defined(WITH_INTEGER64)
						case SUBM*fkt+(t_u64>>1):	case SUBMPOP*fkt+(t_u64>>1):	case SUB*fkt+(t_u64>>1):	v2.u64-=v1.u64; break;
#endif
#if defined(WITH_REAL)
						case SUBM*fkt+(t_real>>1):	case SUBMPOP*fkt+(t_real>>1):	case SUB*fkt+(t_real>>1):	{ real_sub(v2, v1); } break;
#endif

#if defined(WITH_INTEGER8)
						case ANDM*fkt+(t_bool>>1): case AND*fkt+(t_bool>>1): case ANDM*fkt+(t_u8>>1): case AND*fkt+(t_u8>>1):
							v2.u8&=v1.u8; break;
#endif
#if defined(WITH_INTEGER16)
						case ANDM*fkt+(t_u16>>1):	case AND*fkt+(t_u16>>1):	v2.u16&=v1.u16; break;
#endif
#if defined(WITH_INTEGER32)
						case ANDM*fkt+(t_u32>>1):	case AND*fkt+(t_u32>>1):	v2.u32&=v1.u32; break;
#endif
#if defined(WITH_INTEGER64)
						case ANDM*fkt+(t_u64>>1):	case AND*fkt+(t_u64>>1):	v2.u64&=v1.u64; break;
#endif
#if defined(WITH_REAL) && defined(real_and)
						case ANDM*fkt+(t_real>>1):	case AND*fkt+(t_real>>1):	{ real_and(v2, v1); } break;
#endif

#if defined(WITH_INTEGER8)
						case ORM*fkt+(t_bool>>1): case OR*fkt+(t_bool>>1): case ORM*fkt+(t_u8>>1): case OR*fkt+(t_u8>>1):
							v2.u8|=v1.u8; break;
#endif
#if defined(WITH_INTEGER16)
						case ORM*fkt+(t_u16>>1):	case OR*fkt+(t_u16>>1):		v2.u16|=v1.u16; break;
#endif
#if defined(WITH_INTEGER32)
						case ORM*fkt+(t_u32>>1):	case OR*fkt+(t_u32>>1):		v2.u32|=v1.u32; break;
#endif
#if defined(WITH_INTEGER64)
						case ORM*fkt+(t_u64>>1):	case OR*fkt+(t_u64>>1):		v2.u64|=v1.u64; break;
#endif
#if defined(WITH_REAL) && defined(real_or)
						case ORM*fkt+(t_real>>1):	case OR*fkt+(t_real>>1):	{ real_or(v2, v1); } break;
#endif

#if defined(WITH_INTEGER8)
						case XORM*fkt+(t_bool>>1): case XOR*fkt+(t_bool>>1): case XORM*fkt+(t_u8>>1): case XOR*fkt+(t_u8>>1):
							v2.u8^=v1.u8; break;
#endif
#if defined(WITH_INTEGER16)
						case XORM*fkt+(t_u16>>1):	case XOR*fkt+(t_u16>>1):	v2.u16^=v1.u16; break;
#endif
#if defined(WITH_INTEGER32)
						case XORM*fkt+(t_u32>>1):	case XOR*fkt+(t_u32>>1):	v2.u32^=v1.u32; break;
#endif
#if defined(WITH_INTEGER64)
						case XORM*fkt+(t_u64>>1):	case XOR*fkt+(t_u64>>1):	v2.u64^=v1.u64; break;
#endif
#if defined(WITH_REAL) && defined(real_xor)
						case XORM*fkt+(t_real>>1):	case XOR*fkt+(t_real>>1):	{ real_xor(v2, v1); } break;
#endif

#if defined(WITH_INTEGER8)
						case SLM*fkt+(t_u8>>1):		case SL*fkt+(t_u8>>1):		v2.u8<<=(uint8)v1.u8; break;
#endif
#if defined(WITH_INTEGER16)
						case SLM*fkt+(t_u16>>1):	case SL*fkt+(t_u16>>1):		v2.u16<<=(uint8)v1.u8; break;
#endif
#if defined(WITH_INTEGER32)
						case SLM*fkt+(t_u32>>1):	case SL*fkt+(t_u32>>1):		v2.u32<<=(uint8)v1.u8; break;
#endif
#if defined(WITH_INTEGER64)
						case SLM*fkt+(t_u64>>1):	case SL*fkt+(t_u64>>1):		v2.u64<<=(uint8)v1.u8; break;
#endif
#if defined(WITH_REAL) && defined(real_shift_left)
						case SLM*fkt+(t_real>>1):	case SL*fkt+(t_real>>1):	{ real_shift_left(v2, v1); } break;
#endif
					}
#endif
					if(op>=BEGIN_BINARY_OPERATORS_ON_MEMORY)
						set_var(CHOOSE_VI_TYPE(v), &v2);
					if(ADDMPOP!=op && SUBMPOP!=op)
						{ v1=v2; v1_contains_top_stack_element=true; }
				}
			}
			else // op>=BEGIN_UNARY_OPERATORS && op<BEGIN_BINARY_OPERATORS
			{
				if(op>=BEGIN_UNARY_OPERATORS_WITH_BOOLEAN_OPERAND)
					cast_to_bool(&v1);
				switch(op)
				{
#ifdef WITH_OPCODE_POPN
					case POPN:
						// Es darf nicht einfach der Stackpointer auf den neuen Wert gesetzt werden,
						// denn falls z.B. ein PC-Marker auf dem Stack gespeichert ist, muss dieser nicht nur vom
						// Stack gelscht werden, sondern auch die Variable pc_marker gesetzt werden.
						cast(&v1, VarIndexTypeNr);
						while(CHOOSE_VI_TYPE(v1)--)
							if(pop_stack(&v))
							{
								ASSERT(0!=CHOOSE_VI_TYPE(v1), "POPN has POPed pc_marker but not pc", pc);
								CHOOSE_VI_TYPE(v1)--;
								pop_stack(&v);
							}
						break;
#endif
#ifdef WITH_OPCODE_PUSHN
					case PUSHN:
						v.type=get_next_code_nibble();
						if(((signed char*)(&v.u8))-((signed char*)(&v.u64))==0 && ((signed char*)(&v.u16))-((signed char*)(&v.u64))==0 &&
								((signed char*)(&v.u32))-((signed char*)(&v.u64))==0)
							v.u64=0;
						else if(t_u8==v.type || t_i8==v.type || t_bool==v.type || t_void==v.type)
							v.u8=0;
						else if(t_u16==v.type || t_i16==v.type)
							v.u16=0;
						else if(t_u32==v.type || t_i32==v.type)
							v.u32=0;
						else
							v.u64=0;
						cast(&v1, VarIndexTypeNr);
						while(CHOOSE_VI_TYPE(v1)--)
							push_stack(&v);
						break;
#endif
					case CAST:
						cast(&v1, get_next_code_nibble());
						v1_contains_top_stack_element=true;
						break;
#ifdef WITH_OPCODE_NEG
					case NEG:
	#if defined(ASSUME_VALUE_UNION_EQUAL_ADDRESSES) && defined(BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE)
						if(t_bool!=v1.type)
							CHOOSE_LARGEST_UINTEGER(v1)=LARGEST_UINTEGER_TYPE(-LARGEST_SINTEGER_TYPE(CHOOSE_LARGEST_UINTEGER(v1)));
	#else
						switch(v1.type>>1)
						{
		#if defined(WITH_INTEGER8)
							case t_bool>>1: v1.u8=(0!=v1.u8 ? 0 : 1); break;
							case t_u8>>1: v1.u8=-v1.u8; break;
		#endif
		#if defined(WITH_INTEGER16)
							case t_u16>>1: v1.u16=-v1.u16; break;
		#endif
		#if defined(WITH_INTEGER32)
							case t_u32>>1: v1.u32=-((int32)v1.u32); break;
		#endif
		#if defined(WITH_INTEGER64)
							case t_u64>>1: v1.u64=-((int64)v1.u64); break;
		#endif
		#if defined(WITH_REAL) && defined(real_negate)
							case t_real>>1: { real_negate(v1); } break;
		#endif
						}
	#endif
						v1_contains_top_stack_element=true;
						break;
#endif
#ifdef WITH_OPCODE_NOT
					case NOT:
						v1.u8=(0!=v1.u8 ? 0 : 1);
						v1.type=t_bool;
						v1_contains_top_stack_element=true;
						break;
#endif
#ifdef WITH_OPCODE_COM
					case COM:
	#if defined(WITH_REAL)
						if(v1.type>=t_real)
							cast(&v1, LargestAvailableUIntegerTypeNr);
	#endif
	#if defined(ASSUME_VALUE_UNION_EQUAL_ADDRESSES) && defined(BYTECODE_INTERPRETER_RUNS_ON_LITTLE_ENDIAN_MACHINE)
						if(t_bool==v1.type)
							v1.u8=(0!=v1.u8 ? 0 : 1);
						else
							CHOOSE_LARGEST_UINTEGER(v1)=~CHOOSE_LARGEST_UINTEGER(v1);
	#else
						switch(v1.type>>1)
						{
		#if defined(WITH_INTEGER8)
							case t_bool>>1: v1.u8=(0!=v1.u8 ? 0 : 1); break; // Falls der Typ void vorliegt, schadet die Zuweisung an v1.u8 nicht
							case t_u8>>1: v1.u8=~v1.u8; break;
		#endif
		#if defined(WITH_INTEGER16)
							case t_u16>>1: v1.u16=~v1.u16; break;
		#endif
		#if defined(WITH_INTEGER32)
							case t_u32>>1: v1.u32=~v1.u32; break;
		#endif
		#if defined(WITH_INTEGER64)
							case t_u64>>1: v1.u64=~v1.u64; break;
		#endif
						}
	#endif
						v1_contains_top_stack_element=true;
						break;
#endif
					case GETPREVSTORE:
						v1_contains_top_stack_element=true; // kein break, fall through
					case GETPREVSTOREPOP:
					{
						CHOOSE_VI_TYPE(v2)=read_coded_uintegerVI(15);
						get_var(find_previous_stack_element(CHOOSE_VI_TYPE(v2)), &v2);
						goto GETLOCALSTOREPOP_label;
					}
					case GETLOCALSTORE:
						v1_contains_top_stack_element=true; // kein break, fall through
					case GETLOCALSTOREPOP:
					{
						CHOOSE_VI_TYPE(v2)=read_coded_sintegerVI(255);
						if(CHOOSE_SIGNED_VI_TYPE(v2)>=0)
							CHOOSE_SIGNED_VI_TYPE(v2)+=2*MAX_NR_BYTES_FOR_VARIABLE_INDEX + MAX_NR_BYTES_FOR_PROGRAM_COUNTER;
						get_var(CHOOSE_SIGNED_VI_TYPE(v2)+act_function_stackbase, &v2);
						GETLOCALSTOREPOP_label:
						cast(&v2, VarIndexTypeNr);
						set_var(CHOOSE_VI_TYPE(v2), &v1, true, false);
						break;
					}

#ifdef WITH_OPCODE_SETPREVT
					case SETPREVT:
#endif
#ifdef WITH_OPCODE_SETPREV
					case SETPREV: case SETPREVPOP:
					{
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
						SETPREV_label:
#endif
						CHOOSE_VI_TYPE(v2)=read_coded_uintegerVI(15);
#ifdef WITH_OPCODE_SETPREVT
						set_var(find_previous_stack_element(CHOOSE_VI_TYPE(v2)), &v1, SETPREVT!=op);  // Es kann auch
							// op==SETPREVC sein (goto SETPREV_label).
#else
						set_var(find_previous_stack_element(CHOOSE_VI_TYPE(v2)), &v1, true);
#endif
						if(SETPREVPOP!=op)
							v1_contains_top_stack_element=true;
						break;
					}
#endif

#ifdef WITH_OPCODE_SETLOCALT
					case SETLOCALT:
#endif
#ifdef WITH_OPCODE_SETLOCAL
					case SETLOCAL: case SETLOCALPOP:
					{
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
						SETLOCAL_label:
#endif
						CHOOSE_SIGNED_VI_TYPE(v2)=read_coded_sintegerVI(255);
						if(CHOOSE_SIGNED_VI_TYPE(v2)>=0)
							CHOOSE_SIGNED_VI_TYPE(v2)+=2*MAX_NR_BYTES_FOR_VARIABLE_INDEX + MAX_NR_BYTES_FOR_PROGRAM_COUNTER;
#ifdef WITH_OPCODE_SETLOCALT
						set_var(CHOOSE_SIGNED_VI_TYPE(v2)+act_function_stackbase, &v1, SETLOCALT!=op); // Es kann auch
							// op==SETLOCALC sein (goto SETLOCAL_label).
#else
						set_var(CHOOSE_SIGNED_VI_TYPE(v2)+act_function_stackbase, &v1, true);
#endif
						if(SETLOCALPOP!=op)
							v1_contains_top_stack_element=true;
						break;
					}
#endif

					case LOAD:
					case LOADT:
					{
						cast(&v1, VarIndexTypeNr);
						if(LOADT==op)
							v1.type=get_next_code_nibble();
						get_var(CHOOSE_VI_TYPE(v1), &v1, LOADT==op);
						v1_contains_top_stack_element=true;
						break;
					}
#ifdef WITH_OPCODE_GETLOCALI_AND_LOCALADDRI
					case LOCALADDRI:
					case GETLOCALI:
					{
						cast(&v1, VarIndexTypeNr);
						if(CHOOSE_SIGNED_VI_TYPE(v1)>=0)
							CHOOSE_SIGNED_VI_TYPE(v1)+=2*MAX_NR_BYTES_FOR_VARIABLE_INDEX + MAX_NR_BYTES_FOR_PROGRAM_COUNTER;
						CHOOSE_VI_TYPE(v1)+=act_function_stackbase;
						if(LOCALADDRI!=op)
							get_var(CHOOSE_VI_TYPE(v1), &v1);
						v1_contains_top_stack_element=true;
						break;
					}
					case STACKADDRI:
					{
						cast(&v1, VarIndexTypeNr);
						CHOOSE_VI_TYPE(v1)=find_previous_stack_element(CHOOSE_VI_TYPE(v1));
						v1_contains_top_stack_element=true;
						break;
					}
#endif

					case POP:
						break;
					
					case CALLEXT:
						// Achtung: Falls die Funktionsnummer direkt nach dem Opcode erscheinen soll, muss v1 im Fall
						// v1_contains_top_stack_element==true zuvor auf den Stack gebracht werden !
						ASSERT(0!=extern_function_invoker, "No extern_function_invoker defined", pc);
						cast(&v1, ProgramCounterTypeNr);
						extern_function_invoker(CHOOSE_PC_TYPE(v1), *this);
						break;

#ifdef WITH_OPCODE_JMPI
					case JMPI:
						pc=CHOOSE_PC_TYPE(v1);
						break;
#endif

#ifdef WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION
					case IFELSE:
					case BRF:
						v1.u8=!v1.u8; // kein break, fall through.
					case IFNOTELSE:
					case BRT:
					{
						PROGRAM_COUNTER_TYPE d;
						SIGNED_PROGRAM_COUNTER_TYPE d2;
						d=read_coded_integer(255, 2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER, BRF==op || BRT==op);
						PROGRAM_COUNTER_TYPE pc_nach_d=pc, save_pc_vor_opcode=pc_vor_opcode;
						pc=pc_vor_opcode+(PROGRAM_COUNTER_TYPE)d;
						makro_sp=0;
						akt_makro_size=0;
						d2=pc_marker;
						if(get_next_opcode()>=BEGIN_UNARY_OPERATORS)
						{
							pc_marker=d2;
							if(0==v1.u8)
								pc=pc_nach_d;
							else
							{
								if(BRF==op || IFNOTELSE==op)
									v1.u8=!v1.u8;
								v1_contains_top_stack_element=true;
								pc=save_pc_vor_opcode+(PROGRAM_COUNTER_TYPE)d;
							}
						}
						else if(pc_marker=d2, pc=pc_nach_d, ((IFELSE==op || IFNOTELSE==op) &&
									(d2=read_coded_integer(255, 2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER, 1), // d2 als signed lesen
									 0==v1.u8)))
						{		// VOR dem Aufruf read_coded_integer() muss der pc wiederhergestellt werden !
							CHOOSE_PC_TYPE(v)=save_pc_vor_opcode + (SIGNED_PROGRAM_COUNTER_TYPE)d2;
							if(((SIGNED_PROGRAM_COUNTER_TYPE)d2)>=0)
								CHOOSE_PC_TYPE(v) += (PROGRAM_COUNTER_TYPE)d;
								// Wenn d2>=0 ist, wird beim Erreichen der unten definierten PC-Marke (nach Ausfhrung
								// folgender Befehle des IF-Teils) an die PC-Adresse gesprungen, die sich durch Addition
								// von d+d2 zur PC-Adresse ergibt, die der auf den IFELSE bzw. IFNOTELSE-Befehl folgende
								// Befehl hat. Wenn d2<0 ist, wird statt dessen d2 zu dieser PC-Adresse addiert.
							v.type=ProgramCounterTypeNr;
							push_stack(&v);

							CHOOSE_PC_TYPE(v)=pc_marker;
							v.type=t_pc_marker_address;
							push_stack(&v);

							pc_marker = save_pc_vor_opcode+(PROGRAM_COUNTER_TYPE)d;
						}
						else if(0!=v1.u8)
							pc=save_pc_vor_opcode+(PROGRAM_COUNTER_TYPE)d;
						break;
					}

#else
					case BRF:
					case BRAND:
					case BRORNOT:
						v1.u8=!v1.u8; // kein break, fall through
					case BRT:
					case BROR:
					case BRANDNOT:
					{
						PROGRAM_COUNTER_TYPE d;
						d=read_coded_integer(255, 2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER, BRF==op || BRT==op);
						if(v1.u8)
						{
							pc=pc_vor_opcode+(SIGNED_PROGRAM_COUNTER_TYPE)d;
							if(BRF!=op && BRT!=op)
							{
								if(BRANDNOT==op || BRAND==op)
									v1.u8=!v1.u8;
								v1_contains_top_stack_element=true;
							}
						}
						break;
					}

					case IFNOTELSE:
						v1.u8=!v1.u8; // kein break, fall through.
					case IFELSE:
					{
						PROGRAM_COUNTER_TYPE d;
						SIGNED_PROGRAM_COUNTER_TYPE d2;
						d=read_coded_integer(255, 2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER, 0);   // d als unsigned lesen
						d2=read_coded_integer(255, 2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER, 1);  // d2 als signed lesen
						if(v1.u8)
						{
							CHOOSE_PC_TYPE(v)=pc_vor_opcode + (SIGNED_PROGRAM_COUNTER_TYPE)d2;
							if(((SIGNED_PROGRAM_COUNTER_TYPE)d2)>=0)
								CHOOSE_PC_TYPE(v) += (PROGRAM_COUNTER_TYPE)d;
							v.type=ProgramCounterTypeNr;
							push_stack(&v);

							CHOOSE_PC_TYPE(v)=pc_marker;
							v.type=t_pc_marker_address;
							push_stack(&v);

							pc_marker = pc_vor_opcode+(PROGRAM_COUNTER_TYPE)d;
						}
						else
							pc=pc_vor_opcode+(PROGRAM_COUNTER_TYPE)d;
						break;
					}
#endif

#ifdef WITH_OPCODE_GETTYPE
					case GETTYPE:
					{
						v1.u8=v1.type;
						v1.type=t_u8;
						v1_contains_top_stack_element=true;
						break;
					}
#endif
				}
			}
		}
		else  // if(op<BEGIN_UNARY_OPERATORS)
		{
			if(v1_contains_top_stack_element)
				push_stack(&v1);
			v1_contains_top_stack_element=false;
			switch(op)
			{
#ifdef WITH_OPCODE_CALLI
				case CALLI:
					pop_stack(&v1); // kein break, fall through
#endif
				case CALL:
				{
#ifdef WITH_OPCODE_CALLI
					if(CALL==op)
#endif
					{
						v1.u32=(get_next_code_nibble()<<4); // nicht in einer Anweisung, weil Reihenfolge der Auswertung
						v1.u32 |= get_next_code_nibble(); // nicht festgelegt.
						if(((uint8)(v1.u32))>=128)
						{
							uint8 size;
							for((size=(MAX_NR_BYTES_FOR_PROGRAM_COUNTER<<1)-2); size--; )
								v1.u32=(v1.u32<<4) | get_next_code_nibble();
							v1.u32=(PROGRAM_COUNTER_TYPE)(CHOOSE_PC_TYPE(v1)<<1);
						}
						else
						{
							read_byte_block(ProgramHeader.function_address_table()+
											2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER*(PROGRAM_COUNTER_TYPE)((uint8)v1.u32),
								MAX_NR_BYTES_FOR_PROGRAM_COUNTER, (uint8*)(&CHOOSE_PC_TYPE(v1)));
							// Die oberen Bytes von v1.u32 wurden in einer vorigen Anweisung gelscht.
						}
						v1.type=t_u32;
					}
					functioncall(v1);
					break;
				}
				case RET:
					uint8 paraminfo;
					paraminfo=get_code_nibble(act_function_codebase-2)<<4;	// Die beiden get_code_nibble()-Aufrufe drfen nicht in einem Ausdruck erscheinen,
					paraminfo|=get_code_nibble(act_function_codebase-1);	// weil der C++-Standard keine Reihenfolge fr die Auswertung von x | y vorschreibt.
					if(0!=(paraminfo&1))
						pop_stack(&v1); // RETURN-Wert
					/*while(pop_stack(&v), v.type!=t_byte_block_length)
						;*/
					while(sp!=act_function_stackbase)
						pop_stack(&v);
#if defined(WITH_OPCODE_GETLOCAL) || defined(WITH_OPCODE_SETLOCAL)
					pop_stack(&v, false, true);
					cast(&v, VarIndexTypeNr);
					act_function_stackbase=CHOOSE_VI_TYPE(v);
#endif
					pop_stack(&v, false, true);
					cast(&v, ProgramCounterTypeNr);
					act_function_codebase=CHOOSE_VI_TYPE(v);
					pop_stack(&v2, false, true); // an pc darf erst nach der eventuellen Ausfhrung von read_coded_integer()
						// zugewiesen werden.
					while(0!=(paraminfo>>1))
					{
						pop_stack(&v, false, true);
						paraminfo-=2;
					}
					cast(&v2, ProgramCounterTypeNr);
					pc=CHOOSE_PC_TYPE(v2);
					if(0!=(paraminfo&1))
						v1_contains_top_stack_element=true;
					if(0==pc)
					{
#ifdef PRINT_COMMANDS_AT_RUNTIME
						PRINT_NEWLINE_IF_PRINT_COMMANDS_AT_RUNTIME
						PRINT_NEWLINE_IF_PRINT_COMMANDS_AT_RUNTIME
#endif
						return;
					}
					break;
				case SETTYPES:
					VARIABLE_INDEX_TYPE d;
					VARIABLE_INDEX_TYPE addr;
					d=read_coded_integer(255, 2*MAX_NR_BYTES_FOR_VARIABLE_INDEX, 0); // d als unsigned lesen
					addr=act_function_stackbase-d;
					v.u64=0;
					while(d=read_coded_integer(15, 2*MAX_NR_BYTES_FOR_VARIABLE_INDEX, 0), 0!=d) // d als unsigned lesen
					{
						if(d<10)
						{
							while(d--)
							{
								addr-=Value::sbytesize( v.type=get_next_code_nibble() );
								set_var(addr, &v);
							}
						}
						else
						{
							v.type=get_next_code_nibble();
							for(d-=5; --d; ) // ACHTUNG: d=10 erzeugt 4 Variablen vom durch get_next_code_nibble() bestimmten Typ.
							{
								addr-=Value::sbytesize(v.type);
								set_var(addr, &v);
							}
						}
					}
					break;

				case BR:
				{
					SIGNED_PROGRAM_COUNTER_TYPE d;
					d=read_coded_integer(255, 2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER, 1); // d als signed lesen
					pc=pc_vor_opcode+(SIGNED_PROGRAM_COUNTER_TYPE)d;
					break;
				}

				case PUSHC0:
					v1.u64=0; // !!!! evtl. noch Abfrage einbauen, ob &v1.u8==(uint8*)&v1.u64 ist.
					v1.type=get_next_code_nibble();
					v1_contains_top_stack_element=true;
					break;
				case PUSHC0U8:
					v1.u8=0;
					v1.type=t_u8;
					v1_contains_top_stack_element=true;
					break;

				{
					uint8 type, nibblesize;
				case PUSHCS:
						v1.type=get_next_code_nibble();
						if(v1.type>=t_i32_but16used)
						{
							nibblesize=(v1.type>=t_i64_but32used) ? 8 : 4;
							type=v1.type-(t_i32_but16used-Value::t_i32);
							v1.type=type-2; // t_u32 ==> t_u16, t_u64 ==> t_u32, t_i32 ==> t_i16, t_i64 ==> t_i32, 
							// v1.type muss immer korrekt gesetzt werden wegen des folgenden cast()-Aufrufs
							goto PUSHC_label;
						}
						goto PUSHCSU8_label;
				case PUSHCSU8:
						v1.type=t_u8;
						PUSHCSU8_label:
						v1.u64=get_next_code_nibble(); // !!!! evtl. noch Abfrage einbauen, ob &v1.u8==(uint8*)&v1.u64 ist.
						v1_contains_top_stack_element=true;
						break;
#ifdef USE_TABLE_FOR_CONSTANTS
				case PUSHCT:
					v1.u8=(get_next_code_nibble()<<4); // nicht in einer Anweisung, weil Reihenfolge der Auswertung
					v1.u8 |= get_next_code_nibble(); // nicht festgelegt.
					PROGRAM_COUNTER_TYPE cpc;
					if(v1.u8>=ProgramHeader.limit_for_2bytes_constants())
					{
						if(v1.u8>=ProgramHeader.limit_for_4bytes_constants())
						{
							cpc=ProgramHeader.constants8bytes_table()+(((uint16)v1.u8)<<4)+v1.u8; // -(ProgramHeader.limit_for_4bytes_constants*17)
							v1.type=Value::t_i64;
						}
						else
						{
							cpc=ProgramHeader.constants4bytes_table()+(((uint16)v1.u8)<<3)+v1.u8; // -(ProgramHeader.limit_for_2bytes_constants*9)
							v1.type=Value::t_i32;
						}
					}
					else
					{
						if(v1.u8>=ProgramHeader.limit_for_1bytes_constants())
						{
							cpc=ProgramHeader.constants2bytes_table()+(((uint16)v1.u8)<<2)+v1.u8; // -(ProgramHeader.limit_for_1bytes_constants*5)
							v1.type=Value::t_i16;
						}
						else
						{
							cpc=ProgramHeader.constants1bytes_table()+(((uint16)v1.u8)<<1)+v1.u8; // -0
							v1.type=Value::t_i8;
						}
					}
					if(Value::s_is_unsigned_integer(type=get_code_nibble(cpc++)))
						++v1.type; // signed ==> unsigned
					v1.u64=0; // wichtig, falls weniger als 64 Bits gelesen werden.
					read_byte_block(cpc, Value::sbytesize(v1.type), (uint8*)(&v1.u64));
					goto PUSHC_label2;
#endif
				case PUSHC:
					{
						v1.type=type=get_next_code_nibble(); // v1.type muss immer korrekt gesetzt werden wegen des folgenden cast()-Aufrufs
						nibblesize=2;
						if(0==type) // t_i16_but8used
							{ type=t_i16; v1.type=t_i8; } // v1.type muss immer korrekt gesetzt werden wegen des folgenden cast()-Aufrufs
						else if(type>=t_u16_but8used)
						{
							type-=(t_u16_but8used-Value::t_u16);
							v1.type=t_i8 + (type&1); // v1.type muss immer korrekt gesetzt werden wegen des folgenden cast()-Aufrufs
						}
						else
							nibblesize=(Value::sbytesize(type)<<1);
						PUSHC_label:
						for(v1.u64=0; nibblesize--; )
							v1.u64=(v1.u64<<4) | get_next_code_nibble();
#ifdef USE_TABLE_FOR_CONSTANTS
						PUSHC_label2:
#endif
#if defined(WITH_REAL)
						if(t_real!=type)
#endif
							cast(&v1, type); // z.B. Auffllen mit 1-Bits beim up-cast eines negativen signed-Werts.
#ifdef PRINT_COMMANDS_AT_RUNTIME
						bi_print_value(v1);
#endif
						v1_contains_top_stack_element=true;
						break;
					}
				}

#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
				case SETLOCALCPOP:
					op=SETLOCALPOP; // kein break, fall through
				case SETLOCALC:
					v1.i64=read_coded_integer(15, 0, 1); // als signed lesen
					v1.type=t_i64;
					goto SETLOCAL_label;
				case SETPREVCPOP:
					op=SETPREVPOP; // kein break, fall through
				case SETPREVC:
					v1.i64=read_coded_integer(15, 0, 1); // als signed lesen
					v1.type=t_i64;
					goto SETPREV_label;
#endif

				{
					VARIABLE_INDEX_TYPE addr;
				case ADDCPREV: case ADDCPREVPOP:
				case PREVADDC:
						v2.u8=get_next_code_nibble();
						addr=find_previous_stack_element(read_coded_uintegerVI(15));
						goto ADDCPREV_label1;
				case ADDCLOCAL: case ADDCLOCALPOP:
				case LOCALADDC:
						v2.u8=get_next_code_nibble();
						CHOOSE_SIGNED_VI_TYPE(v1)=read_coded_sintegerVI(255);
						addr=act_function_stackbase+CHOOSE_SIGNED_VI_TYPE(v1);
						if(CHOOSE_SIGNED_VI_TYPE(v1)>=0)
							addr+=2*MAX_NR_BYTES_FOR_VARIABLE_INDEX + MAX_NR_BYTES_FOR_PROGRAM_COUNTER;
					ADDCPREV_label1:
						get_var(addr, &v1);
						v2.type=v1.type;
						if(v2.u8<8)
							v2.u64=v1.u64+v2.u8+1; // Ggf. noch das define ASSUME_VALUE_UNION_EQUAL_ADDRESSES abfragen !
						else
							v2.u64=v1.u64-(v2.u8-7);
						set_var(addr, &v2);
						if(ADDCPREV==op || ADDCLOCAL==op)
							v1=v2;
						v1_contains_top_stack_element=(op!=ADDCPREVPOP && op!=ADDCLOCALPOP);
						break;
				}

				case BRK:
				case CONT:
				{
					VARIABLE_INDEX_TYPE d;
					d=read_coded_integer(15, 2*MAX_NR_BYTES_FOR_VARIABLE_INDEX, 0);
					++d;
					do // Die Schleife wird d+1 mal ausgefhrt
					{
						if(op==BRK) // muss innerhalb der do-while-Schleife und vor der inneren while-Schleife stehen !
							pc=pc_marker;
						while(!pop_stack(&v1))
							;
						pop_stack(&v1, CONT==op);
					} while(--d);
					break;
				}
				/*case LOCALADDR:
				{
					CHOOSE_VI_TYPE(v1)=read_coded_sintegerVI(255);
					if(CHOOSE_SIGNED_VI_TYPE(v1)>=0)
						CHOOSE_SIGNED_VI_TYPE(v1)+=2*MAX_NR_BYTES_FOR_VARIABLE_INDEX + MAX_NR_BYTES_FOR_PROGRAM_COUNTER;
					CHOOSE_VI_TYPE(v1)+=act_function_stackbase;
					v1.type=VarIndexTypeNr;
					v1_contains_top_stack_element=true;
					break;
				}*/
				/*case STACKADDR:
				{
					CHOOSE_VI_TYPE(v1)=read_coded_uintegerVI(15);
					cast(&v1, VarIndexTypeNr);
					CHOOSE_VI_TYPE(v1)=find_previous_stack_element(CHOOSE_VI_TYPE(v1));
					v1.type=VarIndexTypeNr;
					v1_contains_top_stack_element=true;
					break;
				}*/
#ifdef WITH_OPCODE_GETPREV
				case STACKADDR:
				case GETPREV:
				case GETPREVLOAD:
				{
					CHOOSE_VI_TYPE(v1)=find_previous_stack_element(read_coded_uintegerVI(15));
					if(STACKADDR==op)
						v1.type=VarIndexTypeNr;
					else
					{
						get_var(CHOOSE_VI_TYPE(v1), &v1);
						if(GETPREVLOAD==op)
						{
							cast(&v1, VarIndexTypeNr);
							get_var(CHOOSE_VI_TYPE(v1), &v1, LOADT==op);
						}
					}
					v1_contains_top_stack_element=true;
					break;
				}
#endif
#ifdef WITH_OPCODE_GETLOCAL
				case LOCALADDR:
				case GETLOCAL:
				case GETLOCALLOAD:
				{
					CHOOSE_VI_TYPE(v1)=read_coded_sintegerVI(255);
					if(CHOOSE_SIGNED_VI_TYPE(v1)>=0)
						CHOOSE_SIGNED_VI_TYPE(v1)+=2*MAX_NR_BYTES_FOR_VARIABLE_INDEX + MAX_NR_BYTES_FOR_PROGRAM_COUNTER;
					if(LOCALADDR==op)
					{
						CHOOSE_VI_TYPE(v1)+=act_function_stackbase;
						v1.type=VarIndexTypeNr;
					}
					else
					{
						get_var(CHOOSE_SIGNED_VI_TYPE(v1)+act_function_stackbase, &v1);
						//if(CHOOSE_SIGNED_VI_TYPE(v1)>=0)
						//	cast(&v, get_code_nibble(act_function_codebase-(CHOOSE_SIGNED_VI_TYPE(v1)+3)));
						if(GETLOCALLOAD==op)
						{
							cast(&v1, VarIndexTypeNr);
							get_var(CHOOSE_VI_TYPE(v1), &v1, LOADT==op);
						}
					}
					v1_contains_top_stack_element=true;
					break;
				}
#endif
			}
		}
#ifdef PRINT_COMMANDS_AT_RUNTIME
		PRINT_NEWLINE_IF_PRINT_COMMANDS_AT_RUNTIME
#endif
	}
}

#undef ASSUME_VALUE_UNION_EQUAL_ADDRESSES

#undef DENSE_PACKED_STRUCT_INIT
#undef DENSE_PACKED_STRUCT_NEW_ELEMENT
#undef DENSE_PACKED_STRUCT_NEW_ARRAY
#undef DENSE_PACKED_STRUCT_END


} // Ende von namespace BytecodeInterpreter

#ifdef MSCPP
#pragma warning(pop)
#endif

#endif // BytecodeInterpreter_h_included
