// BytecodeTargetedCompiler.h

//***********************************************************************************************************//
//***********************************************************************************************************//
//                                                                                                           //
// BytecodeInterpreter.h, BytecodeTargetedCompiler.h, language_defines.h, language_undefines.h, utils.h:     //
// These files contain a bytecode interpreter and a compiler converting a C-like program into bytecode.      //
//                                                                                                           //
// Copyright 2009 by Florian Knigstein, mail@virgusta.eu                                                    //
//                                                                                                           //
//      Commercial use of this code without my permission is prohibited.                                     //
//      You may use this program free of charge for non-commercial purposes provided you agree to the        //
//      following License:                                                                                   //
//                                                                                                           //
//      The Original Work is provided under this License on an "AS IS" BASIS and WITHOUT WARRANTY,           //
//      either express or implied, including, without limitation, the warranties of                          //
//      non-infringement, merchantability or fitness for a particular purpose. THE ENTIRE RISK AS TO THE     //
//      QUALITY OF THE ORIGINAL WORK IS WITH YOU. This DISCLAIMER OF WARRANTY constitutes an essential part  //
//      of this License. No license to the Original Work is granted by this License except under this        //
//      disclaimer.                                                                                          //
//                                                                                                           //
//      Limitation of Liability. Under no circumstances and under no legal theory, whether in tort           //
//      (including negligence), contract, or otherwise, shall the Licensor be liable to anyone for any       //
//      indirect, special, incidental, or consequential damages of any character arising as a result of      //
//      this License or the use of the Original Work including, without limitation, damages for loss of      //
//      goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages    //
//      or losses. This limitation of liability shall not apply to the extent applicable law prohibits such  //
//      limitation.                                                                                          //
//                                                                                                           //
//***********************************************************************************************************//
//***********************************************************************************************************//

// ***** ATTENTION *****
// This code is not well-tested. There may be several bugs.

// Pointers can be created using the keyword 'pointer' and 'cpointer'. Pointers created using 'pointer' are
// only equivalent to ANSI C pointers if the memory the pointer points to really contains a value of the
// type given in the creation of the pointer. In this case, pointers created with 'pointer' are more space-efficient
// than pointers created with 'cpointer'. Pointers created using 'pointer' MUST NOT BE USED if they point
// to a memory position containing a type beeing different from the type given in the creation of the pointer
// (except for the zero pointer).

// Variablen, die mit local deklariert werden, entsprechen den C/C++ lokalen Variablen.
// Variablen, die mit var deklariert werden, entsprechen den C/C++ globalen Variablen, wenn sie auerhalb von
// einer Funktion deklariert werden, und den C/C++ statischen Variablen, wenn sie innerhalb einer Funktion
// deklariert werden.


#define _CRT_SECURE_NO_DEPRECATE 1

#include <stdio.h>
#include <conio.h>
#include <math.h>

#include "utils.h"
#include <stdarg.h>

#include "BytecodeInterpreter.h"

#ifdef MSCPP
#pragma warning( push )
#pragma warning( disable : 4244 ) // Warning: conversion from ... to ... ==> possible loss of data
#pragma warning( disable : 4267 ) // Warning: conversion from ... to ... ==> possible loss of data
#pragma warning( disable : 4715 ) // Warning: not all control paths return a value ==> Nicht notwendig wenn Programmabbruch durch report_error()
#endif

namespace BytecodeTargetedCompiler
{

typedef BytecodeInterpreter::BytecodeInterpreter bi;
typedef BytecodeInterpreter::Value Value;
using BytecodeInterpreter::uint8;
using BytecodeInterpreter::uint16;
using BytecodeInterpreter::uint32;
using BytecodeInterpreter::uint64;
using BytecodeInterpreter::int8;
using BytecodeInterpreter::int16;
using BytecodeInterpreter::int32;
using BytecodeInterpreter::int64;
using BytecodeInterpreter::real_type;


nstring opcode_to_command_string(int op)
{
#define MAKE_COMMAND_AND_STRING(cmd) BytecodeInterpreter::BytecodeInterpreter::cmd, #cmd
	static struct CmdName
	{
		int op;
		const char *name;
	} CommandNames[]=
	{
		MAKE_COMMAND_AND_STRING(CALL),
#ifdef WITH_OPCODE_CALLI
		MAKE_COMMAND_AND_STRING(CALLI),
#endif

		MAKE_COMMAND_AND_STRING(RET),

		MAKE_COMMAND_AND_STRING(BR),

		MAKE_COMMAND_AND_STRING(PUSHC),
#ifdef USE_TABLE_FOR_CONSTANTS
		MAKE_COMMAND_AND_STRING(PUSHCT),
#endif
		MAKE_COMMAND_AND_STRING(PUSHC0U8),
		MAKE_COMMAND_AND_STRING(PUSHC0),
		MAKE_COMMAND_AND_STRING(PUSHCS),
		MAKE_COMMAND_AND_STRING(PUSHCSU8),

		MAKE_COMMAND_AND_STRING(BRK),
		MAKE_COMMAND_AND_STRING(CONT),
		MAKE_COMMAND_AND_STRING(SETTYPES),

#ifdef WITH_OPCODE_GETPREV
		MAKE_COMMAND_AND_STRING(GETPREV),
		MAKE_COMMAND_AND_STRING(GETPREVLOAD),
#endif
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
		MAKE_COMMAND_AND_STRING(SETPREVC),
		MAKE_COMMAND_AND_STRING(SETPREVCPOP),
#endif
#ifdef WITH_OPCODE_GETLOCAL
		MAKE_COMMAND_AND_STRING(GETLOCAL),
		MAKE_COMMAND_AND_STRING(GETLOCALLOAD),
#endif
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
		MAKE_COMMAND_AND_STRING(SETLOCALC),
		MAKE_COMMAND_AND_STRING(SETLOCALCPOP),
#endif
		MAKE_COMMAND_AND_STRING(STACKADDR),
		MAKE_COMMAND_AND_STRING(LOCALADDR),

		MAKE_COMMAND_AND_STRING(ADDCPREV),
		MAKE_COMMAND_AND_STRING(PREVADDC),
		MAKE_COMMAND_AND_STRING(ADDCLOCAL),
		MAKE_COMMAND_AND_STRING(LOCALADDC),
		MAKE_COMMAND_AND_STRING(ADDCPREVPOP),
		MAKE_COMMAND_AND_STRING(ADDCLOCALPOP),

#ifdef WITH_OPCODE_NEG
		MAKE_COMMAND_AND_STRING(NEG),
#endif
#ifdef WITH_OPCODE_COM
		MAKE_COMMAND_AND_STRING(COM),
#endif
		MAKE_COMMAND_AND_STRING(CAST),
#ifdef WITH_OPCODE_GETTYPE
		MAKE_COMMAND_AND_STRING(GETTYPE),
#endif
		MAKE_COMMAND_AND_STRING(POP),

		MAKE_COMMAND_AND_STRING(LOAD),
		MAKE_COMMAND_AND_STRING(LOADT),
#ifdef WITH_OPCODE_SETPREVT
		MAKE_COMMAND_AND_STRING(SETPREVT),
#endif
#ifdef WITH_OPCODE_SETPREV
		MAKE_COMMAND_AND_STRING(SETPREV),
		MAKE_COMMAND_AND_STRING(SETPREVPOP),
		MAKE_COMMAND_AND_STRING(GETPREVSTORE),
		MAKE_COMMAND_AND_STRING(GETPREVSTOREPOP),
#endif
#ifdef WITH_OPCODE_SETLOCALT
		MAKE_COMMAND_AND_STRING(SETLOCALT),
#endif
#ifdef WITH_OPCODE_SETLOCAL
		MAKE_COMMAND_AND_STRING(SETLOCAL),
		MAKE_COMMAND_AND_STRING(SETLOCALPOP),
		MAKE_COMMAND_AND_STRING(GETLOCALSTORE),
		MAKE_COMMAND_AND_STRING(GETLOCALSTOREPOP),
#endif

#ifdef WITH_OPCODE_GETLOCALI_AND_LOCALADDRI
		MAKE_COMMAND_AND_STRING(GETLOCALI),
		MAKE_COMMAND_AND_STRING(LOCALADDRI),
		MAKE_COMMAND_AND_STRING(STACKADDRI),
#endif

#ifdef WITH_OPCODE_PUSHN
		MAKE_COMMAND_AND_STRING(PUSHN),
#endif
#ifdef WITH_OPCODE_POPN
		MAKE_COMMAND_AND_STRING(POPN),
#endif
#ifdef WITH_OPCODE_JMPI
		MAKE_COMMAND_AND_STRING(JMPI),
#endif
		MAKE_COMMAND_AND_STRING(CALLEXT),

#ifdef WITH_OPCODE_NOT
		MAKE_COMMAND_AND_STRING(NOT),
#endif

		MAKE_COMMAND_AND_STRING(BRF),
		MAKE_COMMAND_AND_STRING(BRT),
#ifndef WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION
		MAKE_COMMAND_AND_STRING(BRAND),
		MAKE_COMMAND_AND_STRING(BROR),
		MAKE_COMMAND_AND_STRING(BRANDNOT),
		MAKE_COMMAND_AND_STRING(BRORNOT),
#endif

		MAKE_COMMAND_AND_STRING(IFNOTELSE),
		MAKE_COMMAND_AND_STRING(IFELSE),
#ifdef WITH_OPCODE_LT
		MAKE_COMMAND_AND_STRING(LT),
#endif
#ifdef WITH_OPCODE_LE
		MAKE_COMMAND_AND_STRING(LE),
#endif
#ifdef WITH_OPCODE_GT
		MAKE_COMMAND_AND_STRING(GT),
#endif
#ifdef WITH_OPCODE_GE
		MAKE_COMMAND_AND_STRING(GE),
#endif
#ifdef WITH_OPCODE_EQ
		MAKE_COMMAND_AND_STRING(EQ),
#endif
#ifdef WITH_OPCODE_NE
		MAKE_COMMAND_AND_STRING(NE),
#endif

#ifdef WITH_OPCODE_STORET
		MAKE_COMMAND_AND_STRING(STORET),
#endif
		MAKE_COMMAND_AND_STRING(STORE),
		MAKE_COMMAND_AND_STRING(STORED),
		MAKE_COMMAND_AND_STRING(STOREPOP),
#ifdef WITH_OPCODE_SETLOCALI
		MAKE_COMMAND_AND_STRING(SETLOCALI),
		MAKE_COMMAND_AND_STRING(SETLOCALIT),
#endif

		MAKE_COMMAND_AND_STRING(ADD),
		MAKE_COMMAND_AND_STRING(SUB),
		MAKE_COMMAND_AND_STRING(AND),
		MAKE_COMMAND_AND_STRING(OR),
		MAKE_COMMAND_AND_STRING(XOR),
		MAKE_COMMAND_AND_STRING(SL),

		MAKE_COMMAND_AND_STRING(MUL),
		MAKE_COMMAND_AND_STRING(DIV),
		MAKE_COMMAND_AND_STRING(MOD),
		MAKE_COMMAND_AND_STRING(SR),

		MAKE_COMMAND_AND_STRING(ADDM),
		MAKE_COMMAND_AND_STRING(SUBM),
		MAKE_COMMAND_AND_STRING(ADDMPOP),
		MAKE_COMMAND_AND_STRING(SUBMPOP),
		MAKE_COMMAND_AND_STRING(ANDM),
		MAKE_COMMAND_AND_STRING(ORM),
		MAKE_COMMAND_AND_STRING(XORM),
		MAKE_COMMAND_AND_STRING(SLM),

		MAKE_COMMAND_AND_STRING(MULM),
		MAKE_COMMAND_AND_STRING(DIVM),
		MAKE_COMMAND_AND_STRING(MODM),
		MAKE_COMMAND_AND_STRING(SRM),

		-1, 0
	};
#undef MAKE_COMMAND_AND_STRING
	for(int n=0 ; -1!=CommandNames[n].op && 0!=CommandNames[n].name; n++)
		if(n!=CommandNames[n].op)
			return("****** opcodes in print_command() differ from opcodes in enum Opcodes.");
	for(int n=0 ; -1!=CommandNames[n].op && 0!=CommandNames[n].name; n++)
		if(op==CommandNames[n].op)
			return CommandNames[n].name;
	return "Unknown opcode";
}

nstring printf_string(const char *format, ...)
{
	const int bufsize=1024;
	static char buf[bufsize];
	va_list argptr;

	va_start(argptr, format);
	_vsnprintf(buf, bufsize, format, argptr);
	va_end(argptr);

	return nstring(buf);
}

void report_error(const nstring& message)
{	// report_error() muss so implementiert werden, dass ein Abbruch des Programms erfolgt.
	printf("\n\nERROR:\n");
	printf("%s", message.c_str());
	while(_getch()==0);
	exit(0);
}

struct StructType;
struct Code;
struct RegisterTypeClass;
struct StringConstantsContainer;
struct StringConstants;
struct ProgramStructureElement;
struct FunctionUIDtoFunctionName;
struct FunctionUIDtoFunctionName_Compare;

//#define USE_64_BITS_FOR_UNIQUE_COMMAND_IDENTIFIER

template<class T> struct IS_SIGNED_BUILT_IN_TYPE { };
template<> struct IS_SIGNED_BUILT_IN_TYPE<signed char> { enum { is_signed=1 }; };
template<> struct IS_SIGNED_BUILT_IN_TYPE<unsigned char> { enum { is_signed=0 }; };
template<> struct IS_SIGNED_BUILT_IN_TYPE<short> { enum { is_signed=1 }; };
template<> struct IS_SIGNED_BUILT_IN_TYPE<unsigned short> { enum { is_signed=0 }; };
template<> struct IS_SIGNED_BUILT_IN_TYPE<int> { enum { is_signed=1 }; };
template<> struct IS_SIGNED_BUILT_IN_TYPE<unsigned int> { enum { is_signed=0 }; };
template<> struct IS_SIGNED_BUILT_IN_TYPE<long> { enum { is_signed=1 }; };
template<> struct IS_SIGNED_BUILT_IN_TYPE<unsigned long> { enum { is_signed=0 }; };
template<> struct IS_SIGNED_BUILT_IN_TYPE<long long> { enum { is_signed=1 }; };
template<> struct IS_SIGNED_BUILT_IN_TYPE<unsigned long long> { enum { is_signed=0 }; };

uint64 integer_cast(uint64 x, uint8 typ_nr_vor, uint8 typ_nr_nach)
{
	uint8 i=(Value::t_void==typ_nr_vor) ? 0 : Value::sbytesize(typ_nr_vor);
	if(Value::sbytesize(typ_nr_nach) > i)
	{
		uint8 fillbytes = (Value::s_is_signed_integer(typ_nr_vor) && 0!=(((uint8*)&x)[i-1]&128)) ? 255 : 0;
		while(i<8) // arbeitet auch korrekt, wenn val vom Typ t_void ist.
			((uint8*)&x)[i++]=fillbytes;
	}
	return x;
}

uint64 cast_to_signed(uint64 x, uint8 typ_nr)
{
	if(Value::s_is_signed_integer(typ_nr))
		return x;
	switch(typ_nr)
	{
		case Value::t_u8: return integer_cast(x, Value::t_i8, Value::t_i64);
		case Value::t_u16: return integer_cast(x, Value::t_i16, Value::t_i64);
		case Value::t_u32: return integer_cast(x, Value::t_i32, Value::t_i64);
		case Value::t_u64: return integer_cast(x, Value::t_i64, Value::t_i64);
		default: report_error("Unknown type"); return 0; // return 0 wird nicht ausgefhrt, da Abbruch in report_error().
	}
}

struct Code;

struct Command
{
#ifdef USE_64_BITS_FOR_UNIQUE_COMMAND_IDENTIFIER
	typedef uint64 uid_type;
#else
	typedef uint32 uid_type;
#endif
	struct NewUID
	{
		uid_type uid;
		explicit NewUID(uid_type _uid=0) : uid(_uid) { }
	};
private:
	uid_type uid;
public:
	inline uid_type get_uid() const { return uid; }
	void set_uid(uid_type _uid) { uid=_uid; }
	static Command new_uid_label(uid_type new_uid=0)
	{
		Command cmd;
		cmd.fkt=uid_label;
		if(0!=new_uid)
			cmd.uid=new_uid;
		return cmd;
	}

	static uid_type lowest_unused_uid;
	static uid_type generate_new_uid()
	{
		uid_type uid(lowest_unused_uid);
		lowest_unused_uid++;
		if(lowest_unused_uid<min_not_reserved_uid)
			report_error("No more unique command identifiers available. Use #define USE_64_BITS_FOR_UNIQUE_COMMAND_IDENTIFIER to resolve the problem.");
		return uid;
	}
	static Command GetRawBytesCommand()
	{
		Command cmd;
		cmd.fkt=raw_nibbles;
		cmd.stop_makro=cmd.dont_use_in_makro=true;
		return cmd;
	}

	enum
	{
		maxnibblesize=40,
		min_not_reserved_uid=1 // muss gerade sein
	};
	enum Fkt
	{
		end_of_code, // end_of_code darf nur in assemble_code verwendet werden
		command, label, uid_label, branch_to_label,
		break_statement, continue_statement,
		get_prev_var, get_prev_var_and_load, get_prev_var_and_store, get_prev_var_and_storepop,
		set_prev_var, set_prev_var_const, set_prev_var_const_pop, set_prev_var_and_pop, get_prev_var_addr,
		begin_infos_before_function, function, functioncall, raw_nibbles, makrocall, makro_ret, makro_param_call_fixedsize,
		makro_param_call_varsize,
		addc_to_prev_var, prev_var_addc, addc_to_prev_var_pop,
	} fkt;
	union
	{
		unsigned char code[maxnibblesize]; // Nur gltig, wenn fkt=command.
		uint64 var_data;
	};
	const char *identifier; // identifier darf nicht zusammen mit code[] in eine union gepackt werden, weil in assemble_code() fr
		// GOTO-Statements zuerst code[] gefllt wird, aber identifier spter noch bentigt wird.
	uint8 nibblesize;
	bool stop_makro:1, dont_use_in_makro:1;
	bool is_makro_param:1;
	bool short_branch1:1;
	bool short_branch2:1;
	bool disable_escape_sequences:1;
	union
	{
		uint32 pc;
		uint32 nr_breaks_or_continues;
	};
	union
	{
		int32 sp;
		uint32 pc2;
	};
	union
	{
		int32 spie;
		int32 makro_sp;
	};
	nlist<uid_type> *commands_branching_to_this;
	union
	{
		uid_type branch1_id;
		uint32 nr_used;
	};
	uid_type branch2_id; // Die Commands, die im Code enthalten sind, der  an assemble_code() bergebenen wird,
		// drfen in branch1_id und branch2_id keine Verweise auf Commands mit fkt==command enthalten, sondern branch1_id bzw. branch2_id
		// mssen dann auf Commands mit fkt==uid_label verweisen, welche jeweils vor den Commands mit fkt==command stehen, auf die
		// man eigentlich verweisen mchte. Denn sonst gibt es in der Funktion assemble_code() Probleme, wenn ein Branch auf eine
		// Position erfolgt, welche selbst einen Sprungbefehl (BR, BRT, BRF, BRK, CONT, ...) enthlt, VOR den noch ein POPN-Befehl
		// eingefgt werden muss. Der erstgenannte Branch wrde dann zum zweiten Branch und nicht - wie gewnscht - zum POPN-Befehl
		// vor den zweiten Branch erfolgen.

	uint8 get_opcode() const
	{
		if(fkt!=command || nibblesize<2) return 255;
		return (code[0]<<4) + code[1];
	}
	void replace_opcode(uint8 new_opcode)
	{
		if(fkt==command && nibblesize>=2)
		{
			code[0]=(new_opcode>>4)&15;
			code[1]=new_opcode&15;
		}
	}
	Command(NewUID new_uid=NewUID(0)) : uid(new_uid.uid)
	{
		nibblesize=0;
		fkt=command;
		if(0==uid)
			uid=generate_new_uid();
		pc=pc2=sp=spie=0;
		stop_makro=dont_use_in_makro=is_makro_param=false;
		short_branch1=short_branch2=disable_escape_sequences=false;
		commands_branching_to_this=0;
		makro_sp=0;
	}
	explicit Command(uint8 one_byte_command, NewUID new_uid=NewUID(0)) : uid(new_uid.uid)
	{	// Dieser Konstruktor muss 'explicit' deklariert werden, denn sonst wrde z.B. der Ausdruck expr(10)
		// ungewollt in expr(Command(10)) umgewandelt.
		nibblesize=0;
		add(one_byte_command);
		fkt=command;
		if(0==uid)
			uid=generate_new_uid();
		pc=pc2=sp=spie=0;
		stop_makro=dont_use_in_makro=is_makro_param=false;
		short_branch1=short_branch2=disable_escape_sequences=false;
		commands_branching_to_this=0;
		makro_sp=0;
	}
	Command(const Command& cmd, NewUID new_uid=NewUID(0));
	Command& operator=(const Command& cmd);
	~Command()
	{
		if(0!=commands_branching_to_this)
			delete commands_branching_to_this;
	}
	Command& addnibble(uint8 nibble)
	{
		if(nibble>=16)
			report_error("Internal compiler error: Nibble cannot be >=16");
		if(nibblesize>=maxnibblesize)
			report_error("Command longer than \'maxnibblesize\' nibbles");
		code[nibblesize++]=nibble;
		return *this;
	}
	Command& addnibblefront(uint8 nibble)
	{
		if(nibblesize>=maxnibblesize)
			report_error("Command longer than \'maxnibblesize\' nibbles");
		for(long i=nibblesize; i>0; i--)
			code[i]=code[i-1];
		code[0]=nibble;
		++nibblesize;
		return *this;
	}
	Command& add (uint8 x) { addnibble((x>>4)&15); addnibble(x&15); return *this; }
	Command& add8(uint8 x) { addnibble((x>>4)&15); addnibble(x&15); return *this; }
	Command& add16(uint16 x) { add((((uint16)(x))>>8)&255); add(((uint16)(x))&255); return *this; }
	Command& add32(uint32 x)
	{
		for(uint8 i=4; i--; )
			add((uint8)((((uint32)(x))>>(8*i))&255));
		return *this;
	}
	Command& add64(uint64 x)
	{
		for(uint8 i=8; i--; )
			add((uint8)((((uint64)(x))>>(8*i))&255));
		return *this;
	}
	Command& addreal(real_type x)
	{
		add64(*((uint64*)(&x)));
		return *this;
	}

	Command& addfront (uint8 x) { addnibblefront(x&15); addnibblefront((x>>4)&15); return *this; }
	Command& add8front(uint8 x) { addnibblefront(x&15); addnibblefront((x>>4)&15); return *this; }
	Command& add16front(uint16 x) { addfront(((uint16)(x))&255); addfront((((uint16)(x))>>8)&255); return *this; }
	Command& add32front(uint32 x)
	{
		for(uint8 i=0; i<4; i++)
			addfront((uint8)((((uint32)(x))>>(8*i))&255));
		return *this;
	}
	Command& add64front(uint64 x)
	{
		for(uint8 i=0; i<8; i++)
			addfront((uint8)((((uint64)(x))>>(8*i))&255));
		return *this;
	}
	Command& addrealfront(real_type x)
	{
		add64front(*((uint64*)(&x)));
		return *this;
	}
	Command& set_sp(int32 _sp, int32 _spie=0) { sp=_sp; spie=_spie; return *this; }

	Command& add_coded_integer15(int64 n, uint8 nr_bytes_if_escape, bool _signed);
	Command& add_coded_integer255(int64 n, uint8 nr_bytes_if_escape, bool _signed);
};

Command::uid_type Command::lowest_unused_uid=Command::min_not_reserved_uid;

struct Command_Compare : public BaumKeineAttrBerechnen<Command>
{
	bool less(const Command& c1, const Command& c2) const
	{
		return c1.get_uid() < c2.get_uid();
	}
	bool equal(const Command& c1, const Command& c2) const
	{
		return c1.get_uid() == c2.get_uid();
	}
};

struct FunctionUIDtoFunctionName
{
	Command::uid_type function_uid;
	nstring name;
	FunctionUIDtoFunctionName(Command::uid_type _uid, nstring _name=nstring()) : function_uid(_uid), name(_name) { }
};

struct FunctionUIDtoFunctionName_Compare : public BaumKeineAttrBerechnen<FunctionUIDtoFunctionName>
{
	bool less(const FunctionUIDtoFunctionName& f1, const FunctionUIDtoFunctionName& f2) const
	{
		return f1.function_uid<f2.function_uid;
	}
	bool equal(const FunctionUIDtoFunctionName& f1, const FunctionUIDtoFunctionName& f2) const
	{
		return f1.function_uid==f2.function_uid;
	}
};

struct nstring_LessEqual : public BaumKeineAttrBerechnen<nstring>
{
	bool less(const nstring& s1, const nstring& s2) const
	{
		return s1.compare(s2)<0;
	}
	bool equal(const nstring& s1, const nstring& s2) const
	{
		return s1.compare(s2)==0;
	}
};

struct Compiler
{
	int CompilerPassNr;
	StructType *ActualStructDefinition;
	nlist<nstring> defined_goto_labels_of_actual_function, forward_goto_targets_of_actual_function;
	Code *program_code;
	nlist<ProgramStructureElement*> ProgramStructureStack;
	RegisterTypeClass *RegisterType;
	StringConstantsContainer *StringConstants;
	RotSchwarzBaum<FunctionUIDtoFunctionName, FunctionUIDtoFunctionName_Compare> functionUID2name;
	Compiler(void (*_CFunctionForOneCompilerPass)(Compiler* compiler));
	~Compiler();
	mapping<nstring, PROGRAM_COUNTER_TYPE, nstring_LessEqual> function_begin_addresses;

	//////////////////////////////////////////////////////////////////////////////////////////////////////

	uint8* bytecode;
	uint32 nibble_codesize;
	uint32 required_memory_size;
} *cc;

void report_warning(const nstring& message)
{
	if(1==cc->CompilerPassNr)
	{
		printf("\nWARNING:\n");
		printf("%s\n", message.c_str());
	}
}

struct expr;
struct Statement;

struct Code
{
	nlist<Command> code;
	Code() { }
	operator nlist<Command>&() { return code; }
	Code(const nlist<Command>& _code) : code(_code) { }
	Code(const Code& ex) : code(ex.code) { }
	Code(const Command& cmd) { code.push_back(cmd); }
	Code& operator=(const Code& ex) { if(this!=&ex) { code=ex.code; } return *this; }
	Code(const expr& ex);
	Code& operator=(const expr& ex);
	void add(const Code& x)//, size_t add_sp=0)
		{ code.insert(code.end(), x.code.begin(), x.code.end()); }
	/*void add(const Code& x, size_t add_sp=0)
	{
		nlist<Command>::iterator it=code.insert(code.end(), x.code.begin(), x.code.end());
		if(it!=0 && 0!=add_sp && 0!=x.code.size())
		{
			it->sp+=add_sp;
			Command cmd=Command::new_uid_label();
			cmd.sp=-add_sp;
			code.push_back(cmd);
		}
	}*/
	//void add(const expr& x, size_t add_sp=0);
	void add(const expr& x);
	void addfront(const Code& x) { code.insert(code.begin(), x.code.begin(), x.code.end()); }
	void addfront(const expr& x);
	void add(const Command& cmd) { code.push_back(cmd); }
	void addfront(const Command& cmd) { code.push_front(cmd); }
	void addcmd(uint8 x, int32 add_sp=0) { add(Command(x).set_sp(add_sp)); }
	void addcmdfront(uint8 x, int32 add_sp=0) { addfront(Command(x).set_sp(add_sp)); }
	Command::uid_type get_uid_of_first_command()
	{	// ****** Fgt eine uid-Marke ein ******
		// Es muss auch dann eine uid-Marke eingefgt werden, wenn code.size()!=0 ist, denn:
		// Die Commands, die im Code enthalten sind, der  an assemble_code() bergebenen wird,
		// drfen in branch1_id und branch2_id keine Verweise auf Commands mit fkt==command enthalten, sondern branch1_id bzw. branch2_id
		// mssen dann auf Commands mit fkt==uid_label verweisen, welche jeweils vor den Commands mit fkt==command stehen, auf die
		// man eigentlich verweisen mchte. Denn sonst gibt es in der Funktion assemble_code() Probleme, wenn ein Branch auf eine
		// Position erfolgt, welche selbst einen Sprungbefehl (BR, BRT, BRF, BRK, CONT, ...) enthlt, VOR den noch ein POPN-Befehl
		// eingefgt werden muss. Der erstgenannte Branch wrde dann zum zweiten Branch und nicht - wie gewnscht - zum POPN-Befehl
		// vor den zweiten Branch erfolgen.
		addfront(Command::new_uid_label());
		return code.front().get_uid();
	}
	Command::uid_type get_uid_after_last_command()
	{	// ****** fgt eine uid-Marke ein ******
		add(Command::new_uid_label());
		return code.back().get_uid();
	}
	static Code POPN_Commands(uint32 n, bool with_sp=false)
	{
		if(1==n)
			return Code(Command(bi::POP).set_sp(with_sp ? -1 : 0));
		else
		{
			Code cd;
			if(0==(n & ~0xffLU))
				cd.add(Command(bi::PUSHC).addnibble(Value::t_u8).add8(n));
			else if(0==(n & ~0xffffLU))
				cd.add(Command(bi::PUSHC).addnibble(Value::t_u16).add16(n));
			else if(0==(n & ~0xffffffffLU))
				cd.add(Command(bi::PUSHC).addnibble(Value::t_u32).add32(n));
			else
				cd.add(Command(bi::PUSHC).addnibble(Value::t_u64).add64(n));
			Command cmd=Command(bi::POPN);
			cmd.sp=with_sp ? (-(int32)n) : 0;
			cd.add(cmd);
			return cd;
		}
	}
};

Command::Command(const Command& cmd, NewUID new_uid/*=NewUID(0)*/)
{
	fkt=cmd.fkt;
	uid=(0==new_uid.uid) ? cmd.uid : new_uid.uid;
	pc=cmd.pc;
	pc2=cmd.pc2;
	branch1_id=cmd.branch1_id;
	branch2_id=cmd.branch2_id;
	nr_used=cmd.nr_used;
	nibblesize=cmd.nibblesize;
	identifier=cmd.identifier;
	for(long n=0; n<maxnibblesize; n++)
		code[n]=cmd.code[n];
	pc=cmd.pc;
	nr_breaks_or_continues=cmd.nr_breaks_or_continues;
	sp=cmd.sp;
	spie=cmd.spie;
	makro_sp=cmd.makro_sp;
	stop_makro=cmd.stop_makro;
	dont_use_in_makro=cmd.dont_use_in_makro;
	is_makro_param=cmd.is_makro_param;
	short_branch1=cmd.short_branch1;
	short_branch2=cmd.short_branch2;
	disable_escape_sequences=cmd.disable_escape_sequences;
	if(0!=(commands_branching_to_this=cmd.commands_branching_to_this))
	{
		commands_branching_to_this=new nlist<uid_type>;
		for(nlist<uid_type>::iterator it=cmd.commands_branching_to_this->begin(); it!=cmd.commands_branching_to_this->end(); it++)
			commands_branching_to_this->push_back(*it);
	}
}

Command& Command::operator=(const Command& cmd)
{
	if(this!=&cmd)
	{
		fkt=cmd.fkt;
		uid=cmd.uid;
		pc=cmd.pc;
		pc2=cmd.pc2;
		nr_breaks_or_continues=cmd.nr_breaks_or_continues;
		sp=cmd.sp;
		spie=cmd.spie;
		makro_sp=cmd.makro_sp;
		branch1_id=cmd.branch1_id;
		branch2_id=cmd.branch2_id;
		nr_used=cmd.nr_used;
		nibblesize=cmd.nibblesize;
		identifier=cmd.identifier;
		for(long n=0; n<maxnibblesize; n++)
			code[n]=cmd.code[n];
		stop_makro=cmd.stop_makro;
		dont_use_in_makro=cmd.dont_use_in_makro;
		is_makro_param=cmd.is_makro_param;
		short_branch1=cmd.short_branch1;
		short_branch2=cmd.short_branch2;
		disable_escape_sequences=cmd.disable_escape_sequences;
		if(0!=commands_branching_to_this)
			delete commands_branching_to_this;
		if(0!=(commands_branching_to_this=cmd.commands_branching_to_this))
		{
			commands_branching_to_this=new nlist<uid_type>;
			for(nlist<uid_type>::iterator it=cmd.commands_branching_to_this->begin(); it!=cmd.commands_branching_to_this->end(); it++)
				commands_branching_to_this->push_back(*it);
		}
	}
	return *this;
}

struct Type;
struct ProgramStructureElement
{
	virtual void add(const Code& ex)=0;
	void add(const expr& ex);
	virtual void add_local_variable(const Command& var_label, const expr& init_value)=0;
	virtual VARIABLE_INDEX_TYPE add_uninit_local_variable(const Type *type)=0;
	virtual void end(ProgramStructureElement *superordinate_element)=0;
	virtual ~ProgramStructureElement() { }
};

template<class T> void ExprObjectDestructionHandler(T& ex);
template<> void ExprObjectDestructionHandler<expr>(expr& ex);
template<> void ExprObjectDestructionHandler<Statement>(Statement& ex);
struct local;

struct GeneralExprType
{
	virtual ~GeneralExprType() { }
	virtual GeneralExprType *vcopy() const =0;
	virtual void vset_exported() const=0;
	virtual void vset_not_exported() const=0;
	virtual expr to_expr() const=0;
	virtual bool is_expr_type() const=0;
};

struct expr : public GeneralExprType, Code
{
private: void operator=(const expr& s) { report_error("Assignment expr::operator=() must not be called"); } // Zugriff verhindern. Wenn der Zugriff auf den Standard-Copykonstruktor
		 // erlaubt wre, wrde z.B. der sinnlose Ausdruck expr(5) = expr(7) ohne Fehlermeldung akzeptiert.
private: template<class RHS> void operator=(const RHS& expr) { report_error("Assignment to non l-value"); }
		 // Zugriff verhindern. Wenn der Zugriff auf den Standard-Copykonstruktor erlaubt wre, wrde z.B. der sinnlose Ausdruck
		 // expr(5) = x (wobei x ein Ausdruck ist) ohne Fehlermeldung akzeptiert.
		 // Da MS Visual C++ jedoch das 'private' ignoriert, wird zustzlich noch eine Ausnahme geworfen, um den Fehler anzuzeigen.

private:
	void construct(uint64 x, uint8 bsize, bool _signed)
	{
		exported=false;
		if(0==x)
		{
			if(1==bsize && !_signed)
				{ add(Command(bi::PUSHC0U8).set_sp(1)); type_nr=Value::t_u8; }
			else if(1==bsize)
				add(Command(bi::PUSHC0).set_sp(1).addnibble(type_nr = (_signed ? Value::t_i8 : Value::t_u8)));
			else if(2==bsize)
				add(Command(bi::PUSHC0).set_sp(1).addnibble(type_nr = (_signed ? Value::t_i16 : Value::t_u16)));
			else if(4==bsize)
				add(Command(bi::PUSHC0).set_sp(1).addnibble(type_nr = (_signed ? Value::t_i32 : Value::t_u32)));
			else if(8==bsize)
				add(Command(bi::PUSHC0).set_sp(1).addnibble(type_nr = (_signed ? Value::t_i64 : Value::t_u64)));
			return;
		}
		if(x<16)
		{
			if(1==bsize && !_signed)
				{ add(Command(bi::PUSHCSU8).set_sp(1).addnibble(x)); type_nr=Value::t_u8; }
			else if(1==bsize)
				add(Command(bi::PUSHCS).set_sp(1).addnibble(type_nr = (_signed ? Value::t_i8 : Value::t_u8)).addnibble(x));
			else if(2==bsize)
				add(Command(bi::PUSHCS).set_sp(1).addnibble(type_nr = (_signed ? Value::t_i16 : Value::t_u16)).addnibble(x));
			else if(4==bsize)
				add(Command(bi::PUSHCS).set_sp(1).addnibble(type_nr = (_signed ? Value::t_i32 : Value::t_u32)).addnibble(x));
			else if(8==bsize)
				add(Command(bi::PUSHCS).set_sp(1).addnibble(type_nr = (_signed ? Value::t_i64 : Value::t_u64)).addnibble(x));
			return;
		}
		const uint64 hbs=((uint64)0xffffffffLU)<<32;
		type_nr=Value::t_unknown_type;
		if(bsize>=2 && (_signed ? (0==(x & (hbs|(uint64)0xffffff80)) || (hbs|(uint64)0xffffff80)==(x & (hbs|(uint64)0xffffff80))) :
				(0==(x & ~((uint64)0xffLU)))))
		{
			if(2==bsize)
			{
				add(Command(bi::PUSHC).set_sp(1).addnibble(_signed ? Value::t_i16_but8used : Value::t_u16_but8used).add8(x));
				type_nr=_signed ? Value::t_i16 : Value::t_u16;
			}
			else if(4==bsize)
			{
				add(Command(bi::PUSHC).set_sp(1).addnibble(_signed ? Value::t_i32_but8used : Value::t_u32_but8used).add8(x));
				type_nr=_signed ? Value::t_i32 : Value::t_u32;
			}
			else if(8==bsize)
			{
				add(Command(bi::PUSHC).set_sp(1).addnibble(_signed ? Value::t_i64_but8used : Value::t_u64_but8used).add8(x));
				type_nr=_signed ? Value::t_i64 : Value::t_u64;
			}
		}
		else if(bsize>=4 && (_signed ? (0==(x & (hbs|(uint64)0xffff8000)) || (hbs|(uint64)0xffff8000)==(x & (hbs|(uint64)0xffff8000))) :
				(0==(x & ~((uint64)0xffffLU)))))
		{
			if(4==bsize)
			{
				add(Command(bi::PUSHCS).set_sp(1).addnibble(_signed ? Value::t_i32_but16used : Value::t_u32_but16used).add16(x));
				type_nr=_signed ? Value::t_i32 : Value::t_u32;
			}
			else if(8==bsize)
			{
				add(Command(bi::PUSHCS).set_sp(1).addnibble(_signed ? Value::t_i64_but32used : Value::t_u64_but32used).add32(x));
				type_nr=_signed ? Value::t_i64 : Value::t_u64;
			}
		}
		else if(bsize>=8 && (_signed ? (0==(x & (hbs|(uint64)0x80000000)) || (hbs|(uint64)0x80000000)==(x & (hbs|(uint64)0x80000000))) :
				(0==(x & ~((uint64)0xffffffffLU)))))
		{
			add(Command(bi::PUSHCS).set_sp(1).addnibble(_signed ? Value::t_i64_but32used : Value::t_u64_but32used).add32(x));
			type_nr=_signed ? Value::t_i64 : Value::t_u64;
		}
		else if(1==bsize)
			add(Command(bi::PUSHC).set_sp(1).addnibble(type_nr = (_signed ? Value::t_i8 : Value::t_u8)).add8(x));
		else if(2==bsize)
			add(Command(bi::PUSHC).set_sp(1).addnibble(type_nr = (_signed ? Value::t_i16 : Value::t_u16)).add16(x));
		else if(4==bsize)
			add(Command(bi::PUSHC).set_sp(1).addnibble(type_nr = (_signed ? Value::t_i32 : Value::t_u32)).add32(x));
		else if(8==bsize)
			add(Command(bi::PUSHC).set_sp(1).addnibble(type_nr = (_signed ? Value::t_i64 : Value::t_u64)).add64(x));
	}
	uint8 type_nr; // private, damit nicht ein temporrer expr generiert wird, aus dem type_nr direkt ausgelseen wird, ohne
		// exported=true zu setzen.
	mutable bool code_has_effect;

public:
	mutable bool exported;
	nlist<const local*> assignment_to_locals;
	expr& set_type(uint8 new_type_nr)
	{
		exported=true; // set_type() darf nicht innerhalb eines Konstruktors von expr aufgerufen werden !
		type_nr=new_type_nr;
		return *this;
	}
	uint8 get_type() const
	{
		exported=true; // set_type() darf nicht innerhalb eines Konstruktors von expr aufgerufen werden !
		return type_nr;
	}
	const expr& set_code_has_effect() const { code_has_effect=true; return *this; }
	bool has_code_effect() const { return code_has_effect; }
	bool is_integer_constant() const
	{
		set_exported();
		if(code.size()!=1 || Command::command!=code.front().fkt)
			return false;
		uint8 op=code.front().get_opcode();
		return (op==bi::PUSHC0U8 || op==bi::PUSHC0 || op==bi::PUSHCSU8 || op==bi::PUSHCS ||
			(op==bi::PUSHC && code.front().nibblesize>=3 &&
				(Value::t_i16_but8used==code.front().code[2] || code.front().code[2]>=Value::t_u16_but8used ||
				Value::s_is_integer_type(code.front().code[2]))));
	}
	uint64 get_this_integer_constant() const
	{
		set_exported();
		if(!is_integer_constant())
			report_error("Integer constant required");
		Command cmd=code.front();
		if(Command::command!=cmd.fkt)
			report_error("Integer constant required");

		uint8 op=cmd.get_opcode(), i=2, type, nibblesize;

		if(op==bi::PUSHC0 || op==bi::PUSHC0U8)
			return 0;
		if(op==bi::PUSHCSU8)
			return cmd.code[2];
		if(op==bi::PUSHCS && cmd.code[2]<Value::t_i32_but16used)
			return cmd.code[3];
	
		type=cmd.code[i++];
		if(op==bi::PUSHCS && type>=Value::t_i32_but16used)
			{ nibblesize=(type>=Value::t_i64_but32used) ? 8 : 4; type-=Value::t_i32_but16used-Value::t_i32; }
		else if(op==bi::PUSHC && Value::t_i16_but8used==type)
			{ type=Value::t_i16; nibblesize=2; }
		else if(op==bi::PUSHC && type>=Value::t_u16_but8used)
			{ type-=Value::t_u16_but8used-Value::t_u16; nibblesize=2; }
		else
			nibblesize=(Value::sbytesize(type)<<1);

		if(type!=type_nr) // Es kann nibblesize < (Value::sbytesize(type)<<1)) sein !
			report_error("Internal compiler error, get_this_integer_constant(): type_nr error");
		if(0==nibblesize)
			return 0;
		uint64 x=0;
		uint8 z=cmd.code[i++];
		if(Value::s_is_signed_integer(type_nr) && 0!=(z&8))
			x=~((uint64)0);
		do { x=(x<<4) | z; } while(0!=--nibblesize && (z=cmd.code[i++], true));
		return x;
	}

	GeneralExprType *vcopy() const
		{ set_exported(); return new expr(*this); }
	expr()
		{ exported=false; code_has_effect=true; type_nr=Value::t_unknown_type; } // code_has_effect soll bei leerem Code true sein !
	expr(const Command& cmd) { code.push_back(cmd); exported=false; code_has_effect=false; type_nr=Value::t_unknown_type; }
	expr(const expr& ex) : Code(ex), exported(false), assignment_to_locals(ex.assignment_to_locals), type_nr(ex.type_nr),
			code_has_effect(ex.code_has_effect)
		{ ex.exported=true; } // Struktur expr kopieren, keine Aktionen fr den Bytecode-Interpreter !
	expr& assign(const expr& ex)
	{
		Code::operator=(ex); exported=false; ex.exported=true; assignment_to_locals=ex.assignment_to_locals; type_nr=ex.type_nr;
		return *this;
	}
	//void add(const expr& ex, size_t add_sp=0) { exported=false; ex.exported=true; Code::add(ex, add_sp); }
	void add(const expr& ex)
	{	// Es wird bewusst darauf verzichtet, den Typ this->type_nr auf ex.type_nr zu setzen. Statt dessen muss
		// this->type_nr in jedem Einzelfall korrekt gesetzt werden.
		exported=false; ex.exported=true; Code::add(ex);
		assignment_to_locals.insert(assignment_to_locals.end(), ex.assignment_to_locals.begin(), ex.assignment_to_locals.end());
		// ACHTUNG: Wenn der Code ex nur unter bestimmten Bedingungen ausgefhrt wird, z.B. beim operator&&(), muss die Liste
		// assignment_to_locals ggf. wieder gelscht werden !
		if(!ex.code_has_effect)
			code_has_effect=false;
	}
	void add_effective(const expr& ex)
	{	// Es wird bewusst darauf verzichtet, den Typ this->type_nr auf ex.type_nr zu setzen. Statt dessen muss
		// this->type_nr in jedem Einzelfall korrekt gesetzt werden.
		exported=false; ex.exported=true; Code::add(ex);
		assignment_to_locals.insert(assignment_to_locals.end(), ex.assignment_to_locals.begin(), ex.assignment_to_locals.end());
		// ACHTUNG: Wenn der Code ex nur unter bestimmten Bedingungen ausgefhrt wird, z.B. beim operator&&(), muss die Liste
		// assignment_to_locals ggf. wieder gelscht werden !
	}
	void add_set_type(const expr& ex)
	{
		exported=false; ex.exported=true; Code::add(ex);
		assignment_to_locals.insert(assignment_to_locals.end(), ex.assignment_to_locals.begin(), ex.assignment_to_locals.end());
		// ACHTUNG: Wenn der Code ex nur unter bestimmten Bedingungen ausgefhrt wird, z.B. beim operator&&(), muss die Liste
		// assignment_to_locals ggf. wieder gelscht werden !
		type_nr=ex.type_nr;
		if(!ex.code_has_effect)
			code_has_effect=false;
	}
	//void add(const Code& cd, size_t add_sp=0) { exported=false; Code::add(cd, add_sp); }
	void add(const Code& cd) { exported=false; Code::add(cd); code_has_effect=false; }
	void addcmd(uint8 x, int32 add_sp=0) { Code::addcmd(x, add_sp); code_has_effect=false; }
	void addcmdfront(uint8 x, int32 add_sp=0) { Code::addcmdfront(x, add_sp); code_has_effect=false; }
	void addfront(const expr& ex)
	{
		exported=false; ex.exported=true; Code::addfront(ex);
		assignment_to_locals.insert(assignment_to_locals.end(), ex.assignment_to_locals.begin(), ex.assignment_to_locals.end());
		if(!ex.code_has_effect)
			code_has_effect=false;
	}
	void addfront_effective(const expr& ex)
	{
		exported=false; ex.exported=true; Code::addfront(ex);
		assignment_to_locals.insert(assignment_to_locals.end(), ex.assignment_to_locals.begin(), ex.assignment_to_locals.end());
	}
	void addfront(const Code& cd) { exported=false; Code::addfront(cd); code_has_effect=false; }
	const expr& set_exported() const { exported=true; return *this; }
	expr& set_exported() { exported=true; return *this; }
	void vset_exported() const { set_exported(); }
	void set_not_exported() const { exported=false; }
	void vset_not_exported() const { set_not_exported(); }
	void add(const Command& cmd) { code.push_back(cmd); code_has_effect=false; }
	void add_effective(const Command& cmd) { code.push_back(cmd); }
	void addfront(const Command& cmd) { code.push_front(cmd); code_has_effect=false; }
	void addfront_effective(const Command& cmd) { code.push_front(cmd); }
	expr to_expr() const { exported=true; return *this; }
	bool is_expr_type() const { exported=true; return true; }
	~expr()
	{
		if(!exported)
			ExprObjectDestructionHandler<expr>(*this);
	}
	static expr voidexpr()
	{
		expr ex(Command(bi::PUSHC).set_sp(1).addnibble(Value::t_void));
		ex.type_nr=Value::t_void;
		return ex.set_exported();
	}
	expr(bool x)
	{
		exported=false; code_has_effect=false;
		type_nr=Value::t_bool;
		if(false==x)
			add(Command(bi::PUSHC0).set_sp(1).addnibble(Value::t_bool));
		else
			add(Command(bi::PUSHCS).set_sp(1).addnibble(Value::t_bool).addnibble(x ? 1 : 0));
	}
	expr(unsigned char x) { construct(x, sizeof(x), false); code_has_effect=false; }
	expr(signed char x) { construct(x, sizeof(x), true); code_has_effect=false; }
	expr(unsigned short x) { construct(x, sizeof(x), false); code_has_effect=false; }
	expr(short x) { construct(x, sizeof(x), true); code_has_effect=false; }
	expr(unsigned int x) { construct(x, sizeof(x), false); code_has_effect=false; }
	expr(int x) { construct(x, sizeof(x), true); code_has_effect=false; }
	expr(unsigned long x) { construct(x, sizeof(x), false); code_has_effect=false; }
	expr(long x) { construct(x, sizeof(x), true); code_has_effect=false; }
	expr(unsigned long long x) { construct(x, sizeof(x), false); code_has_effect=false; }
	expr(long long x) { construct(x, sizeof(x), true); code_has_effect=false; }
	expr(float x)
	{
		exported=false; code_has_effect=false; 
		type_nr=Value::t_real;
		add(Command(bi::PUSHC).set_sp(1).addnibble(Value::t_real).addreal((real_type)x));
	}
	expr(double x)
	{
		exported=false; code_has_effect=false; 
		type_nr=Value::t_real;
		add(Command(bi::PUSHC).set_sp(1).addnibble(Value::t_real).addreal((real_type)x));
	}
	static expr get_const_expr(uint64 x)
	{	// Die Funktion shortest_type_if_constant() verlsst sich darauf, dass in dieser get_const_expr()-berladung nur
		// unsigned-Typen erzeugt werden.
		expr tex;
		if(0==x)
			{ tex.add(Command(bi::PUSHC0U8).set_sp(1)); tex.type_nr=Value::t_u8; }
		else if(0==(x & ~((uint64)0xfLU)))
			{ tex.add(Command(bi::PUSHCSU8).set_sp(1).addnibble(x)); tex.type_nr=Value::t_u8; }
		else if(0==(x & ~((uint64)0xffLU)))
			tex.add(Command(bi::PUSHC).set_sp(1).addnibble(tex.type_nr=Value::t_u8).add8(x));
		else if(0==(x & ~((uint64)0xffffLU)))
			tex.add(Command(bi::PUSHC).set_sp(1).addnibble(tex.type_nr=Value::t_u16).add16(x));
		else if(0==(x & ~((uint64)0xffffffffLU)))
			tex.add(Command(bi::PUSHC).set_sp(1).addnibble(tex.type_nr=Value::t_u32).add32(x));
		else
			tex.add(Command(bi::PUSHC).set_sp(1).addnibble(tex.type_nr=Value::t_u64).add64(x));
		return expr(tex);
	}
	static expr get_const_expr(int64 x, bool force_signed_type=false)
	{	// Die Funktion shortest_type_if_constant() verlsst sich darauf, dass in dieser get_const_expr()-berladung nur
		// signed-Typen erzeugt werden, sofern force_signed_type=true ist.
		// Wenn aber force_signed_type==false ist, wird ggf. auf unsigned konvertiert, wenn dadurch ein krzerer Typ mglich wird.
		if(x>=0 && !force_signed_type)
			return get_const_expr((uint64)x);
		const uint64 hbs=((uint64)0xffffffffLU)<<32;
		expr tex;
		if(0==x)
		{
			if(force_signed_type)
				{ tex.add(Command(bi::PUSHC0).addnibble(Value::t_i8).set_sp(1)); tex.type_nr=Value::t_i8; }
			else
				{ tex.add(Command(bi::PUSHC0U8).set_sp(1)); tex.type_nr=Value::t_u8; }
		}
		else if(!force_signed_type && 0==(x & ~((uint64)0xfLU))) // 0<=x<16 (==> x nicht negativ)
			{ tex.add(Command(bi::PUSHCSU8).set_sp(1).addnibble(x)); tex.type_nr=Value::t_u8; }
		else if(0==(x & (hbs|(uint64)0xffffff80)) || (hbs|(uint64)0xffffff80)==(x & (hbs|(uint64)0xffffff80)))
			tex.add(Command(bi::PUSHC).set_sp(1).addnibble(tex.type_nr=Value::t_i8).add8(x));
		else if(0==(x & (hbs|(uint64)0xffff8000)) || (hbs|(uint64)0xffff8000)==(x & (hbs|(uint64)0xffff8000)))
			tex.add(Command(bi::PUSHC).set_sp(1).addnibble(tex.type_nr=Value::t_i16).add16(x));
		else if(0==(x & (hbs|(uint64)0x80000000)) || (hbs|(uint64)0x80000000)==(x & (hbs|(uint64)0x80000000)))
			tex.add(Command(bi::PUSHC).set_sp(1).addnibble(tex.type_nr=Value::t_i32).add32(x));
		else
			tex.add(Command(bi::PUSHC).set_sp(1).addnibble(tex.type_nr=Value::t_i64).add64(x));
		return expr(tex);
	}
	static expr create_with_type(uint64 x, uint8 type)
	{
		if(Value::t_bool==type)
			return expr(x ? true : false); // return expr((bool)x);
		if(Value::t_real==type)
			return expr((real_type)x);
		expr ex;
		ex.construct(x, Value::sbytesize(type), Value::s_is_signed_integer(type));
		return ex;
	}
	static expr create_with_cast_up_type(uint64 x, const expr& ex1, const expr& ex2)
	{
		ex1.set_exported(); ex2.set_exported();
		return create_with_type(x, ex1.type_nr>ex2.type_nr ? ex1.type_nr : ex2.type_nr);
	}
	template<class T> static int get_nr_of_used_bits_template(T x, bool return_one_if_zero=true)
	{
		if(return_one_if_zero && 0==x)
			return 1;
		else if(x>=0)
		{
			int bits=sizeof(T)*8;
			while(0!=bits && 0==((((uint64)1)<<(bits-1)) & x))
				--bits;
			return (IS_SIGNED_BUILT_IN_TYPE<T>::is_signed && bits<sizeof(T)*8) ? (bits+1) : bits;
		}
		else
		{
			int bits=sizeof(T)*8;
			while(0!=bits && 0!=((((uint64)1)<<(bits-1)) & x))
				--bits;
			return (bits<sizeof(T)*8) ? (bits+1) : bits;
		}
	}
	static int get_nr_of_used_bits(uint64 x, uint8 nr_bytes, bool _signed, bool return_one_if_zero=true)
	{
		if(return_one_if_zero && 0==x)
			return 1;
		else if(!_signed || ((int64)x)>=0)
		{
			int bits=nr_bytes*8;
			while(0!=bits && 0==((((uint64)1)<<(bits-1)) & x))
				--bits;
			return (_signed && bits<nr_bytes*8) ? (bits+1) : bits;
		}
		else
		{
			int bits=nr_bytes*8;
			while(0!=bits && 0!=((((uint64)1)<<(bits-1)) & x))
				--bits;
			return (bits<nr_bytes*8) ? (bits+1) : bits;
		}
	}

	static expr get_const_expr(unsigned long x) { return get_const_expr((uint64)x); }
	static expr get_const_expr(long x, bool force_signed_type=false) { return get_const_expr((int64)x, force_signed_type); }
	static expr get_const_expr(unsigned int x) { return get_const_expr((uint64)x); }
	static expr get_const_expr(int x, bool force_signed_type=false) { return get_const_expr((int64)x, force_signed_type); }
	static expr get_const_expr(unsigned short x) { return get_const_expr((uint64)x); }
	static expr get_const_expr(short x, bool force_signed_type=false) { return get_const_expr((int64)x, force_signed_type); }
	static expr get_const_expr(unsigned char x) { return get_const_expr((uint64)x); }
	static expr get_const_expr(signed char x, bool force_signed_type=false) { return get_const_expr((int64)x, force_signed_type); }

	Command::uid_type get_uid_of_first_command()
	{	// ****** fgt eine uid-Marke ein, falls code leer ist ******
		exported=false; // weil ggf. eine uid-Marke eingefgt wird
		addfront_effective(Command::new_uid_label());
		return code.front().get_uid();
	}
	Command::uid_type get_uid_after_last_command()
	{	// ****** fgt eine uid-Marke ein ******
		exported=false; // weil ggf. eine uid-Marke eingefgt wird
		add_effective(Command::new_uid_label());
		return code.back().get_uid();
	}

	bool is_signed_integer_type() const
	{
		set_exported();
		return type_nr!=Value::t_unknown_type && type_nr!=Value::t_nil && type_nr>=Value::t_u8 && type_nr<Value::t_real && 0==(type_nr&1);
	}
	uint8 get_type_of_cast_up(const expr& ex) const
	{
		set_exported(); // Wichtig !
		ex.set_exported();  // Wichtig, weil expr auch oft nur dazu erzeugt werden, um get_type_of_cast_up() aufzurufen.
		if(type_nr==Value::t_nil || ex.type_nr==Value::t_nil)
			return Value::t_nil;
		if(type_nr==Value::t_unknown_type || ex.type_nr==Value::t_unknown_type)
			return Value::t_unknown_type;
		return type_nr > ex.type_nr ? type_nr : ex.type_nr;
	}
	nstring get_type_string() const;

	void add_LOAD_optimize()
	{
		if(code.size()>=1 && code.back().get_opcode()==bi::LOCALADDR)
			code.back().replace_opcode(bi::GETLOCAL);
		else if(code.size()>=1 && code.back().fkt==Command::get_prev_var_addr)
			code.back().fkt=Command::get_prev_var;
		else if(code.size()>=1 && code.back().fkt==Command::get_prev_var)
			code.back().fkt=Command::get_prev_var_and_load;
		else if(code.size()>=1 && code.back().get_opcode()==bi::GETLOCAL)
			code.back().replace_opcode(bi::GETLOCALLOAD);
		else
			addcmd(bi::LOAD);
	}
	void add_STORE_optimize(bool add_STORED=false)
	{
		if(add_STORED)
			addcmd(bi::STORED, -1);
		else if(code.size()>=1 && code.back().get_opcode()==bi::LOCALADDR)
			{ code.back().replace_opcode(bi::SETLOCAL); code.back().sp--; }
		else if(code.size()>=1 && code.back().fkt==Command::get_prev_var_addr)
			{ code.back().fkt=Command::set_prev_var; code.back().sp--; }
		else if(code.size()>=1 && code.back().fkt==Command::get_prev_var)
			{ code.back().fkt=Command::get_prev_var_and_store; code.back().sp--; }
		else if(code.size()>=1 && code.back().get_opcode()==bi::GETLOCAL)
			{ code.back().replace_opcode(bi::GETLOCALSTORE); code.back().sp--; }
		else
			addcmd(bi::STORE, -1);
		code_has_effect=true;
	}

	expr shortest_type_if_constant() const
	{	// Wenn *this eine Integer-Konstante ist, wird auf den kleinstmglichen Typ gecastet.
		// Wenn *this eine signed-Konstante ist, die aber nicht negativ ist, wird nach unsigned gecastet.
		set_exported();
		if(is_integer_constant())
			return (Value::s_is_unsigned_integer(type_nr) || ((int64)get_this_integer_constant())>=0) ?
				get_const_expr((uint64)get_this_integer_constant()) : get_const_expr((int64)get_this_integer_constant());
		else
			return *this;
	}

	expr shortest_type_if_const_keep_signed() const
	{	// Wenn *this eine Integer-Konstante ist, wird auf den kleinstmglichen Typ gecastet.
		// Wenn *this eine signed-Konstante ist, die aber nicht negativ ist, wird nach unsigned gecastet.
		set_exported();
		if(is_integer_constant())
			return Value::s_is_unsigned_integer(type_nr) ? get_const_expr((uint64)get_this_integer_constant()) :
				get_const_expr((int64)get_this_integer_constant(), true);
		else
			return *this;
	}
};

//void Code::add(const expr& ex, size_t add_sp) { ex.exported=true; add(ex.code, add_sp); }
void Code::add(const expr& ex) { ex.exported=true; add(ex.code); }
void Code::addfront(const expr& ex) { ex.exported=true; addfront(ex.code); }
Code::Code(const expr& ex) : code(ex.code) { ex.exported=true; }
Code& Code::operator=(const expr& ex) { ex.exported=true; code=ex.code; return *this; }
void ProgramStructureElement::add(const expr& ex) { ex.set_exported(); add(ex.code); }

Command& Command::add_coded_integer15(int64 n, uint8 nr_bytes_if_escape, bool _signed)
{
	if(_signed)
	{
		if(((0!=(n & (((uint64)1)<<63)))==(0!=(n & (((uint64)1)<<62)))) &&
				2*(n>=0 ? n : (-n)) - (n<0 ? 1 : 0) < ((0==nr_bytes_if_escape) ? 15-8 : 15))
			addnibble((uint8)(2*(n>=0 ? n : (-n)) - (n<0 ? 1 : 0)));
		else
		{
			if(0==nr_bytes_if_escape)
			{
				uint8 usednibbles=(expr::get_nr_of_used_bits_template<int64>((int64)n, true)+3)/4;
				if(usednibbles>8) usednibbles=16;
				addnibble((16==usednibbles) ? 15 : (usednibbles+16-10));
				n=(2*(n>=0 ? n : (-n)) - (n<0 ? 1 : 0));
				while(usednibbles)
					addnibble((n>>(4*(--usednibbles)))&15);
			}
			else
			{
				addnibble(15);
				if(2==nr_bytes_if_escape)
					add16((uint16)(2*(n>=0 ? n : (-n)) - (n<0 ? 1 : 0)));
				else
					add32((uint32)(2*(n>=0 ? n : (-n)) - (n<0 ? 1 : 0)));
			}
		}
	}
	else
	{
		if(((uint64)n) < ((0==nr_bytes_if_escape) ? 15-8 : 15))
			addnibble((uint8)n);
		else
		{
			if(0==nr_bytes_if_escape)
			{
				uint8 usednibbles=(expr::get_nr_of_used_bits_template<uint64>((uint64)n, true)+3)/4;
				if(usednibbles>8) usednibbles=16;
				addnibble((16==usednibbles) ? 15 : (usednibbles+16-10));
				while(usednibbles)
					addnibble((n>>(4*(--usednibbles)))&15);
			}
			else
			{
				addnibble(15);
				if(2==nr_bytes_if_escape)
					add16((uint16)n);
				else
					add32((uint32)n);
			}
		}
	}
	return *this;
}

Command& Command::add_coded_integer255(int64 n, uint8 nr_bytes_if_escape, bool _signed)
{
	if(_signed)
	{
		if(((0!=(n & (((uint64)1)<<63)))==(0!=(n & (((uint64)1)<<62)))) &&
				2*(n>=0 ? n : (-n)) - (n<0 ? 1 : 0) < ((0==nr_bytes_if_escape) ? 255-8 : 255))
			add8((uint8)(2*(n>=0 ? n : (-n)) - (n<0 ? 1 : 0)));
		else
		{
			if(0==nr_bytes_if_escape)
			{
				uint8 usednibbles=(expr::get_nr_of_used_bits_template<int64>((int64)n, true)+3)/4;
				if(usednibbles>8) usednibbles=16;
				addnibble((16==usednibbles) ? 255 : (usednibbles+256-10));
				n=(2*(n>=0 ? n : (-n)) - (n<0 ? 1 : 0));
				while(usednibbles)
					addnibble((n>>(4*(--usednibbles)))&15);
			}
			else
			{
				add8(255);
				if(2==nr_bytes_if_escape)
					add16((uint16)(2*(n>=0 ? n : (-n)) - (n<0 ? 1 : 0)));
				else
					add32((uint32)(2*(n>=0 ? n : (-n)) - (n<0 ? 1 : 0)));
			}
		}
	}
	else
	{
		if(((uint64)n) < ((0==nr_bytes_if_escape) ? 255-8 : 255))
			add8((uint8)n);
		else
		{
			if(0==nr_bytes_if_escape)
			{
				uint8 usednibbles=(expr::get_nr_of_used_bits_template<uint64>((uint64)n, true)+3)/4;
				if(usednibbles>8) usednibbles=16;
				addnibble((16==usednibbles) ? 255 : (usednibbles+256-10));
				while(usednibbles)
					addnibble((n>>(4*(--usednibbles)))&15);
			}
			else
			{
				add8(255);
				if(2==nr_bytes_if_escape)
					add16((uint16)n);
				else
					add32((uint32)n);
			}
		}
	}
	return *this;
}

struct Statement : public expr
{
	Statement(const expr& ex) : expr(ex)
		{ ex.set_exported(); }
	~Statement()
	{
		if(!exported)
			ExprObjectDestructionHandler<Statement>(*this);
	}
};

struct GeneralExpr;
struct PointerExpr;
struct param;
struct Variable;
struct Constant;
struct Type;
struct LogExpr;
struct x_And_y;
struct x_Or_y;
struct BRO_x_And_y_BRC_And_z;
struct BRO_x_Or_y_BRC_Or_z;
struct BRO_x_And_y_BRC_Or_z;
struct x_And_BRO_y_Or_z_BRC;
struct BRO_x_Or_y_BRC_And_z;
struct x_Or_BRO_y_And_z_BRC;
struct PointerType;
struct LocalPlusPlus;
struct LocalMinusMinus;
struct ParamPlusPlus;
struct ParamMinusMinus;
struct VariablePlusPlus;
struct VariableMinusMinus;
struct DrefPointerPlusPlus;
struct DrefPointerMinusMinus;

struct CollectGeneralType
{	// Diese Klasse besitzt Konstruktoren mit jeweils einem Argument von jedem anderen Ausdruck-Typ und Variablen-Typ
	// (expr, PointerExpr, GeneralExpr, var, local, param). Aber es werden absichtlich keine impliziten Konversionen zu anderen
	// Typen definiert (z.B. wird kein operator expr() definiert).
private: void operator=(const CollectGeneralType& gent) { } // Zugriff verhindern
public:
	const GeneralExprType *ex;
	uint64 cval;
	real_type creal;
	bool is_basicconst;
	bool is_expr() const { set_exported(); return 0!=dynamic_cast<const expr*>(ex); }
	bool is_expr_type() const
		{ set_exported(); return ex->is_expr_type(); }
	bool is_PointerExpr() const;
	bool is_pointer_type() const;
	~CollectGeneralType()
	{
		if(0!=ex) delete ex; // Ggf wird hier Bytecode erzeugt !
	}
	CollectGeneralType(const GeneralExprType& genex)
	{	// Kopier-Konstruktor von ****** expr *****, ***** LogExpr *****, ***** PointerExpr *****, ...
		genex.vset_exported();
		ex=(&genex)->vcopy(); // Durch genex.vcopy() wrde NICHT vcopy() einer von GeneralExprType abgeleiteten Klasse aufgerufen.
		// printf("CollectGeneralType::CollectGeneralType   %s\n", get_type_string().c_str());
		cval=0; creal=0;
		if(true==(is_basicconst = 0!=dynamic_cast<const expr*>(&genex) && static_cast<const expr*>(&genex)->is_integer_constant()))
			cval=static_cast<const expr*>(&genex)->get_this_integer_constant();
		set_not_exported();
	}
	CollectGeneralType(const GeneralExpr& genex);
	CollectGeneralType(const CollectGeneralType& genex)
		{ genex.set_exported(); ex=genex.ex->vcopy(); cval=genex.cval; creal=genex.creal; is_basicconst=genex.is_basicconst; }
	const CollectGeneralType& set_exported() const
	{
		ex->vset_exported();
		return *this;
	}
	void set_not_exported() const
	{
		ex->vset_not_exported();
	}
	expr to_expr() const
	{
		set_exported();
		return ex->to_expr();
	}
	PointerExpr to_PointerExpr(const PointerType* requested_pointer_type=0) const;
	nstring get_type_string() const;
	const Type *get_type() const;

	CollectGeneralType(const local& x);
	CollectGeneralType(const param& x);
	CollectGeneralType(const Variable& x);
	CollectGeneralType(const Constant& x);

	template<typename T> CollectGeneralType(const T *x)
		{ report_error("Cannot use a C pointer other than 'const char *'. Please use the type POINTER instead."); }
	CollectGeneralType(const char *x);
	CollectGeneralType(bool x) { ex=new expr(x); cval=x; creal=0; is_basicconst=true; }
	CollectGeneralType(signed char x) { ex=new expr(x); cval=x; creal=0; is_basicconst=true; }
	CollectGeneralType(unsigned char x) { ex=new expr(x); cval=x; creal=0; is_basicconst=true; }
	CollectGeneralType(short x) { ex=new expr(x); cval=x; creal=0; is_basicconst=true; }
	CollectGeneralType(unsigned short x) { ex=new expr(x); cval=x; creal=0; is_basicconst=true; }
	CollectGeneralType(int x) { ex=new expr(x); cval=x; creal=0; is_basicconst=true; }
	CollectGeneralType(unsigned int x) { ex=new expr(x); cval=x; creal=0; is_basicconst=true; }
	CollectGeneralType(long x) { ex=new expr(x); cval=x; creal=0; is_basicconst=true; }
	CollectGeneralType(unsigned long x) { ex=new expr(x); cval=x; creal=0; is_basicconst=true; }
	CollectGeneralType(long long x) { ex=new expr(x); cval=x; creal=0; is_basicconst=true; }
	CollectGeneralType(unsigned long long x) { ex=new expr(x); cval=x; creal=0; is_basicconst=true; }
	CollectGeneralType(float x) { ex=new expr(creal=(real_type)x); cval=0; is_basicconst=true; }
	CollectGeneralType(double x) { ex=new expr(creal=(real_type)x); cval=0; is_basicconst=true; }

	CollectGeneralType(const LogExpr& x);
	CollectGeneralType(const x_And_y& x);
	CollectGeneralType(const x_Or_y& x);
	CollectGeneralType(const BRO_x_And_y_BRC_And_z& x);
	CollectGeneralType(const BRO_x_Or_y_BRC_Or_z& x);
	CollectGeneralType(const BRO_x_And_y_BRC_Or_z& x);
	CollectGeneralType(const x_And_BRO_y_Or_z_BRC& x);
	CollectGeneralType(const BRO_x_Or_y_BRC_And_z& x);
	CollectGeneralType(const x_Or_BRO_y_And_z_BRC& x);

	CollectGeneralType(const LocalPlusPlus& x);
	CollectGeneralType(const LocalMinusMinus& x);
	CollectGeneralType(const ParamPlusPlus& x);
	CollectGeneralType(const ParamMinusMinus& x);
	CollectGeneralType(const VariablePlusPlus& x);
	CollectGeneralType(const VariableMinusMinus& x);
	CollectGeneralType(const DrefPointerPlusPlus& x);
	CollectGeneralType(const DrefPointerMinusMinus& x);
};

GeneralExpr operator+(const CollectGeneralType& x, const CollectGeneralType& y);
GeneralExpr operator-(const CollectGeneralType& x, const CollectGeneralType& y);

expr operator*(const CollectGeneralType& x, const CollectGeneralType& y);
expr operator/(const CollectGeneralType& x, const CollectGeneralType& y);
expr operator%(const CollectGeneralType& x, const CollectGeneralType& y);
expr operator|(const CollectGeneralType& x, const CollectGeneralType& y);
expr operator&(const CollectGeneralType& x, const CollectGeneralType& y);
expr operator^(const CollectGeneralType& x, const CollectGeneralType& y);
expr operator<<(const CollectGeneralType& x, const CollectGeneralType& y);
expr operator>>(const CollectGeneralType& x, const CollectGeneralType& y);
CollectGeneralType operator+(const CollectGeneralType& x);
expr operator-(const CollectGeneralType& x);

/*expr gettype(const expr& x)
{
	x.exported=true;
	expr ex;
	if(Value::t_nil==x.get_type())
		report_error("gettype cannot be applied to type t_nil");
	if(Value::t_unknown_type!=x.get_type())
		ex.add(expr::get_const_expr((uint32)(x.get_type())));
	else
	{
		ex.add(x);
		ex.add(Command(bi::GETTYPE));
		ex.set_type(Value::t_u8;
	}
	return ex;
}*/

expr branch_if(Command::uid_type branch_target_id, bool if_true, bool use_BRANDOR=false,
			   Command::uid_type gen_uid_of_last_command=0, bool negate_condition_bevore_use=false)
{	// Wenn 'distance' nicht negativ ist, so muss 'distance' die Distanz von der PC-Position NACH dem Branch-Befehl
	// bis zur Position sein, an der das Programm bei erfolgtem Sprung fortgesetzt werden soll.
	// Wenn 'distance' jedoch negativ ist, so muss 'distance' die Distanz von der PC-Position,
	// an der der Branch-Befehl eingefgt wird, bis zur Position sein, an der das Programm bei
	// erfolgtem Sprung fortgesetzt werden soll (es wird dann also rckwrts gesprungen).
	uint8 brcode;
	if(if_true)
		brcode=use_BRANDOR ? (negate_condition_bevore_use ? bi::BRORNOT : bi::BROR) : (negate_condition_bevore_use ? bi::BRF: bi::BRT);
	else
		brcode=use_BRANDOR ? (negate_condition_bevore_use ? bi::BRANDNOT : bi::BRAND) : (negate_condition_bevore_use ? bi::BRT: bi::BRF);
	Command cmd(brcode, Command::NewUID(gen_uid_of_last_command));
	cmd.branch1_id=branch_target_id;
	cmd.sp=-1;
	return expr(cmd).set_type(Value::t_nil).set_code_has_effect();
}

expr branch_if_false(Command::uid_type branch_target_id, bool use_BRANDOR=false, Command::uid_type gen_uid_of_last_command=0,
					 bool negate_condition_bevore_use=false)
{
	return branch_if(branch_target_id, false, use_BRANDOR, gen_uid_of_last_command, negate_condition_bevore_use);
}

expr branch_if_true(Command::uid_type branch_target_id, bool use_BRANDOR=false, Command::uid_type gen_uid_of_last_command=0,
					bool negate_condition_bevore_use=false)
{
	return branch_if(branch_target_id, true, use_BRANDOR, gen_uid_of_last_command, negate_condition_bevore_use);
}

expr dowhile_code(const expr& cond, const expr& repeat_code, bool if_not=false, Command::uid_type uid_of_first_loop_command=0,
				  Command::uid_type gen_uid_of_last_command=0)
{
	cond.exported=true; repeat_code.exported=true;
	expr ex(repeat_code);
	cond.set_code_has_effect();
	ex.add(cond);
	ex.add(branch_if(0==uid_of_first_loop_command ? ex.get_uid_of_first_command() : uid_of_first_loop_command, !if_not,
		false, gen_uid_of_last_command));
	ex.set_exported();
	return ex.set_type(Value::t_nil);
}

expr dowhile_code(const expr& cond, const Code& repeat_code, bool if_not=false, Command::uid_type uid_of_first_loop_command=0,
				  Command::uid_type gen_uid_of_last_command=0)
{
	cond.exported=true;
	expr ex;
	ex.add(repeat_code);
	cond.set_code_has_effect();
	ex.add(cond);
	ex.add(branch_if(0==uid_of_first_loop_command ? ex.get_uid_of_first_command() : uid_of_first_loop_command, !if_not,
		false, gen_uid_of_last_command));
	ex.set_exported();
	return ex.set_type(Value::t_nil);
}

expr if_code(const expr& cond, const expr& code_if_true, bool add_pop_to_code_if_true=false, bool if_not=false,
			 bool while_loop=false, Command::uid_type uid_of_first_while_loop_command=0,
			 Command::uid_type uid_of_generated_label_after_last_command=0)
{
	cond.exported=true; code_if_true.exported=true;
	expr ex;
	cond.set_code_has_effect();
	ex.add(cond);
	expr cd(code_if_true);
	if(add_pop_to_code_if_true)
		cd.addcmd(bi::POP, -1);
	if(while_loop)
	{
		Command cmd=Command(if_not ? bi::IFNOTELSE : bi::IFELSE).set_sp(1, 1); // IFELSE erzeugt hier die Funktionalitt einer while-Schleife
		cmd.branch1_id=cd.get_uid_after_last_command();
		cmd.branch2_id=(uid_of_first_while_loop_command==0 ? ex.get_uid_of_first_command() : uid_of_first_while_loop_command);
		ex.add_effective(cmd);
	}
	else
		ex.add(branch_if(cd.get_uid_after_last_command(), if_not));
	ex.add(cd);
	if(while_loop)
	{
		ex.code.back().sp-=2;
		--ex.code.back().spie;
	}
	if(0!=uid_of_generated_label_after_last_command)
		ex.add_effective(Command::new_uid_label(uid_of_generated_label_after_last_command));
	return ex.set_type(Value::t_nil);
}

expr if_code(const expr& cond, const Code& code_if_true, bool add_pop_to_code_if_true=false, bool if_not=false,
			 bool while_loop=false, Command::uid_type uid_of_first_while_loop_command=0,
			 Command::uid_type uid_of_generated_label_after_last_command=0)
{
	cond.exported=true;
	expr ex;
	cond.set_code_has_effect();
	ex.add(cond);
	Code cd(code_if_true);
	if(add_pop_to_code_if_true)
		cd.addcmd(bi::POP, -1);
	if(while_loop)
	{
		Command cmd=Command(if_not ? bi::IFNOTELSE : bi::IFELSE).set_sp(1, 1); // IFELSE erzeugt hier die Funktionalitt einer while-Schleife
		cmd.branch1_id=cd.get_uid_after_last_command();
		cmd.branch2_id=(uid_of_first_while_loop_command==0 ? ex.get_uid_of_first_command() : uid_of_first_while_loop_command);
		ex.add_effective(cmd);
	}
	else
		ex.add(branch_if(cd.get_uid_after_last_command(), if_not));
	ex.add(cd);
	if(while_loop)
	{
		ex.code.back().sp-=2;
		--ex.code.back().spie;
	}
	if(0!=uid_of_generated_label_after_last_command)
		ex.add_effective(Command::new_uid_label(uid_of_generated_label_after_last_command));
	return ex.set_type(Value::t_nil);
}

expr if_else_code(const expr& cond, const expr& code_if_true, const expr& code_if_false,
				  bool add_pop_to_code_if_true=false, bool add_pop_to_code_if_false=false, bool if_not=false)
{
	cond.exported=true; code_if_true.exported=true;
	expr cd(code_if_true), cd2(code_if_false);
	if(add_pop_to_code_if_true)
		cd.addcmd(bi::POP, -1);
	if(add_pop_to_code_if_false)
		cd2.addcmd(bi::POP, -1);
	expr ex;
	cond.set_code_has_effect();
	ex.add(cond);
	Command cmd=Command(if_not ? bi::IFNOTELSE : bi::IFELSE).set_sp(1, 1); // Die zwei durch IFELSE auf dem Stack abgelegten Werte werden unten bercksichtigt.
	cmd.branch1_id=cd.get_uid_after_last_command();
	cmd.branch2_id=cd2.get_uid_after_last_command();
	ex.add_effective(cmd);
	ex.add(cd);
	ex.code.back().sp-=2;
	--ex.code.back().spie;
	ex.add(cd2);
	return ex.set_type(Value::t_nil);
}

expr if_else_code(const expr& cond, const Code& code_if_true, const Code& code_if_false,
				  bool add_pop_to_code_if_true=false, bool add_pop_to_code_if_false=false, bool if_not=false)
{
	cond.exported=true;
	Code cd(code_if_true), cd2(code_if_false);
	if(add_pop_to_code_if_true)
		cd.addcmd(bi::POP, -1);
	if(add_pop_to_code_if_false)
		cd2.addcmd(bi::POP, -1);
	expr ex;
	cond.set_code_has_effect();
	ex.add(cond);
	Command cmd=Command(if_not ? bi::IFNOTELSE : bi::IFELSE).set_sp(1, 1); // Die zwei durch IFELSE auf dem Stack abgelegten Werte werden unten bercksichtigt.
	cmd.branch1_id=cd.get_uid_after_last_command();
	cmd.branch2_id=cd2.get_uid_after_last_command();
	ex.add_effective(cmd);
	ex.add(cd);
	ex.code.back().sp-=2;
	--ex.code.back().spie;
	ex.add(cd2);
	return ex.set_type(Value::t_nil);
}

struct LogExpr
{
private: void operator=(const LogExpr& s) { report_error("Assignment LogExpr::operator=() must not be called"); } // Zugriff verhindern. Wenn der Zugriff auf den
		 // Standard-Copykonstruktor erlaubt wre, wrde z.B. der sinnlose Ausdruck
		 // !expr(true) = !expr(true) ohne Fehlermeldung akzeptiert.
private: template<class RHS> void operator=(const RHS& expr) { report_error("Assignment to non l-value"); }
		 // Zugriff verhindern. Wenn der Zugriff auf den
		 // Standard-Copykonstruktor erlaubt wre, wrde z.B. der sinnlose Ausdruck
		 // (expr(true) && expr(true)) = x (wobei x ein Ausdruck ist) ohne Fehlermeldung akzeptiert.
		 // Da MS Visual C++ jedoch das 'private' ignoriert, wird zustzlich noch eine Ausnahme geworfen, um den Fehler anzuzeigen.

public:
	expr x;
	bool negated;
	mutable bool is_converted_to_bool_type;  // Diese Variable darf nur dann auf true gesetzt werden, wenn es toleriert werden
		// kann, dass to_expr() bzw. operator expr() fr den true-Fall einen Wert liefern, der zwar von Null verschieden
		// ist, aber mglicherweise nicht den Datentyp t_bool hat, sondern ggf. einen Integer- oder Gleitkommatyp.
		// Da jedoch ein LogExpr nur erzeugt wird, wenn im Programm eine logische Operation ausgefhrt wird, bedeutet dies,
		// dass is_converted_to_bool_type nur dann auf true gesetzt werden darf, wenn der von to_expr() oder von operator expr()
		// gelieferte Wert bei der nachfolgenden Verarbeitung nach bool konvertiert wird.
		// ======> WENN is_converted_to_bool_type==true ist, DANN IST x NICHT ZWANGSLUFIG VOM TYP bool. <=======
	void declare_cast_to_bool_not_needed() const { is_converted_to_bool_type=true; }
	LogExpr() { negated=false; is_converted_to_bool_type=false; }
	LogExpr(const LogExpr& ex) : x(ex.x), negated(ex.negated), is_converted_to_bool_type(ex.is_converted_to_bool_type)
		{ ex.x.exported=true; x.exported=false; }
	struct dont_need_cast_to_bool { };
	LogExpr(const LogExpr& ex, dont_need_cast_to_bool dummy)
			: x(ex.x), negated(ex.negated), is_converted_to_bool_type(true)
		{ ex.x.exported=true; x.exported=false; }
	LogExpr(const expr& ex, bool __negated=false) : x(ex), negated(__negated)
		{ ex.exported=true; is_converted_to_bool_type=(Value::t_bool==ex.get_type()); x.exported=false; }
	~LogExpr()
	{
		if(!x.exported)
			ExprObjectDestructionHandler<LogExpr>(*this);
	}
	bool is_expr_type() const { set_exported(); return true; }
	expr to_expr() const
	{
		x.exported=true;
		if(!negated)
		{
			if(is_converted_to_bool_type)
				return x; // Hier wird NICHT x.set_type(Value::t_bool gesetzt, denn is_converted_to_bool_type kann auch
					// von extern auf true gesetzt werden, ohne dass x tatschlich bool ist, nmlich um anzuzeigen, dass
					// eine Konversion nach bool nicht notwendig ist (bei if-Abfragen).
			expr ex(x);
			ex.add_effective(Command(bi::CAST).addnibble(Value::t_bool));
			ex.set_exported();
			ex.set_type(Value::t_bool);
			return ex;
		}
		else
		{
			expr ex(x);
			ex.addcmd(bi::NOT); // durch den NOT-Befehl wird automatisch nach bool konvertiert.
			ex.set_exported();
			ex.set_type(Value::t_bool);
			return ex;
		}
	}
	const LogExpr& set_exported() const { x.exported=true; return *this; }
	void vset_exported() const { set_exported(); }
	void vset_not_exported() const { set_not_exported(); }
	void set_not_exported() const { x.exported=false; }

	LogExpr(const PointerExpr& p);
	LogExpr(const GeneralExpr& genex);
	LogExpr(const VariablePlusPlus &v);
	LogExpr(const VariableMinusMinus &v);
	LogExpr(const LocalPlusPlus &v);
	LogExpr(const LocalMinusMinus &v);
	LogExpr(const ParamPlusPlus& x);
	LogExpr(const ParamMinusMinus& x);
	LogExpr(const DrefPointerPlusPlus& x);
	LogExpr(const DrefPointerMinusMinus& x);
	LogExpr(const local& x);
	LogExpr(const param& x);
	LogExpr(const Variable& x);
	LogExpr(const Constant& x);

	LogExpr(bool y) : x(expr(y)), negated(false), is_converted_to_bool_type(true) { }
	LogExpr(signed char y) : x(expr(y)), negated(false), is_converted_to_bool_type(false) { }
	LogExpr(unsigned char y) : x(expr(y)), negated(false), is_converted_to_bool_type(false) { }
	LogExpr(short y) : x(expr(y)), negated(false), is_converted_to_bool_type(false) { }
	LogExpr(unsigned short y) : x(expr(y)), negated(false), is_converted_to_bool_type(false) { }
	LogExpr(int y) : x(expr(y)), negated(false), is_converted_to_bool_type(false) { }
	LogExpr(unsigned int y) : x(expr(y)), negated(false), is_converted_to_bool_type(false) { }
	LogExpr(long y) : x(expr(y)), negated(false), is_converted_to_bool_type(false) { }
	LogExpr(unsigned long y) : x(expr(y)), negated(false), is_converted_to_bool_type(false) { }
	LogExpr(long long y) : x(expr(y)), negated(false), is_converted_to_bool_type(false) { }
	LogExpr(unsigned long long y) : x(expr(y)), negated(false), is_converted_to_bool_type(false) { }
	LogExpr(float y) : x(expr((real_type)y)), negated(false), is_converted_to_bool_type(false) { }
	LogExpr(double y) : x(expr((real_type)y)), negated(false), is_converted_to_bool_type(false) { }

	template<typename T> LogExpr(const T *x)
		{ report_error("Cannot use a C pointer other than 'const char *'. Please use the type POINTER instead."); }
	LogExpr(const char *y);
};

expr dowhile_code(const LogExpr& cond, const expr& repeat_code, bool if_not=false, Command::uid_type uid_of_first_loop_command=0,
				  Command::uid_type gen_uid_of_last_command=0)
{
	cond.set_exported(); repeat_code.exported=true;
	expr ex;
	ex.add(repeat_code);
	cond.x.set_code_has_effect();
	ex.add(cond.x);
	ex.add(branch_if(0==uid_of_first_loop_command ? ex.get_uid_of_first_command() : uid_of_first_loop_command,
			(!if_not && !cond.negated) || (if_not && cond.negated), false, gen_uid_of_last_command));
	ex.set_exported();
	return ex;
}

expr dowhile_code(const LogExpr& cond, const Code& repeat_code, bool if_not=false, Command::uid_type uid_of_first_loop_command=0,
				  Command::uid_type gen_uid_of_last_command=0)
{
	cond.set_exported();
	expr ex;
	ex.add(repeat_code);
	cond.x.set_code_has_effect();
	ex.add(cond.x);
	ex.add(branch_if(0==uid_of_first_loop_command ? ex.get_uid_of_first_command() : uid_of_first_loop_command,
			(!if_not && !cond.negated) || (if_not && cond.negated), false, gen_uid_of_last_command));
	ex.set_exported();
	return ex;
}

expr if_code(const LogExpr& cond, const expr& code_if_true, bool add_pop_to_code_if_true=false, bool if_not=false,
			 bool while_loop=false, Command::uid_type uid_of_first_while_loop_command=0,
			 Command::uid_type uid_of_generated_label_after_last_command=0)
{
	cond.set_exported();
	code_if_true.exported=true;
	expr ex;
	cond.x.set_code_has_effect();
	ex.add(cond.x);
	expr cd(code_if_true);
	if(add_pop_to_code_if_true)
		cd.addcmd(bi::POP, -1);
	if(while_loop)
	{
		Command cmd=Command(((if_not && !cond.negated) || (!if_not && cond.negated)) ? bi::IFNOTELSE : bi::IFELSE).set_sp(1, 1);
		cmd.branch1_id=cd.get_uid_after_last_command();
		cmd.branch2_id=(uid_of_first_while_loop_command==0 ? ex.get_uid_of_first_command() : uid_of_first_while_loop_command);
		ex.add_effective(cmd);
	}
	else
		ex.add_effective(branch_if(cd.get_uid_after_last_command(), (if_not && !cond.negated) || (!if_not && cond.negated)));
	ex.add(cd);
	if(while_loop)
	{
		ex.code.back().sp-=2;
		--ex.code.back().spie;
	}
	if(0!=uid_of_generated_label_after_last_command)
		ex.add_effective(Command::new_uid_label(uid_of_generated_label_after_last_command));
	return ex;
}

expr if_code(const LogExpr& cond, const Code& code_if_true, bool add_pop_to_code_if_true=false, bool if_not=false,
			 bool while_loop=false, Command::uid_type uid_of_first_while_loop_command=0,
			 Command::uid_type uid_of_generated_label_after_last_command=0)
{
	cond.set_exported();
	expr ex;
	cond.x.set_code_has_effect();
	ex.add(cond.x);
	Code cd(code_if_true);
	if(add_pop_to_code_if_true)
		cd.addcmd(bi::POP, -1);
	if(while_loop)
	{
		Command cmd=Command(((if_not && !cond.negated) || (!if_not && cond.negated)) ? bi::IFNOTELSE : bi::IFELSE).set_sp(1, 1);
			// IFELSE erzeugt hier die Funktionalitt einer while-Schleife
		cmd.branch1_id=cd.get_uid_after_last_command();
		cmd.branch2_id=(uid_of_first_while_loop_command==0 ? ex.get_uid_of_first_command() : uid_of_first_while_loop_command);
		ex.add_effective(cmd);
	}
	else
		ex.add(branch_if(cd.get_uid_after_last_command(), (if_not && !cond.negated) || (!if_not && cond.negated)));
	ex.add(cd);
	if(while_loop)
	{
		ex.code.back().sp-=2;
		--ex.code.back().spie;
	}
	if(0!=uid_of_generated_label_after_last_command)
		ex.add_effective(Command::new_uid_label(uid_of_generated_label_after_last_command));
	return ex;
}

expr if_else_code(const LogExpr& cond, const expr& code_if_true, const expr& code_if_false,
				  bool add_pop_to_code_if_true=false, bool add_pop_to_code_if_false=false, bool if_not=false)
{
	cond.set_exported(); code_if_true.set_exported();
	expr cd(code_if_true), cd2(code_if_false);
	if(add_pop_to_code_if_true)
		cd.addcmd(bi::POP, -1);
	if(add_pop_to_code_if_false)
		cd2.addcmd(bi::POP, -1);
	expr ex;
	cond.x.set_code_has_effect();
	ex.add(cond.x);
	Command cmd=Command(((cond.negated && !if_not) || (!cond.negated && if_not)) ? bi::IFNOTELSE : bi::IFELSE).set_sp(1, 1);
	cmd.branch1_id=cd.get_uid_after_last_command();
	cmd.branch2_id=cd2.get_uid_after_last_command();
	ex.add_effective(cmd);
	ex.add(cd);
	ex.code.back().sp-=2;
	--ex.code.back().spie;
	ex.add(cd2);
	return ex;
}

expr if_else_code(const LogExpr& cond, const Code& code_if_true, const Code& code_if_false,
				  bool add_pop_to_code_if_true=false, bool add_pop_to_code_if_false=false, bool if_not=false)
{
	cond.set_exported();
	Code cd(code_if_true), cd2(code_if_false);
	if(add_pop_to_code_if_true)
		cd.addcmd(bi::POP, -1);
	if(add_pop_to_code_if_false)
		cd2.addcmd(bi::POP, -1);
	expr ex;
	cond.x.set_code_has_effect();
	ex.add(cond.x);
	Command cmd=Command(((cond.negated && !if_not) || (!cond.negated && if_not)) ? bi::IFNOTELSE : bi::IFELSE).set_sp(1, 1);
	cmd.branch1_id=cd.get_uid_after_last_command();
	cmd.branch2_id=cd2.get_uid_after_last_command();
	ex.add_effective(cmd);
	ex.add(cd);
	ex.code.back().sp-=2;
	--ex.code.back().spie;
	ex.add(cd2);
	return ex;
}

struct x_And_y
{	// x && y
private: void operator=(const x_And_y& s) { report_error("Assignment x_And_y::operator=() must not be called"); } // Zugriff verhindern. Wenn der Zugriff auf den Standard-Copykonstruktor
		 // erlaubt wre, wrde z.B. der sinnlose Ausdruck (expr(true) && expr(true)) = (expr(true) && expr(true)) ohne
		 // Fehlermeldung akzeptiert.
private: template<class RHS> void operator=(const RHS& expr) { report_error("Assignment to non l-value"); }
		 // Zugriff verhindern. Wenn der Zugriff auf den
		 // Standard-Copykonstruktor erlaubt wre, wrde z.B. der sinnlose Ausdruck
		 // (expr(true) && expr(true)) = x (wobei x ein Ausdruck ist) ohne Fehlermeldung akzeptiert.
		 // Da MS Visual C++ jedoch das 'private' ignoriert, wird zustzlich noch eine Ausnahme geworfen, um den Fehler anzuzeigen.
public:
	LogExpr x, y;
	mutable bool is_converted_to_bool_type; // Diese Variable wird intern auf true gesetzt, wenn bereits eine Konvertierung auf
		// bool vorgenommen wurde oder wenn x.is_converted_to_bool_type==true und y.is_converted_to_bool_type==true ist.
		// Diese Variable darf von extern nur dann auf true gesetzt werden, wenn es toleriert werden
		// kann, dass to_expr() oder operator LogExpr() fr den true-Fall einen Wert liefern, der zwar von Null verschieden
		// ist, aber mglicherweise nicht den Datentyp t_bool hat, sondern ggf. einen Integer- oder Gleitkommatyp.
		// Da jedoch ein x_Or_y nur erzeugt wird, wenn im Programm eine logische Operation ausgefhrt wird, bedeutet dies,
		// dass no_cast_to_bool_required nur dann auf true gesetzt werden darf, wenn der von to_expr() oder von operator LogExpr()
		// gelieferte Wert bei der nachfolgenden Verarbeitung nach bool konvertiert wird.*/
		// ======> WENN is_converted_to_bool_type==true ist, DANN SIND x.x und y.x NICHT ZWANGSLUFIG VOM TYP bool. <=======
	void declare_cast_to_bool_not_needed() const { is_converted_to_bool_type=true; }
	x_And_y(const LogExpr& ex1, const LogExpr& ex2) : x(ex1), y(ex2)
	{
		ex1.set_exported(); ex2.set_exported();
		set_not_exported();
		is_converted_to_bool_type=ex1.is_converted_to_bool_type && ex2.is_converted_to_bool_type;
	}
	~x_And_y()
	{
		if(!x.x.exported || !y.x.exported)
	    {
		    LogExpr le(*this);
		    le.set_exported();
			ExprObjectDestructionHandler<LogExpr>(le);
        }
	}
	expr to_expr() const
	{
		set_exported();
		return (operator LogExpr()).to_expr().set_type(Value::t_bool);
	}
	bool is_expr_type() const { set_exported(); return true; }
	operator LogExpr() const
	{
		set_exported();
		LogExpr y2(y); y2.set_exported();
		y2.x.assignment_to_locals.clear(); // Der Ausdruck y wird vom Bytecode-Interpreter nur dann ausgewertet, wenn der
			// Ausdruck x true ergibt. Da zur Compilierzeit nicht festgestellt werden kann, ob dies der Fall ist, wird
			// pauschal angenommen, dass y nicht ausgewertet wird, und die Liste y2.assignment_to_locals muss gelscht werden
			// (eine irrtmliche Annahme, dass y ausgewertet wird, wrde zu einer fehlenden Initialisierung der Typeninformation
			// der lokalen Variablen fhren, auf die in der Liste y2.assignment_to_locals verwiesen wird, und somit zu einem
			// fehlerhaften Verhalten des Programms).
		if(y2.negated)
		{
			LogExpr ex(y2.x); // nicht ex.x=y2 oder ex.add(y2), denn hier ist y2.negated==true, und es soll hier KEIN NOT erzeugt werden.
			ex.is_converted_to_bool_type=y2.is_converted_to_bool_type;
			ex.x.addfront(branch_if_true(ex.x.get_uid_after_last_command(), true, 0, !x.negated)); // true ==> BROR verwenden
			ex.x.addfront(x.x);	// x braucht kein bool zu sein, weil die branch-Befehle automatisch nach bool konvertieren.
			ex.is_converted_to_bool_type=is_converted_to_bool_type || y2.is_converted_to_bool_type;
			ex.negated=true;
			return ex;
		}
		else
		{
			LogExpr ex;
			ex.x.assign(LogExpr(y2, LogExpr::dont_need_cast_to_bool()).to_expr());
				// Nicht LogExpr ex(y2), denn ex.negated muss false sein. Falls y2.negated==true ist, wird bei
				// der mit der Zuweisung ex.x=y2 verbundenen Konvertierung von LogExpr nach expr automatisch ein NOT eingefgt.
				// Wenn y2.is_converted_to_bool_type==false und this->is_converted_to_bool_type==false ist
				// und noch kein NOT eingefgt wurde, wird hier zustzlich ein CAST-Opcode nach bool eingefgt.*/
			ex.x.addfront(branch_if_false(ex.x.get_uid_after_last_command(), true, 0, x.negated)); // true ==> BROR verwenden
			ex.x.addfront(x.x);
			ex.is_converted_to_bool_type=is_converted_to_bool_type || y2.is_converted_to_bool_type;
			return ex;
		}
	}
	void set_exported() const
	{
		x.set_exported();
		y.set_exported();
	}
	void vset_exported() const { set_exported(); }
	void vset_not_exported() const { set_not_exported(); }
	void set_not_exported() const
	{
		x.set_not_exported();
		y.set_not_exported();
	}
};

struct x_Or_y
{	// x || y
private: void operator=(const x_Or_y& s) { report_error("Assignment x_Or_y::operator=() must not be called"); } // Zugriff verhindern. Wenn der Zugriff auf den Standard-Copykonstruktor
		 // erlaubt wre, wrde z.B. der sinnlose Ausdruck (expr(true) || expr(true)) = (expr(true) || expr(true)) ohne
		 // Fehlermeldung akzeptiert.
private: template<class RHS> void operator=(const RHS& expr) { report_error("Assignment to non l-value"); }
		 // Zugriff verhindern. Wenn der Zugriff auf den
		 // Standard-Copykonstruktor erlaubt wre, wrde z.B. der sinnlose Ausdruck
		 // (expr(true) && expr(true)) = x (wobei x ein Ausdruck ist) ohne Fehlermeldung akzeptiert.
		 // Da MS Visual C++ jedoch das 'private' ignoriert, wird zustzlich noch eine Ausnahme geworfen, um den Fehler anzuzeigen.

public:
	LogExpr x, y;
	mutable bool is_converted_to_bool_type; // Diese Variable wird intern auf true gesetzt, wenn bereits eine Konvertierung auf
		// bool vorgenommen wurde oder wenn x.is_converted_to_bool_type==true und y.is_converted_to_bool_type==true ist.
		// Diese Variable darf von extern nur dann auf true gesetzt werden, wenn es toleriert werden
		// kann, dass to_expr() oder operator LogExpr() fr den true-Fall einen Wert liefern, der zwar von Null verschieden
		// ist, aber mglicherweise nicht den Datentyp t_bool hat, sondern ggf. einen Integer- oder Gleitkommatyp.
		// Da jedoch ein x_Or_y nur erzeugt wird, wenn im Programm eine logische Operation ausgefhrt wird, bedeutet dies,
		// dass no_cast_to_bool_required nur dann auf true gesetzt werden darf, wenn der von to_expr() oder von operator LogExpr()
		// gelieferte Wert bei der nachfolgenden Verarbeitung nach bool konvertiert wird.*/
		// ======> WENN is_converted_to_bool_type==true ist, DANN SIND x.x und y.x NICHT ZWANGSLUFIG VOM TYP bool. <=======
	void declare_cast_to_bool_not_needed() const { is_converted_to_bool_type=true; }
	x_Or_y(const LogExpr& ex1, const LogExpr& ex2) : x(ex1), y(ex2)
	{
		ex1.set_exported(); ex2.set_exported();
		set_not_exported();
		is_converted_to_bool_type=ex1.is_converted_to_bool_type && ex2.is_converted_to_bool_type;
	}
	~x_Or_y()
	{
		if(!x.x.exported || !y.x.exported)
	    {
		    LogExpr le(*this);
		    le.set_exported();
			ExprObjectDestructionHandler<LogExpr>(le);
        }
	}
	expr to_expr() const
	{
		set_exported();
		return (operator LogExpr()).to_expr().set_type(Value::t_bool);
	}
	bool is_expr_type() const { set_exported(); return true; }
	operator LogExpr() const
	{
		set_exported();
		LogExpr y2(y); y2.set_exported();
		y2.x.assignment_to_locals.clear(); // Der Ausdruck y wird vom Bytecode-Interpreter nur dann ausgewertet, wenn der
			// Ausdruck x false ergibt. Da zur Compilierzeit nicht festgestellt werden kann, ob dies der Fall ist, wird
			// pauschal angenommen, dass y nicht ausgewertet wird, und die Liste y2.assignment_to_locals muss gelscht werden
			// (eine irrtmliche Annahme, dass y ausgewertet wird, wrde zu einer fehlenden Initialisierung der Typeninformation
			// der lokalen Variablen fhren, auf die in der Liste y2.assignment_to_locals verwiesen wird, und somit zu einem
			// fehlerhaften Verhalten des Programms).
		if(y2.negated)
		{
			LogExpr ex(y2.x); // nicht ex.x=y2 oder ex.add(y2), denn hier ist y2.negated==true, und es soll hier KEIN NOT erzeugt werden.
			ex.is_converted_to_bool_type=y2.is_converted_to_bool_type;
			ex.x.addfront(branch_if_false(ex.x.get_uid_after_last_command(), true, 0, !x.negated)); // true ==> BRAND verwenden
			ex.x.addfront(x.x);	// x braucht kein bool zu sein, weil die branch-Befehle automatisch nach bool konvertieren.
			ex.is_converted_to_bool_type=is_converted_to_bool_type || y2.is_converted_to_bool_type;
			ex.negated=true;
			return ex;
		}
		else
		{
			LogExpr ex;
			ex.x.assign(LogExpr(y2, LogExpr::dont_need_cast_to_bool()).to_expr());
				// Nicht LogExpr ex(y2), denn ex.negated muss false sein. Falls y2.negated==true ist, wird bei
				// der mit der Zuweisung ex.x=y2 verbundenen Konvertierung von LogExpr nach expr automatisch ein NOT eingefgt.
				// Wenn y2.is_converted_to_bool_type==false und this->is_converted_to_bool_type==false ist
				// und noch kein NOT eingefgt wurde, wird hier zustzlich ein CAST-Opcode nach bool eingefgt.*/
			ex.x.addfront(branch_if_true(ex.x.get_uid_after_last_command(), true, 0, x.negated)); // true ==> BROR verwenden
			ex.x.addfront(x.x);
			ex.is_converted_to_bool_type=is_converted_to_bool_type || y2.is_converted_to_bool_type;
			return ex;
		}
	}
	void set_exported() const
	{
		x.set_exported();
		y.set_exported();
	}
	void vset_exported() const { set_exported(); }
	void vset_not_exported() const { set_not_exported(); }
	void set_not_exported() const
	{
		x.set_not_exported();
		y.set_not_exported();
	}
};

struct BRO_x_And_y_BRC_And_z
{	// (x && y) && z
private: void operator=(const BRO_x_And_y_BRC_And_z& s) { report_error("Assignment BRO_x_And_y_BRC_And_z::operator=() must not be called"); } // Zugriff verhindern
private: template<class RHS> void operator=(const RHS& expr) { report_error("Assignment to non l-value"); }
		 // Zugriff verhindern. Wenn der Zugriff auf den
		 // Standard-Copykonstruktor erlaubt wre, wrde z.B. der sinnlose Ausdruck
		 // (expr(true) && expr(true)) = x (wobei x ein Ausdruck ist) ohne Fehlermeldung akzeptiert.
		 // Da MS Visual C++ jedoch das 'private' ignoriert, wird zustzlich noch eine Ausnahme geworfen, um den Fehler anzuzeigen.

public:
	x_And_y xy;
	LogExpr z;
 	BRO_x_And_y_BRC_And_z(const x_And_y& ex1, const LogExpr& ex2) : xy(ex1), z(ex2)
	{
		ex1.set_exported(); ex2.set_exported();
		set_not_exported();
	}
	~BRO_x_And_y_BRC_And_z()
	{
		if(!xy.x.x.exported || !xy.y.x.exported || !z.x.exported)
	    {
		    LogExpr le(to_LogExpr());
		    le.set_exported();
			ExprObjectDestructionHandler<LogExpr>(le);
        }
	}
	void declare_cast_to_bool_not_needed() const { xy.declare_cast_to_bool_not_needed(); z.declare_cast_to_bool_not_needed(); }
	operator x_And_y() const
	{
		set_exported();
		return x_And_y(LogExpr(xy), z);
	}
	expr to_expr() const
	{
		set_exported();
		return x_And_y(LogExpr(xy), z).to_expr(); // Der Ausdruck x_And_y(xy.to_expr(), expr(z)).to_expr() wrde zwar
			// funktionieren, aber einen ineffizienteren Bytecode liefern.
	}
	bool is_expr_type() const { set_exported(); return true; }
	LogExpr to_LogExpr() const
	{
		set_exported();
		return LogExpr(x_And_y(LogExpr(xy), z)); // Der Ausdruck x_And_y(xy.to_expr(), expr(z)).to_expr() wrde zwar
			// funktionieren, aber einen ineffizienteren Bytecode liefern.
	}
	void set_exported() const
	{
		xy.set_exported();
		z.set_exported();
	}
	void vset_exported() const { set_exported(); }
	void vset_not_exported() const { set_not_exported(); }
	void set_not_exported() const
	{
		xy.set_not_exported();
		z.set_not_exported();
	}
};

struct BRO_x_Or_y_BRC_Or_z
{	// (x || y) || z
private: void operator=(const BRO_x_Or_y_BRC_Or_z& s) { report_error("Assignment BRO_x_Or_y_BRC_Or_z::operator=() must not be called"); } // Zugriff verhindern
private: template<class RHS> void operator=(const RHS& expr) { report_error("Assignment to non l-value"); }
		 // Zugriff verhindern. Wenn der Zugriff auf den
		 // Standard-Copykonstruktor erlaubt wre, wrde z.B. der sinnlose Ausdruck
		 // (expr(true) && expr(true)) = x (wobei x ein Ausdruck ist) ohne Fehlermeldung akzeptiert.
		 // Da MS Visual C++ jedoch das 'private' ignoriert, wird zustzlich noch eine Ausnahme geworfen, um den Fehler anzuzeigen.

public:
	x_Or_y xy;
	LogExpr z;
	BRO_x_Or_y_BRC_Or_z(const x_Or_y& ex1, const LogExpr& ex2) : xy(ex1), z(ex2)
	{
		ex1.set_exported(); ex2.set_exported();
		set_not_exported();
	}
	~BRO_x_Or_y_BRC_Or_z()
	{
		if(!xy.x.x.exported || !xy.y.x.exported || !z.x.exported)
	    {
		    LogExpr le(to_LogExpr());
		    le.set_exported();
			ExprObjectDestructionHandler<LogExpr>(le);
        }
	}
	void declare_cast_to_bool_not_needed() const { xy.declare_cast_to_bool_not_needed(); z.declare_cast_to_bool_not_needed(); }
	operator x_Or_y() const
	{
		set_exported();
		return x_Or_y(LogExpr(xy), z);
	}
	expr to_expr() const
	{
		set_exported();
		return x_Or_y(LogExpr(xy), z).to_expr(); // Der Ausdruck x_Or_y(xy.to_expr(), expr(z)).to_expr() wrde zwar
			// funktionieren, aber einen ineffizienteren Bytecode liefern.
	}
	bool is_expr_type() const { set_exported(); return true; }
	LogExpr to_LogExpr() const
	{
		set_exported();
		return LogExpr(x_Or_y(LogExpr(xy), z)); // Der Ausdruck x_Or_y(xy.to_expr(), expr(z)).to_expr() wrde zwar
			// funktionieren, aber einen ineffizienteren Bytecode liefern.
	}
	void set_exported() const
	{
		xy.set_exported();
		z.set_exported();
	}
	void vset_exported() const { set_exported(); }
	void vset_not_exported() const { set_not_exported(); }
	void set_not_exported() const
	{
		xy.set_not_exported();
		z.set_not_exported();
	}
};

struct BRO_x_And_y_BRC_Or_z
{	// (x && y) || z
private: void operator=(const BRO_x_And_y_BRC_Or_z& s) { report_error("Assignment BRO_x_And_y_BRC_Or_z::operator=() must not be called"); } // Zugriff verhindern
private: template<class RHS> void operator=(const RHS& expr) { report_error("Assignment to non l-value"); }
		 // Zugriff verhindern. Wenn der Zugriff auf den
		 // Standard-Copykonstruktor erlaubt wre, wrde z.B. der sinnlose Ausdruck
		 // (expr(true) && expr(true)) = x (wobei x ein Ausdruck ist) ohne Fehlermeldung akzeptiert.
		 // Da MS Visual C++ jedoch das 'private' ignoriert, wird zustzlich noch eine Ausnahme geworfen, um den Fehler anzuzeigen.

public:
	x_And_y xy;
	LogExpr z;
	BRO_x_And_y_BRC_Or_z(const x_And_y& ex1, const LogExpr& ex2) : xy(ex1), z(ex2)
	{
		ex1.set_exported(); ex2.set_exported();
		set_not_exported();
	}
	~BRO_x_And_y_BRC_Or_z()
	{
		if(!xy.x.x.exported || !xy.y.x.exported || !z.x.exported)
	    {
		    LogExpr le(to_LogExpr());
		    le.set_exported();
			ExprObjectDestructionHandler<LogExpr>(le);
        }
	}
	void declare_cast_to_bool_not_needed() const { xy.declare_cast_to_bool_not_needed(); z.declare_cast_to_bool_not_needed(); }
	operator x_Or_y() const
	{
		set_exported();
		return x_Or_y(LogExpr(xy), z);
	}
	expr to_expr() const
	{
		set_exported();
		return x_Or_y(LogExpr(xy), z).to_expr(); // Der Ausdruck x_Or_y(xy.to_expr(), expr(z)).to_expr() wrde zwar
			// funktionieren, aber einen ineffizienteren Bytecode liefern.
	}
	bool is_expr_type() const { set_exported(); return true; }
	LogExpr to_LogExpr() const
	{
		set_exported();
		return LogExpr(x_Or_y(LogExpr(xy), z)); // Der Ausdruck x_Or_y(xy.to_expr(), expr(z)).to_expr() wrde zwar
			// funktionieren, aber einen ineffizienteren Bytecode liefern.
	}
	void set_exported() const
	{
		xy.set_exported();
		z.set_exported();
	}
	void vset_exported() const { set_exported(); }
	void vset_not_exported() const { set_not_exported(); }
	void set_not_exported() const
	{
		xy.set_not_exported();
		z.set_not_exported();
	}
};

struct x_And_BRO_y_Or_z_BRC
{	// x && (y || z)
private: void operator=(const x_And_BRO_y_Or_z_BRC& s) { report_error("Assignment x_And_BRO_y_Or_z_BRC::operator=() must not be called"); } // Zugriff verhindern
private: template<class RHS> void operator=(const RHS& expr) { report_error("Assignment to non l-value"); }
		 // Zugriff verhindern. Wenn der Zugriff auf den
		 // Standard-Copykonstruktor erlaubt wre, wrde z.B. der sinnlose Ausdruck
		 // (expr(true) && expr(true)) = x (wobei x ein Ausdruck ist) ohne Fehlermeldung akzeptiert.
		 // Da MS Visual C++ jedoch das 'private' ignoriert, wird zustzlich noch eine Ausnahme geworfen, um den Fehler anzuzeigen.

public:
	LogExpr x;
	x_Or_y yz;
	x_And_BRO_y_Or_z_BRC(const LogExpr& ex1, const x_Or_y& ex2) : x(ex1), yz(ex2)
	{
		ex1.set_exported(); ex2.set_exported();
		set_not_exported();
	}
	~x_And_BRO_y_Or_z_BRC()
	{
		if(!yz.x.x.exported || !yz.y.x.exported || !x.x.exported)
	    {
		    LogExpr le(to_LogExpr());
		    le.set_exported();
			ExprObjectDestructionHandler<LogExpr>(le);
        }
	}
	void declare_cast_to_bool_not_needed() const { yz.declare_cast_to_bool_not_needed(); x.declare_cast_to_bool_not_needed(); }
	operator x_And_y() const
	{
		set_exported();
		return x_And_y(x, LogExpr(yz));
	}
	expr to_expr() const
	{
		set_exported();
		return x_And_y(x, LogExpr(yz)).to_expr(); // Der Ausdruck x_And_y(expr(x), yz.to_expr()).to_expr() wrde zwar
			// funktionieren, aber einen ineffizienteren Bytecode liefern.
	}
	bool is_expr_type() const { set_exported(); return true; }
	LogExpr to_LogExpr() const
	{
		set_exported();
		return LogExpr(x_And_y(x, LogExpr(yz))); // Der Ausdruck x_And_y(expr(x), yz.to_expr()).to_expr() wrde zwar
			// funktionieren, aber einen ineffizienteren Bytecode liefern.
	}
	void set_exported() const
	{
		x.set_exported();
		yz.set_exported();
	}
	void vset_exported() const { set_exported(); }
	void vset_not_exported() const { set_not_exported(); }
	void set_not_exported() const
	{
		x.set_not_exported();
		yz.set_not_exported();
	}
};

struct BRO_x_Or_y_BRC_And_z
{	// (x || y) && z
private: void operator=(const BRO_x_Or_y_BRC_And_z& s) { report_error("Assignment BRO_x_Or_y_BRC_And_z::operator=() must not be called"); } // Zugriff verhindern
private: template<class RHS> void operator=(const RHS& expr) { report_error("Assignment to non l-value"); }
		 // Zugriff verhindern. Wenn der Zugriff auf den
		 // Standard-Copykonstruktor erlaubt wre, wrde z.B. der sinnlose Ausdruck
		 // (expr(true) && expr(true)) = x (wobei x ein Ausdruck ist) ohne Fehlermeldung akzeptiert.
		 // Da MS Visual C++ jedoch das 'private' ignoriert, wird zustzlich noch eine Ausnahme geworfen, um den Fehler anzuzeigen.

public:
	x_Or_y xy;
	LogExpr z;
	BRO_x_Or_y_BRC_And_z(const x_Or_y& ex1, const LogExpr& ex2) : xy(ex1), z(ex2)
	{
		ex1.set_exported(); ex2.set_exported();
		set_not_exported();
	}
	~BRO_x_Or_y_BRC_And_z()
	{
		if(!xy.x.x.exported || !xy.y.x.exported || !z.x.exported)
	    {
		    LogExpr le(to_LogExpr());
		    le.set_exported();
			ExprObjectDestructionHandler<LogExpr>(le);
        }
	}
	void declare_cast_to_bool_not_needed() const { xy.declare_cast_to_bool_not_needed(); z.declare_cast_to_bool_not_needed(); }
	operator x_And_y() const
	{
		set_exported();
		return x_And_y(LogExpr(xy), z);
	}
	expr to_expr() const
	{
		set_exported();
		return x_And_y(LogExpr(xy), z).to_expr(); // Der Ausdruck x_And_y(xy.to_expr(), expr(z)).to_expr() wrde zwar
			// funktionieren, aber einen ineffizienteren Bytecode liefern.
	}
	bool is_expr_type() const { set_exported(); return true; }
	LogExpr to_LogExpr() const
	{
		set_exported();
		return LogExpr(x_And_y(LogExpr(xy), z)); // Der Ausdruck x_And_y(xy.to_expr(), expr(z)).to_expr() wrde zwar
			// funktionieren, aber einen ineffizienteren Bytecode liefern.
	}
	void set_exported() const
	{
		xy.set_exported();
		z.set_exported();
	}
	void vset_exported() const { set_exported(); }
	void vset_not_exported() const { set_not_exported(); }
	void set_not_exported() const
	{
		xy.set_not_exported();
		z.set_not_exported();
	}
};

struct x_Or_BRO_y_And_z_BRC
{	// x || (y && z)
private: void operator=(const x_Or_BRO_y_And_z_BRC& s) { report_error("Assignment x_Or_BRO_y_And_z_BRC::operator=() must not be called"); } // Zugriff verhindern
private: template<class RHS> void operator=(const RHS& expr) { report_error("Assignment to non l-value"); }
		 // Zugriff verhindern. Wenn der Zugriff auf den
		 // Standard-Copykonstruktor erlaubt wre, wrde z.B. der sinnlose Ausdruck
		 // (expr(true) && expr(true)) = x (wobei x ein Ausdruck ist) ohne Fehlermeldung akzeptiert.
		 // Da MS Visual C++ jedoch das 'private' ignoriert, wird zustzlich noch eine Ausnahme geworfen, um den Fehler anzuzeigen.

public:
	LogExpr x;
	x_And_y yz;
	x_Or_BRO_y_And_z_BRC(const LogExpr& ex1, const x_And_y& ex2) : x(ex1), yz(ex2)
	{
		ex1.set_exported(); ex2.set_exported();
		set_not_exported();
	}
	~x_Or_BRO_y_And_z_BRC()
	{
		if(!yz.x.x.exported || !yz.y.x.exported || !x.x.exported)
	    {
		    LogExpr le(to_LogExpr());
		    le.set_exported();
			ExprObjectDestructionHandler<LogExpr>(le);
        }
	}
	void declare_cast_to_bool_not_needed() const { yz.declare_cast_to_bool_not_needed(); x.declare_cast_to_bool_not_needed(); }
	operator x_Or_y() const
	{
		set_exported();
		return x_Or_y(x, LogExpr(yz));
	}
	expr to_expr() const
	{
		set_exported();
		return x_Or_y(x, LogExpr(yz)).to_expr(); // Der Ausdruck x_Or_y(expr(x), yz.to_expr()).to_expr() wrde zwar
			// funktionieren, aber einen ineffizienteren Bytecode liefern.
	}
	bool is_expr_type() const { set_exported(); return true; }
	LogExpr to_LogExpr() const
	{
		set_exported();
		return LogExpr(x_Or_y(x, LogExpr(yz))); // Der Ausdruck x_Or_y(expr(x), yz.to_expr()).to_expr() wrde zwar
			// funktionieren, aber einen ineffizienteren Bytecode liefern.
	}
	void set_exported() const
	{
		x.set_exported();
		yz.set_exported();
	}
	void vset_exported() const { set_exported(); }
	void vset_not_exported() const { set_not_exported(); }
	void set_not_exported() const
	{
		x.set_not_exported();
		yz.set_not_exported();
	}
};

///////////////////////////////////////////////////////////////

LogExpr operator!(const LogExpr& x)
{
	x.set_exported();
	LogExpr ex(x);
	ex.negated=!ex.negated;
	return ex;
}

x_Or_y operator!(const x_And_y& ex)
{
	ex.set_exported();
	return x_Or_y(!ex.x, !ex.y);
}

x_And_y operator!(const x_Or_y& ex)
{
	ex.set_exported();
	return x_And_y(!ex.x, !ex.y);
}

BRO_x_Or_y_BRC_Or_z operator!(const BRO_x_And_y_BRC_And_z& ex)
{
	ex.set_exported();
	return BRO_x_Or_y_BRC_Or_z(!ex.xy, !ex.z);
}

BRO_x_And_y_BRC_And_z operator!(const BRO_x_Or_y_BRC_Or_z& ex)
{
	ex.set_exported();
	return BRO_x_And_y_BRC_And_z(!ex.xy, !ex.z);
}

BRO_x_Or_y_BRC_And_z operator!(const BRO_x_And_y_BRC_Or_z& ex)
{
	ex.set_exported();
	return BRO_x_Or_y_BRC_And_z(!ex.xy, !ex.z);
}

BRO_x_And_y_BRC_Or_z operator!(const BRO_x_Or_y_BRC_And_z& ex)
{
	ex.set_exported();
	return BRO_x_And_y_BRC_Or_z(!ex.xy, !ex.z);
}

x_Or_BRO_y_And_z_BRC operator!(const x_And_BRO_y_Or_z_BRC& ex)
{
	ex.set_exported();
	return x_Or_BRO_y_And_z_BRC(!ex.x, !ex.yz);
}

x_And_BRO_y_Or_z_BRC operator!(const x_Or_BRO_y_And_z_BRC& ex)
{
	ex.set_exported();
	return x_And_BRO_y_Or_z_BRC(!ex.x, !ex.yz);
}

///////////////////////////////////////////////////////////////

x_And_y operator&&(const LogExpr& x, const LogExpr& y)
{
	x.set_exported(); y.set_exported();
	return x_And_y(LogExpr(x), LogExpr(y));
}

x_Or_y operator||(const LogExpr& x, const LogExpr& y)
{
	x.set_exported(); y.set_exported();
	return x_Or_y(x, y);
}

BRO_x_And_y_BRC_And_z operator&&(const x_And_y& xy, const LogExpr& z)
{
	xy.set_exported(); z.set_exported();
	return BRO_x_And_y_BRC_And_z(xy, z);
}

BRO_x_And_y_BRC_And_z operator&&(const LogExpr& x, const x_And_y& yz)
{
	yz.set_exported(); x.set_exported();
	return BRO_x_And_y_BRC_And_z(x_And_y(x, yz.x), yz.y);
}

BRO_x_Or_y_BRC_And_z operator&&(const x_Or_y& xy, const LogExpr& z)
{
	xy.set_exported(); z.set_exported();
	return BRO_x_Or_y_BRC_And_z(xy, z);
}

x_And_BRO_y_Or_z_BRC operator&&(const LogExpr& x, const x_Or_y& yz)
{
	yz.set_exported(); x.set_exported();
	return x_And_BRO_y_Or_z_BRC(x, yz);
}

BRO_x_And_y_BRC_Or_z operator||(const x_And_y& xy, const LogExpr& z)
{
	xy.set_exported(); z.set_exported();
	return BRO_x_And_y_BRC_Or_z(xy, z);
}

x_Or_BRO_y_And_z_BRC operator||(const LogExpr& x, const x_And_y& yz)
{
	yz.set_exported(); x.set_exported();
	return x_Or_BRO_y_And_z_BRC(x, yz);
}

BRO_x_Or_y_BRC_Or_z operator||(const x_Or_y& xy, const LogExpr& z)
{
	xy.set_exported(); z.set_exported();
	return BRO_x_Or_y_BRC_Or_z(xy, z);
}

BRO_x_Or_y_BRC_Or_z operator||(const LogExpr& x, const x_Or_y& yz)
{
	yz.set_exported(); x.set_exported();
	return BRO_x_Or_y_BRC_Or_z(x_Or_y(x, yz.x), yz.y);
}


BRO_x_And_y_BRC_Or_z operator||(const x_And_y& x, const x_And_y& y)
{
	x.set_exported(); y.set_exported();
	return x || y.to_expr();
}
BRO_x_And_y_BRC_Or_z operator||(const x_And_y& x, const x_Or_y& y)
{
	x.set_exported(); y.set_exported();
	return x || y.to_expr();
}
BRO_x_Or_y_BRC_Or_z operator||(const x_Or_y& x, const x_And_y& y)
{
	x.set_exported(); y.set_exported();
	return x || y.to_expr();
}
BRO_x_Or_y_BRC_Or_z operator||(const x_Or_y& x, const x_Or_y& y)
{
	x.set_exported(); y.set_exported();
	return x || y.to_expr();
}


BRO_x_And_y_BRC_And_z operator&&(const x_And_y& x, const x_And_y& y)
{
	x.set_exported(); y.set_exported();
	return x && y.to_expr();
}
BRO_x_And_y_BRC_And_z operator&&(const x_And_y& x, const x_Or_y& y)
{
	x.set_exported(); y.set_exported();
	return x && y.to_expr();
}
BRO_x_Or_y_BRC_And_z operator&&(const x_Or_y& x, const x_And_y& y)
{
	x.set_exported(); y.set_exported();
	return x && y.to_expr();
}
BRO_x_Or_y_BRC_And_z operator&&(const x_Or_y& x, const x_Or_y& y)
{
	x.set_exported(); y.set_exported();
	return x && y.to_expr();
}


////////////////////////////////////////////////////////////////////////////////

struct ProgramStructureGlobalScope : public ProgramStructureElement
{
	void add(const Code& cd) { if(1==cc->CompilerPassNr) cc->program_code->add(cd); }
	void addcmd(uint8 x) { if(1==cc->CompilerPassNr) cc->program_code->addcmd(x); }
	void end(ProgramStructureElement *superordinate_element) { };
	void add_local_variable(const Command& var_label, const expr& init_value)
	{
		init_value.set_exported();
		report_error("Local Variable defined outside of functions.");
	}
	VARIABLE_INDEX_TYPE add_uninit_local_variable(const Type *type)
	{
		report_error("Local Variable defined outside of functions.");
		return 0; // wird nie aufgerufen.
	}
};

typedef void C_void_type;

#define BYTECODE_TARGETED_COMPILER__BEGIN_GLOBAL_SCOPE(name_of_program) \
			BytecodeTargetedCompiler::C_void_type name_of_program(BytecodeTargetedCompiler::Compiler* compiler) \
			{ BytecodeTargetedCompiler::cc=compiler;
			// Die Zuweisung an cc ist eigentlich nicht notwendig, weil dies Compiler::Compiler() schon selbst vornimmt.
			// Aber die durch BEGIN_GLOBAL_SCOPE(name_of_program) erzeugte Funktion name_of_program() soll einen Parameter haben
			// damit der Benutzer sie nicht aus Versehen selbst aufruft, ohne cc gesetzt zu haben.
#define BYTECODE_TARGETED_COMPILER__END_GLOBAL_SCOPE }

#define BYTECODE_TARGETED_COMPILER__END } BytecodeTargetedCompiler::EndProgramStructureElement();

#define BYTECODE_TARGETED_COMPILER__IF(condition) ;BytecodeTargetedCompiler::IF_Element((condition)); {
#define BYTECODE_TARGETED_COMPILER__ELSE } ;BytecodeTargetedCompiler::ElseElement(); {
#define BYTECODE_TARGETED_COMPILER__BEGIN ;BytecodeTargetedCompiler::cc->ProgramStructureStack.push_back(new BytecodeTargetedCompiler::ProgramStructure_Block); {

#define BYTECODE_TARGETED_COMPILER__WHILE(condition) ;BytecodeTargetedCompiler::IF_Element((condition), BytecodeTargetedCompiler::ProgramStructure_IF_BASE::while_loop); {

#define BYTECODE_TARGETED_COMPILER__DO_BEGIN BYTECODE_TARGETED_COMPILER__BEGIN
#define BYTECODE_TARGETED_COMPILER__END_WHILE(condition) } ;BytecodeTargetedCompiler::EndProgramStructureElement((condition));

#define BYTECODE_TARGETED_COMPILER__FOR ;BytecodeTargetedCompiler::cc->ProgramStructureStack.push_back(new BytecodeTargetedCompiler::ProgramStructure_ForLoop(false)); {
#define BYTECODE_TARGETED_COMPILER__DOFOR ;BytecodeTargetedCompiler::cc->ProgramStructureStack.push_back(new BytecodeTargetedCompiler::ProgramStructure_ForLoop(true)); {
#define BYTECODE_TARGETED_COMPILER__DO ;BytecodeTargetedCompiler::DO_StatementHandler();

#define BYTECODE_TARGETED_COMPILER__BREAK ;BytecodeTargetedCompiler::BREAK_StatementHandler(1);
#define BYTECODE_TARGETED_COMPILER__BREAK2 ;BytecodeTargetedCompiler::BREAK_StatementHandler(2);
#define BYTECODE_TARGETED_COMPILER__BREAK3 ;BytecodeTargetedCompiler::BREAK_StatementHandler(3);
#define BYTECODE_TARGETED_COMPILER__BREAK4 ;BytecodeTargetedCompiler::BREAK_StatementHandler(4);
#define BYTECODE_TARGETED_COMPILER__BREAK5 ;BytecodeTargetedCompiler::BREAK_StatementHandler(5);
#define BYTECODE_TARGETED_COMPILER__BREAK6 ;BytecodeTargetedCompiler::BREAK_StatementHandler(6);
#define BYTECODE_TARGETED_COMPILER__BREAK7 ;BytecodeTargetedCompiler::BREAK_StatementHandler(7);
#define BYTECODE_TARGETED_COMPILER__BREAK8 ;BytecodeTargetedCompiler::BREAK_StatementHandler(8);
#define BYTECODE_TARGETED_COMPILER__BREAK9 ;BytecodeTargetedCompiler::BREAK_StatementHandler(9);
#define BYTECODE_TARGETED_COMPILER__BREAK10 ;BytecodeTargetedCompiler::BREAK_StatementHandler(10);

#define BYTECODE_TARGETED_COMPILER__CONTINUE BytecodeTargetedCompiler::;CONTINUE_StatementHandler(1);
#define BYTECODE_TARGETED_COMPILER__CONTINUE2 ;BytecodeTargetedCompiler::CONTINUE_StatementHandler(2);
#define BYTECODE_TARGETED_COMPILER__CONTINUE3 ;BytecodeTargetedCompiler::CONTINUE_StatementHandler(3);
#define BYTECODE_TARGETED_COMPILER__CONTINUE4 ;BytecodeTargetedCompiler::CONTINUE_StatementHandler(4);
#define BYTECODE_TARGETED_COMPILER__CONTINUE5 ;BytecodeTargetedCompiler::CONTINUE_StatementHandler(5);
#define BYTECODE_TARGETED_COMPILER__CONTINUE6 ;BytecodeTargetedCompiler::CONTINUE_StatementHandler(6);
#define BYTECODE_TARGETED_COMPILER__CONTINUE7 ;BytecodeTargetedCompiler::CONTINUE_StatementHandler(7);
#define BYTECODE_TARGETED_COMPILER__CONTINUE8 ;BytecodeTargetedCompiler::CONTINUE_StatementHandler(8);
#define BYTECODE_TARGETED_COMPILER__CONTINUE9 ;BytecodeTargetedCompiler::CONTINUE_StatementHandler(9);
#define BYTECODE_TARGETED_COMPILER__CONTINUE10 ;BytecodeTargetedCompiler::CONTINUE_StatementHandler(10);

ProgramStructureElement *GetActFunctionProgramStructure(char *error_message_if_outside_of_function=0);

nlist<Command>::iterator find_command_from_uid(Code& code, Command::uid_type uid,
											   nlist<Command>::iterator search_start=nlist<Command>::iterator())
{
	nlist<Command>::iterator it;
	if(search_start!=0)
		it=search_start;
	else
		it=code.code.begin(); // search_start muss 0 bleiben !
	while(search_start!=0 || it!=code.code.end())
	{
		if(search_start!=0)
		{
			if(search_start==code.code.begin())
				search_start=nlist<Command>::iterator(0);
			else if((--search_start)->get_uid()==uid)
				return search_start;
		}
		if(it!=code.code.end())
		{
			if(it->get_uid()==uid)
				return it;
			++it;
		}
	}
	return nlist<Command>::iterator();
}

Code write_uninit_vars_typeinfo(const nlist<uint8>& types)
{
	Code cd;
	for(nlist<uint8>::const_iterator it=types.begin(); it!=types.end(); )
	{
		uint32 nrep=0, rep=0;
		nlist<uint8>::const_iterator it2, it3;
		for(it2=it; it2!=types.end(); it2++, nrep++)
		{
			for(it3=it2, rep=0; it3!=types.end() && *it3==*it2; it3++, rep++)
				;
			if(rep>=4)
				break;
		}
		while(nrep)
		{
			cd.add(Command::GetRawBytesCommand().addnibble(nrep>9 ? 9 : nrep));
			for(long i=(nrep>9 ? 9 : nrep); 0!=i ; --i, --nrep)
				cd.add(Command::GetRawBytesCommand().addnibble(*(it++)));
		}
		if(rep>=4)
		{
			cd.add(Command::GetRawBytesCommand().add_coded_integer15(6+rep, MAX_NR_BYTES_FOR_VARIABLE_INDEX, false).addnibble(*it2));
			it=it3;
		}
		else
			break; // dann ist it2==types.end()
	}
	cd.add(Command::GetRawBytesCommand().addnibble(0)); // Ende-Markierung
	return cd;
}

struct ProgramStructure_ForLoop : public ProgramStructureElement
{
	bool init_occured, incr_occured, dofor_loop;
	expr init_code, incr_code;
	uint32 local_variable_count;
	uint32 uninit_vars_created_local_stack_offset;
	nlist<uint8> uninit_vars_created_types;
	ProgramStructure_ForLoop(bool _dofor_loop)
		{ dofor_loop=_dofor_loop; init_occured=incr_occured=false; local_variable_count=uninit_vars_created_local_stack_offset=0; }
	void add(const Code& ex)
	{
		if(uninit_vars_created_types.size()!=0)
		{
			init_code.add(Command(bi::SETTYPES).add_coded_integer255(uninit_vars_created_local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, false));
			init_code.add(write_uninit_vars_typeinfo(uninit_vars_created_types));
			uninit_vars_created_types.clear();
		}
		if(!init_occured)
		{
			if(cc->ProgramStructureStack.size()<2 || cc->ProgramStructureStack.back()!=this || init_occured)
				report_error("FOR loop error");
			init_code.add(ex);
			init_occured=true;
		}
		else
		{
			incr_code.add(ex);
			incr_occured=true;
		}
	}
	void end(ProgramStructureElement *superordinate_element) { }
	void add_local_variable(const Command& var_label, const expr& init_value)
	{
		init_value.set_exported();
		if(init_occured)
			report_error("Local variable cannot be defined after the first statement (init) between FOR and DO");
		if(cc->ProgramStructureStack.size()<2 || cc->ProgramStructureStack.back()!=this)
			report_error("FOR loop error");
		++local_variable_count;
		init_value.set_code_has_effect();
		add(init_value); // Hier MUSS add() aufgerufen werden.
		init_code.add(var_label); // Hier nicht add() aufrufen, damit var_label zu init_code statt incr_code hinzugefgt wird.
		init_occured=true;
	}
	VARIABLE_INDEX_TYPE add_uninit_local_variable(const Type *type);
};

struct ProgramStructure_IF_BASE : public ProgramStructureElement
{
	Code ifcode, elsecode, for_init_code, for_incr_code;
	bool if_is_active, for_incr_code_occured;
	enum SType { if_element, while_loop, dowhile_loop, for_loop, dofor_loop } type;
	nlist<Command::uid_type> jump_command_locations;
	uint32 local_variable_count_for_init, local_variable_count_if, local_variable_count_else;
	uint32 uninit_vars_created_local_stack_offset;
	nlist<uint8> uninit_vars_created_types;

	ProgramStructure_IF_BASE()
	{
		if_is_active=true; type=if_element; for_incr_code_occured=false;
		local_variable_count_if=local_variable_count_else=0;
	}
	void add(const Code& ex)
	{
		if(for_loop==type)
		{
			if(uninit_vars_created_types.size()!=0)
			{
				for_incr_code.add(Command(bi::SETTYPES).add_coded_integer255(uninit_vars_created_local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, false));
				for_incr_code.add(write_uninit_vars_typeinfo(uninit_vars_created_types));
				uninit_vars_created_types.clear();
			}
			if(for_incr_code_occured)
				report_error("More than 3 statemente between FOR and DO occured or no DO occured between FOR and END; required are exactly 3 statements between FOR and DO (initialization, condition, increment).");
			for_incr_code.add(ex);
			for_incr_code_occured=true;
		}
		else if(dofor_loop==type)
			report_error("More than 3 statemente between DOFOR and DO occured or no DO occured between DOFOR and END; required are exactly 3 statements between DOFOR and DO (initialization, condition, increment).");
		else if(if_is_active)
		{
			if(uninit_vars_created_types.size()!=0)
			{
				ifcode.add(Command(bi::SETTYPES).add_coded_integer255(uninit_vars_created_local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, false));
				ifcode.add(write_uninit_vars_typeinfo(uninit_vars_created_types));
				uninit_vars_created_types.clear();
			}
			ifcode.add(ex);
		}
		else
		{
			if(uninit_vars_created_types.size()!=0)
			{
				elsecode.add(Command(bi::SETTYPES).add_coded_integer255(uninit_vars_created_local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, false));
				elsecode.add(write_uninit_vars_typeinfo(uninit_vars_created_types));
				uninit_vars_created_types.clear();
			}
			elsecode.add(ex);
		}
	}
	void add_for_incr_code()
	{
		if(for_loop==type)
			report_error("DO required between FOR and END.");
		if(0!=for_incr_code.code.size())
			ifcode.add(for_incr_code);
	}
	void add_local_variable(const Command& var_label, const expr& init_value)
	{
		init_value.set_exported();
		if(for_loop==type)
			report_error("If a local variable shall be defined between FOR and DO, it must be defined in the first statement after FOR.");
		if(if_is_active)
			++local_variable_count_if;
		else
			++local_variable_count_else;
		add(init_value);
		add(var_label);
	}
	VARIABLE_INDEX_TYPE add_uninit_local_variable(const Type *type);
};

void move_jump_location_to_previous_loop_or_block(Command::uid_type uid, nstring error_message="Internal Compiler Error");

struct ProgramStructure_Block : public ProgramStructureElement
{
	Code code;
	nlist<Command::uid_type> jump_command_locations;
	uint32 local_variable_count;
	uint32 uninit_vars_created_local_stack_offset;
	nlist<uint8> uninit_vars_created_types;
	ProgramStructure_Block() { local_variable_count=uninit_vars_created_local_stack_offset=0; }
	void add(const Code& ex)
	{
		if(uninit_vars_created_types.size()!=0)
		{
			code.add(Command(bi::SETTYPES).add_coded_integer255(uninit_vars_created_local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, false));
			code.add(write_uninit_vars_typeinfo(uninit_vars_created_types));
			uninit_vars_created_types.clear();
		}
		code.add(ex);
	}
	void end(ProgramStructureElement *superordinate_element)
	{
		for(nlist<Command::uid_type>::iterator it=jump_command_locations.begin(); it!=jump_command_locations.end(); it++)
		{
			nlist<Command>::iterator p=find_command_from_uid(code, *it);
			if(!p.is_valid_and_not_end(code.code))
				report_error("Internal Compiler Error");
			move_jump_location_to_previous_loop_or_block(*it, p->fkt==Command::break_statement ? "BREAK occured outside of loop" :
						(p->fkt==Command::continue_statement ? "CONTINUE occured outside of loop" : "GOTO occured outside of loop"));
		}
		if(local_variable_count)
			code.add(Code::POPN_Commands(local_variable_count, true));
		superordinate_element->add(code);
	}
	void add_local_variable(const Command& var_label, const expr& init_value)
	{
		init_value.set_exported();
		++local_variable_count;
		add(init_value);
		add(var_label);
	}
	VARIABLE_INDEX_TYPE add_uninit_local_variable(const Type *type);
};

void move_jump_location_to_previous_loop_or_block(Command::uid_type uid, nstring error_message/*="Internal Compiler Error"*/)
{
	bool found=false;
	if(cc->ProgramStructureStack.size()<2)
		report_error(error_message);
	nlist<ProgramStructureElement*>::iterator it2=cc->ProgramStructureStack.end();
	while(it2!=cc->ProgramStructureStack.begin())
	{
		if(0!=dynamic_cast<ProgramStructure_IF_BASE*>(*(--it2)) &&
				ProgramStructure_IF_BASE::if_element!=static_cast<ProgramStructure_IF_BASE*>(*it2)->type)
		{
			found=true;
			static_cast<ProgramStructure_IF_BASE*>(*it2)->jump_command_locations.push_back(uid);
			break;
		}
		else if(0!=dynamic_cast<ProgramStructure_Block*>(*it2))
		{
			found=true;
			static_cast<ProgramStructure_Block*>(*it2)->jump_command_locations.push_back(uid);
			break;
		}
	}
	if(!found)
		report_error(error_message);
}

void DO_StatementHandler()
{
	if(0==cc->ProgramStructureStack.size() || 0==dynamic_cast<ProgramStructure_IF_BASE*>(cc->ProgramStructureStack.back()))
		report_error("DO occured eighther outside of FOR ... END or DOFOR ... END or there are less than 3 expr-statements between FOR and DO or DOFOR and DO.");
	ProgramStructure_IF_BASE *p=static_cast<ProgramStructure_IF_BASE*>(cc->ProgramStructureStack.back());
	if(ProgramStructure_IF_BASE::for_loop!=p->type && ProgramStructure_IF_BASE::dofor_loop!=p->type)
		report_error("DO occured eighther outside of FOR ... END or DOFOR ... END or multiple DO occured between FOR and END od DOFOR and END.");
	if(ProgramStructure_IF_BASE::for_loop==p->type)
	{
		p->type=ProgramStructure_IF_BASE::while_loop;
		if(!p->for_incr_code_occured)
			report_error("Less than 3 statemente between FOR and DO occured; required are exactly 3 statements (initialization, condition, increment).");
	}
	else
		p->type=ProgramStructure_IF_BASE::dowhile_loop;
}

void BREAK_StatementHandler(long nr_breaks)
{
	if(nr_breaks<=0)
		report_error("n in BREAKN(n) must be positive.");
	if(0==cc->ProgramStructureStack.size())
		report_error("BREAK occured outside of the global scope (between BEGIN_GLOBAL_SCOPE and END_GLOBAL_SCOPE)");
	Command cmd;
	cmd.fkt=Command::break_statement;

	nlist<ProgramStructureElement*>::iterator it;
	for(it=cc->ProgramStructureStack.end() ; it!=cc->ProgramStructureStack.begin() ; )
	{
		if(0!=dynamic_cast<ProgramStructure_IF_BASE*>(*(--it)) &&
				ProgramStructure_IF_BASE::if_element!=static_cast<ProgramStructure_IF_BASE*>(*it)->type)
		{
			cmd.nr_breaks_or_continues=nr_breaks;
			static_cast<ProgramStructure_IF_BASE*>(*it)->jump_command_locations.push_back(cmd.get_uid());
			cc->ProgramStructureStack.back()->add(Code(cmd));
			return;
		}
		else if(0!=dynamic_cast<ProgramStructure_Block*>(*it))
		{	// Das break-Kommando muss zunchst auch in einen ProgramStructure_Block aufgenommen werden, weil eine
			// DO_BEGIN...END_WHILE-Schleife zunchst als ProgramStructure_Block gehandhabt wird und erst beim Erscheinen des
			// END_WHILE-Schlsselworts in ein ProgramStructure_IF_Spec mit dem Typ type==dowhile_loop umgewandelt wird.
			cmd.nr_breaks_or_continues=nr_breaks;
			static_cast<ProgramStructure_Block*>(*it)->jump_command_locations.push_back(cmd.get_uid());
			cc->ProgramStructureStack.back()->add(Code(cmd));
			return;
		}
	}
	report_error("BREAK occured outside of loop.");	
}

void CONTINUE_StatementHandler(long nr_continues)
{
	if(nr_continues<=0)
		report_error("n in CONTINUE(n) must be positive.");
	if(0==cc->ProgramStructureStack.size())
		report_error("CONTINUE occured outside of the global scope (between BEGIN_GLOBAL_SCOPE and END_GLOBAL_SCOPE)");
	Command cmd;
	cmd.fkt=Command::continue_statement;

	nlist<ProgramStructureElement*>::iterator it;
	for(it=cc->ProgramStructureStack.end() ; it!=cc->ProgramStructureStack.begin() ; )
	{
		if(0!=dynamic_cast<ProgramStructure_IF_BASE*>(*(--it)) &&
				ProgramStructure_IF_BASE::if_element!=static_cast<ProgramStructure_IF_BASE*>(*it)->type)
		{
			cmd.nr_breaks_or_continues=nr_continues;
			static_cast<ProgramStructure_IF_BASE*>(*it)->jump_command_locations.push_back(cmd.get_uid());
			cc->ProgramStructureStack.back()->add(Code(cmd));
			return;
		}
		else if(0!=dynamic_cast<ProgramStructure_Block*>(*it))
		{	// Das break-Kommando muss zunchst auch in einen ProgramStructure_Block aufgenommen werden, weil eine
			// DO_BEGIN...END_WHILE-Schleife zunchst als ProgramStructure_Block gehandhabt wird und erst beim Erscheinen des
			// END_WHILE-Schlsselworts in ein ProgramStructure_IF_Spec mit dem Typ type==dowhile_loop umgewandelt wird.
			cmd.nr_breaks_or_continues=nr_continues;
			static_cast<ProgramStructure_Block*>(*it)->jump_command_locations.push_back(cmd.get_uid());
			cc->ProgramStructureStack.back()->add(Code(cmd));
			return;
		}
	}
	report_error("CONTINUE occured outside of loop.");	
}

void ElseElement()
{
	if(0==cc->ProgramStructureStack.size() ||
			0==dynamic_cast<ProgramStructure_IF_BASE*>(cc->ProgramStructureStack.back()))
		report_error("ELSE without IF");
	if(!static_cast<ProgramStructure_IF_BASE*>(cc->ProgramStructureStack.back())->if_is_active)
		report_error("ELSE already occured");
	if(ProgramStructure_IF_BASE::if_element!=static_cast<ProgramStructure_IF_BASE*>(cc->ProgramStructureStack.back())->type)
		report_error("ELSE without IF");
	static_cast<ProgramStructure_IF_BASE*>(cc->ProgramStructureStack.back())->if_is_active=false;
}

template<class T>
struct ProgramStructure_IF_Spec : public ProgramStructure_IF_BASE
{
	T condition;
	ProgramStructure_IF_Spec(const T& cond, SType _type,
							 const expr& _for_init_code=expr(), long __local_variable_count_for_init=0,
							 uint32 __uninit_vars_created_local_stack_offset=0,
							 nlist<uint8> __uninit_vars_created_types=nlist<uint8>(),
							 const expr& _for_incr_code=expr())
							 : condition(cond)
	{
		cond.set_exported(); _for_incr_code.set_exported(); _for_init_code.set_exported();
		type=_type; for_incr_code=_for_incr_code; for_init_code=_for_init_code;
		local_variable_count_for_init=__local_variable_count_for_init;
		uninit_vars_created_local_stack_offset=__uninit_vars_created_local_stack_offset;
		uninit_vars_created_types=__uninit_vars_created_types;
		if(cond.to_expr().set_exported().code.size()==0)
			report_error("Empty loop condition not allowed (use TRUE if condition shall always be fulfilled)");
	}
	void end(ProgramStructureElement *superordinate_element);
};

/*template<>
void ProgramStructure_IF_Spec<expr>::end(ProgramStructureElement *superordinate_element)
{
	condition.set_exported();
	Command::uid_type begin_of_increment_marker_uid=0, end_of_loop_marker_uid=0;

	if(local_variable_count_if)
		ifcode.add(Code::POPN_Commands(local_variable_count_if, true));
	if(local_variable_count_else)
		elsecode.add(Code::POPN_Commands(local_variable_count_else, true));

	for(nlist<Command::uid_type>::iterator it=jump_command_locations.begin(); it!=jump_command_locations.end(); it++)
	{
		if(while_loop==type)
		{
			nlist<Command>::iterator p=find_command_from_uid(ifcode, *it);
			if(p.is_valid_and_not_end(ifcode.code))
			{
				if((Command::break_statement==p->fkt || Command::continue_statement==p->fkt) && 0!=--p->nr_breaks_or_continues)
					move_jump_location_to_previous_loop_or_block(*it, p->fkt==Command::break_statement ? "n in BREAKN(n) is too large." :
								"n in CONTINUE(n) is too large.");
				else if(Command::break_statement==p->fkt)
				{
					if(0==end_of_loop_marker_uid)
						end_of_loop_marker_uid=Command::generate_new_uid();
					p->add(bi::BRK);
					p->branch1_id=ifcode.get_uid_of_first_command();
					p->fkt=Command::command;
				}
				else if(Command::continue_statement==p->fkt)
				{
					if(for_incr_code.code.size()!=0)
					{
						if(0==begin_of_increment_marker_uid)
							begin_of_increment_marker_uid=ifcode.get_uid_after_last_command();
						p->add(bi::BR);
						p->branch1_id=begin_of_increment_marker_uid;
						p->fkt=Command::command;
					}
					else
					{
						p->add(bi::CONT);
						p->branch1_id=ifcode.get_uid_of_first_command();
						p->fkt=Command::command;
					}
				}
			}
		}
		else if(dowhile_loop==type)
		{
			nlist<Command>::iterator p=find_command_from_uid(ifcode, *it);
			if(p.is_valid_and_not_end(ifcode.code))
			{
				if((Command::break_statement==p->fkt || Command::continue_statement==p->fkt) && 0!=--p->nr_breaks_or_continues)
					move_jump_location_to_previous_loop_or_block(*it, p->fkt==Command::break_statement ? "n in BREAKN(n) is too large." :
								"n in CONTINUE(n) is too large.");
				else if(Command::break_statement==p->fkt)
				{
					p->add(bi::BR);
					p->branch1_id=ifcode.get_uid_of_first_command();
					p->fkt=Command::command;
				}
				else if(Command::continue_statement==p->fkt)
				{
					if(0==begin_of_increment_marker_uid)
						begin_of_increment_marker_uid=ifcode.get_uid_after_last_command();
					p->add(bi::BR);
					p->branch1_id=begin_of_increment_marker_uid;
					p->fkt=Command::command;
				}
			}
		}
	}

	add_for_incr_code(); // Muss NACH der obigen Schleife zur Behandlung von BREAK und CONTINUE stehen !

	if(0!=elsecode.code.size())
		superordinate_element->add(if_else_code(condition, ifcode, elsecode, false, false, false));
	else if(dowhile_loop==type)
	{
		superordinate_element->add(for_init_code);
		superordinate_element->add(dowhile_code(condition, ifcode));
		if(0!=end_of_loop_marker_uid)
			superordinate_element->add(Code(Command::new_uid_label(end_of_loop_marker_uid)));
		if(local_variable_count_for_init) // fr den Fall einer DOFOR-Schleife
			superordinate_element->add(Code::POPN_Commands(local_variable_count_for_init, true));
	}
	else
	{
		superordinate_element->add(for_init_code);
		superordinate_element->add(if_code(condition, ifcode, false, false, while_loop==type));
		if(local_variable_count_for_init)
			superordinate_element->add(Code::POPN_Commands(local_variable_count_for_init, true));
	}
}

void IF_Element(const expr& cond, ProgramStructure_IF_BASE::SType type=ProgramStructure_IF_BASE::if_element)
{
	cond.set_exported();
	cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<expr>(cond, type));
}*/

template<>
void ProgramStructure_IF_Spec<LogExpr>::end(ProgramStructureElement *superordinate_element)
{
	condition.set_exported();
	Command::uid_type begin_of_increment_marker_uid=0, end_of_loop_marker_uid=0;

	if(local_variable_count_if)
		ifcode.add(Code::POPN_Commands(local_variable_count_if, true)); // POP fr local_variable_count_for_init erfolgt unten !
	if(local_variable_count_else)
		elsecode.add(Code::POPN_Commands(local_variable_count_else, true));

	for(nlist<Command::uid_type>::iterator it=jump_command_locations.begin(); it!=jump_command_locations.end(); it++)
	{
		if(while_loop==type)
		{
			nlist<Command>::iterator p=find_command_from_uid(ifcode, *it);
			if(p.is_valid_and_not_end(ifcode.code))
			{
				if((Command::break_statement==p->fkt || Command::continue_statement==p->fkt) && 0!=--p->nr_breaks_or_continues)
					move_jump_location_to_previous_loop_or_block(*it, p->fkt==Command::break_statement ? "n in BREAKN(n) is too large." :
								"n in CONTINUE(n) is too large.");
				else if(Command::break_statement==p->fkt)
				{
					if(0==end_of_loop_marker_uid)
						end_of_loop_marker_uid=Command::generate_new_uid();
					p->add(bi::BRK);
					p->branch1_id=ifcode.get_uid_of_first_command();
					p->fkt=Command::command;
				}
				else if(Command::continue_statement==p->fkt)
				{
					if(for_incr_code.code.size()!=0)
					{
						if(0==begin_of_increment_marker_uid)
							begin_of_increment_marker_uid=ifcode.get_uid_after_last_command();
						p->add(bi::BR);
						p->branch1_id=begin_of_increment_marker_uid;
						p->fkt=Command::command;
					}
					else
					{
						p->add(bi::CONT);
						p->branch1_id=ifcode.get_uid_of_first_command();
						p->fkt=Command::command;
					}
				}
			}
		}
		else if(dowhile_loop==type)
		{
			nlist<Command>::iterator p=find_command_from_uid(ifcode, *it);
			if(p.is_valid_and_not_end(ifcode.code))
			{
				if((Command::break_statement==p->fkt || Command::continue_statement==p->fkt) && 0!=--p->nr_breaks_or_continues)
					move_jump_location_to_previous_loop_or_block(*it, p->fkt==Command::break_statement ? "n in BREAKN(n) is too large." :
								"n in CONTINUE(n) is too large.");
				else if(Command::break_statement==p->fkt)
				{
					p->add(bi::BR);
					p->branch1_id=ifcode.get_uid_of_first_command();
					p->fkt=Command::command;
				}
				else if(Command::continue_statement==p->fkt)
				{
					if(0==begin_of_increment_marker_uid)
						begin_of_increment_marker_uid=ifcode.get_uid_after_last_command();
					p->add(bi::BR);
					p->branch1_id=begin_of_increment_marker_uid;
					p->fkt=Command::command;
				}
			}
		}
	}

	add_for_incr_code(); // Muss NACH der obigen Schleife zur Behandlung von BREAK und CONTINUE stehen !

	if(0!=elsecode.code.size())
	{
		if(condition.negated)
			superordinate_element->add(if_else_code(condition.x, elsecode, ifcode, false, false, false));
		else
			superordinate_element->add(if_else_code(condition.x, ifcode, elsecode, false, false, false));
	}
	else if(dowhile_loop==type)
	{
		superordinate_element->add(for_init_code);
		superordinate_element->add(dowhile_code(condition, ifcode));
		if(0!=end_of_loop_marker_uid)
			superordinate_element->add(Code(Command::new_uid_label(end_of_loop_marker_uid)));
		if(local_variable_count_for_init) // fr den Fall einer DOFOR-Schleife
			superordinate_element->add(Code::POPN_Commands(local_variable_count_for_init, true));
	}
	else
	{
		superordinate_element->add(for_init_code);
		superordinate_element->add(if_code(condition, ifcode, false, false, while_loop==type));
		if(local_variable_count_for_init)
			superordinate_element->add(Code::POPN_Commands(local_variable_count_for_init, true));
	}
}

void IF_Element(const LogExpr& cond, ProgramStructure_IF_BASE::SType type=ProgramStructure_IF_BASE::if_element)
{
	cond.set_exported();
	cond.declare_cast_to_bool_not_needed();
	cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<LogExpr>(cond, type));
}

void IF_Element(const x_And_y& cond, ProgramStructure_IF_BASE::SType type=ProgramStructure_IF_BASE::if_element)
{
	cond.set_exported();
	cond.declare_cast_to_bool_not_needed();
	cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<LogExpr>(LogExpr(cond), type));
}

void IF_Element(const x_Or_y& cond, ProgramStructure_IF_BASE::SType type=ProgramStructure_IF_BASE::if_element)
{
	cond.set_exported();
	cond.declare_cast_to_bool_not_needed();
	cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<LogExpr>(LogExpr(cond), type));
}

void IF_Element(const BRO_x_And_y_BRC_And_z& cond, ProgramStructure_IF_BASE::SType type=ProgramStructure_IF_BASE::if_element)
{
	cond.set_exported();
	cond.declare_cast_to_bool_not_needed();
	cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<LogExpr>(cond.to_LogExpr(), type));
}

void IF_Element(const BRO_x_Or_y_BRC_Or_z& cond, ProgramStructure_IF_BASE::SType type=ProgramStructure_IF_BASE::if_element)
{
	cond.set_exported();
	cond.declare_cast_to_bool_not_needed();
	cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<LogExpr>(cond.to_LogExpr(), type));
}

void IF_Element(const BRO_x_And_y_BRC_Or_z& cond, ProgramStructure_IF_BASE::SType type=ProgramStructure_IF_BASE::if_element)
{
	cond.set_exported();
	cond.declare_cast_to_bool_not_needed();
	cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<LogExpr>(cond.to_LogExpr(), type));
}

void IF_Element(const x_And_BRO_y_Or_z_BRC& cond, ProgramStructure_IF_BASE::SType type=ProgramStructure_IF_BASE::if_element)
{
	cond.set_exported();
	cond.declare_cast_to_bool_not_needed();
	cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<LogExpr>(cond.to_LogExpr(), type));
}

void IF_Element(const BRO_x_Or_y_BRC_And_z& cond, ProgramStructure_IF_BASE::SType type=ProgramStructure_IF_BASE::if_element)
{
	cond.set_exported();
	cond.declare_cast_to_bool_not_needed();
	cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<LogExpr>(cond.to_LogExpr(), type));
}

void IF_Element(const x_Or_BRO_y_And_z_BRC& cond, ProgramStructure_IF_BASE::SType type=ProgramStructure_IF_BASE::if_element)
{
	cond.set_exported();
	cond.declare_cast_to_bool_not_needed();
	cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<LogExpr>(cond.to_LogExpr(), type));
}

ProgramStructureElement *EndProgramStructureElement_intern()
{
	if(0==cc->ProgramStructureStack.size())
		report_error("END unexpected");
	ProgramStructureElement *e=cc->ProgramStructureStack.back();
	cc->ProgramStructureStack.pop_back();
	if(0==dynamic_cast<ProgramStructureGlobalScope*>(e) && 0==cc->ProgramStructureStack.size())
		report_error("END unexpected");
	return e;
}

void EndProgramStructureElement()
{
	ProgramStructureElement *e=EndProgramStructureElement_intern();
	if(0!=cc->ProgramStructureStack.size())
		e->end(cc->ProgramStructureStack.back());
	delete e;
}

template<class T>
void EndProgramStructureElementGen(const T& condition)
{
	condition.set_exported();
	ProgramStructureElement *e=EndProgramStructureElement_intern();
	if(0==dynamic_cast<ProgramStructure_Block*>(e))
		report_error("END_WHILE unexpected");
	ProgramStructure_IF_Spec<T> elem(condition, ProgramStructure_IF_Spec<T>::dowhile_loop);
	ProgramStructure_Block *block=static_cast<ProgramStructure_Block*>(e);
	elem.add(block->code);
	elem.local_variable_count_if=block->local_variable_count;
	for(nlist<Command::uid_type>::iterator it=block->jump_command_locations.begin(); it!=block->jump_command_locations.end(); it++)
		elem.jump_command_locations.push_back(*it);
	if(0!=cc->ProgramStructureStack.size())
		elem.end(cc->ProgramStructureStack.back());
	delete e;
}

void EndProgramStructureElement(const LogExpr& condition)
{
	condition.set_exported();
	EndProgramStructureElementGen<LogExpr>(condition);
}
void EndProgramStructureElement(const x_And_y& condition)
{
	condition.set_exported();
	EndProgramStructureElementGen<LogExpr>(LogExpr(condition));
}
void EndProgramStructureElement(const x_Or_y& condition)
{
	condition.set_exported();
	EndProgramStructureElementGen<LogExpr>(LogExpr(condition));
}
void EndProgramStructureElement(const BRO_x_And_y_BRC_And_z& condition)
{
	condition.set_exported();
	condition.declare_cast_to_bool_not_needed();
	EndProgramStructureElementGen<LogExpr>(condition.to_LogExpr());
}
void EndProgramStructureElement(const BRO_x_Or_y_BRC_Or_z& condition)
{
	condition.set_exported();
	condition.declare_cast_to_bool_not_needed();
	EndProgramStructureElementGen<LogExpr>(condition.to_LogExpr());
}
void EndProgramStructureElement(const BRO_x_And_y_BRC_Or_z& condition)
{
	condition.set_exported();
	condition.declare_cast_to_bool_not_needed();
	EndProgramStructureElementGen<LogExpr>(condition.to_LogExpr());
}
void EndProgramStructureElement(const x_And_BRO_y_Or_z_BRC& condition)
{
	condition.set_exported();
	condition.declare_cast_to_bool_not_needed();
	EndProgramStructureElementGen<LogExpr>(condition.to_LogExpr());
}
void EndProgramStructureElement(const BRO_x_Or_y_BRC_And_z& condition)
{
	condition.set_exported();
	condition.declare_cast_to_bool_not_needed();
	EndProgramStructureElementGen<LogExpr>(condition.to_LogExpr());
}
void EndProgramStructureElement(const x_Or_BRO_y_And_z_BRC& condition)
{
	condition.set_exported();
	condition.declare_cast_to_bool_not_needed();
	EndProgramStructureElementGen<LogExpr>(condition.to_LogExpr());
}

struct Type
{
	virtual ~Type() { } // RTTI ermglichen
	virtual checkptr_c<Type> vcopy() const=0;
	virtual VARIABLE_INDEX_TYPE size() const =0;
	virtual bool equal(checkptr_c<Type> type2) const =0;
	virtual nstring get_type_string() const =0;
	virtual bool is_expr() const { return false; }
	virtual bool is_ptr() const { return false; }
	virtual bool is_struct() const { return false; }
	virtual bool is_array() const { return false; }
	virtual bool check_validity_of_refered_types(bool *already_defined=0) const=0;
	bool chk_validity_of_ref_types_variable_creation() const
	{
		bool already_defined;
		if(!check_validity_of_refered_types(&already_defined))
			report_error(nstring("Error creating a variable of type\n   ").append(get_type_string()).
					append(nstring("\nThis type refers to at least one other type already beeing destroyed.")));
		if(!already_defined)
			report_error(nstring("Error creating a variable of type\n   ").append(get_type_string()).
					append(nstring("\nThis type isn't defined or it refers to a type not beeing defined or only beeing defined incompletely.")));
		return false; // wird nie ausgefhrt.
	}
	virtual bool has_only_homogenous_basic_types() const=0;
	virtual uint8 get_type_nr_of_homogenous_basic_types() const=0;
	virtual nlist<uint8> get_flattend_types_list() const=0;
	virtual uint32 get_nr_of_elementary_objects() const=0;
};

struct RegisterTypeClass
{
	nlist<checkptr_c<Type> > registered_types;
	checkptr_c<Type> Register(checkptr_c<Type> type);
	void cleanup();
	~RegisterTypeClass() { cleanup(); }
};

template<class RHS>
void operator,(const Type& t, const RHS &rhs)
{	// Diese Funktion ist notwendig, denn sonst wrde es ohne Fehlermeldung akzeptiert, wenn der
	// Benutzer z.B. versehentlich
	//    local y(__UINT16);
	//    local x=(POINTER(__UINT16), &y);
	// statt
	//    local y(__UINT16);
	//    local x(POINTER(__UINT16), &y)
	// schreibt.
	report_error(printf_string("Cannot use a type name in the comma operator (,). Cannot use the type name\n   %s\nas the lhs of the operator , .",
		t.get_type_string().c_str()));
}

template<class RHS>
void operator,(const RHS &rhs, const Type& t)
{
	report_error(printf_string("Cannot use a type name in the comma operator (,). Cannot use the type name\n   %s\nas the rhs of the operator , .",
		t.get_type_string().c_str()));
}

struct BasicType : public Type
{
private:
	BasicType(const BasicType &bt) { type=bt.type; }
	BasicType& operator=(const BasicType &bt) { type=bt.type; }
	checkptr_UID uid;
public:
	checkptr_c<BasicType> operator&() { return checkptr_c<BasicType>(this, uid); }
	virtual bool is_expr() const { return true; }
	struct CreateBasicType { };
	BasicType(CreateBasicType, uint8 _type) : type(_type) { }
	uint8 type;
	VARIABLE_INDEX_TYPE size() const { return Value::sbytesize(type); }
	static BasicType& get_BasicType_from_type_nr(uint8 type_nr);
	virtual bool check_validity_of_refered_types(bool *already_defined=0) const
	{
		if(already_defined) *already_defined=true;
		return true;
	}

	template<class simple_type>
	expr convert_to_expr_constant_of_this_type(simple_type x) const
	{
		if(Value::t_void==type) return expr::voidexpr();
		if(Value::t_bool==type) return expr(x ? true : false); // return expr((bool)x);
		if(Value::t_uchar==type) return expr((unsigned char)x);
		if(Value::t_char==type) return expr((signed char)x);
		if(Value::t_ushort==type) return expr((unsigned short)x);
		if(Value::t_short==type) return expr((short)x);
		if(Value::t_uint==type) return expr((unsigned int)x);
		if(Value::t_int==type) return expr((int)x);
		if(Value::t_ulong==type) return expr((unsigned long)x);
		if(Value::t_long==type) return expr((long)x);
		if(Value::t_ulonglong==type) return expr((unsigned long long)x);
		if(Value::t_longlong==type) return expr((long long)x);
	}
	checkptr_c<Type> vcopy() const
	{
		report_error("BasicType::vcopy() must not be called");
		return checkptr_c<Type>(); // wird nie aufgerufen.
	}
	bool equal(checkptr_c<Type> type2) const
	{
		return 0!=DynamicCast<BasicType>(type2) && StaticCast<BasicType>(type2)->type==type;
	}
	static BasicType& get_BasicAddressType()
	{
		return get_BasicType_from_type_nr(Value::VarIndexTypeNr);
	}
	nstring get_type_string() const
	{
		switch(type)
		{
			case Value::t_void: return "VOID"; break;
			case Value::t_bool: return "BOOL"; break;
			case Value::t_i8: return "INT8"; break;
			case Value::t_u8: return "UINT8"; break;
			case Value::t_i16: return "INT16"; break;
			case Value::t_u16: return "UINT16"; break;
			case Value::t_i32: return "INT32"; break;
			case Value::t_u32: return "UINT32"; break;
			case Value::t_i64: return "INT64"; break;
			case Value::t_u64: return "UINT64"; break;
			case Value::t_real: return "REAL"; break;
			default: return "Unknown basic type"; break;
		}
	}
	virtual bool has_only_homogenous_basic_types() const { return true; }
	virtual uint8 get_type_nr_of_homogenous_basic_types() const { return type; }
	virtual nlist<uint8> get_flattend_types_list() const { nlist<uint8> lst; lst.push_back(type); return lst; }
	virtual uint32 get_nr_of_elementary_objects() const { return 1; }
};

nstring expr::get_type_string() const
{
	return BasicType::get_BasicType_from_type_nr(type_nr).get_type_string();
}


BasicType __VOID(BasicType::CreateBasicType(), Value::t_void);
BasicType __BOOL(BasicType::CreateBasicType(), Value::t_bool);
BasicType __UINT8(BasicType::CreateBasicType(), Value::t_u8);
BasicType __INT8(BasicType::CreateBasicType(), Value::t_i8);
BasicType __UINT16(BasicType::CreateBasicType(), Value::t_u16);
BasicType __INT16(BasicType::CreateBasicType(), Value::t_i16);
BasicType __UINT32(BasicType::CreateBasicType(), Value::t_u32);
BasicType __INT32(BasicType::CreateBasicType(), Value::t_i32);
BasicType __UINT64(BasicType::CreateBasicType(), Value::t_u64);
BasicType __INT64(BasicType::CreateBasicType(), Value::t_i64);
BasicType __REAL(BasicType::CreateBasicType(), Value::t_real);

//BasicType POINTER(BasicType::CreateBasicType(), bi::VarIndexTypeNr);

BasicType& BasicType::get_BasicType_from_type_nr(uint8 type_nr)
{
	switch(type_nr)
	{
		case Value::t_void: return __VOID; break;
		case Value::t_bool: return __BOOL; break;
		case Value::t_u8: return __UINT8; break;
		case Value::t_i8: return __INT8; break;
		case Value::t_u16: return __UINT16; break;
		case Value::t_i16: return __INT16; break;
		case Value::t_u32: return __UINT32; break;
		case Value::t_i32: return __INT32; break;
		case Value::t_u64: return __UINT64; break;
		case Value::t_i64: return __INT64; break;
		case Value::t_real: return __REAL; break;
		default: report_error("Invalid type nr.");
	}
}

struct StructType : public Type
{
private:
	StructType(const StructType& st)
		{ report_error("Copy constructor of StructType must not be called."); }
	StructType& operator=(const StructType& st)
		{ report_error("StructType::operator=() not implemented."); }
	checkptr_UID uid;
public:
	checkptr_c<StructType> operator&() const { return checkptr_c<StructType>(this, uid); }
	virtual bool is_struct() const { return true; }
	explicit StructType(const char *__struct_name) : struct_name(__struct_name)
		{ is_already_defined=is_beeing_tested_for_beeing_defined=is_invalid=infinite_recursion=all_defined=false; }
	const char *struct_name;
	struct Elem
	{
		checkptr_c<Type> type;
		const char *name;
		Elem(checkptr_c<Type> _type, const char *_name) : type(_type), name(_name) { }
	};
	nlist<Elem> elem;
	mutable bool is_already_defined, all_defined;
	mutable bool is_beeing_tested_for_beeing_defined, is_invalid, infinite_recursion;
	void endstruct_occured()
		{ is_already_defined=true; }
	void add(checkptr_c<Type> type, char *name)
	{
		if(type.get()->equal(&__VOID))
			report_error("Cannot create a STRUCT-element of type void.");
		for(nlist<Elem>::iterator it=elem.begin(); it!=elem.end(); it++)
			if(0==strcmp(it->name, name))
				report_error(nstring("STRUCT(").append(struct_name).append(") contains at least two elements with the identical name \'").
					append(name).append("\'."));
		elem.push_back(Elem(cc->RegisterType->Register(type), name));
	}
	bool equal(checkptr_c<Type> type2) const
	{
		return ::operator==(&*this , type2);
		// Fr jede STRUCT gibt es GENAU ein zugehrendes Objekt vom Typ StructType; somit reicht es
		// fr den Vergleich aus, die Speicheradressen zu vergleichen.
		// StructType-Objekte drfen in vcopy() nicht vervielfltigt werden und nicht noch einmal in
		// cc->RegisterType->Register() dupliziert werden. Ansonsten wrde es im folgenden Fall Probleme geben:
		// Wenn eine Struktur X zuerst definiert wird und einen Zeiger auf X oder eine spter definierte Struktur Y
		// enthlt, wrde z.B. dieser Zeiger auf Y als ein anderer betrachtet werden als ein Zeiger auf Y, nachdem
		// Y tatschlich definiert wurde.
	}
	checkptr_c<Type> vcopy() const
	{
		report_error("StructType::vcopy() must not be called");
		return this; // Damit der Compiler keine Fehlermeldung ausgibt
	}
	bool check_validity_of_refered_types(bool *already_defined=0) const;
	VARIABLE_INDEX_TYPE size() const
	{
		if(is_invalid || !check_validity_of_refered_types())
			report_error(nstring("A Variable of STRUCT(").append(struct_name).append(") is beeing defined, but STRUCT(").
				append(struct_name).append(") contains other STRUCT elements or pointers to STRUCTS or arrays of STRUCTS whose STRUCT types have yet become invisible."));
		if(!is_already_defined)
			report_error(nstring("Variables of type STRUCT(").append(struct_name).append(") cannot be declared without previously defining STRUCT(").
				append(struct_name).append("). This message can also occure if an object of type STRUCT(").append(struct_name).
				append(") is a member of another STRUCT beeing instantiated."));
		if(!all_defined)
			report_error(nstring("A Variable of STRUCT(").append(struct_name).append(") is beeing defined, but STRUCT(").
				append(struct_name).append(") contains other STRUCT elements or pointers to STRUCTS or arrays of STRUCTS whose STRUCT types have not already been defined."));
		if(infinite_recursion)
			report_error(nstring("STRUCT(").append(struct_name).append(") contains at least one element of type STRUCT(").
				append(struct_name).append(") or an array or STRUCT containing element(s) of type STRUCT(").append(struct_name).
				append("). This would lead to an infinite recursion and would cause STRUCT(").append(struct_name).append(") to have infinite size."));
		VARIABLE_INDEX_TYPE size=0;
		for(nlist<Elem>::const_iterator it=elem.begin(); it!=elem.end(); it++)
			size+=it->type->size();
		return size;
	}
	nstring get_type_string() const
	{
		return nstring("STRUCT(").append(struct_name).append(")");
	}
	nlist<Elem>::const_iterator find_elem(const char *elem_name, VARIABLE_INDEX_TYPE* offset=0) const
	{
		if(offset) *offset=0;
		for(nlist<Elem>::const_iterator it=elem.begin(); it!=elem.end(); it++)
		{
			if(0==strcmp(it->name, elem_name))
				return it;
			if(offset) *offset+=it->type->size();
		}
		return nlist<Elem>::iterator();
	}
	VARIABLE_INDEX_TYPE offsetofelem(const char *elem_name) const
	{
		VARIABLE_INDEX_TYPE offset;
		find_elem(elem_name, &offset);
		return offset;
	}
	virtual bool has_only_homogenous_basic_types() const
	{
		uint8 homogenous_type_nr=Value::t_bool; // Falls es sich um einen leeren STRUCT handelt (elem.size()==0), wird in der
			// Funktion get_type_nr_of_homogenous_basic_types() Value::t_bool und in has_only_homogenous_basic_types() true geliefert.
		for(nlist<Elem>::const_iterator it=elem.begin(); it!=elem.end(); it++)
		{
			if(false==it->type->has_only_homogenous_basic_types())
				return false;
			if(elem.begin()==it)
				homogenous_type_nr=it->type->get_type_nr_of_homogenous_basic_types();
			else if(homogenous_type_nr != it->type->get_type_nr_of_homogenous_basic_types())
				return false;
		}
		return true;
	}
	virtual uint8 get_type_nr_of_homogenous_basic_types() const
	{
		uint8 homogenous_type_nr=Value::t_bool; // Falls es sich um einen leeren STRUCT handelt (elem.size()==0), wird in der
			// Funktion get_type_nr_of_homogenous_basic_types() Value::t_bool und in has_only_homogenous_basic_types() true geliefert.
		for(nlist<Elem>::const_iterator it=elem.begin(); it!=elem.end(); it++)
		{
			if(false==it->type->has_only_homogenous_basic_types())
				return Value::t_unknown_type;
			if(elem.begin()==it)
				homogenous_type_nr=it->type->get_type_nr_of_homogenous_basic_types();
			else if(homogenous_type_nr != it->type->get_type_nr_of_homogenous_basic_types())
				return Value::t_unknown_type;
		}
		return homogenous_type_nr;
	}
	virtual nlist<uint8> get_flattend_types_list() const
	{
		nlist<uint8> lst;
		for(nlist<Elem>::const_iterator it=elem.begin(); it!=elem.end(); it++)
		{
			nlist<uint8> l2=it->type->get_flattend_types_list();
			lst.insert(lst.end(), l2.begin(), l2.end());
		}
		return lst;
	}
	virtual uint32 get_nr_of_elementary_objects() const
	{
		uint32 size=0;
		for(nlist<Elem>::const_iterator it=elem.begin(); it!=elem.end(); it++)
			size+=it->type->get_nr_of_elementary_objects();
		return size;
	}
};

struct PointerType : public Type
{
private:
	PointerType& operator=(const PointerType &pt) { report_error("PointerType::operator=() not implemented"); }
	checkptr_UID uid;
public:
	checkptr_c<PointerType> operator&() { return checkptr_c<PointerType>(this, uid); }
	virtual bool is_ptr() const { return true; }
	checkptr_c<Type> type_pointed_to;
	bool C_pointer;
	PointerType(const PointerType &pt) : type_pointed_to(pt.type_pointed_to), C_pointer(pt.C_pointer) { }
	explicit PointerType(checkptr_c<Type> _type_pointed_to, bool is_C_pointer=false)
			: type_pointed_to(cc->RegisterType->Register(_type_pointed_to)), C_pointer(is_C_pointer)
		 { } // Es darf nicht cc->RegisterType->Register(this) aufgerufen werden, weil dies zu einer unendlichen Rekursion
			// PointerType::PointerType ==> cc->RegisterType->Register ==> vcopy ==> PointerType::PointerType ... fhren wrde.
	checkptr_c<Type> vcopy() const
	{
		return &*(new PointerType(*this));
	}
	bool equal(checkptr_c<Type> type2) const
	{
		return 0!=DynamicCast<PointerType>(type2) && type_pointed_to->equal(StaticCast<PointerType>(type2)->type_pointed_to)
				&& C_pointer==StaticCast<PointerType>(type2)->C_pointer;
	}
	VARIABLE_INDEX_TYPE size() const
	{
		return MAX_NR_BYTES_FOR_VARIABLE_INDEX;
	}
	nstring get_type_string() const
	{
		return nstring(C_pointer ? "CPOINTER(" : "POINTER(").append(type_pointed_to->get_type_string()).append(")");
	}
	bool check_validity_of_refered_types(bool *already_defined=0) const
	{
		if(type_pointed_to.is_dead())
			{ if(already_defined) *already_defined=false; return false; }
		else
			return type_pointed_to->check_validity_of_refered_types(already_defined);			
	}
	virtual bool has_only_homogenous_basic_types() const { return true; }
	virtual uint8 get_type_nr_of_homogenous_basic_types() const { return Value::VarIndexTypeNr; }
	virtual nlist<uint8> get_flattend_types_list() const { nlist<uint8> lst; lst.push_back(Value::VarIndexTypeNr); return lst; }
	virtual uint32 get_nr_of_elementary_objects() const { return 1; }
};
#define BYTECODE_TARGETED_COMPILER__POINTER(type) BytecodeTargetedCompiler::PointerType(&type, false)
#define BYTECODE_TARGETED_COMPILER__CPOINTER(type) BytecodeTargetedCompiler::PointerType(&type, true)

struct ArrayType : public Type
{
	ArrayType& operator=(const ArrayType &pt) { report_error("ArrayType::operator=() not implemented"); }
	checkptr_UID uid;
public:
	checkptr_c<ArrayType> operator&() { return checkptr_c<ArrayType>(this, uid); }
	checkptr_c<Type> elem_type;
	uint32 nr_elem;
	virtual bool is_array() const { return true; }
	ArrayType(checkptr_c<Type> _elem_type, uint32 _nr_elem) : elem_type(cc->RegisterType->Register(_elem_type)), nr_elem(_nr_elem)
	{
		if(elem_type.get()->equal(&__VOID))
			report_error("Cannot create an array containing elements of type void.");
	}
	ArrayType(const ArrayType& at) : elem_type(at.elem_type), nr_elem(at.nr_elem) { }
	nstring get_type_string() const
	{
		return printf_string("ARRAY(%s, %u)", elem_type->get_type_string().c_str(), nr_elem);
	}
	checkptr_c<Type> vcopy() const
	{
		return &*new ArrayType(elem_type, nr_elem);
	}
	VARIABLE_INDEX_TYPE size() const
	{
		return elem_type->size()*nr_elem;
	}
	bool equal(checkptr_c<Type> type2) const
	{
		return 0!=DynamicCast<ArrayType>(type2) && elem_type->equal(StaticCast<ArrayType>(type2)->elem_type)
				&& nr_elem==StaticCast<ArrayType>(type2)->nr_elem;
	}
	bool check_validity_of_refered_types(bool *already_defined=0) const
	{
		if(elem_type.is_dead())
			{ if(already_defined) *already_defined=false; return false; }
		else
			return elem_type->check_validity_of_refered_types(already_defined);			
	}
	virtual bool has_only_homogenous_basic_types() const
		{ return elem_type->has_only_homogenous_basic_types(); }
	virtual uint8 get_type_nr_of_homogenous_basic_types() const
		{ return elem_type->get_type_nr_of_homogenous_basic_types(); }
	virtual nlist<uint8> get_flattend_types_list() const
	{
		nlist<uint8> lst;
		nlist<uint8> l2=elem_type->get_flattend_types_list();
		for(uint32 i=0; i<nr_elem; i++)
			lst.insert(lst.end(), l2.begin(), l2.end());
		return lst;
	}
	virtual uint32 get_nr_of_elementary_objects() const
	{
		return nr_elem*elem_type->get_nr_of_elementary_objects();
	}
};

bool StructType::check_validity_of_refered_types(bool *already_defined/*=0*/) const
{	// Wenn invalid, so wird invalid=true gesetzt. Immer wird !invalid geliefert.
	if(already_defined) *already_defined=is_already_defined;
	all_defined=is_already_defined;
	infinite_recursion=false;
	typedef std::pair<checkptr_c<StructType>, nlist<Elem>::const_iterator> stackelem;
	nlist<stackelem> stack;
	stack.push_back(stackelem(&*this, elem.begin()));
	is_beeing_tested_for_beeing_defined=true;
	while(1)
	{
		nlist<Elem>::const_iterator it=stack.back().second;
		if(it==stack.back().first->elem.end())
		{
			stack.back().first->is_beeing_tested_for_beeing_defined=false;
			if(1==stack.size())
				break;
			stack.pop_back();
			++stack.back().second;
			continue;
		}
		if(it->type.is_dead())
		{
			is_invalid=true; all_defined=false;
			if(already_defined) *already_defined=false;
			stack.back().second=++it;
			continue;
		}
		if(0!=DynamicCast<StructType>(it->type))
		{
			checkptr_c<StructType> st=StaticCast<StructType>(it->type);
			//st->other_structs_refering_to_this_struct.push_back(this);
			if(!st->is_already_defined)
			{
				all_defined=false;
				if(already_defined) *already_defined=false;
			}
			if(this==st)
			{
				nlist<stackelem>::iterator it2;
				for(it2=stack.begin(); it2!=stack.end(); it2++)
					if(0!=DynamicCast<PointerType>(it2->second->type))
						break;
				if(it2==stack.end())
					infinite_recursion=true;
			}
			if(!st->is_beeing_tested_for_beeing_defined)
			{
				const nlist<Elem> *e=&st->elem;
				st->is_beeing_tested_for_beeing_defined=true;
				StaticCast<StructType>(it->type)->is_beeing_tested_for_beeing_defined=true;
				stack.push_back(stackelem(st, e->begin()));
				continue;
			}
		}
		if(0!=DynamicCast<PointerType>(it->type))
		{
			checkptr_c<Type> t=it->type;
			bool cont=false;
			while(1)
			{
				if(t.is_dead())
				{
					is_invalid=true; all_defined=false;
					if(already_defined) *already_defined=false;
					stack.back().second=++it;
					cont=true; break;
				}
				if(0!=DynamicCast<PointerType>(t))
					t=StaticCast<PointerType>(t)->type_pointed_to;
				else if(0!=DynamicCast<ArrayType>(t))
					t=StaticCast<ArrayType>(t)->elem_type;
				else
					break;
			}
			if(cont) continue;
			if(0!=DynamicCast<StructType>(t))
			{
				checkptr_c<StructType> st=StaticCast<StructType>(t);
				//st->other_structs_refering_to_this_struct.push_back(this);
				if(!st->is_already_defined)
				{
					all_defined=false;
					if(already_defined) *already_defined=false;
				}
				if(!st->is_beeing_tested_for_beeing_defined)
				{
					const nlist<Elem> *e=&st->elem;
					st->is_beeing_tested_for_beeing_defined=true;
					stack.push_back(stackelem(st, e->begin()));
					continue;
				}
			}
		}
		if(0!=DynamicCast<ArrayType>(it->type))
		{
			checkptr_c<Type> t=it->type;
			bool cont=false;
			while(1)
			{
				if(t.is_dead())
				{
					is_invalid=true; all_defined=false;
					if(already_defined) *already_defined=false;
					stack.back().second=++it;
					cont=true; break;
				}
				if(0!=DynamicCast<PointerType>(t))
					t=StaticCast<PointerType>(t)->type_pointed_to;
				else if(0!=DynamicCast<ArrayType>(t))
					t=StaticCast<ArrayType>(t)->elem_type;
				else
					break;
			}
			if(cont) continue;
			if(0!=DynamicCast<StructType>(t))
			{
				checkptr_c<StructType> st=StaticCast<StructType>(t);
				//st->other_structs_refering_to_this_struct.push_back(this);
				if(!st->is_already_defined)
				{
					all_defined=false;
					if(already_defined) *already_defined=false;
				}
				if(this==st)
				{
					nlist<stackelem>::iterator it2;
					for(it2=stack.begin(); it2!=stack.end(); it2++)
						if(0!=DynamicCast<PointerType>(it2->second->type))
							break;
					if(it2==stack.end())
						infinite_recursion=true;
				}
				if(!st->is_beeing_tested_for_beeing_defined)
				{
					const nlist<Elem> *e=&st->elem;
					st->is_beeing_tested_for_beeing_defined=true;
					stack.push_back(stackelem(st, e->begin()));
					continue;
				}
			}
		}
		stack.back().second=++it;
	}
	is_beeing_tested_for_beeing_defined=false;
	return !is_invalid;
}

checkptr_c<Type> RegisterTypeClass::Register(checkptr_c<Type> type)
{	// Es wird in keinem Fall ein Zeiger auf ein temporres Objekt zurckgeliefert.
	// Falls z.B. type auf ein temporres PointerType-Objekt zeigt, wird ggf. eine Kopie nach registered_types gespeichert
	// und ein Zeiger darauf zurckgeliefert.
	if(0==DynamicCast<PointerType>(type) && 0==DynamicCast<ArrayType>(type))
		return type;
	for(nlist<checkptr_c<Type> >::iterator it=registered_types.begin(); it!=registered_types.end(); )
	{
		nlist<checkptr_c<Type> >::iterator ait=it;
		++it;
		if(ait->is_dead() || !(*ait)->check_validity_of_refered_types())
			registered_types.erase(ait);
	}
	for(nlist<checkptr_c<Type> >::iterator it=registered_types.begin(); it!=registered_types.end(); it++)
		if((*it)->equal(type))
			return *it;
	checkptr_c<Type> pt=type->vcopy();
	registered_types.push_back(pt);
	return pt;
}

void RegisterTypeClass::cleanup()
{
	for(nlist<checkptr_c<Type> >::iterator it=registered_types.begin(); it!=registered_types.end(); it++)
		if(!it->is_dead() && (0!=DynamicCast<PointerType>(*it) || 0!=DynamicCast<ArrayType>(*it)))
			delete *it;
	registered_types.clear(); // Wichtig, da cleanup() nicht nur im Destruktor, sondern auch explizit vom Benutzer aufgerufen werden kann.
}

#define BYTECODE_TARGETED_COMPILER__ARRAY(type, nr_elem) BytecodeTargetedCompiler::ArrayType(&type, nr_elem)

typedef const BytecodeTargetedCompiler::StructType* const_StructType_Pointer;

void throw_error_if_true(bool condition, const nstring& message)
{
	if(condition)
		report_error(message);
}

#define BYTECODE_TARGETED_COMPILER__DECLARE_STRUCT(struct_name) \
	BytecodeTargetedCompiler::throw_error_if_true(0!=BytecodeTargetedCompiler::cc->ActualStructDefinition, "DECLARE_STRUCT isn't allowed inside of STRUCT ... ENDSTRUCT");\
	BytecodeTargetedCompiler::StructType struct_name(#struct_name);
#define BYTECODE_TARGETED_COMPILER__STRUCT(struct_name) \
	BytecodeTargetedCompiler::throw_error_if_true(0!=BytecodeTargetedCompiler::cc->ActualStructDefinition, "STRUCT isn't allowed inside of STRUCT ... ENDSTRUCT definition");\
	BytecodeTargetedCompiler::throw_error_if_true(0==dynamic_cast<BytecodeTargetedCompiler::const_StructType_Pointer>((&struct_name).get()), "Is not a STRUCT");\
	BytecodeTargetedCompiler::cc->ActualStructDefinition=const_cast<BytecodeTargetedCompiler::StructType*>(static_cast<BytecodeTargetedCompiler::const_StructType_Pointer>((&struct_name).get()));\
	BytecodeTargetedCompiler::throw_error_if_true(BytecodeTargetedCompiler::cc->ActualStructDefinition->is_already_defined,\
		BytecodeTargetedCompiler::printf_string("STRUCT(%s) has already beed defined.", #struct_name));
#define BYTECODE_TARGETED_COMPILER__STRUCT_ELEM(type, name) \
	{ BytecodeTargetedCompiler::throw_error_if_true(0==BytecodeTargetedCompiler::cc->ActualStructDefinition, "ELEM not allowed outside of STRUCT ... ENDSTRUCT definition");\
	BytecodeTargetedCompiler::cc->ActualStructDefinition->add(&type, #name); }
#define BYTECODE_TARGETED_COMPILER__END_STRUCT \
	{ BytecodeTargetedCompiler::throw_error_if_true(0==BytecodeTargetedCompiler::cc->ActualStructDefinition, "ENDSTRUCT without preceding corresponding STRUCT");\
	BytecodeTargetedCompiler::cc->ActualStructDefinition->endstruct_occured(); BytecodeTargetedCompiler::cc->ActualStructDefinition=0; }

struct drefpointer;
struct GeneralExpr;
struct DrefPointerPlusPlus;
struct DrefPointerMinusMinus;

void error_conversion_not_possible(const Type *quelltyp, const Type *zieltyp, const char *string_before=0)
{
	report_error(printf_string("%sCannot convert an expression of type\n   %s\ninto an expression of type\n   %s",
		(0==string_before) ? "" : string_before, quelltyp->get_type_string().c_str(), zieltyp->get_type_string().c_str()));
}

void error_conversion_to_elementary_type_not_possible(const Type *quelltyp)
{
	report_error(printf_string("Cannot convert an expression of type\n   %s\ninto an expression of elementary type.",
		quelltyp->get_type_string().c_str()));
}

struct PointerExpr : public GeneralExprType
{
private:
	expr addr; // Der Typ von addr muss immer Value::VarIndexTypeNr sein.
	bool memory_with_fixed_type_info; // Kann und darf nur dann true sein, wenn addr auf einen Speicherbereich zeigt, der
		// mit unvernderlichen, ab Typeninformationen versehen ist. Die Typeninformationen mssen zum Zeitpunkt der Konstruktion
		// von *this bereits vorhanden sein. Dies ist bei globalen Variablen und Konstanten sowie bei an Funktionen bergebenen
		// Parametern der Fall.
	PointerType pt; // Weil hier der PointerType als Element und nicht als Zeiger oder Referenz abgespeichert wird, muss
		// bei den Konstruktoren von PointerExpr nicht cc->RegisterType->Register() aufgerufen werden.
	bool dref;
public:
	bool is_dref() const { set_exported(); return dref; }
	bool is_memory_with_fixed_type_info() const { set_exported(); return memory_with_fixed_type_info; }
	const PointerType& get_pt() const { set_exported(); return pt; }
	const expr& get_addr() const { set_exported(); return addr; }
	bool is_actually_memory_with_fixed_type_info() const
	{	// Liefert nur dann true, wenn die von expression_yielding_address() gelieferte Adresse sicher auf einen Speicherbereich
		// zeigt, der mit unvernderlichen Typeninformationen versehen ist.
		set_exported();
		return dref ? false : memory_with_fixed_type_info;
	}
	GeneralExprType *vcopy() const
		{ set_exported(); return new PointerExpr(pt, addr, dref, memory_with_fixed_type_info); }
	PointerExpr(const PointerExpr& pe) : addr(pe.addr), pt(pe.pt), dref(pe.dref), memory_with_fixed_type_info(pe.memory_with_fixed_type_info)
	{
		pe.set_exported(); check_correct_address_type_nr();
		set_not_exported(); // addr wurde in check_correct_address_type_nr() von addr.get_type() auf exported gesetzt.
	}
	GeneralExpr operator=(const PointerExpr& value) const;
	GeneralExpr operator=(const CollectGeneralType& value) const;
	PointerExpr(const PointerType& ptype, const expr& ex, bool __dref=false, bool __memory_with_fixed_type_info=false)
			: pt(ptype), addr(ex), memory_with_fixed_type_info(__memory_with_fixed_type_info)
	{
		ex.set_exported(); dref=__dref; check_correct_address_type_nr();
		set_not_exported(); // addr wurde in check_correct_address_type_nr() von addr.get_type() auf exported gesetzt.
	}
	PointerExpr(const Type& ptype, const expr &ex, bool __dref=false, bool __memory_with_fixed_type_info=false)
			: pt((check_for_PointerType(&ptype), *static_cast<const PointerType*>(&ptype))), addr(ex),
				memory_with_fixed_type_info(__memory_with_fixed_type_info)
	{
		ex.set_exported();
		dref=__dref;
		check_correct_address_type_nr();
		set_not_exported(); // addr wurde in check_correct_address_type_nr() von addr.get_type() auf exported gesetzt.
	}
	void check_correct_address_type_nr()
	{	// Es ist wichtig, dass Addressen immer genau MAX_NR_BYTES_FOR_VARIABLE_INDEX Bytes haben. Denn wenn z.B. eine
		// lokale Zeigervariable bei der Definition initialisiert wird, drfen nicht weniger Bytes auf dem Stack abgelegt
		// werden, weil der Zeiger z.B. spter inkrementiert werden kann.
		if(Value::VarIndexTypeNr!=addr.get_type())
			report_error("Internal compiler error: Incorrect type of address-expression in PointerExpr.");
	}
	void check_for_PointerType(const Type *ptype)
	{
		if(0==dynamic_cast<const PointerType*>(ptype))
			report_error(printf_string("Cannot construct a PointerExpr from a non-pointer type\n   %s\nexcept for the integer constant 0.", ptype->get_type_string().c_str()));
	}
	PointerExpr operator*() const
	{
		set_exported();
		if(dref)
		{
			if(0==dynamic_cast<const PointerType*>(pt.type_pointed_to.get()))
				report_error(nstring("The dereferentiation operator * can only be applied to a pointer expression. It is actually applied to an expression of type\n   ").
						append(pt.type_pointed_to->get_type_string()));
			if(static_cast<const PointerType*>(pt.type_pointed_to.get())->type_pointed_to->equal(&__VOID))
				report_error("Cannot dereferentiate a pointer to void.");
			PointerExpr pe(*static_cast<const PointerType*>(pt.type_pointed_to.get()), addr, true);
			if(pt.C_pointer)
				pe.addr.add(Command(bi::LOADT).addnibble(Value::VarIndexTypeNr));
			else
				pe.addr.add_LOAD_optimize();
			return pe;
		}
		else
		{
			if(pt.type_pointed_to->equal(&__VOID))
				report_error("Cannot dereferentiate a pointer to void.");
			PointerExpr pe(*this); // wichtig: pt.C_pointer und memory_with_fixed_type_info mssen auch kopiert werden
			pe.dref=true;
			return pe;
		}
	}
	GeneralExpr operator[](const expr& ex) const;
	GeneralExpr operator[](long long x) const;
	GeneralExpr operator[](unsigned long long x) const;
	GeneralExpr operator[](long x) const;
	GeneralExpr operator[](unsigned long x) const;
	GeneralExpr operator[](int x) const;
	GeneralExpr operator[](unsigned int x) const;
	GeneralExpr operator[](short x) const;
	GeneralExpr operator[](unsigned short x) const;
	GeneralExpr operator[](signed char x) const;
	GeneralExpr operator[](unsigned char x) const;
	PointerExpr operator&() const
	{
		set_exported();
		if(!dref)
			report_error(nstring("The address operator & can only be used to get the address of a variable. Cannot apply & to an expression of type\n   ")
					.append(get_actual_type()->get_type_string()));
		return PointerExpr(pt, addr, false, memory_with_fixed_type_info);
	}
	GeneralExpr operator[](const char *element_name) const;
	GeneralExpr operator->*(const char *element_name) const;
	expr to_expr() const // virtuelle Funktion
	{
		set_exported();
		if(!dref || 0!=dynamic_cast<const PointerType*>(pt.type_pointed_to.get()))
			error_conversion_to_elementary_type_not_possible(get_actual_type());
		if(0!=dynamic_cast<const StructType*>(pt.type_pointed_to.get()))
			error_conversion_to_elementary_type_not_possible(get_actual_type());
		if(0!=dynamic_cast<const PointerType*>(pt.type_pointed_to.get()))
			error_conversion_to_elementary_type_not_possible(get_actual_type());
		if(0==dynamic_cast<const BasicType*>(pt.type_pointed_to.get()))
			report_error("Internal compiler error: BasicType expected.");
		expr e(addr);
		if(pt.C_pointer)
			e.add(Command(bi::LOADT).addnibble(static_cast<const BasicType*>(pt.type_pointed_to.get())->type));
		else
			e.add_LOAD_optimize();
		return e.set_type(static_cast<const BasicType*>(pt.type_pointed_to.get())->type);
	}
	bool is_expr_type() const
		{ set_exported(); return 0!=dynamic_cast<const BasicType*>(get_actual_type()); }
	const Type *get_actual_type() const
	{
		set_exported(); // Ist notwendig, weil z.B. in GeneralExpr::operator*() folgendes aufgerufen wird:
			// to_PointerExpr().get_actual_type() . Fr das von to_PointerExpr() erzeugte temporre Objekt muss auch set_exported()
			// aufgerufen werden.
		return dref ? pt.type_pointed_to : cc->RegisterType->Register(&pt);
		// Wenn fr ein temporres PointerExpr()-Objekt get_actual_type() aufgerufen wird und der Rckgabewert abgespeichert wird,
		// darf im Fall dref==false nicht einfach &pt zurckgegeben werden, sondern eine Kopie mit Hilfe von
		// cc->RegisterType->Register(). Denn nach der Zerstrung des temoprren Objekts wre der Zeiger &pt nicht mehr gltig.
	}
	uint8 get_actual_type_nr() const
	{
		const Type *t=get_actual_type();
		if(0==dynamic_cast<const BasicType*>(t))
			report_error("Internal compiler error: BasicType expected.");
		return static_cast<const BasicType*>(t)->type;
	}
	const Type *get_actual_type_pointed_to() const
	{
		set_exported();
		if(0==dynamic_cast<const PointerType*>(get_actual_type()))
			report_error("Pointer required");
		return static_cast<const PointerType*>(get_actual_type())->type_pointed_to;
	}
	nstring get_type_string() const
	{
		set_exported();
		return get_actual_type()->get_type_string();
	}
	
	expr expression_yielding_address() const
	{		// Liefert einen expr, welcher einen cast des Pointers nach expr darstellt.
		set_exported();
		if(!dref)
			return addr;
		if(0==dynamic_cast<const PointerType*>(pt.type_pointed_to.get()))
			report_error("A dereferenced pointer to a non-pointer object cannot be converted to a pointer (address).");
		expr e(addr);
		if(pt.C_pointer)
			e.add(Command(bi::LOADT).addnibble(static_cast<const BasicType*>(pt.type_pointed_to.get())->type));
		else
			e.add_LOAD_optimize();
		return e.set_type(Value::VarIndexTypeNr);
	}
	const PointerExpr& set_exported() const
	{
		addr.set_exported();
		return *this;
	}
	void set_not_exported() const
		{ addr.set_not_exported(); }
	void vset_exported() const { set_exported(); }
	void vset_not_exported() const { set_not_exported(); }
	GeneralExpr PointerExpr::operator+(const expr& ex) const;
	GeneralExpr PointerExpr::operator-(const expr& ex) const;
	GeneralExpr operator+=(const expr& x);
	GeneralExpr operator-=(const expr& x);
	GeneralExpr operator++();
	DrefPointerPlusPlus operator++(int);
	GeneralExpr operator--();
	DrefPointerMinusMinus operator--(int);
	expr operator*=(const expr& x);
	expr operator/=(const expr& x);
	expr operator%=(const expr& x);
	expr operator|=(const expr& x);
	expr operator&=(const expr& x);
	expr operator^=(const expr& x);
	expr operator<<=(const expr& x);
	expr operator>>=(const expr& x);
};

struct GeneralExpr
{
public:
	const GeneralExprType *ex;
	bool is_expr() const { set_exported(); return 0!=dynamic_cast<const expr*>(ex); }
	bool is_expr_type() const
	{
		set_exported();
		return 0!=dynamic_cast<const expr*>(ex) ||
			(is_PointerExpr() && 0!=dynamic_cast<const BasicType*>(to_PointerExpr().get_actual_type()));
	}
	bool is_PointerExpr() const
	{
		set_exported();
		return 0!=dynamic_cast<const PointerExpr*>(ex);
	}
	bool is_pointer_type() const
		{ return 0!=dynamic_cast<const PointerExpr*>(ex) && static_cast<const PointerExpr*>(ex)->get_actual_type()->is_ptr(); }
	~GeneralExpr()
	{
		if(0!=ex) delete ex; // Ggf wird hier Bytecode erzeugt !
	}
	explicit GeneralExpr(const GeneralExprType *genex)
		{ if(0!=genex) ex=genex->vcopy(); else ex=0; set_not_exported(); }
	GeneralExpr(const GeneralExpr& genex)
		{ genex.set_exported(); ex=genex.ex->vcopy(); set_not_exported(); }
	static GeneralExpr from_expr(const expr& __ex)
	{
		__ex.set_exported();
		return GeneralExpr(new expr(__ex));
	}
	static GeneralExpr from_PointerExpr(const PointerExpr& __ex)
	{
		__ex.set_exported();
		return GeneralExpr(new PointerExpr(__ex));
	}
	const GeneralExpr& set_exported() const
	{
		if(0!=dynamic_cast<const expr*>(ex)) // Hier darf nicht is_expr() aufgerufen werden, da is_expr() wieder set_exported() aufruft.
			static_cast<const expr*>(ex)->set_exported();
		if(0!=dynamic_cast<const PointerExpr*>(ex)) // Hier darf nicht is_PointerExpr() aufgerufen werden, da is_expr() wieder set_exported() aufruft.
			static_cast<const PointerExpr*>(ex)->set_exported();
		return *this;
	}
	void set_not_exported() const
	{
		if(0!=dynamic_cast<const expr*>(ex)) // Hier darf nicht is_expr() aufgerufen werden, da is_expr() wieder set_exported() aufruft.
			static_cast<const expr*>(ex)->set_not_exported();
		if(0!=dynamic_cast<const PointerExpr*>(ex)) // Hier darf nicht is_PointerExpr() aufgerufen werden, da is_expr() wieder set_exported() aufruft.
			static_cast<const PointerExpr*>(ex)->set_not_exported();
	}
	void vset_exported() const { set_exported(); }
	void vset_not_exported() const { set_not_exported(); }
	expr to_expr() const
	{
		set_exported();
		if(is_expr()) // nicht is_expr_type() !
			return *static_cast<const expr*>(ex);
		if(is_PointerExpr() && 0!=dynamic_cast<const BasicType*>(to_PointerExpr().get_actual_type()))
			return expr(to_PointerExpr().to_expr());
		report_error("Cannot convert to an expression of elementary type.");
	}
	PointerExpr to_PointerExpr() const
	{
		set_exported();
		if(!is_PointerExpr())
			report_error(printf_string("The non-pointer expression\n   %s\nnot beeing the integer constant 0 cannot be converted to a pointer expression.",
				get_type_string().c_str()));
		return *static_cast<const PointerExpr*>(ex);
	}
	GeneralExpr operator+(const GeneralExpr& obj) const
	{
		obj.set_exported();
		set_exported();
		if(is_expr_type() && obj.is_expr_type())
			return from_expr(BytecodeTargetedCompiler::operator+(to_expr(), obj.to_expr()));
		else if(is_PointerExpr() && obj.is_expr_type())
			return to_PointerExpr()+obj.to_expr();
		else if(obj.is_PointerExpr() && is_expr_type())
			return obj.to_PointerExpr()+to_expr();
		else
			report_error(printf_string("Illegal types for operator +: Cannot add expressions of types\n   %s\nand\n   %s",
				get_type_string().c_str(), obj.get_type_string().c_str()));
	}
	GeneralExpr operator+(const expr &ex) const
	{
		ex.set_exported();
		set_exported();
		if(is_expr_type())
			return from_expr(BytecodeTargetedCompiler::operator+(to_expr(), ex));
		else if(is_PointerExpr())
			return to_PointerExpr()+ex;
		else
			report_error("Illegal types for operator +");
	}
	GeneralExpr operator-(const GeneralExpr& obj) const
	{
		obj.set_exported();
		set_exported();
		if(is_expr_type() && obj.is_expr_type())
			return from_expr(BytecodeTargetedCompiler::operator-(to_expr(), obj.to_expr()));
		else if(is_PointerExpr() && obj.is_expr_type())
			return to_PointerExpr()-obj.to_expr();
		else
			report_error(printf_string("Illegal types for operator -: Cannot subtract an expression of types\n   %s\nfrom an expression of type\n   %s",
				obj.get_type_string().c_str(), get_type_string().c_str()));
	}
	GeneralExpr operator-(const expr &ex) const
	{
		ex.set_exported();
		set_exported();
		if(is_expr_type())
			return from_expr(BytecodeTargetedCompiler::operator-(to_expr(), ex));
		else if(is_PointerExpr())
			return to_PointerExpr()-ex;
		else
			report_error("Illegal types for operator -");
	}
	nstring get_type_string() const
	{
		set_exported();
		if(is_expr_type())
			return to_expr().get_type_string();
		else if(is_PointerExpr())
			return to_PointerExpr().get_type_string();
		else
			report_error("Illegal types for get_type_string()");
	}
	const Type *get_type() const
	{
		set_exported();
		if(is_expr_type())
			return &BasicType::get_BasicType_from_type_nr(to_expr().set_exported().get_type());
		else if(is_PointerExpr())
			return to_PointerExpr().get_actual_type();
		else
			report_error("Illegal types for get_type()");
	}
	PointerExpr operator*() const
	{
		set_exported();
		if(!is_PointerExpr() || 0==dynamic_cast<const PointerType*>(to_PointerExpr().get_actual_type()))
			report_error(printf_string("The operator * can only be applied to pointers. Cannot apply * to an expression of type\n   %s",
				get_type_string().c_str()));
		return *to_PointerExpr();
	}
	PointerExpr operator&() const
	{
		set_exported();
		if(!is_PointerExpr())
			report_error(nstring("The address operator & can only be used to get the address of a variable. Cannot apply & to an expression of type\n   ")
					.append(get_type_string()));
		return &to_PointerExpr();
	}
	GeneralExpr operator[](const GeneralExpr& ex) const
	{
		ex.set_exported();
		set_exported();
		if(!is_PointerExpr() || 0==dynamic_cast<const PointerType*>(to_PointerExpr().get_actual_type()))
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or pointers. Cannot apply [] to an expression of type\n   %s",
				get_type_string().c_str()));
		if(!ex.is_expr_type())
			report_error(nstring("The index y used in the operator [y] must be of elementary type. Cannot evaluate x[y] when y is of type\n   ").
				append(get_type_string()));
		return to_PointerExpr()[ex.to_expr()];
	}
	GeneralExpr operator[](const expr& ex) const
	{
		ex.set_exported();
		set_exported();
		if(!is_PointerExpr() || 0==dynamic_cast<const PointerType*>(to_PointerExpr().get_actual_type()))
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or pointers. Cannot apply [] to an expression of type\n   %s",
				get_type_string().c_str()));
		return to_PointerExpr()[ex];
	}
	GeneralExpr operator[](long long x) const
	{
		set_exported();
		if(!is_PointerExpr() || 0==dynamic_cast<const PointerType*>(to_PointerExpr().get_actual_type()))
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or pointers. Cannot apply [] to an expression of type\n   %s",
				get_type_string().c_str()));
		return to_PointerExpr()[x];
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		set_exported();
		if(!is_PointerExpr() || 0==dynamic_cast<const PointerType*>(to_PointerExpr().get_actual_type()))
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or pointers. Cannot apply [] to an expression of type\n   %s",
				get_type_string().c_str()));
		return to_PointerExpr()[x];
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](const char *element_name) const
	{
		set_exported();
		if(!is_PointerExpr()) // ob es sich wirklich um einen dereferenzierten Zeiger auf eine STRUCT handelt, wird in PointerExpr::operator[] geprft.
			report_error(printf_string("The operator [] used with a string constant index can only be applied to objects of type STRUCT. Cannot apply [\"%s\"] to an expression of type\n   %s",
					element_name, get_type_string().c_str()));
		return to_PointerExpr()[element_name];
	}
	GeneralExpr operator->*(const char *element_name) const
	{
		set_exported();
		if(!is_PointerExpr()) // ob es sich wirklich um einen Zeiger auf eine STRUCT handelt, wird in PointerExpr::operator->*() geprft.
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to an expression of type\n   %s",
				element_name, get_type_string().c_str()));
		return to_PointerExpr()->*element_name;
	}
	GeneralExpr operator=(const GeneralExpr &ex) const
	{
		ex.set_exported();
		set_exported();
		if(is_expr()) // nicht is_expr_type() !
			report_error(nstring("Cannot assign to an expression of type\n   ").append(get_type_string()));
		else if(is_PointerExpr() && ex.is_PointerExpr())
			return to_PointerExpr()=ex.to_PointerExpr();
		else if(is_PointerExpr() && ex.is_expr())
			return to_PointerExpr()=CollectGeneralType(ex.to_expr());
		else
			report_error("Illegal types for operator =");
	}
	GeneralExpr operator=(const PointerExpr &ex) const
	{
		ex.set_exported();
		set_exported();
		if(is_expr()) // nicht is_expr_type() !
			report_error(nstring("Cannot assign to an expression of type\n   ").append(get_type_string()));
		else if(is_PointerExpr())
			return to_PointerExpr()=ex;
		else
			report_error("Illegal types for operator =");
	}
	GeneralExpr operator=(const expr &ex) const
	{
		ex.set_exported();
		set_exported();
		if(is_expr()) // nicht is_expr_type() !
			report_error(nstring("Cannot assign to an expression of type\n   ").append(get_type_string()));
		else if(is_PointerExpr())
			return from_expr(to_PointerExpr()=ex);
		else
			report_error("Illegal types for operator =");
	}
	operator expr() const
	{
		set_exported();
		return to_expr();
	}
	GeneralExpr operator+=(const expr& x);
	GeneralExpr operator-=(const expr& x);
	GeneralExpr operator++();
	DrefPointerPlusPlus operator++(int);
	GeneralExpr operator--();
	DrefPointerMinusMinus operator--(int);
	expr operator*=(const expr& x);
	expr operator/=(const expr& x);
	expr operator%=(const expr& x);
	expr operator|=(const expr& x);
	expr operator&=(const expr& x);
	expr operator^=(const expr& x);
	expr operator<<=(const expr& x);
	expr operator>>=(const expr& x);
};

GeneralExpr PointerExpr::operator+(const expr& ex) const
{
	ex.set_exported();
	set_exported();
	if(0!=dynamic_cast<const PointerType*>(get_actual_type()))
	{
		if(get_actual_type_pointed_to()->equal(&__VOID))
			report_error("Cannot apply + to a pointer to void.");
		expr ex2(ex);
		if(ex2.get_type()>Value::VarIndexTypeNr)
		{
			ex2.add_effective(Command(bi::CAST).addnibble(Value::VarIndexTypeNr));
			ex2.set_type(Value::VarIndexTypeNr);
		}
		if(1==get_actual_type_pointed_to()->size())
			return GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(),
						ex2 + expression_yielding_address(), false, memory_with_fixed_type_info));
		else if(ex2.get_type()<Value::SignedVarIndexTypeNr)
			return GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(),
						(ex2*expr((SIGNED_VARIABLE_INDEX_TYPE)get_actual_type_pointed_to()->size())).shortest_type_if_constant()
							+ expression_yielding_address(),
						false, memory_with_fixed_type_info));
		else
			return GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(),
						ex2*expr::get_const_expr((int64)get_actual_type_pointed_to()->size(), true) + expression_yielding_address(),
						false, memory_with_fixed_type_info));
	}
	if(0!=dynamic_cast<const StructType*>(get_actual_type()))
		report_error("Cannot add an expression of elementary type to a struct-object.");
	if(0==dynamic_cast<const BasicType*>(get_actual_type()))
		report_error("Internal compiler error: BasicType expected.");
	return GeneralExpr::from_expr(to_expr()+ex);
}

GeneralExpr operator+(const expr& ex, const PointerExpr& p)
{
	ex.set_exported();
	p.set_exported();
	return p+ex;
}

GeneralExpr PointerExpr::operator-(const expr& ex) const
{
	ex.set_exported();
	set_exported();
	if(0!=dynamic_cast<const PointerType*>(get_actual_type()))
	{
		if(get_actual_type_pointed_to()->equal(&__VOID))
			report_error("Cannot apply - to a pointer to void.");
		expr ex2(ex);
		if(ex2.get_type()>Value::VarIndexTypeNr)
		{
			ex2.add_effective(Command(bi::CAST).addnibble(Value::VarIndexTypeNr));
			ex2.set_type(Value::VarIndexTypeNr);
		}
		if(1==get_actual_type_pointed_to()->size())
			return GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(),
					expression_yielding_address() - ex2, false, memory_with_fixed_type_info));
		else if(ex2.get_type()<Value::SignedVarIndexTypeNr)
			return GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(),
					expression_yielding_address()
						- (ex2*expr((SIGNED_VARIABLE_INDEX_TYPE)get_actual_type_pointed_to()->size())).shortest_type_if_constant(),
					false, memory_with_fixed_type_info));
		else
			return GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(),
					expression_yielding_address() - ex2*expr::get_const_expr((int64)get_actual_type_pointed_to()->size(), true),
					false, memory_with_fixed_type_info));
	}
	if(0!=dynamic_cast<const StructType*>(get_actual_type()))
		report_error("Cannot add an expression of elementary type to a struct-object.");
	if(0==dynamic_cast<const BasicType*>(get_actual_type()))
		report_error("Internal compiler error: BasicType expected.");
	return GeneralExpr::from_expr(to_expr()+ex);
}

GeneralExpr operator+(const expr &ex, const GeneralExpr& gex)
{
	ex.set_exported();
	gex.set_exported();
	return gex+ex;
}

expr operator-(const expr &ex, const GeneralExpr& gex)
{
	ex.set_exported();
	gex.set_exported();
	if(!gex.is_expr_type())
		report_error(printf_string("Illegal type for rhs of operator- : Cannot use operator- with a lhs type\n   %s\nand a rhs type\n   %s",
			ex.get_type_string().c_str(), gex.get_type_string().c_str()));
	return ex-gex.to_expr();
}

GeneralExpr PointerExpr::operator=(const CollectGeneralType& value) const
{
	set_exported();
	value.set_exported();
	if(!dref)
		report_error(printf_string("Assignment to temporary expressions of type pointer not being dereferentiated is not possible. Cannot assign an expression of type\n   %s\nto an expression of type\n   %s",
			value.get_type_string().c_str(), get_type_string().c_str()));
	if(value.is_expr_type())
	{	// Voraussetzung: dref==true
		if(get_actual_type()->is_ptr())
		{
			if(value.is_basicconst && 0==value.cval)
			{	// Zuweisung des Nullzeigers
				expr ex(pt.C_pointer ? expr((VARIABLE_INDEX_TYPE)0) : expr::get_const_expr(0));
				ex.add(addr);
				ex.add_STORE_optimize(pt.C_pointer);
				return GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(), ex));
			}
			report_error(printf_string("Cannot assign an expression of type\n   %s\nto an expression of type\n   %s",
				value.get_type_string().c_str(), get_type_string().c_str()));
		}
		if(0!=dynamic_cast<const StructType*>(pt.type_pointed_to.get()))
			report_error("Cannot assign an expression of elementary type to a struct-object.");
		if(0!=dynamic_cast<const PointerType*>(pt.type_pointed_to.get()))
			report_error(printf_string("Cannot assign the expression\n   %s\nof elementary type not beeing the integer constant 0 to a pointer of type\n   %s",
				get_type_string().c_str(), pt.get_type_string().c_str()));
		if(0==dynamic_cast<const BasicType*>(pt.type_pointed_to.get()))
			report_error("Internal compiler error: BasicType expected.");
		expr ex(pt.C_pointer ? value.to_expr() : value.to_expr().shortest_type_if_constant());
		if(pt.C_pointer &&
				static_cast<const BasicType*>(pt.type_pointed_to.get())->type!=value.to_expr().get_type())
			ex.add_effective(Command(bi::CAST).addnibble(static_cast<const BasicType*>(pt.type_pointed_to.get())->type));
		ex.add(addr);
		ex.add_STORE_optimize(pt.C_pointer);
			// Auch wenn der Zeiger auf Speicher mit unvernderlichen Typeninformationen zeigt, muss im Falle true==pt.C_pointer
			// ber bi::STORED auf den Speicher zugegriffen werden, denn die Daten knnen dann trotzdem unabhngig von den
			// Variablen(typen)grenzen beschrieben werden.
		return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(pt.type_pointed_to.get())->type)); // ber einen
			// POINTER darf ein Speicherbereich nur dann beschrieben werden, 
	}
	else if(value.is_PointerExpr())
	{	// Voraussetzung: dref==true
		PointerExpr pe(value.to_PointerExpr());
		if(!get_actual_type()->equal(pe.get_actual_type()))
		{
			if(get_actual_type()->equal(&__BOOL) && 0!=dynamic_cast<const PointerType*>(value.get_type()))
			{
				// Konversion und Zuweisung eines Zeigers an einen dereferenzierten Zeiger auf bool.
				expr ex;
				ex.add(pe.expression_yielding_address());
				if(pt.C_pointer)
					ex.add_effective(Command(bi::CAST).addnibble(Value::t_bool));
				ex.add(addr);
				ex.add_STORE_optimize(pt.C_pointer);
				ex.set_type(Value::t_bool);
				return GeneralExpr::from_expr(ex);
			}
			if(get_actual_type()->is_ptr() && value.is_basicconst && 0==value.cval)
			{	// Zuweisung des Nullzeigers
				expr ex(pt.C_pointer ? expr((VARIABLE_INDEX_TYPE)0) : expr::get_const_expr(0));
				ex.add(addr);
				ex.add_STORE_optimize(pt.C_pointer);
				return GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(), ex));
			}
			if(get_actual_type()->is_ptr() && get_actual_type_pointed_to()->equal(&__VOID) && pe.get_actual_type()->is_ptr())
			{
				// Im Folgenden wird ein Zeiger an einen dereferenzierten Zeiger auf einen Zeiger auf __VOID zugewiesen.
				expr ex(pe.expression_yielding_address());
				ex.add(addr); // Nicht expression_yielding_address()! dref==true
				ex.add_STORE_optimize(pt.C_pointer);
				return GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(), ex, false,
					pe.is_actually_memory_with_fixed_type_info()));
			}
			report_error(printf_string("Type mismatch in assignment to the object the pointer points to. Cannot assign an expression of type\n   %s\nto an expression of type\n   %s",
				pe.get_actual_type()->get_type_string().c_str(), get_actual_type()->get_type_string().c_str()));
		}
		if(pe.dref && 0!=dynamic_cast<const BasicType*>(pe.pt.type_pointed_to.get()))
		{
			if(!pt.type_pointed_to->equal(pe.pt.type_pointed_to))
				error_conversion_not_possible(pe.get_actual_type(), get_actual_type(), "operator=(): ");
			// Zuweisung eines expr an einen dereferenzierten Zeiger auf ein expr
			return GeneralExpr::from_expr(operator=(pe.to_expr())); // Rekursiver Aufruf von operator=() mit einem expr-Parameter
		}
		if(0!=dynamic_cast<const StructType*>(pe.get_actual_type()) || 0!=dynamic_cast<const ArrayType*>(pe.get_actual_type()))
			report_error(printf_string("Cannot assign objects of type STRUCT or ARRAY to variables or dereferenced pointers; please implement the assignment by assigning element-wise. Cannot assign an expression of type\n   %s\nto a variable.",
				pe.get_actual_type()->get_type_string().c_str()));
		if(0==dynamic_cast<const PointerType*>(pe.get_actual_type()))
			error_conversion_not_possible(pe.get_actual_type(), get_actual_type(), "\'=\': ");
		// Im Folgenden wird ein Zeiger an einen dereferenzierten Zeiger auf einen Zeiger zugewiesen.
		expr ex(pe.expression_yielding_address());
		ex.add(addr); // Nicht expression_yielding_address()! dref==true
		ex.add_STORE_optimize(pt.C_pointer);
		return GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(), ex, false, pe.is_actually_memory_with_fixed_type_info()));
	}
	else
		report_error("Unknown type of CollectGeneralType");
}

GeneralExpr PointerExpr::operator=(const PointerExpr& pe) const
{	// Muss definiert werden, obwohl es schon  PointerExpr::operator=(const CollectGeneralType& value) const  gibt,
	// denn ansonsten wird vom Compiler eine Standard-Zuweisungsroutine generiert.
	set_exported();
	pe.set_exported();
	if(!dref)
		report_error(printf_string("Assignment to temporary expressions of type pointer not being dereferentiated is not possible. Cannot assign an expression of type\n   %s\nto an expression of type\n   %s",
			pe.get_type_string().c_str(), get_type_string().c_str()));
	if(!get_actual_type()->equal(pe.get_actual_type()))
	{
		if(get_actual_type()->equal(&__BOOL) && 0!=dynamic_cast<const PointerType*>(pe.get_actual_type()))
		{
			// Konversion und Zuweisung eines Zeigers an einen dereferenzierten Zeiger auf bool.
			expr ex;
			ex.add(pe.expression_yielding_address());
			if(pt.C_pointer)
				ex.add_effective(Command(bi::CAST).addnibble(Value::t_bool));
			ex.add(addr);
			ex.add_STORE_optimize(pt.C_pointer);
			ex.set_type(Value::t_bool);
			return GeneralExpr::from_expr(ex);
		}
		if(get_actual_type()->is_ptr() && get_actual_type_pointed_to()->equal(&__VOID) && pe.get_actual_type()->is_ptr())
		{
			// Im Folgenden wird ein Zeiger an einen dereferenzierten Zeiger auf einen Zeiger auf __VOID zugewiesen.
			expr ex(pe.expression_yielding_address());
			ex.add(addr); // Nicht expression_yielding_address()! dref==true
			ex.add_STORE_optimize(pt.C_pointer);
			return GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(), ex, false, pe.is_actually_memory_with_fixed_type_info()));
		}
		report_error(printf_string("Type mismatch in assignment to the object the pointer points to. Cannot assign an expression of type\n   %s\nto an expression of type\n   %s",
			pe.get_actual_type()->get_type_string().c_str(), get_actual_type()->get_type_string().c_str()));
	}
	if(pe.dref && 0!=dynamic_cast<const BasicType*>(pe.pt.type_pointed_to.get()))
	{
		if(!pt.type_pointed_to->equal(pe.pt.type_pointed_to))
			error_conversion_not_possible(pe.get_actual_type(), get_actual_type(), "operator=(): ");
		// Zuweisung eines expr an einen dereferenzierten Zeiger auf ein expr
		return GeneralExpr::from_expr(operator=(pe.to_expr())); // Rekursiver Aufruf von operator=() mit einem expr-Parameter
	}
	if(0!=dynamic_cast<const StructType*>(pe.get_actual_type()) || 0!=dynamic_cast<const ArrayType*>(pe.get_actual_type()))
		report_error(printf_string("Cannot assign objects of type STRUCT or ARRAY to variables or dereferenced pointers; please implement the assignment by assigning element-wise. Cannot assign an expression of type\n   %s\nto a variable.",
			pe.get_actual_type()->get_type_string().c_str()));
	if(0==dynamic_cast<const PointerType*>(pe.get_actual_type()))
		error_conversion_not_possible(pe.get_actual_type(), get_actual_type(), "\'=\': ");
	// Im Folgenden wird ein Zeiger an einen dereferenzierten Zeiger auf einen Zeiger zugewiesen.
	expr ex(pe.expression_yielding_address());
	ex.add(addr); // Nicht expression_yielding_address()! dref==true
	ex.add_STORE_optimize(pt.C_pointer);
	return GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(), ex, false, pe.is_actually_memory_with_fixed_type_info()));
}

GeneralExpr PointerExpr::operator[](const expr& ex) const
{
	ex.set_exported();
	set_exported();
	return GeneralExpr::from_PointerExpr(*GeneralExpr((*this)+ex));
}

GeneralExpr PointerExpr::operator[](long long x) const
{
	set_exported();
	if(0==dynamic_cast<const PointerType*>(get_actual_type()))
		report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or pointers. Cannot apply [] to an expression of type\n   %s",
			get_type_string().c_str()));
	if(get_actual_type_pointed_to()->equal(&__VOID))
		report_error("Cannot apply [] to a pointer to void.");
	if(0==x)
		return GeneralExpr::from_PointerExpr(*GeneralExpr::from_PointerExpr(*this));
	int64 y=(x*((int64)get_actual_type_pointed_to()->size()));
	return GeneralExpr::from_PointerExpr(*GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(),
		expression_yielding_address() + expr::get_const_expr((uint64)y), false, memory_with_fixed_type_info)));
}
GeneralExpr PointerExpr::operator[](unsigned long long x) const
{
	set_exported();
	if(0==dynamic_cast<const PointerType*>(get_actual_type()))
		report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or pointers. Cannot apply [] to an expression of type\n   %s",
			get_type_string().c_str()));
	if(get_actual_type_pointed_to()->equal(&__VOID))
		report_error("Cannot apply [] to a pointer to void.");
	if(0==x)
		return GeneralExpr::from_PointerExpr(*GeneralExpr::from_PointerExpr(*this));
	uint64 y=(x*((uint64)get_actual_type_pointed_to()->size()));
	return GeneralExpr::from_PointerExpr(*GeneralExpr::from_PointerExpr(PointerExpr(*get_actual_type(),
		expression_yielding_address() + expr::get_const_expr((uint64)y), false, memory_with_fixed_type_info)));
}
GeneralExpr PointerExpr::operator[](int x) const
{
	return operator[]((long long)x);
}
GeneralExpr PointerExpr::operator[](unsigned int x) const
{
	return operator[]((unsigned long long)x);
}
GeneralExpr PointerExpr::operator[](short x) const
{
	return operator[]((long long)x);
}
GeneralExpr PointerExpr::operator[](unsigned short x) const
{
	return operator[]((unsigned long long)x);
}
GeneralExpr PointerExpr::operator[](signed char x) const
{
	return operator[]((long long)x);
}
GeneralExpr PointerExpr::operator[](unsigned char x) const
{
	return operator[]((unsigned long long)x);
}

GeneralExpr PointerExpr::operator[](const char *element_name) const
{
	set_exported();
	const Type *type=get_actual_type();
	if(0==dynamic_cast<const StructType*>(type))
		report_error(printf_string("The operator [] used with a string constant index can only be applied to objects of type STRUCT. Cannot apply [\"%s\"] to an expression of type\n   %s",
			element_name, type->get_type_string().c_str()));
	nlist<StructType::Elem>::const_iterator it=static_cast<const StructType*>(type)->find_elem(element_name);
	if(!it.is_valid_and_not_end(static_cast<const StructType*>(type)->elem))
		report_error(printf_string("operator[]: '%s' in not an element of %s", element_name, type->get_type_string().c_str()));
	expr ex2;
	ex2.add(addr); // Es muss dref==true sein, sonst knnte get_actual_type() kein STRUCT sein.
	ex2.add(expr::get_const_expr(static_cast<const StructType*>(type)->offsetofelem(element_name)));
	ex2.addcmd(bi::ADD, -1);
	return GeneralExpr::from_PointerExpr(
		PointerExpr(PointerType(it->type, pt.C_pointer), ex2.set_type(Value::VarIndexTypeNr), true, memory_with_fixed_type_info));
}

GeneralExpr PointerExpr::operator->*(const char *element_name) const
{
	set_exported();
	if(0==dynamic_cast<const PointerType*>(get_actual_type()))
		report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to an expression of type\n   %s",
			element_name, get_type_string().c_str()));
	const Type *type=get_actual_type_pointed_to();
	if(0==dynamic_cast<const StructType*>(type))
		report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to an expression of type\n   %s",
			element_name, type->get_type_string().c_str()));
	nlist<StructType::Elem>::const_iterator it=static_cast<const StructType*>(type)->find_elem(element_name);
	if(!it.is_valid_and_not_end(static_cast<const StructType*>(type)->elem))
		report_error(printf_string("operator->*: '%s' in not an element of %s", element_name, type->get_type_string().c_str()));
	expr ex2;
	ex2.add(expression_yielding_address());
	ex2.add(expr::get_const_expr(static_cast<const StructType*>(type)->offsetofelem(element_name)));
	ex2.addcmd(bi::ADD, -1);
	return GeneralExpr::from_PointerExpr(
		PointerExpr(PointerType(it->type, static_cast<const PointerType*>(get_actual_type())->C_pointer),
			ex2.set_type(Value::VarIndexTypeNr), true));
}

GeneralExpr PointerExpr::operator+=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!dref)
		report_error(printf_string("operator += : l-value required on the left hand side. Cannot evaluate += with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(is_expr_type())
	{
		expr ex(addr); // dref==true
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::ADDM, -1);
		ex.set_code_has_effect();
		return GeneralExpr::from_expr(ex.set_type(get_actual_type_nr()));
	}
	else // *this ist ein dereferenzierter Zeiger auf einen Zeiger. Da dieser Fall selten eintritt, wird auf ADDM verzichtet.
		return operator=(operator+(x));
}

struct DrefPointerPlusPlus
{
private:
	void operator=(const DrefPointerPlusPlus& lpp) { report_error("Assignment DrefPointerPlusPlus::operator=() must not be called"); } // Zugriff verhindern
	DrefPointerPlusPlus(const DrefPointerPlusPlus& dp) : p(dp.p) { report_error("Copy constructor DrefPointerPlusPlus::DrefPointerPlusPlus() must not be called"); } // Zugriff verhindern
	friend DrefPointerPlusPlus PointerExpr::operator++(int);
	friend DrefPointerPlusPlus GeneralExpr::operator++(int);
public:
	const PointerExpr& p; // Hier muss eine Referenz verwendet werden, da ansonsten im Konstruktor LocalMinusMinus::LocalMinusMinus()
		// der Konstruktor local::local aufgerufen wird, welcher nicht lediglich den Speicher der Struktur local kopiert, sondern
		// annimmt, der Benutzer htte eine neue local angelegt und somit entsprechend Speicher auf dem Stack reserviert.
	mutable bool exported;
	const DrefPointerPlusPlus& set_exported() const { exported=true; return *this; }
	DrefPointerPlusPlus(const PointerExpr& _p) : p(_p)
	{
		if(!p.is_dref())
			report_error("Internal compiler Error in DrefPointerPlusPlus::DrefPointerPlusPlus : dref=true expected.");
		exported=false;
	}
	~DrefPointerPlusPlus()
	{
		if(!exported)
		{	// Der folgende Code ist identisch wie beim Prfix Operator -- und auch effizienter als der Code des Postfix Operators --,
			// kann jedoch nur verwendet werden, wenn der Wert des Ausdrucks nicht weiter verwendet wird.
			if(!p.get_actual_type()->is_expr() && !p.get_actual_type()->is_ptr())
				report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
			expr s;
			if(p.get_actual_type()->is_ptr())
			{
				if(static_cast<const PointerType*>(p.get_actual_type())->type_pointed_to->equal(&__VOID))
					report_error("Cannot apply ++ to a pointer to void.");
				s.assign(expr::get_const_expr((VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(p.get_actual_type())->type_pointed_to->size()));
			}
			else // get_actual_type()->is_expr()==true
				s.assign(expr::get_const_expr((uint32)1));
			// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator--() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.

			expr ex(p.get_addr()); // dref==true
			ex.add(s);
			ex.addcmd(bi::ADDM, -1);
			ex.set_code_has_effect();
			ex.set_exported();
			if(p.get_actual_type()->is_ptr())
				ExprObjectDestructionHandler<expr>(ex.set_type(Value::VarIndexTypeNr));
			else if(p.get_actual_type()->is_expr())
				ExprObjectDestructionHandler<expr>(ex.set_type(static_cast<const BasicType*>(p.get_actual_type())->type));
		}
	}
	GeneralExpr to_GeneralExpr() const
	{
		set_exported();
		if(!p.get_actual_type()->is_expr() && !p.get_actual_type()->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
		expr s;
		if(p.get_actual_type()->is_ptr())
		{
			if(static_cast<const PointerType*>(p.get_actual_type())->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply ++ to a pointer to void.");
			s.assign(expr::get_const_expr(static_cast<const PointerType*>(p.get_actual_type())->type_pointed_to->size()));
		}
		else // get_actual_type()->is_expr()==true
			s.assign(expr::get_const_expr((uint32)1));

		expr ex;
		ex.add(p.get_addr()); // Es darf keine Methode benutzt werden, bei der get_addr() zweimal eingefgt wird. Denn
		ex.add(s);			  // in get_addr() knnten Seiteneffekte auftreten.
		ex.addcmd(bi::ADDM, -1);
		ex.add(s);
		ex.addcmd(bi::SUB, -1);
		ex.set_exported();
		ex.set_code_has_effect();
		if(p.get_actual_type()->is_ptr())
			return GeneralExpr::from_PointerExpr(PointerExpr(*p.get_actual_type(), ex.set_type(Value::VarIndexTypeNr), false, true));
		else if(p.get_actual_type()->is_expr())
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(p.get_actual_type())->type));
		else
			report_error("DrefPointerPlusPlus Error");
	}
	operator expr() const
		{ return to_GeneralExpr().to_expr(); }
	PointerExpr to_PointerExpr() const
		{ return to_GeneralExpr().to_PointerExpr(); }
	GeneralExpr operator->*(const char *element_name) const
	{	// ->* : Dereferenzierung des Zeigers und Elementauswahl des sich ergebenden STRUCT-Objekts.
		set_exported();
		if(0==dynamic_cast<const PointerType*>(p.get_actual_type()) ||
				0==dynamic_cast<const StructType*>(static_cast<const PointerType*>(p.get_actual_type())->type_pointed_to.get()))
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to a variable of type\n   %s",
				element_name, p.get_actual_type()->get_type_string().c_str()));
		return (to_PointerExpr())->*element_name;
	}
	GeneralExpr operator[](const expr& ex) const
	{
		set_exported();
		if(!p.get_actual_type()->is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.get_actual_type()->get_type_string().c_str()));
		return (to_PointerExpr())[ex];
	}
	GeneralExpr operator[](long long x) const
	{
		set_exported();
		if(!p.get_actual_type()->is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.get_actual_type()->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		set_exported();
		if(!p.get_actual_type()->is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.get_actual_type()->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }
};

struct DrefPointerMinusMinus
{
private:
	void operator=(const DrefPointerMinusMinus& lpp) { report_error("Assignment DrefPointerMinusMinus::operator=() must not be called"); } // Zugriff verhindern
	DrefPointerMinusMinus(const DrefPointerMinusMinus& dp) : p(dp.p) { report_error("Copy constructor DrefPointerMinusMinus::DrefPointerMinusMinus() must not be called"); } // Zugriff verhindern
	friend DrefPointerMinusMinus PointerExpr::operator--(int);
	friend DrefPointerMinusMinus GeneralExpr::operator--(int);
public:
	const PointerExpr& p; // Hier muss eine Referenz verwendet werden, da ansonsten im Konstruktor LocalMinusMinus::LocalMinusMinus()
		// der Konstruktor local::local aufgerufen wird, welcher nicht lediglich den Speicher der Struktur local kopiert, sondern
		// annimmt, der Benutzer htte eine neue local angelegt und somit entsprechend Speicher auf dem Stack reserviert.
	mutable bool exported;
	const DrefPointerMinusMinus& set_exported() const { exported=true; return *this; }
	DrefPointerMinusMinus(const PointerExpr& _p) : p(_p)
	{
		if(!p.is_dref())
			report_error("Internal compiler Error in DrefPointerMinusMinus::DrefPointerMinusMinus : dref=true expected.");
		exported=false;
	}
	~DrefPointerMinusMinus()
	{
		if(!exported)
		{	// Der folgende Code ist identisch wie beim Prfix Operator -- und auch effizienter als der Code des Postfix Operators --,
			// kann jedoch nur verwendet werden, wenn der Wert des Ausdrucks nicht weiter verwendet wird.
			if(!p.get_actual_type()->is_expr() && !p.get_actual_type()->is_ptr())
				report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
			expr s;
			if(p.get_actual_type()->is_ptr())
			{
				if(static_cast<const PointerType*>(p.get_actual_type())->type_pointed_to->equal(&__VOID))
					report_error("Cannot apply ++ to a pointer to void.");
				s.assign(expr::get_const_expr((VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(p.get_actual_type())->type_pointed_to->size()));
			}
			else // get_actual_type()->is_expr()==true
				s.assign(expr::get_const_expr((uint32)1));
			// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator--() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.

			expr ex(p.get_addr()); // dref==true
			ex.add(s);
			ex.addcmd(bi::SUBM, -1);
			ex.set_exported();
			ex.set_code_has_effect();
			if(p.get_actual_type()->is_ptr())
				ExprObjectDestructionHandler<expr>(ex.set_type(Value::VarIndexTypeNr));
			else if(p.get_actual_type()->is_expr())
				ExprObjectDestructionHandler<expr>(ex.set_type(static_cast<const BasicType*>(p.get_actual_type())->type));
		}
	}
	GeneralExpr to_GeneralExpr() const
	{
		set_exported();
		if(!p.get_actual_type()->is_expr() && !p.get_actual_type()->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
		expr s;
		if(p.get_actual_type()->is_ptr())
		{
			if(static_cast<const PointerType*>(p.get_actual_type())->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply ++ to a pointer to void.");
			s.assign(expr::get_const_expr(static_cast<const PointerType*>(p.get_actual_type())->type_pointed_to->size()));
		}
		else // get_actual_type()->is_expr()==true
			s.assign(expr::get_const_expr((uint32)1));

		expr ex;
		ex.add(p.get_addr()); // Es darf keine Methode benutzt werden, bei der get_addr() zweimal eingefgt wird. Denn
		ex.add(s);			  // in get_addr() knnten Seiteneffekte auftreten.
		ex.addcmd(bi::SUBM, -1);
		ex.add(s);
		ex.addcmd(bi::ADD, -1);
		ex.set_exported();
		ex.set_code_has_effect();
		if(p.get_actual_type()->is_ptr())
			return GeneralExpr::from_PointerExpr(PointerExpr(*p.get_actual_type(), ex.set_type(Value::VarIndexTypeNr), false, true));
		else if(p.get_actual_type()->is_expr())
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(p.get_actual_type())->type));
		else
			report_error("DrefPointerMinusMinus Error");
	}
	operator expr() const
		{ return to_GeneralExpr().to_expr(); }
	PointerExpr to_PointerExpr() const
		{ return to_GeneralExpr().to_PointerExpr(); }
	GeneralExpr operator->*(const char *element_name) const
	{	// ->* : Dereferenzierung des Zeigers und Elementauswahl des sich ergebenden STRUCT-Objekts.
		set_exported();
		if(0==dynamic_cast<const PointerType*>(p.get_actual_type()) ||
				0==dynamic_cast<const StructType*>(static_cast<const PointerType*>(p.get_actual_type())->type_pointed_to.get()))
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to a variable of type\n   %s",
				element_name, p.get_actual_type()->get_type_string().c_str()));
		return (to_PointerExpr())->*element_name;
	}
	GeneralExpr operator[](const expr& ex) const
	{
		set_exported();
		if(!p.get_actual_type()->is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.get_actual_type()->get_type_string().c_str()));
		return (to_PointerExpr())[ex];
	}
	GeneralExpr operator[](long long x) const
	{
		set_exported();
		if(!p.get_actual_type()->is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.get_actual_type()->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		set_exported();
		if(!p.get_actual_type()->is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.get_actual_type()->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }
};

GeneralExpr PointerExpr::operator++()
{
	if(!dref)
		report_error(printf_string("operator ++ : l-value required on the left hand side. Cannot apply ++ to an operand of type\n   %s",
			get_type_string().c_str()));
	if(is_expr_type())
	{
		expr ex(addr); // dref==true
		ex.add(expr::get_const_expr(1));
		ex.addcmd(bi::ADDM, -1);
		ex.set_code_has_effect();
		return GeneralExpr::from_expr(ex.set_type(get_actual_type_nr()));
	}
	else // *this ist ein dereferenzierter Zeiger auf einen Zeiger. Da dieser Fall selten eintritt, wird auf ADDM verzichtet.
		return operator=(operator+(expr::get_const_expr(1)));
}
DrefPointerPlusPlus PointerExpr::operator++(int)
{
	if(!dref)
		report_error(printf_string("operator ++ : l-value required on the left hand side. Cannot apply ++ to an operand of type\n   %s",
			get_type_string().c_str()));
	return DrefPointerPlusPlus(*this);
}
GeneralExpr PointerExpr::operator--()
{
	if(!dref)
		report_error(printf_string("operator -- : l-value required on the left hand side. Cannot apply -- to an operand of type\n   %s",
			get_type_string().c_str()));
	if(is_expr_type())
	{
		expr ex(addr); // dref==true
		ex.add(expr::get_const_expr(1));
		ex.addcmd(bi::SUBM, -1);
		ex.set_code_has_effect();
		return GeneralExpr::from_expr(ex.set_type(get_actual_type_nr()));
	}
	else // *this ist ein dereferenzierter Zeiger auf einen Zeiger. Da dieser Fall selten eintritt, wird auf ADDM verzichtet.
		return operator=(operator-(expr::get_const_expr(1)));
}
DrefPointerMinusMinus PointerExpr::operator--(int)
{
	if(!dref)
		report_error(printf_string("operator -- : l-value required on the left hand side. Cannot apply -- to an operand of type\n   %s",
			get_type_string().c_str()));
	return DrefPointerMinusMinus(*this);
}

GeneralExpr PointerExpr::operator-=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!dref)
		report_error(printf_string("operator -= : l-value required on the left hand side. Cannot evaluate -= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(is_expr_type())
	{
		expr ex(addr); // dref==true
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::SUBM, -1);
		ex.set_code_has_effect();
		return GeneralExpr::from_expr(ex.set_type(get_actual_type_nr()));
	}
	else // *this ist ein dereferenzierter Zeiger auf einen Zeiger. Da dieser Fall selten eintritt, wird auf SUBM verzichtet.
		return operator=(operator-(x));
}
expr PointerExpr::operator*=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!dref)
		report_error(printf_string("operator *= : l-value required on the left hand side. Cannot evaluate *= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("PointerExpr::operator *= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	expr ex(addr); // dref==true
	ex.add(x.shortest_type_if_const_keep_signed());
	ex.addcmd(bi::MULM, -1);
	ex.set_code_has_effect();
	return ex.set_type(get_actual_type_nr());
}
expr PointerExpr::operator/=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!dref)
		report_error(printf_string("operator /= : l-value required on the left hand side. Cannot evaluate /= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("PointerExpr::operator /= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
#ifndef WITH_INTEGER64_DIVISION
	if(get_actual_type_nr()==Value::t_u64 || get_actual_type_nr()==Value::t_i64)
		report_error("64 Bit division is deactivated in the bytecode interpreter. To activate it, #define WITH_INTEGER64_DIVISION");
#endif
	expr ex(addr); // dref==true
	ex.add(x.shortest_type_if_const_keep_signed());
	ex.addcmd(bi::DIVM, -1);
	ex.set_code_has_effect();
	return ex.set_type(get_actual_type_nr());
}
expr PointerExpr::operator%=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!dref)
		report_error(printf_string("operator %= : l-value required on the left hand side. Cannot evaluate %= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("PointerExpr::operator %= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
#ifndef WITH_INTEGER64_MODULO
	if(get_actual_type_nr()==Value::t_u64 || get_actual_type_nr()==Value::t_i64)
		report_error("64 Bit modulo is deactivated in the bytecode interpreter. To activate it, #define WITH_INTEGER64_MODULO");
#endif
	expr ex(addr); // dref==true
	ex.add(x.shortest_type_if_const_keep_signed());
	ex.addcmd(bi::MODM, -1);
	ex.set_code_has_effect();
	return ex.set_type(get_actual_type_nr());
}
expr PointerExpr::operator|=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!dref)
		report_error(printf_string("operator |= : l-value required on the left hand side. Cannot evaluate |= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("PointerExpr::operator |= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	expr ex(addr); // dref==true
	ex.add(x.shortest_type_if_constant());
	ex.addcmd(bi::ORM, -1);
	ex.set_code_has_effect();
	return ex.set_type(get_actual_type_nr());
}
expr PointerExpr::operator&=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!dref)
		report_error(printf_string("operator &= : l-value required on the left hand side. Cannot evaluate &= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("PointerExpr::operator &= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	expr ex(addr); // dref==true
	ex.add(x.shortest_type_if_constant());
	ex.addcmd(bi::ANDM, -1);
	ex.set_code_has_effect();
	return ex.set_type(get_actual_type_nr());
}
expr PointerExpr::operator^=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!dref)
		report_error(printf_string("operator ^= : l-value required on the left hand side. Cannot evaluate += with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("PointerExpr::operator ^= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	expr ex(addr); // dref==true
	ex.add(x.shortest_type_if_constant());
	ex.addcmd(bi::XORM, -1);
	ex.set_code_has_effect();
	return ex.set_type(get_actual_type_nr());
}
expr PointerExpr::operator<<=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!dref)
		report_error(printf_string("operator <<= : l-value required on the left hand side. Cannot evaluate <<= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("PointerExpr::operator <<= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	expr ex(addr); // dref==true
	ex.add(x.shortest_type_if_constant());
	ex.addcmd(bi::SLM, -1);
	ex.set_code_has_effect();
	return ex.set_type(get_actual_type_nr());
}
expr PointerExpr::operator>>=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!dref)
		report_error(printf_string("operator >>= : l-value required on the left hand side. Cannot evaluate >>= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("PointerExpr::operator >>= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	expr ex(addr); // dref==true
	ex.add(x.shortest_type_if_constant());
	ex.addcmd(bi::SRM, -1);
	ex.set_code_has_effect();
	return ex.set_type(get_actual_type_nr());
}

GeneralExpr GeneralExpr::operator+=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator += : l-value required on the left hand side. Cannot evaluate += with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(is_expr_type())
	{
		PointerExpr pe(to_PointerExpr());
		expr ex(pe.get_addr()); // pe.dref==true
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::ADDM, -1);
		ex.set_code_has_effect();
		return GeneralExpr::from_expr(ex.set_type(pe.get_actual_type_nr()));
	}
	else
		return operator=(operator+(x));
}
GeneralExpr GeneralExpr::operator-=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator -= : l-value required on the left hand side. Cannot evaluate -= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(is_expr_type())
	{
		PointerExpr pe(to_PointerExpr());
		expr ex(pe.get_addr()); // pe.dref==true
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::SUBM, -1);
		ex.set_code_has_effect();
		return GeneralExpr::from_expr(ex.set_type(pe.get_actual_type_nr()));
	}
	else
		return operator=(operator-(x));
}

GeneralExpr GeneralExpr::operator++()
{
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator ++ : l-value required on the left hand side. Cannot apply ++ to an operand of type\n   %s",
			get_type_string().c_str()));
	return ++to_PointerExpr();
}
DrefPointerPlusPlus GeneralExpr::operator++(int)
{
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator ++ : l-value required on the left hand side. Cannot apply ++ to an operand of type\n   %s",
			get_type_string().c_str()));
	return to_PointerExpr()++;
}
GeneralExpr GeneralExpr::operator--()
{
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator -- : l-value required on the left hand side. Cannot apply -- to an operand of type\n   %s",
			get_type_string().c_str()));
	return --to_PointerExpr();
}
DrefPointerMinusMinus GeneralExpr::operator--(int)
{
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator -- : l-value required on the left hand side. Cannot apply -- to an operand of type\n   %s",
			get_type_string().c_str()));
	return to_PointerExpr()--;
}

expr GeneralExpr::operator*=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator *= : l-value required on the left hand side. Cannot evaluate *= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("operator *= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	PointerExpr pe(to_PointerExpr());
	expr ex(pe.get_addr()); // pe.dref==true
	ex.add(x.shortest_type_if_const_keep_signed());
	ex.addcmd(bi::MULM, -1);
	ex.set_code_has_effect();
	return ex.set_type(pe.get_actual_type_nr());
}
expr GeneralExpr::operator/=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator /= : l-value required on the left hand side. Cannot evaluate /= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("operator /= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	PointerExpr pe(to_PointerExpr());
#ifndef WITH_INTEGER64_DIVISION
	if(pe.get_actual_type_nr()==Value::t_u64 || pe.get_actual_type_nr()==Value::t_i64)
		report_error("64 Bit division is deactivated in the bytecode interpreter. To activate it, #define WITH_INTEGER64_DIVISION");
#endif
	expr ex(pe.get_addr()); // pe.dref==true
	ex.add(x.shortest_type_if_const_keep_signed());
	ex.addcmd(bi::DIVM, -1);
	ex.set_code_has_effect();
	return ex.set_type(pe.get_actual_type_nr());
}
expr GeneralExpr::operator%=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator %= : l-value required on the left hand side. Cannot evaluate %= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("operator %= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	PointerExpr pe(to_PointerExpr());
#ifndef WITH_INTEGER64_MODULO
	if(pe.get_actual_type_nr()==Value::t_u64 || pe.get_actual_type_nr()==Value::t_i64)
		report_error("64 Bit modulo is deactivated in the bytecode interpreter. To activate it, #define WITH_INTEGER64_MODULO");
#endif
	expr ex(pe.get_addr()); // pe.dref==true
	ex.add(x.shortest_type_if_const_keep_signed());
	ex.addcmd(bi::MODM, -1);
	ex.set_code_has_effect();
	return ex.set_type(pe.get_actual_type_nr());
}
expr GeneralExpr::operator|=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator |= : l-value required on the left hand side. Cannot evaluate |= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("operator |= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	PointerExpr pe(to_PointerExpr());
	expr ex(pe.get_addr()); // pe.dref==true
	ex.add(x.shortest_type_if_constant());
	ex.addcmd(bi::ORM, -1);
	ex.set_code_has_effect();
	return ex.set_type(pe.get_actual_type_nr());
}
expr GeneralExpr::operator&=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator &= : l-value required on the left hand side. Cannot evaluate &= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("operator &= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	PointerExpr pe(to_PointerExpr());
	expr ex(pe.get_addr()); // pe.dref==true
	ex.add(x.shortest_type_if_constant());
	ex.addcmd(bi::ANDM, -1);
	ex.set_code_has_effect();
	return ex.set_type(pe.get_actual_type_nr());
}
expr GeneralExpr::operator^=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator ^= : l-value required on the left hand side. Cannot evaluate ^= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("operator ^= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	PointerExpr pe(to_PointerExpr());
	expr ex(pe.get_addr()); // pe.dref==true
	ex.add(x.shortest_type_if_constant());
	ex.addcmd(bi::XORM, -1);
	ex.set_code_has_effect();
	return ex.set_type(pe.get_actual_type_nr());
}
expr GeneralExpr::operator<<=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator <<= : l-value required on the left hand side. Cannot evaluate <<= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("operator <<= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	PointerExpr pe(to_PointerExpr());
	expr ex(pe.get_addr()); // pe.dref==true
	ex.add(x.shortest_type_if_constant());
	ex.addcmd(bi::SLM, -1);
	ex.set_code_has_effect();
	return ex.set_type(pe.get_actual_type_nr());
}
expr GeneralExpr::operator>>=(const expr& x)
{
	x.set_exported();
	set_exported();
	if(!is_PointerExpr() || !to_PointerExpr().is_dref())
		report_error(printf_string("operator >>= : l-value required on the left hand side. Cannot evaluate >>= with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	if(!is_expr_type())
		report_error(printf_string("operator >>= : Cannot evaluate with operands of types\n   %s\nand\n   %s",
			get_type_string().c_str(), x.get_type_string().c_str()));
	PointerExpr pe(to_PointerExpr());
	expr ex(pe.get_addr()); // pe.dref==true
	ex.add(x.shortest_type_if_constant());
	ex.addcmd(bi::SRM, -1);
	ex.set_code_has_effect();
	return ex.set_type(pe.get_actual_type_nr());
}

struct VariablePlusPlus;
struct VariableMinusMinus;

uint64 real2uint64_endianness_independent(real_type x)
{
	uint64 y=(4==sizeof(real_type)) ? (*((uint32*)&x)) : (*((uint64*)&x));
	uint64 z=0;
	unsigned short test_little_endian=1u;
	bool compiler_runs_on_little_endian=((unsigned char)1u) == *((unsigned char*)(&test_little_endian));
	for(long j=0; j<sizeof(real_type); j++)
	{
		long k=compiler_runs_on_little_endian ? j : (sizeof(real_type)-1-j);
		z|=((y>>(8*k))&255)<<(8*j);
	}
	return z;
}

PointerExpr STRING(const char *C_char_pointer);

struct DataEndMarker { };
struct DataInitialization
{
private:
	DataInitialization() { }
	friend PointerExpr STRING(const char *C_char_pointer);
public:
	bool end;
	uint64 ival;
	real_type rval;
	uint8 type_nr;
	char *str;
	DataInitialization(bool x) : ival((uint64)x), rval(0), type_nr(Value::t_bool) , str(0), end(false) { end_marker_occured=false; }
	DataInitialization(signed char x) : ival((uint64)x), rval(0), type_nr(Value::t_char) , str(0), end(false) { end_marker_occured=false; }
	DataInitialization(unsigned char x) : ival((uint64)x), rval(0), type_nr(Value::t_uchar) , str(0), end(false) { end_marker_occured=false; }
	DataInitialization(short x) : ival((uint64)x), rval(0), type_nr(Value::t_short) , str(0), end(false) { end_marker_occured=false; }
	DataInitialization(unsigned short x) : ival((uint64)x), rval(0), type_nr(Value::t_ushort) , str(0), end(false) { end_marker_occured=false; }
	DataInitialization(int x) : ival((uint64)x), rval(0), type_nr(Value::t_int) , str(0), end(false) { end_marker_occured=false; }
	DataInitialization(unsigned int x) : ival((uint64)x), rval(0), type_nr(Value::t_uint) , str(0), end(false) { end_marker_occured=false; }
	DataInitialization(long x) : ival((uint64)x), rval(0), type_nr(Value::t_long) , str(0), end(false) { end_marker_occured=false; }
	DataInitialization(unsigned long x) : ival((uint64)x), rval(0), type_nr(Value::t_ulong) , str(0), end(false) { end_marker_occured=false; }
	DataInitialization(long long x) : ival((uint64)x), rval(0), type_nr(Value::t_longlong) , str(0), end(false) { end_marker_occured=false; }
	DataInitialization(unsigned long long x) : ival((uint64)x), rval(0), type_nr(Value::t_ulonglong) , str(0), end(false) { end_marker_occured=false; }
	DataInitialization(float x) : rval((real_type)x), ival(0), type_nr(Value::t_real) , str(0), end(false) { end_marker_occured=false; }
	DataInitialization(double x) : rval((real_type)x), ival(0), type_nr(Value::t_real) , str(0), end(false) { end_marker_occured=false; }
	//DataInitialization(char *__str) : rval(0), ival(0), type_nr(Value::t_void), str(__str) , end(false) { end_marker_occured=false; }
	DataInitialization(const DataEndMarker& dem) : rval(0), ival(0), type_nr(Value::t_void), str(0), end(true) { end_marker_occured=true; }
	static bool end_marker_occured;

	static uint32 get_size(const DataInitialization *di)
	{
		if(!di->end && !end_marker_occured)
			report_error("DataInitialization without DataEndMarker");
		uint32 size=0;
		while(!di->end)
			{ di++; size++; }
		return size;
	}
};
bool DataInitialization::end_marker_occured=false;

struct Variable
{
private:
	Variable(const Variable& v)
		{ report_error("Copy constructor Variable::Variable(const Variable& v) must not be called."); } // Zugriff verhindern
	friend struct VariablePlusPlus;
	friend struct VariableMinusMinus;
public:
	static VARIABLE_INDEX_TYPE var_mem_count[3*Value::anz_user_types+3];
	static VARIABLE_INDEX_TYPE var_mem_begin[3*Value::anz_user_types+3];
	VARIABLE_INDEX_TYPE var_addr;
	const Type *type;
	static nlist<uint64> init_values[Value::anz_user_types+1];
	static nlist<uint8> type_info[3];
	void construct(const Type *__type, nlist<uint64> InitValues=nlist<uint64>())
	{
		type=__type;
		if(0!=InitValues.size() && InitValues.size()!=type->get_flattend_types_list().size())
			report_error(printf_string("The number %d of initialization parameters of elementary type used to initialize a variable of type\n   %s\ndiffers from the required number %u",
				InitValues.size(), type->get_type_string().c_str(), type->get_nr_of_elementary_objects()));
		if(type->has_only_homogenous_basic_types())
		{
			uint8 type_nr;
			if(0!=dynamic_cast<const BasicType*>(type))
				type_nr=static_cast<const BasicType*>(type)->type;
			else if(0!=dynamic_cast<const PointerType*>(type))
				type_nr=Value::VarIndexTypeNr;
			else if(type->has_only_homogenous_basic_types())
				type_nr=type->get_type_nr_of_homogenous_basic_types();

			if(Value::t_void==type_nr)
				report_error("Cannot create a variable of type void");
			long t2=(0==InitValues.size()) ? 1 : 0, t=(Value::anz_user_types+1)*t2+type_nr;
			var_addr=var_mem_count[t];
			var_mem_count[t]+=type->size();
			if(1==cc->CompilerPassNr&& 0!=InitValues.size())
				init_values[type_nr].insert(init_values[type_nr].end(), InitValues.begin(), InitValues.end());
		}
		else
		{
			long t2=(0==InitValues.size()) ? 1 : 0, t=(Value::anz_user_types+1)*t2+Value::anz_user_types;
			var_addr=var_mem_count[t];
			var_mem_count[t]+=type->size();
			if(1==cc->CompilerPassNr)
			{
				nlist<uint64>::iterator ivalit=InitValues.begin();
				nlist<uint8> lst=type->get_flattend_types_list();
				for(nlist<uint8>::iterator it=lst.begin(); it!=lst.end(); it++)
				{
					uint8 type_nr=*it;
					if(ivalit!=InitValues.end())
						for(long i=0; i<Value::sbytesize(type_nr); i++)
							init_values[Value::anz_user_types].push_back(((*ivalit)>>(8*i)) & 255);

					switch(type_nr)
					{
						case Value::t_bool:
						case Value::t_real:
							type_info[t2].push_back(1); break;
						case Value::t_u8:
						case Value::t_u16:
						case Value::t_u32:
						case Value::t_u64:
							type_info[t2].push_back(2); break;
						case Value::t_i8:
						case Value::t_i16:
						case Value::t_i32:
						case Value::t_i64:
							type_info[t2].push_back(3); break;
					}
					for(long i=1; i<Value::sbytesize(type_nr); i++)
						type_info[t2].push_back(0);
					if(ivalit!=InitValues.end())
						++ivalit;
				}
			}
		}
	}
	void construct(const Type *__type, uint64 x)
	{
		if(0==dynamic_cast<const BasicType*>(__type))
			report_error("construct() called with one initialization parameter for a non-elementary type.");
		nlist<uint64> lst;
		if(0!=x)
		{
			if(static_cast<const BasicType*>(__type)->type==Value::t_real)
				lst.push_back(real2uint64_endianness_independent((real_type)x));
			else
				lst.push_back(x);
		}
		construct(__type, lst);
	}
	void construct(const Type *__type, int64 x)
	{
		if(0==dynamic_cast<const BasicType*>(__type))
			report_error("construct() called with one initialization parameter for a non-elementary type.");
		nlist<uint64> lst;
		if(0!=x)
		{
			if(static_cast<const BasicType*>(__type)->type==Value::t_real)
				lst.push_back(real2uint64_endianness_independent((real_type)x));
			else
				lst.push_back(x);
		}
		construct(__type, lst);
	}
	void construct_from_real(const Type *__type, real_type x)
	{
		if(0==dynamic_cast<const BasicType*>(__type))
			report_error("construct_from_real() called with one initialization parameter for a non-elementary type.");
		nlist<uint64> lst;
		if(0!=x)
		{
			if(static_cast<const BasicType*>(__type)->type==Value::t_real)
				lst.push_back(real2uint64_endianness_independent(x));
			else if(Value::s_is_signed_integer(static_cast<const BasicType*>(__type)->type))
				lst.push_back((int64)x);
			else
				lst.push_back((uint64)x);
		}
		construct(__type, lst);
	}
	Variable(const BasicType& type, unsigned long long init_value) { construct(&type, init_value); }
	Variable(const BasicType& type, long long init_value) { construct(&type, init_value); }
	Variable(const BasicType& type, unsigned long init_value) { construct(&type, (uint64)init_value); }
	Variable(const BasicType& type, long init_value) { construct(&type, (int64)init_value); }
	Variable(const BasicType& type, unsigned int init_value) { construct(&type, (uint64)init_value); }
	Variable(const BasicType& type, int init_value) { construct(&type, (int64)init_value); }
	Variable(const BasicType& type, unsigned short init_value) { construct(&type, (uint64)init_value); }
	Variable(const BasicType& type, short init_value) { construct(&type, (int64)init_value); }
	Variable(const BasicType& type, unsigned char init_value) { construct(&type, (uint64)init_value); }
	Variable(const BasicType& type, signed char init_value) { construct(&type, (int64)init_value); }
	Variable(const BasicType& type, bool init_value) { construct(&type, (uint64)init_value); }
#ifndef real_type_is_integer_type
	Variable(const BasicType& type, real_type init_value) { construct_from_real(&type, init_value); }
#endif
	Variable(const BasicType& type) { construct(&type); }

	Variable(const CollectGeneralType& init_value)
	{
		init_value.set_exported();
		if(init_value.is_expr_type())
		{
			if(!init_value.is_basicconst)
				report_error("Global variables can only have constants as initialisers.");
			construct(&BasicType::get_BasicType_from_type_nr(init_value.to_expr().get_type()), init_value.cval);
		}
		else if(init_value.is_PointerExpr())
		{
			if(!init_value.is_basicconst || !Value::s_is_integer_type(init_value.to_expr().get_type()) || 0!=init_value.cval)
				report_error("Global variables of type pointer can only have the integer constant 0 as initialiser.");
			construct(init_value.get_type(), (uint64)0); // Nullzeiger
		}
		else
			report_error("Unknown type in CollectGeneralType");
	}

	bool is_expr() const { return 0!=dynamic_cast<const BasicType*>(type); }
	bool is_ptr() const { return 0!=dynamic_cast<const PointerType*>(type); }

	Variable(const PointerType& __type)
	{
		__type.chk_validity_of_ref_types_variable_creation();
		type=cc->RegisterType->Register(&__type);
		construct(type); // Zeiger wird automatisch mit 0 initialisiert. Hier darf nicht construct(type, (uint64)0) aufgerufen werden !
	}

	nlist<uint64> prepare_data_init(const DataInitialization *di) const
	{	// ACHTUNG: type muss schon initialisiert sein !
		if(DataInitialization::get_size(di)!=type->get_nr_of_elementary_objects())
			report_error(printf_string("The number %d of initialization parameters of elementary type used to initialize a variable of type\n   %s\ndiffers from the required number %u",
				DataInitialization::get_size(di), type->get_type_string().c_str(), type->get_nr_of_elementary_objects()));
		nlist<uint8> types=type->get_flattend_types_list();
		nlist<uint64> init_values;
		for(nlist<uint8>::iterator it=types.begin(); it!=types.end() && !di->end; it++, di++)
		{
			if(Value::t_real==*it)
			{
				if(Value::t_real==di->type_nr)
					init_values.push_back(real2uint64_endianness_independent(di->rval));
				else if(Value::s_is_signed_integer(di->type_nr))
					init_values.push_back(real2uint64_endianness_independent((real_type)((int64)di->ival)));
				else
					init_values.push_back(real2uint64_endianness_independent((real_type)((uint64)di->ival)));					
			}
			else if(Value::s_is_signed_integer(*it))
			{
				if(Value::t_real==di->type_nr)
					init_values.push_back((uint64)((int64)di->rval));
				else
					init_values.push_back((uint64)di->ival);
			}
			else
			{
				if(Value::t_real==di->type_nr)
					init_values.push_back((uint64)di->rval);
				else
					init_values.push_back((uint64)di->ival);
			}
		}
		return init_values;
	}

	Variable(const ArrayType& __type)
	{
		__type.chk_validity_of_ref_types_variable_creation();
		type=cc->RegisterType->Register(&__type);
		construct(type); // hier darf NICHT mehr &__type verwendet werden !
	}
	Variable(const ArrayType& __type, const DataInitialization *di)
	{
		__type.chk_validity_of_ref_types_variable_creation();
		type=cc->RegisterType->Register(&__type);
		construct(type, prepare_data_init(di)); // hier darf NICHT mehr &__type verwendet werden !
	}
	Variable(const StructType& __type)
	{
		__type.chk_validity_of_ref_types_variable_creation();
		type=cc->RegisterType->Register(&__type);
		construct(type); // hier darf NICHT mehr &__type verwendet werden !
	}
	Variable(const StructType& __type, const DataInitialization *di)
	{
		__type.chk_validity_of_ref_types_variable_creation();
		type=cc->RegisterType->Register(&__type);
		construct(type, prepare_data_init(di)); // hier darf NICHT mehr &__type verwendet werden !
	}
	
	expr expression_yielding_address() const
	{
		if(!type->is_ptr())
			report_error("expression_yielding_address() called for non-pointer.");
		expr ex(expr((VARIABLE_INDEX_TYPE)var_addr));
		return ex.set_type(Value::VarIndexTypeNr);
	}
	PointerExpr to_PointerExpr() const
	{
		if(!is_ptr())
			report_error(printf_string("Conversion to Pointer not possible. Cannot convert a variable of type\n   %s\ninto a pointer type.",
				type->get_type_string().c_str()));
		return PointerExpr(*type, expression_yielding_address(), false, true); // PointerExpr::memory_with_fixed_type==true
	}

	operator expr() const
	{
		if(!is_expr()) // Ein Pointer knnte zwar in einen __BOOL umgewandelt werden. Hierfr ist jedoch die Umwandlung nach LogExpr vorgesehen.
			report_error(printf_string("Cannot convert a variable of type\n   %s\ninto an expression of elementary type.",
				type->get_type_string().c_str()));
		expr ex(expr::get_const_expr((uint32)var_addr));
		ex.addcmd(bi::LOAD);
		return ex;
	}

	expr operator=(const Variable& v) const
	{	// Es ist wichtig, diesen Zuweisungsoperator (mit Quelltyp=Zieltyp) zu definieren, weil er vom Standard-Zuweisungsoperator abweicht.
		return operator=(CollectGeneralType(v));
	}

	expr get_assignment_code(const expr& value) const
	{
		value.set_exported();
		uint8 type_nr=value.get_type();
		expr ex(value.shortest_type_if_constant());
		// Value braucht nicht auf den passenden Typ gecastet zu werden, weil in der an die Funktion bergebenen Variablen immer
		// bereits der richtige Typ steht und ein cast ggf. vom Kommando SETLOCAL durchgefhrt wird.
		if(0!=dynamic_cast<const BasicType*>(type))
			type_nr=static_cast<const BasicType*>(type)->type;
		if(0!=dynamic_cast<const PointerType*>(type))
			type_nr=Value::VarIndexTypeNr;
		ex.add(expr::get_const_expr((uint32)var_addr));
		ex.addcmd(bi::STORE, -1);
		ex.set_code_has_effect();
		return ex.set_type(type_nr);
	}

	GeneralExpr operator=(const CollectGeneralType& value) const
	{
		value.set_exported();
		if(value.is_expr_type())
		{
			if(type->is_ptr())
			{
				if(value.is_basicconst && 0==value.cval) // Zuweisung des Nullzeigers
					return GeneralExpr::from_PointerExpr(PointerExpr(*type, get_assignment_code(expr((VARIABLE_INDEX_TYPE)0))));
				report_error(printf_string("Cannot assign an expression of type\n   %s\nto an expression of type\n   %s",
					value.get_type_string().c_str(), type->get_type_string().c_str()));
			}
			if(0!=dynamic_cast<const StructType*>(type))
				report_error("Cannot assign an expression of elementary type to a struct-object.");
			if(0!=dynamic_cast<const PointerType*>(type))
				report_error(printf_string("Cannot assign the expression\n   %s\nof elementary type not beeing the integer constant 0 to a pointer of type\n   %s",
					value.get_type_string().c_str(), type->get_type_string().c_str()));
			if(0==dynamic_cast<const BasicType*>(type))
				report_error("Internal compiler error: BasicType expected.");
			return GeneralExpr::from_expr(get_assignment_code(value.to_expr()));
		}
		else if(value.is_PointerExpr())
		{
			PointerExpr pe(value.to_PointerExpr());
			if(!type->equal(pe.get_actual_type()))
			{
				if(type->equal(&__BOOL) && 0!=dynamic_cast<const PointerType*>(value.get_type()))
				{
					// Konversion und Zuweisung eines Zeigers an eine bool-Variable
					expr ex;
					ex.add(pe.expression_yielding_address());
					ex.add_effective(Command(bi::CAST).addnibble(Value::t_bool));
					ex.set_type(Value::t_bool);
					return GeneralExpr::from_expr(get_assignment_code(ex));
				}
				if(type->is_ptr() && value.is_basicconst && 0==value.cval) // Zuweisung des Nullzeigers
					return GeneralExpr::from_PointerExpr(PointerExpr(*type, get_assignment_code(expr((VARIABLE_INDEX_TYPE)0))));
				if(type->is_ptr() && static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID) && pe.get_actual_type()->is_ptr())
				{
					// Im Folgenden wird ein Zeiger an einen Zeiger auf __VOID zugewiesen.
					return GeneralExpr::from_PointerExpr(PointerExpr(*type, get_assignment_code(pe.expression_yielding_address())));
				}
				report_error(printf_string("Type mismatch in assignment to the object the pointer points to. Cannot assign an expression of type\n   %s\nto an expression of type\n   %s",
					pe.get_actual_type()->get_type_string().c_str(), type->get_type_string().c_str()));
			}
			if(pe.is_dref() && 0!=dynamic_cast<const BasicType*>(pe.get_pt().type_pointed_to.get()))
			{
				if(!type->equal(pe.get_pt().type_pointed_to))
					error_conversion_not_possible(pe.get_actual_type(), type, "operator=(): ");
				// Zuweisung eines expr an einen dereferenzierten Zeiger auf ein expr
				return GeneralExpr::from_expr(operator=(pe.to_expr())); // Rekursiver Aufruf von operator=() mit einem expr-Parameter
			}
			if(0!=dynamic_cast<const StructType*>(pe.get_actual_type()) || 0!=dynamic_cast<const ArrayType*>(pe.get_actual_type()))
				report_error(printf_string("Cannot assign objects of type STRUCT or ARRAY to variables or dereferenced pointers; please implement the assignment by assigning element-wise. Cannot assign an expression of type\n   %s\nto a variable.",
					pe.get_actual_type()->get_type_string().c_str()));
			if(0==dynamic_cast<const PointerType*>(pe.get_actual_type()))
				error_conversion_not_possible(pe.get_actual_type(), type, "\'=\': ");
			// Im Folgenden wird ein Zeiger an einen Zeiger zugewiesen.
			return GeneralExpr::from_PointerExpr(PointerExpr(*type, get_assignment_code(pe.expression_yielding_address()),
				false, pe.is_actually_memory_with_fixed_type_info()));
		}
		else
			report_error("Unknown type of CollectGeneralType");
	}

	PointerExpr operator&()
	{
		return PointerExpr(PointerType(type), expr((VARIABLE_INDEX_TYPE)var_addr).set_type(Value::VarIndexTypeNr),
			false, true); // PointerExpr::memory_with_fixed_type_info==true
	}

	PointerExpr operator*()
	{
		if(0==dynamic_cast<const PointerType*>(type))
			report_error("The dereferentiation operator * can only be applied to expressions of type pointer.");
		if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
			report_error("Cannot dereferentiate a pointer to void.");
		expr ex;
		ex.add(expr::get_const_expr((uint32)var_addr));
		ex.addcmd(bi::LOAD, 0);
		return PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr), true);
	}

	GeneralExpr operator+=(const CollectGeneralType& value)
	{
		value.set_exported();
		if(value.is_PointerExpr() && !value.is_expr_type()) // Ein PointerExpr kann auch einen BasicType ergeben (falls dref=true).
			report_error(printf_string("operator+=() : Cannot apply a pointer on the right hand side. Cannot apply the type\n   %s\non the right hand side.",
				value.get_type_string().c_str()));
		if(!value.is_expr_type()) // nicht value.is_expr() !
			report_error("Unknown Type in CollectGeneralType()");
		if(!type->is_expr() && !type->is_ptr())
			report_error(printf_string("Cannot add to an object of type ARRAY or STRUCT. Cannot add an expression of type\n   %s",
				value.get_type_string().c_str()));
		expr s;
		if(type->is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply += to a pointer to void.");
			expr v2(value.to_expr().shortest_type_if_constant());
			// v2 muss nicht ggf. auf Value::VarIndexTypeNr gecastet werden, weil Voraussetzung zur Anwendung des operator+=() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
			if(1==static_cast<const PointerType*>(type)->type_pointed_to->size())
				s.assign(v2);
			else if(v2.get_type()<Value::SignedVarIndexTypeNr)
				s.assign(v2*expr((SIGNED_VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(type)->type_pointed_to->size()));
			else
				s.assign(v2*expr::get_const_expr((int64)static_cast<const PointerType*>(type)->type_pointed_to->size(), true));
			expr ex(expr::get_const_expr(var_addr));
			ex.add(s.shortest_type_if_constant());
			ex.addcmd(bi::ADDM, -1);
			ex.set_code_has_effect();
			return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr), false, true));
			// PointerExpr::memory_with_fixed_type==true
		}
		else // type->is_expr()==true
		{
			s.assign(value.to_expr().shortest_type_if_constant()); // s muss nicht ggf. auf Value::VarIndexTypeNr gecastet werden, weil Voraussetzung zur Anwendung des operator+=() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
			expr ex(expr::get_const_expr((uint32)var_addr));
			ex.add(s.shortest_type_if_constant());
			ex.addcmd(bi::ADDM, -1);
			ex.set_code_has_effect();
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
	}
	GeneralExpr operator++()
	{
		if(!type->is_expr() && !type->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
		if(type->is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply ++ to a pointer to void.");
			expr ex(expr::get_const_expr(var_addr));
			ex.add(expr::get_const_expr((VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(type)->type_pointed_to->size()));
			ex.addcmd(bi::ADDM, -1);
			ex.set_code_has_effect();
			return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr), false, true));
			// PointerExpr::memory_with_fixed_type==true
		}
		else // type->is_expr()==true
		{
			// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator++() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
			expr ex(expr::get_const_expr((uint32)var_addr));
			ex.add(expr::get_const_expr((uint32)1));
			ex.addcmd(bi::ADDM, -1);
			ex.set_code_has_effect();
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
	}
	VariablePlusPlus operator++(int);

	GeneralExpr operator-=(const CollectGeneralType& value)
	{
		value.set_exported();
		if(value.is_PointerExpr() && !value.is_expr_type()) // Ein PointerExpr kann auch einen BasicType ergeben (falls dref=true).
			report_error(printf_string("operator+=() : Cannot apply a pointer on the right hand side. Cannot apply the type\n   %s\non the right hand side.",
				value.get_type_string().c_str()));
		if(!value.is_expr_type()) // nicht value.is_expr() !
			report_error("Unknown Type in CollectGeneralType()");
		if(!type->is_expr() && !type->is_ptr())
			report_error(printf_string("Cannot subtract from an object of type ARRAY or STRUCT. Cannot subtract an expression of type\n   %s",
				value.get_type_string().c_str()));
		expr s;
		if(type->is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply += to a pointer to void.");
			expr v2(value.to_expr().shortest_type_if_constant());
			// v2 muss nicht ggf. auf Value::VarIndexTypeNr gecastet werden, weil Voraussetzung zur Anwendung des operator+=() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
			if(1==static_cast<const PointerType*>(type)->type_pointed_to->size())
				s.assign(v2);
			else if(v2.get_type()<Value::SignedVarIndexTypeNr)
				s.assign(v2*expr((SIGNED_VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(type)->type_pointed_to->size()));
			else
				s.assign(v2*expr::get_const_expr((int64)static_cast<const PointerType*>(type)->type_pointed_to->size(), true));
			expr ex(expr::get_const_expr(var_addr));
			ex.add(s.shortest_type_if_constant());
			ex.addcmd(bi::SUBM, -1);
			ex.set_code_has_effect();
			return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr), false, true));
			// PointerExpr::memory_with_fixed_type==true
		}
		else // type->is_expr()==true
		{
			s.assign(value.to_expr().shortest_type_if_constant()); // s muss nicht ggf. auf Value::VarIndexTypeNr gecastet werden, weil Voraussetzung zur Anwendung des operator+=() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
			expr ex(expr::get_const_expr((uint32)var_addr));
			ex.add(s.shortest_type_if_constant());
			ex.addcmd(bi::SUBM, -1);
			ex.set_code_has_effect();
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
	}
	GeneralExpr operator--()
	{
		if(!type->is_expr() && !type->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
		if(type->is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply ++ to a pointer to void.");
			expr ex(expr::get_const_expr(var_addr));
			ex.add(expr::get_const_expr((VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(type)->type_pointed_to->size()));
			ex.addcmd(bi::SUBM, -1);
			ex.set_code_has_effect();
			return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr), false, true));
			// PointerExpr::memory_with_fixed_type==true
		}
		else // type->is_expr()==true
		{
			// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator++() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
			expr ex(expr::get_const_expr((uint32)var_addr));
			ex.add(expr::get_const_expr((uint32)1));
			ex.addcmd(bi::SUBM, -1);
			ex.set_code_has_effect();
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
	}
	VariableMinusMinus operator--(int);

	expr operator*=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator*=(): Multiplication can only be done with two operands of elementary type. Cannot apply << to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.exported=true;
		expr ex(expr::get_const_expr((uint32)var_addr));
		ex.add(x.shortest_type_if_const_keep_signed());
		ex.addcmd(bi::MULM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator/=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator/=(): Division can only be done with two operands of elementary type. Cannot apply << to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.exported=true;
#ifndef WITH_INTEGER64_DIVISION
		if(static_cast<const BasicType*>(type)->type==Value::t_u64 || static_cast<const BasicType*>(type)->type==Value::t_i64)
			report_error("64 Bit division is deactivated in the bytecode interpreter. To activate it, #define WITH_INTEGER64_DIVISION");
#endif
		expr ex(expr::get_const_expr((uint32)var_addr));
		ex.add(x.shortest_type_if_const_keep_signed());
		ex.addcmd(bi::DIVM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator%=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator%=(): Modulo can only be done with two operands of elementary type. Cannot apply << to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.exported=true;
#ifndef WITH_INTEGER64_MODULO
		if(static_cast<const BasicType*>(type)->type==Value::t_u64 || static_cast<const BasicType*>(type)->type==Value::t_i64)
			report_error("64 Bit modulo is deactivated in the bytecode interpreter. To activate it, #define WITH_INTEGER64_MODULO");
#endif
		expr ex(expr::get_const_expr((uint32)var_addr));
		ex.add(x.shortest_type_if_const_keep_signed());
		ex.addcmd(bi::MODM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator|=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator|=(): This operation can only be done with two operands of elementary type. Cannot apply << to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.exported=true;
		expr ex(expr::get_const_expr((uint32)var_addr));
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::ORM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator&=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator&=(): This operation can only be done with two operands of elementary type. Cannot apply << to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.exported=true;
		expr ex(expr::get_const_expr((uint32)var_addr));
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::ANDM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator^=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator^=(): This operation can only be done with two operands of elementary type. Cannot apply << to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.exported=true;
		expr ex(expr::get_const_expr((uint32)var_addr));
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::XORM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator<<=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator<<=(): Shift can only be done with two operands of elementary type. Cannot apply << to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.exported=true;
		expr ex(expr::get_const_expr((uint32)var_addr));
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::SLM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator>>=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator>>=(): Shift can only be done with two operands of elementary type. Cannot apply << to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.exported=true;
		expr ex(expr::get_const_expr((uint32)var_addr));
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::SRM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	GeneralExpr operator[](const char *element_name) const
	{	// [] : Elementauswahl eines STRUCTs.
		if(0==dynamic_cast<const StructType*>(type))
			report_error(printf_string("The operator [] used with a string constant index can only be applied to objects of type STRUCT. Cannot apply [\"%s\"] to a variable of type\n   %s",
				element_name, type->get_type_string().c_str()));
		nlist<StructType::Elem>::const_iterator it=static_cast<const StructType*>(type)->find_elem(element_name);
		if(!it.is_valid_and_not_end(static_cast<const StructType*>(type)->elem))
			report_error(printf_string("operator[]: '%s' in not an element of %s", element_name, type->get_type_string().c_str()));
		expr ex2;
		ex2.add(expr((VARIABLE_INDEX_TYPE)(static_cast<const StructType*>(type)->offsetofelem(element_name)+var_addr)));
		return GeneralExpr::from_PointerExpr(
			PointerExpr(PointerType(it->type), ex2.set_type(Value::VarIndexTypeNr), true, true));
	}
	GeneralExpr operator->*(const char *element_name) const
	{	// ->* : Dereferenzierung des Zeigers und Elementauswahl des sich ergebenden STRUCT-Objekts.
		if(0==dynamic_cast<const PointerType*>(type) ||
				0==dynamic_cast<const StructType*>(static_cast<const PointerType*>(type)->type_pointed_to.get()))
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to a variable of type\n   %s",
				element_name, type->get_type_string().c_str()));
		const Type *ptype=static_cast<const PointerType*>(type)->type_pointed_to;
		const StructType *stype=static_cast<const StructType*>(ptype);
		nlist<StructType::Elem>::const_iterator it=stype->find_elem(element_name);
		if(!it.is_valid_and_not_end(stype->elem))
			report_error(printf_string("operator->*: '%s' in not an element of %s", element_name, type->get_type_string().c_str()));
		expr ex(expr::get_const_expr((uint32)var_addr));
		ex.addcmd(bi::LOAD, 0);
		ex.add(expr::get_const_expr(stype->offsetofelem(element_name)));
		ex.addcmd(bi::ADD, -1);
		//return GeneralExpr::from_PointerExpr(PointerExpr(PointerType(it->type), ex.set_type(Value::VarIndexTypeNr), true));
		return GeneralExpr::from_PointerExpr(
			PointerExpr(PointerType(it->type, static_cast<const PointerType*>(type)->C_pointer),
			ex.set_type(Value::VarIndexTypeNr), true));
	}
	GeneralExpr operator[](const expr& ex) const
	{
		ex.set_exported();
		if(0!=dynamic_cast<const ArrayType*>(type))
		{
			expr ex2;
			ex2.add(expr((VARIABLE_INDEX_TYPE)var_addr));
			return PointerExpr(PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr),
				false, true)[ex]; // PointerExpr::memory_with_fixed_type==true
		}
		else if(is_ptr())
			return PointerExpr(*type, expression_yielding_address().set_type(Value::VarIndexTypeNr), false, true)[ex];
		else
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or to pointers. Cannot apply [] to a variable of type\n   %s",
				type->get_type_string().c_str()));
	}
	GeneralExpr operator[](long long x) const
	{
		if(0!=dynamic_cast<const ArrayType*>(type))
		{
			if(x<0 || x>=static_cast<const ArrayType*>(type)->nr_elem)
				report_error(printf_string("Array index is out of range. Cannot access index %d of the array\n   %s", (int)x, type->get_type_string().c_str()));
			expr ex2;
			ex2.add(expr((VARIABLE_INDEX_TYPE)var_addr));
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(
					PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr), true, true));
			return PointerExpr(PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr),
				false, true)[x];
		}
		else if(is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply [] to a pointer to void.");
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(*type,
					expression_yielding_address().set_type(Value::VarIndexTypeNr), true, true));
			else
				return PointerExpr(*type, expression_yielding_address().set_type(Value::VarIndexTypeNr), false, true)[x];
		}
		else
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or to pointers. Cannot apply [] to a variable of type\n   %s",
				type->get_type_string().c_str()));
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		if(0!=dynamic_cast<const ArrayType*>(type))
		{
			if(x<0 || x>=static_cast<const ArrayType*>(type)->nr_elem)
				report_error(printf_string("Array index is out of range. Cannot access index %d of the array\n   %s", (int)x, type->get_type_string().c_str()));
			expr ex2;
			ex2.add(expr((VARIABLE_INDEX_TYPE)var_addr));
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(
					PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr), true, true));
			return PointerExpr(PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr),
				false, true)[x];
		}
		else if(is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply [] to a pointer to void.");
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(*type,
						expression_yielding_address().set_type(Value::VarIndexTypeNr), true, true));
			else
				return PointerExpr(*type, expression_yielding_address().set_type(Value::VarIndexTypeNr), false, true)[x];
		}
		else
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or to pointers. Cannot apply [] to a variable of type\n   %s",
				type->get_type_string().c_str()));
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }
};

VARIABLE_INDEX_TYPE Variable::var_mem_count[3*Value::anz_user_types+3];
VARIABLE_INDEX_TYPE Variable::var_mem_begin[3*Value::anz_user_types+3];
nlist<uint64> Variable::init_values[Value::anz_user_types+1];
nlist<uint8> Variable::type_info[3];

struct VariablePlusPlus
{
private: void operator=(const VariablePlusPlus& lpp) { report_error("Assignment VariablePlusPlus::operator=() must not be called"); } // Zugriff verhindern
public:
	const Variable& v; // Hier muss eine Referenz verwendet werden, da ansonsten im Konstruktor LocalMinusMinus::LocalMinusMinus()
		// der Konstruktor local::local aufgerufen wird, welcher nicht lediglich den Speicher der Struktur local kopiert, sondern
		// annimmt, der Benutzer htte eine neue local angelegt und somit entsprechend Speicher auf dem Stack reserviert.
	mutable bool exported;
	const VariablePlusPlus& set_exported() const { exported=true; return *this; }
	VariablePlusPlus(const Variable& _v) : v(_v) { exported=false; }
	~VariablePlusPlus()
	{
		if(!exported)
		{	// Der folgende Code ist identisch wie beim Prfix Operator -- und auch effizienter als der Code des Postfix Operators --,
			// kann jedoch nur verwendet werden, wenn der Wert des Ausdrucks nicht weiter verwendet wird.
			if(!v.type->is_expr() && !v.type->is_ptr())
				report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
			expr s;
			if(v.type->is_ptr())
			{
				if(static_cast<const PointerType*>(v.type)->type_pointed_to->equal(&__VOID))
					report_error("Cannot apply ++ to a pointer to void.");
				s.assign(expr::get_const_expr((VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(v.type)->type_pointed_to->size()));
			}
			else // type->is_expr()==true
				s.assign(expr::get_const_expr((uint32)1));
			// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator--() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.

			expr ex(expr::get_const_expr((uint32)v.var_addr));
			ex.add(s);
			ex.addcmd(bi::ADDM, -1);
			ex.set_code_has_effect();
			ex.set_exported();
			if(v.type->is_ptr())
				ExprObjectDestructionHandler<expr>(ex.set_type(Value::VarIndexTypeNr));
			else if(v.type->is_expr())
				ExprObjectDestructionHandler<expr>(ex.set_type(static_cast<const BasicType*>(v.type)->type));
		}
	}
	GeneralExpr to_GeneralExpr() const
	{
		set_exported();
		if(!v.type->is_expr() && !v.type->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
		expr s;
		if(v.type->is_ptr())
		{
			if(static_cast<const PointerType*>(v.type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply ++ to a pointer to void.");
			s.assign(expr::get_const_expr(static_cast<const PointerType*>(v.type)->type_pointed_to->size()));
		}
		else // type->is_expr()==true
			s.assign(expr::get_const_expr((uint32)1));

		expr ex;
		ex.add(expr::get_const_expr((uint32)v.var_addr));
		ex.add(s);
		ex.addcmd(bi::ADDM, -1);
		ex.add(s);
		ex.addcmd(bi::SUB, -1);
		ex.set_exported();
		ex.set_code_has_effect();
		if(v.type->is_ptr())
			return GeneralExpr::from_PointerExpr(PointerExpr(*v.type, ex.set_type(Value::VarIndexTypeNr), false, true));
		else if(v.type->is_expr())
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(v.type)->type));
		else
			report_error("VariablePlusPlus Error");
	}
	operator expr() const
		{ return to_GeneralExpr().to_expr(); }
	PointerExpr to_PointerExpr() const
		{ return to_GeneralExpr().to_PointerExpr(); }
	GeneralExpr operator->*(const char *element_name) const
	{	// ->* : Dereferenzierung des Zeigers und Elementauswahl des sich ergebenden STRUCT-Objekts.
		set_exported();
		if(0==dynamic_cast<const PointerType*>(v.type) ||
				0==dynamic_cast<const StructType*>(static_cast<const PointerType*>(v.type)->type_pointed_to.get()))
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to a variable of type\n   %s",
				element_name, v.type->get_type_string().c_str()));
		return (to_PointerExpr())->*element_name;
	}
	GeneralExpr operator[](const expr& ex) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[ex];
	}
	GeneralExpr operator[](long long x) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }
};

struct VariableMinusMinus
{
private: void operator=(const VariableMinusMinus& lpp) { report_error("Assignment VariableMinusMinus::operator=() must not be called"); } // Zugriff verhindern
public:
	const Variable& v; // Hier muss eine Referenz verwendet werden, da ansonsten im Konstruktor LocalMinusMinus::LocalMinusMinus()
		// der Konstruktor local::local aufgerufen wird, welcher nicht lediglich den Speicher der Struktur local kopiert, sondern
		// annimmt, der Benutzer htte eine neue local angelegt und somit entsprechend Speicher auf dem Stack reserviert.
	mutable bool exported;
	const VariableMinusMinus& set_exported() const { exported=true; return *this; }
	VariableMinusMinus(const Variable& _v) : v(_v) { exported=false; }
	~VariableMinusMinus()
	{
		if(!exported)
		{	// Der folgende Code ist identisch wie beim Prfix Operator -- und auch effizienter als der Code des Postfix Operators --,
			// kann jedoch nur verwendet werden, wenn der Wert des Ausdrucks nicht weiter verwendet wird.
			if(!v.type->is_expr() && !v.type->is_ptr())
				report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
			expr s;
			if(v.type->is_ptr())
			{
				if(static_cast<const PointerType*>(v.type)->type_pointed_to->equal(&__VOID))
					report_error("Cannot apply ++ to a pointer to void.");
				s.assign(expr::get_const_expr((VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(v.type)->type_pointed_to->size()));
			}
			else // type->is_expr()==true
				s.assign(expr::get_const_expr((uint32)1));
			// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator--() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.

			expr ex(expr::get_const_expr((uint32)v.var_addr));
			ex.add(s);
			ex.addcmd(bi::SUBM, -1);
			ex.set_exported();
			ex.set_code_has_effect();
			if(v.type->is_ptr())
				ExprObjectDestructionHandler<expr>(ex.set_type(Value::VarIndexTypeNr));
			else if(v.type->is_expr())
				ExprObjectDestructionHandler<expr>(ex.set_type(static_cast<const BasicType*>(v.type)->type));
		}
	}
	GeneralExpr to_GeneralExpr() const
	{
		set_exported();
		if(!v.type->is_expr() && !v.type->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
		expr s;
		if(v.type->is_ptr())
		{
			if(static_cast<const PointerType*>(v.type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply ++ to a pointer to void.");
			s.assign(expr::get_const_expr(static_cast<const PointerType*>(v.type)->type_pointed_to->size()));
		}
		else // type->is_expr()==true
			s.assign(expr::get_const_expr((uint32)1));

		expr ex;
		ex.add(expr::get_const_expr((uint32)v.var_addr));
		ex.add(s);
		ex.addcmd(bi::SUBM, -1);
		ex.add(s);
		ex.addcmd(bi::ADD, -1);
		ex.set_exported();
		ex.set_code_has_effect();
		if(v.type->is_ptr())
			return GeneralExpr::from_PointerExpr(PointerExpr(*v.type, ex.set_type(Value::VarIndexTypeNr), false, true));
		else if(v.type->is_expr())
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(v.type)->type));
		else
			report_error("VariableMinusMinus Error");
	}
	operator expr() const
		{ return to_GeneralExpr().to_expr(); }
	PointerExpr to_PointerExpr() const
		{ return to_GeneralExpr().to_PointerExpr(); }
	GeneralExpr operator->*(const char *element_name) const
	{	// ->* : Dereferenzierung des Zeigers und Elementauswahl des sich ergebenden STRUCT-Objekts.
		set_exported();
		if(0==dynamic_cast<const PointerType*>(v.type) ||
				0==dynamic_cast<const StructType*>(static_cast<const PointerType*>(v.type)->type_pointed_to.get()))
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to a variable of type\n   %s",
				element_name, v.type->get_type_string().c_str()));
		return (to_PointerExpr())->*element_name;
	}
	GeneralExpr operator[](const expr& ex) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[ex];
	}
	GeneralExpr operator[](long long x) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }
};

VariablePlusPlus Variable::operator++(int)
{
	return VariablePlusPlus(*this);
}

VariableMinusMinus Variable::operator--(int)
{
	return VariableMinusMinus(*this);
}

struct Constant
{
private:
	Constant(const Constant& v) { report_error("Copy constructor Constant::Constant(const Constant& v) must not be called"); } // Zugriff verhindern
	friend struct VariablePlusPlus;
	friend struct VariableMinusMinus;
public:
	VARIABLE_INDEX_TYPE var_addr;
	const Type *type;
	static nlist<uint64> init_values[Value::anz_user_types+1];
	void construct(const Type *__type, nlist<uint64> InitValues=nlist<uint64>())
	{
		type=__type;
		if(InitValues.size()!=type->get_flattend_types_list().size())
			report_error(printf_string("The number %d of initialization parameters of elementary type used to initialize an constant of type\n   %s\ndiffers from the required number %u",
				InitValues.size(), type->get_type_string().c_str(), type->get_nr_of_elementary_objects()));
		if(type->has_only_homogenous_basic_types())
		{
			uint8 type_nr;
			if(0!=dynamic_cast<const BasicType*>(type))
				type_nr=static_cast<const BasicType*>(type)->type;
			else if(0!=dynamic_cast<const PointerType*>(type))
				type_nr=Value::VarIndexTypeNr;
			else if(type->has_only_homogenous_basic_types())
				type_nr=type->get_type_nr_of_homogenous_basic_types();

			if(Value::t_void==type_nr)
				report_error("Cannot create a variable of type void");
			const long t2=2, t=(Value::anz_user_types+1)*t2+type_nr;
			var_addr=Variable::var_mem_count[t];
			Variable::var_mem_count[t]+=type->size();
			if(1==cc->CompilerPassNr)
				init_values[type_nr].insert(init_values[type_nr].end(), InitValues.begin(), InitValues.end());
		}
		else
		{
			long t2=2, t=(Value::anz_user_types+1)*t2+Value::anz_user_types;
			var_addr=Variable::var_mem_count[t];
			Variable::var_mem_count[t]+=type->size();
			if(1==cc->CompilerPassNr)
			{
				nlist<uint64>::iterator ivalit=InitValues.begin();
				nlist<uint8> lst=type->get_flattend_types_list();
				for(nlist<uint8>::iterator it=lst.begin(); it!=lst.end(); it++)
				{
					uint8 type_nr=*it;
					if(ivalit!=InitValues.end())
						for(long i=0; i<Value::sbytesize(type_nr); i++)
							init_values[Value::anz_user_types].push_back(((*ivalit)>>(8*i)) & 255);

					switch(type_nr)
					{
						case Value::t_bool:
						case Value::t_real:
							Variable::type_info[t2].push_back(1); break;
						case Value::t_u8:
						case Value::t_u16:
						case Value::t_u32:
						case Value::t_u64:
							Variable::type_info[t2].push_back(2); break;
						case Value::t_i8:
						case Value::t_i16:
						case Value::t_i32:
						case Value::t_i64:
							Variable::type_info[t2].push_back(3); break;
					}
					for(long i=1; i<Value::sbytesize(type_nr); i++)
						Variable::type_info[t2].push_back(0);
					if(ivalit!=InitValues.end())
						++ivalit;
				}
			}
		}
	}
	void construct(const Type *__type, uint64 x)
	{
		if(0==dynamic_cast<const BasicType*>(__type))
			report_error("construct() called with one initialization parameter for a non-elementary type.");
		nlist<uint64> lst;
		if(0!=x)
		{
			if(static_cast<const BasicType*>(__type)->type==Value::t_real)
				lst.push_back(real2uint64_endianness_independent((real_type)x));
			else
				lst.push_back(x);
		}
		construct(__type, lst);
	}
	void construct(const Type *__type, int64 x)
	{
		if(0==dynamic_cast<const BasicType*>(__type))
			report_error("construct() called with one initialization parameter for a non-elementary type.");
		nlist<uint64> lst;
		if(0!=x)
		{
			if(static_cast<const BasicType*>(__type)->type==Value::t_real)
				lst.push_back(real2uint64_endianness_independent((real_type)x));
			else
				lst.push_back(x);
		}
		construct(__type, lst);
	}
	void construct_from_real(const Type *__type, real_type x)
	{
		if(0==dynamic_cast<const BasicType*>(__type))
			report_error("construct_from_real() called with one initialization parameter for a non-elementary type.");
		nlist<uint64> lst;
		if(0!=x)
		{
			if(static_cast<const BasicType*>(__type)->type==Value::t_real)
				lst.push_back(real2uint64_endianness_independent(x));
			else if(Value::s_is_signed_integer(static_cast<const BasicType*>(__type)->type))
				lst.push_back((int64)x);
			else
				lst.push_back((uint64)x);
		}
		construct(__type, lst);
	}
	Constant(const BasicType& type, unsigned long long init_value) { construct(&type, init_value); }
	Constant(const BasicType& type, long long init_value) { construct(&type, init_value); }
	Constant(const BasicType& type, unsigned long init_value) { construct(&type, (uint64)init_value); }
	Constant(const BasicType& type, long init_value) { construct(&type, (int64)init_value); }
	Constant(const BasicType& type, unsigned int init_value) { construct(&type, (uint64)init_value); }
	Constant(const BasicType& type, int init_value) { construct(&type, (int64)init_value); }
	Constant(const BasicType& type, unsigned short init_value) { construct(&type, (uint64)init_value); }
	Constant(const BasicType& type, short init_value) { construct(&type, (int64)init_value); }
	Constant(const BasicType& type, unsigned char init_value) { construct(&type, (uint64)init_value); }
	Constant(const BasicType& type, signed char init_value) { construct(&type, (int64)init_value); }
	Constant(const BasicType& type, bool init_value) { construct(&type, (uint64)init_value); }
#ifndef real_type_is_integer_type
	Constant(const BasicType& type, real_type init_value) { construct_from_real(&type, init_value); }
#endif
	Constant(const BasicType& type) { construct(&type); }

	Constant(const CollectGeneralType& init_value)
	{
		init_value.set_exported();
		if(init_value.is_expr_type())
		{
			if(!init_value.is_basicconst)
				report_error("Global variables can only have constants as initialisers.");
			construct(&BasicType::get_BasicType_from_type_nr(init_value.to_expr().get_type()), init_value.cval);
		}
		else if(init_value.is_PointerExpr())
		{
			if(!init_value.is_basicconst || !Value::s_is_integer_type(init_value.to_expr().get_type()) || 0!=init_value.cval)
				report_error("Global variables of type pointer can only have the integer constant 0 as initialiser.");
			construct(init_value.get_type(), (uint64)0); // Nullzeiger
		}
		else
			report_error("Unknown type in CollectGeneralType");
	}

	bool is_expr() const { return 0!=dynamic_cast<const BasicType*>(type); }
	bool is_ptr() const { return 0!=dynamic_cast<const PointerType*>(type); }

	Constant(const PointerType& __type)
	{
		__type.chk_validity_of_ref_types_variable_creation();
		type=cc->RegisterType->Register(&__type);
		construct(type); // Zeiger wird automatisch mit 0 initialisiert. Hier darf nicht construct(type, (uint64)0) aufgerufen werden !
	}


	nlist<uint64> prepare_data_init(const DataInitialization *di) const
	{	// ACHTUNG: type muss schon initialisiert sein !
		if(DataInitialization::get_size(di)!=type->get_nr_of_elementary_objects())
			report_error(printf_string("The number %d of initialization parameters of elementary type used to initialize a variable of type\n   %s\ndiffers from the required number %u",
				DataInitialization::get_size(di), type->get_type_string().c_str(), type->get_nr_of_elementary_objects()));
		nlist<uint8> types=type->get_flattend_types_list();
		nlist<uint64> init_values;
		for(nlist<uint8>::iterator it=types.begin(); it!=types.end() && !di->end; it++, di++)
		{
			if(Value::t_real==*it)
			{
				if(Value::t_real==di->type_nr)
					init_values.push_back(real2uint64_endianness_independent(di->rval));
				else if(Value::s_is_signed_integer(di->type_nr))
					init_values.push_back(real2uint64_endianness_independent((real_type)((int64)di->ival)));
				else
					init_values.push_back(real2uint64_endianness_independent((real_type)((uint64)di->ival)));					
			}
			else if(Value::s_is_signed_integer(*it))
			{
				if(Value::t_real==di->type_nr)
					init_values.push_back((uint64)((int64)di->rval));
				else
					init_values.push_back((uint64)di->ival);
			}
			else
			{
				if(Value::t_real==di->type_nr)
					init_values.push_back((uint64)di->rval);
				else
					init_values.push_back((uint64)di->ival);
			}
		}
		return init_values;
	}
	Constant(const ArrayType& __type)
	{
		__type.chk_validity_of_ref_types_variable_creation();
		type=cc->RegisterType->Register(&__type);
		construct(type); // hier darf NICHT mehr &__type verwendet werden !
	}
	Constant(const ArrayType& __type, const DataInitialization *di)
	{
		__type.chk_validity_of_ref_types_variable_creation();
		type=cc->RegisterType->Register(&__type);
		construct(type, prepare_data_init(di)); // hier darf NICHT mehr &__type verwendet werden !
	}
	Constant(const StructType& __type, const DataInitialization *di)
	{
		__type.chk_validity_of_ref_types_variable_creation();
		type=cc->RegisterType->Register(&__type);
		construct(type, prepare_data_init(di)); // hier darf NICHT mehr &__type verwendet werden !
	}
	
	expr expression_yielding_address() const
	{
		if(!type->is_ptr())
			report_error("expression_yielding_address() called for non-pointer.");
		expr ex(expr((VARIABLE_INDEX_TYPE)var_addr));
		return ex.set_type(Value::VarIndexTypeNr);
	}
	PointerExpr to_PointerExpr() const
	{
		if(!is_ptr())
			report_error(printf_string("Conversion to Pointer not possible. Cannot convert a variable of type\n   %s\ninto a pointer type.",
				type->get_type_string().c_str()));
		return PointerExpr(*type, expression_yielding_address(), false, true); // PointerExpr::memory_with_fixed_type==true
	}

	operator expr() const
	{
		if(!is_expr()) // Ein Pointer knnte zwar in einen __BOOL umgewandelt werden. Hierfr ist jedoch die Umwandlung nach LogExpr vorgesehen.
			report_error(printf_string("Cannot convert a variable of type\n   %s\ninto an expression of elementary type.",
				type->get_type_string().c_str()));
		expr ex(expr::get_const_expr((uint32)var_addr));
		ex.addcmd(bi::LOAD);
		return ex;
	}

	PointerExpr operator&()
	{
		return PointerExpr(PointerType(type), expr((VARIABLE_INDEX_TYPE)var_addr).set_type(Value::VarIndexTypeNr),
			false, true); // PointerExpr::memory_with_fixed_type_info==true
	}

	PointerExpr operator*()
	{
		if(0==dynamic_cast<const PointerType*>(type))
			report_error("The dereferentiation operator * can only be applied to expressions of type pointer.");
		if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
			report_error("Cannot dereferentiate a pointer to void.");
		expr ex;
		ex.add(expr::get_const_expr((uint32)var_addr));
		ex.addcmd(bi::LOAD, 0);
		return PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr), true);
	}

	void operator=(const Constant& value)
	{
		report_error("Assignment to constant in not possible.");
	}
	void operator=(const CollectGeneralType& value)
	{
		report_error("Assignment to constant in not possible.");
	}

	void operator+=(const expr& x)
	{
		report_error("Assignment to constant in not possible.");
	}
	void operator++()
	{
		report_error("Increment of constant in not possible.");
	}
	void operator++(int)
	{
		report_error("Increment of constant in not possible.");
	}

	void operator-=(const expr& x)
	{
		report_error("Assignment to constant in not possible.");
	}
	void operator--()
	{
		report_error("Decrement of constant in not possible.");
	}
	void operator--(int)
	{
		report_error("Decrement of constant in not possible.");
	}

	void operator*=(const expr& x)
	{
		report_error("Assignment to constant in not possible.");
	}

	void operator/=(const expr& x)
	{
		report_error("Assignment to constant in not possible.");
	}

	void operator%=(const expr& x)
	{
		report_error("Assignment to constant in not possible.");
	}

	void operator|=(const expr& x)
	{
		report_error("Assignment to constant in not possible.");
	}

	void operator&=(const expr& x)
	{
		report_error("Assignment to constant in not possible.");
	}

	void operator^=(const expr& x)
	{
		report_error("Assignment to constant in not possible.");
	}

	void operator<<=(const expr& x)
	{
		report_error("Assignment to constant in not possible.");
	}

	void operator>>=(const expr& x)
	{
		report_error("Assignment to constant in not possible.");
	}

	GeneralExpr operator[](const char *element_name) const
	{	// [] : Elementauswahl eines STRUCTs.
		if(0==dynamic_cast<const StructType*>(type))
			report_error(printf_string("The operator [] used with a string constant index can only be applied to objects of type STRUCT. Cannot apply [\"%s\"] to a variable of type\n   %s",
				element_name, type->get_type_string().c_str()));
		nlist<StructType::Elem>::const_iterator it=static_cast<const StructType*>(type)->find_elem(element_name);
		if(!it.is_valid_and_not_end(static_cast<const StructType*>(type)->elem))
			report_error(printf_string("operator[]: '%s' in not an element of %s", element_name, type->get_type_string().c_str()));
		expr ex2;
		ex2.add(expr((VARIABLE_INDEX_TYPE)(static_cast<const StructType*>(type)->offsetofelem(element_name)+var_addr)));
		return GeneralExpr::from_PointerExpr(
			PointerExpr(PointerType(it->type), ex2.set_type(Value::VarIndexTypeNr), true, true));
	}
	GeneralExpr operator->*(const char *element_name) const
	{	// ->* : Dereferenzierung des Zeigers und Elementauswahl des sich ergebenden STRUCT-Objekts.
		if(0==dynamic_cast<const PointerType*>(type) ||
				0==dynamic_cast<const StructType*>(static_cast<const PointerType*>(type)->type_pointed_to.get()))
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to a variable of type\n   %s",
				element_name, type->get_type_string().c_str()));
		const Type *ptype=static_cast<const PointerType*>(type)->type_pointed_to;
		const StructType *stype=static_cast<const StructType*>(ptype);
		nlist<StructType::Elem>::const_iterator it=stype->find_elem(element_name);
		if(!it.is_valid_and_not_end(stype->elem))
			report_error(printf_string("operator->*: '%s' in not an element of %s", element_name, type->get_type_string().c_str()));
		expr ex(expr::get_const_expr((uint32)var_addr));
		ex.addcmd(bi::LOAD, 0);
		ex.add(expr::get_const_expr(stype->offsetofelem(element_name)));
		ex.addcmd(bi::ADD, -1);
		//return GeneralExpr::from_PointerExpr(PointerExpr(PointerType(it->type), ex.set_type(Value::VarIndexTypeNr), true));
		return GeneralExpr::from_PointerExpr(
			PointerExpr(PointerType(it->type, static_cast<const PointerType*>(type)->C_pointer),
			ex.set_type(Value::VarIndexTypeNr), true));
	}
	GeneralExpr operator[](const expr& ex) const
	{
		ex.set_exported();
		if(0!=dynamic_cast<const ArrayType*>(type))
		{
			expr ex2;
			ex2.add(expr((VARIABLE_INDEX_TYPE)var_addr));
			return PointerExpr(PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr),
				false, true)[ex]; // PointerExpr::memory_with_fixed_type==true
		}
		else if(is_ptr())
			return PointerExpr(*type, expression_yielding_address().set_type(Value::VarIndexTypeNr), false, true)[ex];
		else
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or to pointers. Cannot apply [] to a variable of type\n   %s",
				type->get_type_string().c_str()));
	}
	GeneralExpr operator[](long long x) const
	{
		if(0!=dynamic_cast<const ArrayType*>(type))
		{
			if(x<0 || x>=static_cast<const ArrayType*>(type)->nr_elem)
				report_error(printf_string("Array index is out of range. Cannot access index %d of the array\n   %s", (int)x, type->get_type_string().c_str()));
			expr ex2;
			ex2.add(expr((VARIABLE_INDEX_TYPE)var_addr));
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(
					PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr), true, true));
			return PointerExpr(PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr),
				false, true)[x];
		}
		else if(is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply [] to a pointer to void.");
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(*type,
					expression_yielding_address().set_type(Value::VarIndexTypeNr), true, true));
			else
				return PointerExpr(*type, expression_yielding_address().set_type(Value::VarIndexTypeNr), false, true)[x];
		}
		else
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or to pointers. Cannot apply [] to a variable of type\n   %s",
				type->get_type_string().c_str()));
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		if(0!=dynamic_cast<const ArrayType*>(type))
		{
			if(x<0 || x>=static_cast<const ArrayType*>(type)->nr_elem)
				report_error(printf_string("Array index is out of range. Cannot access index %d of the array\n   %s", (int)x, type->get_type_string().c_str()));
			expr ex2;
			ex2.add(expr((VARIABLE_INDEX_TYPE)var_addr));
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(
					PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr), true, true));
			return PointerExpr(PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr),
				false, true)[x];
		}
		else if(is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply [] to a pointer to void.");
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(*type,
						expression_yielding_address().set_type(Value::VarIndexTypeNr), true, true));
			else
				return PointerExpr(*type, expression_yielding_address().set_type(Value::VarIndexTypeNr), false, true)[x];
		}
		else
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or to pointers. Cannot apply [] to a variable of type\n   %s",
				type->get_type_string().c_str()));
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }
};

nlist<uint64> Constant::init_values[Value::anz_user_types+1];

} // Ende namespace BytecodeTargetedCompiler

template<>
struct LessEqual<std::pair<nstring, BytecodeTargetedCompiler::Constant*> > :
	public BaumKeineAttrBerechnen<std::pair<nstring, BytecodeTargetedCompiler::Constant*> >
{
	bool less(const std::pair<nstring, BytecodeTargetedCompiler::Constant*>& x, const std::pair<nstring, BytecodeTargetedCompiler::Constant*>& y) const
		{ return 0<x.first.compare(y.first); }
	bool equal(const std::pair<nstring, BytecodeTargetedCompiler::Constant*>& x, const std::pair<nstring, BytecodeTargetedCompiler::Constant*>& y) const
		{ return 0==x.first.compare(y.first); }
};

namespace BytecodeTargetedCompiler
{

struct StringConstantsContainer
{
	RotSchwarzBaum<std::pair<nstring, Constant*> > strs;
	~StringConstantsContainer() { clear(); }
	void clear()
	{
		for(RotSchwarzBaum<std::pair<nstring, Constant*> >::iterator it=strs.begin(); it!=strs.end(); it++)
			delete it->second;
		strs.clear();
	}
};

PointerExpr STRING(const char *C_char_pointer)
{
	if(0==C_char_pointer)
		return PointerExpr(BYTECODE_TARGETED_COMPILER__POINTER(__UINT8), expr::get_const_expr(0), false);
	RotSchwarzBaum<std::pair<nstring, Constant*> >::iterator it;
	it=cc->StringConstants->strs.suche(std::pair<nstring, Constant*>(C_char_pointer, 0));
	if(it!=0)
		return &(*(it->second))[0]; // Wenn ein String mehrmals auftritt, wird er nur einmal abgespeichert.
	uint32 len=0;
	while(C_char_pointer[len]) len++;
	DataInitialization *ch=new DataInitialization[2+len];
	for(len=0; C_char_pointer[len]; len++)
		ch[len]=DataInitialization((uint8)C_char_pointer[len]);
	ch[len]=DataInitialization((uint8)0);
	ch[len+1]=DataEndMarker();
	it=cc->StringConstants->strs.insert(std::pair<nstring, Constant*>(C_char_pointer,
			new Constant(BYTECODE_TARGETED_COMPILER__ARRAY(__UINT8, 1+len), ch)));
	delete[] ch;
	return &(*(it->second))[0];
}

#define Function_ADDPARAM(n) ;arg##n.set_exported(); args.push_front(arg##n);
#define Function_ADDPARAM2 ;Function_ADDPARAM(1); Function_ADDPARAM(0);
#define Function_ADDPARAM4 ;Function_ADDPARAM(3); Function_ADDPARAM(2); Function_ADDPARAM2;
#define Function_ADDPARAM6 ;Function_ADDPARAM(5); Function_ADDPARAM(4); Function_ADDPARAM4;
#define Function_ADDPARAM8 ;Function_ADDPARAM(7); Function_ADDPARAM(6); Function_ADDPARAM6;
#define Function_ADDPARAM10 ;Function_ADDPARAM(9); Function_ADDPARAM(8); Function_ADDPARAM8;
#define Function_ADDPARAM12 ;Function_ADDPARAM(11); Function_ADDPARAM(10); Function_ADDPARAM10;
#define Function_ADDPARAM14 ;Function_ADDPARAM(13); Function_ADDPARAM(12); Function_ADDPARAM12;
#define Function_ADDPARAM16 ;Function_ADDPARAM(15); Function_ADDPARAM(14); Function_ADDPARAM14;
#define Function_ADDPARAM18 ;Function_ADDPARAM(17); Function_ADDPARAM(16); Function_ADDPARAM16;
#define Function_ADDPARAM20 ;Function_ADDPARAM(19); Function_ADDPARAM(18); Function_ADDPARAM18;
#define Function_ADDPARAM22 ;Function_ADDPARAM(21); Function_ADDPARAM(20); Function_ADDPARAM20;
#define Function_ADDPARAM24 ;Function_ADDPARAM(23); Function_ADDPARAM(22); Function_ADDPARAM22;
#define Function_ADDPARAM26 ;Function_ADDPARAM(25); Function_ADDPARAM(24); Function_ADDPARAM24;
#define Function_ADDPARAM28 ;Function_ADDPARAM(27); Function_ADDPARAM(26); Function_ADDPARAM26;
#define Function_ADDPARAM30 ;Function_ADDPARAM(29); Function_ADDPARAM(28); Function_ADDPARAM28;
#define Function_ADDPARAM32 ;Function_ADDPARAM(31); Function_ADDPARAM(30); Function_ADDPARAM30;
#define Function_ADDPARAM34 ;Function_ADDPARAM(33); Function_ADDPARAM(32); Function_ADDPARAM32;
#define Function_ADDPARAM36 ;Function_ADDPARAM(35); Function_ADDPARAM(34); Function_ADDPARAM34;
#define Function_ADDPARAM38 ;Function_ADDPARAM(37); Function_ADDPARAM(36); Function_ADDPARAM36;
#define Function_ADDPARAM40 ;Function_ADDPARAM(39); Function_ADDPARAM(38); Function_ADDPARAM38;
#define Function_ADDPARAM42 ;Function_ADDPARAM(41); Function_ADDPARAM(40); Function_ADDPARAM40;
#define Function_ADDPARAM44 ;Function_ADDPARAM(43); Function_ADDPARAM(42); Function_ADDPARAM42;
#define Function_ADDPARAM46 ;Function_ADDPARAM(45); Function_ADDPARAM(44); Function_ADDPARAM44;
#define Function_ADDPARAM48 ;Function_ADDPARAM(47); Function_ADDPARAM(46); Function_ADDPARAM46;
#define Function_ADDPARAM50 ;Function_ADDPARAM(49); Function_ADDPARAM(48); Function_ADDPARAM48;
#define Function_ADDPARAM52 ;Function_ADDPARAM(51); Function_ADDPARAM(50); Function_ADDPARAM50;
#define Function_ADDPARAM54 ;Function_ADDPARAM(53); Function_ADDPARAM(52); Function_ADDPARAM52;
#define Function_ADDPARAM56 ;Function_ADDPARAM(55); Function_ADDPARAM(54); Function_ADDPARAM54;
#define Function_ADDPARAM58 ;Function_ADDPARAM(57); Function_ADDPARAM(56); Function_ADDPARAM56;
#define Function_ADDPARAM60 ;Function_ADDPARAM(59); Function_ADDPARAM(58); Function_ADDPARAM58;
#define Function_ADDPARAM62 ;Function_ADDPARAM(61); Function_ADDPARAM(60); Function_ADDPARAM60;
#define Function_ADDPARAM64 ;Function_ADDPARAM(63); Function_ADDPARAM(62); Function_ADDPARAM62;
#define Function_ADDPARAM_even_nr_args(NrArgs) ;Function_ADDPARAM##NrArgs;
#define Function_ADDPARAM_odd_nr_args(NrArgs) ;Function_ADDPARAM(NrArgs); Function_ADDPARAM##NrArgs;

#define Function_FuncOPHeader(n) ,const PointerOrElementaryType& arg##n
#define Function_FuncOPHeader2 operator()(const PointerOrElementaryType& arg0 Function_FuncOPHeader(1)
#define Function_FuncOPHeader4 Function_FuncOPHeader2 Function_FuncOPHeader(2) Function_FuncOPHeader(3)
#define Function_FuncOPHeader6 Function_FuncOPHeader4 Function_FuncOPHeader(4) Function_FuncOPHeader(5)
#define Function_FuncOPHeader8 Function_FuncOPHeader6 Function_FuncOPHeader(6) Function_FuncOPHeader(7)
#define Function_FuncOPHeader10 Function_FuncOPHeader8 Function_FuncOPHeader(8) Function_FuncOPHeader(9)
#define Function_FuncOPHeader12 Function_FuncOPHeader10 Function_FuncOPHeader(10) Function_FuncOPHeader(11)
#define Function_FuncOPHeader14 Function_FuncOPHeader12 Function_FuncOPHeader(12) Function_FuncOPHeader(13)
#define Function_FuncOPHeader16 Function_FuncOPHeader14 Function_FuncOPHeader(14) Function_FuncOPHeader(15)
#define Function_FuncOPHeader18 Function_FuncOPHeader16 Function_FuncOPHeader(16) Function_FuncOPHeader(17)
#define Function_FuncOPHeader20 Function_FuncOPHeader18 Function_FuncOPHeader(18) Function_FuncOPHeader(19)
#define Function_FuncOPHeader22 Function_FuncOPHeader20 Function_FuncOPHeader(20) Function_FuncOPHeader(21)
#define Function_FuncOPHeader24 Function_FuncOPHeader22 Function_FuncOPHeader(22) Function_FuncOPHeader(23)
#define Function_FuncOPHeader26 Function_FuncOPHeader24 Function_FuncOPHeader(24) Function_FuncOPHeader(25)
#define Function_FuncOPHeader28 Function_FuncOPHeader26 Function_FuncOPHeader(26) Function_FuncOPHeader(27)
#define Function_FuncOPHeader30 Function_FuncOPHeader28 Function_FuncOPHeader(28) Function_FuncOPHeader(29)
#define Function_FuncOPHeader32 Function_FuncOPHeader30 Function_FuncOPHeader(30) Function_FuncOPHeader(31)
#define Function_FuncOPHeader34 Function_FuncOPHeader32 Function_FuncOPHeader(32) Function_FuncOPHeader(33)
#define Function_FuncOPHeader36 Function_FuncOPHeader34 Function_FuncOPHeader(34) Function_FuncOPHeader(35)
#define Function_FuncOPHeader38 Function_FuncOPHeader36 Function_FuncOPHeader(36) Function_FuncOPHeader(37)
#define Function_FuncOPHeader40 Function_FuncOPHeader38 Function_FuncOPHeader(38) Function_FuncOPHeader(39)
#define Function_FuncOPHeader42 Function_FuncOPHeader40 Function_FuncOPHeader(40) Function_FuncOPHeader(41)
#define Function_FuncOPHeader44 Function_FuncOPHeader42 Function_FuncOPHeader(42) Function_FuncOPHeader(43)
#define Function_FuncOPHeader46 Function_FuncOPHeader44 Function_FuncOPHeader(44) Function_FuncOPHeader(45)
#define Function_FuncOPHeader48 Function_FuncOPHeader46 Function_FuncOPHeader(46) Function_FuncOPHeader(47)
#define Function_FuncOPHeader50 Function_FuncOPHeader48 Function_FuncOPHeader(48) Function_FuncOPHeader(49)
#define Function_FuncOPHeader52 Function_FuncOPHeader50 Function_FuncOPHeader(50) Function_FuncOPHeader(51)
#define Function_FuncOPHeader54 Function_FuncOPHeader52 Function_FuncOPHeader(52) Function_FuncOPHeader(53)
#define Function_FuncOPHeader56 Function_FuncOPHeader54 Function_FuncOPHeader(54) Function_FuncOPHeader(55)
#define Function_FuncOPHeader58 Function_FuncOPHeader56 Function_FuncOPHeader(56) Function_FuncOPHeader(57)
#define Function_FuncOPHeader60 Function_FuncOPHeader58 Function_FuncOPHeader(58) Function_FuncOPHeader(59)
#define Function_FuncOPHeader62 Function_FuncOPHeader60 Function_FuncOPHeader(60) Function_FuncOPHeader(61)
#define Function_FuncOPHeader64 Function_FuncOPHeader62 Function_FuncOPHeader(62) Function_FuncOPHeader(63)

#define Function_FuncOP_even_nr_args(NrArgs) GeneralExpr Function_FuncOPHeader##NrArgs)\
	{\
		nlist<PointerOrElementaryType> args;\
		Function_ADDPARAM_even_nr_args(NrArgs);\
		return functioncall_helper(args);\
	}
#define Function_FuncOP_odd_nr_args(NrArgsMinus1) GeneralExpr Function_FuncOPHeader##NrArgsMinus1 Function_FuncOPHeader(NrArgsMinus1))\
	{\
		nlist<PointerOrElementaryType> args;\
		Function_ADDPARAM_odd_nr_args(NrArgsMinus1);\
		return functioncall_helper(args);\
	}
#define Function_4FuncOPs(NrArgs, NrArgsPlus2)\
			Function_FuncOP_even_nr_args(NrArgs); Function_FuncOP_odd_nr_args(NrArgs);\
			Function_FuncOP_even_nr_args(NrArgsPlus2); Function_FuncOP_odd_nr_args(NrArgsPlus2);

#define VoidFunction_FuncOP_even_nr_args(NrArgs) void Function_FuncOPHeader##NrArgs)\
	{\
		nlist<PointerOrElementaryType> args;\
		Function_ADDPARAM_even_nr_args(NrArgs);\
		functioncall_helper(args);\
	}
#define VoidFunction_FuncOP_odd_nr_args(NrArgsMinus1) void Function_FuncOPHeader##NrArgsMinus1 Function_FuncOPHeader(NrArgsMinus1))\
	{\
		nlist<PointerOrElementaryType> args;\
		Function_ADDPARAM_odd_nr_args(NrArgsMinus1);\
		functioncall_helper(args);\
	}
#define VoidFunction_4FuncOPs(NrArgs, NrArgsPlus2)\
			VoidFunction_FuncOP_even_nr_args(NrArgs); VoidFunction_FuncOP_odd_nr_args(NrArgs);\
			VoidFunction_FuncOP_even_nr_args(NrArgsPlus2); VoidFunction_FuncOP_odd_nr_args(NrArgsPlus2);


#define PARAMS_ADDPARAM(n) const Type& type##n
#define PARAMS_ADDPARAM2 PARAMS_ADDPARAM(0) , PARAMS_ADDPARAM(1)
#define PARAMS_ADDPARAM4 PARAMS_ADDPARAM2 , PARAMS_ADDPARAM(2) , PARAMS_ADDPARAM(3)
#define PARAMS_ADDPARAM6 PARAMS_ADDPARAM4 , PARAMS_ADDPARAM(4) , PARAMS_ADDPARAM(5)
#define PARAMS_ADDPARAM8 PARAMS_ADDPARAM6 , PARAMS_ADDPARAM(6) , PARAMS_ADDPARAM(7)
#define PARAMS_ADDPARAM10 PARAMS_ADDPARAM8 , PARAMS_ADDPARAM(8) , PARAMS_ADDPARAM(9)
#define PARAMS_ADDPARAM12 PARAMS_ADDPARAM10 , PARAMS_ADDPARAM(10) , PARAMS_ADDPARAM(11)
#define PARAMS_ADDPARAM14 PARAMS_ADDPARAM12 , PARAMS_ADDPARAM(12) , PARAMS_ADDPARAM(13)
#define PARAMS_ADDPARAM16 PARAMS_ADDPARAM14 , PARAMS_ADDPARAM(14) , PARAMS_ADDPARAM(15)
#define PARAMS_ADDPARAM18 PARAMS_ADDPARAM16 , PARAMS_ADDPARAM(16) , PARAMS_ADDPARAM(17)
#define PARAMS_ADDPARAM20 PARAMS_ADDPARAM18 , PARAMS_ADDPARAM(18) , PARAMS_ADDPARAM(19)
#define PARAMS_ADDPARAM22 PARAMS_ADDPARAM20 , PARAMS_ADDPARAM(20) , PARAMS_ADDPARAM(21)
#define PARAMS_ADDPARAM24 PARAMS_ADDPARAM22 , PARAMS_ADDPARAM(22) , PARAMS_ADDPARAM(23)
#define PARAMS_ADDPARAM26 PARAMS_ADDPARAM24 , PARAMS_ADDPARAM(24) , PARAMS_ADDPARAM(25)
#define PARAMS_ADDPARAM28 PARAMS_ADDPARAM26 , PARAMS_ADDPARAM(26) , PARAMS_ADDPARAM(27)
#define PARAMS_ADDPARAM30 PARAMS_ADDPARAM28 , PARAMS_ADDPARAM(28) , PARAMS_ADDPARAM(29)
#define PARAMS_ADDPARAM32 PARAMS_ADDPARAM30 , PARAMS_ADDPARAM(30) , PARAMS_ADDPARAM(31)
#define PARAMS_ADDPARAM34 PARAMS_ADDPARAM32 , PARAMS_ADDPARAM(32) , PARAMS_ADDPARAM(33)
#define PARAMS_ADDPARAM36 PARAMS_ADDPARAM34 , PARAMS_ADDPARAM(34) , PARAMS_ADDPARAM(35)
#define PARAMS_ADDPARAM38 PARAMS_ADDPARAM36 , PARAMS_ADDPARAM(36) , PARAMS_ADDPARAM(37)
#define PARAMS_ADDPARAM40 PARAMS_ADDPARAM38 , PARAMS_ADDPARAM(38) , PARAMS_ADDPARAM(39)
#define PARAMS_ADDPARAM42 PARAMS_ADDPARAM40 , PARAMS_ADDPARAM(40) , PARAMS_ADDPARAM(41)
#define PARAMS_ADDPARAM44 PARAMS_ADDPARAM42 , PARAMS_ADDPARAM(42) , PARAMS_ADDPARAM(43)
#define PARAMS_ADDPARAM46 PARAMS_ADDPARAM44 , PARAMS_ADDPARAM(44) , PARAMS_ADDPARAM(45)
#define PARAMS_ADDPARAM48 PARAMS_ADDPARAM46 , PARAMS_ADDPARAM(46) , PARAMS_ADDPARAM(47)
#define PARAMS_ADDPARAM50 PARAMS_ADDPARAM48 , PARAMS_ADDPARAM(48) , PARAMS_ADDPARAM(49)
#define PARAMS_ADDPARAM52 PARAMS_ADDPARAM50 , PARAMS_ADDPARAM(50) , PARAMS_ADDPARAM(51)
#define PARAMS_ADDPARAM54 PARAMS_ADDPARAM52 , PARAMS_ADDPARAM(52) , PARAMS_ADDPARAM(53)
#define PARAMS_ADDPARAM56 PARAMS_ADDPARAM54 , PARAMS_ADDPARAM(54) , PARAMS_ADDPARAM(55)
#define PARAMS_ADDPARAM58 PARAMS_ADDPARAM56 , PARAMS_ADDPARAM(56) , PARAMS_ADDPARAM(57)
#define PARAMS_ADDPARAM60 PARAMS_ADDPARAM58 , PARAMS_ADDPARAM(58) , PARAMS_ADDPARAM(59)
#define PARAMS_ADDPARAM62 PARAMS_ADDPARAM60 , PARAMS_ADDPARAM(60) , PARAMS_ADDPARAM(61)
#define PARAMS_ADDPARAM64 PARAMS_ADDPARAM62 , PARAMS_ADDPARAM(62) , PARAMS_ADDPARAM(63)
#define PARAMS_ADDPARAM_even_nr_args(NrArgs) PARAMS( PARAMS_ADDPARAM##NrArgs )
#define PARAMS_ADDPARAM_odd_nr_args(NrArgs)  PARAMS( PARAMS_ADDPARAM##NrArgs, PARAMS_ADDPARAM(NrArgs) )

#define PARAMS_PUSHPARAM(n) ;types.push_back(cc->RegisterType->Register(&type##n));
#define PARAMS_PUSHPARAM2 ;PARAMS_PUSHPARAM(0); PARAMS_PUSHPARAM(1);
#define PARAMS_PUSHPARAM4 PARAMS_PUSHPARAM2; PARAMS_PUSHPARAM(2); PARAMS_PUSHPARAM(3);
#define PARAMS_PUSHPARAM6 PARAMS_PUSHPARAM4; PARAMS_PUSHPARAM(4); PARAMS_PUSHPARAM(5);
#define PARAMS_PUSHPARAM8 PARAMS_PUSHPARAM6; PARAMS_PUSHPARAM(6); PARAMS_PUSHPARAM(7);
#define PARAMS_PUSHPARAM10 PARAMS_PUSHPARAM8; PARAMS_PUSHPARAM(8); PARAMS_PUSHPARAM(9);
#define PARAMS_PUSHPARAM12 PARAMS_PUSHPARAM10; PARAMS_PUSHPARAM(10); PARAMS_PUSHPARAM(11);
#define PARAMS_PUSHPARAM14 PARAMS_PUSHPARAM12; PARAMS_PUSHPARAM(12); PARAMS_PUSHPARAM(13);
#define PARAMS_PUSHPARAM16 PARAMS_PUSHPARAM14; PARAMS_PUSHPARAM(14); PARAMS_PUSHPARAM(15);
#define PARAMS_PUSHPARAM18 PARAMS_PUSHPARAM16; PARAMS_PUSHPARAM(16); PARAMS_PUSHPARAM(17);
#define PARAMS_PUSHPARAM20 PARAMS_PUSHPARAM18; PARAMS_PUSHPARAM(18); PARAMS_PUSHPARAM(19);
#define PARAMS_PUSHPARAM22 PARAMS_PUSHPARAM20; PARAMS_PUSHPARAM(20); PARAMS_PUSHPARAM(21);
#define PARAMS_PUSHPARAM24 PARAMS_PUSHPARAM22; PARAMS_PUSHPARAM(22); PARAMS_PUSHPARAM(23);
#define PARAMS_PUSHPARAM26 PARAMS_PUSHPARAM24; PARAMS_PUSHPARAM(24); PARAMS_PUSHPARAM(25);
#define PARAMS_PUSHPARAM28 PARAMS_PUSHPARAM26; PARAMS_PUSHPARAM(26); PARAMS_PUSHPARAM(27);
#define PARAMS_PUSHPARAM30 PARAMS_PUSHPARAM28; PARAMS_PUSHPARAM(28); PARAMS_PUSHPARAM(29);
#define PARAMS_PUSHPARAM32 PARAMS_PUSHPARAM30; PARAMS_PUSHPARAM(30); PARAMS_PUSHPARAM(31);
#define PARAMS_PUSHPARAM34 PARAMS_PUSHPARAM32; PARAMS_PUSHPARAM(32); PARAMS_PUSHPARAM(33);
#define PARAMS_PUSHPARAM36 PARAMS_PUSHPARAM34; PARAMS_PUSHPARAM(34); PARAMS_PUSHPARAM(35);
#define PARAMS_PUSHPARAM38 PARAMS_PUSHPARAM36; PARAMS_PUSHPARAM(36); PARAMS_PUSHPARAM(37);
#define PARAMS_PUSHPARAM40 PARAMS_PUSHPARAM38; PARAMS_PUSHPARAM(38); PARAMS_PUSHPARAM(39);
#define PARAMS_PUSHPARAM42 PARAMS_PUSHPARAM40; PARAMS_PUSHPARAM(40); PARAMS_PUSHPARAM(41);
#define PARAMS_PUSHPARAM44 PARAMS_PUSHPARAM42; PARAMS_PUSHPARAM(42); PARAMS_PUSHPARAM(43);
#define PARAMS_PUSHPARAM46 PARAMS_PUSHPARAM44; PARAMS_PUSHPARAM(44); PARAMS_PUSHPARAM(45);
#define PARAMS_PUSHPARAM48 PARAMS_PUSHPARAM46; PARAMS_PUSHPARAM(46); PARAMS_PUSHPARAM(47);
#define PARAMS_PUSHPARAM50 PARAMS_PUSHPARAM48; PARAMS_PUSHPARAM(48); PARAMS_PUSHPARAM(49);
#define PARAMS_PUSHPARAM52 PARAMS_PUSHPARAM50; PARAMS_PUSHPARAM(50); PARAMS_PUSHPARAM(51);
#define PARAMS_PUSHPARAM54 PARAMS_PUSHPARAM52; PARAMS_PUSHPARAM(52); PARAMS_PUSHPARAM(53);
#define PARAMS_PUSHPARAM56 PARAMS_PUSHPARAM54; PARAMS_PUSHPARAM(54); PARAMS_PUSHPARAM(55);
#define PARAMS_PUSHPARAM58 PARAMS_PUSHPARAM56; PARAMS_PUSHPARAM(56); PARAMS_PUSHPARAM(57);
#define PARAMS_PUSHPARAM60 PARAMS_PUSHPARAM58; PARAMS_PUSHPARAM(58); PARAMS_PUSHPARAM(59);
#define PARAMS_PUSHPARAM62 PARAMS_PUSHPARAM60; PARAMS_PUSHPARAM(60); PARAMS_PUSHPARAM(61);
#define PARAMS_PUSHPARAM64 PARAMS_PUSHPARAM62; PARAMS_PUSHPARAM(62); PARAMS_PUSHPARAM(63);
#define PARAMS_PUSHPARAM_even_nr_args(NrArgs) { ;PARAMS_PUSHPARAM##NrArgs; }
#define PARAMS_PUSHPARAM_odd_nr_args(NrArgs) { ;PARAMS_PUSHPARAM##NrArgs; PARAMS_PUSHPARAM(NrArgs); }

#define PARAMS_DEFINE_4_CALLING_SCHEMAS(nr_args, nr_args_plus2) \
	PARAMS_ADDPARAM_even_nr_args(nr_args) PARAMS_PUSHPARAM_even_nr_args(nr_args) \
	PARAMS_ADDPARAM_odd_nr_args(nr_args) PARAMS_PUSHPARAM_odd_nr_args(nr_args) \
	PARAMS_ADDPARAM_even_nr_args(nr_args_plus2) PARAMS_PUSHPARAM_even_nr_args(nr_args_plus2) \
	PARAMS_ADDPARAM_odd_nr_args(nr_args_plus2) PARAMS_PUSHPARAM_odd_nr_args(nr_args_plus2)

struct PARAMS
{
	nlist<const Type*> types;
	PARAMS() { }
	PARAMS(const Type& type0)
		{ types.push_back(cc->RegisterType->Register(&type0)); }
	PARAMS(const Type& type0, const Type& type1)
		{ types.push_back(cc->RegisterType->Register(&type0)); types.push_back(cc->RegisterType->Register(&type1)); }
	PARAMS_ADDPARAM_odd_nr_args(2) PARAMS_PUSHPARAM_odd_nr_args(2) // 3 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(4, 6) // 4-7 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(8, 10) // 8-11 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(12, 14) // 12-15 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(16, 18) // 16-19 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(20, 22) // 20-23 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(24, 26) // 24-27 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(28, 30) // 28-31 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(32, 34) // 32-35 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(36, 38) // 36-39 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(40, 42) // 40-43 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(44, 46) // 44-47 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(48, 50) // 48-51 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(52, 54) // 52-55 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(56, 58) // 56-59 Parameter
	PARAMS_DEFINE_4_CALLING_SCHEMAS(60, 62) // 60-63 Parameter
	PARAMS_ADDPARAM_even_nr_args(64) PARAMS_PUSHPARAM_even_nr_args(64) // 64 Parameter
};

GeneralExpr cast(const CollectGeneralType& ex, const Type& type);

struct PointerOrElementaryType
{
	private: const Type *type; // type==0, wenn expr
	public:
	expr ex; // Wenn PointerOrElementaryType aus einem PointerExpr erzeugt wird, steht in ex der CAST des Pointers in
		// den Integer, der den Pointer reprsentiert. ****** Dieser Integer muss immer den Typ Value::VarIndexTypeNr haben. ******
	void set_exported() const { ex.set_exported(); }
	void set_not_exported() const { ex.set_not_exported(); }
	void vset_exported() const { set_exported(); }
	void vset_not_exported() const { set_not_exported(); }
	const Type *get_type() const { return 0==type ? (Type*)(&BasicType::get_BasicType_from_type_nr(ex.get_type())).get() : type; }
	nstring get_type_string() const { return get_type()->get_type_string(); }
	operator expr() const { return ex; }

	template<typename T> PointerOrElementaryType(const T *x)
		{ report_error("Cannot use a C pointer other than 'const char *' for a function call. Please use the type POINTER instead."); }
	PointerOrElementaryType(const char *x) { construct(STRING(x)); }
	PointerOrElementaryType(bool x) : ex(x) { type=0; }
		// Auch ein bool soll sich in eine IntegerKONSTANTE konvertieren lassen (um sie an einen Zeiger zuzuweisen (Nullzeiger)).
	PointerOrElementaryType(signed char x) : ex(x) { type=0; }
	PointerOrElementaryType(unsigned char x) : ex(x) { type=0; }
	PointerOrElementaryType(short x) : ex(x) { type=0; }
	PointerOrElementaryType(unsigned short x) : ex(x) { type=0; }
	PointerOrElementaryType(int x) : ex(x) { type=0; }
	PointerOrElementaryType(unsigned int x) : ex(x) { type=0; }
	PointerOrElementaryType(long x) : ex(x) { type=0; }
	PointerOrElementaryType(unsigned long x) : ex(x) { type=0; }
	PointerOrElementaryType(long long x) : ex(x) { type=0; }
	PointerOrElementaryType(unsigned long long x) : ex(x) { type=0; }
	PointerOrElementaryType(float x) : ex((real_type)x) { type=0; }
	PointerOrElementaryType(double x) : ex((real_type)x) { type=0; }

	PointerOrElementaryType(const expr& elemex) : ex(elemex) { elemex.set_exported(); type=0; }

	PointerOrElementaryType(const PointerExpr& ptr) { ptr.set_exported(); construct(ptr); }
	void construct(const PointerExpr& ptr)
	{
		ptr.set_exported();
		if(ptr.get_actual_type()->is_expr())
		{
			ex.assign(ptr.to_expr());
			type=0;
		}
		else if(ptr.get_actual_type()->is_ptr())
		{	// ****** Die Addresse muss immer den Typ Value::VarIndexTypeNr haben. ******
			ex.assign(cast(ptr, BasicType::get_BasicType_from_type_nr(Value::VarIndexTypeNr)));
			type=ptr.get_actual_type();
		}
		else
			report_error(printf_string("Objects of type STRUCT or ARRAY cannot be taken as function arguments. Cannot take an object of type\n   %s\nas function argument.",
				ptr.get_actual_type()));
		set_not_exported();
	}
	PointerOrElementaryType(const GeneralExpr& genex)
	{
		genex.set_exported();
		if(genex.is_expr_type())
			{ ex.assign(genex.to_expr()); type=0; }
		else if(genex.is_PointerExpr())
			construct(genex.to_PointerExpr());
		else
			report_error(printf_string("Objects of type STRUCT or ARRAY cannot be taken as function arguments. Cannot take an object of type\n   %s\nas function argument.",
				genex.get_type()->get_type_string().c_str()));
		set_not_exported();
	}
	PointerOrElementaryType(const LogExpr& value) : ex(value.to_expr()) { value.set_exported(); type=0; }
	PointerOrElementaryType(const x_And_y& value) : ex(value.to_expr()) { value.set_exported(); type=0; }
	PointerOrElementaryType(const x_Or_y& value) : ex(value.to_expr()) { value.set_exported(); type=0; }
	PointerOrElementaryType(const BRO_x_And_y_BRC_And_z& value) : ex(value.to_expr()) { value.set_exported(); type=0; }
	PointerOrElementaryType(const BRO_x_Or_y_BRC_Or_z& value) : ex(value.to_expr()) { value.set_exported(); type=0; }
	PointerOrElementaryType(const BRO_x_And_y_BRC_Or_z& value) : ex(value.to_expr()) { value.set_exported(); type=0; }
	PointerOrElementaryType(const x_And_BRO_y_Or_z_BRC& value) : ex(value.to_expr()) { value.set_exported(); type=0; }
	PointerOrElementaryType(const BRO_x_Or_y_BRC_And_z& value) : ex(value.to_expr()) { value.set_exported(); type=0; }
	PointerOrElementaryType(const x_Or_BRO_y_And_z_BRC& value) : ex(value.to_expr()) { value.set_exported(); type=0; }

	PointerOrElementaryType(const local& x);
	PointerOrElementaryType(const param& x);
	PointerOrElementaryType(const Variable& x);
	PointerOrElementaryType(const Constant& x);

	PointerOrElementaryType(const LocalPlusPlus& x);
	PointerOrElementaryType(const LocalMinusMinus& x);
	PointerOrElementaryType(const ParamPlusPlus& x);
	PointerOrElementaryType(const ParamMinusMinus& x);
	PointerOrElementaryType(const VariablePlusPlus& x);
	PointerOrElementaryType(const VariableMinusMinus& x);
	PointerOrElementaryType(const DrefPointerPlusPlus& x);
	PointerOrElementaryType(const DrefPointerMinusMinus& x);
};

CollectGeneralType::CollectGeneralType(const GeneralExpr& genex)
{
	genex.set_exported();
	if(genex.is_expr_type())
		ex=genex.to_expr().vcopy();
	else if(genex.is_PointerExpr())
		ex=genex.to_PointerExpr().vcopy();
	else
		report_error("Unknow type in GeneralExpr");
	cval=0; creal=0;
	if(true==(is_basicconst = genex.is_expr_type() && genex.to_expr().is_integer_constant()))
		cval=genex.to_expr().get_this_integer_constant();
	set_not_exported();
}

CollectGeneralType::CollectGeneralType(const char *x)
{
	ex=new PointerExpr(STRING(x));
	cval=0; creal=0; is_basicconst=false;
	set_not_exported();
}

bool CollectGeneralType::is_PointerExpr() const
{
	set_exported();
	return 0!=dynamic_cast<const PointerExpr*>(ex);
}

bool CollectGeneralType::is_pointer_type() const
{
	return 0!=dynamic_cast<const PointerExpr*>(ex) && static_cast<const PointerExpr*>(ex)->get_actual_type()->is_ptr();
}

PointerExpr CollectGeneralType::to_PointerExpr(const PointerType* requested_pointer_type/*=0*/) const
{
	set_exported();
	if(0==requested_pointer_type)
	{
		if(!is_PointerExpr())
			report_error(printf_string("Cannot convert an expression of type\n   %s\nto a pointer expression.",
				get_type_string().c_str()));
		return *static_cast<const PointerExpr*>(ex);
	}
	if(is_expr_type())
	{
		if(!is_basicconst || 0!=cval)
			report_error(printf_string("Cannot convert an expression of type\n   %s\nto a pointer expression. The only non-pointer expression convertable to a pointer expression is the integer constant 0.",
				get_type_string().c_str()));
		return PointerExpr(*requested_pointer_type, expr((VARIABLE_INDEX_TYPE)0), false); // Nullzeiger
	}
	if(!is_PointerExpr() || 
			(!static_cast<const PointerExpr*>(ex)->get_actual_type()->equal(requested_pointer_type) &&
			!requested_pointer_type->type_pointed_to->equal(&__VOID)))
		report_error(printf_string("Cannot convert an expression of type\n   %s\ninto a pointer expression of type\n   %s",
			get_type_string().c_str(), requested_pointer_type->get_type_string().c_str()));
	return *static_cast<const PointerExpr*>(ex);
}

nstring CollectGeneralType::get_type_string() const
{
	if(is_expr_type())
		return to_expr().get_type_string();
	if(is_PointerExpr())
		return to_PointerExpr().get_type_string();
	report_error("Unknow type in CollectGeneralType");
}

const Type *CollectGeneralType::get_type() const
{
	set_exported();
	if(is_expr_type())
		return &BasicType::get_BasicType_from_type_nr(to_expr().set_exported().get_type());
	else if(is_PointerExpr())
		return to_PointerExpr().get_actual_type();
	else
		report_error("Illegal types for get_type()");
}

CollectGeneralType::CollectGeneralType(const LogExpr& x)
	{ x.set_exported(); ex=new expr(x.to_expr()); cval=0; creal=0; is_basicconst=false; set_not_exported(); }
CollectGeneralType::CollectGeneralType(const x_And_y& x)
	{ x.set_exported(); ex=new expr(x.to_expr()); cval=0; creal=0; is_basicconst=false; set_not_exported(); }
CollectGeneralType::CollectGeneralType(const x_Or_y& x)
	{ x.set_exported(); ex=new expr(x.to_expr()); cval=0; creal=0; is_basicconst=false; set_not_exported(); }
CollectGeneralType::CollectGeneralType(const BRO_x_And_y_BRC_And_z& x)
	{ x.set_exported(); ex=new expr(x.to_expr()); cval=0; creal=0; is_basicconst=false; set_not_exported(); }
CollectGeneralType::CollectGeneralType(const BRO_x_Or_y_BRC_Or_z& x)
	{ x.set_exported(); ex=new expr(x.to_expr()); cval=0; creal=0; is_basicconst=false; set_not_exported(); }
CollectGeneralType::CollectGeneralType(const BRO_x_And_y_BRC_Or_z& x)
	{ x.set_exported(); ex=new expr(x.to_expr()); cval=0; creal=0; is_basicconst=false; set_not_exported(); }
CollectGeneralType::CollectGeneralType(const x_And_BRO_y_Or_z_BRC& x)
	{ x.set_exported(); ex=new expr(x.to_expr()); cval=0; creal=0; is_basicconst=false; set_not_exported(); }
CollectGeneralType::CollectGeneralType(const BRO_x_Or_y_BRC_And_z& x)
	{ x.set_exported(); ex=new expr(x.to_expr()); cval=0; creal=0; is_basicconst=false; set_not_exported(); }
CollectGeneralType::CollectGeneralType(const x_Or_BRO_y_And_z_BRC& x)
	{ x.set_exported(); ex=new expr(x.to_expr()); cval=0; creal=0; is_basicconst=false; set_not_exported(); }

struct Function
{
	bool extern_function;
	PROGRAM_COUNTER_TYPE extern_function_number;
	Command::uid_type uid, header_uid;
	size_t akt_arg, akt_arg_bytecount;
	nlist<const Type*> argtypes;
	nlist<const Type*>::iterator akt_arg_it;
	const char *functionname;
	bool is_already_defined;
	const Type* return_type;
	Function(const char *_functionname, const Type& __return_type, const PARAMS& parameters, bool is_extern_function,
				PROGRAM_COUNTER_TYPE __extern_function_number=0)
			: functionname(_functionname)
	{
		__return_type.chk_validity_of_ref_types_variable_creation();
		return_type=cc->RegisterType->Register(&__return_type);
		if(!return_type->is_expr() && !return_type->is_ptr())
			report_error(printf_string("Function %s : A function can only return objects with elementary types of pointers to any type.",
				_functionname));
		extern_function=is_extern_function;
		akt_arg=akt_arg_bytecount=0;
		uid=Command::generate_new_uid();
		header_uid=Command::generate_new_uid();
		is_already_defined=false;
		argtypes=parameters.types;
		akt_arg_it=argtypes.begin();
		extern_function_number=__extern_function_number;
		cc->functionUID2name.insert(FunctionUIDtoFunctionName(uid, functionname));
	}
	void Return(const CollectGeneralType& retval)
	{
		retval.set_exported();
		expr ex;
		if(return_type->is_expr())
			ex.assign(cast(retval, *return_type).to_expr());
		else if(return_type->is_ptr())
			ex.assign(cast(retval, *return_type).to_PointerExpr().expression_yielding_address());
		Command cmd(bi::RET);
		cmd.set_sp(-1);
		ex.add(cmd);
		ex.set_code_has_effect();
		Statement stm(ex); // Erzeugung von Code beim Aufruf des Destruktors von stm
	}
	GeneralExpr functioncall_helper(const nlist<PointerOrElementaryType>& args)
	{
		nlist<PointerOrElementaryType>::const_iterator it;
		if(args.size()!=argtypes.size())
			report_error(printf_string("Function %s must be called with %d arguments.", functionname, argtypes.size()));
		long arg_count=0;
		for(akt_arg_it=argtypes.begin(), it=args.begin(); akt_arg_it!=argtypes.end(); akt_arg_it++, it++, ++arg_count)
		{
			it->set_exported(); // wichtig !
			it->ex.set_code_has_effect();
			if(((*akt_arg_it)->is_expr()) ? (!(*it).get_type()->is_expr()) : (!(*akt_arg_it)->equal((*it).get_type())))
			{
				if((*akt_arg_it)->is_ptr() && (*it).get_type()->is_expr())
				{
					if(!expr(*it).is_integer_constant() || 0!=expr(*it).get_this_integer_constant())
						report_error(printf_string("Call of Function %s: Parameter number %u has been declared to be of type\n   %s\nbut in the function call appears a parameter of type\n   %s\nnot beeing convertable to the required type. \
The only non-pointer expression convertable to a pointer is the integer constant 0.",
							functionname, 1+arg_count, (*akt_arg_it)->get_type_string().c_str(), (*it).get_type_string().c_str()));
					else
						; // OK: Konversion der Konstante 0 in den Nullzeiger.
				}
				else
					report_error(printf_string("Call of Function %s: Parameter number %u has been declared to be of type\n   %s\nbut in the function call appears a parameter of type\n   %s\nnot beeing convertable to the required type.",
						functionname, 1+arg_count, (*akt_arg_it)->get_type_string().c_str(), (*it).get_type_string().c_str()));
			}
		}
		expr ex;
		for(it=args.end(), akt_arg_it=argtypes.end(); it!=args.begin(); )
		{
			--akt_arg_it; --it;
			if((*akt_arg_it)->is_ptr() && (*it).get_type()->is_expr())
				ex.add(expr::get_const_expr(0)); // Nullzeiger. Dass (*it).get_type()->is_expr() die Konstante 0 ist, wurde schon
					// oben sichergestellt. Es darf die Konstante 0 vom Typ int8 statt vom Typ VARIABLE_INDEX_TYPE eingefgt werden,
					// weil der Bytecodeinterpreter sie beim Zugriff automatisch nach VARIABLE_INDEX_TYPE konvertiert.
			else
				ex.add(expr(*it).shortest_type_if_constant()); // Falls *it nicht den korrekten Typ hat, muss hier kein bi::CAST-Kommando eingefgt werden, denn
			// wenn der Bytecode-Interpreter auf das Funktionsargument zugereift, fhrt er automatisch eine entsprechende Konvertierung durch.
		}
		akt_arg_it=argtypes.begin();
		if(extern_function)
		{
			ex.add(expr::get_const_expr(extern_function_number));
			ex.addcmd(bi::CALLEXT, -argtypes.size());
		}
		else
		{
			Command cmd;
			cmd.fkt=Command::functioncall;
			cmd.branch1_id=uid;
			cmd.branch2_id=header_uid;
			cmd.sp=1-(int32)argtypes.size();
			ex.add(cmd);
		}
		if(return_type->is_expr())
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(return_type)->type));
		else if(return_type->is_ptr())
			return GeneralExpr::from_PointerExpr(PointerExpr(*return_type, ex.set_type(Value::VarIndexTypeNr), false));
		else
			report_error(printf_string("Function %s : A function can only return objects with elementary types of pointers to any type.",
				functionname));
	}
	GeneralExpr operator()()
	{
		nlist<PointerOrElementaryType> args;
		return functioncall_helper(args);
	}
	GeneralExpr operator()(const PointerOrElementaryType& arg0)
	{
		nlist<PointerOrElementaryType> args;
		arg0.set_exported(); args.push_front(arg0);
		return functioncall_helper(args);
	}
	GeneralExpr operator()(const PointerOrElementaryType& arg0, const PointerOrElementaryType& arg1)
	{
		nlist<PointerOrElementaryType> args;
		arg1.set_exported(); args.push_front(arg1);
		arg0.set_exported(); args.push_front(arg0);
		return functioncall_helper(args);
	}
	Function_FuncOP_odd_nr_args(2); // 3 Parameter
	Function_4FuncOPs(4, 6);	// 4 bis 7 Parameter
	Function_4FuncOPs(8, 10);	// 8 bis 11 Parameter
	Function_4FuncOPs(12, 14);	// 12 bis 13 Parameter
	Function_4FuncOPs(16, 18);	// 16 bis 19 Parameter
	Function_4FuncOPs(20, 22);	// 20 bis 23 Parameter
	Function_4FuncOPs(24, 26);	// 24 bis 27 Parameter
	Function_4FuncOPs(28, 30);	// 28 bis 31 Parameter
	Function_4FuncOPs(32, 34);	// 32 bis 35 Parameter
	Function_4FuncOPs(36, 38);	// 36 bis 39 Parameter
	Function_4FuncOPs(40, 42);	// 40 bis 43 Parameter
	Function_4FuncOPs(44, 46);	// 44 bis 47 Parameter
	Function_4FuncOPs(48, 50);	// 48 bis 51 Parameter
	Function_4FuncOPs(52, 54);	// 52 bis 55 Parameter
	Function_4FuncOPs(56, 58);	// 56 bis 59 Parameter
	Function_4FuncOPs(60, 62);	// 60 bis 63 Parameter
	Function_FuncOP_even_nr_args(64); // 64 Parameter
};

struct ProgramStructure_Function : public ProgramStructureElement
{
private: Code code;
public:
	Function& func;
	nlist<Command::uid_type> jump_command_locations;
	uint32 local_variable_count, byte_addressed_local_variable_count, max_byte_addressed_local_variable_count;
	uint32 first_uninit_vars_created_bytesize, uninit_vars_created_local_stack_offset, uninit_vars_created_bytesize;
	nlist<uint8> first_uninit_vars_created_types, uninit_vars_created_types;
	bool first_uninit_vars_section;

	ProgramStructure_Function(Function& funcobj) : func(funcobj)
	{
		local_variable_count=byte_addressed_local_variable_count=max_byte_addressed_local_variable_count=0;
		uninit_vars_created_local_stack_offset=first_uninit_vars_created_bytesize=uninit_vars_created_bytesize=0;
		first_uninit_vars_section=true;
	}
	VARIABLE_INDEX_TYPE get_next_unused_local_var_address(const Type *type)
	{
		if((byte_addressed_local_variable_count+=type->size()) > max_byte_addressed_local_variable_count)
			max_byte_addressed_local_variable_count=byte_addressed_local_variable_count;
		return byte_addressed_local_variable_count;
	}
	void free_local_var_memory(VARIABLE_INDEX_TYPE size)
	{
		byte_addressed_local_variable_count-=size;
		if(first_uninit_vars_section)
		{
			uint32 addr=0;
			for(nlist<uint8>::iterator it=first_uninit_vars_created_types.begin(); it!=first_uninit_vars_created_types.end(); it++)
				if((addr+=Value::sbytesize(*it)) > byte_addressed_local_variable_count)
				{
					while(it!=first_uninit_vars_created_types.end())
					{
						nlist<uint8>::iterator aktit=it; it++;
						first_uninit_vars_created_bytesize-=Value::sbytesize(*aktit);
						first_uninit_vars_created_types.erase(aktit);
					}
					break;
				}
		}
		else
		{
			uint32 addr=uninit_vars_created_local_stack_offset;
			for(nlist<uint8>::iterator it=uninit_vars_created_types.begin(); it!=uninit_vars_created_types.end(); it++)
				if((addr+=Value::sbytesize(*it)) > byte_addressed_local_variable_count)
				{
					while(it!=uninit_vars_created_types.end())
					{
						nlist<uint8>::iterator aktit=it; it++;
						uninit_vars_created_bytesize-=Value::sbytesize(*aktit);
						uninit_vars_created_types.erase(aktit);
					}
					break;
				}
		}
	}
	void add(const Code& ex)
	{
		if(!first_uninit_vars_section && uninit_vars_created_types.size()!=0)
		{
			code.add(Command(bi::SETTYPES).add_coded_integer255(uninit_vars_created_local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, false));
			code.add(write_uninit_vars_typeinfo(uninit_vars_created_types));
			uninit_vars_created_types.clear();
			uninit_vars_created_bytesize=0;
		}
		code.add(ex);
		first_uninit_vars_section=false;
	}
	void end(ProgramStructureElement *superordinate_element)
	{
		if(func.akt_arg!=func.argtypes.size())
			report_error(printf_string("Function %s: Too few formal params defined (required: exactly %u)", func.functionname, func.argtypes.size()));
		for(nlist<Command::uid_type>::iterator it=jump_command_locations.begin(); it!=jump_command_locations.end(); it++)
		{
			nlist<Command>::iterator p=find_command_from_uid(code, *it);
			if(!p.is_valid_and_not_end(code.code))
				report_error("Internal Compiler Error");
			bool found=false;
			if(cc->ProgramStructureStack.size()<2)
				report_error(p->fkt==Command::break_statement ? "BREAK occured outside of loop" :
						(p->fkt==Command::continue_statement ? "CONTINUE occured outside of loop" : "GOTO occured outside of loop"));
			nlist<ProgramStructureElement*>::iterator it2=cc->ProgramStructureStack.end();
			while(it2!=cc->ProgramStructureStack.begin())
			{
				if(0!=dynamic_cast<ProgramStructure_IF_BASE*>(*(--it2)) &&
						ProgramStructure_IF_BASE::if_element!=static_cast<ProgramStructure_IF_BASE*>(*it2)->type)
				{
					found=true;
					static_cast<ProgramStructure_IF_BASE*>(*it2)->jump_command_locations.push_back(*it);
					break;
				}
				else if(0!=dynamic_cast<ProgramStructure_Block*>(*it2))
				{
					found=true;
					static_cast<ProgramStructure_Block*>(*it2)->jump_command_locations.push_back(*it);
					break;
				}
			}
			if(!found)
				report_error(p->fkt==Command::break_statement ? "BREAK occured outside of loop" :
						(p->fkt==Command::continue_statement ? "CONTINUE occured outside of loop" : "GOTO occured outside of loop"));
		}

		// Am Ende des Codes muss hier nicht POPN ausgefhrt werden, weil bei nicht-void-Funktionen immer ein
		// RETURN notwendig ist und hierbei gleichzeitig ein POPN ausgefhrt wird. Statt dessen wird im folgenden ein
		// Label eingefgt, in welchem der Stackpointer entsprechend einem POPN-Befehl gendert wird.
		Command subsp=Command::new_uid_label();
		subsp.sp=-(int32)local_variable_count;
		code.add(subsp);

#ifdef AUTO_TYPEINFO_FOR_UNINITIALIZED_LOCALS
		code.addfront(Command::GetRawBytesCommand().add_coded_integer255(
			max_byte_addressed_local_variable_count-first_uninit_vars_created_bytesize, MAX_NR_BYTES_FOR_VARIABLE_INDEX, false));
		code.addfront(write_uninit_vars_typeinfo(first_uninit_vars_created_types)); // auch im Fall 0==first_uninit_vars_created_types.size() !
#endif

		Command fktcmd=Command::NewUID(func.uid);
		fktcmd.fkt=Command::function;
		code.addfront(fktcmd);

		// Vor den ausfhrbaren Code wird in einem Byte die Zahl der Parameter und die Information, ob die Funktion einen
		// Rckgabewert liefert, und danach die Typinformationen der Funktionsargumente abgespeichert.
		code.addfront(Command::GetRawBytesCommand().add8(func.argtypes.size()*2+1)); // +0 ==> keine void-Funktion (Rckgabewert wird geliefert)
		for(nlist<const Type*>::iterator it=func.argtypes.begin(); it!=func.argtypes.end(); it++)
			if(0!=dynamic_cast<const BasicType*>(*it))
				code.addfront(Command::GetRawBytesCommand().addnibble(static_cast<const BasicType*>(*it)->type));
			else if(0!=dynamic_cast<const PointerType*>(*it))
				code.addfront(Command::GetRawBytesCommand().addnibble(Value::VarIndexTypeNr));
			else
				report_error("Illegal type of function parameter.");

		Command headercmd=Command::NewUID(func.header_uid);
		headercmd.fkt=Command::begin_infos_before_function;
		code.addfront(headercmd);

		superordinate_element->add(code);
	}
	void add_local_variable(const Command& var_label, const expr& init_value)
	{
		init_value.set_exported();
		++local_variable_count;
		add(init_value);
		add(var_label);
	}
	VARIABLE_INDEX_TYPE add_uninit_local_variable(const Type *type);
};

#define BYTECODE_TARGETED_COMPILER__DECLARE_FUNCTION(return_type, name, parameters) ;BytecodeTargetedCompiler::Function name(#name, return_type, parameters, false);
#define BYTECODE_TARGETED_COMPILER__function(name) ;BytecodeTargetedCompiler::FunctionDefinitionHandler(name); {
#define BYTECODE_TARGETED_COMPILER__DECLARE_EXTERN_FUNCTION(return_type, name, parameters) ;BytecodeTargetedCompiler::Function name(#name, return_type, parameters, true, extern_function_names::name);

void FunctionDefinitionHandler(Function &func)
{
	if(func.extern_function)
		report_error(printf_string("The function %s is declared extern and therefore no code can be defined for this function.",
			func.functionname));
	if(cc->ProgramStructureStack.size()!=1 || 0==dynamic_cast<ProgramStructureGlobalScope*>(*(--(cc->ProgramStructureStack.end()))))
		report_error("Function defined within a function or outside of the global scope (between BEGIN_GLOBAL_SCOPE and END_GLOBAL_SCOPE)");
	if(func.is_already_defined)
		report_error(printf_string("The function %s is already defined", func.functionname));
	cc->defined_goto_labels_of_actual_function.clear();
	cc->forward_goto_targets_of_actual_function.clear();
	cc->ProgramStructureStack.push_back(new ProgramStructure_Function(func));
	func.is_already_defined=true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////

struct VoidFunction
{
	bool extern_function;
	PROGRAM_COUNTER_TYPE extern_function_number;
	Command::uid_type uid, header_uid;
	size_t akt_arg, akt_arg_bytecount;
	nlist<const Type*> argtypes;
	nlist<const Type*>::iterator akt_arg_it;
	const char *functionname;
	bool is_already_defined;
	VoidFunction(const char *_functionname, const PARAMS& parameters, bool is_extern_function,
					PROGRAM_COUNTER_TYPE __extern_function_number=0)
		: functionname(_functionname)
	{
		extern_function=is_extern_function;
		akt_arg=akt_arg_bytecount=0;
		uid=Command::generate_new_uid();
		header_uid=Command::generate_new_uid();
		is_already_defined=false;
		argtypes=parameters.types;
		akt_arg_it=argtypes.begin();
		extern_function_number=__extern_function_number;
			// Function::next_extern_function_number nur inkrementieren, wenn true==is_extern_function.
		cc->functionUID2name.insert(FunctionUIDtoFunctionName(uid, functionname));
	}
	void Return()
	{
		expr ex;
		Command cmd(bi::RET);
		ex.add_effective(cmd);
		Statement stm(ex); // Erzeugung von Code beim Aufruf des Destruktors von stm
	}
	void functioncall_helper(const nlist<PointerOrElementaryType>& args)
	{
		nlist<PointerOrElementaryType>::const_iterator it;
		if(args.size()!=argtypes.size())
			report_error(printf_string("Function %s must be called with %d arguments.", functionname, argtypes.size()));
		long arg_count=0;
		for(akt_arg_it=argtypes.begin(), it=args.begin(); akt_arg_it!=argtypes.end(); akt_arg_it++, it++, ++arg_count)
		{
			it->set_exported(); // wichtig !
			it->ex.set_code_has_effect();
			if(((*akt_arg_it)->is_expr()) ? (!(*it).get_type()->is_expr()) : (!(*akt_arg_it)->equal((*it).get_type())))
			{
				if((*akt_arg_it)->is_ptr() && (*it).get_type()->is_expr())
				{
					if(!expr(*it).is_integer_constant() || 0!=expr(*it).get_this_integer_constant())
						report_error(printf_string("Call of Function %s: Parameter number %u has been declared to be of type\n   %s\nbut in the function call appears a parameter of type\n   %s\nnot beeing convertable to the required type. \
The only non-pointer expression convertable to a pointer is the integer constant 0.",
							functionname, 1+arg_count, (*akt_arg_it)->get_type_string().c_str(), (*it).get_type_string().c_str()));
					else
						; // OK: Konversion der Konstante 0 in den Nullzeiger.
				}
				else
					report_error(printf_string("Call of Function %s: Parameter number %u has been declared to be of type\n   %s\nbut in the function call appears a parameter of type\n   %s\nnot beeing convertable to the required type.",
						functionname, 1+arg_count, (*akt_arg_it)->get_type_string().c_str(), (*it).get_type_string().c_str()));
			}
		}
		expr ex;
		for(it=args.end(), akt_arg_it=argtypes.end(); it!=args.begin(); )
		{
			--akt_arg_it; --it;
			if((*akt_arg_it)->is_ptr() && (*it).get_type()->is_expr())
				ex.add(expr::get_const_expr(0)); // Nullzeiger. Dass (*it).get_type()->is_expr() die Konstante 0 ist, wurde schon
					// oben sichergestellt. Es darf die Konstante 0 vom Typ int8 statt vom Typ VARIABLE_INDEX_TYPE eingefgt werden,
					// weil der Bytecodeinterpreter sie beim Zugriff automatisch nach VARIABLE_INDEX_TYPE konvertiert.
			else
				ex.add(expr(*it).shortest_type_if_constant()); // Falls *it nicht den korrekten Typ hat, muss hier kein bi::CAST-Kommando eingefgt werden, denn
			// wenn der Bytecode-Interpreter auf das Funktionsargument zugereift, fhrt er automatisch eine entsprechende Konvertierung durch.
		}
		akt_arg_it=argtypes.begin();
		if(extern_function)
		{
			ex.add(expr::get_const_expr(extern_function_number));
			ex.addcmd(bi::CALLEXT, -1-argtypes.size());
		}
		else
		{
			Command cmd;
			cmd.fkt=Command::functioncall;
			cmd.branch1_id=uid;
			cmd.branch2_id=header_uid;
			cmd.sp=-(int32)argtypes.size();
			ex.add(cmd);
		}
		ex.set_type(Value::t_nil);
		Statement stm=ex; // Mit dem Aufruf des Destruktors von Statement wird Code erzeugt.
	}
	void operator()()
	{
		nlist<PointerOrElementaryType> args;
		functioncall_helper(args);
	}
	void operator()(const PointerOrElementaryType& arg0)
	{
		nlist<PointerOrElementaryType> args;
		arg0.set_exported(); args.push_front(arg0);
		functioncall_helper(args);
	}
	void operator()(const PointerOrElementaryType& arg0, const PointerOrElementaryType& arg1)
	{
		nlist<PointerOrElementaryType> args;
		arg1.set_exported(); args.push_front(arg1);
		arg0.set_exported(); args.push_front(arg0);
		functioncall_helper(args);
	}
	VoidFunction_FuncOP_odd_nr_args(2); // 3 Parameter
	VoidFunction_4FuncOPs(4, 6);	// 4 bis 7 Parameter
	VoidFunction_4FuncOPs(8, 10);	// 8 bis 11 Parameter
	VoidFunction_4FuncOPs(12, 14);	// 12 bis 13 Parameter
	VoidFunction_4FuncOPs(16, 18);	// 16 bis 19 Parameter
	VoidFunction_4FuncOPs(20, 22);	// 20 bis 23 Parameter
	VoidFunction_4FuncOPs(24, 26);	// 24 bis 27 Parameter
	VoidFunction_4FuncOPs(28, 30);	// 28 bis 31 Parameter
	VoidFunction_4FuncOPs(32, 34);	// 32 bis 35 Parameter
	VoidFunction_4FuncOPs(36, 38);	// 36 bis 39 Parameter
	VoidFunction_4FuncOPs(40, 42);	// 40 bis 43 Parameter
	VoidFunction_4FuncOPs(44, 46);	// 44 bis 47 Parameter
	VoidFunction_4FuncOPs(48, 50);	// 48 bis 51 Parameter
	VoidFunction_4FuncOPs(52, 54);	// 52 bis 55 Parameter
	VoidFunction_4FuncOPs(56, 58);	// 56 bis 59 Parameter
	VoidFunction_4FuncOPs(60, 62);	// 60 bis 63 Parameter
	VoidFunction_FuncOP_even_nr_args(64); // 64 Parameter
};

struct ProgramStructure_VoidFunction : public ProgramStructureElement
{
	VoidFunction& func;
	Code code;
	nlist<Command::uid_type> jump_command_locations;
	uint32 local_variable_count, byte_addressed_local_variable_count, max_byte_addressed_local_variable_count;
	uint32 first_uninit_vars_created_bytesize, uninit_vars_created_local_stack_offset, uninit_vars_created_bytesize;
	nlist<uint8> first_uninit_vars_created_types, uninit_vars_created_types;
	bool first_uninit_vars_section;

	ProgramStructure_VoidFunction(VoidFunction& funcobj) : func(funcobj)
	{
		local_variable_count=byte_addressed_local_variable_count=max_byte_addressed_local_variable_count=0;
		uninit_vars_created_local_stack_offset=first_uninit_vars_created_bytesize=uninit_vars_created_bytesize=0;
		first_uninit_vars_section=true;
	}
	VARIABLE_INDEX_TYPE get_next_unused_local_var_address(const Type *type)
	{
		if((byte_addressed_local_variable_count+=type->size()) > max_byte_addressed_local_variable_count)
			max_byte_addressed_local_variable_count=byte_addressed_local_variable_count;
		return byte_addressed_local_variable_count;
	}
	void free_local_var_memory(VARIABLE_INDEX_TYPE size)
	{
		byte_addressed_local_variable_count-=size;
		if(first_uninit_vars_section)
		{
			uint32 addr=0;
			for(nlist<uint8>::iterator it=first_uninit_vars_created_types.begin(); it!=first_uninit_vars_created_types.end(); it++)
				if((addr+=Value::sbytesize(*it)) > byte_addressed_local_variable_count)
				{
					while(it!=first_uninit_vars_created_types.end())
					{
						nlist<uint8>::iterator aktit=it; it++;
						first_uninit_vars_created_bytesize-=Value::sbytesize(*aktit);
						first_uninit_vars_created_types.erase(aktit);
					}
					break;
				}
		}
		else
		{
			uint32 addr=uninit_vars_created_local_stack_offset;
			for(nlist<uint8>::iterator it=uninit_vars_created_types.begin(); it!=uninit_vars_created_types.end(); it++)
				if((addr+=Value::sbytesize(*it)) > byte_addressed_local_variable_count)
				{
					while(it!=uninit_vars_created_types.end())
					{
						nlist<uint8>::iterator aktit=it; it++;
						uninit_vars_created_bytesize-=Value::sbytesize(*aktit);
						uninit_vars_created_types.erase(aktit);
					}
					break;
				}
		}
	}
	void add(const Code& ex)
	{
		if(!first_uninit_vars_section && uninit_vars_created_types.size()!=0)
		{
			code.add(Command(bi::SETTYPES).add_coded_integer255(uninit_vars_created_local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, false));
			code.add(write_uninit_vars_typeinfo(uninit_vars_created_types));
			uninit_vars_created_types.clear();
			uninit_vars_created_bytesize=0;
		}
		code.add(ex);
		first_uninit_vars_section=false;
	}

	void end(ProgramStructureElement *superordinate_element)
	{
		if(func.akt_arg!=func.argtypes.size())
			report_error(printf_string("VoidFunction %s: Too few formal params defined (required: exactly %u)", func.functionname, func.argtypes.size()));
		for(nlist<Command::uid_type>::iterator it=jump_command_locations.begin(); it!=jump_command_locations.end(); it++)
		{
			nlist<Command>::iterator p=find_command_from_uid(code, *it);
			if(!p.is_valid_and_not_end(code.code))
				report_error("Internal Compiler Error");
			bool found=false;
			if(cc->ProgramStructureStack.size()<2)
				report_error(p->fkt==Command::break_statement ? "BREAK occured outside of loop" :
						(p->fkt==Command::continue_statement ? "CONTINUE occured outside of loop" : "GOTO occured outside of loop"));
			nlist<ProgramStructureElement*>::iterator it2=cc->ProgramStructureStack.end();
			while(it2!=cc->ProgramStructureStack.begin())
			{
				if(0!=dynamic_cast<ProgramStructure_IF_BASE*>(*(--it2)) &&
						ProgramStructure_IF_BASE::if_element!=static_cast<ProgramStructure_IF_BASE*>(*it2)->type)
				{
					found=true;
					static_cast<ProgramStructure_IF_BASE*>(*it2)->jump_command_locations.push_back(*it);
					break;
				}
				else if(0!=dynamic_cast<ProgramStructure_Block*>(*it2))
				{
					found=true;
					static_cast<ProgramStructure_Block*>(*it2)->jump_command_locations.push_back(*it);
					break;
				}
			}
			if(!found)
				report_error(p->fkt==Command::break_statement ? "BREAK occured outside of loop" :
						(p->fkt==Command::continue_statement ? "CONTINUE occured outside of loop" : "GOTO occured outside of loop"));
		}

		Command subsp=Command::new_uid_label();
		subsp.sp=-(int32)local_variable_count;
		code.add(subsp);

#ifdef AUTO_TYPEINFO_FOR_UNINITIALIZED_LOCALS
		code.addfront(Command::GetRawBytesCommand().add_coded_integer255(
			max_byte_addressed_local_variable_count-first_uninit_vars_created_bytesize, MAX_NR_BYTES_FOR_VARIABLE_INDEX, false));
		code.addfront(write_uninit_vars_typeinfo(first_uninit_vars_created_types)); // auch im Fall 0==first_uninit_vars_created_types.size() !
#endif

		Command fktcmd=Command::NewUID(func.uid);
		fktcmd.fkt=Command::function;
		code.addfront(fktcmd);

		if(0==func.argtypes.size())
			code.addcmd(bi::RET);
		else
		{
			Command cmd(bi::RET);
			code.add(cmd);
		}

		// Vor den ausfhrbaren Code wird in einem Byte die Zahl der Parameter und die Information, ob die Funktion einen
		// Rckgabewert liefert, und danach die Typinformationen der Funktionsargumente abgespeichert.
		code.addfront(Command::GetRawBytesCommand().add8(func.argtypes.size()*2+0)); // +0 ==> void (kein Rckgabewert)
		for(nlist<const Type*>::iterator it=func.argtypes.begin(); it!=func.argtypes.end(); it++)
			if(0!=dynamic_cast<const BasicType*>(*it))
				code.addfront(Command::GetRawBytesCommand().addnibble(static_cast<const BasicType*>(*it)->type));
			else if(0!=dynamic_cast<const PointerType*>(*it))
				code.addfront(Command::GetRawBytesCommand().addnibble(Value::VarIndexTypeNr));
			else
				report_error("Illegal type of function parameter.");

		Command headercmd=Command::NewUID(func.header_uid);
		headercmd.fkt=Command::begin_infos_before_function;
		code.addfront(headercmd);

		superordinate_element->add(code);
	}
	void add_local_variable(const Command& var_label, const expr& init_value)
	{
		init_value.set_exported();
		++local_variable_count;
		add(init_value);
		add(var_label);
	}
	VARIABLE_INDEX_TYPE add_uninit_local_variable(const Type *type);
};

VARIABLE_INDEX_TYPE ProgramStructure_Function::add_uninit_local_variable(const Type *type)
{
	ProgramStructureElement* it=GetActFunctionProgramStructure("Local variable defined outside of a function.");
	if(dynamic_cast<ProgramStructure_Function*>(it)!=this)
		report_error("Error in ProgramStructure_Function::add_uninit_local_variable()");
	nlist<uint8> lst=type->get_flattend_types_list();
	if(first_uninit_vars_section)
	{
		if(first_uninit_vars_created_bytesize!=byte_addressed_local_variable_count)
			report_error("Internal compiler error in function ProgramStructure_Function::add_uninit_local_variable()");
		first_uninit_vars_created_types.insert(first_uninit_vars_created_types.end(), lst.begin(), lst.end());
		first_uninit_vars_created_bytesize+=type->size();
	}
	else
	{
		if(0==uninit_vars_created_types.size())
		{
			uninit_vars_created_bytesize=0;
			uninit_vars_created_local_stack_offset=byte_addressed_local_variable_count;
		}
		else if(uninit_vars_created_local_stack_offset+uninit_vars_created_bytesize!=byte_addressed_local_variable_count)
			report_error("Internal compiler error in function ProgramStructure_Function::add_uninit_local_variable()");
		uninit_vars_created_types.insert(uninit_vars_created_types.end(), lst.begin(), lst.end());
		uninit_vars_created_bytesize+=type->size();
	}
	return get_next_unused_local_var_address(type);
}

VARIABLE_INDEX_TYPE ProgramStructure_VoidFunction::add_uninit_local_variable(const Type *type)
{
	ProgramStructureElement* it=GetActFunctionProgramStructure("Local variable defined outside of a function.");
	if(dynamic_cast<ProgramStructure_VoidFunction*>(it)!=this)
		report_error("Error in ProgramStructure_Function::add_uninit_local_variable()");
	nlist<uint8> lst=type->get_flattend_types_list();
	if(first_uninit_vars_section)
	{
		if(first_uninit_vars_created_bytesize!=byte_addressed_local_variable_count)
			report_error("Internal compiler error in function ProgramStructure_Function::add_uninit_local_variable()");
		first_uninit_vars_created_types.insert(first_uninit_vars_created_types.end(), lst.begin(), lst.end());
		first_uninit_vars_created_bytesize+=type->size();
	}
	else
	{
		if(0==uninit_vars_created_types.size())
		{
			uninit_vars_created_bytesize=0;
			uninit_vars_created_local_stack_offset=byte_addressed_local_variable_count;
		}
		else if(uninit_vars_created_local_stack_offset+uninit_vars_created_bytesize!=byte_addressed_local_variable_count)
			report_error("Internal compiler error in function ProgramStructure_Function::add_uninit_local_variable()");
		uninit_vars_created_types.insert(uninit_vars_created_types.end(), lst.begin(), lst.end());
		uninit_vars_created_bytesize+=type->size();
	}
	return get_next_unused_local_var_address(type);
}

VARIABLE_INDEX_TYPE ProgramStructure_ForLoop::add_uninit_local_variable(const Type *type)
{
	if(init_occured)
		report_error("Local variable cannot be defined after the first statement (init) between FOR and DO");
	ProgramStructureElement* it=GetActFunctionProgramStructure("Local variable defined outside of a function.");
	if(0==it || cc->ProgramStructureStack.size()<2 || cc->ProgramStructureStack.back()!=this)
		report_error("FOR loop error");
	if(0!=dynamic_cast<ProgramStructure_Function*>(it))
	{
		nlist<uint8> lst=type->get_flattend_types_list();
		if(0==uninit_vars_created_types.size())
			uninit_vars_created_local_stack_offset=
				static_cast<ProgramStructure_Function*>(it)->byte_addressed_local_variable_count;
		uninit_vars_created_types.insert(uninit_vars_created_types.end(), lst.begin(), lst.end());
		return static_cast<ProgramStructure_Function*>(it)->get_next_unused_local_var_address(type);
	}
	else
	{
		nlist<uint8> lst=type->get_flattend_types_list();
		if(0==uninit_vars_created_types.size())
			uninit_vars_created_local_stack_offset=
				static_cast<ProgramStructure_VoidFunction*>(it)->byte_addressed_local_variable_count;
		uninit_vars_created_types.insert(uninit_vars_created_types.end(), lst.begin(), lst.end());
		return static_cast<ProgramStructure_VoidFunction*>(it)->get_next_unused_local_var_address(type);
	}
}

VARIABLE_INDEX_TYPE ProgramStructure_Block::add_uninit_local_variable(const Type *type)
{
	ProgramStructureElement* it=GetActFunctionProgramStructure("Local variable defined outside of a function.");
	if(0==it || cc->ProgramStructureStack.size()<2 || cc->ProgramStructureStack.back()!=this)
		report_error("ProgramStructure_Block loop error");
	if(0!=dynamic_cast<ProgramStructure_Function*>(it))
	{
		nlist<uint8> lst=type->get_flattend_types_list();
		if(0==uninit_vars_created_types.size())
			uninit_vars_created_local_stack_offset=
				static_cast<ProgramStructure_Function*>(it)->byte_addressed_local_variable_count;
		uninit_vars_created_types.insert(uninit_vars_created_types.end(), lst.begin(), lst.end());
		return static_cast<ProgramStructure_Function*>(it)->get_next_unused_local_var_address(type);
	}
	else
	{
		nlist<uint8> lst=type->get_flattend_types_list();
		if(0==uninit_vars_created_types.size())
			uninit_vars_created_local_stack_offset=
				static_cast<ProgramStructure_VoidFunction*>(it)->byte_addressed_local_variable_count;
		uninit_vars_created_types.insert(uninit_vars_created_types.end(), lst.begin(), lst.end());
		return static_cast<ProgramStructure_VoidFunction*>(it)->get_next_unused_local_var_address(type);
	}
}

VARIABLE_INDEX_TYPE ProgramStructure_IF_BASE::add_uninit_local_variable(const Type *type)
{
	ProgramStructureElement* it=GetActFunctionProgramStructure("Local variable defined outside of a function.");
	if(0==it || cc->ProgramStructureStack.size()<2 || cc->ProgramStructureStack.back()!=this)
		report_error("ProgramStructure_Block loop error");
	if(0!=dynamic_cast<ProgramStructure_Function*>(it))
	{
		nlist<uint8> lst=type->get_flattend_types_list();
		if(0==uninit_vars_created_types.size())
			uninit_vars_created_local_stack_offset=
				static_cast<ProgramStructure_Function*>(it)->byte_addressed_local_variable_count;
		uninit_vars_created_types.insert(uninit_vars_created_types.end(), lst.begin(), lst.end());
		return static_cast<ProgramStructure_Function*>(it)->get_next_unused_local_var_address(type);
	}
	else
	{
		nlist<uint8> lst=type->get_flattend_types_list();
		if(0==uninit_vars_created_types.size())
			uninit_vars_created_local_stack_offset=
				static_cast<ProgramStructure_VoidFunction*>(it)->byte_addressed_local_variable_count;
		uninit_vars_created_types.insert(uninit_vars_created_types.end(), lst.begin(), lst.end());
		return static_cast<ProgramStructure_VoidFunction*>(it)->get_next_unused_local_var_address(type);
	}
}

struct ParamPlusPlus;
struct ParamMinusMinus;

struct param
{
private:
	param(const param& v) { report_error("Constructor param::param must not be called"); } // Zugriff verhindern
	friend struct ParamPlusPlus;
	friend struct ParamMinusMinus;
public:
	VARIABLE_INDEX_TYPE nr;
	const Type *type;

	bool is_expr() const { return 0!=dynamic_cast<const BasicType*>(type); }
	bool is_ptr() const { return 0!=dynamic_cast<const PointerType*>(type); }

	param(const BasicType& _type)
	{
		if(Value::t_void==_type.type || Value::t_nil==_type.type || Value::t_unknown_type==_type.type)
			report_error("Function parameter cannot have the type void.");
		type=&_type;
		if(cc->ProgramStructureStack.size()!=2 || (0==dynamic_cast<ProgramStructure_VoidFunction*>(*(--(cc->ProgramStructureStack.end()))) &&
				0==dynamic_cast<ProgramStructure_Function*>(*(--(cc->ProgramStructureStack.end())))))
			report_error("param may only be defined within a function and in the lowest program structure level within the function.");
		if(0!=dynamic_cast<ProgramStructure_VoidFunction*>(*(--(cc->ProgramStructureStack.end()))))
		{
			ProgramStructure_VoidFunction *psfunc=static_cast<ProgramStructure_VoidFunction*>(*(--(cc->ProgramStructureStack.end())));
			VoidFunction &f=psfunc->func;
			if(++f.akt_arg>f.argtypes.size() || f.akt_arg_it==f.argtypes.end())
				report_error(printf_string("Function %s: Too many formal params defined (required: exactly %u)", f.functionname, f.argtypes.size()));
			nr=(VARIABLE_INDEX_TYPE)f.akt_arg_bytecount;
			f.akt_arg_bytecount+=type->size();
			if(!(*(f.akt_arg_it))->equal(type))
				report_error(printf_string("Function %s: The formal parameter number %u has been declared to have type\n   %s\nbut in the function definition it has type\n   %s.\n",
					f.functionname, f.akt_arg, (*(f.akt_arg_it))->get_type_string().c_str(), type->get_type_string().c_str()));
			++f.akt_arg_it;
		}
		else
		{
			ProgramStructure_Function *psfunc=static_cast<ProgramStructure_Function*>(*(--(cc->ProgramStructureStack.end())));
			Function &f=psfunc->func;
			if(++f.akt_arg>f.argtypes.size() || f.akt_arg_it==f.argtypes.end())
				report_error(printf_string("Function %s: Too many formal params defined (required: exactly %u)", f.functionname, f.argtypes.size()));
			nr=(VARIABLE_INDEX_TYPE)f.akt_arg_bytecount;
			f.akt_arg_bytecount+=type->size();
			if(!(*(f.akt_arg_it))->equal(type))
				report_error(printf_string("Function %s: The formal parameter number %u has been declared to have type\n   %s\nbut in the function definition it has type\n   %s.\n",
					f.functionname, f.akt_arg, (*(f.akt_arg_it))->get_type_string().c_str(), type->get_type_string().c_str()));
			++f.akt_arg_it;
		}
	}
	param(const PointerType& _type)
	{
		type=cc->RegisterType->Register(&_type); // Es darf NICHT type=&_type verwendet werden !
		if(cc->ProgramStructureStack.size()!=2 || (0==dynamic_cast<ProgramStructure_VoidFunction*>(*(--(cc->ProgramStructureStack.end()))) &&
				0==dynamic_cast<ProgramStructure_Function*>(*(--(cc->ProgramStructureStack.end())))))
			report_error("param may only be defined within a function and in the lowest program structure level within the function.");
		if(0!=dynamic_cast<ProgramStructure_VoidFunction*>(*(--(cc->ProgramStructureStack.end()))))
		{
			ProgramStructure_VoidFunction *psfunc=static_cast<ProgramStructure_VoidFunction*>(*(--(cc->ProgramStructureStack.end())));
			VoidFunction &f=psfunc->func;
			if(++f.akt_arg>f.argtypes.size() || f.akt_arg_it==f.argtypes.end())
				report_error(printf_string("Function %s: Too many formal params defined (required: exactly %u)", f.functionname, f.argtypes.size()));
			nr=(VARIABLE_INDEX_TYPE)f.akt_arg_bytecount;
			f.akt_arg_bytecount+=type->size();
			if(!(*(f.akt_arg_it))->equal(type))
				report_error(printf_string("Function %s: The formal parameter number %u has been declared to have type\n   %s\nbut in the function definition it has type\n   %s.\n",
					f.functionname, 1+f.akt_arg, (*(f.akt_arg_it))->get_type_string().c_str(), type->get_type_string().c_str()));
			++f.akt_arg_it;
		}
		else
		{
			ProgramStructure_Function *psfunc=static_cast<ProgramStructure_Function*>(*(--(cc->ProgramStructureStack.end())));
			Function &f=psfunc->func;
			if(++f.akt_arg>f.argtypes.size() || f.akt_arg_it==f.argtypes.end())
				report_error(printf_string("Function %s: Too many formal params defined (required: exactly %u)", f.functionname, f.argtypes.size()));
			nr=(VARIABLE_INDEX_TYPE)f.akt_arg_bytecount;
			f.akt_arg_bytecount+=type->size();
			if(!(*(f.akt_arg_it))->equal(type))
				report_error(printf_string("Function %s: The formal parameter number %u has been declared to have type\n   %s\nbut in the function definition it has type\n   %s.\n",
					f.functionname, 1+f.akt_arg, (*(f.akt_arg_it))->get_type_string().c_str(), type->get_type_string().c_str()));
			++f.akt_arg_it;
		}
	}
	operator expr() const
	{
		if(!is_expr()) // Ein Pointer knnte zwar in einen __BOOL umgewandelt werden. Hierfr ist jedoch die Umwandlung nach LogExpr vorgesehen.
			report_error("Cannot convert to an expression of elementary type.");
		expr ex(Command(bi::GETLOCAL).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
	}
	PointerExpr to_PointerExpr() const
	{
		if(!is_ptr())
			report_error(printf_string("Conversion to Pointer not possible. Cannot convert the type\n   %s\to a pointer type.",
				type->get_type_string().c_str()));
		return PointerExpr(*type, expression_yielding_address(), false, true); // PointerExpr::memory_with_fixed_type==true
	}
	expr expression_yielding_address() const
	{
		if(!type->is_ptr())
			report_error("expression_yielding_address() called for non-pointer.");
		expr ex(Command(bi::GETLOCAL).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		return ex.set_type(Value::VarIndexTypeNr);
	}
	GeneralExpr operator->*(const char *element_name) const
	{	// ->* : Dereferenzierung des Zeigers und Elementauswahl des sich ergebenden STRUCT-Objekts.
		if(0==dynamic_cast<const PointerType*>(type) ||
				0==dynamic_cast<const StructType*>(static_cast<const PointerType*>(type)->type_pointed_to.get()))
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to a variable of type\n   %s",
				element_name, type->get_type_string().c_str()));
		const Type *ptype=static_cast<const PointerType*>(type)->type_pointed_to;
		const StructType *stype=static_cast<const StructType*>(ptype);
		nlist<StructType::Elem>::const_iterator it=stype->find_elem(element_name);
		if(!it.is_valid_and_not_end(stype->elem))
			report_error(printf_string("operator->*: '%s' in not an element of %s", element_name, type->get_type_string().c_str()));

		expr ex(Command(bi::GETLOCAL).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		ex.add(expr::get_const_expr(stype->offsetofelem(element_name)));
		ex.addcmd(bi::ADD, -1);
		return GeneralExpr::from_PointerExpr(PointerExpr(PointerType(it->type, static_cast<const PointerType*>(type)->C_pointer),
			ex.set_type(Value::VarIndexTypeNr), true));
	}
	GeneralExpr operator[](const expr& ex) const
	{
		ex.set_exported();
		if(0!=dynamic_cast<const ArrayType*>(type))
		{
			expr ex2;
			ex2.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			return PointerExpr(PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr),
				false, true)[ex]; // PointerExpr::memory_with_fixed_type==true
		}
		else if(is_ptr())
			return PointerExpr(*type, expression_yielding_address().set_type(Value::VarIndexTypeNr), false, true)[ex];
		else
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or to pointers. Cannot apply [] to a variable of type\n   %s",
				type->get_type_string().c_str()));
	}
	GeneralExpr operator[](long long x) const
	{
		if(0!=dynamic_cast<const ArrayType*>(type))
		{
			if(x<0 || x>=static_cast<const ArrayType*>(type)->nr_elem)
				report_error(printf_string("Array index is out of range. Cannot access index %d of the array\n   %s", (int)x, type->get_type_string().c_str()));
			expr ex2;
			ex2.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(
					PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr), true, true));
			return PointerExpr(PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr),
				false, true)[x];
		}
		else if(is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply [] to a pointer to void.");
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(*type,
					expression_yielding_address().set_type(Value::VarIndexTypeNr), true, true));
			else
				return PointerExpr(*type, expression_yielding_address().set_type(Value::VarIndexTypeNr), false, true)[x];
		}
		else
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or to pointers. Cannot apply [] to a variable of type\n   %s",
				type->get_type_string().c_str()));
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		if(0!=dynamic_cast<const ArrayType*>(type))
		{
			if(x<0 || x>=static_cast<const ArrayType*>(type)->nr_elem)
				report_error(printf_string("Array index is out of range. Cannot access index %d of the array\n   %s", (int)x, type->get_type_string().c_str()));
			expr ex2;
			ex2.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(
					PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr), true, true));
			return PointerExpr(PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr),
				false, true)[x];
		}
		else if(is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply [] to a pointer to void.");
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(*type,
					expression_yielding_address().set_type(Value::VarIndexTypeNr), true, true));
			else
				return PointerExpr(*type, expression_yielding_address().set_type(Value::VarIndexTypeNr), false, true)[x];
		}
		else
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or to pointers. Cannot apply [] to a variable of type\n   %s",
				type->get_type_string().c_str()));
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }

	GeneralExpr operator=(const param& x) const
	{	// Es ist wichtig, diesen Zuweisungsoperator (mit Quelltyp (local) = Zieltyp (local)) zu definieren, weil er vom Standard-Zuweisungsoperator abweicht.
		return operator=(CollectGeneralType(x));
	}

	expr get_assignment_code(const expr& value) const
	{
		value.set_exported();

		uint8 type_nr=value.get_type();
		// Value braucht nicht auf den passenden Typ gecastet zu werden, weil in der an die Funktion bergebenen Variablen immer
		// bereits der richtige Typ steht und ein cast ggf. vom Kommando SETLOCAL durchgefhrt wird.
		if(0!=dynamic_cast<const BasicType*>(type))
			type_nr=static_cast<const BasicType*>(type)->type;
		if(0!=dynamic_cast<const PointerType*>(type))
			type_nr=Value::VarIndexTypeNr;

#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
		if(value.is_integer_constant())
		{
			expr ex;
			ex.add(Command(bi::SETLOCALC).add_coded_integer15(value.get_this_integer_constant(), 0, true).
					add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.set_code_has_effect();
			ex.set_type(type_nr);
			return ex;
		}
#endif
		expr ex(value.shortest_type_if_constant());
		ex.add(Command(bi::SETLOCAL).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(0));
		ex.set_code_has_effect();
		// Weil SETLOCAL und nicht SETLOCALT verwendet wird, wird im Falle eines Pointers
		// automatisch auf Value::VarIndexTypeNr gecastet.
		return ex.set_type(type_nr);
	}

	GeneralExpr operator=(const CollectGeneralType& value) const
	{
		value.set_exported();
		if(value.is_expr_type())
		{
			if(type->is_ptr())
			{
				if(value.is_basicconst && 0==value.cval) // Zuweisung des Nullzeigers
					return GeneralExpr::from_PointerExpr(PointerExpr(*type, get_assignment_code(expr((VARIABLE_INDEX_TYPE)0))));
				report_error(printf_string("Cannot assign an expression of type\n   %s\nto an expression of type\n   %s",
					value.get_type_string().c_str(), type->get_type_string().c_str()));
			}
			if(0!=dynamic_cast<const StructType*>(type))
				report_error("Cannot assign an expression of elementary type to a struct-object.");
			if(0!=dynamic_cast<const PointerType*>(type))
				report_error(printf_string("Cannot assign the expression\n   %s\nof elementary type not beeing the integer constant 0 to a pointer of type\n   %s",
					value.get_type_string().c_str(), type->get_type_string().c_str()));
			if(0==dynamic_cast<const BasicType*>(type))
				report_error("Internal compiler error: BasicType expected.");
			return GeneralExpr::from_expr(get_assignment_code(value.to_expr()));
		}
		else if(value.is_PointerExpr())
		{
			PointerExpr pe(value.to_PointerExpr());
			if(!type->equal(pe.get_actual_type()))
			{
				if(type->equal(&__BOOL) && 0!=dynamic_cast<const PointerType*>(value.get_type()))
				{
					// Konversion und Zuweisung eines Zeigers an eine bool-Variable
					expr ex;
					ex.add(pe.expression_yielding_address());
					ex.add_effective(Command(bi::CAST).addnibble(Value::t_bool));
					ex.set_type(Value::t_bool);
					return GeneralExpr::from_expr(get_assignment_code(ex));
				}
				if(type->is_ptr() && value.is_basicconst && 0==value.cval) // Zuweisung des Nullzeigers
					return GeneralExpr::from_PointerExpr(PointerExpr(*type, get_assignment_code(expr((VARIABLE_INDEX_TYPE)0))));
				if(type->is_ptr() && static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID) && pe.get_actual_type()->is_ptr())
				{
					// Im Folgenden wird ein Zeiger an einen Zeiger auf __VOID zugewiesen.
					return GeneralExpr::from_PointerExpr(PointerExpr(*type, get_assignment_code(pe.expression_yielding_address())));
				}
				report_error(printf_string("Type mismatch in assignment to the object the pointer points to. Cannot assign an expression of type\n   %s\nto an expression of type\n   %s",
					pe.get_actual_type()->get_type_string().c_str(), type->get_type_string().c_str()));
			}
			if(pe.is_dref() && 0!=dynamic_cast<const BasicType*>(pe.get_pt().type_pointed_to.get()))
			{
				if(!type->equal(pe.get_pt().type_pointed_to))
					error_conversion_not_possible(pe.get_actual_type(), type, "operator=(): ");
				// Zuweisung eines expr an einen dereferenzierten Zeiger auf ein expr
				return GeneralExpr::from_expr(operator=(pe.to_expr())); // Rekursiver Aufruf von operator=() mit einem expr-Parameter
			}
			if(0!=dynamic_cast<const StructType*>(pe.get_actual_type()) || 0!=dynamic_cast<const ArrayType*>(pe.get_actual_type()))
				report_error(printf_string("Cannot assign objects of type STRUCT or ARRAY to variables or dereferenced pointers; please implement the assignment by assigning element-wise. Cannot assign an expression of type\n   %s\nto a variable.",
					pe.get_actual_type()->get_type_string().c_str()));
			if(0==dynamic_cast<const PointerType*>(pe.get_actual_type()))
				error_conversion_not_possible(pe.get_actual_type(), type, "\'=\': ");
			// Im Folgenden wird ein Zeiger an einen Zeiger zugewiesen.
			return GeneralExpr::from_PointerExpr(PointerExpr(*type, get_assignment_code(pe.expression_yielding_address()),
				false, pe.is_actually_memory_with_fixed_type_info()));
		}
		else
			report_error("Unknown type of CollectGeneralType");
	}

	PointerExpr operator&()
	{
		expr ex(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		return PointerExpr(PointerType(type), ex.set_type(Value::VarIndexTypeNr), false, true);
	}
	PointerExpr operator*()
	{
		if(0==dynamic_cast<const PointerType*>(type))
			report_error("The dereferentiation operator * can only be applied to expressions of type pointer.");
		if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
			report_error("Cannot dereferentiate a pointer to void.");
		expr ex(Command(bi::GETLOCAL).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		return PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr), true);
	}
	GeneralExpr operator+=(const CollectGeneralType& value)
	{
		value.set_exported();
		if(value.is_PointerExpr() && !value.is_expr_type()) // Ein PointerExpr kann auch einen BasicType ergeben (falls dref=true).
			report_error(printf_string("operator+=() : Cannot apply a pointer on the right hand side. Cannot apply the type\n   %s\non the right hand side.",
				value.get_type_string().c_str()));
		if(!value.is_expr_type()) // nicht value.is_expr() !
			report_error("Unknown Type in CollectGeneralType()");
		if(!type->is_expr() && !type->is_ptr())
			report_error(printf_string("Cannot add to an object of type ARRAY or STRUCT. Cannot add an expression of type\n   %s",
				value.get_type_string().c_str()));
		expr s;
		if(type->is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply += to a pointer to void.");
			expr v2(value.to_expr().shortest_type_if_constant());
			// v2 muss nicht ggf. auf Value::VarIndexTypeNr gecastet werden, weil Voraussetzung zur Anwendung des operator+=() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
			if(1==static_cast<const PointerType*>(type)->type_pointed_to->size())
				s.assign(v2);
			else if(v2.get_type()<Value::SignedVarIndexTypeNr)
				s.assign(v2*expr((SIGNED_VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(type)->type_pointed_to->size()));
			else
				s.assign(v2*expr::get_const_expr((int64)static_cast<const PointerType*>(type)->type_pointed_to->size(), true));
				// get_const_expr() muss einen signed Typ erzeugen, denn value knnte negativ sein.

			expr ex;
			int64 c=s.is_integer_constant() ? s.get_this_integer_constant() : 0; // get_this_integer_constant() darf nur aufgerufen werden, wenn tatschlich Konstante
			if(s.is_integer_constant() && ((s.get_type()==Value::t_u8 && c<=8) || (s.get_type()==Value::t_i8 && (c<=8 && c>=-8))))
			{
				if(0==c)
					return GeneralExpr::from_PointerExpr(to_PointerExpr());
				ex.add(Command(bi::ADDCLOCAL).addnibble(c>=0 ? (c-1) : (7-c))
					.add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.set_code_has_effect();
			}
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(s.shortest_type_if_constant());
				ex.addcmd(bi::ADDM, -1);
				ex.set_code_has_effect();
			}
			return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr), false, true));
		}
		else // type->is_expr()==true
		{
			expr s(value.to_expr().shortest_type_if_constant());
			expr ex;
			int64 c=s.is_integer_constant() ? s.get_this_integer_constant() : 0; // get_this_integer_constant() darf nur aufgerufen werden, wenn tatschlich Konstante
			if(s.is_integer_constant() && ((s.get_type()==Value::t_u8 && c<=8) || (s.get_type()==Value::t_i8 && (c<=8 && c>=-8))))
			{
				if(0==c)
					return GeneralExpr::from_PointerExpr(to_PointerExpr());
				ex.add(Command(bi::ADDCLOCAL).addnibble(c>=0 ? (c-1) : (7-c))
					.add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.set_code_has_effect();
			}
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(s.shortest_type_if_constant());
				ex.addcmd(bi::ADDM, -1);
				ex.set_code_has_effect();
			}
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
	}
	GeneralExpr operator++()
	{
		if(!type->is_expr() && !type->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
		if(type->is_ptr() && static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
			report_error("Cannot apply ++ to a pointer to void.");
		if(type->is_ptr())
		{
			expr ex;
			if(static_cast<const PointerType*>(type)->type_pointed_to->size()<=8)
			{
				ex.add(Command(bi::ADDCLOCAL).addnibble(static_cast<const PointerType*>(type)->type_pointed_to->size()-1)
					.add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.set_code_has_effect();
			}
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(expr::get_const_expr((VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(type)->type_pointed_to->size()));
				ex.addcmd(bi::ADDM, -1);
				ex.set_code_has_effect();
			}
			return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr), false, true));
		}
		else // type->is_expr()==true
		{
			// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator++() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
			expr ex;
			ex.add(Command(bi::ADDCLOCAL).addnibble(0).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.set_code_has_effect();
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
	}
	ParamPlusPlus operator++(int);

	GeneralExpr operator-=(const CollectGeneralType& value)
	{
		value.set_exported();
		if(value.is_PointerExpr() && !value.is_expr_type()) // Ein PointerExpr kann auch einen BasicType ergeben (falls dref=true).
			report_error(printf_string("operator-=() : Cannot apply a pointer on the right hand side. Cannot apply the type\n   %s\non the right hand side.",
				value.get_type_string().c_str()));
		if(!value.is_expr_type()) // nicht value.is_expr() !
			report_error("Unknown Type in CollectGeneralType()");
		if(!type->is_expr() && !type->is_ptr())
			report_error(printf_string("Cannot sub from an object of type ARRAY or STRUCT. Cannot add an expression of type\n   %s",
				value.get_type_string().c_str()));
		expr s;
		if(type->is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply += to a pointer to void.");
			expr v2(value.to_expr().shortest_type_if_constant());
			// v2 muss nicht ggf. auf Value::VarIndexTypeNr gecastet werden, weil Voraussetzung zur Anwendung des operator+=() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
			if(1==static_cast<const PointerType*>(type)->type_pointed_to->size())
				s.assign(v2);
			else if(v2.get_type()<Value::SignedVarIndexTypeNr)
				s.assign(v2*expr((SIGNED_VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(type)->type_pointed_to->size()));
			else
				s.assign(v2*expr::get_const_expr((int64)static_cast<const PointerType*>(type)->type_pointed_to->size(), true));
				// get_const_expr() muss einen signed Typ erzeugen, denn value knnte negativ sein.

			expr ex;
			int64 c=s.is_integer_constant() ? s.get_this_integer_constant() : 0; // get_this_integer_constant() darf nur aufgerufen werden, wenn tatschlich Konstante
			if(s.is_integer_constant() && ((s.get_type()==Value::t_u8 && c<=8) || (s.get_type()==Value::t_i8 && (c<=8 && c>=-8))))
			{
				if(0==c)
					return GeneralExpr::from_PointerExpr(to_PointerExpr());
				ex.add(Command(bi::ADDCLOCAL).addnibble(c>=0 ? (7+c) : (-c-1))
					.add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.set_code_has_effect();
			}
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(s.shortest_type_if_constant());
				ex.addcmd(bi::SUBM, -1);
				ex.set_code_has_effect();
			}
			return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr), false, true));
		}
		else // type->is_expr()==true
		{
			expr s(value.to_expr().shortest_type_if_constant());
			expr ex;
			int64 c=s.is_integer_constant() ? s.get_this_integer_constant() : 0; // get_this_integer_constant() darf nur aufgerufen werden, wenn tatschlich Konstante
			if(s.is_integer_constant() && ((s.get_type()==Value::t_u8 && c<=8) || (s.get_type()==Value::t_i8 && (c<=8 && c>=-8))))
			{
				if(0==c)
					return GeneralExpr::from_PointerExpr(to_PointerExpr());
				ex.add(Command(bi::ADDCLOCAL).addnibble(c>=0 ? (7+c) : (-c-1))
					.add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.set_code_has_effect();
			}
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(s.shortest_type_if_constant());
				ex.addcmd(bi::SUBM, -1);
				ex.set_code_has_effect();
			}
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
	}
	GeneralExpr operator--()
	{
		if(!type->is_expr() && !type->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
		if(type->is_ptr() && static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
			report_error("Cannot apply ++ to a pointer to void.");
		if(type->is_ptr())
		{
			expr ex;
			if(static_cast<const PointerType*>(type)->type_pointed_to->size()<=8)
			{
				ex.add(Command(bi::ADDCLOCAL).addnibble(7+static_cast<const PointerType*>(type)->type_pointed_to->size())
					.add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.set_code_has_effect();
			}
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(expr::get_const_expr((VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(type)->type_pointed_to->size()));
				ex.addcmd(bi::SUBM, -1);
				ex.set_code_has_effect();
			}
			return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr), false, true));
		}
		else // type->is_expr()==true
		{
			// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator++() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
			expr ex;
			ex.add(Command(bi::ADDCLOCAL).addnibble(8).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.set_code_has_effect();
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
	}
	ParamMinusMinus operator--(int);

	expr operator*=(const expr& x)
	{
		if(!is_expr())
			report_error(printf_string("Cannot apply operator*=() to a function parameter with the non-elementary type\n   %s",
			type->get_type_string().c_str()));
		x.set_exported();
		expr ex;
		ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		ex.add(x.shortest_type_if_const_keep_signed());
		ex.addcmd(bi::MULM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator/=(const expr& x)
	{
		if(!is_expr())
			report_error(printf_string("Cannot apply operator/=() to a function parameter with the non-elementary type\n   %s",
			type->get_type_string().c_str()));
		x.set_exported();
#ifndef WITH_INTEGER64_DIVISION
		if(static_cast<const BasicType*>(type)->type==Value::t_u64 || static_cast<const BasicType*>(type)->type==Value::t_i64)
			report_error("64 Bit division is deactivated in the bytecode interpreter. To activate it, #define WITH_INTEGER64_DIVISION");
#endif
		expr ex;
		ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		ex.add(x.shortest_type_if_const_keep_signed());
		ex.addcmd(bi::DIVM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator%=(const expr& x)
	{
		if(!is_expr())
			report_error(printf_string("Cannot apply operator%=() to a function parameter with the non-elementary type\n   %s",
			type->get_type_string().c_str()));
		x.set_exported();
#ifndef WITH_INTEGER64_MODULO
		if(static_cast<const BasicType*>(type)->type==Value::t_u64 || static_cast<const BasicType*>(type)->type==Value::t_i64)
			report_error("64 Bit modulo is deactivated in the bytecode interpreter. To activate it, #define WITH_INTEGER64_MODULO");
#endif
		expr ex;
		ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		ex.add(x.shortest_type_if_const_keep_signed());
		ex.addcmd(bi::MODM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator|=(const expr& x)
	{
		if(!is_expr())
			report_error(printf_string("Cannot apply operator|=() to a function parameter with the non-elementary type\n   %s",
			type->get_type_string().c_str()));
		x.set_exported();
		expr ex;
		ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::ORM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator&=(const expr& x)
	{
		if(!is_expr())
			report_error(printf_string("Cannot apply operator&=() to a function parameter with the non-elementary type\n   %s",
			type->get_type_string().c_str()));
		x.set_exported();
		expr ex;
		ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::ANDM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator^=(const expr& x)
	{
		if(!is_expr())
			report_error(printf_string("Cannot apply operator^=() to a function parameter with the non-elementary type\n   %s",
			type->get_type_string().c_str()));
		x.set_exported();
		expr ex;
		ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::XORM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator<<=(const expr& x)
	{
		if(!is_expr())
			report_error(printf_string("Cannot apply operator<<=() to a function parameter with the non-elementary type\n   %s",
			type->get_type_string().c_str()));
		x.set_exported();
		expr ex;
		ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::SLM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}

	expr operator>>=(const expr& x)
	{
		if(!is_expr())
			report_error(printf_string("Cannot apply operator>>=() to a function parameter with the non-elementary type\n   %s",
			type->get_type_string().c_str()));
		x.set_exported();
		expr ex;
		ex.add(Command(bi::LOCALADDR).add_coded_integer255(nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		ex.add(x.shortest_type_if_constant());
		ex.addcmd(bi::SRM, -1);
		ex.set_code_has_effect();
		return ex.set_type(static_cast<const BasicType*>(type)->type);
	}
};

struct ParamPlusPlus
{
private: void operator=(const ParamPlusPlus& lpp) { report_error("Assignment ParamPlusPlus::operator=() must not be called"); } // Zugriff verhindern
public:
	const param& v; // Hier muss eine Referenz verwendet werden, da ansonsten im Konstruktor LocalMinusMinus::LocalMinusMinus()
		// der Konstruktor local::local aufgerufen wird, welcher nicht lediglich den Speicher der Struktur local kopiert, sondern
		// annimmt, der Benutzer htte eine neue local angelegt und somit entsprechend Speicher auf dem Stack reserviert.
	mutable bool exported;
	const ParamPlusPlus& set_exported() const { exported=true; return *this; }
	ParamPlusPlus(const param& _v) : v(_v) { exported=false; }
	~ParamPlusPlus()
	{
		if(!exported)
		{	// Der folgende Code ist identisch wie beim Prfix Operator -- und auch effizienter als der Code des Postfix Operators --,
			// kann jedoch nur verwendet werden, wenn der Wert des Ausdrucks nicht weiter verwendet wird.
			if(!v.type->is_expr() && !v.type->is_ptr())
				report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");
			int64 d;
			if(v.type->is_ptr())
			{
				if(static_cast<const PointerType*>(v.type)->type_pointed_to->equal(&__VOID))
					report_error("Cannot apply ++ to a pointer to void.");
				d=static_cast<const PointerType*>(v.type)->type_pointed_to->size();
			}
			else // type->is_expr()==true
				d=1;
			// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator--() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.

			expr ex;
			if(d<=8)
				ex.add(Command(bi::ADDCLOCALPOP).addnibble(d-1).add_coded_integer255(v.nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(0));
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(v.nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::ADDMPOP, -2);
			}
			ex.set_code_has_effect();
			Statement stm(ex); // Erzeugung von Code beim Aufruf des Destruktors von stm
		}
	}
	GeneralExpr to_GeneralExpr() const
	{
		set_exported();
		if(!v.type->is_expr() && !v.type->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");

		int64 d;
		if(v.type->is_ptr())
		{
			if(static_cast<const PointerType*>(v.type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply ++ to a pointer to void.");
			d=static_cast<const PointerType*>(v.type)->type_pointed_to->size();
		}
		else // type->is_expr()==true
			d=1;

		expr ex;
		if(d<=8) // d==0 ist unzulssig
			ex.add(Command(bi::LOCALADDC).addnibble(d-1).add_coded_integer255(v.nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		else
		{
			ex.add(Command(bi::LOCALADDR).add_coded_integer255(v.nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.add(expr::get_const_expr(d));
			ex.addcmd(bi::ADDM, -1);
			ex.add(expr::get_const_expr(d));
			ex.addcmd(bi::SUB, -1);
		}
		ex.set_exported();
		ex.set_code_has_effect();
		if(v.type->is_ptr())
			return GeneralExpr::from_PointerExpr(PointerExpr(*v.type, ex.set_type(Value::VarIndexTypeNr)));
		else if(v.type->is_expr())
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(v.type)->type));
		else
			report_error("ParamPlusPlus Error");
	}
	operator expr() const
		{ return to_GeneralExpr().to_expr(); }
	PointerExpr to_PointerExpr() const
		{ return to_GeneralExpr().to_PointerExpr(); }
	GeneralExpr operator->*(const char *element_name) const
	{	// ->* : Dereferenzierung des Zeigers und Elementauswahl des sich ergebenden STRUCT-Objekts.
		set_exported();
		if(0==dynamic_cast<const PointerType*>(v.type) ||
				0==dynamic_cast<const StructType*>(static_cast<const PointerType*>(v.type)->type_pointed_to.get()))
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to a variable of type\n   %s",
				element_name, v.type->get_type_string().c_str()));
		return (to_PointerExpr())->*element_name;
	}
	GeneralExpr operator[](const expr& ex) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[ex];
	}
	GeneralExpr operator[](long long x) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }
};

struct ParamMinusMinus
{
private: void operator=(const ParamMinusMinus& lpp) { report_error("Assignment ParamMinusMinus::operator=() must not be called"); } // Zugriff verhindern
public:
	const param& v; // Hier muss eine Referenz verwendet werden, da ansonsten im Konstruktor LocalMinusMinus::LocalMinusMinus()
		// der Konstruktor local::local aufgerufen wird, welcher nicht lediglich den Speicher der Struktur local kopiert, sondern
		// annimmt, der Benutzer htte eine neue local angelegt und somit entsprechend Speicher auf dem Stack reserviert.
	mutable bool exported;
	const ParamMinusMinus& set_exported() const { exported=true; return *this; }
	ParamMinusMinus(const param& _v) : v(_v) { exported=false; }
	~ParamMinusMinus()
	{
		if(!exported)
		{	// Der folgende Code ist identisch wie beim Prfix Operator -- und auch effizienter als der Code des Postfix Operators --,
			// kann jedoch nur verwendet werden, wenn der Wert des Ausdrucks nicht weiter verwendet wird.
			if(!v.type->is_expr() && !v.type->is_ptr())
				report_error("operator--(): Cannot increment an object of type ARRAY or STRUCT.");
			expr s;
			if(v.type->is_ptr())
			{
				if(static_cast<const PointerType*>(v.type)->type_pointed_to->equal(&__VOID))
					report_error("Cannot apply -- to a pointer to void.");
				s.assign(expr::get_const_expr((VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(v.type)->type_pointed_to->size()));
			}
			else // type->is_expr()==true
				s.assign(expr::get_const_expr((uint32)1));
			// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator--() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.

			expr ex;
			ex.add(Command(bi::LOCALADDR).add_coded_integer255(v.nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.add(s);
			ex.addcmd(bi::SUBM, -1);
			ex.set_exported();
			ex.set_code_has_effect();
			if(v.type->is_ptr())
				ExprObjectDestructionHandler<expr>(ex.set_type(Value::VarIndexTypeNr));
			else if(v.type->is_expr())
				ExprObjectDestructionHandler<expr>(ex.set_type(static_cast<const BasicType*>(v.type)->type));
		}
	}
	GeneralExpr to_GeneralExpr() const
	{
		set_exported();
		if(!v.type->is_expr() && !v.type->is_ptr())
			report_error("operator--(): Cannot increment an object of type ARRAY or STRUCT.");
		expr s;
		if(v.type->is_ptr())
		{
			if(static_cast<const PointerType*>(v.type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply -- to a pointer to void.");
			s.assign(expr::get_const_expr(static_cast<const PointerType*>(v.type)->type_pointed_to->size()));
		}
		else // type->is_expr()==true
			s.assign(expr::get_const_expr((uint32)1));

		expr ex;
		ex.add(Command(bi::LOCALADDR).add_coded_integer255(v.nr, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		ex.add(s);
		ex.addcmd(bi::SUBM, -1);
		ex.add(s);
		ex.addcmd(bi::ADD, -1);
		ex.set_exported();
		ex.set_code_has_effect();
		if(v.type->is_ptr())
			return GeneralExpr::from_PointerExpr(PointerExpr(*v.type, ex.set_type(Value::VarIndexTypeNr), false, true));
		else if(v.type->is_expr())
			return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(v.type)->type));
		else
			report_error("ParamMinusMinus Error");
	}
	operator expr() const
		{ return to_GeneralExpr().to_expr(); }
	PointerExpr to_PointerExpr() const
		{ return to_GeneralExpr().to_PointerExpr(); }
	GeneralExpr operator->*(const char *element_name) const
	{	// ->* : Dereferenzierung des Zeigers und Elementauswahl des sich ergebenden STRUCT-Objekts.
		set_exported();
		if(0==dynamic_cast<const PointerType*>(v.type) ||
				0==dynamic_cast<const StructType*>(static_cast<const PointerType*>(v.type)->type_pointed_to.get()))
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to a variable of type\n   %s",
				element_name, v.type->get_type_string().c_str()));
		return (to_PointerExpr())->*element_name;
	}
	GeneralExpr operator[](const expr& ex) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[ex];
	}
	GeneralExpr operator[](long long x) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }
};

ParamPlusPlus param::operator++(int)
{
	return ParamPlusPlus(*this);
}

ParamMinusMinus param::operator--(int)
{
	return ParamMinusMinus(*this);
}

#define BYTECODE_TARGETED_COMPILER__DECLARE_VOID_FUNCTION(name, parameters) ;BytecodeTargetedCompiler::VoidFunction name(#name, parameters, false);
#define BYTECODE_TARGETED_COMPILER__void_function(name) ;BytecodeTargetedCompiler::VoidFunctionDefinitionHandler(name); {
#define BYTECODE_TARGETED_COMPILER__DECLARE_EXTERN_VOID_FUNCTION(name, parameters) ;BytecodeTargetedCompiler::VoidFunction name(#name, parameters, true, extern_function_names::name);

void VoidFunctionDefinitionHandler(VoidFunction &func)
{
	if(func.extern_function)
		report_error(printf_string("The function %s is declared extern and therefore no code can be defined for this function.",
			func.functionname));
	if(cc->ProgramStructureStack.size()!=1 || 0==dynamic_cast<ProgramStructureGlobalScope*>(*(--(cc->ProgramStructureStack.end()))))
		report_error("VoidFunction defined within a function or out of the global scope (between BEGIN_GLOBAL_SCOPE and END_GLOBAL_SCOPE)");
	if(func.is_already_defined)
		report_error(printf_string("The function %s is already defined", func.functionname));
	cc->defined_goto_labels_of_actual_function.clear();
	cc->forward_goto_targets_of_actual_function.clear();
	cc->ProgramStructureStack.push_back(new ProgramStructure_VoidFunction(func));
	func.is_already_defined=true;
}

void RETURN(const CollectGeneralType& ex)
{
	ex.set_exported();
	if(cc->ProgramStructureStack.size()<2 || (0==dynamic_cast<ProgramStructure_Function*>(*(++(cc->ProgramStructureStack.begin()))) &&
			0==dynamic_cast<ProgramStructure_VoidFunction*>(*(++(cc->ProgramStructureStack.begin())))))
		report_error("RETURN occured outside of function");
	if(0==dynamic_cast<ProgramStructure_Function*>(*(++(cc->ProgramStructureStack.begin()))))
		report_error(printf_string("The void_function \'%s\' returns a value",
			static_cast<ProgramStructure_VoidFunction*>(*(++(cc->ProgramStructureStack.begin())))->func.functionname));
	ProgramStructure_Function *psfunc=static_cast<ProgramStructure_Function*>(*(++(cc->ProgramStructureStack.begin())));
	Function &f=psfunc->func;
	f.Return(ex); // Hier wird Code erzeugt. Dieser Code wird aber nicht
		// unbedingt in *psfunc eingefgt (z.B. wenn innerhalb einer IF-Abfrage ein RETURN erfolgt).
}

void RETURN()
{
	if(cc->ProgramStructureStack.size()<2 || (0==dynamic_cast<ProgramStructure_Function*>(*(++(cc->ProgramStructureStack.begin()))) &&
			0==dynamic_cast<ProgramStructure_VoidFunction*>(*(++(cc->ProgramStructureStack.begin())))))
		report_error("RETURN occured outside of function");
	if(0==dynamic_cast<ProgramStructure_VoidFunction*>(*(++(cc->ProgramStructureStack.begin()))))
		report_error(printf_string("The Non-void function \'%s\' must return a value.",
			static_cast<ProgramStructure_Function*>(*(++(cc->ProgramStructureStack.begin())))->func.functionname));
	ProgramStructure_VoidFunction *psfunc=static_cast<ProgramStructure_VoidFunction*>(*(++(cc->ProgramStructureStack.begin())));
	VoidFunction &f=psfunc->func;
	f.Return(); // Beim hier wird Code erzeugt. Dieser Code wird aber nicht
		// unbedingt in *psfunc eingefgt (z.B. wenn innerhalb einer IF-Abfrage ein RETURN erfolgt).
}

///////////////////////////////////////////////////////////////////////////////////////////

struct LocalPlusPlus;
struct LocalMinusMinus;

ProgramStructureElement *GetActFunctionProgramStructure(char *error_message_if_outside_of_function/*=0*/)
{
	if(cc->ProgramStructureStack.size()<2 || 0==dynamic_cast<ProgramStructureGlobalScope*>(cc->ProgramStructureStack.front()) ||
			(0==dynamic_cast<ProgramStructure_Function*>(*(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin()))) &&
			0==dynamic_cast<ProgramStructure_VoidFunction*>(*(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin())))))
		report_error(error_message_if_outside_of_function);
	return *(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin()));
}

struct local
{
private:
	void construct(uint8 type_nr, const expr& init_value)
	{
		init_value.set_exported();
		if(Value::t_void==type_nr || Value::t_nil==type_nr)
			report_error("local variable cannot have the type void.");
		if(Value::t_unknown_type==type_nr)
			report_error("The compiler could not deduce the type of the local variable to be created.");
		type=&BasicType::get_BasicType_from_type_nr(type_nr);
		if(0==cc->ProgramStructureStack.size())
			report_error("local variable defined outside of function");
		Command cmd=Command::new_uid_label();
		cmd.sp=0;
		expr ex(init_value);
		if(type_nr!=init_value.get_type())
			ex.add(Code(Command(bi::CAST).addnibble(type_nr)));
		cc->ProgramStructureStack.back()->add_local_variable(cmd, ex);
		uid=cmd.get_uid();
		var_mem_initialized=false;
		created_at_psstack_deep=cc->ProgramStructureStack.size();
	}
public:
	Command::uid_type uid;
	VARIABLE_INDEX_TYPE local_stack_offset;
	const Type *type; // Wenn uid==0 ist, MUSS type!=0 sein.
	long created_at_psstack_deep;
	mutable bool var_mem_initialized;

	bool is_expr() const { return 0!=dynamic_cast<const BasicType*>(type); }
	bool is_ptr() const { return 0!=dynamic_cast<const PointerType*>(type); }

	local(const BasicType& _type, bool init_value) { construct(_type.type, _type.convert_to_expr_constant_of_this_type(init_value)); }
	local(const BasicType& _type, unsigned char init_value) { construct(_type.type, _type.convert_to_expr_constant_of_this_type(init_value)); }
	local(const BasicType& _type, signed char init_value) { construct(_type.type, _type.convert_to_expr_constant_of_this_type(init_value)); }
	local(const BasicType& _type, unsigned short init_value) { construct(_type.type, _type.convert_to_expr_constant_of_this_type(init_value)); }
	local(const BasicType& _type, short init_value) { construct(_type.type, _type.convert_to_expr_constant_of_this_type(init_value)); }
	local(const BasicType& _type, unsigned int init_value) { construct(_type.type, _type.convert_to_expr_constant_of_this_type(init_value)); }
	local(const BasicType& _type, int init_value) { construct(_type.type, _type.convert_to_expr_constant_of_this_type(init_value)); }
	local(const BasicType& _type, unsigned long init_value) { construct(_type.type, _type.convert_to_expr_constant_of_this_type(init_value)); }
	local(const BasicType& _type, long init_value) { construct(_type.type, _type.convert_to_expr_constant_of_this_type(init_value)); }
	local(const BasicType& _type, unsigned long long init_value) { construct(_type.type, _type.convert_to_expr_constant_of_this_type(init_value)); }
	local(const BasicType& _type, long long init_value) { construct(_type.type, _type.convert_to_expr_constant_of_this_type(init_value)); }
#ifndef real_type_is_integer_type
	local(const BasicType& _type, real_type init_value) { construct(_type.type, _type.convert_to_expr_constant_of_this_type(init_value)); }
#endif

	local(const BasicType& _type, const CollectGeneralType& init_value)
	{
		init_value.set_exported();
		if(0==cc->ProgramStructureStack.size())
			report_error("local variable defined outside of function");
		if(!init_value.is_expr_type())
		{
			if(_type.type==Value::t_bool && init_value.is_PointerExpr() && init_value.to_PointerExpr().get_actual_type()->is_ptr())
				construct(_type.type, init_value.to_PointerExpr().expression_yielding_address());
			else
				report_error(printf_string("Type mismatch initializing a local variable. Cannot initialize a variable of type\n   %s\nwith an expression of type\n   %s",
					_type.get_type_string().c_str(), init_value.get_type_string().c_str()));
		}
		else
			construct(_type.type, init_value.to_expr());
		var_mem_initialized=false;
		created_at_psstack_deep=cc->ProgramStructureStack.size();
	}
	local(const local& v)
	{
		type=v.type;
		if(0==cc->ProgramStructureStack.size())
			report_error("local variable defined outside of function");
		Command cmd=Command::new_uid_label();
		cmd.sp=0;
		cc->ProgramStructureStack.back()->add_local_variable(cmd, expr(v));
		uid=cmd.get_uid();
		var_mem_initialized=false;
		created_at_psstack_deep=cc->ProgramStructureStack.size();
	}
	local(const BasicType& _type)
	{
		if(Value::t_void==_type.type || Value::t_nil==_type.type || Value::t_unknown_type==_type.type)
			report_error("local variable cannot have the type void.");
		type=&_type;
		uid=0;
		local_stack_offset=cc->ProgramStructureStack.back()->add_uninit_local_variable(type);
		var_mem_initialized=false;
		created_at_psstack_deep=cc->ProgramStructureStack.size();
	}
	local(const CollectGeneralType& init_value)
	{
		init_value.set_exported();
		if(0==cc->ProgramStructureStack.size())
			report_error("local variable defined outside of function");
		if(init_value.is_expr_type())
		{
			expr ex(init_value.to_expr());
			if(Value::t_nil==ex.get_type())
				report_error("Cannot create a local variable of type t_nil.");
			if(Value::t_unknown_type==ex.get_type())
				report_error("The compiler could not deduce the type of the local variable to be created.");
			type=&BasicType::get_BasicType_from_type_nr(ex.get_type());
			Command cmd=Command::new_uid_label();
			cmd.sp=0;
			cc->ProgramStructureStack.back()->add_local_variable(cmd, ex);
			uid=cmd.get_uid();
		}
		else if(init_value.is_PointerExpr())
		{
			type=init_value.to_PointerExpr().get_actual_type();
			Command cmd=Command::new_uid_label();
			cmd.sp=0;
			cc->ProgramStructureStack.back()->add_local_variable(cmd, init_value.to_PointerExpr().expression_yielding_address());
			uid=cmd.get_uid();
		}
		else
			report_error("Unknown type in CollectGeneralType");
		var_mem_initialized=false;
		created_at_psstack_deep=cc->ProgramStructureStack.size();
	}
	local(const PointerType &pt)
	{
		pt.chk_validity_of_ref_types_variable_creation();
		type=static_cast<const PointerType*>(cc->RegisterType->Register(&pt).get());
		uid=0;
		local_stack_offset=cc->ProgramStructureStack.back()->add_uninit_local_variable(type);
		var_mem_initialized=false;
		created_at_psstack_deep=cc->ProgramStructureStack.size();
	}
	local(const PointerType &pt, const CollectGeneralType& init_value)
	{	// Es gengt nicht, fr den Initialisierungsparameter nur den Typ const PointerExpr& zuzulassen. Sonst wrde z.B.
		// ein param oder ein var (struct Variable) nicht fr die Initialisierung akzeptiert, weil diese keine implizite Konvertierung
		// nach PointerExpr besitzen.
		init_value.set_exported();
		pt.chk_validity_of_ref_types_variable_creation();
		if(0==cc->ProgramStructureStack.size())
			report_error("local variable defined outside of function");
		if(!pt.equal(init_value.get_type()) && !(init_value.is_basicconst && 0==init_value.cval) &&
				!(pt.type_pointed_to->equal(&__VOID) && init_value.is_PointerExpr() && init_value.to_PointerExpr().get_actual_type()->is_ptr()))
			report_error(printf_string("Type mismatch initializing a pointer variable. Cannot initialize a variable of type\n   %s\nwith an expression of type\n   %s",
				pt.get_type_string().c_str(), init_value.get_type_string().c_str()));
		type=static_cast<const PointerType*>(cc->RegisterType->Register(&pt).get());
		Command cmd=Command::new_uid_label();
		cmd.sp=0;
		expr ex(init_value.to_PointerExpr(&pt).expression_yielding_address());
		if(Value::VarIndexTypeNr!=ex.get_type())
			ex.add_effective(Command(bi::CAST).addnibble(Value::VarIndexTypeNr));
		cc->ProgramStructureStack.back()->add_local_variable(cmd, ex.set_type(Value::VarIndexTypeNr));
		uid=cmd.get_uid();
		var_mem_initialized=false;
		created_at_psstack_deep=cc->ProgramStructureStack.size();
	}
	local(const ArrayType &at)
	{	// ArrayType::ArrayType() stellt bereits sicher, dass es sich nicht um ein Array von __VOID-Objekten handelt.
		at.chk_validity_of_ref_types_variable_creation();
		type=cc->RegisterType->Register(&at);
		uid=0;
		local_stack_offset=cc->ProgramStructureStack.back()->add_uninit_local_variable(type);
		var_mem_initialized=false;
		created_at_psstack_deep=cc->ProgramStructureStack.size();
	}
	local(const StructType &st)
	{
		st.chk_validity_of_ref_types_variable_creation();
		type=cc->RegisterType->Register(&st);
		uid=0;
		local_stack_offset=cc->ProgramStructureStack.back()->add_uninit_local_variable(type);
		var_mem_initialized=false;
		created_at_psstack_deep=cc->ProgramStructureStack.size();
	}
	GeneralExpr operator[](const char *element_name) const
	{	// [] : Elementauswahl eines STRUCTs.
		if(0==dynamic_cast<const StructType*>(type))
			report_error(printf_string("The operator [] used with a string constant index can only be applied to objects of type STRUCT. Cannot apply [\"%s\"] to a variable of type\n   %s",
				element_name, type->get_type_string().c_str()));
		nlist<StructType::Elem>::const_iterator it=static_cast<const StructType*>(type)->find_elem(element_name);
		if(!it.is_valid_and_not_end(static_cast<const StructType*>(type)->elem))
			report_error(printf_string("operator[]: '%s' in not an element of %s", element_name, type->get_type_string().c_str()));
		expr ex2;
		ex2.add(Command(bi::LOCALADDR).add_coded_integer255(static_cast<const StructType*>(type)->offsetofelem(element_name)-(int64)local_stack_offset,
				MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
		return GeneralExpr::from_PointerExpr(PointerExpr(PointerType(it->type), ex2.set_type(Value::VarIndexTypeNr), true));
	}
	GeneralExpr operator->*(const char *element_name) const
	{	// ->* : Dereferenzierung des Zeigers und Elementauswahl des sich ergebenden STRUCT-Objekts.
		if(0==dynamic_cast<const PointerType*>(type) ||
				0==dynamic_cast<const StructType*>(static_cast<const PointerType*>(type)->type_pointed_to.get()))
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to a variable of type\n   %s",
				element_name, type->get_type_string().c_str()));
		const Type *ptype=static_cast<const PointerType*>(type)->type_pointed_to;
		const StructType *stype=static_cast<const StructType*>(ptype);
		nlist<StructType::Elem>::const_iterator it=stype->find_elem(element_name);
		if(!it.is_valid_and_not_end(stype->elem))
			report_error(printf_string("operator->*: '%s' in not an element of %s", element_name, type->get_type_string().c_str()));
		if(0==uid)
		{
			expr ex(Command(bi::GETLOCAL).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			// Der Wert auf dem Stack nach der Ausfhrung von GETLOCAL hat den Typ VarIndedTypeNr
			ex.add(expr::get_const_expr(stype->offsetofelem(element_name)));
			ex.addcmd(bi::ADD, -1);
			return GeneralExpr::from_PointerExpr(PointerExpr(PointerType(it->type, static_cast<const PointerType*>(type)->C_pointer),
				ex.set_type(Value::VarIndexTypeNr), true));
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var;
			cmd.sp=1; // Command::get_prev_var wird in die Kommandos PUSHC und GETLOCAL umgewandelt.
			expr ex(cmd);
			ex.add(expr::get_const_expr(stype->offsetofelem(element_name)));
			ex.addcmd(bi::ADD, -1);
			return GeneralExpr::from_PointerExpr(PointerExpr(PointerType(it->type, static_cast<const PointerType*>(type)->C_pointer),
				ex.set_type(Value::VarIndexTypeNr), true));
		}
	}
	GeneralExpr operator[](const expr& ex) const
	{
		ex.set_exported();
		if(0!=dynamic_cast<const ArrayType*>(type))
		{
			expr ex2;
			ex2.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			return PointerExpr(PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr))[ex];
		}
		else if(is_ptr())
			return PointerExpr(*type, expression_yielding_address().set_type(Value::VarIndexTypeNr))[ex];
		else
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or to pointers. Cannot apply [] to a variable of type\n   %s",
				type->get_type_string().c_str()));
	}
	GeneralExpr operator[](long long x) const
	{
		if(0!=dynamic_cast<const ArrayType*>(type))
		{
			if(x<0 || x>=static_cast<const ArrayType*>(type)->nr_elem)
				report_error(printf_string("Array index is out of range. Cannot access index %d of the array\n   %s", (int)x, type->get_type_string().c_str()));
			expr ex2;
			ex2.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(
					PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr), true));
			return PointerExpr(PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr))[x];
		}
		else if(is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply [] to a pointer to void.");
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(*type,
					expression_yielding_address().set_type(Value::VarIndexTypeNr), true));
			else
				return PointerExpr(*type, expression_yielding_address().set_type(Value::VarIndexTypeNr))[x];
		}
		else
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or to pointers. Cannot apply [] to a variable of type\n   %s",
				type->get_type_string().c_str()));
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		if(0!=dynamic_cast<const ArrayType*>(type))
		{
			if(x<0 || x>=static_cast<const ArrayType*>(type)->nr_elem)
				report_error(printf_string("Array index is out of range. Cannot access index %d of the array\n   %s", (int)x, type->get_type_string().c_str()));
			expr ex2;
			ex2.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(
					PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr), true));
			return PointerExpr(PointerType(static_cast<const ArrayType*>(type)->elem_type), ex2.set_type(Value::VarIndexTypeNr))[x];
		}
		else if(is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply [] to a pointer to void.");
			if(0==x)
				return GeneralExpr::from_PointerExpr(PointerExpr(*type,
					expression_yielding_address().set_type(Value::VarIndexTypeNr), true));
			else
				return PointerExpr(*type, expression_yielding_address().set_type(Value::VarIndexTypeNr))[x];
		}
		else
			report_error(printf_string("The operator [] used with an integer type index can only be applied to arrays or to pointers. Cannot apply [] to a variable of type\n   %s",
				type->get_type_string().c_str()));
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }
	~local()
	{
		if(0==uid)
		{
			ProgramStructureElement* p=GetActFunctionProgramStructure("Local variable defined outside of a function.");
			if(0!=dynamic_cast<ProgramStructure_Function*>(p))
				static_cast<ProgramStructure_Function*>(p)->free_local_var_memory(type->size());
			else if(0!=dynamic_cast<ProgramStructure_VoidFunction*>(p))
				static_cast<ProgramStructure_VoidFunction*>(p)->free_local_var_memory(type->size());
		}
	}
	operator expr() const
	{
		if(0==dynamic_cast<const BasicType*>(type)) // Ein Pointer knnte zwar in einen __BOOL umgewandelt werden. Hierfr ist jedoch die Umwandlung nach LogExpr vorgesehen.
			error_conversion_to_elementary_type_not_possible(type);
		if(0==uid)
		{
			expr ex(Command(bi::GETLOCAL).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var;
			cmd.sp=1; // Command::get_prev_var wird in die Kommandos PUSHC und GETLOCAL umgewandelt.
			return expr(cmd).set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
	}
	PointerExpr to_PointerExpr() const
	{
		if(!is_ptr())
			report_error(printf_string("Conversion to Pointer not possible. Cannot convert the type\n   %s\to a pointer type.",
				type->get_type_string().c_str()));
		return PointerExpr(*type, expression_yielding_address());
	}
	void assignment_check_var_mem_initialized() const
	{
		if(0==uid && cc->ProgramStructureStack.size()==created_at_psstack_deep &&
				0!=dynamic_cast<ProgramStructureGlobalScope*>(cc->ProgramStructureStack.front()) &&
				(0!=dynamic_cast<ProgramStructure_Function*>(*(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin()))) ||
				0!=dynamic_cast<ProgramStructure_VoidFunction*>(*(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin())))))
			var_mem_initialized=true;
	}
	GeneralExpr operator=(const local& v) const
	{	// Es ist wichtig, diesen Zuweisungsoperator (mit Quelltyp (local) = Zieltyp (local)) zu definieren, weil er vom Standard-Zuweisungsoperator abweicht.
		return operator=(CollectGeneralType(v));
	}

	expr get_assignment_code(const expr& value) const
	{
		value.set_exported();
		uint8 type_nr=value.get_type();
		if(0==uid)
		{
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
			if(value.is_integer_constant())
			{
				expr ex;
				ex.add(Command(bi::SETLOCALC).add_coded_integer15(value.get_this_integer_constant(), 0, true).
						add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.set_code_has_effect();
				ex.set_type(type_nr);
				return ex;
			}
#endif
			expr ex(value.shortest_type_if_constant());
			/*if(!var_mem_initialized && 0!=dynamic_cast<const BasicType*>(type) && static_cast<const BasicType*>(type)->type!=type_nr)
				ex.add_effective(Command(bi::CAST).addnibble(type_nr=static_cast<const BasicType*>(type)->type));
			if(0!=dynamic_cast<const PointerType*>(type) && Value::VarIndexTypeNr!=type_nr)
				ex.add_effective(Command(bi::CAST).addnibble(type_nr=Value::VarIndexTypeNr));*/

			ex.add(Command(bi::SETLOCAL).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(0));
			ex.set_code_has_effect();
			ex.assignment_to_locals.push_back(this);

			if(0!=dynamic_cast<const BasicType*>(type))
				type_nr=static_cast<const BasicType*>(type)->type;
			if(0!=dynamic_cast<const PointerType*>(type))
				type_nr=Value::VarIndexTypeNr;
			return ex.set_type(type_nr);
		}
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
		else if(value.is_integer_constant())
		{
			expr ex;
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::set_prev_var_const;
			cmd.var_data=value.get_this_integer_constant();
			cmd.sp=1; // Command::set_prev_var wird in die Kommandos PUSHC und SETLOCAL umgewandelt.
			ex.add(cmd);
			ex.set_code_has_effect();
			if(0!=dynamic_cast<const BasicType*>(type))
				type_nr=static_cast<const BasicType*>(type)->type;
			if(0!=dynamic_cast<const PointerType*>(type))
				type_nr=Value::VarIndexTypeNr;
			return ex.set_type(type_nr);
		}
#endif
		else
		{
			expr ex(value.shortest_type_if_constant());
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::set_prev_var;
			cmd.sp=0; // Command::set_prev_var wird in die Kommandos PUSHC und SETLOCAL umgewandelt.
			ex.add(cmd);
			ex.set_code_has_effect();
			if(0!=dynamic_cast<const BasicType*>(type))
				type_nr=static_cast<const BasicType*>(type)->type;
			if(0!=dynamic_cast<const PointerType*>(type))
				type_nr=Value::VarIndexTypeNr;
			return ex.set_type(type_nr);
		}
	}

	GeneralExpr operator=(const CollectGeneralType& value) const
	{
		value.set_exported();
		if(value.is_expr_type())
		{
			if(type->is_ptr())
			{
				if(value.is_basicconst && 0==value.cval) // Zuweisung des Nullzeigers
					return GeneralExpr::from_PointerExpr(PointerExpr(*type, get_assignment_code(expr((VARIABLE_INDEX_TYPE)0))));
				report_error(printf_string("Cannot assign an expression of type\n   %s\nto an expression of type\n   %s",
					value.get_type_string().c_str(), type->get_type_string().c_str()));
			}
			if(0!=dynamic_cast<const StructType*>(type))
				report_error("Cannot assign an expression of elementary type to a struct-object.");
			if(0!=dynamic_cast<const PointerType*>(type))
				report_error(printf_string("Cannot assign the expression\n   %s\nof elementary type not beeing the integer constant 0 to a pointer of type\n   %s",
					value.get_type_string().c_str(), type->get_type_string().c_str()));
			if(0==dynamic_cast<const BasicType*>(type))
				report_error("Internal compiler error: BasicType expected.");
			return GeneralExpr::from_expr(get_assignment_code(value.to_expr()));
		}
		else if(value.is_PointerExpr())
		{
			PointerExpr pe(value.to_PointerExpr());
			if(!type->equal(pe.get_actual_type()))
			{
				if(type->equal(&__BOOL) && 0!=dynamic_cast<const PointerType*>(value.get_type()))
				{
					// Konversion und Zuweisung eines Zeigers an eine bool-Variable
					expr ex;
					ex.add(pe.expression_yielding_address());
					ex.add_effective(Command(bi::CAST).addnibble(Value::t_bool));
					ex.set_type(Value::t_bool);
					return GeneralExpr::from_expr(get_assignment_code(ex));
				}
				if(type->is_ptr() && value.is_basicconst && 0==value.cval) // Zuweisung des Nullzeigers
					return GeneralExpr::from_PointerExpr(PointerExpr(*type, get_assignment_code(expr((VARIABLE_INDEX_TYPE)0))));
				if(type->is_ptr() && static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID) && pe.get_actual_type()->is_ptr())
				{
					// Im Folgenden wird ein Zeiger an einen Zeiger auf __VOID zugewiesen.
					return GeneralExpr::from_PointerExpr(PointerExpr(*type, get_assignment_code(pe.expression_yielding_address())));
				}
				report_error(printf_string("Type mismatch in assignment to the object the pointer points to. Cannot assign an expression of type\n   %s\nto an expression of type\n   %s",
					pe.get_actual_type()->get_type_string().c_str(), type->get_type_string().c_str()));
			}
			if(pe.is_dref() && 0!=dynamic_cast<const BasicType*>(pe.get_pt().type_pointed_to.get()))
			{
				if(!type->equal(pe.get_pt().type_pointed_to))
					error_conversion_not_possible(pe.get_actual_type(), type, "operator=(): ");
				// Zuweisung eines expr an einen dereferenzierten Zeiger auf ein expr
				return GeneralExpr::from_expr(operator=(pe.to_expr())); // Rekursiver Aufruf von operator=() mit einem expr-Parameter
			}
			if(0!=dynamic_cast<const StructType*>(pe.get_actual_type()) || 0!=dynamic_cast<const ArrayType*>(pe.get_actual_type()))
				report_error(printf_string("Cannot assign objects of type STRUCT or ARRAY to variables or dereferenced pointers; please implement the assignment by assigning element-wise. Cannot assign an expression of type\n   %s\nto a variable.",
					pe.get_actual_type()->get_type_string().c_str()));
			if(0==dynamic_cast<const PointerType*>(pe.get_actual_type()))
				error_conversion_not_possible(pe.get_actual_type(), type, "\'=\': ");
			// Im Folgenden wird ein Zeiger an einen Zeiger zugewiesen.
			return GeneralExpr::from_PointerExpr(PointerExpr(*type, get_assignment_code(pe.expression_yielding_address()),
				false, pe.is_actually_memory_with_fixed_type_info()));
		}
		else
			report_error("Unknown type of CollectGeneralType");
	}

	GeneralExpr operator+=(const CollectGeneralType& value)
	{
		value.set_exported();
		if(value.is_PointerExpr() && !value.is_expr_type()) // Ein PointerExpr kann auch einen BasicType ergeben (falls dref=true).
			report_error(printf_string("operator+=() : Cannot apply a pointer on the right hand side. Cannot apply the type\n   %s\non the right hand side.",
				value.get_type_string().c_str()));
		if(!value.is_expr_type()) // nicht value.is_expr() !
			report_error("Unknown Type in CollectGeneralType()");
		if(!type->is_expr() && !type->is_ptr())
			report_error(printf_string("Cannot add to an object of type ARRAY or STRUCT. Cannot add an expression of type\n   %s",
				value.get_type_string().c_str()));
		expr s;
		if(type->is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply += to a pointer to void.");
			expr v2(value.to_expr().shortest_type_if_constant());
			// v2 muss nicht ggf. auf Value::VarIndexTypeNr gecastet werden, weil Voraussetzung zur Anwendung des operator+=() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
			if(1==static_cast<const PointerType*>(type)->type_pointed_to->size())
				s.assign(v2);
			else if(v2.get_type()<Value::SignedVarIndexTypeNr)
				s.assign(v2*expr((SIGNED_VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(type)->type_pointed_to->size()));
			else
				s.assign(v2*expr::get_const_expr((int64)static_cast<const PointerType*>(type)->type_pointed_to->size(), true));
		}
		else // type->is_expr()==true
			s.assign(value.to_expr().shortest_type_if_constant()); // s muss nicht ggf. auf Value::VarIndexTypeNr gecastet werden, weil Voraussetzung zur Anwendung des operator+=() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.

		if(0==uid)
		{
			expr ex;
			int64 c=s.is_integer_constant() ? s.get_this_integer_constant() : 0; // get_this_integer_constant() darf nur aufgerufen werden, wenn tatschlich Konstante
			if(s.is_integer_constant() && ((s.get_type()==Value::t_u8 && c<=8) || (s.get_type()==Value::t_i8 && (c<=8 && c>=-8))))
				ex.add(Command(bi::ADDCLOCAL).addnibble(c>=0 ? (c-1) : (7-c)).
					add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(s.shortest_type_if_constant());
				ex.addcmd(bi::ADDM, -1);
			}
			ex.set_code_has_effect();
			if(type->is_ptr())
				return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr)));
			else
				return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
		else
		{
			expr ex;
			int64 c=s.is_integer_constant() ? s.get_this_integer_constant() : 0; // get_this_integer_constant() darf nur aufgerufen werden, wenn tatschlich Konstante
			if(s.is_integer_constant() && ((s.get_type()==Value::t_u8 && c<=8) || (s.get_type()==Value::t_i8 && (c<=8 && c>=-8))))
			{
				Command cmd;
				cmd.branch1_id=uid;
				cmd.fkt=Command::addc_to_prev_var;
				cmd.sp=1;
				cmd.var_data=c;
				ex.add(cmd);
			}
			else
			{
				Command cmd;
				cmd.branch1_id=uid;
				cmd.fkt=Command::get_prev_var_addr;
				cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
				ex.add(cmd);
				ex.add(s.shortest_type_if_constant());
				ex.addcmd(bi::ADDM, -1);
			}
			ex.set_code_has_effect();
			if(type->is_ptr())
				return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr)));
			else
				return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
	}
	GeneralExpr operator++()
	{
		if(!type->is_expr() && !type->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");

		int64 d;
		if(type->is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply ++ to a pointer to void.");
			d=static_cast<const PointerType*>(type)->type_pointed_to->size();
		}
		else // type->is_expr()==true
			d=1;

		if(0==uid)
		{
			expr ex;
			if(d<=8) // d==0 ist unzulssig
				ex.add(Command(bi::ADDCLOCAL).addnibble(d-1).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::ADDM, -1);
			}
			ex.set_code_has_effect();
			if(type->is_ptr())
				return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr)));
			else
				return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
		else
		{
			expr ex;
			if(d<=8)
			{
				Command cmd;
				cmd.branch1_id=uid;
				cmd.fkt=Command::addc_to_prev_var;
				cmd.sp=1;
				cmd.var_data=d;
				ex.add(cmd);
			}
			else
			{
				// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator++() ist,
				// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
				Command cmd;
				cmd.branch1_id=uid;
				cmd.fkt=Command::get_prev_var_addr;
				cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
				ex.add(cmd);
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::ADDM, -1);
			}
			ex.set_code_has_effect();
			if(type->is_ptr())
				return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr)));
			else
				return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
	}
	LocalPlusPlus operator++(int);

	GeneralExpr operator-=(const CollectGeneralType& value)
	{
		value.set_exported();
		if(value.is_PointerExpr() && !value.is_expr_type()) // Ein PointerExpr kann auch einen BasicType ergeben (falls dref=true).
			report_error(printf_string("operator-=() : Cannot apply a pointer on the right hand side. Cannot apply the type\n   %s\non the right hand side.",
				value.get_type_string().c_str()));
		if(!value.is_expr_type()) // nicht value.is_expr() !
			report_error("Unknown Type in CollectGeneralType()");
		if(!type->is_expr() && !type->is_ptr())
			report_error(printf_string("Cannot sub from an object of type ARRAY or STRUCT. Cannot add an expression of type\n   %s",
				value.get_type_string().c_str()));
		expr s;
		if(type->is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply += to a pointer to void.");
			expr v2(value.to_expr().shortest_type_if_constant());
			// v2 muss nicht ggf. auf Value::VarIndexTypeNr gecastet werden, weil Voraussetzung zur Anwendung des operator+=() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
			if(1==static_cast<const PointerType*>(type)->type_pointed_to->size())
				s.assign(v2);
			else if(v2.get_type()<Value::SignedVarIndexTypeNr)
				s.assign(v2*expr((SIGNED_VARIABLE_INDEX_TYPE)static_cast<const PointerType*>(type)->type_pointed_to->size()));
			else
				s.assign(v2*expr::get_const_expr((int64)static_cast<const PointerType*>(type)->type_pointed_to->size(), true));
		}
		else // type->is_expr()==true
			s.assign(value.to_expr().shortest_type_if_constant()); // s muss nicht ggf. auf Value::VarIndexTypeNr gecastet werden, weil Voraussetzung zur Anwendung des operator+=() ist,
			// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.

		if(0==uid)
		{
			expr ex;
			int64 c=s.is_integer_constant() ? s.get_this_integer_constant() : 0; // get_this_integer_constant() darf nur aufgerufen werden, wenn tatschlich Konstante
			if(s.is_integer_constant() && ((s.get_type()==Value::t_u8 && c<=8) || (s.get_type()==Value::t_i8 && (c<=8 && c>=-8))))
				ex.add(Command(bi::ADDCLOCAL).addnibble(c>=0 ? (7+c) : (-c-1)).
					add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(s.shortest_type_if_constant());
				ex.addcmd(bi::SUBM, -1);
			}
			ex.set_code_has_effect();
			if(type->is_ptr())
				return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr)));
			else
				return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
		else
		{
			expr ex;
			int64 c=s.is_integer_constant() ? s.get_this_integer_constant() : 0; // get_this_integer_constant() darf nur aufgerufen werden, wenn tatschlich Konstante
			if(s.is_integer_constant() && ((s.get_type()==Value::t_u8 && c<=8) || (s.get_type()==Value::t_i8 && (c<=8 && c>=-8))))
			{
				Command cmd;
				cmd.branch1_id=uid;
				cmd.fkt=Command::addc_to_prev_var;
				cmd.sp=1;
				cmd.var_data=-c;
				ex.add(cmd);
			}
			else
			{
				Command cmd;
				cmd.branch1_id=uid;
				cmd.fkt=Command::get_prev_var_addr;
				cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
				ex.add(cmd);
				ex.add(s.shortest_type_if_constant());
				ex.addcmd(bi::SUBM, -1);
			}
			ex.set_code_has_effect();
			if(type->is_ptr())
				return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr)));
			else
				return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
	}
	GeneralExpr operator--()
	{
		if(!type->is_expr() && !type->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");

		int64 d;
		if(type->is_ptr())
		{
			if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply ++ to a pointer to void.");
			d=static_cast<const PointerType*>(type)->type_pointed_to->size();
		}
		else // type->is_expr()==true
			d=1;

		if(0==uid)
		{
			expr ex;
			if(d<=8) // d==0 ist unzulssig
				ex.add(Command(bi::ADDCLOCAL).addnibble(7+d).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::SUBM, -1);
			}
			ex.set_code_has_effect();
			if(type->is_ptr())
				return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr)));
			else
				return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
		else
		{
			expr ex;
			if(d<=8)
			{
				Command cmd;
				cmd.branch1_id=uid;
				cmd.fkt=Command::addc_to_prev_var;
				cmd.sp=1;
				cmd.var_data=-d;
				ex.add(cmd);
			}
			else
			{
				// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator++() ist,
				// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
				Command cmd;
				cmd.branch1_id=uid;
				cmd.fkt=Command::get_prev_var_addr;
				cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
				ex.add(cmd);
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::SUBM, -1);
			}
			ex.set_code_has_effect();
			if(type->is_ptr())
				return GeneralExpr::from_PointerExpr(PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr)));
			else
				return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(type)->type));
		}
	}
	LocalMinusMinus operator--(int);
	expr operator*=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator*=(): Multiplication can only be done with two operands of elementary type. Cannot multiply two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.set_exported();
		if(0==uid)
		{
			expr ex;
			ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.add(x.shortest_type_if_const_keep_signed());
			ex.addcmd(bi::MULM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var_addr;
			cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
			expr ex=cmd;
			ex.add(x.shortest_type_if_const_keep_signed());
			ex.addcmd(bi::MULM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
	}

	expr operator/=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator/=(): Division can only be done with two operands of elementary type. Cannot divide two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.set_exported();
#ifndef WITH_INTEGER64_DIVISION
		if(static_cast<const BasicType*>(type)->type==Value::t_u64 || static_cast<const BasicType*>(type)->type==Value::t_i64)
			report_error("64 Bit division is deactivated in the bytecode interpreter. To activate it, #define WITH_INTEGER64_DIVISION");
#endif
		if(0==uid)
		{
			expr ex;
			ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.add(x.shortest_type_if_const_keep_signed());
			ex.addcmd(bi::DIVM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var_addr;
			cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
			expr ex=cmd;
			ex.add(x.shortest_type_if_const_keep_signed());
			ex.addcmd(bi::DIVM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
	}

	expr operator%=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator%(): Modulo can only be done with two operands of elementary type. Cannot apply % to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.set_exported();
#ifndef WITH_INTEGER64_MODULO
		if(static_cast<const BasicType*>(type)->type==Value::t_u64 || static_cast<const BasicType*>(type)->type==Value::t_i64)
			report_error("64 Bit modulo is deactivated in the bytecode interpreter. To activate it, #define WITH_INTEGER64_MODULO");
#endif
		if(0==uid)
		{
			expr ex;
			ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.add(x.shortest_type_if_const_keep_signed());
			ex.addcmd(bi::MODM, -1);
			ex.set_code_has_effect();
			return ex.set_type(static_cast<const BasicType*>(type)->type);
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var_addr;
			cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
			expr ex=cmd;
			ex.add(x.shortest_type_if_const_keep_signed());
			ex.addcmd(bi::MODM, -1);
			ex.set_code_has_effect();
			return ex.set_type(static_cast<const BasicType*>(type)->type);
		}
	}

	expr operator|=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator|=(): | can only be done with two operands of elementary type. Cannot apply | to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.set_exported();
		if(0==uid)
		{
			expr ex;
			ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.add(x.shortest_type_if_constant());
			ex.addcmd(bi::ORM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var_addr;
			cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
			expr ex=cmd;
			ex.add(x.shortest_type_if_constant());
			ex.addcmd(bi::ORM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
	}

	expr operator&=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator&=(): & can only be done with two operands of elementary type. Cannot apply & to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.set_exported();
		if(0==uid)
		{
			expr ex;
			ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.add(x.shortest_type_if_constant());
			ex.addcmd(bi::ANDM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var_addr;
			cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
			expr ex=cmd;
			ex.add(x.shortest_type_if_constant());
			ex.addcmd(bi::ANDM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
	}

	expr operator^=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator^=(): ^ can only be done with two operands of elementary type. Cannot apply ^ to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.set_exported();
		if(0==uid)
		{
			expr ex;
			ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.add(x.shortest_type_if_constant());
			ex.addcmd(bi::XORM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var_addr;
			cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
			expr ex=cmd;
			ex.add(x.shortest_type_if_constant());
			ex.addcmd(bi::XORM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
	}

	expr operator<<=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator<<=(): Shift can only be done with two operands of elementary type. Cannot apply << to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.set_exported();
		if(0==uid)
		{
			expr ex;
			ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.add(x.shortest_type_if_constant());
			ex.addcmd(bi::SLM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var_addr;
			cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
			expr ex=cmd;
			ex.add(x.shortest_type_if_constant());
			ex.addcmd(bi::SLM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
	}

	expr operator>>=(const expr& x)
	{
		if(!type->is_expr())
			report_error(printf_string("operator>>=(): Shift can only be done with two operands of elementary type. Cannot apply >> to two expressions with types\n   %s\nand\n   %s",
				type->get_type_string().c_str(), x.get_type_string().c_str()));
		x.set_exported();
		if(0==uid)
		{
			expr ex;
			ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			ex.add(x.shortest_type_if_constant());
			ex.addcmd(bi::SRM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var_addr;
			cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
			expr ex=cmd;
			ex.add(x.shortest_type_if_constant());
			ex.addcmd(bi::SRM, -1);
			ex.set_code_has_effect();
			return ex.set_type(0!=dynamic_cast<const BasicType*>(type) ? static_cast<const BasicType*>(type)->type : Value::t_unknown_type);
		}
	}

	///////////////////////////////////////////////////////////////////////////////////

	expr expression_yielding_address() const
	{
		if(!type->is_ptr())
			report_error("expression_yielding_address() called for non-pointer.");
		if(0==uid)
		{
			expr ex(Command(bi::GETLOCAL).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			return ex.set_type(Value::VarIndexTypeNr);
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var;
			cmd.sp=1; // Command::get_prev_var wird in die Kommandos PUSHC und GETLOCAL umgewandelt.
			return expr(cmd).set_type(Value::VarIndexTypeNr);
		}
	}
	PointerExpr operator*()
	{
		if(0==dynamic_cast<const PointerType*>(type))
			report_error("The dereferentiation operator * can only be applied to variables of type pointer or to pointer expressions.");
		if(static_cast<const PointerType*>(type)->type_pointed_to->equal(&__VOID))
			report_error("Cannot dereferentiate a pointer to void.");
		if(0==uid)
		{
			expr ex(Command(bi::GETLOCAL).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			return PointerExpr(*type, ex.set_type(Value::VarIndexTypeNr), true);
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var;
			cmd.sp=1; // Command::get_prev_var wird in die Kommandos PUSHC und GETLOCAL umgewandelt.
			return PointerExpr(*type, expr(cmd).set_type(Value::VarIndexTypeNr), true);
		}
	}
	PointerExpr operator&()
	{
		if(0==type)
			report_error("Unknown type");
		if(0==uid)
		{
			expr ex;
			ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			return PointerExpr(PointerType(type), ex.set_type(Value::VarIndexTypeNr));
		}
		else
		{
			Command cmd;
			cmd.branch1_id=uid;
			cmd.fkt=Command::get_prev_var_addr;
			cmd.sp=1; // Command::get_prev_var wird in die Kommandos PUSHC und GETLOCAL umgewandelt.
			return PointerExpr(PointerType(type), expr(cmd).set_type(Value::VarIndexTypeNr));
		}
	}
};

struct LocalPlusPlus
{
private: void operator=(const LocalPlusPlus& lpp) { report_error("Assignment LocalPlusPlus::operator=() must not be called"); } // Zugriff verhindern
public:
	const local& v; // Hier muss eine Referenz verwendet werden, da ansonsten im Konstruktor LocalPlusPlus::LocalPlusPlus()
		// der Konstruktor local::local aufgerufen wird, welcher nicht lediglich den Speicher der Struktur local kopiert, sondern
		// annimmt, der Benutzer htte eine neue local angelegt und somit entsprechend Speicher auf dem Stack reserviert.
	mutable bool exported;
	const LocalPlusPlus& set_exported() const { exported=true; return *this; }
	LocalPlusPlus(const local& _v) : v(_v) { exported=false; }
	~LocalPlusPlus()
	{
		if(!exported)
		{	// Der folgende Code ist identisch wie beim Prfix Operator ++ und auch effizienter als der Code des Postfix Operators ++,
			// kann jedoch nur verwendet werden, wenn der Wert des Ausdrucks nicht weiter verwendet wird.
			if(!v.type->is_expr() && !v.type->is_ptr())
				report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");

			int64 d;
			if(v.type->is_ptr())
			{
				if(static_cast<const PointerType*>(v.type)->type_pointed_to->equal(&__VOID))
					report_error("Cannot apply ++ to a pointer to void.");
				d=static_cast<const PointerType*>(v.type)->type_pointed_to->size();
			}
			else // type->is_expr()==true
				d=1;

			if(0==v.uid)
			{
				expr ex;
				if(d<=8) // d==0 ist unzulssig
					ex.add(Command(bi::ADDCLOCALPOP).addnibble(d-1).add_coded_integer255(-(int32)v.local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(0));
				else
				{
					ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)v.local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
					ex.add(expr::get_const_expr(d));
					ex.addcmd(bi::ADDMPOP, -2);
				}
				ex.set_code_has_effect();
				Statement stm(ex); // Erzeugung von Code beim Aufruf des Destruktors von stm
			}
			else
			{
				if(d<=8)
				{
					Command cmd;
					cmd.branch1_id=v.uid;
					cmd.fkt=Command::addc_to_prev_var_pop;
					cmd.sp=0;
					cmd.var_data=d;
					expr ex=cmd;
					ex.set_code_has_effect();
					Statement stm(ex); // Erzeugung von Code beim Aufruf des Destruktors von stm
				}
				else
				{
					// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator++() ist,
					// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
					Command cmd;
					cmd.branch1_id=v.uid;
					cmd.fkt=Command::get_prev_var_addr;
					cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
					expr ex=cmd;
					ex.add(expr::get_const_expr(d));
					ex.addcmd(bi::ADDM, -1);
					ex.set_code_has_effect();
					if(v.type->is_ptr())
						ExprObjectDestructionHandler<expr>(ex.set_type(Value::VarIndexTypeNr));
					else if(v.type->is_expr())
						ExprObjectDestructionHandler<expr>(ex.set_type(static_cast<const BasicType*>(v.type)->type));
				}
			}
		}
	}
	GeneralExpr to_GeneralExpr() const
	{
		set_exported();
		if(!v.type->is_expr() && !v.type->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");

		int64 d;
		if(v.type->is_ptr())
		{
			if(static_cast<const PointerType*>(v.type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply ++ to a pointer to void.");
			d=static_cast<const PointerType*>(v.type)->type_pointed_to->size();
		}
		else // type->is_expr()==true
			d=1;

		if(0==v.uid)
		{
			expr ex;
			if(d<=8) // d==0 ist unzulssig
				ex.add(Command(bi::LOCALADDC).addnibble(d-1).add_coded_integer255(-(int32)v.local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)v.local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::ADDM, -1);
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::SUB, -1);
			}
			ex.set_exported();
			ex.set_code_has_effect();
			if(v.type->is_ptr())
				return GeneralExpr::from_PointerExpr(PointerExpr(*v.type, ex.set_type(Value::VarIndexTypeNr)));
			else if(v.type->is_expr())
				return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(v.type)->type));
			else
				report_error("LocalPlusPlus Error");
		}
		else
		{
			expr ex;
			if(d<=8)
			{
				Command cmd;
				cmd.branch1_id=v.uid;
				cmd.fkt=Command::prev_var_addc;
				cmd.sp=1;
				cmd.var_data=d;
				ex.add(cmd);
			}
			else
			{
				Command cmd;
				cmd.branch1_id=v.uid;
				cmd.fkt=Command::get_prev_var_addr;
				cmd.sp=1; // Command::get_prev_var wird in die Kommandos PUSHC und GETLOCAL umgewandelt.
				ex.add(cmd);
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::ADDM, -1);
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::SUB, -1);
			}
			ex.set_exported();
			ex.set_code_has_effect();
			if(v.type->is_ptr())
				return GeneralExpr::from_PointerExpr(PointerExpr(*v.type, ex.set_type(Value::VarIndexTypeNr)));
			else if(v.type->is_expr())
				return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(v.type)->type));
			else
				report_error("LocalPlusPlus Error");
		}
	}
	operator expr() const
		{ return to_GeneralExpr().to_expr(); }
	PointerExpr to_PointerExpr() const
		{ return to_GeneralExpr().to_PointerExpr(); }
	GeneralExpr operator->*(const char *element_name) const
	{	// ->* : Dereferenzierung des Zeigers und Elementauswahl des sich ergebenden STRUCT-Objekts.
		set_exported();
		if(0==dynamic_cast<const PointerType*>(v.type) ||
				0==dynamic_cast<const StructType*>(static_cast<const PointerType*>(v.type)->type_pointed_to.get()))
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to a variable of type\n   %s",
				element_name, v.type->get_type_string().c_str()));
		return (to_PointerExpr())->*element_name;
	}
	GeneralExpr operator[](const expr& ex) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[ex];
	}
	GeneralExpr operator[](long long x) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }
};

struct LocalMinusMinus
{
private: void operator=(const LocalMinusMinus& lpp) { report_error("Assignment LocalMinusMinus::operator=() must not be called"); } // Zugriff verhindern
public:
	const local& v; // Hier muss eine Referenz verwendet werden, da ansonsten im Konstruktor LocalMinusMinus::LocalMinusMinus()
		// der Konstruktor local::local aufgerufen wird, welcher nicht lediglich den Speicher der Struktur local kopiert, sondern
		// annimmt, der Benutzer htte eine neue local angelegt und somit entsprechend Speicher auf dem Stack reserviert.
	mutable bool exported;
	const LocalMinusMinus& set_exported() const { exported=true; return *this; }
	LocalMinusMinus(const local& _v) : v(_v) { exported=false; }
	~LocalMinusMinus()
	{
		if(!exported)
		{	// Der folgende Code ist identisch wie beim Prfix Operator ++ und auch effizienter als der Code des Postfix Operators ++,
			// kann jedoch nur verwendet werden, wenn der Wert des Ausdrucks nicht weiter verwendet wird.
			if(!v.type->is_expr() && !v.type->is_ptr())
				report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");

			int64 d;
			if(v.type->is_ptr())
			{
				if(static_cast<const PointerType*>(v.type)->type_pointed_to->equal(&__VOID))
					report_error("Cannot apply ++ to a pointer to void.");
				d=static_cast<const PointerType*>(v.type)->type_pointed_to->size();
			}
			else // type->is_expr()==true
				d=1;

			if(0==v.uid)
			{
				expr ex;
				if(d<=8) // d==0 ist unzulssig
					ex.add(Command(bi::ADDCLOCALPOP).addnibble(7+d).add_coded_integer255(-(int32)v.local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(0));
				else
				{
					ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)v.local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
					ex.add(expr::get_const_expr(d));
					ex.addcmd(bi::SUBMPOP, -2);
				}
				ex.set_code_has_effect();
				Statement stm(ex); // Erzeugung von Code beim Aufruf des Destruktors von stm
			}
			else
			{
				if(d<=8)
				{
					Command cmd;
					cmd.branch1_id=v.uid;
					cmd.fkt=Command::addc_to_prev_var_pop;
					cmd.sp=0;
					cmd.var_data=-d;
					expr ex=cmd;
					ex.set_code_has_effect();
					Statement stm(ex); // Erzeugung von Code beim Aufruf des Destruktors von stm
				}
				else
				{
					// s muss nicht ggf. auf den korrekten Typ gecastet werden, weil Voraussetzung zur Anwendung des operator++() ist,
					// dass die Variable bereits vorher initialisiert ist und daher automatisch eine Konversion zum korrekten Typ stattfindet.
					Command cmd;
					cmd.branch1_id=v.uid;
					cmd.fkt=Command::get_prev_var_addr;
					cmd.sp=1; // Command::get_prev_var_addr wird in die Kommandos PUSHC und LOCALADDR umgewandelt.
					expr ex=cmd;
					ex.add(expr::get_const_expr(d));
					ex.addcmd(bi::SUBM, -1);
					ex.set_code_has_effect();
					if(v.type->is_ptr())
						ExprObjectDestructionHandler<expr>(ex.set_type(Value::VarIndexTypeNr));
					else if(v.type->is_expr())
						ExprObjectDestructionHandler<expr>(ex.set_type(static_cast<const BasicType*>(v.type)->type));
				}
			}
		}
	}
	GeneralExpr to_GeneralExpr() const
	{
		set_exported();
		if(!v.type->is_expr() && !v.type->is_ptr())
			report_error("operator++(): Cannot increment an object of type ARRAY or STRUCT.");

		int64 d;
		if(v.type->is_ptr())
		{
			if(static_cast<const PointerType*>(v.type)->type_pointed_to->equal(&__VOID))
				report_error("Cannot apply ++ to a pointer to void.");
			d=static_cast<const PointerType*>(v.type)->type_pointed_to->size();
		}
		else // type->is_expr()==true
			d=1;

		if(0==v.uid)
		{
			expr ex;
			if(d<=8) // d==0 ist unzulssig
				ex.add(Command(bi::LOCALADDC).addnibble(7+d).add_coded_integer255(-(int32)v.local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
			else
			{
				ex.add(Command(bi::LOCALADDR).add_coded_integer255(-(int32)v.local_stack_offset, MAX_NR_BYTES_FOR_VARIABLE_INDEX, true).set_sp(1));
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::SUBM, -1);
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::ADD, -1);
			}
			ex.set_exported();
			ex.set_code_has_effect();
			if(v.type->is_ptr())
				return GeneralExpr::from_PointerExpr(PointerExpr(*v.type, ex.set_type(Value::VarIndexTypeNr)));
			else if(v.type->is_expr())
				return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(v.type)->type));
			else
				report_error("LocalPlusPlus Error");
		}
		else
		{
			expr ex;
			if(d<=8)
			{
				Command cmd;
				cmd.branch1_id=v.uid;
				cmd.fkt=Command::prev_var_addc;
				cmd.sp=1;
				cmd.var_data=-d;
				ex.add(cmd);
			}
			else
			{
				Command cmd;
				cmd.branch1_id=v.uid;
				cmd.fkt=Command::get_prev_var_addr;
				cmd.sp=1; // Command::get_prev_var wird in die Kommandos PUSHC und GETLOCAL umgewandelt.
				ex.add(cmd);
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::SUBM, -1);
				ex.add(expr::get_const_expr(d));
				ex.addcmd(bi::ADD, -1);
			}
			ex.set_exported();
			ex.set_code_has_effect();
			if(v.type->is_ptr())
				return GeneralExpr::from_PointerExpr(PointerExpr(*v.type, ex.set_type(Value::VarIndexTypeNr)));
			else if(v.type->is_expr())
				return GeneralExpr::from_expr(ex.set_type(static_cast<const BasicType*>(v.type)->type));
			else
				report_error("LocalPlusPlus Error");
		}
	}
	operator expr() const
		{ return to_GeneralExpr().to_expr(); }
	PointerExpr to_PointerExpr() const
		{ return to_GeneralExpr().to_PointerExpr(); }
	GeneralExpr operator->*(const char *element_name) const
	{	// ->* : Dereferenzierung des Zeigers und Elementauswahl des sich ergebenden STRUCT-Objekts.
		set_exported();
		if(0==dynamic_cast<const PointerType*>(v.type) ||
				0==dynamic_cast<const StructType*>(static_cast<const PointerType*>(v.type)->type_pointed_to.get()))
			report_error(printf_string("The operator ->* used with a rhs string constant index can only be applied to objects of type POINTER(STRUCT). Cannot apply ->* \"%s\" to a variable of type\n   %s",
				element_name, v.type->get_type_string().c_str()));
		return (to_PointerExpr())->*element_name;
	}
	GeneralExpr operator[](const expr& ex) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[ex];
	}
	GeneralExpr operator[](long long x) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](unsigned long long x) const
	{
		set_exported();
		if(!v.is_ptr())
			report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", v.type->get_type_string().c_str()));
		return (to_PointerExpr())[x];
	}
	GeneralExpr operator[](long x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned long x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](int x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned int x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](short x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned short x) const { return operator[]((unsigned long long)x); }
	GeneralExpr operator[](signed char x) const { return operator[]((long long)x); }
	GeneralExpr operator[](unsigned char x) const { return operator[]((unsigned long long)x); }
};

LocalPlusPlus local::operator++(int)
{
	return LocalPlusPlus(*this);
}

LocalMinusMinus local::operator--(int)
{
	return LocalMinusMinus(*this);
}

PointerExpr operator*(const LocalPlusPlus& p)
{
	p.set_exported();
	if(!p.v.is_ptr())
		report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.v.type->get_type_string().c_str()));
	return *(p.to_PointerExpr());
}
PointerExpr operator*(const LocalMinusMinus& p)
{
	p.set_exported();
	if(!p.v.is_ptr())
		report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.v.type->get_type_string().c_str()));
	return *(p.to_PointerExpr());
}
PointerExpr operator*(const VariablePlusPlus& p)
{
	p.set_exported();
	if(!p.v.is_ptr())
		report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.v.type->get_type_string().c_str()));
	return *(p.to_PointerExpr());
}
PointerExpr operator*(const VariableMinusMinus& p)
{
	p.set_exported();
	if(!p.v.is_ptr())
		report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.v.type->get_type_string().c_str()));
	return *(p.to_PointerExpr());
}
PointerExpr operator*(const DrefPointerPlusPlus& p)
{
	p.set_exported();
	if(!p.p.get_actual_type()->is_ptr())
		report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.p.get_actual_type()->get_type_string().c_str()));
	return *(p.to_PointerExpr());
}
PointerExpr operator*(const DrefPointerMinusMinus& p)
{
	p.set_exported();
	if(!p.p.get_actual_type()->is_ptr())
		report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.p.get_actual_type()->get_type_string().c_str()));
	return *(p.to_PointerExpr());
}
PointerExpr operator*(const ParamPlusPlus& p)
{
	p.set_exported();
	if(!p.v.is_ptr())
		report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.v.type->get_type_string().c_str()));
	return *(p.to_PointerExpr());
}
PointerExpr operator*(const ParamMinusMinus& p)
{
	p.set_exported();
	if(!p.v.is_ptr())
		report_error(printf_string("operator * : Cannot apply * to an expression of type\n   %s", p.v.type->get_type_string().c_str()));
	return *(p.to_PointerExpr());
}

CollectGeneralType::CollectGeneralType(const local& x)
{
	if(x.is_expr())
		ex=new expr(x);
	else if(x.is_ptr())
		ex=new PointerExpr(x.to_PointerExpr());
	else
		report_error(printf_string("Cannot create a CollectGeneralType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.type->get_type_string().c_str()));
	cval=0; creal=0; is_basicconst=false;
	set_not_exported();
}

CollectGeneralType::CollectGeneralType(const param& x)
{
	if(x.is_expr())
		ex=new expr(x);
	else if(x.is_ptr())
		ex=new PointerExpr(x.to_PointerExpr());
	else
		report_error(printf_string("Cannot create a CollectGeneralType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.type->get_type_string().c_str()));
	cval=0; creal=0; is_basicconst=false;
	set_not_exported();
}

CollectGeneralType::CollectGeneralType(const Variable& x)
{
	if(x.is_expr())
		ex=new expr(x);
	else if(x.is_ptr())
		ex=new PointerExpr(x.to_PointerExpr());
	else
		report_error(printf_string("Cannot create a CollectGeneralType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.type->get_type_string().c_str()));
	cval=0; creal=0; is_basicconst=false;
	set_not_exported();
}

CollectGeneralType::CollectGeneralType(const Constant& x)
{
	if(x.is_expr())
		ex=new expr(x);
	else if(x.is_ptr())
		ex=new PointerExpr(x.to_PointerExpr());
	else
		report_error(printf_string("Cannot create a CollectGeneralType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.type->get_type_string().c_str()));
	cval=0; creal=0; is_basicconst=false;
	set_not_exported();
}

CollectGeneralType::CollectGeneralType(const LocalPlusPlus& x)
{
	x.set_exported();
	if(x.v.is_expr())
		ex=new expr(x); // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.v.is_ptr())
		ex=new PointerExpr(x.to_PointerExpr());
	else
		report_error(printf_string("Cannot create a CollectGeneralType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.v.type->get_type_string().c_str()));
	cval=0; creal=0; is_basicconst=false;
	set_not_exported();
}

CollectGeneralType::CollectGeneralType(const LocalMinusMinus& x)
{
	x.set_exported();
	if(x.v.is_expr())
		ex=new expr(x); // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.v.is_ptr())
		ex=new PointerExpr(x.to_PointerExpr());
	else
		report_error(printf_string("Cannot create a CollectGeneralType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.v.type->get_type_string().c_str()));
	cval=0; creal=0; is_basicconst=false;
	set_not_exported();
}

CollectGeneralType::CollectGeneralType(const ParamPlusPlus& x)
{
	x.set_exported();
	if(x.v.is_expr())
		ex=new expr(x); // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.v.is_ptr())
		ex=new PointerExpr(x.to_PointerExpr());
	else
		report_error(printf_string("Cannot create a CollectGeneralType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.v.type->get_type_string().c_str()));
	cval=0; creal=0; is_basicconst=false;
	set_not_exported();
}

CollectGeneralType::CollectGeneralType(const ParamMinusMinus& x)
{
	x.set_exported();
	if(x.v.is_expr())
		ex=new expr(x); // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.v.is_ptr())
		ex=new PointerExpr(x.to_PointerExpr());
	else
		report_error(printf_string("Cannot create a CollectGeneralType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.v.type->get_type_string().c_str()));
	cval=0; creal=0; is_basicconst=false;
	set_not_exported();
}

CollectGeneralType::CollectGeneralType(const VariablePlusPlus& x)
{
	x.set_exported();
	if(x.v.is_expr())
		ex=new expr(x); // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.v.is_ptr())
		ex=new PointerExpr(x.to_PointerExpr());
	else
		report_error(printf_string("Cannot create a CollectGeneralType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.v.type->get_type_string().c_str()));
	cval=0; creal=0; is_basicconst=false;
}

CollectGeneralType::CollectGeneralType(const VariableMinusMinus& x)
{
	x.set_exported();
	if(x.v.is_expr())
		ex=new expr(x); // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.v.is_ptr())
		ex=new PointerExpr(x.to_PointerExpr());
	else
		report_error(printf_string("Cannot create a CollectGeneralType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.v.type->get_type_string().c_str()));
	cval=0; creal=0; is_basicconst=false;
}

CollectGeneralType::CollectGeneralType(const DrefPointerPlusPlus& x)
{
	x.set_exported();
	if(x.p.get_actual_type()->is_expr())
		ex=new expr(x); // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.p.get_actual_type()->is_ptr())
		ex=new PointerExpr(x.to_PointerExpr());
	else
		report_error(printf_string("Cannot create a CollectGeneralType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.p.get_actual_type()->get_type_string().c_str()));
	cval=0; creal=0; is_basicconst=false;
}

CollectGeneralType::CollectGeneralType(const DrefPointerMinusMinus& x)
{
	x.set_exported();
	if(x.p.get_actual_type()->is_expr())
		ex=new expr(x); // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.p.get_actual_type()->is_ptr())
		ex=new PointerExpr(x.to_PointerExpr());
	else
		report_error(printf_string("Cannot create a CollectGeneralType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.p.get_actual_type()->get_type_string().c_str()));
	cval=0; creal=0; is_basicconst=false;
}

PointerOrElementaryType::PointerOrElementaryType(const LocalPlusPlus& x)
{
	x.set_exported();
	if(x.v.is_expr())
		{ ex.assign(expr(x)); type=0; } // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.v.is_ptr())
		construct(PointerExpr(x.to_PointerExpr()));
	else
		report_error(printf_string("Cannot create a PointerOrElementaryType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.v.type->get_type_string().c_str()));
	set_not_exported();
}

PointerOrElementaryType::PointerOrElementaryType(const LocalMinusMinus& x)
{
	x.set_exported();
	if(x.v.is_expr())
		{ ex.assign(expr(x)); type=0; } // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.v.is_ptr())
		construct(PointerExpr(x.to_PointerExpr()));
	else
		report_error(printf_string("Cannot create a PointerOrElementaryType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.v.type->get_type_string().c_str()));
	set_not_exported();
}

PointerOrElementaryType::PointerOrElementaryType(const ParamPlusPlus& x)
{
	x.set_exported();
	if(x.v.is_expr())
		{ ex.assign(expr(x)); type=0; } // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.v.is_ptr())
		construct(PointerExpr(x.to_PointerExpr()));
	else
		report_error(printf_string("Cannot create a PointerOrElementaryType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.v.type->get_type_string().c_str()));
	set_not_exported();
}

PointerOrElementaryType::PointerOrElementaryType(const ParamMinusMinus& x)
{
	x.set_exported();
	if(x.v.is_expr())
		{ ex.assign(expr(x)); type=0; } // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.v.is_ptr())
		construct(PointerExpr(x.to_PointerExpr()));
	else
		report_error(printf_string("Cannot create a PointerOrElementaryType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.v.type->get_type_string().c_str()));
	set_not_exported();
}

PointerOrElementaryType::PointerOrElementaryType(const VariablePlusPlus& x)
{
	x.set_exported();
	if(x.v.is_expr())
		{ ex.assign(expr(x)); type=0; } // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.v.is_ptr())
		construct(PointerExpr(x.to_PointerExpr()));
	else
		report_error(printf_string("Cannot create a PointerOrElementaryType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.v.type->get_type_string().c_str()));
	set_not_exported();
}

PointerOrElementaryType::PointerOrElementaryType(const VariableMinusMinus& x)
{
	x.set_exported();
	if(x.v.is_expr())
		{ ex.assign(expr(x)); type=0; } // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.v.is_ptr())
		construct(PointerExpr(x.to_PointerExpr()));
	else
		report_error(printf_string("Cannot create a PointerOrElementaryType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.v.type->get_type_string().c_str()));
	set_not_exported();
}

PointerOrElementaryType::PointerOrElementaryType(const DrefPointerPlusPlus& x)
{
	x.set_exported();
	if(x.p.get_actual_type()->is_expr())
		{ ex.assign(expr(x)); type=0; } // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.p.get_actual_type()->is_ptr())
		construct(PointerExpr(x.to_PointerExpr()));
	else
		report_error(printf_string("Cannot create a PointerOrElementaryType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.p.get_actual_type()->get_type_string().c_str()));
	set_not_exported();
}

PointerOrElementaryType::PointerOrElementaryType(const DrefPointerMinusMinus& x)
{
	x.set_exported();
	if(x.p.get_actual_type()->is_expr())
		{ ex.assign(expr(x)); type=0; } // NICHT x.v.to_expr(); dies wrde keine Inkrementierung durchfhren
	else if(x.p.get_actual_type()->is_ptr())
		construct(PointerExpr(x.to_PointerExpr()));
	else
		report_error(printf_string("Cannot create a PointerOrElementaryType object from an expression of type STRUCT or ARRAY. Cannot create it from an expression of type\n   %s",
			x.p.get_actual_type()->get_type_string().c_str()));
	set_not_exported();
}

GeneralExpr cast(const CollectGeneralType& ex, const Type& type)
{
	ex.set_exported();
	if(ex.is_expr_type()) // nicht ex.is_expr() !
	{
		expr x(ex.to_expr());
		if(0!=dynamic_cast<const BasicType*>(&type))
		{
			if(x.get_type()==static_cast<const BasicType*>(&type)->type)
				return GeneralExpr::from_expr(x);
			expr ex;
			if(x.is_integer_constant())
				ex.add(expr::create_with_type(x.get_this_integer_constant(), static_cast<const BasicType*>(&type)->type));
			else
			{
				ex.add(x);
				ex.add_effective(Command(bi::CAST).addnibble(static_cast<const BasicType*>(&type)->type));
			}
			ex.set_type(static_cast<const BasicType*>(&type)->type);
			return GeneralExpr::from_expr(ex);
		}
		else if(0!=dynamic_cast<const PointerType*>(&type))
		{
			expr ex;
			if(x.is_integer_constant())
				ex.add(expr::create_with_type(x.get_this_integer_constant(), Value::VarIndexTypeNr));
			else
			{
				ex.add(x);
				ex.add_effective(Command(bi::CAST).addnibble(Value::VarIndexTypeNr));
			}
			ex.set_type(Value::VarIndexTypeNr);
			return GeneralExpr::from_PointerExpr(PointerExpr(type, ex));
		}
		else
			error_conversion_not_possible(&BasicType::get_BasicType_from_type_nr(x.get_type()), &type, "cast: ");
	}
	else if(ex.is_PointerExpr())
	{
		PointerExpr x(ex.to_PointerExpr());
		if(0!=dynamic_cast<const BasicType*>(&type))
		{
			if(0!=dynamic_cast<const BasicType*>(x.get_actual_type()))
				return cast(x.to_expr(), type);
			else if(0!=dynamic_cast<const PointerType*>(x.get_actual_type()))
				return cast(x.expression_yielding_address(), type);
			else
				error_conversion_not_possible(x.get_actual_type(), &type, "cast: ");
		}
		else if(0!=dynamic_cast<const PointerType*>(&type))
		{
			if(0!=dynamic_cast<const BasicType*>(x.get_actual_type()))
				return GeneralExpr::from_PointerExpr(PointerExpr(type, x.to_expr(), false));
			else if(0!=dynamic_cast<const PointerType*>(x.get_actual_type()))
				return GeneralExpr::from_PointerExpr(PointerExpr(type, x.expression_yielding_address(), false));
			else
				error_conversion_not_possible(x.get_actual_type(), &type, "cast: ");
		}
		else
			error_conversion_not_possible(x.get_actual_type(), &type, "cast: ");
	}
}

GeneralExpr operator+(const CollectGeneralType& x, const CollectGeneralType& y)
{
	x.set_exported(); y.set_exported();
	if(x.is_expr_type() && y.is_expr_type())
	{
		expr ex, xex=x.to_expr(), yex=y.to_expr();
		if(xex.is_integer_constant() && yex.is_integer_constant())
			return GeneralExpr::from_expr(expr::create_with_cast_up_type(
				xex.get_this_integer_constant() + yex.get_this_integer_constant(), xex, yex));
		else if(xex.get_type()<=yex.get_type())
		{
			ex.add(xex.shortest_type_if_constant());
			ex.add(yex);
		}
		else
		{
			ex.add(xex);
			ex.add(yex.shortest_type_if_constant());
		}
		ex.addcmd(bi::ADD, -1);
		ex.set_type(x.to_expr().get_type_of_cast_up(y.to_expr()));
		return GeneralExpr::from_expr(ex);
	}
	if(y.is_PointerExpr())
		return y.to_PointerExpr()+x.to_expr();
	return x.to_PointerExpr()+y.to_expr();
}

GeneralExpr operator-(const CollectGeneralType& x, const CollectGeneralType& y)
{
	x.set_exported(); y.set_exported();
	if(x.is_expr_type() && y.is_expr_type())
	{
		expr ex, xex=x.to_expr(), yex=y.to_expr();
		if(xex.is_integer_constant() && yex.is_integer_constant())
			return GeneralExpr::from_expr(expr::create_with_cast_up_type(
				xex.get_this_integer_constant() - yex.get_this_integer_constant(), xex, yex));
		else if(xex.get_type()<=yex.get_type())
		{
			ex.add(xex.shortest_type_if_constant());
			ex.add(yex);
		}
		else
		{
			ex.add(xex);
			ex.add(yex.shortest_type_if_constant());
		}
		ex.addcmd(bi::SUB, -1);
		ex.set_type(x.to_expr().get_type_of_cast_up(y.to_expr()));
		return GeneralExpr::from_expr(ex);
	}
	if(x.is_PointerExpr())
		return x.to_PointerExpr()-y.to_expr();
	report_error(printf_string("operator-(): Subtraction x-y is not defined for types\n      x: %s\nand   y: %s",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
}

expr operator*(const CollectGeneralType& x, const CollectGeneralType& y)
{
	if(!x.is_expr_type() || !y.is_expr_type())
		report_error(printf_string("Operator * : This operator cannot be used with operands of types\n   %s\nand\n   %s",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
	x.set_exported(); y.set_exported();
	expr ex, xex=x.to_expr(), yex=y.to_expr();
	if(xex.is_integer_constant() && yex.is_integer_constant())
	{
		Type& cast_up_type=BasicType::get_BasicType_from_type_nr(xex.get_type_of_cast_up(yex));
		return GeneralExpr::from_expr(expr::create_with_cast_up_type(
			(cast(xex, cast_up_type).to_expr().get_this_integer_constant()) *
			(cast(yex, cast_up_type).to_expr().get_this_integer_constant()), xex, yex));
	}
	else if(xex.get_type()<=yex.get_type())
	{
		ex.add(xex.shortest_type_if_constant());
		ex.add(yex);
	}
	else
	{
		ex.add(xex);
		ex.add(yex.shortest_type_if_constant());
	}
	ex.addcmd(bi::MUL, -1);
	ex.set_type(x.to_expr().get_type_of_cast_up(y.to_expr()));
	return ex;
}

expr operator/(const CollectGeneralType& x, const CollectGeneralType& y)
{
	if(!x.is_expr_type() || !y.is_expr_type())
		report_error(printf_string("Operator / : This operator cannot be used with operands of types\n   %s\nand\n   %s",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
	x.set_exported(); y.set_exported();
#ifndef WITH_INTEGER64_DIVISION
	if(x.to_expr().get_type_of_cast_up(y.to_expr())==Value::t_u64 || x.to_expr().get_type_of_cast_up(y.to_expr())==Value::t_i64)
		report_error("64 Bit division is deactivated in the bytecode interpreter. To activate it, #define WITH_INTEGER64_DIVISION");
#endif
	// Die Flle, dass sowohl x als auch y Konstanten sind, werden derzeit nicht abgehandelt, weil dann einige Fallunterscheidungen
	// bezglich der Typen notwendig wren.
	expr ex;
	if(x.to_expr().get_type()<=y.to_expr().get_type())
	{
		ex.add(x.to_expr().shortest_type_if_constant());
		ex.add(y.to_expr());
	}
	else
	{
		ex.add(x.to_expr());
		ex.add(y.to_expr().shortest_type_if_constant());
	}
	ex.addcmd(bi::DIV, -1);
	ex.set_type(x.to_expr().get_type_of_cast_up(y.to_expr()));
	return ex;
}

expr operator%(const CollectGeneralType& x, const CollectGeneralType& y)
{
	if(!x.is_expr_type() || !y.is_expr_type())
		report_error(printf_string("Operator % : This operator cannot be used with operands of types\n   %s\nand\n   %s",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
	x.set_exported(); y.set_exported();
	// Die Flle, dass sowohl x als auch y Konstanten sind, werden derzeit nicht abgehandelt, weil dann einige Fallunterscheidungen
	// bezglich der Typen notwendig wren.
#ifndef WITH_INTEGER64_MODULO
	if(x.to_expr().get_type_of_cast_up(y.to_expr())==Value::t_u64 || x.to_expr().get_type_of_cast_up(y.to_expr())==Value::t_i64)
		report_error("64 Bit modulo is deactivated in the bytecode interpreter. To activate it, #define WITH_INTEGER64_MODULO");
#endif
	expr ex;
	if(x.to_expr().get_type()<=y.to_expr().get_type())
	{
		ex.add(x.to_expr().shortest_type_if_constant());
		ex.add(y.to_expr());
	}
	else
	{
		ex.add(x.to_expr());
		ex.add(y.to_expr().shortest_type_if_constant());
	}
	ex.addcmd(bi::MOD, -1);
	ex.set_type(x.to_expr().get_type_of_cast_up(y.to_expr()));
	return ex;
}

expr operator|(const CollectGeneralType& x, const CollectGeneralType& y)
{
	if(!x.is_expr_type() || !y.is_expr_type())
		report_error(printf_string("Operator | : This operator cannot be used with operands of types\n   %s\nand\n   %s",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
	x.set_exported(); y.set_exported();
	expr ex, xex=x.to_expr(), yex=y.to_expr();
	if(xex.is_integer_constant() && yex.is_integer_constant())
	{
		Type& cast_up_type=BasicType::get_BasicType_from_type_nr(xex.get_type_of_cast_up(yex));
		return GeneralExpr::from_expr(expr::create_with_cast_up_type(
			(cast(xex, cast_up_type).to_expr().get_this_integer_constant()) |
			(cast(yex, cast_up_type).to_expr().get_this_integer_constant()), xex, yex));
	}
	else if(xex.get_type()<=yex.get_type())
	{
		ex.add(xex.shortest_type_if_constant());
		ex.add(yex);
	}
	else
	{
		ex.add(xex);
		ex.add(yex.shortest_type_if_constant());
	}
	ex.addcmd(bi::OR, -1);
	ex.set_type(x.to_expr().get_type_of_cast_up(y.to_expr()));
	return ex;
}

expr operator&(const CollectGeneralType& x, const CollectGeneralType& y)
{
	if(!x.is_expr_type() || !y.is_expr_type())
		report_error(printf_string("Operator & : This operator cannot be used with operands of types\n   %s\nand\n   %s",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
	x.set_exported(); y.set_exported();
	expr ex, xex=x.to_expr(), yex=y.to_expr();
	if(xex.is_integer_constant() && yex.is_integer_constant())
	{
		Type& cast_up_type=BasicType::get_BasicType_from_type_nr(xex.get_type_of_cast_up(yex));
		return GeneralExpr::from_expr(expr::create_with_cast_up_type(
			(cast(xex, cast_up_type).to_expr().get_this_integer_constant()) &
			(cast(yex, cast_up_type).to_expr().get_this_integer_constant()), xex, yex));
	}
	else if(xex.get_type()<=yex.get_type())
	{
		ex.add(xex.shortest_type_if_constant());
		ex.add(yex);
	}
	else
	{
		ex.add(xex);
		ex.add(yex.shortest_type_if_constant());
	}
	ex.addcmd(bi::AND, -1);
	ex.set_type(x.to_expr().get_type_of_cast_up(y.to_expr()));
	return ex;
}

expr operator^(const CollectGeneralType& x, const CollectGeneralType& y)
{
	if(!x.is_expr_type() || !y.is_expr_type())
		report_error(printf_string("Operator ^ : This operator cannot be used with operands of types\n   %s\nand\n   %s",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
	x.set_exported(); y.set_exported();
	expr ex, xex=x.to_expr(), yex=y.to_expr();
	if(xex.is_integer_constant() && yex.is_integer_constant())
	{
		Type& cast_up_type=BasicType::get_BasicType_from_type_nr(xex.get_type_of_cast_up(yex));
		return GeneralExpr::from_expr(expr::create_with_cast_up_type(
			(cast(xex, cast_up_type).to_expr().get_this_integer_constant()) ^
			(cast(yex, cast_up_type).to_expr().get_this_integer_constant()), xex, yex));
	}
	else if(xex.get_type()<=yex.get_type())
	{
		ex.add(xex.shortest_type_if_constant());
		ex.add(yex);
	}
	else
	{
		ex.add(xex);
		ex.add(yex.shortest_type_if_constant());
	}
	ex.addcmd(bi::XOR, -1);
	ex.set_type(x.to_expr().get_type_of_cast_up(y.to_expr()));
	return ex;
}

expr operator<<(const CollectGeneralType& x, const CollectGeneralType& y)
{
	if(!x.is_expr_type() || !y.is_expr_type())
		report_error(printf_string("Operator << : This operator cannot be used with operands of types\n   %s\nand\n   %s",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
	x.set_exported(); y.set_exported();
	expr ex, xex=x.to_expr(), yex=y.to_expr();
	if(xex.is_integer_constant() && yex.is_integer_constant())
	{
		return GeneralExpr::from_expr(expr::create_with_type(
			(xex.get_this_integer_constant()) << (yex.get_this_integer_constant()), xex.get_type()));
	}
	else
	{
		ex.add(xex);
		ex.add(yex.shortest_type_if_constant());
	}
	ex.addcmd(bi::SL, -1);
	ex.set_type(x.to_expr().get_type_of_cast_up(y.to_expr()));
	return ex;
}

expr operator>>(const CollectGeneralType& x, const CollectGeneralType& y)
{
	if(!x.is_expr_type() || !y.is_expr_type())
		report_error(printf_string("Operator >> : This operator cannot be used with operands of types\n   %s\nand\n   %s",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
	x.set_exported(); y.set_exported();
	expr ex, xex=x.to_expr(), yex=y.to_expr();
	if(xex.is_integer_constant() && yex.is_integer_constant())
	{
		if(Value::s_is_signed_integer(xex.get_type()))
			return GeneralExpr::from_expr(expr::create_with_type(
				((int64)xex.get_this_integer_constant()) >> (yex.get_this_integer_constant()), xex.get_type()));
		else
			return GeneralExpr::from_expr(expr::create_with_type(
				((uint64)xex.get_this_integer_constant()) >> (yex.get_this_integer_constant()), xex.get_type()));
	}
	else
	{
		ex.add(xex);
		ex.add(yex.shortest_type_if_constant());
	}
	ex.addcmd(bi::SR, -1);
	ex.set_type(x.to_expr().get_type_of_cast_up(y.to_expr()));
	return ex;
}

CollectGeneralType operator+(const CollectGeneralType& x)
{
	x.set_exported();
	return x;
}

expr operator-(const CollectGeneralType& x)
{
	if(!x.is_expr_type())
		report_error(printf_string("Operator - : This operator cannot be used with an operand of type\n   %s",
			x.get_type_string().c_str()));
	x.set_exported();
	if(x.to_expr().is_integer_constant())
		return GeneralExpr::from_expr(expr::create_with_type(-(int64)x.to_expr().get_this_integer_constant(), x.to_expr().get_type()));
	expr ex;
	ex.add(x.to_expr());;
	ex.addcmd(bi::NEG);
	ex.set_type(x.to_expr().get_type());
	return ex;
}

expr operator~(const CollectGeneralType& x)
{
	if(!x.is_expr_type())
		report_error(printf_string("Operator ~ : This operator cannot be used with an operand of type\n   %s",
			x.get_type_string().c_str()));
	x.set_exported();
	if(x.to_expr().is_integer_constant())
		return GeneralExpr::from_expr(expr::create_with_type(~x.to_expr().get_this_integer_constant(), x.to_expr().get_type()));
	expr ex;
	ex.add(x.to_expr());;
	ex.addcmd(bi::COM);
	ex.set_type(x.to_expr().get_type());
	return ex;
}

GeneralExpr operator,(const CollectGeneralType& x, const CollectGeneralType& y)
{		// Der erste Ausdruck wird ausgewertet und verworfen, der zweite wird ausgewertet und zurckgegeben
	x.set_exported(); y.set_exported();
	expr ex;
	if(x.is_expr_type())
		{ if(!x.to_expr().is_integer_constant()) { ex.add(x.to_expr()); ex.addcmd(bi::POP, -1); } }
	else if(x.is_PointerExpr())
	{
		if(!x.to_PointerExpr().expression_yielding_address().is_integer_constant())
			{ ex.add(x.to_PointerExpr().expression_yielding_address()); ex.addcmd(bi::POP, -1); }
	}
	if(y.is_expr_type())
	{
		ex.add_set_type(y.to_expr());
		return GeneralExpr::from_expr(ex);
	}
	else if(y.is_PointerExpr())
	{
		PointerExpr p(y.to_PointerExpr());
		ex.add_set_type(p.get_addr());
		return GeneralExpr::from_PointerExpr(PointerExpr(p.get_pt(), ex, p.is_dref(), p.is_memory_with_fixed_type_info()));
	}
}

GeneralExpr COND(const LogExpr& condition,
				 const CollectGeneralType& expression_if_true, const CollectGeneralType& expression_if_false)
{	// Entspricht condition ? expression_if_true : expression_if_false
	condition.set_exported();
	expression_if_true.set_exported();
	expression_if_false.set_exported();
	const LogExpr &cnd=condition;
	const CollectGeneralType &ext=expression_if_true, &exf=expression_if_false;
	if((ext.is_expr_type()!=exf.is_expr_type()) || (ext.is_pointer_type()!=exf.is_pointer_type()))
		report_error(printf_string("COND: Either both 'expression_if_true' and 'expression_if_false' must be expressions of elementary type or both must be pointers to the same type. \
Cannot convert the type\n   %s\nof 'expression_if_true' to the type\n   %s\nof 'expression_if_false'.",
			expression_if_true.get_type_string().c_str(), expression_if_false.get_type_string().c_str()));
	if(ext.is_expr_type() && exf.is_expr_type())
	{
		const BasicType* cast_up_type=&BasicType::get_BasicType_from_type_nr(
			expression_if_true.to_expr().get_type_of_cast_up(expression_if_false.to_expr()));
		expr et(cast(expression_if_true.to_expr(), *cast_up_type));
		expr ef(cast(expression_if_false.to_expr(), *cast_up_type));
		Command::uid_type uid=Command::generate_new_uid();
		Command cmd;
		cmd.add(bi::BR);
		cmd.branch1_id=uid;
		cmd.fkt=Command::command;
		et.add_effective(cmd);
		et.add_effective(Command::new_uid_label().set_sp(-1)); // muss nach et.add(cmd) erfolgen !
		expr ex(if_code(condition, et));
		ex.add(ef);
		ex.add_effective(Command::new_uid_label(uid));
		et.set_exported(); ef.set_exported();
		return GeneralExpr::from_expr(ex.set_type(expression_if_true.to_expr().get_type_of_cast_up(expression_if_false.to_expr())));
	}
	if(ext.is_pointer_type() && exf.is_pointer_type())
	{
		PointerExpr pt(ext.to_PointerExpr()), pf(exf.to_PointerExpr());
		if(!pt.get_actual_type()->equal(pf.get_actual_type()))
			report_error(printf_string("COND: If both 'expression_if_true' and 'expression_if_false' are pointers, they must point to the same type. \
Cannot convert the type\n   %s\nof 'expression_if_true' to the type\n   %s\nof 'expression_if_false'.",
				expression_if_true.get_type_string().c_str(), expression_if_false.get_type_string().c_str()));
		Command::uid_type uid=Command::generate_new_uid();
		Command cmd;
		cmd.add(bi::BR);
		cmd.branch1_id=uid;
		cmd.fkt=Command::command;
		expr et(pt.expression_yielding_address());
		et.add(cmd);
		et.add_effective(Command::new_uid_label().set_sp(-1)); // muss nach et.add(cmd) erfolgen !
		expr ex(if_code(condition, et));
		ex.add(pf.expression_yielding_address());
		ex.add_effective(Command::new_uid_label(uid));
		return GeneralExpr::from_PointerExpr(PointerExpr(*pt.get_actual_type(), ex.set_type(Value::VarIndexTypeNr)));
	}
	report_error("Internal compiler error in function COND()");
}

GeneralExpr COND(const BRO_x_And_y_BRC_And_z& x, const CollectGeneralType& expression_if_true, const CollectGeneralType& expression_if_false)
{
	x.set_exported(); expression_if_true.set_exported(); expression_if_false.set_exported();
	return COND(x.to_LogExpr(), expression_if_true, expression_if_false);
}
GeneralExpr COND(const BRO_x_Or_y_BRC_Or_z& x, const CollectGeneralType& expression_if_true, const CollectGeneralType& expression_if_false)
{
	x.set_exported(); expression_if_true.set_exported(); expression_if_false.set_exported();
	return COND(x.to_LogExpr(), expression_if_true, expression_if_false);
}
GeneralExpr COND(const BRO_x_And_y_BRC_Or_z& x, const CollectGeneralType& expression_if_true, const CollectGeneralType& expression_if_false)
{
	x.set_exported(); expression_if_true.set_exported(); expression_if_false.set_exported();
	return COND(x.to_LogExpr(), expression_if_true, expression_if_false);
}
GeneralExpr COND(const x_And_BRO_y_Or_z_BRC& x, const CollectGeneralType& expression_if_true, const CollectGeneralType& expression_if_false)
{
	x.set_exported(); expression_if_true.set_exported(); expression_if_false.set_exported();
	return COND(x.to_LogExpr(), expression_if_true, expression_if_false);
}
GeneralExpr COND(const BRO_x_Or_y_BRC_And_z& x, const CollectGeneralType& expression_if_true, const CollectGeneralType& expression_if_false)
{
	x.set_exported(); expression_if_true.set_exported(); expression_if_false.set_exported();
	return COND(x.to_LogExpr(), expression_if_true, expression_if_false);
}
GeneralExpr COND(const x_Or_BRO_y_And_z_BRC& x, const CollectGeneralType& expression_if_true, const CollectGeneralType& expression_if_false)
{
	x.set_exported(); expression_if_true.set_exported(); expression_if_false.set_exported();
	return COND(x.to_LogExpr(), expression_if_true, expression_if_false);
}

LogExpr operator!=(const CollectGeneralType& x, const CollectGeneralType& y)
{
	x.set_exported(); y.set_exported();
	if((x.is_expr_type() && y.is_expr_type()) || (x.is_pointer_type() && y.is_pointer_type()))
	{
		if(x.is_expr_type() && x.to_expr().is_integer_constant() && x.to_expr().get_this_integer_constant()==0)
		{
			LogExpr ex;
			ex.x.add_set_type(y.is_pointer_type() ? y.to_PointerExpr().expression_yielding_address() : y.to_expr());
			ex.negated=false;
			ex.is_converted_to_bool_type=false;
			return ex;
		}
		if(y.is_expr_type() && y.to_expr().is_integer_constant() && y.to_expr().get_this_integer_constant()==0)
		{
			LogExpr ex;
			ex.x.add_set_type(x.is_pointer_type() ? x.to_PointerExpr().expression_yielding_address() : x.to_expr());
			ex.negated=false;
			ex.is_converted_to_bool_type=false;
			return ex;
		}
		LogExpr ex;
		ex.x.add(x.is_pointer_type() ? x.to_PointerExpr().expression_yielding_address() : x.to_expr().shortest_type_if_constant());
		ex.x.add(y.is_pointer_type() ? y.to_PointerExpr().expression_yielding_address() : y.to_expr().shortest_type_if_constant());
		ex.x.addcmd(bi::NE, -1); // Durch den NE-Befehl wird automatisch eine Konvertierung nach bool vorgenommen
		ex.negated=false;
		ex.is_converted_to_bool_type=true;
		ex.x.set_type(Value::t_bool);
		return ex;
	}
	if((x.is_pointer_type() && y.is_expr() && y.is_basicconst && Value::s_is_integer_type(y.to_expr().get_type())) ||
		(y.is_pointer_type() && x.is_expr() && x.is_basicconst && Value::s_is_integer_type(x.to_expr().get_type())))
	{
		if((y.is_expr() && y.is_basicconst && 0!=y.cval) || (x.is_expr() && x.is_basicconst && 0!=x.cval))
			report_error("operator==(): A pointer can only be compares against another pointer or the integer constant 0.");
		LogExpr ex;
		ex.x.add(x.is_pointer_type() ? x.to_PointerExpr().expression_yielding_address() : y.to_PointerExpr().expression_yielding_address());
		ex.negated=false;
		ex.is_converted_to_bool_type=false;
		ex.x.set_type(Value::t_bool);
		return ex;
	}
	report_error(printf_string("operator!=(): Comparision != is not defined for types\n      x: %s\nand   y: %s\nA pointer can only be compared to another pointer or to the integer CONSTANT 0.",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
}

LogExpr operator==(const CollectGeneralType& x, const CollectGeneralType& y)
{
	x.set_exported(); y.set_exported();
	if((x.is_expr_type() && y.is_expr_type()) || (x.is_pointer_type() && y.is_pointer_type()))
	{
		if(x.is_expr_type() && x.to_expr().is_integer_constant() && x.to_expr().get_this_integer_constant()==0)
		{
			LogExpr ex;
			ex.x.add_set_type(y.is_pointer_type() ? y.to_PointerExpr().expression_yielding_address() : y.to_expr());
			ex.negated=true;
			ex.is_converted_to_bool_type=false;
			return ex;
		}
		if(y.is_expr_type() && y.to_expr().is_integer_constant() && y.to_expr().get_this_integer_constant()==0)
		{
			LogExpr ex;
			ex.x.add_set_type(x.is_pointer_type() ? x.to_PointerExpr().expression_yielding_address() : x.to_expr());
			ex.negated=true;
			ex.is_converted_to_bool_type=false;
			return ex;
		}
		LogExpr ex;
		ex.x.add(x.is_pointer_type() ? x.to_PointerExpr().expression_yielding_address() : x.to_expr().shortest_type_if_constant());
		ex.x.add(y.is_pointer_type() ? y.to_PointerExpr().expression_yielding_address() : y.to_expr().shortest_type_if_constant());
		ex.x.addcmd(bi::NE, -1); // Durch den NE-Befehl wird automatisch eine Konvertierung nach bool vorgenommen
		ex.negated=true;
		ex.is_converted_to_bool_type=true;
		ex.x.set_type(Value::t_bool);
		return ex;
	}
	if((x.is_pointer_type() && y.is_expr() && y.is_basicconst && Value::s_is_integer_type(y.to_expr().get_type())) ||
		(y.is_pointer_type() && x.is_expr() && x.is_basicconst && Value::s_is_integer_type(x.to_expr().get_type())))
	{
		if((y.is_expr() && y.is_basicconst && 0!=y.cval) || (x.is_expr() && x.is_basicconst && 0!=x.cval))
			report_error("operator==(): A pointer can only be compares against another pointer or the integer constant 0.");
		LogExpr ex;
		ex.x.add(x.is_pointer_type() ? x.to_PointerExpr().expression_yielding_address() : y.to_PointerExpr().expression_yielding_address());
		ex.negated=true;
		ex.is_converted_to_bool_type=false;
		ex.x.set_type(Value::t_bool);
		return ex;
	}
	report_error(printf_string("operator==(): Comparision == is not defined for types\n      x: %s\nand   y: %s\nA pointer can only be compared to another pointer or to the integer CONSTANT 0.",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
}

LogExpr operator>(const CollectGeneralType& x, const CollectGeneralType& y)
{
	x.set_exported(); y.set_exported();
	if((x.is_expr_type() && y.is_expr_type()) || (x.is_pointer_type() && y.is_pointer_type()))
	{
		LogExpr ex;
		ex.x.add(y.is_pointer_type() ? y.to_PointerExpr().expression_yielding_address() : y.to_expr().shortest_type_if_const_keep_signed());
		ex.x.add(x.is_pointer_type() ? x.to_PointerExpr().expression_yielding_address() : x.to_expr().shortest_type_if_const_keep_signed());
		ex.x.addcmd(bi::LT, -1); // Durch den LT-Befehl wird automatisch eine Konvertierung nach bool vorgenommen
		ex.is_converted_to_bool_type=true;
		ex.negated=false;
		ex.x.set_type(Value::t_bool);
		return ex;
	}
	report_error(printf_string("operator>(): Comparision > is not defined for types\n      x: %s\nand   y: %s\nA pointer can only be compared to another pointer with >.",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
}

LogExpr operator<(const CollectGeneralType& x, const CollectGeneralType& y)
{
	x.set_exported(); y.set_exported();
	if((x.is_expr_type() && y.is_expr_type()) || (x.is_pointer_type() && y.is_pointer_type()))
	{
		LogExpr ex;
		ex.x.add(x.is_pointer_type() ? x.to_PointerExpr().expression_yielding_address() : x.to_expr().shortest_type_if_const_keep_signed());
		ex.x.add(y.is_pointer_type() ? y.to_PointerExpr().expression_yielding_address() : y.to_expr().shortest_type_if_const_keep_signed());
		ex.x.addcmd(bi::LT, -1); // Durch den LT-Befehl wird automatisch eine Konvertierung nach bool vorgenommen
		ex.is_converted_to_bool_type=true;
		ex.negated=false;
		ex.x.set_type(Value::t_bool);
		return ex;
	}
	report_error(printf_string("operator>(): Comparision > is not defined for types\n      x: %s\nand   y: %s\nA pointer can only be compared to another pointer with >.",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
}

LogExpr operator>=(const CollectGeneralType& x, const CollectGeneralType& y)
{
	x.set_exported(); y.set_exported();
	if((x.is_expr_type() && y.is_expr_type()) || (x.is_pointer_type() && y.is_pointer_type()))
	{
		LogExpr ex;
		ex.x.add(x.is_pointer_type() ? x.to_PointerExpr().expression_yielding_address() : x.to_expr().shortest_type_if_const_keep_signed());
		ex.x.add(y.is_pointer_type() ? y.to_PointerExpr().expression_yielding_address() : y.to_expr().shortest_type_if_const_keep_signed());
		ex.x.addcmd(bi::LT, -1); // Durch den LT-Befehl wird automatisch eine Konvertierung nach bool vorgenommen
		ex.is_converted_to_bool_type=true;
		ex.negated=true;
		ex.x.set_type(Value::t_bool);
		return ex;
	}
	report_error(printf_string("operator>(): Comparision > is not defined for types\n      x: %s\nand   y: %s\nA pointer can only be compared to another pointer with >.",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
}

LogExpr operator<=(const CollectGeneralType& x, const CollectGeneralType& y)
{
	x.set_exported(); y.set_exported();
	if((x.is_expr_type() && y.is_expr_type()) || (x.is_pointer_type() && y.is_pointer_type()))
	{
		LogExpr ex;
		ex.x.add(y.is_pointer_type() ? y.to_PointerExpr().expression_yielding_address() : y.to_expr().shortest_type_if_const_keep_signed());
		ex.x.add(x.is_pointer_type() ? x.to_PointerExpr().expression_yielding_address() : x.to_expr().shortest_type_if_const_keep_signed());
		ex.x.addcmd(bi::LT, -1); // Durch den LT-Befehl wird automatisch eine Konvertierung nach bool vorgenommen
		ex.is_converted_to_bool_type=true;
		ex.negated=true;
		ex.x.set_type(Value::t_bool);
		return ex;
	}
	report_error(printf_string("operator>(): Comparision > is not defined for types\n      x: %s\nand   y: %s\nA pointer can only be compared to another pointer with >.",
			x.get_type_string().c_str(), y.get_type_string().c_str()));
}

LogExpr::LogExpr(const GeneralExpr& genex)
{
	genex.set_exported();
	negated=false; is_converted_to_bool_type=false;
	if(genex.is_expr_type())
		x.assign(genex.to_expr());
	else if(genex.is_PointerExpr())
		x.assign(genex.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

LogExpr::LogExpr(const PointerExpr& p)
{
	p.set_exported();
	negated=false; is_converted_to_bool_type=false;
	if(p.is_expr_type())
		x.assign(p.to_expr());
	else
		x.assign(p.expression_yielding_address());
	set_not_exported();
}

LogExpr::LogExpr(const char *y) :x(expr::get_const_expr(0!=y)), negated(false), is_converted_to_bool_type(true)
	{ set_not_exported(); }

LogExpr::LogExpr(const local& y)
{
	negated=false; is_converted_to_bool_type=false;
	if(y.is_expr())
		x.assign(y);
	else if(y.is_ptr())
		x.assign(y.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

LogExpr::LogExpr(const param& y)
{
	negated=false; is_converted_to_bool_type=false;
	if(y.is_expr())
		x.assign(y);
	else if(y.is_ptr())
		x.assign(y.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

LogExpr::LogExpr(const Variable& y)
{
	negated=false; is_converted_to_bool_type=false;
	if(y.is_expr())
		x.assign(y);
	else if(y.is_ptr())
		x.assign(y.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

LogExpr::LogExpr(const Constant& y)
{
	negated=false; is_converted_to_bool_type=false;
	if(y.is_expr())
		x.assign(y);
	else if(y.is_ptr())
		x.assign(y.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

LogExpr::LogExpr(const VariablePlusPlus &v)
{
	v.set_exported();
	negated=false; is_converted_to_bool_type=false;
	if(v.v.is_expr())
		x.assign(v.to_GeneralExpr().to_expr());
	else if(v.v.is_ptr())
		x.assign(v.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

LogExpr::LogExpr(const VariableMinusMinus &v)
{
	v.set_exported();
	negated=false; is_converted_to_bool_type=false;
	if(v.v.is_expr())
		x.assign(v.to_GeneralExpr().to_expr());
	else if(v.v.is_ptr())
		x.assign(v.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

LogExpr::LogExpr(const DrefPointerPlusPlus &v)
{
	v.set_exported();
	negated=false; is_converted_to_bool_type=false;
	if(v.p.get_actual_type()->is_expr())
		x.assign(v.to_GeneralExpr().to_expr());
	else if(v.p.get_actual_type()->is_ptr())
		x.assign(v.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

LogExpr::LogExpr(const DrefPointerMinusMinus &v)
{
	v.set_exported();
	negated=false; is_converted_to_bool_type=false;
	if(v.p.get_actual_type()->is_expr())
		x.assign(v.to_GeneralExpr().to_expr());
	else if(v.p.get_actual_type()->is_ptr())
		x.assign(v.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

LogExpr::LogExpr(const ParamPlusPlus &v)
{
	v.set_exported();
	negated=false; is_converted_to_bool_type=false;
	if(v.v.is_expr())
		x.assign(v.v);
	else if(v.v.is_ptr())
		x.assign(v.v.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

LogExpr::LogExpr(const ParamMinusMinus &v)
{
	v.set_exported();
	negated=false; is_converted_to_bool_type=false;
	if(v.v.is_expr())
		x.assign(v.v);
	else if(v.v.is_ptr())
		x.assign(v.v.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

LogExpr::LogExpr(const LocalPlusPlus &v)
{
	v.set_exported();
	negated=false; is_converted_to_bool_type=false;
	if(v.v.is_expr())
		x.assign(v.v);
	else if(v.v.is_ptr())
		x.assign(v.v.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

LogExpr::LogExpr(const LocalMinusMinus &v)
{
	v.set_exported();
	negated=false; is_converted_to_bool_type=false;
	if(v.v.is_expr())
		x.assign(v.v);
	else if(v.v.is_ptr())
		x.assign(v.v.to_PointerExpr().expression_yielding_address());
	else
		report_error("Illegal types for GeneralExpr");
	set_not_exported();
}

PointerOrElementaryType::PointerOrElementaryType(const local& x)
{
	if(x.is_expr())
		{ ex.assign(x); type=0; }
	else if(x.is_ptr())
		construct(x.to_PointerExpr());
	else
		report_error("Illegal type for function call");
	set_not_exported();
}

PointerOrElementaryType::PointerOrElementaryType(const param& x)
{
	if(x.is_expr())
		{ ex.assign(x); type=0; }
	else if(x.is_ptr())
		construct(x.to_PointerExpr());
	else
		report_error("Objects of type ARRAY or STRUCT cannot be used as function parameters. You may use a pointer to such a type.");
	set_not_exported();
}

PointerOrElementaryType::PointerOrElementaryType(const Variable& x)
{
	if(x.is_expr())
		{ ex.assign(x); type=0; }
	else if(x.is_ptr())
		construct(x.to_PointerExpr());
	else
		report_error("Objects of type ARRAY or STRUCT cannot be used as function parameters. You may use a pointer to such a type.");
	set_not_exported();
}

PointerOrElementaryType::PointerOrElementaryType(const Constant& x)
{
	if(x.is_expr())
		{ ex.assign(x); type=0; }
	else if(x.is_ptr())
		construct(x.to_PointerExpr());
	else
		report_error("Objects of type ARRAY or STRUCT cannot be used as function parameters. You may use a pointer to such a type.");
	set_not_exported();
}

template<class T>
void ExprObjectDestructionHandler(T& ex)
{
	ex.set_exported();
	if(cc->ProgramStructureStack.size()>=2 && 0!=dynamic_cast<ProgramStructure_ForLoop*>(cc->ProgramStructureStack.back()))
	{
		ProgramStructure_ForLoop *p=static_cast<ProgramStructure_ForLoop*>(cc->ProgramStructureStack.back());
		if(!p->init_occured || (p->dofor_loop && !p->incr_occured))
		{
			expr ex2(ex.to_expr());
			if(ex2.code.size()!=0)
				ex2.addcmd(bi::POP, -1);
			ex2.set_exported();
			p->add(ex2);
		}
		else
		{
			cc->ProgramStructureStack.pop_back();
			cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<T>(ex,
				p->dofor_loop ? ProgramStructure_IF_BASE::dofor_loop : ProgramStructure_IF_BASE::for_loop,
				p->init_code, p->local_variable_count,
				p->uninit_vars_created_local_stack_offset, p->uninit_vars_created_types,
				p->incr_code));
		}
	}
	else
		ex.to_expr();	// Beim Aufruf des Destruktors des erzeugten temporren Objekts wird ExprObjectDestructionHandler<expr>()
			// aufgerufen.
}
template<>
void ExprObjectDestructionHandler<expr>(expr& ex)
{
	ex.set_exported();
	long sumsp=0;
	for(nlist<Command>::iterator it=ex.code.begin(); it!=ex.code.end(); it++)
		sumsp+=it->sp;
	bool control_command=(0==sumsp);
	if(cc->ProgramStructureStack.size()>=2 && 0!=dynamic_cast<ProgramStructure_ForLoop*>(cc->ProgramStructureStack.back()))
	{
		if(control_command)
			report_error("Unexpected statement between FOR and DO or DOFOR and DO.");
		ProgramStructure_ForLoop *p=static_cast<ProgramStructure_ForLoop*>(cc->ProgramStructureStack.back());
		if(!p->init_occured || (p->dofor_loop && !p->incr_occured))
		{
			if(ex.code.size()!=0)
				ex.addcmd(bi::POP, -1);
			p->add(ex);
		}
		else
		{
			cc->ProgramStructureStack.pop_back();
			cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<LogExpr>(LogExpr(ex),
				p->dofor_loop ? ProgramStructure_IF_BASE::dofor_loop : ProgramStructure_IF_BASE::for_loop,
				p->init_code, p->local_variable_count,
				p->uninit_vars_created_local_stack_offset, p->uninit_vars_created_types,
				p->incr_code));
			if(0==cc->forward_goto_targets_of_actual_function.size())
				for(nlist<const local*>::const_iterator it=p->init_code.assignment_to_locals.begin(); it!=p->init_code.assignment_to_locals.end(); it++)
					(*it)->assignment_check_var_mem_initialized();
			// Fr die Elemente der Liste p->incr_code.assignment_to_locals sowie fr den Krper einer do-while-Schleife
			// (und erst recht fr eine while-Schleife) darf nicht ohne weiteres assignment_check_var_mem_initialized()
			// aufgerufen werden, weil die zugehrenden Anweisungen dann ggf. nicht ausgefhrt werden, wenn BREAK oder CONTINUE
			// Befehle innerhalb der Schleife auftauchen.
		}
	}
	else if(cc->ProgramStructureStack.size()<2 || 0==dynamic_cast<ProgramStructureGlobalScope*>(cc->ProgramStructureStack.front()) ||
		(0==dynamic_cast<ProgramStructure_Function*>(*(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin()))) &&
		0==dynamic_cast<ProgramStructure_VoidFunction*>(*(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin())))))
		report_error("Code occured outside of BEGIN_GLOBAL_SCOPE ... END_GLOBAL_SCOPE or outside of a function.");
	else
	{
		if(!control_command && ex.code.size()!=0)
		{
			if(!ex.has_code_effect())
				report_warning("Code has no effect.");
			Command cmd=ex.code.back();
			if(cmd.fkt==Command::command)
			{
				uint8 op=cmd.get_opcode();
				if(bi::STORE==op)
					{ ex.code.back().replace_opcode(bi::STOREPOP); ex.code.back().sp--; }
				else if(bi::GETLOCALSTORE==op)
					{ ex.code.back().replace_opcode(bi::GETLOCALSTOREPOP); ex.code.back().sp--; }
				else if(bi::SETLOCAL==op)
					{ ex.code.back().replace_opcode(bi::SETLOCALPOP); ex.code.back().sp--; }
				else if(bi::ADDM==op)
					{ ex.code.back().replace_opcode(bi::ADDMPOP); ex.code.back().sp--; }
				else if(bi::SUBM==op)
					{ ex.code.back().replace_opcode(bi::SUBMPOP); ex.code.back().sp--; }
				else if(bi::ADDCLOCAL==op || bi::LOCALADDC==op)
					{ ex.code.back().replace_opcode(bi::ADDCLOCALPOP); ex.code.back().sp--; }
				else if(bi::SUBM==op)
					{ ex.code.back().replace_opcode(bi::SUBMPOP); ex.code.back().sp--; }
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
				else if(bi::SETLOCALC==op)
					{ ex.code.back().replace_opcode(bi::SETLOCALCPOP); ex.code.back().sp--; }
#endif
				else
					ex.addcmd(bi::POP, -1);
			}
			else if(cmd.fkt==Command::set_prev_var)
				{ ex.code.back().fkt=Command::set_prev_var_and_pop; ex.code.back().sp--; }
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
			else if(cmd.fkt==Command::set_prev_var_const)
				{ ex.code.back().fkt=Command::set_prev_var_const_pop; ex.code.back().sp--; }
#endif
			else if(cmd.fkt==Command::get_prev_var_and_store)
				{ ex.code.back().fkt=Command::get_prev_var_and_storepop; ex.code.back().sp--; }
			else if(cmd.fkt==Command::addc_to_prev_var || cmd.fkt==Command::prev_var_addc)
				{ ex.code.back().fkt=Command::addc_to_prev_var_pop; ex.code.back().sp--; }
			else
				ex.addcmd(bi::POP, -1);
		}
		cc->ProgramStructureStack.back()->add(ex);
		if(0==cc->forward_goto_targets_of_actual_function.size())
			for(nlist<const local*>::const_iterator it=ex.assignment_to_locals.begin(); it!=ex.assignment_to_locals.end(); it++)
				(*it)->assignment_check_var_mem_initialized();
	}
}

template<>
void ExprObjectDestructionHandler<Statement>(Statement& ex)
{
	ex.set_exported();
	long sumsp=0;
	for(nlist<Command>::iterator it=ex.code.begin(); it!=ex.code.end(); it++)
		sumsp+=it->sp;
	bool control_command=(0==sumsp);
	if(cc->ProgramStructureStack.size()>=2 && 0!=dynamic_cast<ProgramStructure_ForLoop*>(cc->ProgramStructureStack.back()))
	{
		if(control_command)
			report_error("Unexpected statement between FOR and DO or DOFOR and DO.");
		ProgramStructure_ForLoop *p=static_cast<ProgramStructure_ForLoop*>(cc->ProgramStructureStack.back());
		if(!p->init_occured || (p->dofor_loop && !p->incr_occured))
			p->add(ex);
		else
		{
			cc->ProgramStructureStack.pop_back();
			cc->ProgramStructureStack.push_back(new ProgramStructure_IF_Spec<LogExpr>(LogExpr(ex),
				p->dofor_loop ? ProgramStructure_IF_BASE::dofor_loop : ProgramStructure_IF_BASE::for_loop,
				p->init_code, p->local_variable_count,
				p->uninit_vars_created_local_stack_offset, p->uninit_vars_created_types,
				p->incr_code));
		}
	}
	else if(cc->ProgramStructureStack.size()<2 || 0==dynamic_cast<ProgramStructureGlobalScope*>(cc->ProgramStructureStack.front()) ||
		(0==dynamic_cast<ProgramStructure_Function*>(*(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin()))) &&
		0==dynamic_cast<ProgramStructure_VoidFunction*>(*(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin())))))
		report_error("Code occured outside of BEGIN_GLOBAL_SCOPE ... END_GLOBAL_SCOPE or outside of a function.");
	else
		cc->ProgramStructureStack.back()->add(ex);
}

void label_handler(const char *identifier)
{
	if(cc->ProgramStructureStack.size()<2 || 0==dynamic_cast<ProgramStructureGlobalScope*>(cc->ProgramStructureStack.front()) ||
			(0==dynamic_cast<ProgramStructure_Function*>(*(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin()))) &&
			0==dynamic_cast<ProgramStructure_VoidFunction*>(*(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin())))))
		report_error("label defined outside of BEGIN_GLOBAL_SCOPE ... END_GLOBAL_SCOPE or outside of a function.");
	Command cmd;
	cmd.fkt=Command::label;
	cmd.identifier=identifier;
	nlist<ProgramStructureElement*>::iterator it=(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin()));
	cmd.branch1_id=(0!=dynamic_cast<ProgramStructure_Function*>(*it) ? static_cast<ProgramStructure_Function*>(*it)->func.uid :
		static_cast<ProgramStructure_VoidFunction*>(*it)->func.uid);
	for(nlist<nstring>::iterator it=cc->forward_goto_targets_of_actual_function.begin(); it!=cc->forward_goto_targets_of_actual_function.end(); )
	{
		nlist<nstring>::iterator actit=it;
		++it;
		if(!actit->compare(identifier))
			cc->forward_goto_targets_of_actual_function.erase(actit);
	}
	cc->defined_goto_labels_of_actual_function.push_back(identifier);
	expr x(cmd);	// Beim Aufruf des expr-Destruktors wird der expr eingefgt.
}

#define BYTECODE_TARGETED_COMPILER__label(identifier) BytecodeTargetedCompiler::label_handler(#identifier)

void goto_handler(const char *identifier)
{
	if(cc->ProgramStructureStack.size()<2 || 0==dynamic_cast<ProgramStructureGlobalScope*>(cc->ProgramStructureStack.front()) ||
		(0==dynamic_cast<ProgramStructure_Function*>(*(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin()))) &&
		0==dynamic_cast<ProgramStructure_VoidFunction*>(*(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin())))))
		report_error("GOTO occured outside of BEGIN_GLOBAL_SCOPE ... END_GLOBAL_SCOPE or outside of a function.");
	Command cmd;
	cmd.fkt=Command::branch_to_label;
	cmd.identifier=identifier;
	nlist<ProgramStructureElement*>::iterator it=(++nlist<ProgramStructureElement*>::iterator(cc->ProgramStructureStack.begin()));
	cmd.branch1_id=(0!=dynamic_cast<ProgramStructure_Function*>(*it) ? static_cast<ProgramStructure_Function*>(*it)->func.uid :
		static_cast<ProgramStructure_VoidFunction*>(*it)->func.uid);
	bool found=false;
	for(nlist<nstring>::iterator it=cc->defined_goto_labels_of_actual_function.begin(); it!=cc->defined_goto_labels_of_actual_function.end(); it++)
		if(!it->compare(identifier))
			{ found=true; break; }
	if(!found)
		cc->forward_goto_targets_of_actual_function.push_back(identifier);
	expr x(cmd);	// Beim Aufruf des expr-Destruktors wird der expr eingefgt.
}
#define BYTECODE_TARGETED_COMPILER__GOTO(identifier) BytecodeTargetedCompiler::goto_handler(#identifier)

int32 SIZEOF(bool x) { return sizeof(x); }
int32 SIZEOF(signed char x) { return sizeof(x); }
int32 SIZEOF(unsigned char x) { return sizeof(x); }
int32 SIZEOF(short x) { return sizeof(x); }
int32 SIZEOF(unsigned short x) { return sizeof(x); }
int32 SIZEOF(int x) { return sizeof(x); }
int32 SIZEOF(unsigned int x) { return sizeof(x); }
int32 SIZEOF(long x) { return sizeof(x); }
int32 SIZEOF(unsigned long x) { return sizeof(x); }
int32 SIZEOF(long long x) { return sizeof(x); }
int32 SIZEOF(unsigned long long x) { return sizeof(x); }
int32 SIZEOF(float x) { return sizeof(x); }
int32 SIZEOF(double x) { return sizeof(x); }

int32 SIZEOF(const Type& type)
{
	return type.size();
}

int32 SIZEOF(const CollectGeneralType& ex)
{
	if(ex.is_expr_type())
		return BasicType::get_BasicType_from_type_nr(ex.to_expr().get_type()).size();
	if(ex.is_PointerExpr())
		return ex.to_PointerExpr().get_actual_type()->size();
	report_error("Unknown type of CollectGeneralType");
}

int32 SIZEOF(const local& v)
{
	return v.type->size();
}

int32 SIZEOF(const Variable& v)
{
	return v.type->size();
}

int32 SIZEOF(const Constant& v)
{
	return v.type->size();
}

int32 SIZEOF(const param& v)
{
	return v.type->size();
}

void InitializeGlobalScope()
{
	Command::lowest_unused_uid=Command::min_not_reserved_uid;
	cc->ActualStructDefinition=0;
	cc->RegisterType->cleanup(); // wichtig, weil whrend des ersten Compilerdurchlaufs in RegisterType z.B. Zeiger auf STRUCT-Definitionsobjekte
		// (StructType) abgespeichert werden, welche jedoch sptestens mit END_GLOBAL_SCOPE gelscht werden und somit zu Beginn
		// des zweiten Compilerdurchlaufs nicht mehr existieren.
	cc->StringConstants->clear();

	DataInitialization::end_marker_occured=false;

	for(long i=0; i<Value::anz_user_types+1; i++)
	{
		Constant::init_values[i].clear();
		Variable::init_values[i].clear();
	}
	for(long i=0; i<3; i++)
		Variable::type_info[i].clear();

	if(0==cc->CompilerPassNr)
	{
		for(int i=0; i<3*Value::anz_user_types+3; i++)
			Variable::var_mem_count[i]=0;
	}
	else
	{
		VARIABLE_INDEX_TYPE addr=0;
		for(int i=0; i<3*Value::anz_user_types+3; i++)
		{
			VARIABLE_INDEX_TYPE addr_vor=addr;
			addr+=Variable::var_mem_count[i];
			Variable::var_mem_count[i]=Variable::var_mem_begin[i]=addr_vor;
			//printf("vb[%u]=%u\n", i, Variable::var_mem_begin[i]);
		}
	}

	// Im Folgenden wird an Adresse 0 eine __VOID-Variable angelegt, um eine Reservierung fr den Nullzeiger vorzunehmen.
	const long t2=0, t=(Value::anz_user_types+1)*t2+Value::t_void;
	Variable::var_mem_count[t]++;
	if(1==cc->CompilerPassNr)
		Variable::init_values[Value::t_void].push_back(2); // Adresse Null belegen.
}

int add_branch_distance(Command &cmd, int32 distance, bool force_wide_distance, bool _signed)
{
	if(_signed)
	{
		if(!force_wide_distance && 2*(distance>=0 ? distance : (-distance)) - (distance<0 ? 1 : 0) < 255)
		{
			cmd.add((uint8)(2*(distance>=0 ? distance : (-distance)) - (distance<0 ? 1 : 0)));
			return 2;
		}
		else
		{
			cmd.add(255);
#if MAX_NR_BYTES_FOR_PROGRAM_COUNTER==2
			cmd.add16((uint16)(2*(distance>=0 ? distance : (-distance)) - (distance<0 ? 1 : 0)));
			return 6;
#else
			cmd.add32((uint32)(2*(distance>=0 ? distance : (-distance)) - (distance<0 ? 1 : 0)));
			return 10;
#endif
		}
	}
	else
	{
		if(!force_wide_distance && distance<256)
		{
			cmd.add((uint8)(distance));
			return 2;
		}
		else
		{
			cmd.add(255);
#if MAX_NR_BYTES_FOR_PROGRAM_COUNTER==2
			cmd.add16((uint16)(distance));
			return 6;
#else
			cmd.add32((uint32)(distance));
			return 10;
#endif
		}
	}
}

typedef ListAndTree<Command, Command_Compare> CommandListType;

uint32 get_sp_before_command(CommandListType& code, CommandListType::iterator it)
{
	if(it==code.begin() || it==0)
		return 0;
	return (--CommandListType::iterator(it))->sp;
}

CommandListType::iterator find_branch_target(CommandListType& code, Command::uid_type label_uid, const char *label_identifier, bool& error,
												CommandListType::iterator search_start=CommandListType::iterator())
{
	error=false;
	bool subordinate1=false, subordinate2=false;
	uint32 sub1sp_vor, sub2sp_vor;
	uint32 sp=0;
	if(search_start!=0)
		sp=search_start->sp;
	CommandListType::iterator it, temp;
	if(search_start!=0)
		it=search_start;
	else
		it=code.begin(); // search_start muss 0 bleiben !
	while(search_start!=0 || it!=code.end())
	{
		if(search_start!=0)
		{
			if(search_start==code.begin())
				search_start=CommandListType::iterator(0);
			else
			{
				/*printf("*** uid=%d   sp=%d   sp_vor=%d\n", search_start->get_uid(), search_start->sp, get_sp_before_command(code, search_start));
				temp=search_start; --temp;
				printf("### uid=%d   sp=%d   sp_vor=%d\n\n", temp->get_uid(), temp->sp, get_sp_before_command(code, temp));*/

				temp=search_start;
				if(get_sp_before_command(code, --temp) > get_sp_before_command(code, search_start) && !subordinate1)
				{
					//printf("==> %d  %d\n", get_sp_before_command(code, temp) , get_sp_before_command(code, search_start));
					subordinate1=true;
					sub1sp_vor=get_sp_before_command(code, search_start);
				}
				else if(subordinate1 && get_sp_before_command(code, temp)<=sub1sp_vor)
					subordinate1=false;

				if((--search_start)->fkt==Command::label && search_start->branch1_id==label_uid &&
						!strcmp(search_start->identifier, label_identifier))
				{
					if(subordinate1) error=true;
					return ++search_start;
				}
			}
		}
		if(it!=code.end())
		{
			if(it->fkt==Command::label && it->branch1_id==label_uid && !strcmp(it->identifier, label_identifier))
			{
				if(subordinate2) error=true;
				return ++it;
			}
			temp=it;
			if(get_sp_before_command(code, ++temp) > get_sp_before_command(code, it) && !subordinate2)
			{
				subordinate2=true;
				sub2sp_vor=get_sp_before_command(code, it);
			}
			else if(subordinate2 && get_sp_before_command(code, temp)<=sub2sp_vor)
				subordinate2=false;
			++it;
		}
	}
	return CommandListType::iterator();
}

nlist<uint8> make_nibblelist(void *memory, uint32 size)
{
	uint8 *mem=(uint8*)memory;
	nlist<uint8> lst;
	while(size--)
	{
		lst.push_back(((*mem)>>4)&15);
		lst.push_back((*mem)&15);
		++mem;
	}
	return lst;
}

nlist<uint8> make_nibblelist(const nlist<uint8>& lst)
{
	nlist<uint8> l;
	for(nlist<uint8>::const_iterator it=lst.begin(); it!=lst.end(); it++)
	{
		uint8 x=*it;
		l.push_back((x>>4)&15);
		l.push_back(x&15);
	}
	return l;
}

nlist<uint8> make_nibblelist(const nlist<uint16>& lst)
{
	nlist<uint8> l;
	for(nlist<uint16>::const_iterator it=lst.begin(); it!=lst.end(); it++)
	{
		uint16 x=*it;
		for(uint8 i=0; i<2 ; i++)
		{
			l.push_back((x>>(8*i+4))&15);
			l.push_back((x>>(8*i))&15);
		}
	}
	return l;
}

nlist<uint8> make_nibblelist(const nlist<uint32>& lst)
{
	nlist<uint8> l;
	for(nlist<uint32>::const_iterator it=lst.begin(); it!=lst.end(); it++)
	{
		uint32 x=*it;
		for(uint8 i=0; i<4 ; i++)
		{
			l.push_back((x>>(8*i+4))&15);
			l.push_back((x>>(8*i))&15);
		}
	}
	return l;
}

struct OpcodeOrMakro
{
	uint8 opcode;
	Command::uid_type uid;
	uint32 nr_used;
	uint64 code;
	uint8 nr_codenibbles;
};

struct OpcodeOrMakro_Compare : public BaumKeineAttrBerechnen<OpcodeOrMakro>
{
	bool less(const OpcodeOrMakro& c1, const OpcodeOrMakro& c2) const
	{
		Command::uid_type uid1=(0==c1.uid) ? c1.opcode : (c1.uid+256);
		Command::uid_type uid2=(0==c2.uid) ? c2.opcode : (c2.uid+256);
		return uid1<uid2;
	}
	bool equal(const OpcodeOrMakro& c1, const OpcodeOrMakro& c2) const
	{
		Command::uid_type uid1=(0==c1.uid) ? c1.opcode : (c1.uid+256);
		Command::uid_type uid2=(0==c2.uid) ? c2.opcode : (c2.uid+256);
		return uid1==uid2;
	}
};

struct FrequencySortedOpcode
{
	OpcodeOrMakro *op;
	uint32 nr_used; // nur gltig, wenn op==0
	nlist<FrequencySortedOpcode> children;
	uint8 codenibble;
	void calc_nr_used()
	{
		if(0==op)
		{
			nr_used=0;
			for(nlist<FrequencySortedOpcode>::iterator it=children.begin(); it!=children.end(); it++)
				nr_used+=it->nr_used;
		}
		else
			nr_used=op->nr_used;
	}
};

struct FrequencySortedOpcode_Compare : public BaumKeineAttrBerechnen<FrequencySortedOpcode>
{
	bool less(const FrequencySortedOpcode& c1, const FrequencySortedOpcode& c2) const
	{
		return (0==c1.op ? c1.nr_used : c1.op->nr_used) < (0==c2.op ? c2.nr_used : c2.op->nr_used);
	}
	bool equal(const FrequencySortedOpcode& c1, const FrequencySortedOpcode& c2) const
	{
		return (0==c1.op ? c1.nr_used : c1.op->nr_used) == (0==c2.op ? c2.nr_used : c2.op->nr_used);
	}
};

struct CountFunctionCalls
{
	Command::uid_type function_header_uid, function_uid;
	uint32 nr_used;
	uint8 alignment;
	uint8 function_nr;
};

struct CountFunctionCalls_Compare : public BaumKeineAttrBerechnen<CountFunctionCalls>
{
	bool less(const CountFunctionCalls& f1, const CountFunctionCalls& f2) const
	{
		return f1.function_header_uid<f2.function_header_uid;
	}
	bool equal(const CountFunctionCalls& f1, const CountFunctionCalls& f2) const
	{
		return f1.function_header_uid==f2.function_header_uid;
	}
};

struct FrequencySortedFunctioncall
{
	CountFunctionCalls *func;
};

struct FrequencySortedFunctioncall_Compare : public BaumKeineAttrBerechnen<FrequencySortedFunctioncall>
{
	bool less(const FrequencySortedFunctioncall& f1, const FrequencySortedFunctioncall& f2) const
	{
		return f1.func->nr_used<f2.func->nr_used;
	}
	bool equal(const FrequencySortedFunctioncall& f1, const FrequencySortedFunctioncall& f2) const
	{
		return f1.func->nr_used==f2.func->nr_used;
	}
};

int32 get_listpos(CommandListType& lst, CommandListType::iterator it)
{
	uint32 pos=0;
	if(it==lst.end())
		return lst.size();
	for(CommandListType::iterator it2=lst.begin(); it2!=lst.end(); it2++, pos++)
		if(it2==it)
			return pos;
	return -1;
}

nstring code_element_to_string(CommandListType::iterator it)
{
	switch(it->fkt)
	{
	case Command::command:
		return opcode_to_command_string(it->get_opcode());
		break;
	case Command::functioncall:
		return "function_call";
		break;
	case Command::function:
		return "function";
		break;
	case Command::begin_infos_before_function:
		return "begin_infos_before_function";
		break;
	case Command::raw_nibbles:
		return "raw_nibbles";
		break;
	case Command::makrocall:
		return "makro_call";
		break;
	default:
		return printf_string("Fkt=%u\n", it->fkt);
		break;
	}
}

struct EncodeNibblesUsingEscapeSequences
{
private:
	void new_coded_nibble(uint8 nibble)
	{
		if(!only_simulate)
			cd.push_back(nibble);
		simulate_nibblesize++;
	}
	void push_back_special(uint8 nibble, int param_size=0)
	{	// nibble muss bi::makro_escape_ret, bi::makro_escape_fixedsize_param oder bi::makro_escape_varsize_param sein.
		// param_size wird nur im Fall nibble==bi::makro_escape_fixedsize_param benutzt.
		if(last_nibble_was_escape)
		{
			new_coded_nibble(bi::makro_escape_nibble);
			new_coded_nibble(bi::makro_escape_nibble);
			last_nibble_was_escape=false;
		}
		new_coded_nibble(bi::makro_escape_nibble);
		new_coded_nibble(nibble);
		if(bi::makro_escape_fixedsize_param==nibble)
		{
			if(0==param_size)
				report_error("Error in EncodeNibblesUsingEscapeSequences::push_back_makro_param_call_fixedsize(): param_size must not be zero.");
			if(param_size<16)
				new_coded_nibble(nibble);
			else
				new_coded_nibble(0);
			{
				PROGRAM_COUNTER_TYPE x=param_size;
				for(long i=0; i<MAX_NR_BYTES_FOR_PROGRAM_COUNTER*2; i++)
				{
					new_coded_nibble(x&15);
					x>>=4;
				}
			}
		}
	}
	bool only_simulate, save_only_simulate;
	unsigned long simulate_nibblesize, save_simulate_nibblesize;
	bool use_escape_sequences, save_use_escape_sequences;
public:
	nlist<uint8> cd;
	uint32 save_cd_size;
	bool last_nibble_was_escape, save_last_nibble_was_escape;
	EncodeNibblesUsingEscapeSequences()
		{ last_nibble_was_escape=false; only_simulate=false; simulate_nibblesize=0; use_escape_sequences=false; }
	void set_only_simulate() { only_simulate=true; }
	void enable_encoding(bool __enable) { use_escape_sequences=__enable; }
	void push_back(uint8 nibble)
	{
		if(last_nibble_was_escape)
		{
			last_nibble_was_escape=false;
			if(nibble!=bi::makro_escape_escape && nibble!=bi::makro_escape_ret && nibble!=bi::makro_escape_fixedsize_param &&
					nibble!=bi::makro_escape_varsize_param)
			{
				new_coded_nibble(bi::makro_escape_nibble);
				new_coded_nibble(nibble);
			}
			else
			{
				new_coded_nibble(bi::makro_escape_nibble);
				new_coded_nibble(bi::makro_escape_nibble);
				if(!use_escape_sequences || bi::makro_escape_nibble!=nibble)
					new_coded_nibble(nibble);
				else
					last_nibble_was_escape=true;
			}
		}
		else if(!use_escape_sequences || bi::makro_escape_nibble!=nibble)
			new_coded_nibble(nibble);
		else
			last_nibble_was_escape=true;
	}
	void push_back_makro_ret()
		{ push_back_special(bi::makro_escape_ret); }
	void push_back_makro_param_call_fixedsize(int param_size)
		{ push_back_special(bi::makro_escape_fixedsize_param, param_size); }
	void push_back_makro_param_call_varsize()
		{ push_back_special(bi::makro_escape_varsize_param); }
	void push_back_nibblelist(nlist<uint8> lst)
	{
		for(nlist<uint8>::iterator it=lst.begin(); it!=lst.end(); it++)
			push_back(*it);
	}
	unsigned long nibblesize(uint8 next_nibble, bool next_is_special_function=false, bool no_escape_nibble=false) const
	{	// Es wird die Codegre VOR dem Einfgen des nchsten Nibbles (next_nibble) geliefert. Falls jedoch das
		// vorangehende Nibble gleich bi::makro_escape_nibble war, kann nur mit Zuhilfenahme des Nibbles 'next_nibble' entschieden
		// werden, ob das vorangehende Nibble durch zwei Nibbles (bi::makro_escape_nibble, bi::makro_escape_escape) codiert
		// werden muss.
		if(!last_nibble_was_escape)
			return only_simulate ? simulate_nibblesize : cd.size();
		if(no_escape_nibble)
			return only_simulate ? (simulate_nibblesize+1) : (cd.size()+1);
		return (next_is_special_function || next_nibble==bi::makro_escape_nibble || next_nibble==bi::makro_escape_ret ||
			next_nibble==bi::makro_escape_fixedsize_param || next_nibble==bi::makro_escape_varsize_param) ?
			(only_simulate ? (simulate_nibblesize+2) : (cd.size()+2)) : (only_simulate ? (simulate_nibblesize+1) : (cd.size()+1));
	}
	bool will_previous_nibble_be_coded_with_escape_escape(uint8 next_nibble, bool next_is_special_function=false) const
	{
		if(!last_nibble_was_escape)
			return false;;
		return next_is_special_function || next_nibble==bi::makro_escape_nibble || next_nibble==bi::makro_escape_ret ||
			next_nibble==bi::makro_escape_fixedsize_param || next_nibble==bi::makro_escape_varsize_param;
	}
	void increase_simulated_size(uint32 anz_nibbles_to_add)
	{		// Im Zweifelsfall, d.h. wenn last_nibble_was_escape==true ist, wird eher ein zu groer Wert fr die Gre erzeugt.
		if(!only_simulate)
			return;
		if(last_nibble_was_escape)
		{
			simulate_nibblesize+=2; // Escape-Nibble + Nutz-Nibble
			last_nibble_was_escape=false;
		}
		simulate_nibblesize+=anz_nibbles_to_add;
	}
	void simulate_known_number_of_unknown_nibbles(uint32 anz_uncoded_nibbles)
	{
		increase_simulated_size(2*anz_uncoded_nibbles); // Im ungnstigsten Fall betrgt die Gre der codierten Nibbles
			// das Doppelte der Gre der uncodierten Nibbles
	}
	void reset_simulated_size()
	{
		simulate_nibblesize=0;
	}
	void set_simulated_size(uint32 size)
	{
		simulate_nibblesize=size;
	}
	void flush()
	{
		if(last_nibble_was_escape)
		{
			new_coded_nibble(bi::makro_escape_nibble);
			new_coded_nibble(bi::makro_escape_escape);
			last_nibble_was_escape=false;
		}
	}
	void store_state()
	{
		save_only_simulate=only_simulate;
		save_simulate_nibblesize=simulate_nibblesize;
		save_use_escape_sequences=use_escape_sequences;
		save_cd_size=cd.size();
		save_last_nibble_was_escape=last_nibble_was_escape;
	}
	void restore_state()
	{
		only_simulate=save_only_simulate;
		simulate_nibblesize=save_simulate_nibblesize;
		use_escape_sequences=save_use_escape_sequences;
		last_nibble_was_escape=save_last_nibble_was_escape;
		while(cd.size()>save_cd_size)
			cd.erase(--cd.end());
	}
};

uint8 get_first_code_nibble_of_command(CommandListType::iterator it, RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare> &opcodes)
{	// Wenn it->fkt==Command::begin_infos_before_function oder it->fkt==Command::function oder it->fkt==Command::end_of_code ist,
	// wird nicht das erste Codenibble dieses oder des folgenden
	// Kommandos geliefert, sondern ein Nibble, welches verschieden von bi::makro_escape_nibble, bi::makro_escape_ret,
	// bi::makro_escape_fixedsize_param und =bi::makro_escape_varsize_param ist.
	if(Command::command==it->fkt)
	{
		uint8 opcode=(it->code[0]<<4) | it->code[1];
		OpcodeOrMakro op; op.opcode=opcode; op.uid=0; op.nr_used=0;
		RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
		if(it2==0 || 0==it2->nr_codenibbles)
			report_error("Internal compiler error: Error in assemble_code()");
		return (it2->code>>(4*(it2->nr_codenibbles-1)))&15;
	}
	else switch(it->fkt)
	{
		case Command::functioncall:
		{
			OpcodeOrMakro op; op.opcode=bi::CALL; op.uid=0; op.nr_used=0;
			RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
			if(it2==0 || 0==it2->nr_codenibbles)
				report_error("Internal compiler error: Error in assemble_code()");
			return (it2->code>>(4*(it2->nr_codenibbles-1)))&15;
		}
		case Command::begin_infos_before_function:
		case Command::function:
		case Command::end_of_code:
		{
			for(long i=0; i<16; i++)
				if(i!=bi::makro_escape_nibble && i!=bi::makro_escape_ret && i!=bi::makro_escape_fixedsize_param &&
						i!=bi::makro_escape_varsize_param)
					return i;
		}
		case Command::raw_nibbles:
			return it->code[0]; // it->nibblesize>0 vorausgesetzt.
		case Command::makrocall:
		{
			OpcodeOrMakro op; op.opcode=0; op.uid=it->branch1_id; op.nr_used=0;
			RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
			if(it2==0 || 0==it2->nr_codenibbles)
				report_error("Internal compiler error: Error in assemble_code()");
			return (it2->code>>(4*(it2->nr_codenibbles-1)))&15;
		}
		case Command::makro_ret:
		case Command::makro_param_call_fixedsize:
		case Command::makro_param_call_varsize:
			return bi::makro_escape_nibble;
		default:
			report_error("Internal compiler error in function get_first_code_nibble_of_command(): Invalid it->fkt");
	}
}

struct UID_LessEqual : public BaumKeineAttrBerechnen<Command::uid_type>
{
	bool less(const Command::uid_type& c1, const Command::uid_type& c2) const
	{
		return c1<c2;
	}
	bool equal(const Command::uid_type& c1, const Command::uid_type& c2) const
	{
		return c1==c2;
	}
};

struct makro_information
{
	nlist<Command::uid_type> orig_start_uids;
	Command::uid_type start;
	Command::uid_type end;
	uint32 nr_called;
	uint32 orig_makro_sp;
};

int32 find_commandlist_distance(CommandListType& code, CommandListType::iterator search_start, Command::uid_type target_uid)
{
	CommandListType::iterator it=search_start, temp;
	int32 search_start_diff=0, it_diff=0;
	while(search_start!=0 || it!=code.end())
	{
		if(search_start!=0)
		{
			if(search_start==code.begin())
				search_start=CommandListType::iterator(0);
			else
			{
				--search_start_diff;
				if((--search_start)->get_uid()==target_uid)
					return search_start_diff;
			}
		}
		if(it!=code.end())
		{
			if(it->get_uid()==target_uid)
				return it_diff;
			++it;
			++it_diff;
		}
	}
	return 0;
}

void check_for_branches_outside_makro(CommandListType& code, CommandListType::iterator it, CommandListType::iterator& it3,
									  CommandListType::iterator it2, CommandListType::iterator& it4, long& anz)
{
	restart:
	CommandListType::iterator it5, it6, it7;
	anz=0;
	for(it5=it, it6=it2; it5!=it3; it5++, it6++, anz++)
	{
		if(it6==it4)
			report_error("Internal compiler error: check_for_branches_outside_makro()");
		if(Command::command==it5->fkt)
		{
			uint8 op=it5->get_opcode();
			if(bi::BRT==op || bi::BRF==op || bi::BR==op || bi::BRAND==op || bi::BROR==op || bi::BRANDNOT==op || bi::BRORNOT==op)
			{
				for(it7=it; it7!=it3; it7++)
					if(it7->get_uid()==it5->branch1_id)
						break;
				if(it7==it3 && !(it3!=code.end() && it3->get_uid()==it5->branch1_id))
				{
					it3=it5;
					it4=it6;
					goto restart;
				}
			}
		}
		if(it5!=it && 0!=it5->commands_branching_to_this)
		{
			for(nlist<Command::uid_type>::iterator it8=it5->commands_branching_to_this->begin();
					it8!=it5->commands_branching_to_this->end(); it8++)
			{
				for(it7=it; it7!=it3; it7++)
					if(it7->get_uid()==*it8)
						break;
				if(it7==it3)
				{
					it3=it5;
					it4=it6;
					goto restart;
				}
			}
		}
	}
}

struct PUSHC_Command_LessEqual : public BaumKeineAttrBerechnen<Command>
{
	bool less(const Command& c1, const Command& c2) const
	{
		if(c1.fkt!=Command::command || c2.fkt!=Command::command)
			report_error("Internal compiler error: assemble_code()");
		uint32 i;
		for(i=0; i<c1.nibblesize && i<c2.nibblesize; i++)
			if(c1.code[i]<c2.code[i])
				return true;
			else if(c1.code[i]>c2.code[i])
				return false;
		return i<c2.nibblesize;
	}
	bool equal(const Command& c1, const Command& c2) const
	{
		if(c1.fkt!=Command::command || c2.fkt!=Command::command)
			report_error("Internal compiler error: assemble_code()");
		uint32 i;
		for(i=0; i<c1.nibblesize && i<c2.nibblesize; i++)
			if(c1.code[i]!=c2.code[i])
				return false;
		return i==c1.nibblesize && i==c2.nibblesize;
	}
};

uint8 assumed_nibblesize_for_PUSHCT_Command=2;
uint8 assumed_nibblesize_for_PUSHC_Command=2;
uint8 assumed_nibblesize_for_PUSHCS_Command=2;

struct PUSHC_Command_SavingSorted_LessEqual : public BaumKeineAttrBerechnen<std::pair<Command, uint32> >
{
	static uint64 get_integer_constant(const Command& cmd, uint8 *ptype=0)
	{
		uint8 op=cmd.get_opcode(), i=2, type, nibblesize;
		if(bi::PUSHC!=op && bi::PUSHC0!=op && bi::PUSHCS!=op && bi::PUSHC0U8!=op && bi::PUSHCSU8!=op)
			report_error("Internal compiler error: assemble_code()");
		if(bi::PUSHC0U8!=op && bi::PUSHCSU8!=op && cmd.nibblesize<3)
			report_error("Internal compiler error: assemble_code()");

		if(op==bi::PUSHC0 || op==bi::PUSHC0U8)
		{
			if(ptype)
				*ptype=(op==bi::PUSHC0U8) ? Value::t_u8 : cmd.code[2];
			return 0;
		}
		if(op==bi::PUSHCSU8)
		{
			if(ptype)
				*ptype=Value::t_u8;
			return cmd.code[2];
		}
		if(op==bi::PUSHCS && cmd.code[2]<Value::t_i32_but16used)
		{
			if(ptype)
			{
				*ptype=cmd.code[2];
				if(*ptype>=Value::t_i32_but16used)
					*ptype-=(Value::t_i32_but16used-Value::t_i32);
			}
			return cmd.code[3];
		}
	
		type=cmd.code[i++];
		if(op==bi::PUSHCS && type>=Value::t_i32_but16used)
			{ nibblesize=(type>=Value::t_i64_but32used) ? 8 : 4; type-=Value::t_i32_but16used-Value::t_i32; }
		else if(op==bi::PUSHC && Value::t_i16_but8used==type)
			{ type=Value::t_i16; nibblesize=2; }
		else if(op==bi::PUSHC && type>=Value::t_u16_but8used)
			{ type-=Value::t_u16_but8used-Value::t_u16; nibblesize=2; }
		else
			nibblesize=(Value::sbytesize(type)<<1);

		if(ptype)
			*ptype=type;
		if(0==nibblesize)
			return 0;
		uint64 x=0;
		uint8 z=cmd.code[i++];
		if(Value::s_is_signed_integer(type) && 0!=(z&8))
			x=~((uint64)0);
		do { x=(x<<4) | z; } while(0!=--nibblesize && (z=cmd.code[i++], true));
		return x;
	}
	static int32 get_nr_bytes(const Command& c)
	{
		uint8 type;
		uint64 ic=get_integer_constant(c, &type);
		int32 nrbits=expr::get_nr_of_used_bits(ic, Value::sbytesize(type), Value::s_is_signed_integer(type));
		return (nrbits<=8 ? 1 : (nrbits<=16 ? 2 : (nrbits<=32 ? 4 : 8)));
	}
	static int32 get_saved_nibbles(const Command& c)
	{
		int32 ns=((c.get_opcode()==bi::PUSHC) ? assumed_nibblesize_for_PUSHC_Command :
				((c.get_opcode()==bi::PUSHCS) ? assumed_nibblesize_for_PUSHCS_Command : 2)) + c.nibblesize-2;
		return (ns-assumed_nibblesize_for_PUSHCT_Command-2)*((int32)c.nr_used) - (1+get_nr_bytes(c)*2);
	}
	bool less(const std::pair<Command, uint32>& c1, const std::pair<Command, uint32>& c2) const
	{
		if(c1.first.fkt!=Command::command || c2.first.fkt!=Command::command)
			report_error("Internal compiler error: assemble_code()");
		uint32 i;
		if(get_saved_nibbles(c1.first)<get_saved_nibbles(c2.first))
			return true;
		if(get_saved_nibbles(c1.first)>get_saved_nibbles(c2.first))
			return false;
		for(i=0; i<c1.first.nibblesize && i<c2.first.nibblesize; i++)
			if(c1.first.code[i]<c2.first.code[i])
				return true;
			else if(c1.first.code[i]>c2.first.code[i])
				return false;
		return i<c2.first.nibblesize;
	}
	bool equal(const std::pair<Command, uint32>& c1, const std::pair<Command, uint32>& c2) const
	{
		if(c1.first.fkt!=Command::command || c2.first.fkt!=Command::command)
			report_error("Internal compiler error: assemble_code()");
		if(get_saved_nibbles(c1.first)!=get_saved_nibbles(c2.first))
			return false;
		uint32 i;
		for(i=0; i<c1.first.nibblesize && i<c2.first.nibblesize; i++)
			if(c1.first.code[i]!=c2.first.code[i])
				return false;
		return i==c1.first.nibblesize && i==c2.first.nibblesize;
	}
};

nlist<uint8> assemble_code_makro_trial(const Code& orig_code, RotSchwarzBaum<Command::uid_type, UID_LessEqual> &forbidden_makros)
{
	if(0!=cc->ActualStructDefinition) report_error("STRUCT occured without corresponding ENDSTRUCT.");\

	RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare> opcodes;
	EncodeNibblesUsingEscapeSequences cd;
	RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare> count_functioncalls;
	RotSchwarzBaum<Command, PUSHC_Command_LessEqual> constants;

	for(long i=0; i<bi::MAX_OPCODE_NUMBER; i++)
	{
		OpcodeOrMakro opc;
		opc.nr_used=0;
		opc.opcode=i;
		opc.uid=0;
		opc.nr_codenibbles=0;
		opcodes.push_back(opc);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	CommandListType code;
	int32 sp=0, spie=0;
	for(nlist<Command>::const_iterator origit=orig_code.code.begin(); origit!=orig_code.code.end(); origit++)
	{
		uint8 opcode=(origit->code[0]<<4) | origit->code[1];
		
		/*if(origit->fkt==Command::command && (opcode==bi::IFELSE || opcode==bi::IFNOTELSE))
			printf("uid=%d   fkt=%d    opcode=%u    sp=%d    d_sp=%d   bruid1=%d   bruid2=%d\n", origit->get_uid(), origit->fkt, (origit->code[0]<<4) | origit->code[1], sp, origit->sp,
				origit->branch1_id, origit->branch2_id);
		else if(origit->fkt==Command::command && (opcode==bi::BR || opcode==bi::BRT || opcode==bi::BRF || opcode==bi::BRAND || opcode==bi::BROR || opcode==bi::BRANDNOT || opcode==bi::BRORNOT))
			printf("uid=%d   fkt=%d    opcode=%u    sp=%d    d_sp=%d   bruid1=%d\n", origit->get_uid(), origit->fkt, (origit->code[0]<<4) | origit->code[1], sp, origit->sp,
				origit->branch1_id);
		else
			printf("uid=%d   fkt=%d    opcode=%u    sp=%d    d_sp=%d\n", origit->get_uid(), origit->fkt, (origit->code[0]<<4) | origit->code[1], sp, origit->sp);
		*/
		Command cmd=*origit;
		sp = (cmd.sp+=sp);
		spie = (cmd.spie+=spie);
		code.push_back(cmd);
	}
	//printf("sp=%d\n", sp);
	if(0!=sp)
		report_error("Internal Compiler Error: Stack pointer not zero at the end.");

	Command eoc_command;
	eoc_command.fkt=Command::end_of_code;
	code.push_back(eoc_command); // Dies muss eingefgt werden, damit es keinen Fehler gibt, wenn in einer if-Abfrage
		// ber einen branch-Befehl an das Ende des Codes gesprungen wird, wobei zuvor ein RETURN-Statement auftrat.

	long nr_inserted_castbool_for_BRANDOR=0;
	long nr_inserted_branch0_for_BRTF=0;

	for(CommandListType::iterator it=code.begin(); it!=code.end(); it++)
	{	// Im dieser ersten Schleife werden vor Befehle, die effektiv einen Sprung ausfhren, falls erforderlich
		// POPN-Kommandos eingefgt. Die Werte branch1_id und branch2_id drfen in dieser ersten Schleife noch nicht
		// auf jeweils den Befehl gesetzt werden, der der uid-Marken folgt, denn falls das Ziel branch1_id bzw. branch2_id
		// auch einen Sprungbefehl enthlt, dessen ggf. erforderlicher POPN-Befehl erst in einem spteren
		// Durchlauf der Schleife eingefgt wird, wrden sonst die branch1_id bzw. branch2_id nicht wie gewnscht auf den
		// neu erzeugten POPN-Befehl verweisen.
		uint8 opcode=(it->code[0]<<4) | it->code[1];
		if(Command::command==it->fkt)
			switch(opcode)
			{
				case bi::BRT: case bi::BRF: case bi::BR:
				case bi::BRAND: case bi::BROR: case bi::BRANDNOT: case bi::BRORNOT:
				{
					CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
					if(!target.is_valid_and_not_end(code))
						report_error("Internal Compiler Error: Branch to undefined uid.");
					int32 dsp=it->sp - (target==code.begin() ? 0 : ((--(CommandListType::iterator(target)))->sp));
					if(bi::BROR==opcode || bi::BRAND==opcode || bi::BRORNOT==opcode || bi::BRANDNOT==opcode)
						++dsp;
					if(0!=dsp)
					{
						if(bi::BR!=opcode)
							report_error("Internal Compiler Error: Conditional branch to target of other stack deep.");
						if(dsp<0)
							report_error("Internal Compiler Error: Branch to target in subordinate program structure.");
						else if(1==dsp)
							code.insert(it, Command(bi::POP));
						else
						{
							if(dsp<256)
								code.insert(it, expr(uint8(dsp)).set_exported().code.front());
							else
								code.insert(it, expr(VARIABLE_INDEX_TYPE(dsp)).set_exported().code.front());
							code.insert(it, Command(bi::POPN));
						}
					}
					break;
				}
				case bi::BRK: case bi::CONT:
				{
					CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
					if(!target.is_valid_and_not_end(code))
						report_error("Internal Compiler Error: Branch to undefined uid.");
					int32 dsp=it->spie - (target==code.begin() ? 0 : ((--(CommandListType::iterator(target)))->spie));
					if(dsp<0)
						report_error("Internal Compiler Error: Branch to target in subordinate program structure.");
					it->add_coded_integer15(dsp, MAX_NR_BYTES_FOR_VARIABLE_INDEX, false);
					break;
				}
#ifdef USE_TABLE_FOR_CONSTANTS
				case bi::PUSHC:
				case bi::PUSHCS:
					RotSchwarzBaum<Command, PUSHC_Command_LessEqual>::iterator cit=constants.suche(*it);
					if(cit!=0)
						cit->nr_used++;
					else
					{
						Command ccmd(*it);
						ccmd.nr_used=1;
						constants.insert(ccmd);
					}
					break;
#endif
			}
		else
			switch(it->fkt)
			{
				case Command::label:
					for(CommandListType::iterator it2=code.begin(); it2!=code.end(); it2++)
						if(it!=it2 && Command::label==it2->fkt && it->branch1_id==it2->branch1_id &&
								0==strcmp(it->identifier, it2->identifier))
							report_error(printf_string("The label %s is defined more than one time within a function.", it->identifier));
					break;
				case Command::branch_to_label:
				{
					bool error;
					CommandListType::iterator target=find_branch_target(code, it->branch1_id, it->identifier, error, it);
					if(error)
						report_error("Error in GOTO statement: branch to target in subordinate program structure.");
					if(!target.is_valid_and_not_end(code))
						report_error(printf_string("The identifier \'%s\' belonging to the goto statement is undefined.", it->identifier));
					int32 dsp=it->sp - (target==code.begin() ? 0 : ((--(CommandListType::iterator(target)))->sp));
					it->add(bi::BR);
					if(0!=dsp)
					{
						if(dsp<0)
							report_error("Error in GOTO statement: branch to target in subordinate program structure.");
						else if(1==dsp)
							code.insert(it, Command(bi::POP).set_sp(get_sp_before_command(code, it)));
						else
						{
							if(dsp<256)
								code.insert(it, expr(uint8(dsp)).set_exported().code.front().set_sp(get_sp_before_command(code, it)));
							else
								code.insert(it,
									expr(VARIABLE_INDEX_TYPE(dsp)).set_exported().code.front().set_sp(get_sp_before_command(code, it)));
							code.insert(it, Command(bi::POPN).set_sp(get_sp_before_command(code, it)));
						}
					}
					break;
				}
				case Command::get_prev_var:
				case Command::get_prev_var_and_load:
				case Command::get_prev_var_and_store:
				case Command::get_prev_var_and_storepop:
				case Command::set_prev_var:
				case Command::set_prev_var_const:
				case Command::set_prev_var_const_pop:
				case Command::set_prev_var_and_pop:
				case Command::get_prev_var_addr:
				case Command::addc_to_prev_var:
				case Command::prev_var_addc:
				case Command::addc_to_prev_var_pop:
				{
					CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
					if(!target.is_valid_and_not_end(code))
						report_error("Internal compiler error: Reference to undefined uid");
					if(Command::set_prev_var_and_pop==it->fkt || Command::get_prev_var_and_storepop==it->fkt ||
							 Command::set_prev_var_const_pop==it->fkt || Command::addc_to_prev_var_pop==it->fkt)
						++it->sp;
					int32 itsp1=it->sp, itsp=it->sp-1;
					uint64 data;
					switch(it->fkt)
					{
						case Command::get_prev_var:				it->add(bi::GETPREV); break;
						case Command::get_prev_var_and_load:	it->add(bi::GETPREVLOAD); break;
						case Command::get_prev_var_and_store:	it->add(bi::GETPREVSTORE); break;
						case Command::get_prev_var_and_storepop:it->add(bi::GETPREVSTOREPOP); break;
						case Command::get_prev_var_addr:		it->add(bi::STACKADDR); break;
						case Command::set_prev_var:				it->add(bi::SETPREV); break;
						case Command::set_prev_var_and_pop:		it->add(bi::SETPREVPOP); break;
#ifdef WITH_OPCODE_SETPREVC_AND_SETLOCALC
						case Command::set_prev_var_const:
							data=it->var_data; // vor anderen add, weil vardata union mit code[] ist
							it->add(bi::SETPREVC);
							it->add_coded_integer15(data, 0, true);
							break;
						case Command::set_prev_var_const_pop:
							data=it->var_data; // vor anderen add, weil vardata union mit code[] ist
							it->add(bi::SETPREVCPOP);
							it->add_coded_integer15(data, 0, true);
							break;
#endif
						case Command::addc_to_prev_var:
							data=it->var_data; // vor anderen add, weil vardata union mit code[] ist
							it->add(bi::ADDCPREV);
							it->addnibble(((int64)data)>=0 ? (data-1) : (7-((int64)data))); // data==0 ist nicht zulssig
							break;
						case Command::prev_var_addc:
							data=it->var_data; // vor anderen add, weil vardata union mit code[] ist
							it->add(bi::PREVADDC);
							it->addnibble(((int64)data)>=0 ? (data-1) : (7-((int64)data))); // data==0 ist nicht zulssig
							break;
						case Command::addc_to_prev_var_pop:
							data=it->var_data; // vor anderen add, weil vardata union mit code[] ist
							it->add(bi::ADDCPREVPOP);
							it->addnibble(((int64)data)>=0 ? (data-1) : (7-((int64)data))); // data==0 ist nicht zulssig
							break;
					}
					it->add_coded_integer15(itsp-target->sp, MAX_NR_BYTES_FOR_VARIABLE_INDEX, false);
					it->sp=itsp1;
					if(Command::set_prev_var_and_pop==it->fkt || Command::get_prev_var_and_storepop==it->fkt ||
							 Command::set_prev_var_const_pop==it->fkt || Command::addc_to_prev_var_pop==it->fkt)
						--it->sp;
					break;
				}
			}
	}

#ifdef USE_TABLE_FOR_CONSTANTS
	RotSchwarzBaum<std::pair<Command, uint32>, PUSHC_Command_SavingSorted_LessEqual> saving_sorted_constants;
	for(RotSchwarzBaum<Command, PUSHC_Command_LessEqual>::iterator cit=constants.minimizer(); cit!=0; cit++)
		if(PUSHC_Command_SavingSorted_LessEqual::get_saved_nibbles(*cit)>0)
			saving_sorted_constants.insert(std::pair<Command, uint32>(*cit, 0));
	uint32 constant_nr=0, saving=0;
	uint32 bytesize_constant_nr_limit[4]; // bytesize_constant_nr_limit[3] wird spter auch benutzt !
	for(long i=0; i<4; i++)
	{
		int constant_count=0;
		for(RotSchwarzBaum<std::pair<Command, uint32>, PUSHC_Command_SavingSorted_LessEqual>::iterator
				it=saving_sorted_constants.maximizer(); it!=0 && constant_count<256; it--, constant_count++)
			if(PUSHC_Command_SavingSorted_LessEqual::get_nr_bytes(it->first)==(1<<i))
			{
				it->second=constant_nr++;
				saving+=PUSHC_Command_SavingSorted_LessEqual::get_saved_nibbles(it->first);
			}
		bytesize_constant_nr_limit[i]=constant_nr;
		//printf("==> %d\n", constant_nr);
	}
	//printf("Saved: %d\n", saving);
#endif

	// Ab hier darf it->spie (Command::spie) nicht mehr verwendet werden, da es eine union mit it->makro_sp bildet und ab nun
	// statt dessen das Element it->makro_sp verwendet wird.

	for(CommandListType::iterator it=code.begin(); it!=code.end(); it++)
	{
		it->makro_sp=0;
		it->is_makro_param=false;
		it->makro_sp=0;
		uint8 opcode=(it->code[0]<<4) | it->code[1];
		if(Command::command==it->fkt)
			switch(opcode)
			{
				case bi::BRT: case bi::BRF: case bi::BR:
				case bi::BRAND: case bi::BROR: case bi::BRANDNOT: case bi::BRORNOT:
				{
					CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
					while(target.is_valid_and_not_end(code) && (Command::label==target->fkt || Command::uid_label==target->fkt))
						++target; // Command-Objekte mit fkt==Command::label werden spter gelscht
					if(!target.is_valid_and_not_end(code))
						report_error("Internal Compiler Error: Branch to undefined uid.");
					it->branch1_id=target->get_uid();
#ifdef ALLOW_BRANCHES_INSIDE_MAKROS
					if(0==target->commands_branching_to_this)
						target->commands_branching_to_this=new nlist<Command::uid_type>();
					target->commands_branching_to_this->push_back(it->get_uid());
#else
					it->stop_makro=it->dont_use_in_makro=true;
					target->stop_makro=true;
#endif
					break;
				}
				case bi::BRK: case bi::CONT:
				{
					CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
					if(!target.is_valid_and_not_end(code))
						report_error("Internal Compiler Error: Branch to undefined uid.");
					target->stop_makro=true;
					it->stop_makro=it->dont_use_in_makro=true;
					break;
				}
				case bi::IFELSE: case bi::IFNOTELSE:
				{
					CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
					while(target.is_valid_and_not_end(code) && (Command::label==target->fkt || Command::uid_label==target->fkt))
						++target; // Command-Objekte mit fkt==Command::label werden spter gelscht
					if(!target.is_valid_and_not_end(code))
						report_error("Internal Compiler Error: Branch to undefined uid.");
					it->branch1_id=target->get_uid();
					target->stop_makro=true;

					target=code.find(Command(Command::NewUID(it->branch2_id)));
					while(target.is_valid_and_not_end(code) && (Command::label==target->fkt || Command::uid_label==target->fkt))
						++target; // Command-Objekte mit fkt==Command::label werden spter gelscht
					if(!target.is_valid_and_not_end(code))
						report_error("Internal Compiler Error: Branch to undefined uid.");
					it->branch2_id=target->get_uid();
					target->stop_makro=true;

					it->stop_makro=it->dont_use_in_makro=true;
					break;
				}
				case bi::RET:
					it->stop_makro=it->dont_use_in_makro=true;
					break;
#ifdef USE_TABLE_FOR_CONSTANTS
				case bi::PUSHC:
				case bi::PUSHCS:
					RotSchwarzBaum<Command, PUSHC_Command_LessEqual>::iterator cit=constants.suche(*it);
					if(cit==0)
						report_error("Internal compiler error: assemble_code()");
					Command ccmd(*it);
					ccmd.nr_used=cit->nr_used;
					RotSchwarzBaum<std::pair<Command, uint32>, PUSHC_Command_SavingSorted_LessEqual>::iterator cit2=
						saving_sorted_constants.suche(std::pair<Command, uint32>(ccmd, 0));
					if(cit2!=0) // cit2 ist 0, wenn die Zahl der gesparten Nibbles nicht positiv ist
					{
						it->nibblesize=0;
						it->add(bi::PUSHCT);
						it->add8((uint8)cit2->second);
					}
					break;
#endif
			}
		else
			switch(it->fkt)
			{
				case Command::label:
					break;
				case Command::branch_to_label:
				{
					bool error;
					CommandListType::iterator target=find_branch_target(code, it->branch1_id, it->identifier, error, it);
					if(error)
						report_error("Error in GOTO statement: branch to target in subordinate program structure.");
					while(target.is_valid_and_not_end(code) && (Command::label==target->fkt || Command::uid_label==target->fkt))
						++target; // Command-Objekte mit fkt==Command::label werden spter gelscht
					if(!target.is_valid_and_not_end(code))
						report_error(printf_string("The identifier \'%s\' belonging to the goto statement is undefined.", it->identifier));
					it->branch1_id=target->get_uid();
					it->fkt=Command::command;
					target->stop_makro=true;
					it->stop_makro=it->dont_use_in_makro=true;
					break;
				}
				case Command::get_prev_var:
				case Command::get_prev_var_and_load:
				case Command::get_prev_var_and_store:
				case Command::get_prev_var_and_storepop:
				case Command::set_prev_var:
				case Command::set_prev_var_const:
				case Command::set_prev_var_const_pop:
				case Command::set_prev_var_and_pop:
				case Command::get_prev_var_addr:
				case Command::addc_to_prev_var:
				case Command::prev_var_addc:
				case Command::addc_to_prev_var_pop:
				{
					it->fkt=Command::command;
					break;
				}
				case Command::function:
				case Command::functioncall:
				case Command::begin_infos_before_function:
				case Command::raw_nibbles:
					it->stop_makro=it->dont_use_in_makro=true;
					if(it->fkt==Command::functioncall)
					{
						CountFunctionCalls cnt;
						cnt.function_uid=it->branch1_id;
						cnt.function_header_uid=it->branch2_id; // function_header_uid ist der Sortierschlssel
						cnt.nr_used=1;
						RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare>::iterator it2=count_functioncalls.suche(cnt);
						if(it2==0)
						{
							cnt.alignment=FUNCTION_ADDRESS_ALIGNMENT_NR_OF_RIGHT_ZERO_BITS;
							count_functioncalls.insert(cnt);
						}
						else
							++it2->nr_used;
					}
					break;
			}
	}

#ifdef WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION
	// In der folgenden Schleife werden BRAND und BROR Opcodes in BRF und BRT Opcodes umgewandelt und ggf.
	// zustzliche Befehle eingefgt, um ein korrektes Verhalten der resultierenden Sprungbefehle bezglich der
	// Eigenschaft zu erreichen, ob der boolsche Operand der BRF oder BRT-Befehle bei einem ausgefhrten Sprung
	// vor der Ausfhrung des nchsten Befehls wieder auf den Stack zurckgeschrieben wird.
	// Die Arbeiten der folgenden Schleife drfen NICHT gemeinsam in der vorigen Schleife erledigt werden, weil
	// vorher alle Sprungzeile, die auf einen Befehl verweisen, welcher einem bestimmten angegebenen Befehl folgt
	// (im Fall 0!=(it->branch1_id & 1)), in der vorrigen Schleife so abgendert werden mssen, dass sie direkt
	// auf den Befehl verweisen, zu dem gesprungen werden soll.
	for(CommandListType::iterator it=code.begin(); it!=code.end(); it++)
	{
		uint8 opcode=(it->code[0]<<4) | it->code[1];
		if(Command::command==it->fkt)
			switch(opcode)
			{
				case bi::BRT: case bi::BRF: case bi::BR:
				case bi::BRAND: case bi::BROR:
				{
					CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
					if(bi::BRAND==opcode || bi::BROR==opcode)
					{
						if(Command::command==target->fkt && ((target->code[0]<<4) | target->code[1])<bi::BEGIN_UNARY_OPERATORS)
						{
							target=code.insert(target,
								Command(bi::CAST).set_sp(get_sp_before_command(code, target)).addnibble(Value::t_bool));
							it->branch1_id=target->get_uid();
							++nr_inserted_castbool_for_BRANDOR;
						}	
						if(bi::BRAND==opcode)
						{
							it->code[0]=(bi::BRF>>4)&15;
							it->code[1]=bi::BRF&15;
						}
						else if(bi::BROR==opcode)
						{
							it->code[0]=(bi::BRT>>4)&15;
							it->code[1]=bi::BRT&15;
						}
					}
					else if(bi::BRT==opcode || bi::BRF==opcode)
					{
						if(Command::command==target->fkt && ((target->code[0]<<4) | target->code[1])>=bi::BEGIN_UNARY_OPERATORS)
						{
							Command cmd(bi::BR);
							cmd.fkt=Command::command;
							cmd.branch1_id=target->get_uid();
							cmd.sp=get_sp_before_command(code, target);
							target=code.insert(target, cmd);
							it->branch1_id=target->get_uid();
							++nr_inserted_branch0_for_BRTF;
						}
					}
					break;
				}
				case bi::IFELSE: case bi::IFNOTELSE:
				case bi::BRANDNOT: case bi::BRORNOT:
				{
					CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
					if(bi::BRANDNOT==opcode || bi::BRORNOT==opcode)
					{
						if(Command::command==target->fkt && ((target->code[0]<<4) | target->code[1])<bi::BEGIN_UNARY_OPERATORS)
						{
							target=code.insert(target,
								Command(bi::CAST).set_sp(get_sp_before_command(code, target)).addnibble(Value::t_bool));
							it->branch1_id=target->get_uid();
							++nr_inserted_castbool_for_BRANDOR;
						}	
						if(bi::BRANDNOT==opcode)
						{
							it->code[0]=(bi::IFNOTELSE>>4)&15;
							it->code[1]=bi::IFNOTELSE&15;
						}
						else if(bi::BRORNOT==opcode)
						{
							it->code[0]=(bi::IFELSE>>4)&15;
							it->code[1]=bi::IFELSE&15;
						}
						it->branch2_id=0;
					}
					else if(bi::IFELSE==opcode || bi::IFNOTELSE==opcode)
					{
						if(Command::command==target->fkt && ((target->code[0]<<4) | target->code[1])>=bi::BEGIN_UNARY_OPERATORS)
						{
							Command cmd(bi::BR);
							cmd.fkt=Command::command;
							cmd.branch1_id=target->get_uid();
							cmd.sp=get_sp_before_command(code, target);
							target=code.insert(target, cmd);
							it->branch1_id=target->get_uid();
							++nr_inserted_branch0_for_BRTF;
						}
					}
					break;
				}
			}
	}
#endif
	for(CommandListType::iterator it=code.begin(); it!=code.end(); )
	{
		//printf("fkt=%d    opcode=%u\n", it->fkt, (it->code[0]<<4) | it->code[1]);
		CommandListType::iterator aktit=it;
		++it;
		if(Command::label==aktit->fkt || Command::uid_label==aktit->fkt)
			code.erase(aktit);
		else if(Command::command!=aktit->fkt && Command::functioncall!=aktit->fkt && Command::function!=aktit->fkt &&
				Command::begin_infos_before_function!=aktit->fkt && Command::end_of_code!=aktit->fkt && Command::raw_nibbles!=aktit->fkt)
			report_error("Fehler in assemble_code(): Hier msste fkt=Command::command sein.");
		else if(Command::command==aktit->fkt || Command::functioncall==aktit->fkt)
		{
			OpcodeOrMakro op;
			op.opcode=Command::functioncall==aktit->fkt ? bi::CALL : aktit->get_opcode();
			op.uid=0;
			op.nr_used=0;
			RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
			if(it2==0)
				report_error("Error in assemble_code()");
			++it2->nr_used;
		}
	}

	RotSchwarzBaum<FrequencySortedFunctioncall, FrequencySortedFunctioncall_Compare> freqsortedcalls;
	for(RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare>::iterator it=count_functioncalls.begin();
			it!=count_functioncalls.end(); it++)
	{
		FrequencySortedFunctioncall fc;
		fc.func=&*it;
		freqsortedcalls.insert(fc);
	}

	uint32 function_nr=0;
	for(RotSchwarzBaum<FrequencySortedFunctioncall, FrequencySortedFunctioncall_Compare>::iterator it=freqsortedcalls.maximizer();
			it!=0; it--)
	{
		if(function_nr>=128 || it->func->nr_used<=1)
		{
			it->func->function_nr=255; // Markierung dafr, dass die Addresse und nicht die Nummer der Funktion angegeben werden soll.
			++it->func->alignment;  // weil ein Bit weniger zur Codierung der Adresse zur Verfgung steht.
		}
		else
			it->func->function_nr=function_nr++;
	}

	//print_code(code, opcodes);

	/*FILE *file2=fopen("d:\\makros.txt", "wb");
	if(0==file2)
		report_error("Cannot open debug file2.");

	for(CommandListType::iterator it=code.begin(); it!=code.end(); it++)
	{
		fprintf(file2, "UID=0x%x  Command=%s   %s%s", it->get_uid(), code_element_to_string(it).c_str(),
			0xcccccccc!=it->branch1_id ? printf_string("br1=%x   ", it->branch1_id).c_str() : "",
			0xcccccccc!=it->branch2_id ? printf_string("br2=%x   ", it->branch2_id).c_str() : "");
		if(Command::function==it->fkt)
		{
			RotSchwarzBaum<FunctionUIDtoFunctionName, FunctionUIDtoFunctionName_Compare>::iterator it2=
				cc->functionUID2name.suche(FunctionUIDtoFunctionName(it->get_uid()));
			if(it2!=0)
				fprintf(file2, "   %s", it2->name.c_str());
		}
		else if(Command::functioncall==it->fkt)
		{
			RotSchwarzBaum<FunctionUIDtoFunctionName, FunctionUIDtoFunctionName_Compare>::iterator it2=
				cc->functionUID2name.suche(FunctionUIDtoFunctionName(it->branch1_id));
			if(it2!=0)
				fprintf(file2, "   %s", it2->name.c_str());
		}
		for(long i=(Command::function==it->fkt) ? 2 : 0; i<it->nibblesize; i++)
			fprintf(file2, "%.2x ", it->code[i]);
		fprintf(file2, "\r\n");
	}

	fprintf(file2, "-------------------------------------\r\n");
	fprintf(file2, "BEGIN MAKROS\r\n");
	fprintf(file2, "-------------------------------------\r\n");

	fprintf(file2, "Size before makro generation: %u\r\n", code.size());*/

	nlist<makro_information> generated_makros;

#ifdef USE_CODE_MAKRO_COMPRESSION
	{
		CommandListType::iterator nextit;
		for(CommandListType::iterator it=code.begin(); it!=code.end(); it=nextit)
		{
			nextit=it; ++nextit;
			if(forbidden_makros.suche(it->get_uid())!=0)
				continue;
			if(!it->dont_use_in_makro && it->makro_sp<bi::max_makro_recursion_depth)
			{
				CommandListType::iterator it2=it;
				++it2;
				while(it2!=code.end())
				{
					if(!it2->dont_use_in_makro && it2->makro_sp<bi::max_makro_recursion_depth)
					{
						long anz=0;
						CommandListType::iterator it3=it, it4=it2, it5;
						while(it3!=code.end() && it3!=it2 && it4!=code.end())
						{
							if(Command::command==it3->fkt || Command::command==it4->fkt)
							{
								if(Command::command!=it3->fkt || Command::command!=it4->fkt)
									break;
								long i;
								if(it3->nibblesize!=it4->nibblesize)
									break;
#ifdef ALLOW_BRANCHES_INSIDE_MAKROS
								uint8 op=it3->get_opcode();
								if(it4->get_opcode()!=op) // Abfrage notwendig fr den Fall eines branches.
									break;
								if(bi::BRT==op || bi::BRF==op || bi::BR==op || bi::BRAND==op || bi::BROR==op ||
											bi::BRANDNOT==op || bi::BRORNOT==op)
								{
									if(find_commandlist_distance(code, it3, it3->branch1_id)!=
											find_commandlist_distance(code, it4, it4->branch1_id))
										break;
								}
								else
#endif
								{
									for(i=0; i<it3->nibblesize && i<it4->nibblesize && it3->code[i]==it4->code[i]; i++)
										;
									if(i!=it3->nibblesize || i!=it4->nibblesize)
										break;
								}
							}
							else if(Command::makrocall==it3->fkt || Command::makrocall==it4->fkt)
							{
								if(Command::makrocall!=it3->fkt || Command::makrocall!=it4->fkt)
									break;
								if(it3->branch1_id!=it4->branch1_id)
									break;
							}
							else if(Command::raw_nibbles==it3->fkt || Command::raw_nibbles==it4->fkt)
							{
								if(Command::raw_nibbles!=it3->fkt || Command::raw_nibbles!=it4->fkt)
									break;
								long i;
								if(it3->nibblesize!=it4->nibblesize)
									break;
								for(i=0; i<it3->nibblesize && i<it4->nibblesize && it3->code[i]==it4->code[i]; i++)
									;
								if(i!=it3->nibblesize || i!=it4->nibblesize)
									break;
							}
							else
								break;

#ifdef ALLOW_BRANCHES_INSIDE_MAKROS
							if(it3!=it && (0!=it3->commands_branching_to_this || it4->commands_branching_to_this))
							{
								if(0==it3->commands_branching_to_this || 0==it4->commands_branching_to_this)
									break;
								nlist<Command::uid_type>::iterator uit1, uit2;
								for(uit1=it3->commands_branching_to_this->begin(), uit2=it4->commands_branching_to_this->begin();
										uit1!=it3->commands_branching_to_this->end() &&
										uit2!=it4->commands_branching_to_this->end(); uit1++, uit2++)
									if(find_commandlist_distance(code, it3, *uit1)!=find_commandlist_distance(code, it4, *uit2))
										break;
								if(uit1!=it3->commands_branching_to_this->end() || uit2!=it4->commands_branching_to_this->end())
									break;
							}
#endif

							++anz;
							++it3;
							++it4;
							if(it3->stop_makro || it->dont_use_in_makro || it4->stop_makro || it4->dont_use_in_makro)
								break;
						}
						if(anz>=2 && (check_for_branches_outside_makro(code, it, it3, it2, it4, anz), anz>=2))
						{
							//printf("-------------------------------------\n");
							/*fprintf(file2, "-------------------------------------\r\n");*/
							//fprintf(file2, "Code size=%u\r\n", code.size());
							//fprintf(file2, "Pos1 = %d   Cmd=%s\r\n", get_listpos(code, it), code_element_to_string(it).c_str());
							//fprintf(file2, "Pos2 = %d   Cmd=%s\r\n", get_listpos(code, it2), code_element_to_string(it2).c_str());
							Command cmd(Command::NewUID(it->get_uid()));
							cmd.fkt=Command::makrocall;

							CommandListType::iterator first_makro_command;
							makro_information mi;
							OpcodeOrMakro opc;

							nlist<Command::uid_type> new_uids;
							for(it5=it; it5!=it3; it5++)
								new_uids.push_back(Command::generate_new_uid());
							Command::uid_type endofmakro_uid=Command::generate_new_uid();
							nlist<Command::uid_type>::iterator uidit=new_uids.begin();
							for(it5=it; it5!=it3; ++it5, ++uidit)
							{
								if(uidit==new_uids.end())
									report_error("Internal compiler error: assemble_code()");
								Command c(*it5, Command::NewUID(*uidit));
#ifdef ALLOW_BRANCHES_INSIDE_MAKROS
								if(Command::command==c.fkt)
								{
									uint8 op=c.get_opcode();
									if(bi::BRT==op || bi::BRF==op || bi::BR==op || bi::BRAND==op || bi::BROR==op ||
											bi::BRANDNOT==op || bi::BRORNOT==op)
									{
										int32 d=find_commandlist_distance(code, it5, it5->branch1_id);
										nlist<Command::uid_type>::iterator uidit2=uidit;
										while(d>0)
										{
											--d;
											if(uidit2==new_uids.end())
												report_error("Internal compiler error: assemble_code()");
											++uidit2;
										}
										while(d<0)
										{
											++d;
											if(uidit2==new_uids.begin())
												report_error("Internal compiler error: assemble_code()");
											--uidit2;
										}
										if(uidit2==new_uids.end())
											c.branch1_id=endofmakro_uid;
										else
											c.branch1_id=*uidit2;
									}
								}
#endif
								if(it5==it)
								{
									c.stop_makro=true;
									c.makro_sp=1+it->makro_sp;
								}
								code.push_back(c); // makro_sp wird unten zugewiesen
								if(it5==it)
								{
									first_makro_command=--code.end();
									cmd.branch1_id=code.back().get_uid();

									mi.orig_start_uids.push_back(it->get_uid());
									mi.orig_start_uids.push_back(it2->get_uid());
									mi.start=c.get_uid();
									mi.nr_called=2;
									mi.orig_makro_sp=it->makro_sp;
									opc.opcode=0;
									opc.nr_used=2;
									opc.uid=code.back().get_uid();
								}
							}
							Command endofmakro=Command(Command::NewUID(endofmakro_uid));
							endofmakro.fkt=Command::makro_ret;
							endofmakro.makro_sp=1+it->makro_sp;
							endofmakro.stop_makro=endofmakro.dont_use_in_makro=true;
							code.push_back(endofmakro);
							mi.end=endofmakro.get_uid();
							//fprintf(file2, "Makro call command has UID %x\r\n", cmd.get_uid());
							cmd.stop_makro=it->stop_makro;
							cmd.makro_sp=it->makro_sp;
							code.insert(it, cmd);
							CommandListType::iterator itendmarker=it3;
							//printf("itendmarker=%x\n", itendmarker->get_uid());
							nextit=itendmarker;
							if(it2==nextit)
								{ nextit=it4; itendmarker=it4; }
							cmd.set_uid(it2->get_uid());
							//fprintf(file2, "Makro call command has UID %x\r\n", cmd.get_uid());
							cmd.stop_makro=it2->stop_makro;
							cmd.makro_sp=it2->makro_sp;
							code.insert(it2, cmd);
							if(it3==it2)
								itendmarker=--CommandListType::iterator(it2); // nextit hier nicht verndern !
							while(it2!=it4)
							{
								CommandListType::iterator aktit=it2; ++it2;
								//printf("Erase %d\n", aktit->get_opcode());
								//fprintf(file2, "2) Erase Command with UID %x  Listpos=%u   Cmd=%s\r\n", aktit->get_uid(), get_listpos(code, aktit),
								//	code_element_to_string(aktit).c_str());
								code.erase(aktit);
							}
							//printf("anz = %u, 2    size=%u\n", anz, code.size());
							//fprintf(file2, "anz = %u, 2    size=%u\r\n", anz, code.size());

							long anz2=2;
							while(it2!=first_makro_command)
							{
								if(!it2->dont_use_in_makro && it2->makro_sp<bi::max_makro_recursion_depth)
								{
									long anz=0;
									CommandListType::iterator it3=it, it4=it2, it5;
									while(it3!=itendmarker && it4!=first_makro_command)
									{
										if(Command::command==it3->fkt || Command::command==it4->fkt)
										{
											if(Command::command!=it3->fkt || Command::command!=it4->fkt)
												break;
											long i;
											if(it3->nibblesize!=it4->nibblesize)
												break;
#ifdef ALLOW_BRANCHES_INSIDE_MAKROS
											uint8 op=it3->get_opcode();
											if(bi::BRT==op || bi::BRF==op || bi::BR==op || bi::BRAND==op || bi::BROR==op ||
														bi::BRANDNOT==op || bi::BRORNOT==op)
											{
												if(find_commandlist_distance(code, it3, it3->branch1_id)!=
														find_commandlist_distance(code, it4, it4->branch1_id))
													break;
											}
											else
#endif
											{
												for(i=0; i<it3->nibblesize && i<it4->nibblesize && it3->code[i]==it4->code[i]; i++)
													;
												if(i!=it3->nibblesize || i!=it4->nibblesize)
													break;
											}
										}
										else if(Command::makrocall==it3->fkt || Command::makrocall==it4->fkt)
										{
											if(Command::makrocall!=it3->fkt || Command::makrocall!=it4->fkt)
												break;
											if(it3->branch1_id!=it4->branch1_id)
												break;
										}
										else if(Command::raw_nibbles==it3->fkt || Command::raw_nibbles==it4->fkt)
										{
											if(Command::raw_nibbles!=it3->fkt || Command::raw_nibbles!=it4->fkt)
												break;
											long i;
											if(it3->nibblesize!=it4->nibblesize)
												break;
											for(i=0; i<it3->nibblesize && i<it4->nibblesize && it3->code[i]==it4->code[i]; i++)
												;
											if(i!=it3->nibblesize || i!=it4->nibblesize)
												break;
										}
										else
											break;

#ifdef ALLOW_BRANCHES_INSIDE_MAKROS
										if(0!=it3->commands_branching_to_this || it4->commands_branching_to_this)
										{
											if(0==it3->commands_branching_to_this || 0==it4->commands_branching_to_this)
												break;
											nlist<Command::uid_type>::iterator uit1, uit2;
											for(uit1=it3->commands_branching_to_this->begin(), uit2=it4->commands_branching_to_this->begin();
													uit1!=it3->commands_branching_to_this->end() &&
													uit2!=it4->commands_branching_to_this->end(); uit1++, uit2++)
												if(find_commandlist_distance(code, it3, *uit1)!=find_commandlist_distance(code, it4, *uit2))
													break;
											if(uit1!=it3->commands_branching_to_this->end() || uit2!=it4->commands_branching_to_this->end())
												break;
										}
#endif

										++it3;
										++it4;
										++anz;
										if(it3->stop_makro || it->dont_use_in_makro || it4->stop_makro || it4->dont_use_in_makro)
											break;
									}
									if(it3==itendmarker)
									{
										//fprintf(file2, "Pos%d = %d   Cmd=%s\r\n", anz2+1, get_listpos(code, it2), code_element_to_string(it2).c_str());
										++opc.nr_used;
										if(it2==nextit)
											nextit=it4;
										//fprintf(file2, "Makro call command has UID %x\r\n", it2->get_uid());
										mi.orig_start_uids.push_back(it2->get_uid());
										cmd.set_uid(it2->get_uid());
										cmd.stop_makro=it2->stop_makro;
										cmd.makro_sp=it2->makro_sp;
										code.insert(it2, cmd);
										if(1+it2->makro_sp > first_makro_command->makro_sp)
											first_makro_command->makro_sp=1+it2->makro_sp;
										while(it2!=it4)
										{
											CommandListType::iterator aktit=it2; ++it2;
											//printf("Erase %d\n", aktit->get_opcode());
											//fprintf(file2, "%d) Erase Command with UID %x  Listpos=%u   Cmd=%s\r\n", anz2+1, aktit->get_uid(), get_listpos(code, aktit),
												//code_element_to_string(aktit).c_str());
											code.erase(aktit);
										}
										//fprintf(file2, "anz = %u, %u    size=%u\r\n", anz, ++anz2, code.size());
										++mi.nr_called;
										//printf("anz = %u, %u    size=%u\n", anz, ++anz2, code.size());
									}
									else
										++it2;
								}
								else
									++it2;
							}

							while(it!=itendmarker)
							{
								CommandListType::iterator aktit=it; ++it;
								//printf("Erase %d\n", aktit->get_opcode());
								//fprintf(file2, "1) Erase Command with UID %x  Listpos=%u   Cmd=%s\r\n", aktit->get_uid(), get_listpos(code, aktit),
								//	code_element_to_string(aktit).c_str());
								code.erase(aktit);
							}

							for(it3=first_makro_command; ++it3!=code.end(); )
								it3->makro_sp=first_makro_command->makro_sp;

							opcodes.push_back(opc);
							generated_makros.push_back(mi);
							//fprintf(file2, "OPC insert uid %x\r\n", opc.uid);

							//fprintf(file2, "anz = %u, %u    size=%u\r\n", anz, ++anz2, code.size());
							//printf("anz = %u, %u    size=%u\n", anz, ++anz2, code.size());
							//fprintf(file2, ".....................................\r\n");
							break;
						}
						else
							++it2;
					}
					else
						++it2;
				}
			}
		}
	}
	//fprintf(file2, "-------------------------------------\r\n");
	//fprintf(file2, "END MAKROS\r\n");
	//fprintf(file2, "-------------------------------------\r\n");
#endif

	/*for(CommandListType::iterator it=code.begin(); it!=code.end(); it++)
	{
		fprintf(file2, "UID=0x%x  Command=%s   %s%s", it->get_uid(), code_element_to_string(it).c_str(),
			0xcccccccc!=it->branch1_id ? printf_string("br1=%x   ", it->branch1_id).c_str() : "",
			0xcccccccc!=it->branch2_id ? printf_string("br2=%x   ", it->branch2_id).c_str() : "");
		if(Command::function==it->fkt)
		{
			RotSchwarzBaum<FunctionUIDtoFunctionName, FunctionUIDtoFunctionName_Compare>::iterator it2=
				cc->functionUID2name.suche(FunctionUIDtoFunctionName(it->get_uid()));
			if(it2!=0)
				fprintf(file2, "   %s", it2->name.c_str());
		}
		else if(Command::functioncall==it->fkt)
		{
			RotSchwarzBaum<FunctionUIDtoFunctionName, FunctionUIDtoFunctionName_Compare>::iterator it2=
				cc->functionUID2name.suche(FunctionUIDtoFunctionName(it->branch1_id));
			if(it2!=0)
				fprintf(file2, "   %s", it2->name.c_str());
		}
		for(long i=(Command::function==it->fkt) ? 2 : 0; i<it->nibblesize; i++)
			fprintf(file2, "%.2x ", it->code[i]);
		fprintf(file2, "\r\n");
	}

	if(file2)
		fclose(file2);*/

	RotSchwarzBaum<FrequencySortedOpcode, FrequencySortedOpcode_Compare> FrequencySortedOpcodesTree;
	for(RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it=opcodes.begin(); it!=opcodes.end(); it++)
	{
		it->nr_codenibbles=0;
		if(0!=it->nr_used)
		{
			FrequencySortedOpcode fop;
			fop.op=&*it;
			fop.calc_nr_used();
			FrequencySortedOpcodesTree.insert(fop);
		}
	}
	typedef RotSchwarzBaum<FrequencySortedOpcode, FrequencySortedOpcode_Compare>::iterator FreqTreeIterType;
	/*for(FreqTreeIterType it=FrequencySortedOpcodesTree.begin();
			it!=FrequencySortedOpcodesTree.end(); it++)
		printf("Opcode %d   used %u\n", it->op->opcode, it->op->nr_used);*/

	//printf("size=%d\n", FrequencySortedOpcodesTree.size());
	FrequencySortedOpcode huffman_tree;
	bool start_treesize_smaller16=FrequencySortedOpcodesTree.size()<16;
	if(FrequencySortedOpcodesTree.size()>16 && 0!=((FrequencySortedOpcodesTree.size()-16)%15))
	{
		FrequencySortedOpcode fop;
		fop.op=0;
		long i=1+(FrequencySortedOpcodesTree.size()-16)%15;
		FreqTreeIterType it, it2;
		for(it=FrequencySortedOpcodesTree.minimizer(); 0!=i-- && it!=0; )
			++it;
		--it;
		if(it==0) it=FrequencySortedOpcodesTree.maximizer();
		it2=it;
		uint8 codenibble=0;
		while(it!=0)
			if(0==it->op)
				--it;
			else
			{
				it->codenibble=codenibble++;
				fop.children.push_back(*it);
				FreqTreeIterType aktit=it; --it;
				if(it2==aktit) --it2;
				FrequencySortedOpcodesTree.erase(aktit);
			}
		it=it2;
		while(it!=0)
		{
			it->codenibble=codenibble++;
			fop.children.push_back(*it);
			FreqTreeIterType aktit=it; --it;
			FrequencySortedOpcodesTree.erase(aktit);
		}
		fop.calc_nr_used();
		if(FrequencySortedOpcodesTree.size()==0) // nicht size()<16 !
			huffman_tree=fop;
		else
			FrequencySortedOpcodesTree.insert(fop);
		//printf("size=%d\n", FrequencySortedOpcodesTree.size());
	}
	while(FrequencySortedOpcodesTree.size()>=16 || start_treesize_smaller16)
	{
		start_treesize_smaller16=false;
		FrequencySortedOpcode fop;
		fop.op=0;
		long i=16;
		FreqTreeIterType it, it2;
		for(it=FrequencySortedOpcodesTree.minimizer(); 0!=i-- && it!=0; )
			++it;
		if(it!=0) --it;
		if(it==0) it=FrequencySortedOpcodesTree.maximizer();
		it2=it;
		uint8 codenibble=0;
		while(it!=0) // In dieser Schleife erst die Elemente mit 0!=it->op in FrequencySortedOpcodesTree einfgen.
			if(0==it->op)
				--it;
			else
			{
				it->codenibble=codenibble++;
				fop.children.push_back(*it);
				FreqTreeIterType aktit=it; --it;
				if(it2==aktit) --it2;
				FrequencySortedOpcodesTree.erase(aktit);
			}
		it=it2;
		while(it!=0)
		{
			it->codenibble=codenibble++;
			fop.children.push_back(*it);
			FreqTreeIterType aktit=it; --it;
			FrequencySortedOpcodesTree.erase(aktit);
		}
		fop.calc_nr_used();
		if(FrequencySortedOpcodesTree.size()==0) // nicht size()<16 !
			huffman_tree=fop;
		else
			FrequencySortedOpcodesTree.insert(fop);
		//printf("size=%d\n", FrequencySortedOpcodesTree.size());
	}
	//printf("size=%d\n", FrequencySortedOpcodesTree.size());

	enum { max_tree_depth=8 };
	uint32 nr_elem_per_depth[max_tree_depth+1];
	uint32 nr_children_per_depth[max_tree_depth+1];

	/*FILE *file=fopen("d:\\huffman.txt", "wb");
	if(0==file)
		report_error("Cannot open debug file.");*/

	for(long i=0; i<=max_tree_depth; i++)
		nr_elem_per_depth[i]=nr_children_per_depth[i]=0;

	{
		typedef std::pair<FrequencySortedOpcode*, nlist<FrequencySortedOpcode>::iterator> sPair;
		nlist<sPair> tree_traversion_stack;
		uint8 nbcode[max_tree_depth+1];
		uint64 code[max_tree_depth+1];
		for(long i=0; i<=max_tree_depth; i++)
			code[i]=(uint64)((int64)-1);
		nlist<FrequencySortedOpcode>::iterator it2=huffman_tree.children.begin();
		tree_traversion_stack.push_back(sPair(&huffman_tree, 0));
		bool nextit=false;
		long depth=0;
		while(1)
		{
			while(it2==tree_traversion_stack.back().first->children.end())
			{
				it2=tree_traversion_stack.back().second;
				tree_traversion_stack.pop_back();
				if(0==tree_traversion_stack.size())
					{ nextit=true; break; }
				++it2;
				--depth;
			}
			if(nextit)
				break;
			//for(long i=0; i<depth; i++)
			//	printf("   ");
			//	fprintf(file, "   ");
			//fprintf(file, "Opcode %d   used %u\r\n", 0==it2->op ? -1 : (0==it2->op->uid ? it2->op->opcode : 1000+it2->op->uid), it2->nr_used);
			//printf("Opcode %d   used %u\r\n", 0==it2->op ? -1 : (0==it2->op->uid ? it2->op->opcode : 1000+it2->op->uid), it2->nr_used);
			if(0!=it2->op)
			{
				nr_elem_per_depth[depth<=max_tree_depth ? depth : max_tree_depth]++;
				it2->op->nr_codenibbles=1+depth;  // it2->op->code wird bei einem spteren Baum-Durchlauf gesetzt.
				++it2;
			}
			else
			{
				if(depth<=max_tree_depth)
					nr_children_per_depth[depth]++;
				tree_traversion_stack.push_back(sPair(&*it2, it2));
				nbcode[depth]=it2->codenibble;
				it2=it2->children.begin();
				++depth;
				if(depth>max_tree_depth)
					report_error(printf_string("The sum of numbers of opcodes and makros is too large to be coded with a huffman-like code using maximal %d code bits.",
						depth*4));
			}
		}
	}

	//for(long i=0; i<=max_tree_depth; i++)
	//	printf("depth=%d    elem=%u    children=%u\n", i, nr_elem_per_depth[i], nr_children_per_depth[i]);


	/////////////////////////////////////////////////////////////////////////////////////////

	{
		typedef std::pair<FrequencySortedOpcode*, nlist<FrequencySortedOpcode>::iterator> sPair;
		nlist<sPair> tree_traversion_stack;
		uint64 Code[max_tree_depth+1];
		for(long i=0; i<=max_tree_depth; i++)
		{
			Code[i]=0;
			for(long j=0; j<i; j++)
				Code[i]+=(nr_elem_per_depth[j]<<(4*(i-j)));
		}
		nlist<FrequencySortedOpcode>::iterator it2=huffman_tree.children.begin();
		tree_traversion_stack.push_back(sPair(&huffman_tree, 0));
		bool nextit=false;
		long depth=0;
		while(1)
		{
			while(it2==tree_traversion_stack.back().first->children.end())
			{
				it2=tree_traversion_stack.back().second;
				tree_traversion_stack.pop_back();
				if(0==tree_traversion_stack.size())
					{ nextit=true; break; }
				++it2;
				--depth;
			}
			if(nextit)
				break;
			if(0!=it2->op)
			{
				it2->op->code=Code[depth]++; // it2->op->nr_codenibbles wird bei einem frheren Baum-Durchlauf gesetzt.
				++it2;
			}
			else
			{
				tree_traversion_stack.push_back(sPair(&*it2, it2));
				it2=it2->children.begin();
				++depth;
			}
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////////

	EncodeNibblesUsingEscapeSequences sim;
	sim.set_only_simulate();
	for(CommandListType::iterator it=code.begin(); it!=code.end(); it++)
	{
		it->disable_escape_sequences=false;
		sim.enable_encoding(0!=it->makro_sp && !it->is_makro_param);
		it->pc=sim.nibblesize(get_first_code_nibble_of_command(it, opcodes));
		if(Command::command==it->fkt)
		{
			uint8 opcode=(it->code[0]<<4) | it->code[1];
			OpcodeOrMakro op; op.opcode=opcode; op.uid=0; op.nr_used=0;
			RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
			if(it2==0 || 0==it2->nr_codenibbles)
				report_error("Internal compiler error: Error in assemble_code()");
			for(long i=0; i<it2->nr_codenibbles; i++)
				sim.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15); // Codierten Opcode schreiben, auch bei branches.
			switch(opcode)
			{
				case bi::BRT: case bi::BRF: case bi::BR:
				case bi::BRAND: case bi::BROR: // diese Flle knnen nur dann auftreten, wenn 
				case bi::BRANDNOT: case bi::BRORNOT: // WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION nicht definiert ist
				{
					sim.simulate_known_number_of_unknown_nibbles(2+2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER);
					break;
				}
				case bi::IFELSE: case bi::IFNOTELSE:
				{
					if(0==it->branch2_id)
						sim.simulate_known_number_of_unknown_nibbles(2+2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER);
					else
						sim.simulate_known_number_of_unknown_nibbles(4+4*MAX_NR_BYTES_FOR_PROGRAM_COUNTER);
					break;
				}
				default:
					for(long i=2; i<it->nibblesize; i++)
						sim.push_back(it->code[i]);
					break;
			}
		}
		else switch(it->fkt)
		{
			case Command::functioncall:
			{
				OpcodeOrMakro op; op.opcode=bi::CALL; op.uid=0; op.nr_used=0;
				RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
				if(it2==0 || 0==it2->nr_codenibbles)
					report_error("Internal compiler error: Error in assemble_code()");
				for(long i=0; i<it2->nr_codenibbles; i++)
					sim.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15);

				CountFunctionCalls cnt;
				cnt.function_header_uid=it->branch2_id; // function_header_uid ist der Sortierschlssel
				RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare>::iterator it3=count_functioncalls.suche(cnt);
				if(it3==0)
					report_error("Internal compiler error: Error in assemble_code()");
				sim.increase_simulated_size((255!=it3->function_nr) ? 2 : 2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER);
				// Ein Funktionsaufruf darf nicht in einem Makro vorkommen. Daher darf increase_simulated_size() aufgerufen werden.
				break;
			}
			case Command::begin_infos_before_function:
			{
				CountFunctionCalls cnt;
				cnt.function_header_uid=it->get_uid(); // function_header_uid ist der Sortierschlssel
				RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare>::iterator it2=count_functioncalls.suche(cnt);
				if(it2!=0 && 0!=it2->alignment) // it2==0 tritt auf, wenn eine definierte Funktion nirgenswo aufgerufen wird.
					sim.increase_simulated_size((1LU<<it2->alignment)-1);
				break;
			}
			case Command::raw_nibbles:
				for(long i=0; i<it->nibblesize; i++)
					sim.push_back(it->code[i]);
				break;
			case Command::makrocall:
			{
				OpcodeOrMakro op; op.opcode=0; op.uid=it->branch1_id; op.nr_used=0;
				RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
				if(it2==0 || 0==it2->nr_codenibbles)
					report_error("Internal compiler error: Error in assemble_code()");
				for(long i=0; i<it2->nr_codenibbles; i++)
					sim.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15);
				break;
			}
			case Command::makro_ret:
				sim.push_back_makro_ret();
				break;
			case Command::makro_param_call_fixedsize:
				sim.push_back_makro_param_call_fixedsize(it->var_data);
				break;
			case Command::makro_param_call_varsize:
				sim.push_back_makro_param_call_varsize();
				break;
		}
	}
	//printf("End pc = %u\n", pc);

	for(long optijump=0; optijump<5; optijump++)
	{
		sim.reset_simulated_size();
		bool t1=(0==(optijump&1));
		for(CommandListType::iterator it=code.begin(); it!=code.end(); it++)
		{
			sim.enable_encoding(0!=it->makro_sp && !it->is_makro_param && !it->disable_escape_sequences);
			if(t1)
			{
				it->pc2=sim.nibblesize(get_first_code_nibble_of_command(it, opcodes));
				//printf("it->pc2=%d\n", it->pc2);
			}
			else
			{
				it->pc=sim.nibblesize(get_first_code_nibble_of_command(it, opcodes));
				//printf("it->pc=%d\n", it->pc);
			}
			if(optijump>=2 && it->pc2!=it->pc)
				optijump&=1; // Die Schleife muss so lange durchlaufen werden, bis keine nderungen mehr an it->pc und it->pc2 auftreten.

			if(Command::command==it->fkt)
			{
				uint8 opcode=(it->code[0]<<4) | it->code[1];
				OpcodeOrMakro op; op.opcode=opcode; op.uid=0; op.nr_used=0;
				RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
				if(it2==0 || 0==it2->nr_codenibbles)
					report_error("Internal compiler error: Error in assemble_code()");
				if(it->disable_escape_sequences)
				{
					sim.flush();
					sim.increase_simulated_size(5); // Die Sequenz zum Deaktivieren der Escape-Nibble-Erkennung bentigt 5 Nibbles
				}
				for(long i=0; i<it2->nr_codenibbles; i++)
					sim.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15); // Codierten Opcode schreiben, auch bei branches.
				switch(opcode)
				{
					case bi::BRT: case bi::BRF: case bi::BR:
					case bi::BRAND: case bi::BROR: // diese Flle knnen nur dann auftreten, wenn 
					case bi::BRANDNOT: case bi::BRORNOT: // WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION nicht definiert ist
					{
						CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
						if(!target.is_valid_and_not_end(code))
							report_error("Internal Compiler Error: Branch to undefined uid.");
						uint32 limit=(bi::BRT==opcode || bi::BRF==opcode || bi::BR==opcode) ? 127 : 254;
						it->short_branch1= (t1 ? target->pc : target->pc2) <= (t1 ? it->pc : it->pc2)+limit &&
							(t1 ? target->pc : target->pc2)+limit>=(t1 ? it->pc : it->pc2);
						Command cmd(*it);
						add_branch_distance(cmd, (t1 ? target->pc : target->pc2)-(t1 ? it->pc : it->pc2),
							!it->short_branch1, bi::BRT==opcode || bi::BRF==opcode || bi::BR==opcode);
						for(long i=2; i<cmd.nibblesize; i++)
							sim.push_back(cmd.code[i]);
						/*if(0!=optijump && !it->disable_escape_sequences)
							{ it->disable_escape_sequences=true; optijump&=1; }*/
						if(0!=optijump && ++CommandListType::iterator(it)!=code.end() &&
								(sim.nibblesize(get_first_code_nibble_of_command(++CommandListType::iterator(it), opcodes))-
								(t1 ? it->pc2 : it->pc)) >
								((t1 ? (++CommandListType::iterator(it))->pc : (++CommandListType::iterator(it))->pc2)-
								(t1 ? it->pc : it->pc2)))
							{ it->disable_escape_sequences=true; optijump&=1; }
						/*{
							printf("==> %d   %d   %d   %d\n",
								sim.nibblesize(get_first_code_nibble_of_command(++CommandListType::iterator(it), opcodes)),
								(t1 ? it->pc2 : it->pc),
								(t1 ? (++CommandListType::iterator(it))->pc : (++CommandListType::iterator(it))->pc2),
								(t1 ? it->pc : it->pc2));
							it->disable_escape_sequences=true; optijump&=1;
						}*/
						break;
					}
					case bi::IFELSE: case bi::IFNOTELSE:
					{
						CommandListType::iterator target1=code.find(Command(Command::NewUID(it->branch1_id)));
						CommandListType::iterator target2=code.find(Command(Command::NewUID(it->branch2_id)));
						if(!target1.is_valid_and_not_end(code) || (it->branch2_id!=0 && !target2.is_valid_and_not_end(code)))
							report_error("Internal Compiler Error: Branch to undefined uid.");
						long d1=(t1 ? target1->pc : target1->pc2) - (t1 ? it->pc : it->pc2);
						if(d1<0)
							report_error("Interner Compiler-Fehler: Sprungweite fr den else-Teil eines IFELSE- oder IFNOTELSE-Befehls ist negativ.");
						it->short_branch1=(d1<255);
						Command cmd(*it);
						add_branch_distance(cmd, d1, !it->short_branch1, false); // Sprungweite als unsigned
						if(0!=it->branch2_id)
						{
							long d2=(t1 ? target2->pc : target2->pc2) - (t1 ? it->pc : it->pc2);
							if(d2-d1>=0) d2-=d1;
							it->short_branch2= (d2<0 ? (-d2) : d2)<=127;
							add_branch_distance(cmd, (d2-d1>=0) ? (d2-d1) : d2, !it->short_branch2, true); // signed
						}
						for(long i=2; i<cmd.nibblesize; i++)
							sim.push_back(cmd.code[i]);
						if(0!=optijump && ++CommandListType::iterator(it)!=code.end() &&
								(sim.nibblesize(get_first_code_nibble_of_command(++CommandListType::iterator(it), opcodes))-
								(t1 ? it->pc2 : it->pc)) >
								((t1 ? (++CommandListType::iterator(it))->pc : (++CommandListType::iterator(it))->pc2)-
								(t1 ? it->pc : it->pc2)))
							{ it->disable_escape_sequences=true; optijump&=1; }
						break;
					}
					default:
					{
						for(long i=2; i<it->nibblesize; i++)
							sim.push_back(it->code[i]);
						break;
					}
				}
			}
			else switch(it->fkt)
			{
				case Command::functioncall:
				{
					OpcodeOrMakro op; op.opcode=bi::CALL; op.uid=0; op.nr_used=0;
					RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
					if(it2==0 || 0==it2->nr_codenibbles)
						report_error("Internal compiler error: Error in assemble_code()");
					for(long i=0; i<it2->nr_codenibbles; i++)
						sim.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15);

					CountFunctionCalls cnt;
					cnt.function_header_uid=it->branch2_id; // function_header_uid ist der Sortierschlssel
					RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare>::iterator it3=count_functioncalls.suche(cnt);
					if(it3==0)
						report_error("Internal compiler error: Error in assemble_code()");
					sim.increase_simulated_size((255!=it3->function_nr) ? 2 : 2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER);
					// Ein Funktionsaufruf darf nicht in einem Makro vorkommen. Daher darf increase_simulated_size() aufgerufen werden.
					break;
				}
				case Command::begin_infos_before_function:
				{
					CountFunctionCalls cnt;
					cnt.function_header_uid=it->get_uid(); // function_header_uid ist der Sortierschlssel
					RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare>::iterator it2=count_functioncalls.suche(cnt);
					if(it2!=0 && 0!=it2->alignment)
						sim.increase_simulated_size((1LU<<it2->alignment)-1);
					break;
				}
				case Command::raw_nibbles:
					for(long i=0; i<it->nibblesize; i++)
						sim.push_back(it->code[i]);
					break;
				case Command::makrocall:
				{
					OpcodeOrMakro op; op.opcode=0; op.uid=it->branch1_id; op.nr_used=0;
					RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
					if(it2==0 || 0==it2->nr_codenibbles)
						report_error("Internal compiler error: Error in assemble_code()");
					for(long i=0; i<it2->nr_codenibbles; i++)
						sim.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15);
					break;
				}
				case Command::makro_ret:
					sim.push_back_makro_ret();
					break;
				case Command::makro_param_call_fixedsize:
					sim.push_back_makro_param_call_fixedsize(it->var_data);
					break;
				case Command::makro_param_call_varsize:
					sim.push_back_makro_param_call_varsize();
					break;
			}
		}
		//printf("End pc = %u\n", pc);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	nlist<VARIABLE_INDEX_TYPE> begin_addrs_of_elem_types;
	for(int j=0; j<3; j++)
		for(int i=0; i<Value::anz_user_types; i++)
			begin_addrs_of_elem_types.push_back(Variable::var_mem_begin[i+j*(1+Value::anz_user_types)]);
	begin_addrs_of_elem_types.push_back(Variable::var_mem_count[1+(Value::anz_user_types-1)+2*(1+Value::anz_user_types)]);

	uint32 init_value_size0=0;
	for(long i=0; i<Value::anz_user_types; i++)
		init_value_size0+=Value::sbytesize(i)*Variable::init_values[i].size();
	init_value_size0+=Variable::init_values[Value::anz_user_types].size();

	uint32 init_value_size1=0;
	for(long i=0; i<Value::anz_user_types; i++)
		init_value_size1+=Value::sbytesize(i)*Constant::init_values[i].size();
	init_value_size1+=Constant::init_values[Value::anz_user_types].size();

	bi::BytecodeProgramHeader ProgramHeader;
	ProgramHeader.global_begin_addrs_of_elementary_types_start()=2*sizeof(ProgramHeader);
	ProgramHeader.addr_of_first_global_non_elementary_variable(0)=Variable::var_mem_begin[Value::anz_user_types];
	ProgramHeader.addr_of_first_global_non_elementary_variable(1)=Variable::var_mem_begin[2*Value::anz_user_types+1];
	ProgramHeader.addr_of_first_global_non_elementary_variable(2)=Variable::var_mem_begin[3*Value::anz_user_types+2];
	ProgramHeader.global_var_type_info_start(0)=sizeof(ProgramHeader)+sizeof(VARIABLE_INDEX_TYPE)*begin_addrs_of_elem_types.size();
	ProgramHeader.global_var_type_info_start(1)=ProgramHeader.global_var_type_info_start(0)+(Variable::type_info[0].size()+1+3)/4;
	ProgramHeader.global_var_type_info_start(2)=ProgramHeader.global_var_type_info_start(1)+(Variable::type_info[1].size()+1+3)/4;
	ProgramHeader.global_var_initvalues_start(0)=ProgramHeader.global_var_type_info_start(2)+(Variable::type_info[2].size()+1+3)/4;
	ProgramHeader.global_var_initvalues_start(1)=ProgramHeader.global_var_initvalues_start(0)+init_value_size0;
	ProgramHeader.global_var_type_info_start(0)*=2; // Nibbles
	ProgramHeader.global_var_type_info_start(1)*=2; // Nibbles
	ProgramHeader.global_var_type_info_start(2)*=2; // Nibbles
	// ProgramHeader.global_var_initvalues_start(i) ist Byte-basiert.

	ProgramHeader.opcode_transformation_tables(0)=2*(ProgramHeader.global_var_initvalues_start(1)+init_value_size1);
	ProgramHeader.limit_for_4bit_opcodes()=nr_elem_per_depth[0];
	ProgramHeader.opcode_transformation_tables(1)=ProgramHeader.opcode_transformation_tables(0)+
		2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER*nr_elem_per_depth[0];
	ProgramHeader.limit_for_8bit_opcodes()=nr_elem_per_depth[1];
	for(long i=0; i<6; i++)
	{
		ProgramHeader.opcode_transformation_tables(2+i)=ProgramHeader.opcode_transformation_tables(1+i)+
			2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER*(0==i ? ProgramHeader.limit_for_8bit_opcodes() : ProgramHeader.limit_for_nbit_opcodes(i-1));
		if(i<5)
			ProgramHeader.limit_for_nbit_opcodes(i)=nr_elem_per_depth[(2+i>=max_tree_depth) ? (max_tree_depth-1) : (2+i)];
	}
	
	ProgramHeader.function_address_table()=ProgramHeader.opcode_transformation_tables(7)+
		2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER*nr_elem_per_depth[7];

	//ProgramHeader.constants_table()=ProgramHeader.function_address_table() + function_nr*2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER;
	//code_begin=ProgramHeader.constants_table() + constant_nr*2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER;
#ifdef USE_TABLE_FOR_CONSTANTS
	ProgramHeader.limit_for_1bytes_constants()=bytesize_constant_nr_limit[0];
	ProgramHeader.limit_for_2bytes_constants()=bytesize_constant_nr_limit[1];
	ProgramHeader.limit_for_4bytes_constants()=bytesize_constant_nr_limit[2];
	ProgramHeader.constants1bytes_table()=ProgramHeader.function_address_table() + function_nr*2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER;
	ProgramHeader.constants2bytes_table()=ProgramHeader.constants1bytes_table() + bytesize_constant_nr_limit[0]*3;
	ProgramHeader.constants4bytes_table()=ProgramHeader.constants2bytes_table() +
		(bytesize_constant_nr_limit[1]-bytesize_constant_nr_limit[0])*5;
	ProgramHeader.constants8bytes_table()=ProgramHeader.constants4bytes_table() +
		(bytesize_constant_nr_limit[2]-bytesize_constant_nr_limit[1])*9;
	PROGRAM_COUNTER_TYPE code_begin=ProgramHeader.constants8bytes_table() + (bytesize_constant_nr_limit[3]-bytesize_constant_nr_limit[2])*17;

	ProgramHeader.constants2bytes_table() -= (ProgramHeader.limit_for_1bytes_constants()*5);
	ProgramHeader.constants4bytes_table() -= (ProgramHeader.limit_for_2bytes_constants()*9);
	ProgramHeader.constants8bytes_table() -= (ProgramHeader.limit_for_4bytes_constants()*17);
#else
	PROGRAM_COUNTER_TYPE code_begin=ProgramHeader.function_address_table() + function_nr*2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER;
#endif
	
	cd.push_back_nibblelist(make_nibblelist(&ProgramHeader, sizeof(ProgramHeader)));
	cd.push_back_nibblelist(make_nibblelist(begin_addrs_of_elem_types));

	//printf("TYPEINFO:\n");
	for(int j=0; j<3; j++)
	{
		Variable::type_info[j].push_back(3); // damit das Ende der Type-Info der letzten Variable gefunden wird
		while(0!=(Variable::type_info[j].size() & 3))
			Variable::type_info[j].push_back(0);
		for(nlist<uint8>::iterator it=Variable::type_info[j].begin(); it!=Variable::type_info[j].end(); it++)
		{
			uint8 temp=*it;
			//printf("Type Info = %u\n", temp);
			if(++it==Variable::type_info[j].end())
				report_error("Internal Compiler Error in assemble_code()");
			//printf("Type Info = %u\n", *it);
			cd.push_back(temp + ((*it)<<2));
			//printf("WRITE %u\n", temp + ((*it)<<2));
		}
		//printf("-----------------------------\n2.1 ####### %u   %u\n----------------\n", cd.size(), cd.size()/2);
	}

	unsigned short test_little_endian=1u;
	bool compiler_runs_on_little_endian=((unsigned char)1u) == *((unsigned char*)(&test_little_endian));
	for(long i=0; i<Value::anz_user_types; i++)
		for(nlist<uint64>::iterator it=Variable::init_values[i].begin(); it!=Variable::init_values[i].end(); it++)
		{
			for(long j=0; j<Value::sbytesize(i); j++)
			{
				cd.push_back(((*it)>>(8*j+4))&15); // Innerhalb eines Bytes Big Endian, aber bei einer aus mehreren Bytes bestehenden
				cd.push_back(((*it)>>(8*j))&15);   // Zahl werden die Bytes im Little Endian Format abgelegt.
			}
		}

	for(nlist<uint64>::iterator it=Variable::init_values[Value::anz_user_types].begin();
			it!=Variable::init_values[Value::anz_user_types].end(); it++)
	{
		cd.push_back(((*it)>>4)&15); // Innerhalb eines Bytes Big Endian, aber bei einer aus mehreren Bytes bestehenden
		cd.push_back((*it)&15);   // Zahl werden die Bytes im Little Endian Format abgelegt.
	}
			
	//printf("3 ####### %u   %u\n", cd.size(), cd.size()/2);
	for(long i=0; i<Value::anz_user_types; i++)
		for(nlist<uint64>::iterator it=Constant::init_values[i].begin(); it!=Constant::init_values[i].end(); it++)
		{
			for(long j=0; j<Value::sbytesize(i); j++)
			{
				cd.push_back(((*it)>>(8*j+4))&15); // Innerhalb eines Bytes Big Endian, aber bei einer aus mehreren Bytes bestehenden
				cd.push_back(((*it)>>(8*j))&15);   // Zahl werden die Bytes im Little Endian Format abgelegt.
			}
		}

	//printf("4 ####### %u   %u\n", cd.size(), cd.size()/2);
	for(nlist<uint64>::iterator it=Constant::init_values[Value::anz_user_types].begin();
			it!=Constant::init_values[Value::anz_user_types].end(); it++)
	{
		cd.push_back(((*it)>>4)&15); // Innerhalb eines Bytes Big Endian, aber bei einer aus mehreren Bytes bestehenden
		cd.push_back((*it)&15);   // Zahl werden die Bytes im Little Endian Format abgelegt.
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	sim.set_simulated_size(code_begin);
	for(CommandListType::iterator it=code.begin(); it!=code.end(); it++)
	{
		it->pc=sim.nibblesize(get_first_code_nibble_of_command(it, opcodes));
		//printf("***** %d\n", it->pc);
		sim.enable_encoding(0!=it->makro_sp && !it->is_makro_param && !it->disable_escape_sequences);
		/*printf("UID=0x%x  pc=%.3x   Command=%s   %s%s\r\n", it->get_uid(), it->pc, code_element_to_string(it).c_str(),
			0xcccccccc!=it->branch1_id ? printf_string("br1=%x   ", it->branch1_id).c_str() : "",
			0xcccccccc!=it->branch2_id ? printf_string("br2=%x   ", it->branch2_id).c_str() : "");*/
		if(Command::command==it->fkt)
		{
			uint8 opcode=(it->code[0]<<4) | it->code[1];
			OpcodeOrMakro op; op.opcode=opcode; op.uid=0; op.nr_used=0;
			RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
			if(it2==0 || 0==it2->nr_codenibbles)
				report_error("Internal compiler error: Error in assemble_code()");
			if(it->disable_escape_sequences)
			{
				sim.flush();
				sim.increase_simulated_size(5); // Die Sequenz zum Deaktivieren der Escape-Nibble-Erkennung bentigt 5 Nibbles
			}
			for(long i=0; i<it2->nr_codenibbles; i++)
				sim.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15); // Codierten Opcode schreiben, auch bei branches.
			switch(opcode)
			{
				case bi::BRT: case bi::BRF: case bi::BR:
				case bi::BRAND: case bi::BROR: // diese Flle knnen nur dann auftreten, wenn 
				case bi::BRANDNOT: case bi::BRORNOT: // WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION nicht definiert ist
				{
					CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
					if(!target.is_valid_and_not_end(code))
						report_error("Internal Compiler Error: Branch to undefined uid.");
					Command cmd(*it);
					add_branch_distance(cmd, target->pc2 - it->pc2, !it->short_branch1, bi::BRT==opcode || bi::BRF==opcode || bi::BR==opcode);
					for(long i=2; i<cmd.nibblesize; i++)
						sim.push_back(cmd.code[i]);
					break;
				}
				case bi::IFELSE: case bi::IFNOTELSE:
				{
					CommandListType::iterator target1=code.find(Command(Command::NewUID(it->branch1_id)));
					CommandListType::iterator target2=code.find(Command(Command::NewUID(it->branch2_id)));
					if(!target1.is_valid_and_not_end(code) || (it->branch2_id!=0 && !target2.is_valid_and_not_end(code)))
						report_error("Internal Compiler Error: Branch to undefined uid.");
					long d1=target1->pc2 - it->pc2;
					if(d1<0)
						report_error("Interner Compiler-Fehler: Sprungweite fr den else-Teil eines IFELSE- oder IFNOTELSE-Befehls ist negativ.");
					Command cmd(*it);
					add_branch_distance(cmd, d1, !it->short_branch1, false); // unsigned
					if(it->branch2_id!=0)
					{
						long d2=target2->pc2 - it->pc2;
						add_branch_distance(cmd, (d2-d1>=0) ? (d2-d1) : d2, !it->short_branch2, true); // signed
					}
					for(long i=2; i<cmd.nibblesize; i++)
						sim.push_back(cmd.code[i]);
					break;
				}
				default:
					for(long i=2; i<it->nibblesize; i++)
						sim.push_back(it->code[i]);
					break;
			}
		}
		else switch(it->fkt)
		{
			case Command::functioncall:
			{
				OpcodeOrMakro op; op.opcode=bi::CALL; op.uid=0; op.nr_used=0;
				RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
				if(it2==0 || 0==it2->nr_codenibbles)
					report_error("Internal compiler error: Error in assemble_code()");
				for(long i=0; i<it2->nr_codenibbles; i++)
					sim.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15);

				CountFunctionCalls cnt;
				cnt.function_header_uid=it->branch2_id; // function_header_uid ist der Sortierschlssel
				RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare>::iterator it3=count_functioncalls.suche(cnt);
				if(it3==0)
					report_error("Internal compiler error: Error in assemble_code()");
				sim.increase_simulated_size((255!=it3->function_nr) ? 2 : 2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER);
				break;
			}
			case Command::function:
			{
				RotSchwarzBaum<FunctionUIDtoFunctionName, FunctionUIDtoFunctionName_Compare>::iterator funcit=
					cc->functionUID2name.suche(FunctionUIDtoFunctionName(it->get_uid()));
				if(funcit==0)
					report_error("Internal compiler error: Error in assemble_code()");
				cc->function_begin_addresses[funcit->name]=it->pc;
				break;
			}
			case Command::begin_infos_before_function:
			{
				CountFunctionCalls cnt;
				cnt.function_header_uid=it->get_uid(); // function_header_uid ist der Sortierschlssel
				RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare>::iterator it2=count_functioncalls.suche(cnt);
				if(it2!=0 && 0!=it2->alignment)
				{
					CommandListType::iterator it3=it;
					PROGRAM_COUNTER_TYPE pc2=sim.nibblesize(0); // keine Codierung mit Escape-Sequenzen, weil Makros nach allen Funktionen
					while(++it3!=code.end() && it3->fkt!=Command::function)
					{
						if(it3->fkt!=Command::raw_nibbles)
							report_error("Internal compiler error: Error in assemble_code()");
						pc2+=it3->nibblesize; // Weil Funktionen grundstzlich vor allen Makros definiert werden, wird hier
						// keine Codierung durch Escape-Sequenzen vorgenommen, so dass die it3->nibblesize einfach aufaddiert
						// werden knnen.
					}
					sim.increase_simulated_size(((pc2+(1LU<<it2->alignment)-1) & (~(uint32)((1LU<<it2->alignment)-1))) - pc2);
				}
				break;
			}
			case Command::raw_nibbles:
				for(long i=0; i<it->nibblesize; i++)
					sim.push_back(it->code[i]);
				break;
			case Command::makrocall:
			{
				OpcodeOrMakro op; op.opcode=0; op.uid=it->branch1_id; op.nr_used=0;
				RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
				if(it2==0 || 0==it2->nr_codenibbles)
					report_error("Internal compiler error: Error in assemble_code()");
				for(long i=0; i<it2->nr_codenibbles; i++)
					sim.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15);
				break;
			}
			case Command::makro_ret:
				sim.push_back_makro_ret();
				break;
			case Command::makro_param_call_fixedsize:
				sim.push_back_makro_param_call_fixedsize(it->var_data);
				break;
			case Command::makro_param_call_varsize:
				sim.push_back_makro_param_call_varsize();
				break;
		}
	}

	EncodeNibblesUsingEscapeSequences sim2;
	sim2.set_only_simulate();
	for(CommandListType::iterator it=code.begin(); it!=code.end(); it++)
	{
		it->pc2=sim2.nibblesize(0);
		sim2.enable_encoding(false);
		if(Command::command==it->fkt)
		{
			uint8 opcode=(it->code[0]<<4) | it->code[1];
			OpcodeOrMakro op; op.opcode=opcode; op.uid=0; op.nr_used=0;
			RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
			if(it2==0 || 0==it2->nr_codenibbles)
				report_error("Internal compiler error: Error in assemble_code()");
			for(long i=0; i<it2->nr_codenibbles; i++)
				sim2.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15);
			switch(opcode)
			{
				case bi::BRT: case bi::BRF: case bi::BR:
				case bi::BRAND: case bi::BROR: // diese Flle knnen nur dann auftreten, wenn 
				case bi::BRANDNOT: case bi::BRORNOT: // WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION nicht definiert ist
				{
					CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
					if(!target.is_valid_and_not_end(code))
						report_error("Internal Compiler Error: Branch to undefined uid.");
					Command cmd(*it);
					add_branch_distance(cmd, target->pc - it->pc, !it->short_branch1, bi::BRT==opcode || bi::BRF==opcode || bi::BR==opcode);
					for(long i=2; i<cmd.nibblesize; i++)
						sim2.push_back(cmd.code[i]);
					break;
				}
				case bi::IFELSE: case bi::IFNOTELSE:
				{
					CommandListType::iterator target1=code.find(Command(Command::NewUID(it->branch1_id)));
					CommandListType::iterator target2=code.find(Command(Command::NewUID(it->branch2_id)));
					if(!target1.is_valid_and_not_end(code) || (it->branch2_id!=0 && !target2.is_valid_and_not_end(code)))
						report_error("Internal Compiler Error: Branch to undefined uid.");
					long d1=target1->pc - it->pc;
					if(d1<0)
						report_error("Interner Compiler-Fehler: Sprungweite fr den else-Teil eines IFELSE- oder IFNOTELSE-Befehls ist negativ.");
					Command cmd(*it);
					add_branch_distance(cmd, d1, !it->short_branch1, false); // unsigned
					if(it->branch2_id!=0)
					{
						long d2=target2->pc - it->pc;
						add_branch_distance(cmd, (d2-d1>=0) ? (d2-d1) : d2, !it->short_branch2, true); // signed
					}
					for(long i=2; i<cmd.nibblesize; i++)
						sim2.push_back(cmd.code[i]);
					break;
				}
				default:
					for(long i=2; i<it->nibblesize; i++)
						sim2.push_back(it->code[i]);
					break;
			}
		}
		else switch(it->fkt)
		{
			case Command::functioncall:
			{
				OpcodeOrMakro op; op.opcode=bi::CALL; op.uid=0; op.nr_used=0;
				RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
				if(it2==0 || 0==it2->nr_codenibbles)
					report_error("Internal compiler error: Error in assemble_code()");
				for(long i=0; i<it2->nr_codenibbles; i++)
					sim2.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15);

				CountFunctionCalls cnt;
				cnt.function_header_uid=it->branch2_id; // function_header_uid ist der Sortierschlssel
				RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare>::iterator it3=count_functioncalls.suche(cnt);
				if(it3==0)
					report_error("Internal compiler error: Error in assemble_code()");
				sim2.increase_simulated_size((255!=it3->function_nr) ? 2 : 2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER);
				break;
			}
			case Command::function:
				break;
			case Command::begin_infos_before_function:
			{
				CountFunctionCalls cnt;
				cnt.function_header_uid=it->get_uid(); // function_header_uid ist der Sortierschlssel
				RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare>::iterator it2=count_functioncalls.suche(cnt);
				if(it2!=0 && 0!=it2->alignment)
				{
					CommandListType::iterator it3=it;
					PROGRAM_COUNTER_TYPE pc2=sim2.nibblesize(0); // keine Codierung mit Escape-Sequenzen, weil Makros nach allen Funktionen
					while(++it3!=code.end() && it3->fkt!=Command::function)
					{
						if(it3->fkt!=Command::raw_nibbles)
							report_error("Internal compiler error: Error in assemble_code()");
						pc2+=it3->nibblesize; // Weil Funktionen grundstzlich vor allen Makros definiert werden, wird hier
						// keine Codierung durch Escape-Sequenzen vorgenommen, so dass die it3->nibblesize einfach aufaddiert
						// werden knnen.
					}
					sim2.increase_simulated_size(((pc2+(1LU<<it2->alignment)-1) & (~(uint32)((1LU<<it2->alignment)-1))) - pc2);
				}
				break;
			}
			case Command::raw_nibbles:
				for(long i=0; i<it->nibblesize; i++)
					sim2.push_back(it->code[i]);
				break;
			case Command::makrocall:
			{
				OpcodeOrMakro op; op.opcode=0; op.uid=it->branch1_id; op.nr_used=0;
				RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
				if(it2==0 || 0==it2->nr_codenibbles)
					report_error("Internal compiler error: Error in assemble_code()");
				for(long i=0; i<it2->nr_codenibbles; i++)
					sim2.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15);
				break;
			}
			case Command::makro_ret:
				sim2.push_back_makro_ret();
				break;
			case Command::makro_param_call_fixedsize:
				sim2.push_back_makro_param_call_fixedsize(it->var_data);
				break;
			case Command::makro_param_call_varsize:
				sim2.push_back_makro_param_call_varsize();
				break;
		}
	}

	//printf("End pc = %u\n", pc);
	//printf("---------------------------------------------\n");

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	nlist<PROGRAM_COUNTER_TYPE> transform_opcode_nbit[8];

	{
		typedef std::pair<FrequencySortedOpcode*, nlist<FrequencySortedOpcode>::iterator> sPair;
		nlist<sPair> tree_traversion_stack;
		nlist<FrequencySortedOpcode>::iterator it2=huffman_tree.children.begin();
		tree_traversion_stack.push_back(sPair(&huffman_tree, 0));
		bool nextit=false;
		long depth=0;
		while(1)
		{
			while(it2==tree_traversion_stack.back().first->children.end())
			{
				it2=tree_traversion_stack.back().second;
				tree_traversion_stack.pop_back();
				if(0==tree_traversion_stack.size())
					{ nextit=true; break; }
				++it2;
				--depth;
			}
			if(nextit)
				break;
			if(0!=it2->op)
			{
				if(it2->op->uid<bi::MAX_OPCODE_NUMBER)
					transform_opcode_nbit[depth].push_back(it2->op->opcode);
				else
				{
					CommandListType::iterator target=code.find(Command(Command::NewUID(it2->op->uid)));
					if(!target.is_valid_and_not_end(code))
						report_error("Internal Compiler Error: Call of makro with undefined uid.");
					transform_opcode_nbit[depth].push_back(target->pc);
				}
				++it2;
			}
			else
			{
				tree_traversion_stack.push_back(sPair(&*it2, it2));
				it2=it2->children.begin();
				++depth;
			}
		}
	}

	//fprintf(file, "-----------------------------------\r\n");

	/*for(RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it=opcodes.begin(); it!=opcodes.end(); it++)
	{
		fprintf(file, "c=%u   uid=%u   code= ", it->opcode, it->uid);
		for(long i=0; i<it->nr_codenibbles; i++)
			fprintf(file, "%u, ", (it->code>>(4*(it->nr_codenibbles-1-i)))&15);
		fprintf(file, "\r\n");
	}
	if(file)
		fclose(file);*/

	for(long i=0; i<8; i++)
		cd.push_back_nibblelist(make_nibblelist(transform_opcode_nbit[i]));

	for(RotSchwarzBaum<FrequencySortedFunctioncall, FrequencySortedFunctioncall_Compare>::iterator it=freqsortedcalls.maximizer();
			it!=0; it--)
	{
		//printf("==> %d    %d\n", it->func->function_nr, it->func->nr_used);
		if(255!=it->func->function_nr)
		{
			CommandListType::iterator target=code.find(Command(Command::NewUID(it->func->function_uid)));
			if(!target.is_valid_and_not_end(code))
			{
				RotSchwarzBaum<FunctionUIDtoFunctionName, FunctionUIDtoFunctionName_Compare>::iterator funcit=
					cc->functionUID2name.suche(FunctionUIDtoFunctionName(it->func->function_uid));
				report_error(printf_string("Call of undefined function '%s'.",
					funcit!=0 ? funcit->name.c_str() : ""));
			}
			PROGRAM_COUNTER_TYPE tpc=target->pc;
			for(uint8 size=MAX_NR_BYTES_FOR_PROGRAM_COUNTER; size--; )
			{
				cd.push_back((tpc>>4)&15);
				cd.push_back(tpc&15);
				tpc>>=8;
			}
		}
	}

#ifdef USE_TABLE_FOR_CONSTANTS
	constant_nr=0;
	for(long i=0; i<4; i++)
	{
		//printf("cd.size()=%u\n", cd.nibblesize(0));
		int constant_count=0;
		for(RotSchwarzBaum<std::pair<Command, uint32>, PUSHC_Command_SavingSorted_LessEqual>::iterator
				it=saving_sorted_constants.maximizer(); it!=0 && constant_count<256; it--, constant_count++)
			if(PUSHC_Command_SavingSorted_LessEqual::get_nr_bytes(it->first)==(1<<i))
			{
				uint8 type;
				uint64 c=PUSHC_Command_SavingSorted_LessEqual::get_integer_constant(it->first, &type);
				uint8 nrbytes=1<<i;
				cd.push_back(type);
				for(long j=0; j<nrbytes; j++)
				{
					cd.push_back((c>>4)&15);
					cd.push_back(c&15);
					c>>=8;
				}
			}
	}
	//printf("cd.size()=%u\n", cd.nibblesize(0));

	{
		OpcodeOrMakro op; op.opcode=bi::PUSHCT; op.uid=0; op.nr_used=0;
		RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
		if(it2!=0)
			assumed_nibblesize_for_PUSHCT_Command=it2->nr_codenibbles;
		op.opcode=bi::PUSHC;
		it2=opcodes.suche(op);
		if(it2!=0)
			assumed_nibblesize_for_PUSHC_Command=it2->nr_codenibbles;
		op.opcode=bi::PUSHCS;
		it2=opcodes.suche(op);
		if(it2!=0)
			assumed_nibblesize_for_PUSHCS_Command=it2->nr_codenibbles;
	}
#endif


	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef USE_CODE_MAKRO_COMPRESSION
	for(nlist<makro_information>::iterator it=generated_makros.begin(); it!=generated_makros.end(); it++)
	{
		CommandListType::iterator begin=code.find(Command(Command::NewUID(it->start)));
		CommandListType::iterator end=code.find(Command(Command::NewUID(it->end)));
		OpcodeOrMakro op; op.opcode=0; op.uid=it->start; op.nr_used=0;
		RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator opit=opcodes.suche(op);
		if(begin==0 || end==0 || opit==0)
			report_error("Error in assemble_code()");
		uint32 size, saved;
		size=end->pc-begin->pc + 2 + 2*MAX_NR_BYTES_FOR_PROGRAM_COUNTER + it->nr_called*opit->nr_codenibbles;
		++size;
		if(0==it->orig_makro_sp)
			saved=it->nr_called*(end->pc2-begin->pc2);
		else
			saved=it->nr_called*(end->pc-begin->pc);
		if(size>saved)
		{
			for(nlist<Command::uid_type>::iterator it2=it->orig_start_uids.begin(); it2!=it->orig_start_uids.end(); it2++)
				if(forbidden_makros.suche(*it2)==0)
					forbidden_makros.insert(*it2);
			//printf("size=%d   saved=%d\n", size, saved);
		}
	}
#endif

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	//FILE *prg_file=fopen("d:\\prg2.txt", "wb");

	for(CommandListType::iterator it=code.begin(); it!=code.end(); it++)
	{
#ifdef PRINT_COMMANDS_AT_RUNTIME
		/*fprintf(prg_file, "UID=0x%x  pc=%.3x   Command=%s   %s%s", it->get_uid(), it->pc, code_element_to_string(it).c_str(),
			0xcccccccc!=it->branch1_id ? printf_string("br1=%x   ", it->branch1_id).c_str() : "",
			0xcccccccc!=it->branch2_id ? printf_string("br2=%x   ", it->branch2_id).c_str() : "");
		if(Command::function==it->fkt)
		{
			RotSchwarzBaum<FunctionUIDtoFunctionName, FunctionUIDtoFunctionName_Compare>::iterator it2=
				cc->functionUID2name.suche(FunctionUIDtoFunctionName(it->get_uid()));
			if(it2!=0)
				fprintf(file2, "   %s", it2->name.c_str());
		}
		else if(Command::functioncall==it->fkt)
		{
			RotSchwarzBaum<FunctionUIDtoFunctionName, FunctionUIDtoFunctionName_Compare>::iterator it2=
				cc->functionUID2name.suche(FunctionUIDtoFunctionName(it->branch1_id));
			if(it2!=0)
				fprintf(file2, "   %s", it2->name.c_str());
		}
		for(long i=(Command::function==it->fkt) ? 2 : 0; i<it->nibblesize; i++)
			fprintf(file2, "%.2x ", it->code[i]);
		fprintf(file2, "\r\n");*/
#endif
		if(cd.nibblesize(get_first_code_nibble_of_command(it, opcodes))!=it->pc)
		{
			//printf("==> %d    %d\n", cd.nibblesize(get_first_code_nibble_of_command(it, opcodes)), it->pc);
			report_error("Error in assemble_code()");
		}
		cd.enable_encoding(0!=it->makro_sp && !it->is_makro_param && !it->disable_escape_sequences);
		if(Command::command==it->fkt)
		{
			uint8 opcode=(it->code[0]<<4) | it->code[1];
			OpcodeOrMakro op; op.opcode=opcode; op.uid=0; op.nr_used=0;
			RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
			if(it2==0 || 0==it2->nr_codenibbles)
				report_error("Internal compiler error: Error in assemble_code()");
/*#ifdef PRINT_COMMANDS_AT_RUNTIME
			bi::print_command(opcode, it->pc, it==code.begin() ? 0 : (-((--(CommandListType::iterator(it)))->sp)));
			printf("Opcode=%u", opcode);
			printf("\n");
#endif*/

			for(long i=0; i<it2->nr_codenibbles; i++)
				cd.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15);
			switch(opcode)
			{
				case bi::BRT: case bi::BRF: case bi::BR:
				case bi::BRAND: case bi::BROR: // diese Flle knnen nur dann auftreten, wenn 
				case bi::BRANDNOT: case bi::BRORNOT: // WITH_BRAND_BROR_BRANDNOT_BRORNOT_OPTIMIZATION nicht definiert ist
				{
					nlist<uint8>::iterator it2=0;
					if(it->disable_escape_sequences && ++CommandListType::iterator(it)!=code.end())
					{
						cd.flush();
						cd.push_back(bi::makro_escape_nibble);
						cd.push_back(bi::makro_escape_fixedsize_param);
						cd.push_back(15);
						cd.push_back(0); // wird gleich korrigiert
						it2=--cd.cd.end();
					}
					CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
					if(!target.is_valid_and_not_end(code))
						report_error("Internal Compiler Error: Branch to undefined uid.");
					add_branch_distance(*it, target->pc - it->pc, !it->short_branch1, bi::BRT==opcode || bi::BRF==opcode || bi::BR==opcode);
					for(long i=2; i<it->nibblesize; i++)
						cd.push_back(it->code[i]);
					if(it2!=0)
						*it2=(++CommandListType::iterator(it))->pc - cd.nibblesize(0);
					break;
				}
				case bi::IFELSE: case bi::IFNOTELSE:
				{
					nlist<uint8>::iterator it2=0;
					if(it->disable_escape_sequences && ++CommandListType::iterator(it)!=code.end())
					{
						cd.flush();
						cd.push_back(bi::makro_escape_nibble);
						cd.push_back(bi::makro_escape_fixedsize_param);
						cd.push_back(15);
						cd.push_back(0); // wird gleich korrigiert
						it2=--cd.cd.end();
					}
					CommandListType::iterator target1=code.find(Command(Command::NewUID(it->branch1_id)));
					CommandListType::iterator target2=code.find(Command(Command::NewUID(it->branch2_id)));
					if(!target1.is_valid_and_not_end(code) || (it->branch2_id!=0 && !target2.is_valid_and_not_end(code)))
						report_error("Internal Compiler Error: Branch to undefined uid.");
					long d1=target1->pc - it->pc;
					if(d1<0)
						report_error("Interner Compiler-Fehler: Sprungweite fr den else-Teil eines IFELSE- oder IFNOTELSE-Befehls ist negativ.");
					add_branch_distance(*it, d1, !it->short_branch1, false); // unsigned
					if(it->branch2_id!=0)
					{
						long d2=target2->pc - it->pc;
						add_branch_distance(*it, (d2-d1>=0) ? (d2-d1) : d2, !it->short_branch2, true); // signed
					}
					for(long i=2; i<it->nibblesize; i++)
						cd.push_back(it->code[i]);
					if(it2!=0)
						*it2=(++CommandListType::iterator(it))->pc - cd.nibblesize(0);
					break;
				}
				default:
				{
					for(long i=2; i<it->nibblesize; i++)
						cd.push_back(it->code[i]);
					break;
				}
			}
		}
		else
		{
			switch(it->fkt)
			{
				case Command::functioncall:
				{
					OpcodeOrMakro op; op.opcode=bi::CALL; op.uid=0; op.nr_used=0;
					RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
					if(it2==0 || 0==it2->nr_codenibbles)
						report_error("Internal compiler error: Error in assemble_code()");

					CommandListType::iterator target=code.find(Command(Command::NewUID(it->branch1_id)));
					if(!target.is_valid_and_not_end(code) || !(++target).is_valid_and_not_end(code)) // ++target: Command::function berspringen
					{
						RotSchwarzBaum<FunctionUIDtoFunctionName, FunctionUIDtoFunctionName_Compare>::iterator funcit=
							cc->functionUID2name.suche(FunctionUIDtoFunctionName(it->branch1_id));
						report_error(printf_string("Call of undefined function '%s'.",
							funcit!=0 ? funcit->name.c_str() : ""));
					}
					while(target.is_valid_and_not_end(code) && (Command::label==target->fkt || Command::uid_label==target->fkt))
						++target;
					if(!target.is_valid_and_not_end(code))
					{
						RotSchwarzBaum<FunctionUIDtoFunctionName, FunctionUIDtoFunctionName_Compare>::iterator funcit=
							cc->functionUID2name.suche(FunctionUIDtoFunctionName(it->branch1_id));
						report_error(printf_string("Call of undefined function '%s'.",
							funcit!=0 ? funcit->name.c_str() : ""));
					}
					it->add(bi::CALL);
					uint32 tpc=target->pc;

					CountFunctionCalls cnt;
					cnt.function_header_uid=it->branch2_id; // function_header_uid ist der Sortierschlssel
					RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare>::iterator it3=count_functioncalls.suche(cnt);
					if(it3==0)
						report_error("Internal compiler error: Error in assemble_code()");
					if(255!=it3->function_nr)
					{
						it->addnibble((it3->function_nr>>4)&15);
						it->addnibble(it3->function_nr&15);
					}
					else
					{
						if(0!=(tpc&1))
							report_error("Internal compiler error: Error in assemble_code()");
						tpc=(tpc>>1) | (1LU<<((8*MAX_NR_BYTES_FOR_PROGRAM_COUNTER)-1));
						for(uint8 size=(MAX_NR_BYTES_FOR_PROGRAM_COUNTER<<1); size--; )
						{
							it->addnibble((tpc>>(MAX_NR_BYTES_FOR_PROGRAM_COUNTER*8-4)) & 15);
							tpc<<=4;
						}
					}
					it->fkt=Command::command;
					for(long i=0; i<it2->nr_codenibbles; i++)
						cd.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15);
					for(long i=2; i<it->nibblesize; i++)
						cd.push_back(it->code[i]);
					break;
				}
				case Command::function:
					break;
				case Command::begin_infos_before_function:
				{
					CountFunctionCalls cnt;
					cnt.function_header_uid=it->get_uid(); // function_header_uid ist der Sortierschlssel
					RotSchwarzBaum<CountFunctionCalls, CountFunctionCalls_Compare>::iterator it2=count_functioncalls.suche(cnt);
					while(cd.nibblesize(0)<(++CommandListType::iterator(it))->pc)
					{		// Makros kommen erst nach den Funktionen; daher kann cd.nibblesize() mit dem Parameter 0 aufgerufen werden.
						for(long i=0; i<16; i++)
						{
							if(i!=bi::makro_escape_nibble && i!=bi::makro_escape_ret && i!=bi::makro_escape_fixedsize_param &&
									i!=bi::makro_escape_varsize_param)
							{
								cd.push_back(i); // Fllnibble zum Alignment des Anfangs der Funktion
								break;
							}
						}
					}
					break;
				}
				case Command::raw_nibbles:
					for(long i=0; i<it->nibblesize; i++)
						cd.push_back(it->code[i]);
					break;
				case Command::makrocall:
				{
					OpcodeOrMakro op; op.opcode=0; op.uid=it->branch1_id; op.nr_used=0;
					RotSchwarzBaum<OpcodeOrMakro, OpcodeOrMakro_Compare>::iterator it2=opcodes.suche(op);
					if(it2==0 || 0==it2->nr_codenibbles)
						report_error("Internal compiler error: Error in assemble_code()");
					for(long i=0; i<it2->nr_codenibbles; i++)
						cd.push_back((it2->code>>(4*(it2->nr_codenibbles-1-i)))&15);
					break;
				}
				case Command::makro_ret:
					cd.push_back_makro_ret();
					break;
				case Command::makro_param_call_fixedsize:
					cd.push_back_makro_param_call_fixedsize(it->var_data);
					break;
				case Command::makro_param_call_varsize:
					cd.push_back_makro_param_call_varsize();
					break;
			}
		}
	}

	//if(prg_file)
	//	fclose(prg_file);

	for(nlist<uint8>::iterator it=cd.cd.begin(); it!=cd.cd.end(); it++)
		if(*it>=16)
			report_error("Internal compiler error: Nibble cannot be >=16.");

#ifdef _DEBUG
	if(nr_inserted_branch0_for_BRTF || nr_inserted_castbool_for_BRANDOR)
	{
		report_warning(printf_string("\n********* WARNUNG: **********\nnr_inserted_branch0_for_BRTF=%u\nnr_inserted_castbool_for_BRANDOR=%u\n\n",
			nr_inserted_branch0_for_BRTF, nr_inserted_castbool_for_BRANDOR));
		while(_getch()==0);
	}
#endif
	return cd.cd;
}

nlist<uint8> assemble_code(const Code& orig_code)
{
	assumed_nibblesize_for_PUSHCT_Command=2; // wird ggf. von assemble_code_makro_trial() gendert 
	assumed_nibblesize_for_PUSHC_Command=2;
	assumed_nibblesize_for_PUSHCS_Command=2;
#ifdef USE_CODE_MAKRO_COMPRESSION
	uint32 prev_nr_forbidden=0;
	enum { anz_trials=10 };
	RotSchwarzBaum<Command::uid_type, UID_LessEqual> forbidden_makros, save_forbidden_makros[anz_trials];
	uint32 sizes[anz_trials];
	nlist<uint8> code;
	for(long i=0; i<anz_trials; i++)
	{
		save_forbidden_makros[i]=forbidden_makros;
		code=assemble_code_makro_trial(orig_code, forbidden_makros);
		//printf("Anz forbidden = %d   size=%d\n", forbidden_makros.size(), code.size());
		prev_nr_forbidden=forbidden_makros.size();
		sizes[i]=code.size();
	}
	uint32 min_size=0xffffffff, min_index=0;
	for(long i=0; i<anz_trials; i++)
		if(sizes[i]<=min_size)
		{
			min_size=sizes[i];
			min_index=i;
		}
	return assemble_code_makro_trial(orig_code, save_forbidden_makros[min_index]);
#else
	RotSchwarzBaum<Command::uid_type, UID_LessEqual> forbidden_makros;
	nlist<uint8> code=assemble_code_makro_trial(orig_code, forbidden_makros);
  #ifdef USE_TABLE_FOR_CONSTANTS
	code=assemble_code_makro_trial(orig_code, forbidden_makros);
		// mit aktualisierten Werten von assumed_nibblesize_for_PUSHCT_Command, assumed_nibblesize_for_PUSHC_Command und
		// assumed_nibblesize_for_PUSHCS_Command neu berechnen.
  #endif
	return code;
#endif
}

Compiler::~Compiler()
{
	if(program_code) delete program_code;
	if(RegisterType) delete RegisterType;
	if(StringConstants) delete StringConstants;
	if(bytecode) delete[] bytecode;
}

Compiler::Compiler(void (*_CFunctionForOneCompilerPass)(Compiler* compiler))
{
	cc=this; // Muss hier durchgefhrt werden, obwohl dies _CFunctionForOneCompilerPass() wiederholt, denn
		// InitializeGlobalScope() greift auf cc zu.
	ActualStructDefinition=0;
	program_code=new Code;
	RegisterType=new RegisterTypeClass;
	StringConstants=new StringConstantsContainer;
	for(CompilerPassNr=0; CompilerPassNr<2; CompilerPassNr++)
	{
		InitializeGlobalScope();
		ProgramStructureStack.push_back(new ProgramStructureGlobalScope);
		_CFunctionForOneCompilerPass(this);
		EndProgramStructureElement();
	}
	nlist<uint8> codelist=assemble_code(*program_code);
	if(codelist.size()==0)
		report_error("Internal compiler error");
	bytecode=new uint8[(codelist.size()+1)/2];
	uint32 codesize=0;
	for(nlist<uint8>::iterator it=codelist.begin(); it!=codelist.end(); it++, codesize++)
	{
		bytecode[codesize]=(*it)<<4;
		if(++it==codelist.end())
			break;
		bytecode[codesize] |= (*it);
	}
	nibble_codesize=codelist.size();
	required_memory_size=Variable::var_mem_count[2*(Value::anz_user_types+1)];
}

#undef Function_ADDPARAM
#undef Function_ADDPARAM2
#undef Function_ADDPARAM4
#undef Function_ADDPARAM6
#undef Function_ADDPARAM8
#undef Function_ADDPARAM10
#undef Function_ADDPARAM12
#undef Function_ADDPARAM14
#undef Function_ADDPARAM16
#undef Function_ADDPARAM18
#undef Function_ADDPARAM20
#undef Function_ADDPARAM22
#undef Function_ADDPARAM24
#undef Function_ADDPARAM26
#undef Function_ADDPARAM28
#undef Function_ADDPARAM30
#undef Function_ADDPARAM32
#undef Function_ADDPARAM34
#undef Function_ADDPARAM36
#undef Function_ADDPARAM38
#undef Function_ADDPARAM40
#undef Function_ADDPARAM42
#undef Function_ADDPARAM44
#undef Function_ADDPARAM46
#undef Function_ADDPARAM48
#undef Function_ADDPARAM50
#undef Function_ADDPARAM52
#undef Function_ADDPARAM54
#undef Function_ADDPARAM56
#undef Function_ADDPARAM58
#undef Function_ADDPARAM60
#undef Function_ADDPARAM62
#undef Function_ADDPARAM64
#undef Function_ADDPARAM_even_nr_args
#undef Function_ADDPARAM_odd_nr_args

#undef Function_FuncOPHeader
#undef Function_FuncOPHeader2
#undef Function_FuncOPHeader4
#undef Function_FuncOPHeader6
#undef Function_FuncOPHeader8
#undef Function_FuncOPHeader10
#undef Function_FuncOPHeader12
#undef Function_FuncOPHeader14
#undef Function_FuncOPHeader16
#undef Function_FuncOPHeader18
#undef Function_FuncOPHeader20
#undef Function_FuncOPHeader22
#undef Function_FuncOPHeader24
#undef Function_FuncOPHeader26
#undef Function_FuncOPHeader28
#undef Function_FuncOPHeader30
#undef Function_FuncOPHeader32
#undef Function_FuncOPHeader34
#undef Function_FuncOPHeader36
#undef Function_FuncOPHeader38
#undef Function_FuncOPHeader40
#undef Function_FuncOPHeader42
#undef Function_FuncOPHeader44
#undef Function_FuncOPHeader46
#undef Function_FuncOPHeader48
#undef Function_FuncOPHeader50
#undef Function_FuncOPHeader52
#undef Function_FuncOPHeader54
#undef Function_FuncOPHeader56
#undef Function_FuncOPHeader58
#undef Function_FuncOPHeader60
#undef Function_FuncOPHeader62
#undef Function_FuncOPHeader64

#undef Function_FuncOP_even_nr_args
#undef Function_FuncOP_odd_nr_args
#undef Function_4FuncOPs
#undef VoidFunction_FuncOP_even_nr_args
#undef VoidFunction_FuncOP_odd_nr_args
#undef VoidFunction_4FuncOPs

#undef PARAMS_ADDPARAM
#undef PARAMS_ADDPARAM2
#undef PARAMS_ADDPARAM4
#undef PARAMS_ADDPARAM6
#undef PARAMS_ADDPARAM8
#undef PARAMS_ADDPARAM10
#undef PARAMS_ADDPARAM12
#undef PARAMS_ADDPARAM14
#undef PARAMS_ADDPARAM16
#undef PARAMS_ADDPARAM18
#undef PARAMS_ADDPARAM20
#undef PARAMS_ADDPARAM22
#undef PARAMS_ADDPARAM24
#undef PARAMS_ADDPARAM26
#undef PARAMS_ADDPARAM28
#undef PARAMS_ADDPARAM30
#undef PARAMS_ADDPARAM32
#undef PARAMS_ADDPARAM34
#undef PARAMS_ADDPARAM36
#undef PARAMS_ADDPARAM38
#undef PARAMS_ADDPARAM40
#undef PARAMS_ADDPARAM42
#undef PARAMS_ADDPARAM44
#undef PARAMS_ADDPARAM46
#undef PARAMS_ADDPARAM48
#undef PARAMS_ADDPARAM50
#undef PARAMS_ADDPARAM52
#undef PARAMS_ADDPARAM54
#undef PARAMS_ADDPARAM56
#undef PARAMS_ADDPARAM58
#undef PARAMS_ADDPARAM60
#undef PARAMS_ADDPARAM62
#undef PARAMS_ADDPARAM64
#undef PARAMS_ADDPARAM_even_nr_args
#undef PARAMS_ADDPARAM_odd_nr_args

#undef PARAMS_PUSHPARAM
#undef PARAMS_PUSHPARAM2
#undef PARAMS_PUSHPARAM4
#undef PARAMS_PUSHPARAM6
#undef PARAMS_PUSHPARAM8
#undef PARAMS_PUSHPARAM10
#undef PARAMS_PUSHPARAM12
#undef PARAMS_PUSHPARAM14
#undef PARAMS_PUSHPARAM16
#undef PARAMS_PUSHPARAM18
#undef PARAMS_PUSHPARAM20
#undef PARAMS_PUSHPARAM22
#undef PARAMS_PUSHPARAM24
#undef PARAMS_PUSHPARAM26
#undef PARAMS_PUSHPARAM28
#undef PARAMS_PUSHPARAM30
#undef PARAMS_PUSHPARAM32
#undef PARAMS_PUSHPARAM34
#undef PARAMS_PUSHPARAM36
#undef PARAMS_PUSHPARAM38
#undef PARAMS_PUSHPARAM40
#undef PARAMS_PUSHPARAM42
#undef PARAMS_PUSHPARAM44
#undef PARAMS_PUSHPARAM46
#undef PARAMS_PUSHPARAM48
#undef PARAMS_PUSHPARAM50
#undef PARAMS_PUSHPARAM52
#undef PARAMS_PUSHPARAM54
#undef PARAMS_PUSHPARAM56
#undef PARAMS_PUSHPARAM58
#undef PARAMS_PUSHPARAM60
#undef PARAMS_PUSHPARAM62
#undef PARAMS_PUSHPARAM64
#undef PARAMS_PUSHPARAM_even_nr_args
#undef PARAMS_PUSHPARAM_odd_nr_args

#undef PARAMS_DEFINE_4_CALLING_SCHEMAS

} // Ende von namespace BytecodeTargetedCompiler

#ifdef MSCPP
#pragma warning(pop) 
#endif
